package MyANEX_IO2;
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.5: Copyright (C) 2019 Kiyoshi Ezawa
#	Version 0.6: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyANEX_IO2.pm:
#	This Perl package consists of subroutines for data inputs, (test) outputs,
#	and some other processings, 
#	that are typically used by some scripts of 'ANEX_P' BUT not found in "MyANEX_IO.pm".
#	[CREATION STARTED on 2020/04/13, ENDED on 2020/04/dd.]
#

use strict;
use warnings;

use MyFile_manipu qw(
	open_input_file2
	open_output_file
	); # ADDED 'open_output_file' in creation of 'detect_purge_cands.ver0.5.pl' (on 2019/06/17). #
	# list_sessions_ext
	# tfm_labels2indices


use MyANEX_Main qw(
	initialize_multidim_storage
	);
	# nonrec_mlt_shifts_gapblocks
	# printout_multidim_storage


=Set aside for later use.
use MyANEX_IO qw(
	print_out_coord_frame 
	print_out_nr_mshifts_gblks 
	);
=cut # END of "Set aside for later use."



use Exporter;

our @ISA =qw(Exporter);

our @EXPORT_OK = qw( 

	readin_coord_frame
	readin_sgl_qnt_on_multidim_space
	find_maxvalue_in_multidim_storage
	cmpt_log_total_qnt_in_multidim_storage
	);

sub readin_coord_frame (\$); # Added on 2020/04/14.
sub readin_sgl_qnt_on_multidim_space (\*\@\@\@$$$); # Added on 2020/04/14. 

##sub find_maxvalue_in_multidim_storage (\@\@\@$$); # Added on 2020/04/14.
sub find_maxvalue_in_multidim_storage (\@\@\@$$$); # => REVISED on 2020/04/18.

sub cmpt_log_total_qnt_in_multidim_storage (\@\@\@$$$$); # Added on 2020/04/15.


	# Global Variables that specify the level of tests on the subroutines.
	# (0: Nothing, 1: brief, 2: a bit detailed, 3: quite detailed, 5: very detailed)

	# For 'find_maxvalue_in_multidim_storage'. #

our $TEST_LEVEL_fnd_mxv_mdms = 0;
##our $TEST_LEVEL_fnd_mxv_mdms = 1;
##our $TEST_LEVEL_fnd_mxv_mdms = 5;


	# For 'cmpt_log_total_qnt_in_multidim_storage'. #

our $TEST_LEVEL_cmpt_ltot_mdms = 0;
##our $TEST_LEVEL_cmpt_ltot_mdms = 1;
##our $TEST_LEVEL_cmpt_ltot_mdms = 5;


	##### (I) DEFINITIONS of the Subroutines For Inputs. #####

# readin_coord_frame:
#	This subroutine reads in a file
#	output by the subroutine, 'print_out_coord_frame', in "MyANEX_IO.pm".
#	[CREATION STARTED on 2020/04/13, ENDED on 2020/04/14.]
#
# INPUTS: (\$infile), where
#
#	$infile is the path to an input file (recording a coordinate frame).
#
# OUTPUTS: (\@bds_gb_coords_raw, \@org_gb_coords_offset, \@bds_gb_coords_offset), where
#
#	@bds_gb_coords_raw stores the RAW (i.e., UN-offset) coordinates of the boundaries of the gap-blocks.
#	More specifically,
#	($lbd_coord, $rbd_coord) = @{$bds_bl_coords_raw[$bl]}
#	are the left- and right-bounds of the RAW coordinate of the $bl th block.
#	[In the UN-offset format, the origin of each block always has the coorinate 0 (zero).]
#
#	@org_gb_coords_offset stores the coordinate-origins of the gap-blocks in the OFFSET format;
#	@bds_gb_coords_offset stores the coordinates of the boundaries of the gap-blocks in the OFFSET format.
#	[For more specific descriptions on the OFFSET coordinates,
#	see, e.g., the header of 'mk_orgs_bds_gb_coords' in "MyANEX_Supple.pm".]
#
sub readin_coord_frame (\$) {

	my ($rf_infile) = @_;

	my $infile = ${$rf_infile};
	unless (( -f $infile ) and ( -r $infile )) {
	   print STDOUT "[ERROR 'readin_coord_frame'] CanNOT FIND a readable input file '$infile': $! !\n";
	   return (undef, undef, undef);
	}

	local $/ = "\n"; # Use the newline as the input-record-separator.

	my $fh_in = open_input_file2 ($infile);
	unless (defined $fh_in) {
	    print STDOUT "[ERROR 'readin_coord_frame'] 'open_input_file2' FAILED on the input file '$infile': $! !\n";
	    return (undef, undef);
	}

	while (my $line = <$fh_in>) { # Remove decorative lines on the top. #
	    if (uc ($line) =~ m/^COORDINATE/) { last; }
	} # End of "Remove decorative lines on the top."

			# Retrieve the RAW coordinates of the boundaries. #

	my @bds_gb_coords_raw = ();

	while (my $line = <$fh_in>) { # Main while-loop (over input lines). #

	    my ($gb) = ($line =~ m/^X(\d+)\t/);
	    unless (defined $gb) { next; }

	    $line =~ s/\s+$//;

	    my ($rlabel, $lbd_coord, $rbd_coord) = split /\t/, $line;

	    $bds_gb_coords_raw[$gb] = [$lbd_coord, $rbd_coord];

	} # End of the Main while-loop (over input lines). #

	close ($fh_in);


			# Make OFFSET coordinates out of the RAW coordinates. #

	my @orgs_gb_coords_offset = my @bds_gb_coords_offset = ();

	my $ct_gbs = @bds_gb_coords_raw;
	for (my $gb=0; $gb < $ct_gbs; $gb++) { # Main for-loop (over gap-blocks). #

	    my ($lbd_coord_raw, $rbd_coord_raw) = @{$bds_gb_coords_raw[$gb]};

	    my $org = ($lbd_coord_raw < 0) ? - $lbd_coord_raw : 0 ; # What's Important is to always keep the offset coordinates NON-negative.

	    my $lbd_coord_offset = $lbd_coord_raw + $org;
	    my $rbd_coord_offset = $rbd_coord_raw + $org;

	    $orgs_gb_coords_offset[$gb] = $org;
	    $bds_gb_coords_offset[$gb] = [$lbd_coord_offset, $rbd_coord_offset];

	} # End of the Main for-loop (over gap-blocks). #

	return (\@bds_gb_coords_raw, \@orgs_gb_coords_offset, \@bds_gb_coords_offset);

}	# End of the subroutine, 'readin_coord_frame'.


# readin_sgl_qnt_on_multidim_space:
#	This subroutine reads in the values of a single quantity
#	defined on a multi-dimentional space (actually, a hyperrectangle),
#	which are output by 'printout_multidim_storage' in "MyANEX_Main.pm",
#	and stores the retrieved values into a multi-dimensional storage,
#	as an element specified by the input index.
#	[CREATION STARTED on 2020/04/13, ENDED on 2020/04/14.]
#
#	[NOTE (on 2020/04/14)]:
#	The subroutine ends if it encounters a header line (other than the 1st one if ($if_skip_1st_header ==1)),
#	or if it runs out of input lines.
#
# INPUTS: (\*${$fh}, \@orgs_gb_coords_offset, \@bds_gb_coords_offset, \@multidim_storage, $indx_sbj, $string_undef, $ifskip_1st_header), where
#
#	$fh is the reference to the file-handle for the file that the values of the subject quantity in the multi-dimensional space (i.e., hyperrectangle) are recorded;
#
#	$org_gb_coords_offset[$gb] is the position of the "origin" of the coordinate of the $gb th gap-block (in the OFFSET format).
#
#	($lbd_coord, $rbd_coord) = @{$bds_gb_coords_offset[$gb]}
#	are the left- and right-bounds of the coordinate of the $gb th gap-block (in the OFFSET format).
#	[ 0 <= $lbd_coord <= $rbd_coord MUST ALWAYS hold.]
#
#	@storage_for_pnt = @{$mltdim_storage[$k_0]->...->[$k_{B-1}]} (where B = scalar (@bds_bl_coords))
#	is the storage for some information on the point with the coordinate,
#	($k_0, ..., $k_{B-1}) within the hyperrectangle in question.
#	The retrieved value will be stores as the $indx_sbj th element of @storage_for_pnt.
#	[NOTE: @multidim_storage may be EMPTY; in that case, it MUST be initialized.]
#
#	$string_undef is the string designated for indicating that the component is UNDEFINED.
#	(Commonly used ones are: 'undef', 'n/a', and 'N/A'.)
#
#	$if_skip_1st_header	= 1 if you want to skip the 1st header (beginning w/ 'X0' or '\X0' (when #{gap-blocks} == 1)), as well as ALL lines above it(, if any),
#				= 0 otherwise.
#
# OUTPUTS: $fl_end, where
#
#	$fl_end	= 1 if the subroutine ended without any apparent problems,
#		= 0 if the subroutine failed.
#
sub readin_sgl_qnt_on_multidim_space (\*\@\@\@$$$) {

	my ($fh, $orgs_gb_coords_offset, $bds_gb_coords_offset, $multidim_storage, $indx_sbj, $string_undef, $if_skip_1st_header) = @_;

	my $ct_gbs = @{$orgs_gb_coords_offset};
	unless ($ct_gbs>0) {
	    print STDOUT "[INFORMATION 'readin_sgl_qnt_on_multidim_space'] The input space is ZERO DIMENSIONAL!! ... I will END the subroutine without doing anything ...\n\n";
	    return 1;
	}

	unless (@{$multidim_storage}>0) {

		# The multi-dimensional storage is EMPTY. => Initialize it. #

	    my $mdim_storage0 = initialize_multidim_storage (@{$bds_gb_coords_offset});
	    unless (defined $mdim_storage0) {
		print STDOUT "[ERROR 'readin_sgl_qnt_on_multidim_space'] The subroutine, 'initialize_multidim_storage', FAILED: $! !\n\n";
		return 0;
	    }
	    @{$multidim_storage} = @{$mdim_storage0};
	}

	my ($lbd_last_coord, $rbd_last_coord) = @{$bds_gb_coords_offset->[$ct_gbs-1]};
	my $ct_last_coords = $rbd_last_coord - $lbd_last_coord +1;

	local $/ = "\n"; # Use a newline character as the input-record-separator. #

	my $header_beg = ($ct_gbs>1) ? 'X0' : '\X0';
	my $len_hb = length ($header_beg);

	if ($if_skip_1st_header) {
		# Skip the 1st header and ALL the lines above it(, if any). #

	    while (my $line = <$fh>) {
		my $beg_line = substr ($line, 0, $len_hb);
		if ($beg_line eq $header_beg) { last; }
	    }
	}

	while (my $line = <$fh>) { # The Main Outer while-loop (over input lines). #

	    my $beg_line = substr ($line, 0, $len_hb);
	    if ($beg_line eq $header_beg) { last; }

			# Skip lines NOT recording the values of the subject quantity. #
	    if ($ct_gbs>1) {
		unless ($line =~ m/^(\-)?\d/) { next; }

	    } else { # ($ct_gbs ==1) #
		unless ($line =~ m/^\t/) { next; }
	    }


			# Properly Parse the Input Line. #

	    $line =~ s/\s+$//;

	    my @fields = split /\t/, $line;

	    unless (@fields == $ct_last_coords+1) {
		my $ct_flds = @fields;
		my $ct_flds0 = $ct_last_coords+1;
		print STDOUT "[WARNING 'readin_sgl_qnt_on_multidim_space'] Encountered a STORANGE RECORD LINE (w/ #{fields} = $ct_flds, NOT $ct_flds0 as expected!!):\n";
		print STDOUT "'$line': $! !\n\n";
		next;
	    }

	    my $label = shift @fields;

	    my @curr_pa_coords_raw = split /:/, $label;


			# Prepare an Appropriate "Parent" Storage. #

	    my $parent_storage = $multidim_storage;
	    for (my $i=0; $i <$ct_gbs-1; $i++) {
		my $coord_offset = $curr_pa_coords_raw[$i] + $orgs_gb_coords_offset->[$i];
		$parent_storage = $parent_storage->[$coord_offset];
	    }

			# Store the values in the current line into appropriate point-storages. #

	    for (my $f = 0; $f < $ct_last_coords; $f++) {
		my $val_sbj = $fields[$f];
		if ($val_sbj eq $string_undef) { next; } # Skip if the value is UNDEFINED. #
		my $last_coord = $lbd_last_coord + $f;
		$parent_storage->[$last_coord][$indx_sbj] = $val_sbj;
	    }

	} # End of the Main Outer while-loop (over input lines). #

	return 1;

}	# End of the subroutine, 'readin_sgl_qnt_on_multidim_space'. #



	##### (II) DEFINITIONS of the Subroutines For (test) Outputs. #####


	##### (III) DEFINITIONS of the Subroutines For Other Processes. #####

# find_maxvalue_in_multidim_storage:
#	This subroutine searches through a multi-dimensional storage,
#	and finds the maximum value of a specified element.
#	It was created by taking advantage of the framework of 'printout_multidim_storage' in "MyANEX_Main.pm".	
#	[CREATION STARTED on 2020/04/13, ENDED on 2020/04/15.]
#
#	[REVISED on 2020/04/18]:
#	Now, the subroutine also examines the degree of degeneracy,
#	and SKIPs those points with {degree of degeneracy} is undef, $string_undef or 0.
#
#	[NOTE (on 2020/04/14)]:
#	This subroutine does NOT identify the points where the specified element is maximum.
#
# [ INPUTS: (\@org_bl_coords, \@bds_bl_coords, \@multidim_storage, $indx_sbj, $string_undef), where ] ... OBSOLETE as of 2020/04/18.
# INPUTS: (\@org_bl_coords, \@bds_bl_coords, \@multidim_storage, $indx_sbj, $indx_dgn, $string_undef), where
#
#	$org_bl_coords[$bl] is the position of the "origin" of the coordinate of the $bl th block.
#	(Although the "origin" should actually be ZERO, it is SHIFTED to this value, SO THAT the range of the coordinate will become entirely non-negaitve. (That is, The coordinates follow a OFFSET format.);
#
#	($lbd_coord, $rbd_coord) = @{$bds_bl_coords[$bl]}
#	are the left- and right-bounds of the coordinate of the $bl th block.
#	[ASSUME that 0 <= $lbd_coord <= $rbd_coord holds;
#	In the present OFFSET format, this should always be satisfied.]
#
#	@storage_for_pnt = @{$mltdim_storage[$k_0]->...->[$k_{B-1}]} (where B = scalar (@bds_bl_coords))
#	is the storage for some information on the point with the coordinate,
#	($k_0, ..., $k_{B-1}) within the hyperrectangle in question.
#
#	The quantity in question is stored as the $indx_sbj th element of @storage_for_pnt.
#
#			# (Added in REVISION on 2020/04/18.) #
#	The degree of degeneracy is stored as the $indx_dgn th element of @storage_for_pnt.
#	[NOTE: If $indx_dgn = undef, the subroutine ASSUMES that the degree of degeneracy is 1 for ALL lattice points in the hyperrectangle.]
#			# End of "(Added in REVISION on 2020/04/18.)" #
#
#	$string_undef is the string designated for indicating that the component is UNDEFINED.
#	(Commonly used ones are: 'undef', 'n/a', and 'N/A'.)
#
# OUTPUTS: ($fl_end, $maxvalue), where
#
#	$fl_end	= 1 if the subroutine ended without any apparent problems,
#		= 0 if it failed.
#
#	$maxvalue is the maximum value of the specified element.
#
# [ sub find_maxvalue_in_multidim_storage (\@\@\@$$) {
#	my ($org_bl_coords, $bds_bl_coords, $mltdim_storage, $indx_sbj, $string_undef) = @_; ] ... OBSOLETE as of 2020/04/18. 
#
sub find_maxvalue_in_multidim_storage (\@\@\@$$$) {

	my ($org_bl_coords, $bds_bl_coords, $mltdim_storage, $indx_sbj, $indx_dgn, $string_undef) = @_;

	my $ct_blks = @{$bds_bl_coords}; 
	unless (@{$org_bl_coords} == $ct_blks) {
	    my $ct_orgs = @{$org_bl_coords};
	    print STDOUT "[ERROR 'find_maxvalue_in_multidim_storage'] The number of 'origin's (=$ct_orgs) CLASHES with the number of block-coordinate-boundary-pairs (=$ct_blks): $! !\n\n";
	    return (0, undef);
	}
    (print STDOUT "< Subroutine 'find_maxvalue_in_multidim_storage' (#{coordinates} = $ct_blks) >\n\n") if ($TEST_LEVEL_fnd_mxv_mdms >0);

	if ($ct_blks == 0) { # (ADDED on 2020/04/15.) #
	    print STDOUT "[INFORMATION 'find_maxvalue_in_multidim_storage'] The number of blocks (=$ct_blks) is ZERO!! ... I will end this subroutine without doing anything ...\n\n";
	    return (1, undef);
	}

		# Examine the coordinate consistencies. #

	my @errors1 = my @errors2 = ();
	for (my $k=0; $k < $ct_blks; $k++) { # Preliminary for-loop (over blocks).
	    my ($lb, $rb) = @{$bds_bl_coords->[$k]};

	    if ($lb < 0) { push @errors1, $k; }
		
	    if ($rb < $lb) { push @errors2, $k; }
	
	} # End of the preliminary for-loop (over blocks).

	if ((@errors1>0) or (@errors2>0)) {
	    print STDOUT "[ERROR 'find_maxvalue_in_multidim_storage'] The blocks (", join (', ', @errors1), ") has NEGATIVE left-boundaries, and the blocks (", join (', ', @errors2), ") has their right-bound ON THE LEFT OF their left-bounds: $! !\n\n";
	    return (0, undef);
	}

		# Initialize the current parent. #
	my @curr_parent = ();
	for (my $k=0; $k < $ct_blks-1; $k++) { $curr_parent[$k] = $bds_bl_coords->[$k][0]; }

		# & Initialize the stack (as a nested series of single anonymous arrays specified by the coordinates).

	my $sbj_storage0 = $mltdim_storage;
	my @stack = ($sbj_storage0);
	for (my $k=0; $k < $ct_blks-1; $k++) { # 2nd for-loop (over blocks).
	    my $coord = $curr_parent[$k];
	    my $child_storage = $sbj_storage0->[$coord];
	    push @stack, $child_storage;
	    $sbj_storage0 = $child_storage;
	} # End of the 2nd for-loop (over blocks).
	my $parent_storage0 = pop @stack;

	my ($lb_last, $rb_last) = @{$bds_bl_coords->[$ct_blks-1]};
	my $org_last = $org_bl_coords->[$ct_blks-1];


	my @labels_point = ();
	for (my $k=0; $k < $ct_blks; $k++) { push @labels_point, "X$k"; }
	my $label_point_vector = join (', ', @labels_point);
	$label_point_vector = "($label_point_vector)";

=For the predecessor, 'printout_multidim_storage'. (1)
		# Print out the Header. #

	my @labels_corner = ();
	for (my $k=0; $k < $ct_blks-1; $k++) { push @labels_corner, "X$k"; }
	my $labels_corner = join (':', @labels_corner);
	$labels_corner .= '\X';
	$labels_corner .= $ct_blks-1;
	my @last_coords = ();
	for (my $k=$lb_last; $k <= $rb_last; $k++) { push @last_coords, $k - $org_last; }
	print $fh join ("\t", $labels_corner, @last_coords), "\n\n";

=cut # End of "For the predecessor, 'printout_multidim_storage'. (1)"


		# Examine the subject components of the children of the starting parent. #

	my $maxvalue;
	for (my $k = $lb_last; $k <= $rb_last; $k++) {

			# Added in REVISION on 2020/04/18. (1) #
	    my $ddgn = (defined $indx_dgn) ? $parent_storage0->[$k][$indx_dgn] : 1 ;
	    unless (defined $ddgn) { next; }

	    ##unless ($ddgn eq $string_undef) { next; } ## A Serious BUG Fixed on 2020/04/19. ##
	    if ($ddgn eq $string_undef) { next; }

	    if ($ddgn == 0) { next; } # Skip those points with {degree of degeneracy} == 0.
			# End of "Added in REVISION on 2020/04/18. (1)" #

	    my $component_sbj = $parent_storage0->[$k][$indx_sbj];

			# Modified in REVISION on 2020/04/18. (1) #
	    ##unless (defined $component_sbj) { next; }
	    ##if ($component_sbj eq $string_undef) { next; }
	    unless ((defined $component_sbj) and ($component_sbj ne $string_undef)) { 
		my $out_component_sbj = (defined $component_sbj) ? $component_sbj : "'UNDEF'";
		print STDOUT "[WARNING 'find_maxvalue_in_multidim_storage'] The lattice point, $label_point_vector = (", join (', ', @curr_parent, $k), ") has an ANOMALOUS subject component, $out_component_sbj, while its degree of degeneracy is $ddgn !! ... I will skip this point ...\n\n";
		next;
	    }
			# End of "Modified in REVISION on 2020/04/18. (1)" #

	    if ((defined $maxvalue) and ($maxvalue >= $component_sbj)) { next; }
	    $maxvalue = $component_sbj;
	}

=For the predecessor, 'printout_multidim_storage'. (2)
		# A BUG FIXED on 2019/03/06. (c)
	my @curr_pa_coords0 = ();
	for (my $i=0; $i < $ct_blks -1; $i++) {
	    $curr_pa_coords0[$i] = $curr_parent[$i] - $org_bl_coords->[$i];
	}
	my $cnct_curr_pa_coords0 = join (':', @curr_pa_coords0);
	##my $cnct_curr_parent = join (':', @curr_parent);

	my @components0 = ();
	for (my $k = $lb_last; $k <= $rb_last; $k++) {
	    my $component_sbj = $parent_storage0->[$k][$indx_po];
	    unless (defined $component_sbj) { $component_sbj = $string_undef; }
	    push @components0, $component_sbj;
	}
		# A BUG FIXED on 2019/03/06. (d)
	print $fh join ("\t", $cnct_curr_pa_coords0, @components0), "\n";
	##print $fh join ("\t", $cnct_curr_parent, @components0), "\n";

=cut # End of "For the predecessor, 'printout_multidim_storage'. (2)"

	if ($ct_blks < 2) {
			# This is necessary for the subroutine to end normally even if ($ct_blks < 2). #
			# (ADDED on 2020/04/15.)
	    return (1, $maxvalue);
	}


		# Main: Examine the subject components from the entire storage. #

	my $indx_curr_crd = $ct_blks -2;

	while (1) { # Main while-loop (to construct the output).

	    my $curr_sbj_coord = $curr_parent[$indx_curr_crd];
	    my ($lbd, $rbd) = @{$bds_bl_coords->[$indx_curr_crd]};

	    if ($curr_sbj_coord == $rbd) {
		if ($indx_curr_crd == 0) { last; }

		$curr_parent[$indx_curr_crd] = $lbd;
		$indx_curr_crd--;
		pop @stack;
		next;
	    }

		# A BUG fixed (on 2019/02/22).
	    $curr_parent[$indx_curr_crd]++;
	    $curr_sbj_coord = $curr_parent[$indx_curr_crd];
	    ##$curr_sbj_coord = $curr_parent[$indx_curr_crd]++;

	    my $grand_parent_storage = $stack[$#stack];

	    my $parent_storage = $grand_parent_storage->[$curr_sbj_coord];

	    while ($indx_curr_crd < $ct_blks - 2) { # Inner while-loop.
		push @stack, $parent_storage;
		$indx_curr_crd++;
		$curr_sbj_coord = $curr_parent[$indx_curr_crd];
		$parent_storage = $parent_storage->[$curr_sbj_coord];

	    } # End of the inner while-loop.

			### RESTART FROM HERE ###

			# Examine the subject components at the current points. #

	    for (my $k = $lb_last; $k <= $rb_last; $k++) { # Main inner for-loop. #

			# Added in REVISION on 2020/04/18. (2) #
	        my $ddgn = (defined $indx_dgn) ? $parent_storage->[$k][$indx_dgn] : 1 ;
	        unless (defined $ddgn) { next; }
	        if ($ddgn eq $string_undef) { next; }
	        if ($ddgn == 0) { next; } # Skip those points with {degree of degeneracy} == 0.
			# End of "Added in REVISION on 2020/04/18. (2)" #

		my $component_sbj = $parent_storage->[$k][$indx_sbj];

			# Modified in REVISION on 2020/04/18. (2) #
		##unless (defined $component_sbj) { next; }
		##if ($component_sbj eq $string_undef) { next; }
	        if (!(defined $component_sbj) or ($component_sbj eq $string_undef)) { 
		    my $out_component_sbj = (defined $component_sbj) ? $component_sbj : "'UNDEF'";
		    print STDOUT "[WARNING 'find_maxvalue_in_multidim_storage'] The lattice point, $label_point_vector = (", join (', ', @curr_parent, $k), ") has an ANOMALOUS subject component, $out_component_sbj, while its degree of degeneracy is $ddgn !! ... I will skip this point ...\n\n";
		    next;
	        }
			# End of "Modified in REVISION on 2020/04/18. (2)" #

		if ((defined $maxvalue) and ($maxvalue >= $component_sbj)) { next; }
		$maxvalue = $component_sbj;

	    } # End of the main inner for-loop. #

=For the predecessor, 'printout_multidim_storage'. (3)
		# Print out the components at the current points. #

		# A BUG FIXED on 2019/03/06. (a)
	    my @curr_pa_coords = ();
	    for (my $i = 0; $i < $ct_blks -1; $i++) {
		$curr_pa_coords[$i] = $curr_parent[$i] - $org_bl_coords->[$i];
	    }
	    my $cnct_curr_pa_coords = join (':', @curr_pa_coords);
	    ##my $cnct_curr_parent = join (':', @curr_parent);

	    my @components = ();
	    for (my $k = $lb_last; $k <= $rb_last; $k++) {
		my $component_sbj = $parent_storage->[$k][$indx_po];
		unless (defined $component_sbj) { $component_sbj = $string_undef; }
		push @components, $component_sbj;
	    }

		# A BUG FIXED on 2019/03/06. (b)
	    print $fh join ("\t", $cnct_curr_pa_coords, @components), "\n";
	    ##print $fh join ("\t", $cnct_curr_parent, @components), "\n";

=cut # END of "For the predecessor, 'printout_multidim_storage'. (3)"


	} # End of the main while-loop (to construct the output).


    (print STDOUT "< Subroutine 'find_maxvalue_in_multidim_storage' Ended (w/ \$maxvalue = $maxvalue) />\n\n") if ($TEST_LEVEL_fnd_mxv_mdms >0);

	return (1, $maxvalue);

}	# End of the subroutine, 'find_maxvalue_in_multidim_storage'. #



# cmpt_log_total_qnt_in_multidim_storage:
#	This subroutine computes the logarithm of the total value of a quantity,
#	whose logarithms (at lattice points) are stored as a specific element in a multi-dimensional storage.
#	(NOTE that the degrees of degeneracy MUST usually be stored as well, as another specific element in the multi-dimensional storage.)
#	The subroutine was created by taking advantage of the framework of 'find_maxvalue_in_multidim_storage' above.
#	[CREATION STARTED on 2020/04/13, ENDED on 2020/04/15.]
#
#	[NOTE (added on 2020/04/15)]:
#	The weight is given as 1/{degree of degeneracy};
#	{degree of degeneracy} == 0 means that the point MUST be EXCLUDED from the total computation.
#
#
# INPUTS: (\@org_bl_coords, \@bds_bl_coords, \@multidim_storage, $indx_sbj, $indx_dgn, $string_undef, $ref_log_qnt), where
#
#	$org_bl_coords[$bl] is the position of the "origin" of the coordinate of the $bl th block.
#	(Although the "origin" should actually be ZERO, it is SHIFTED to this value, SO THAT the range of the coordinate will become entirely non-negaitve. (That is, The coordinates follow a OFFSET format.);
#
#	($lbd_coord, $rbd_coord) = @{$bds_bl_coords[$bl]}
#	are the left- and right-bounds of the coordinate of the $bl th block.
#	[ASSUME that 0 <= $lbd_coord <= $rbd_coord holds;
#	In the present OFFSET format, this should always be satisfied.]
#
#	@storage_for_pnt = @{$mltdim_storage[$k_0]->...->[$k_{B-1}]} (where B = scalar (@bds_bl_coords))
#	is the storage for some information on the point with the coordinate,
#	($k_0, ..., $k_{B-1}) within the hyperrectangle in question.
#
#	The logarithm of the quantity in question is stored as the $indx_sbj th element of @storage_for_pnt.
#
#	The degree of degeneracy is stored as the $indx_dgn th element of @storage_for_pnt.
#	[NOTE: If $indx_dgn = undef, the subroutine ASSUMES that the degree of degeneracy is 1 for ALL lattice points in the hyperrectangle.]
#
#	$string_undef is the string designated for indicating that the component is UNDEFINED.
#	(Commonly used ones are: 'undef', 'n/a', and 'N/A'.)
#
#	$ref_log_qnt is a "reference" value of the logarithm of the subject quantity.
#	[Typically, it is the maximum value of the log-quantity over the lattice points; or sometimes it may be somewhat larger than such maximum.]
#
# OUTPUTS: ($fl_end, $log_totalval_qnt, $total_wgts), where
#
#	$fl_end	= 1 if the subroutine ended without any apparent problems,
#		= 0 if it failed.
#
#	$log_totalval_qnt is the logarithm of the total value of the quantity whose logarithm is specified by the $indx_sbj.
#
#	$total_wgts is the total of the weights (= 1/{degree of degeneracy}).
#
# [ sub find_maxvalue_in_multidim_storage (\@\@\@$$) {
#	my ($org_bl_coords, $bds_bl_coords, $mltdim_storage, $indx_sbj, $string_undef) = @_; ] ... for the predecessor, 'find_maxvalue_in_multidim_storage'.
#
sub cmpt_log_total_qnt_in_multidim_storage (\@\@\@$$$$) {

	my ($org_bl_coords, $bds_bl_coords, $mltdim_storage, $indx_sbj, $indx_dgn, $string_undef, $ref_log_qnt) = @_;

# INPUTS: (\@org_bl_coords, \@bds_bl_coords, \@multidim_storage, $indx_sbj, $indx_dgn, $string_undef, $ref_log_qnt), where


	my $ct_blks = @{$bds_bl_coords}; 
	unless (@{$org_bl_coords} == $ct_blks) {
	    my $ct_orgs = @{$org_bl_coords};
	    print STDOUT "[ERROR 'cmpt_log_total_qnt_in_multidim_storage'] The number of 'origin's (=$ct_orgs) CLASHES with the number of block-coordinate-boundary-pairs (=$ct_blks): $! !\n\n";
	    return (0, undef, undef);
	}
    (print STDOUT "< Subroutine 'cmpt_log_total_qnt_in_multidim_storage' (#{coordinates} = $ct_blks) >\n\n") if ($TEST_LEVEL_cmpt_ltot_mdms >0);

	if ($ct_blks == 0) { # (ADDED on 2020/04/15.) #
	    print STDOUT "[INFORMATION 'cmpt_log_total_qnt_in_multidim_storage'] The number of blocks (=$ct_blks) is ZERO!! ... I will end this subroutine without doing anything ...\n\n";
	    return (1, undef, undef);
	}

		# Examine the coordinate consistencies. #

	my @errors1 = my @errors2 = ();
	for (my $k=0; $k < $ct_blks; $k++) { # Preliminary for-loop (over blocks).
	    my ($lb, $rb) = @{$bds_bl_coords->[$k]};

	    if ($lb < 0) { push @errors1, $k; }
		
	    if ($rb < $lb) { push @errors2, $k; }
	
	} # End of the preliminary for-loop (over blocks).

	if ((@errors1>0) or (@errors2>0)) {
	    print STDOUT "[ERROR 'cmpt_log_total_qnt_in_multidim_storage'] The blocks (", join (', ', @errors1), ") has NEGATIVE left-boundaries, and the blocks (", join (', ', @errors2), ") has their right-bound ON THE LEFT OF their left-bounds: $! !\n\n";
	    return (0, undef);
	}

		# Initialize the current parent. #
	my @curr_parent = ();
	for (my $k=0; $k < $ct_blks-1; $k++) { $curr_parent[$k] = $bds_bl_coords->[$k][0]; }

		# & Initialize the stack (as a nested series of single anonymous arrays specified by the coordinates).

	my $sbj_storage0 = $mltdim_storage;
	my @stack = ($sbj_storage0);
	for (my $k=0; $k < $ct_blks-1; $k++) { # 2nd for-loop (over blocks).
	    my $coord = $curr_parent[$k];
	    my $child_storage = $sbj_storage0->[$coord];
	    push @stack, $child_storage;
	    $sbj_storage0 = $child_storage;
	} # End of the 2nd for-loop (over blocks).
	my $parent_storage0 = pop @stack;

	my ($lb_last, $rb_last) = @{$bds_bl_coords->[$ct_blks-1]};
	my $org_last = $org_bl_coords->[$ct_blks-1];


	my @labels_point = ();
	for (my $k=0; $k < $ct_blks; $k++) { push @labels_point, "X$k"; }
	my $label_point_vector = join (', ', @labels_point);
	$label_point_vector = "($label_point_vector)";

=For the predecessor, 'printout_multidim_storage'. (1)
		# Print out the Header. #

	my @labels_corner = ();
	for (my $k=0; $k < $ct_blks-1; $k++) { push @labels_corner, "X$k"; }
	my $labels_corner = join (':', @labels_corner);
	$labels_corner .= '\X';
	$labels_corner .= $ct_blks-1;
	my @last_coords = ();
	for (my $k=$lb_last; $k <= $rb_last; $k++) { push @last_coords, $k - $org_last; }
	print $fh join ("\t", $labels_corner, @last_coords), "\n\n";

=cut # End of "For the predecessor, 'printout_multidim_storage'. (1)"


		# Sum the ratios of the subject quantities to the reference at the children of the starting parent. #

	my $total_ratios = my $total_wgts = 0;

	for (my $k = $lb_last; $k <= $rb_last; $k++) {

    if ($TEST_LEVEL_cmpt_ltot_mdms > 2) { # (Added in DEBUGGING on 2020/04/19.) (Pre-Main 1) # 
	print STDOUT "\n{ $label_point_vector = (", join (', ', @curr_parent, $k), "): }\n\n";
    } # End of "if ($TEST_LEVEL_cmpt_ltot_mdms > 2) { # (Added in DEBUGGING on 2020/04/19.) (Pre-Main 1) # ...}

	    my $ddgn = (defined $indx_dgn) ? $parent_storage0->[$k][$indx_dgn] : 1 ;
	    unless (defined $ddgn) { next; }

	    ##unless ($ddgn eq $string_undef) { next; } # A Serious BUG Fixed on 2020/04/19. #
	    if ($ddgn eq $string_undef) { next; }

	    if ($ddgn == 0) { next; } # Skip those points with {degree of degeneracy} == 0.

	    my $log_qnt_sbj = $parent_storage0->[$k][$indx_sbj];

	    unless ((defined $log_qnt_sbj) and ($log_qnt_sbj ne $string_undef)) { 
		my $out_log_qnt_sbj = (defined $log_qnt_sbj) ? $log_qnt_sbj : "'UNDEF'";
		print STDOUT "[WARNING 'cmpt_log_total_qnt_in_multidim_storage'] The lattice point, $label_point_vector = (", join (', ', @curr_parent, $k), ") has an ANOMALOUS subject log-quantity, $out_log_qnt_sbj, while its degree of degeneracy is $ddgn !! ... I will skip this point ...\n\n";
		next;
	    }

    if ($TEST_LEVEL_cmpt_ltot_mdms > 2) { # (Pre-Main 2) #
	my $out_log_qnt_sbj = (defined $log_qnt_sbj) ? $log_qnt_sbj : "'UNDEF'";
	print STDOUT '($ddgn, $log_qnt_sbj) = ', "($ddgn, $out_log_qnt_sbj) .\n\n\n";
    } # End of "if ($TEST_LEVEL_cmpt_ltot_mdms > 2) { # (Pre-Main 2) # ...}

	    $total_ratios += exp ($log_qnt_sbj - $ref_log_qnt) /$ddgn;
	    $total_wgts += 1/$ddgn;

	}

=For the predecessor, 'printout_multidim_storage'. (2)
		# A BUG FIXED on 2019/03/06. (c)
	my @curr_pa_coords0 = ();
	for (my $i=0; $i < $ct_blks -1; $i++) {
	    $curr_pa_coords0[$i] = $curr_parent[$i] - $org_bl_coords->[$i];
	}
	my $cnct_curr_pa_coords0 = join (':', @curr_pa_coords0);
	##my $cnct_curr_parent = join (':', @curr_parent);

	my @components0 = ();
	for (my $k = $lb_last; $k <= $rb_last; $k++) {
	    my $component_sbj = $parent_storage0->[$k][$indx_po];
	    unless (defined $component_sbj) { $component_sbj = $string_undef; }
	    push @components0, $component_sbj;
	}
		# A BUG FIXED on 2019/03/06. (d)
	print $fh join ("\t", $cnct_curr_pa_coords0, @components0), "\n";
	##print $fh join ("\t", $cnct_curr_parent, @components0), "\n";

=cut # End of "For the predecessor, 'printout_multidim_storage'. (2)"

	if ($ct_blks < 2) {
			# This is necessary for the subroutine to end normally even if ($ct_blks < 2). #
			# (ADDED on 2020/04/15.)

	    my $log_totalval_qnt = ($total_ratios > 0) ? $ref_log_qnt + log ($total_ratios) : undef ;

	    return (1, $log_totalval_qnt, $total_wgts);
	}


		# Main: Examine the subject components from the entire storage. #

	my $indx_curr_crd = $ct_blks -2;

	while (1) { # Main while-loop (to construct the output).

	    my $curr_sbj_coord = $curr_parent[$indx_curr_crd];
	    my ($lbd, $rbd) = @{$bds_bl_coords->[$indx_curr_crd]};

	    if ($curr_sbj_coord == $rbd) {
		if ($indx_curr_crd == 0) { last; }

		$curr_parent[$indx_curr_crd] = $lbd;
		$indx_curr_crd--;
		pop @stack;
		next;
	    }

		# A BUG fixed (on 2019/02/22).
	    $curr_parent[$indx_curr_crd]++;
	    $curr_sbj_coord = $curr_parent[$indx_curr_crd];
	    ##$curr_sbj_coord = $curr_parent[$indx_curr_crd]++;

	    my $grand_parent_storage = $stack[$#stack];

	    my $parent_storage = $grand_parent_storage->[$curr_sbj_coord];

	    while ($indx_curr_crd < $ct_blks - 2) { # Inner while-loop.
		push @stack, $parent_storage;
		$indx_curr_crd++;
		$curr_sbj_coord = $curr_parent[$indx_curr_crd];
		$parent_storage = $parent_storage->[$curr_sbj_coord];

	    } # End of the inner while-loop.


			### RESTART FROM HERE ###

			# Sum the ratios of the subject quantities to the reference at the children of the current parent. #

	    for (my $k = $lb_last; $k <= $rb_last; $k++) {

    if ($TEST_LEVEL_cmpt_ltot_mdms > 2) { # (Added in DEBUGGING on 2020/04/19.) (Pre-Main 1) # 
	print STDOUT "\n{ $label_point_vector = (", join (', ', @curr_parent, $k), "): }\n\n";
    } # End of "if ($TEST_LEVEL_cmpt_ltot_mdms > 2) { # (Added in DEBUGGING on 2020/04/19.) (Pre-Main 1) # ...}

	        my $ddgn = (defined $indx_dgn) ? $parent_storage->[$k][$indx_dgn] : 1 ;
	        unless (defined $ddgn) { next; }
	        if ($ddgn eq $string_undef) { next; }
	        if ($ddgn == 0) { next; } # Skip those points with {degree of degeneracy} == 0.

	        my $log_qnt_sbj = $parent_storage->[$k][$indx_sbj];

	        if (!(defined $log_qnt_sbj) or ($log_qnt_sbj eq $string_undef)) { 
		    my $out_log_qnt_sbj = (defined $log_qnt_sbj) ? $log_qnt_sbj : "'UNDEF'";
		    print STDOUT "[WARNING 'cmpt_log_total_qnt_in_multidim_storage'] The lattice point, $label_point_vector = (", join (', ', @curr_parent, $k), ") has an ANOMALOUS subject log-quantity, $out_log_qnt_sbj, while its degree of degeneracy is $ddgn !! ... I will skip this point ...\n\n";
		    next;
	        }

    if ($TEST_LEVEL_cmpt_ltot_mdms > 2) { # (Pre-Main 2) #
	my $out_log_qnt_sbj = (defined $log_qnt_sbj) ? $log_qnt_sbj : "'UNDEF'";
	print STDOUT '($ddgn, $log_qnt_sbj) = ', "($ddgn, $out_log_qnt_sbj) .\n\n\n";
    } # End of "if ($TEST_LEVEL_cmpt_ltot_mdms > 2) { # (Pre-Main 2) # ...}

	        $total_ratios += exp ($log_qnt_sbj - $ref_log_qnt) /$ddgn;
	        $total_wgts += 1/$ddgn;
	    }

=For the predecessor, 'find_maxvalue_in_multidim_storage'.
			# Examine the subject components at the current points. #

	    for (my $k = $lb_last; $k <= $rb_last; $k++) { # Main inner for-loop. #

		my $component_sbj = $parent_storage->[$k][$indx_sbj];
		unless (defined $component_sbj) { next; }
		if ($component_sbj eq $string_undef) { next; }

		if ((defined $maxvalue) and ($maxvalue >= $component_sbj)) { next; }
		$maxvalue = $component_sbj;

	    } # End of the main inner for-loop. #
=cut # END of "For the predecessor, 'find_maxvalue_in_multidim_storage'."

=For the predecessor, 'printout_multidim_storage'. (3)
		# Print out the components at the current points. #

		# A BUG FIXED on 2019/03/06. (a)
	    my @curr_pa_coords = ();
	    for (my $i = 0; $i < $ct_blks -1; $i++) {
		$curr_pa_coords[$i] = $curr_parent[$i] - $org_bl_coords->[$i];
	    }
	    my $cnct_curr_pa_coords = join (':', @curr_pa_coords);
	    ##my $cnct_curr_parent = join (':', @curr_parent);

	    my @components = ();
	    for (my $k = $lb_last; $k <= $rb_last; $k++) {
		my $component_sbj = $parent_storage->[$k][$indx_po];
		unless (defined $component_sbj) { $component_sbj = $string_undef; }
		push @components, $component_sbj;
	    }

		# A BUG FIXED on 2019/03/06. (b)
	    print $fh join ("\t", $cnct_curr_pa_coords, @components), "\n";
	    ##print $fh join ("\t", $cnct_curr_parent, @components), "\n";

=cut # END of "For the predecessor, 'printout_multidim_storage'. (3)"


	} # End of the main while-loop (to construct the output).


	my $log_totalval_qnt = ($total_ratios > 0) ? $ref_log_qnt + log ($total_ratios) : undef ;

	my $out_log_totalval_qnt = (defined $log_totalval_qnt) ? $log_totalval_qnt : "'undef'";

    (print STDOUT "< Subroutine 'cmpt_log_total_qnt_in_multidim_storage' Ended (w/ \$log_totalval_qnt = $out_log_totalval_qnt, \$total_wgts = $total_wgts) />\n\n") if ($TEST_LEVEL_cmpt_ltot_mdms >0);

	return (1, $log_totalval_qnt, $total_wgts);

}	# End of the subroutine, 'cmpt_log_total_qnt_in_multidim_storage'. #


1;		# a "spell" which makes the module "use"-able and/or "require"-able #


