package MyDetect_purge_cands;
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyDetect_purge_cands.pm:
#	This package is a collection of subroutines
#	necessary for detecting the candidates of purge-like errors.
#	[CREATION Started on May 3 (Tue), 2016, ENDED on Mmm dd (www), 2016.]
#
#	[REVISED on 2019/06/15]
#	A new subroutine, 'detect_purge_cands2', was created
#	by MODIFYing the existing 'detect_purge_cands'.
#

use strict;
use warnings;

use MyTaxonomy qw(
	if_ancestor fetch_ancestry calculate_abs_depth
	fetch_ext_offsprings_nr fetch_all_offsprings_nr
	);

use MyTreeManipu_II qw( );
	# extract_sgltree_from_file output_tree 

use MyTreeManipu_III qw( make_node2abs_depth );

use MyLikelihoods_respattern_msa_dna qw(
	calc_node2ntfreqs_td
	calc_prob_descresstats_cond_ancstat2
	calc_joint_prob_ancresstats_and_ancstat2
	calc_clm_probs
	);


=set aside for potential use
use MyTreeManipu_IV qw( bi2multifurcated_top construct_neat_subtree );
use MyPAML_parser qw( retrieve_aa_freqs prepare_paml_ctrl );
=cut

use MyProbDistri qw(
	binomial_cum_via_betai
	);

use MyArrayManipu qw(
	merge_integer_intervals
	);

	# Added in creation of 'detect_purge_cands2' (on 2019/06/15). #
use MyArrayManipu_num qw(
	union_nmarray
	binsearch_num
	); # Added 'binsearch_num' in revision of 'detect_purge_cands2' (on 2019/06/16).
	# End of "Added in creation of 'detect_purge_cands2' (on 2019/06/15)." #

use Exporter;

our @ISA =qw(Exporter);

our @EXPORT_OK = qw( 
	sum_arrays_over_branches
	calc_br2cw_expcts_sbsts_ue
	calc_br2cw_expcts_sbsts_le
	calc_br2cw_expcts_sbsts_ave
	calc_br2cw_obscts_sbsts
	calc_br2cw_expcts_obscts_sbsts
	calc_pv_obscts_sbsts_wd_br
	calc_modpv_obscts_sbsts_wd_br
	mk_node2sw_gstats
	detect_purge_cands
	merge_sitesets
	assign_sgn_wds_to_regions
	detect_purge_cands2
	);
	# Added 'merge_sitesets' & 'detect_purge_cands2' on 2019/06/15. #
	# Added 'assign_sgn_wds_to_regions' on 2019/06/15.


	### Default values of the Parameters ###

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;

our @NUCLEOTIDES = @MyLikelihoods_respattern_msa_dna::NUCLEOTIDES; # Unless otherwise stated, nucleotide MSA scores will be calculated according to this order (or index assignment) of the nucleotide bases.
our $CT_NUCLS = @NUCLEOTIDES;

our %NUCL2INDX; # = ($nucleotide => {index in @NUCLEOTIDES}, ...)
for (my $k=0; $k<$CT_NUCLS; $k++) { $NUCL2INDX{$NUCLEOTIDES[$k]} = $k; }

our $LOG_ZERO = $MyLikelihoods_respattern_msa_dna::LOG_ZERO;


	## The following parameters specify the verboseness of the test output.
	## (0, silent; 1, very modest; 2, modest; 3, medium; 4, verbose; 5, very verbose; 6, extremely verbose.)

	# For 'calc_prob_descresstats_cond_ancstat2'. #
our $TEST_LEVEL_cpdca = 0;
##our $TEST_LEVEL_cpdca = 6;

	# For 'detect_purge_cands'. #
our $TEST_LEVEL = 0;
###our $TEST_LEVEL = 5;


### 5 important hash tables which describe a tree ###
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.


# sum_arrays_over_branches:
#	This subroutine inputs arrays associated with the branches,
#	and outputs an array which is the summation of the input arrays over branches.
# 	NOTE: We assume that all the input arrays have the same size.
#	[CREATED on May 6 (Fri), 2016.]
#
# INPUT: \%br2array, where
#	%br2array = ($br_id => \@array, ...) for ALL branches.
#
# OUTPUT: \@sum_array, where
#	$sum_array[$i] = Sum_{$br_id in keys (%br2array)} $br2array{$br_id}->[$i].
#
sub sum_arrays_over_branches (\%) {

	my ($br2array) = @_;

	my @sum_array = ();
	my $size_array0;

	foreach my $br_id (sort {$a <=> $b} keys %{$br2array}) { # Main outer foreach-loop (over branches). #

	    my $array = $br2array->{$br_id};
	    my $size_array = @{$array};

	    unless (defined $size_array0) { # Initialize the summation. #
		$size_array0 = $size_array;
		@sum_array = @{$array};
		next;
	    }

	    unless ($size_array0 == $size_array) {
		print STDOUT "[ERROR 'sum_arrays_over_branches'] The array associated with branch ID $br_id has the size $size_array, which CLASHES with the previous array size (=$size_array): $! !\n\n";
		return undef;
	    }

			# Add the current array to the summed array. #

	    for (my $i=0; $i<$size_array0; $i++) { $sum_array[$i] += $array->[$i]; }

	}					# End of the main outer foreach-loop (over branches). #

	return \@sum_array;

}	# End of the subroutine 'sum_arrays_over_branches'. #


# calc_br2cw_expcts_sbsts_ue:
#	This subroutine calculates the expected numbers of substitutions (actually substitutional differences) at all sites
#	based on the semi-posterior probabilities of the residues at the upper-end of each branch. 
#	[CREATED on May 3 (Tue), 2016.]
#
#	[Formula]
#	D_exp_1A (b, c) = 1 - [Sum_{\omega in \Omega} P_U (\omega; b, c) P(\omega -> \omega; b)] / [Sum_{\omega'} P_U(\omega'; b, c)].
#	Here, P_U(\omega; b, c) is the joint probability that the extant sequences on the "ancestral side" of branch b have the states in column c, in addition to that the sequence at the upper-end of b has \omega in column c.
#	P(\omega -> \omega' ; b) is the transition probability to \omega' (at the lower-end), given \omega (at the upper-end), along branch b.
#
# INPUT: (\%br2tr_mtrx, \%br2set_sw_jnt_probs), where
#
#	@{$br2tr_mtrx{$br_id}} is the transition matrix
#	that describes the transition probabilities between both ends of the branch $br_id.
#	More specifically,
#	$br2tr_mtrx{$br_id}->[$i][$j] is the probability that the node $br_id shows $NUCLEOTIDES[$j]
#	given that the parent node of $br_id shows $NUCLEOTIDES[$i].
#
#	%br2set_sw_jnt_probs = ($br_id => \@set_sw_jnt_probs, ...) for ALL branches, where 
#	$set_sw_jnt_probs[$s]->[$i] is the probability that
#	all the extant sequences on the "ancestral side" of $br_id show the residue states that are actually found in the $s th site,
#	in addition to that $NUCLEOTIDES[$i] occupies the $s th site of the UPPER-end node of $br_id.
#	(It is an output of 'calc_joint_prob_ancresstats_and_ancstat2' in 'MyLikelihoods_respattern_msa_dna.pm'.)
#
# OUTPUT: \%br2set_cw_expcts_sbsts_ue, where
#
#	%br2set_cw_expcts_sbsts_ue = ($br_id => \@set_cw_expcts_sbcts_ue, ...) for ALL branches, where
#	$set_cw_expcts_sbsts_ue[$s] is the expected number of substitutions (more precisely, substitutional differences) along the branch $br_id in the $s th column, based on the semi-posterior probabilities of residues at the upper-end of $br_id.
#
sub calc_br2cw_expcts_sbsts_ue (\%\%) {

	my ($br2tr_mtrx, $br2set_sw_jnt_probs) = @_;

	my %br2set_cw_expcts_sbsts_ue;

	foreach my $br_id (sort {$a <=> $b} keys %{$br2tr_mtrx}) { # Outer foreach-loop (over branches). #

	    my $tr_mtrx = $br2tr_mtrx->{$br_id};
	    my $set_sw_jnt_probs = $br2set_sw_jnt_probs->{$br_id};
	    my $ct_clms = @{$set_sw_jnt_probs};

	    my @set_cw_expcts_sbsts_ue = ();

	    for (my $s=0; $s<$ct_clms; $s++) { # Middle for-loop (over sites = columns). #

		my $jnt_probs = $set_sw_jnt_probs->[$s];

		my $num = my $denom = 0;

		for (my $i=0; $i<$CT_NUCLS; $i++) { # Inner for-loop (over nucleotides). #
		    my $jnt_prb = $jnt_probs->[$i];

			# A BUG fixed on May 4 (Wed), 2016. #
		    $num += $jnt_prb * $tr_mtrx->[$i][$i];
		    $denom += $jnt_prb;
		    ##$num += $jnt_prb;
		    ##$denom += $jnt_prb * $tr_mtrx->[$i][$i];

		}			# End of the inner for-loop (over nucleotides). #

			# A BUG fixed on May 4 (Wed), 2016. #
		$set_cw_expcts_sbsts_ue[$s] = 1 - $num/$denom;
		##$set_cw_expcts_sbsts_ue[$s] = $num/$denom;

	    }				# End of the middle for-loop (over sites = columns). #

	    $br2set_cw_expcts_sbsts_ue{$br_id} = \@set_cw_expcts_sbsts_ue;

	}			# End of the outer foreach-loop (over branches). #

	return \%br2set_cw_expcts_sbsts_ue;

}	# End of the subroutine, 'calc_br2cw_expcts_sbsts_ue'. #


# calc_br2cw_expcts_sbsts_le:
#	This subroutine calculates the expected numbers of substitutions (actually substitutional differences) at all sites
#	based on the semi-posterior probabilities of the residues at the lower-end of each branch. 
#	[CREATED on May 3 (Tue), 2016.]
#
#	[Formula]
#	D_exp_1B (b, c) = 1 - Sum_{\omega in \Omega} P_{(L)} [\omega -> \omega | b, c],
#	with
#	P_{(L)} [\omega -> \omega | b, c] =: 
#	P (\omega; n_U(b))* P(\omega->\omega; b)*P_L (\omega; b, c) /[Sum_{\omega'} P(\omega'; n_L(b)) * P_L (\omega'; b, c) ] .
#	Here, P_L (\omega; b, c) is the conditional probability that the extant sequences on the "descendant side" of branch b have the states in column c, conditioned on to that the sequence at the lower-end of b has \omega in column c.
#	P(\omega -> \omega' ; b) is the transition probability to \omega' (at the lower-end), given \omega (at the upper-end), along branch b.
#	P (\omega; n) is the frequency of the residue \omega at the node n,
#	and n_U(b) and n_L(b) are the nodes at the upper-end and the lower-end, respectivley, of the branch $b.
#
# INPUT: (\%br2tr_mtrx, \%node2set_sw_cond_probs, \%node2ntfreqs, \%node2pa), where
#
#	@{$br2tr_mtrx{$br_id}} is the transition matrix
#	that describes the transition probabilities between both ends of the branch $br_id.
#	More specifically,
#	$br2tr_mtrx{$br_id}->[$i][$j] is the probability that the node $br_id shows $NUCLEOTIDES[$j]
#	given that the parent node of $br_id shows $NUCLEOTIDES[$i].
#
#	%node2set_sw_cond_probs = ($node_id => \@set_sw_cond_probs, ...) for ALL nodes, where
#	$set_sw_cond_probs[$s]->[$i] is the probability that
#	all the descendant sequences of $node_id show the residue states that are actually found in the $s th site,
#	given that $NUCLEOTIDES[$i] occupies the $s th site of the ancestral node, $node_id.
#	But $set_sw_cond_probs[$s] = [] (empty) if $node_id refers to an external node and if the node has its $s th site occupied by a gap.
#	(It is an output of 'calc_prob_descresstats_cond_ancstat2' in 'MyLikelihoods_respattern_msa_dna.pm'.)
#
#	%node2ntfreqs = ($node_id => \@ntfreqs, ...), where
#	$ntfreqs[$i] is the frequency of $NUCLEOTIDES[$i] at the node, $node_id.
#
#	%node2pa = ($node_id => $parent_node_id, ...) for all nodes but the root.
#
# OUTPUT: \%br2set_cw_expcts_sbsts_le, where
#
#	%br2set_cw_expcts_sbsts_le = ($br_id => \@set_cw_expcts_sbcts_le, ...) for ALL branches, where
#	$set_cw_expcts_sbsts_le[$s] is the expected number of substitutions (more precisely, substitutional differences) along the branch $br_id in the $s th column, based on the semi-posterior probabilities of residues at the lower-end of $br_id.
#
sub calc_br2cw_expcts_sbsts_le (\%\%\%\%) {

	my ($br2tr_mtrx, $node2set_sw_cond_probs, $node2ntfreqs, $node2pa) = @_;

	my @cond_probs_allgaps = ();
	for (1 .. $CT_NUCLS) { push @cond_probs_allgaps, 1; }

	my %br2set_cw_expcts_sbsts_le;

	foreach my $br_id (sort {$a <=> $b} keys %{$br2tr_mtrx}) { # Outer foreach-loop (over branches). #

	    my $tr_mtrx = $br2tr_mtrx->{$br_id};

	    my $set_sw_cond_probs = $node2set_sw_cond_probs->{$br_id};
	    my $ct_sites = @{$set_sw_cond_probs};

	    my $ntfreqs_le = $node2ntfreqs->{$br_id};

	    my $node_ue = $node2pa->{$br_id}; # The upper-end. #
	    my $ntfreqs_ue = $node2ntfreqs->{$node_ue};

	    my @set_cw_expcts_sbsts_le = ();

	    for (my $s=0; $s<$ct_sites; $s++) { # Middle for-loop (over sites = columns). #

		my $cond_probs = $set_sw_cond_probs->[$s];
		((defined $cond_probs) and (@{$cond_probs}>0)) or ($cond_probs = \@cond_probs_allgaps);

		my $num = my $denom = 0;

		for (my $i=0; $i<$CT_NUCLS; $i++) { # Inner for-loop (over nucleotides). #

		    $num += $ntfreqs_ue->[$i] * $tr_mtrx->[$i][$i] * $cond_probs->[$i];
		    $denom += $ntfreqs_le->[$i] * $cond_probs->[$i];
		}			# End of the inner for-loop (over nucleotides). #

			# A BUG fixed on May 4 (Wed), 2016. #
		$set_cw_expcts_sbsts_le[$s] = 1 - $num /$denom ;
		##$set_cw_expcts_sbsts_le[$s] = $num /$denom ;

	    }				# End of the middle for-loop (over sites = columns). #

	    $br2set_cw_expcts_sbsts_le{$br_id} = \@set_cw_expcts_sbsts_le;

	}						# End of the outer foreach-loop (over branches). #

	return \%br2set_cw_expcts_sbsts_le;

}	# End of the subroutine 'calc_br2cw_expcts_sbsts_le'. #


# calc_br2cw_expcts_sbsts_ave:
#	This subroutine calculates two types of the expected numbers of substitutions (actually substitutional differences) at all sites,
#	one based on the semi-posterior probabilities of the residues at the upper-end of each branch
#	and the other based on the semi-posterior probabilities of the residues at the lower-end of each branch.
#	Then, the subroutine calculates the average (i.e., the arithmetic mean) of the two types of expected numbers.
#	[CREATED on May 3 (Tue), 2016.]
#
# INPUT: (\%br2tr_mtrx, \%br2set_sw_jnt_probs, \%node2set_sw_cond_probs, \%node2ntfreqs, \%node2pa), where
#
#	@{$br2tr_mtrx{$br_id}} is the transition matrix
#	that describes the transition probabilities between both ends of the branch $br_id.
#	More specifically,
#	$br2tr_mtrx{$br_id}->[$i][$j] is the probability that the node $br_id shows $NUCLEOTIDES[$j]
#	given that the parent node of $br_id shows $NUCLEOTIDES[$i].
#
#	%br2set_sw_jnt_probs = ($br_id => \@set_sw_jnt_probs, ...) for ALL branches, where 
#	$set_sw_jnt_probs[$s]->[$i] is the probability that
#	all the extant sequences on the "ancestral side" of $br_id show the residue states that are actually found in the $s th site,
#	in addition to that $NUCLEOTIDES[$i] occupies the $s th site of the UPPER-end node of $br_id.
#	(It is an output of 'calc_joint_prob_ancresstats_and_ancstat2' in 'MyLikelihoods_respattern_msa_dna.pm'.)
#
#	%node2set_sw_cond_probs = ($node_id => \@set_sw_cond_probs, ...) for ALL nodes, where
#	$set_sw_cond_probs[$s]->[$i] is the probability that
#	all the descendant sequences of $node_id show the residue states that are actually found in the $s th site,
#	given that $NUCLEOTIDES[$i] occupies the $s th site of the ancestral node, $node_id.
#	But $set_sw_cond_probs[$s] = [] (empty) if $node_id refers to an external node and if the node has its $s th site occupied by a gap.
#	(It is an output of 'calc_prob_descresstats_cond_ancstat2' in 'MyLikelihoods_respattern_msa_dna.pm'.)
#
#	%node2ntfreqs = ($node_id => \@ntfreqs, ...), where
#	$ntfreqs[$i] is the frequency of $NUCLEOTIDES[$i] at the node, $node_id.
#
#	%node2pa = ($node_id => $parent_node_id, ...) for all nodes but the root.
#
#
# OUTPUT: (\%br2set_cw_expcts_sbsts_ave, \%br2set_cw_expcts_sbsts_ue, \%br2set_cw_expcts_sbsts_le), where
#
#	%br2set_cw_expcts_sbsts_ave = ($br_id => \@set_cw_expcts_sbcts_ave, ...) for ALL branches, where
#	$set_cw_expcts_sbsts_ave[$s] = ($set_cw_expcts_sbsts_ue[$s] + $set_cw_expcts_sbsts_le[$s])/2 .
#
#	%br2set_cw_expcts_sbsts_ue = ($br_id => \@set_cw_expcts_sbcts_ue, ...) for ALL branches, where
#	$set_cw_expcts_sbsts_ue[$s] is the expected number of substitutions (more precisely, substitutional differences) along the branch $br_id in the $s th column, based on the semi-posterior probabilities of residues at the upper-end of $br_id;
#
#	%br2set_cw_expcts_sbsts_le = ($br_id => \@set_cw_expcts_sbcts_le, ...) for ALL branches, where
#	$set_cw_expcts_sbsts_le[$s] is the expected number of substitutions (more precisely, substitutional differences) along the branch $br_id in the $s th column, based on the semi-posterior probabilities of residues at the lower-end of $br_id;
#
sub calc_br2cw_expcts_sbsts_ave (\%\%\%\%\%) {

	my ($br2tr_mtrx, $br2set_sw_jnt_probs, $node2set_sw_cond_probs, $node2ntfreqs, $node2pa) = @_;

	my $br2set_cw_expcts_sbsts_ue = calc_br2cw_expcts_sbsts_ue (%{$br2tr_mtrx}, %{$br2set_sw_jnt_probs});
	unless (defined $br2set_cw_expcts_sbsts_ue) {
	    print STDOUT "[ERROR 'calc_br2cw_expcts_sbsts_ave'] Subroutine 'calc_br2cw_expcts_sbsts_ue' FAILED: $! !\n\n";
	    return (undef, undef, undef);
	}

	my $br2set_cw_expcts_sbsts_le = calc_br2cw_expcts_sbsts_le (%{$br2tr_mtrx}, %{$node2set_sw_cond_probs}, %{$node2ntfreqs}, %{$node2pa}) ;
	unless (defined $br2set_cw_expcts_sbsts_le) {
	    print STDOUT "[ERROR 'calc_br2cw_expcts_sbsts_ave'] Subroutine 'calc_br2cw_expcts_sbsts_le' FAILED: $! !\n\n";
	    return (undef, undef, undef);
	}

	my %br2set_cw_expcts_sbsts_ave;

	foreach my $br_id (sort {$a <=> $b} keys %{$br2set_cw_expcts_sbsts_ue}) { # Outer foreach-loop (over branches). #

	    my $set_cw_expcts_sbsts_ue = $br2set_cw_expcts_sbsts_ue->{$br_id};
	    my $set_cw_expcts_sbsts_le = $br2set_cw_expcts_sbsts_le->{$br_id};

	    my $ct_clms = @{$set_cw_expcts_sbsts_ue};
	    unless (@{$set_cw_expcts_sbsts_le} == $ct_clms) {
		print STDOUT "[ERROR 'calc_br2cw_expcts_sbsts_ave'] The number of columns in ", '@{$set_cw_expcts_sbsts_le} (= ', scalar (@{$set_cw_expcts_sbsts_le}), ") CLASHES with that of colums in ", '@{$set_cw_expcts_sbsts_ue} (= ', "$ct_clms): $! !\n\n";
		return (undef, undef, undef);
	    }

	    my @set_cw_expcts_sbsts_ave = ();

	    for (my $s=0; $s<$ct_clms; $s++) { # Inner for-loop (over columns). #

		my $ct_ue = $set_cw_expcts_sbsts_ue->[$s];
		my $ct_le = $set_cw_expcts_sbsts_le->[$s];

		my $ct_ave = ($ct_ue + $ct_le)/2;

		$set_cw_expcts_sbsts_ave[$s] = $ct_ave;

	    }			# End of the inner for-loop (over columns). #

	    $br2set_cw_expcts_sbsts_ave{$br_id} = \@set_cw_expcts_sbsts_ave;

	}					# End of the outer foreach-loop (over branches). #

	return (\%br2set_cw_expcts_sbsts_ave, $br2set_cw_expcts_sbsts_ue, $br2set_cw_expcts_sbsts_le);

}	# End of the subroutine 'calc_br2cw_expcts_sbsts_ave'. #


# calc_br2cw_obscts_sbsts:
#	This subroutine calculates the "observed" numbers of substitutions (actually substitutional differences) at all sites,
#	along all branches.
#	[CREATED on May 4 (Wed), 2016.]
#
#	[FORMULA]
#	The "observed" number of substitutional differences along branch b in column c is:
#	D_obs (b, c) = 1 - Sum_{\omega in \Omega} P_(U,L) [\omega, \omega | b, c],
#	where
#	P_(U,L) [\omega, \omega' | b, c] =: P_U (\omega; b, c ) * P(\omega -> \omega'; b) * P_L (\omega'; b, c) / P(c)
#	is the joint probability that the upper-end and the lower-end of branch b have the residue states \omega and \omega', respectively, in the column c.
#
#	Here, P_U (\omega; b, c) is the joint probability that the extant sequence on the "ancestral side" of branch b
#	shows the actual residue states in column c, in addition to that the upper-end of branch b shows the residue state \omega.
#	P_L (\omega, b, c) is the conditional probability that the extant sequence on the "descendant side" of branch b
#	shows the actual residue states in column c, conditioned on that the lower-end of branch b shows the residue state \omega.
#	P (c) is the probability that the residue state of column c actually occurs under the given substitition model.
#	P(\omega -> \omega'; b) is the transition probability from \omega (at the upper-end) to \omega' (at the lower-end) along branch b.
#
# INPUT: (\%br2tr_mtrx, \%br2set_sw_jnt_probs, \%node2set_sw_cond_probs, \@set_clm_probs), where
#
#	@{$br2tr_mtrx{$br_id}} is the transition matrix
#	that describes the transition probabilities between both ends of the branch $br_id.
#	More specifically,
#	$br2tr_mtrx{$br_id}->[$i][$j] is the probability that the node $br_id shows $NUCLEOTIDES[$j]
#	given that the parent node of $br_id shows $NUCLEOTIDES[$i].
#
#	%br2set_sw_jnt_probs = ($br_id => \@set_sw_jnt_probs, ...) for ALL branches, where 
#	$set_sw_jnt_probs[$s]->[$i] is the probability that
#	all the extant sequences on the "ancestral side" of $br_id show the residue states that are actually found in the $s th site,
#	in addition to that $NUCLEOTIDES[$i] occupies the $s th site of the UPPER-end node of $br_id.
#	(It is an output of 'calc_joint_prob_ancresstats_and_ancstat2' in 'MyLikelihoods_respattern_msa_dna.pm'.)
#
#	%node2set_sw_cond_probs = ($node_id => \@set_sw_cond_probs, ...) for ALL nodes, where
#	$set_sw_cond_probs[$s]->[$i] is the probability that
#	all the descendant sequences of $node_id show the residue states that are actually found in the $s th site,
#	given that $NUCLEOTIDES[$i] occupies the $s th site of the ancestral node, $node_id.
#	But $set_sw_cond_probs[$s] = [] (empty) if $node_id refers to an external node and if the node has its $s th site occupied by a gap.
#	(It is an output of 'calc_prob_descresstats_cond_ancstat2' in 'MyLikelihoods_respattern_msa_dna.pm'.)
#
#	$set_clm_probs[$c] is the occurrence probability of the $c th column
#	under the given substitution model.
#	(It is an output of 'calc_clm_probs' in 'MyLikelihoods_respattern_msa_dna.pm'.)
#
# OUTPUT: \%br2set_cw_obscts_sbsts, where
#	%br2set_cw_obscts_sbsts = ($br_id => \@set_cw_obscts_sbcts, ...) for ALL branches, where
#	$set_cw_obscts_sbsts[$c] is the "observed" number of substitutions (more precisely, substitutional differences)
#	along the branch $br_id in the $c th column, estimated using ALL extant residue states in c.
#
sub calc_br2cw_obscts_sbsts (\%\%\%\@) {

	my ($br2tr_mtrx, $br2set_sw_jnt_probs, $node2set_sw_cond_probs, $set_clm_probs) = @_;

	my $ct_clms = @{$set_clm_probs};

	my @cond_probs_allgaps = ();
	for (1 .. $CT_NUCLS) { push @cond_probs_allgaps, 1; }

	my %br2set_cw_obscts_sbsts;

	foreach my $br_id (sort {$a <=> $b} keys %{$br2tr_mtrx}) {  # Outer foreach-loop (over branches). #

	    my $tr_mtrx = $br2tr_mtrx->{$br_id};

	    my $sw_jnt_probs = $br2set_sw_jnt_probs->{$br_id};
	    my $sw_cond_probs = $node2set_sw_cond_probs->{$br_id};

	    unless ($ct_clms == @{$sw_jnt_probs}) {
		my $ct_clms2 = @{$sw_jnt_probs};
		print STDOUT "[ERROR 'calc_br2cw_obscts_sbsts'] The size of ", '@{$sw_jnt_probs} (= ', "$ct_clms2) CLASHES with the size of ", '@{$set_clm_probs} (= ', "$ct_clms): $! !\n\n";
		return undef;
	    }
	    unless ($ct_clms == @{$sw_cond_probs}) {
		my $ct_clms3 = @{$sw_cond_probs};
		print STDOUT "[ERROR 'calc_br2cw_obscts_sbsts'] The size of ", '@{$sw_cond_probs} (= ', "$ct_clms3) CLASHES with the size of ", '@{$set_clm_probs} (= ', "$ct_clms): $! !\n\n";
		return undef;
	    }

	    my @set_cw_obscts_sbsts = ();

	    for (my $c=0; $c<$ct_clms; $c++) { # Middle for-loop (over columns). #

		my $jnt_probs = $sw_jnt_probs->[$c];
		my $cond_probs = $sw_cond_probs->[$c];

			# ADDED to silence the warning output when encountering the external node occupied by a gap. #
		((defined $cond_probs) and (@{$cond_probs}>0)) or ($cond_probs = \@cond_probs_allgaps);

		my $denom = $set_clm_probs->[$c];

		my $num = 0;
		for (my $i=0; $i<$CT_NUCLS; $i++) { $num += $jnt_probs->[$i] * $tr_mtrx->[$i][$i] * $cond_probs->[$i]; }

		$set_cw_obscts_sbsts[$c] = 1 - $num/$denom;

	    }				# End of the middle for-loop (over columns). #

	    $br2set_cw_obscts_sbsts{$br_id} = \@set_cw_obscts_sbsts;

	}					# End of the outer foreach-loop (over branches). #

	return \%br2set_cw_obscts_sbsts;

}	# End of the subroutine 'calc_br2cw_obscts_sbsts'. #


# calc_br2cw_expcts_obscts_sbsts:
#	This MAIN subroutine puts together a number of subroutines
#	and calculates the expected and "observed" numbers of substitutional differences
#	in each column along each branch, from a given input MSA.
#	The subroutine leaves the input MSA intact.
#	[CREATED on May 4 (Wed), 2016.]
#
# INPUT: (\@in_msa, \@seqnames, \%name2id, \%node2ch, \%node2pa, \%br2tr_mtrx, \%node2ntfreqs, $gap), where
#
#	${$in_msa[$k]} is the $k th sequence embedded in the input MSA;
#	$seqnames[$k] is the name (or ID) of the $k th sequence in the input MSA;
#	%name2id = ($node_name => $node_id, ...) for the exterior nodes and the top node (i.e., root) of the tree;
#	%node2ch = ($node_id => \@child_nodes, ...) for the interior nodes of the tree;
#	%node2pa = ($node_id => $parent_id, ...) for all nodes but the root;
#
#	@{$br2tr_mtrx{$br_id}} is the transition matrix
#	that describes the transition probabilities between both ends of the branch $br_id.
#	More specifically,
#	$br2tr_mtrx{$br_id}->[$i][$j] is the probability that the node $br_id shows $NUCLEOTIDES[$j]
#	given that the parent node of $br_id shows $NUCLEOTIDES[$i].
#
#	%node2ntfreqs = ($node_id => \@ntfreqs, ...), where
#	$ntfreqs[$i] is the frequency of $NUCLEOTIDES[$i] at the node, $node_id.
#
#	$gap is the character representing a gap.
#
# OUTPUT: (\%br2set_cw_expcts_sbsts_ave, \%br2set_cw_obscts_sbsts), where
#
#	%br2set_cw_expcts_sbsts_ave = ($br_id => \@set_cw_expcts_sbcts_ave, ...) for ALL branches, where
#	$set_cw_expcts_sbsts_ave[$s] is the expected number of substitutions (more precisely, substitutional differences) along the branch $br_id in the $s th column, which is the average (i.e., arithmetic mean) of the upper-end-based and the lower-end-based numbers.
#
#	%br2set_cw_obscts_sbsts = ($br_id => \@set_cw_obscts_sbcts, ...) for ALL branches, where
#	$set_cw_obscts_sbsts[$c] is the "observed" number of substitutions (more precisely, substitutional differences)
#	along the branch $br_id in the $c th column, estimated using ALL extant residue states in c.
#
sub calc_br2cw_expcts_obscts_sbsts (\@\@\%\%\%\%\%$) {

	my ($in_msa, $seqnames, $name2id, $node2ch, $node2pa, $br2tr_mtrx, $node2ntfreqs, $gap) = @_;


			# (Pre) Retrieve the nucleotide frequencies at the root (i.e., the top node). #

	my $topnode_id = $name2id->{$TOP_NODE};
	my $ntfreqs_root = $node2ntfreqs->{$topnode_id};

			# (1) Compute the conditional probabilities of the residue states of the "descendant side." #

	my ($node2set_sw_cond_probs, $node2set_sw_extcond_probs)
	    = calc_prob_descresstats_cond_ancstat2 (@{$in_msa}, @{$seqnames}, %{$name2id}, %{$node2ch}, %{$br2tr_mtrx}, $gap);
	unless (defined $node2set_sw_cond_probs) {
	    print STDOUT "[ERROR 'calc_br2cw_expcts_obscts_sbsts'] The subroutine 'calc_prob_descresstats_cond_ancstat2' FAILED: $! !\n\n";
	    return (undef, undef);
	}

			# (2) Compute the joint probabilities of the residue states of the "ancestral side." #

	my ($br2set_sw_jnt_probs, $br2set_sw_extjnt_probs) 
	    = calc_joint_prob_ancresstats_and_ancstat2 (@{$in_msa}, @{$seqnames}, %{$name2id}, %{$node2ch}, %{$br2tr_mtrx}, $gap, @{$ntfreqs_root}, %{$node2set_sw_extcond_probs}) ;
	unless (defined $br2set_sw_jnt_probs) {
	    print STDOUT "[ERROR 'calc_br2cw_expcts_obscts_sbsts'] The subroutine 'calc_joint_prob_ancresstats_and_ancstat2' FAILED: $! !\n\n";
	    return (undef, undef);
	}

			# (3) Compute the probabilities of the MSA columns. #

	my $set_clm_probs = calc_clm_probs (@{$ntfreqs_root}, %{$node2set_sw_cond_probs}, $topnode_id) ;
	unless (defined $set_clm_probs) {
	    print STDOUT "[ERROR 'calc_br2cw_expcts_obscts_sbsts'] The subroutine 'calc_clm_probs' FAILED: $! !\n\n";
	    return (undef, undef);
	}

			# (4) Compute the expected numbers of substitutional differences per site. #

	my ($br2set_cw_expcts_sbsts_ave, $br2set_cw_expcts_sbsts_ue, $br2set_cw_expcts_sbsts_le)
	    = calc_br2cw_expcts_sbsts_ave (%{$br2tr_mtrx}, %{$br2set_sw_jnt_probs}, %{$node2set_sw_cond_probs}, %{$node2ntfreqs}, %{$node2pa});
	unless (defined $br2set_cw_expcts_sbsts_ave) {
	    print STDOUT "[ERROR 'calc_br2cw_expcts_obscts_sbsts'] The subroutine 'calc_br2cw_expcts_sbsts_ave' FAILED: $! !\n\n";
	    return (undef, undef);
	}

			# (5) Compute the "observed" numbers of substitutional differences per site. #

	my $br2set_cw_obscts_sbsts = calc_br2cw_obscts_sbsts (%{$br2tr_mtrx}, %{$br2set_sw_jnt_probs}, %{$node2set_sw_cond_probs}, @{$set_clm_probs}) ;
	unless (defined $br2set_cw_obscts_sbsts) {
	    print STDOUT "[ERROR 'calc_br2cw_expcts_obscts_sbsts'] The subroutine 'calc_br2cw_obscts_sbsts' FAILED: $! !\n\n";
	    return (undef, undef);
	}

			# (Fin) Return the output data. #

	return ($br2set_cw_expcts_sbsts_ave, $br2set_cw_obscts_sbsts);

}	# End of the subroutine 'calc_br2cw_expcts_obscts_sbsts'. #


# calc_pv_obscts_sbsts_wd_br:
#	This subroutine calculates the expected and "observed" numbers of
#	substitutional diffferences along a specified branch in a specified windown of MSA columns.
#	Then, it calculates the P-value for the "observed" difference in the light of the expected difference.
#	[CREATED on May 6 (Fri), 2016.]
#
#	[Formula]
#	The expected and "observed" differences in a window (W, with size |W|) along a branch (b),
#	denoted as  D_exp (b, W) and D_obs (b, W), respectively, are just the summations of
#	the corresponding column-wise differences over the columns belonging to the window.
#	The P-value is calculated under the binomial distribution with |W| trials:
#
#	P[X >= D_obs (b, W)] = Sum_{x=D_obs (b, W)}^{|W|} |W|!/[x!(|W|-x)!] * (p_D)^x * (1-p_D)^(|W|-x) ,
#
#	where p_D =: D_exp(b, W)/|W| is the "success" probability.
#
# INPUT: (\@window, $br_id, \%br2set_cw_expcts_sbsts, \%br2set_cw_obscts_sbsts), where
#
#	@window stores the column numbers belonging to the specified window.
#
#	$br_id is the ID of the specified branch.
#
#	%br2set_cw_expcts_sbsts = ($br_id => \@set_cw_expcts_sbcts, ...) for ALL branches, where
#	$set_cw_expcts_sbsts[$s] is the expected number of substitutions (more precisely, substitutional differences) along the branch $br_id in the $s th column, calculated in some way.
#
#	%br2set_cw_obscts_sbsts = ($br_id => \@set_cw_obscts_sbcts, ...) for ALL branches, where
#	$set_cw_obscts_sbsts[$c] is the "observed" number of substitutions (more precisely, substitutional differences)
#	along the branch $br_id in the $c th column, estimated in some way.
#
# OUTPUT: ($size_wd, $d_exp_wd_br, $d_obs_wd_br, $pv_wd_br), where
#
#	$size_wd is the number of columns in the window.
#
#	$d_exp_wd_br is the expected number of substitutional differences in the window (and along the branch);
#	$d_obs_wd_br is the "observed" number of substitutional differences in the window (and along the branch);
#	$pv_wd_br is the P-value calculated for the window (and along the branch).
#
sub calc_pv_obscts_sbsts_wd_br (\@$\%\%) {

	my ($window, $br_id, $br2set_cw_expcts_sbsts, $br2set_cw_obscts_sbsts) = @_;

	my $size_wd = @{$window};

	my $cw_expcts_sbsts_br = $br2set_cw_expcts_sbsts->{$br_id};
	my $cw_obscts_sbsts_br = $br2set_cw_obscts_sbsts->{$br_id};

	my $d_exp_wd_br = my $d_obs_wd_br = 0;

	foreach my $clm (@{$window}) { # Main foreach-loop (over columns in the window). #

	    $d_exp_wd_br += $cw_expcts_sbsts_br->[$clm];
	    $d_obs_wd_br += $cw_obscts_sbsts_br->[$clm];

	}			# End of the main foreach-loop (over columns in the window). #

	my $p_D = $d_exp_wd_br/$size_wd;

	my $int_d_obs = int ($d_obs_wd_br);
	my $frac_d_obs = $d_obs_wd_br - $int_d_obs;

	my $cmpl_pv1 = ($d_obs_wd_br<1) ? 0 : binomial_cum_via_betai ($size_wd, $int_d_obs -1, $p_D); # Prob (X >= $int_d_obs)
	my $cmpl_pv2 = binomial_cum_via_betai ($size_wd, $int_d_obs, $p_D); # Prob (X >= $int_d_obs + 1)

	my $pv_wd_br = (1 - $frac_d_obs) * (1 - $cmpl_pv1) + $frac_d_obs * (1 - $cmpl_pv2);


	return ($size_wd, $d_exp_wd_br, $d_obs_wd_br, $pv_wd_br);

}	# End of the subroutine 'calc_pv_obscts_sbsts_wd'. #



# calc_modpv_obscts_sbsts_wd_br:
#	This modified version of the subroutine, "calc_pv_obscts_sbsts_wd_br,"
#	calculates the MODIFIED expected and and "observed" numbers of
#	substitutional differences. Then, it calculates the P-value for the "observed" difference
#	in the light of the MODIFIED expected difference.
#	[CREATED on May 6 (Fri), 2016.]
#
#	[Formula]
#	The MODIFIED expectd difference, denoted as D_mexp (b, W), is calculated as:
#
#	D_mexp (b, W) =: D_exp (b, W) * [Sum_{b': all branches except b} D_obs (b', W)] / [Sum_{b': all branches except b} D_exp (b', W)].
#
#	The P-value is calculated in nearly the same way as in 'calc_pv_obscts_sbsts_wd_br',
#	with p_D now given by D_mexp (b, W) /|W|.
#
#
# calc_pv_obscts_sbsts_wd_br:
#	This subroutine calculates the expected and "observed" numbers of
#	substitutional diffferences along a specified branch in a specified windown of MSA columns.
#	Then, it calculates the P-value for the "observed" difference in the light of the expected difference.
#	[CREATED on May 6 (Fri), 2016.]
#
#	[Formula]
#	The expected and "observed" differences in a window (W, with size |W|) along a branch (b),
#	denoted as  D_exp (b, W) and D_obs (b, W), respectively, are just the summations of
#	the corresponding column-wise differences over the columns belonging to the window.
#	The P-value is calculated under the binomial distribution with |W| trials:
#
#	P[X >= D_obs (b, W)] = Sum_{x=D_obs (b, W)}^{|W|} |W|!/[x!(|W|-x)!] * (p_D)^x * (1-p_D)^(|W|-x) ,
#
#	where p_D =: D_exp(b, W)/|W| is the "success" probability.
#
#
# INPUT: (\@window, $br_id, \%br2set_cw_expcts_sbsts, \%br2set_cw_obscts_sbsts, \@cw_sumexpcts_sbsts, \@cw_sumobscts_sbsts), where
#
#	@window stores the column numbers belonging to the specified window.
#
#	$br_id is the ID of the specified branch.
#
#	%br2set_cw_expcts_sbsts = ($br_id => \@set_cw_expcts_sbcts, ...) for ALL branches, where
#	$set_cw_expcts_sbsts[$s] is the expected number of substitutions (more precisely, substitutional differences) along the branch $br_id in the $s th column, calculated in some way.
#
#	%br2set_cw_obscts_sbsts = ($br_id => \@set_cw_obscts_sbcts, ...) for ALL branches, where
#	$set_cw_obscts_sbsts[$c] is the "observed" number of substitutions (more precisely, substitutional differences)
#	along the branch $br_id in the $c th column, estimated in some way.
#
#	$cw_sumexpcts_sbsts[$c] = Sum_{$br_id: all branches} $br2set_cw_expcts_sbsts{$br_id}->[$c].
#
#	$cw_sumobscts_sbsts[$c] = sum_{$br_id; all branches} $br2set_cw_obscts_sbsts{$br_id}->[$c].
#
#
# OUTPUT: ($size_wd, $d_exp_wd_br, $d_obs_wd_br, $sumd_exp_wd, $sumd_obs_wd, $d_mexp_wd_br, $modpv_wd_br), where
#
#	$size_wd is the number of columns in the window.
#
#	$d_exp_wd_br is the expected number of substitutional differences in the window (and along the branch);
#	$d_obs_wd_br is the "observed" number of substitutional differences in the window (and along the branch);
#
#	$sumd_exp_wd is the summation of the expected number of substitutional differences in the window over all branches;
#	$sumd_obs_wd is the summation of the "observed" number of substitutional differences in the window over all branches;
#	$d_mexp_wd_br is the MODIFIED expected number of substitutional differences in the window (and along the branch).
#	More precisely, $d_mexp_wd_br =: $d_exp_wd_br * ($sumd_obs_wd - $d_obs_wd_br)/($sumd_exp_wd - $d_exp_wd_br).
#
#	$modpv_wd_br is the P-value calculated for the window (and along the branch), using the MODIFIED expected difference.
#
sub calc_modpv_obscts_sbsts_wd_br (\@$\%\%\@\@) {

	my ($window, $br_id, $br2set_cw_expcts_sbsts, $br2set_cw_obscts_sbsts, $cw_sumexpcts_sbsts, $cw_sumobscts_sbsts) = @_;

	my $size_wd = @{$window};

	my $cw_expcts_sbsts_br = $br2set_cw_expcts_sbsts->{$br_id};
	my $cw_obscts_sbsts_br = $br2set_cw_obscts_sbsts->{$br_id};

	my $d_exp_wd_br = my $d_obs_wd_br = 0;
	my $sumd_exp_wd = my $sumd_obs_wd = 0; # ADDED in creation of 'calc_modpv_obscts_sbsts_wd_br'. #

	foreach my $clm (@{$window}) { # Main foreach-loop (over columns in the window). #

	    $d_exp_wd_br += $cw_expcts_sbsts_br->[$clm];
	    $d_obs_wd_br += $cw_obscts_sbsts_br->[$clm];

		# ADDED in creation of 'calc_modpv_obscts_sbsts_wd_br'. #
	    $sumd_exp_wd += $cw_sumexpcts_sbsts->[$clm];
	    $sumd_obs_wd += $cw_sumobscts_sbsts->[$clm];

	}			# End of the main foreach-loop (over columns in the window). #

	my $d_mexp_wd_br = $d_exp_wd_br * ($sumd_obs_wd - $d_obs_wd_br) / ($sumd_exp_wd - $d_exp_wd_br); # ADDED in creation of 'calc_modpv_obscts_sbsts_wd_br'. #


		# MODIFIED in creation of 'calc_modpv_obscts_sbsts_wd_br'. #
	my $p_D = $d_mexp_wd_br/$size_wd;
	##my $p_D = $d_exp_wd_br/$size_wd;

	my $int_d_obs = int ($d_obs_wd_br);
	my $frac_d_obs = $d_obs_wd_br - $int_d_obs;

	my $cmpl_pv1 = ($d_obs_wd_br<1) ? 0 : binomial_cum_via_betai ($size_wd, $int_d_obs -1, $p_D); # Prob (X >= $int_d_obs)
	my $cmpl_pv2 = binomial_cum_via_betai ($size_wd, $int_d_obs, $p_D); # Prob (X >= $int_d_obs + 1)

		# RENAMED in creation of 'calc_modpv_obscts_sbsts_wd'. #
	my $modpv_wd_br = (1 - $frac_d_obs) * (1 - $cmpl_pv1) + $frac_d_obs * (1 - $cmpl_pv2);
	##my $pv_wd_br = (1 - $frac_d_obs) * (1 - $cmpl_pv1) + $frac_d_obs * (1 - $cmpl_pv2);


	return ($size_wd, $d_exp_wd_br, $d_obs_wd_br, $sumd_exp_wd, $sumd_obs_wd, $d_mexp_wd_br, $modpv_wd_br);

}	# End of the subroutine 'calc_modpv_obscts_sbsts_wd'. #


# mk_node2sw_gstats:
#	This satellite subroutine of 'detect_purge_cands'
#	constructs a hash table associating each node with the site-wise gap states
#	according to an input ancestral state that is phylogenetically compatible with an entire MSA.
#	[CREATION STARTED on Jan 9, 2018, ENDED on Jan dd, 2018.]
#
# INPUT: (\@info_indel_hst, $char_abs, $char_pres, $beg_msa, $end_msa, \@set_all_nodes), where
#
#	@info_indel_hst = (\@set_cls_gblocks, \@totcts_indels_for_cls, \@indel_hst_for_cls),
#	gives information on an indel history giving rise to the input MSA gap configuration, where
#	
#	@{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end) is for the $gb th block in the $cl th cluster (i.e., gapped segment) in the input MSA, with
#		$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#		$len_gb being the length (in terms of #{columns}) of the block,
#		and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention),
#	
#	$totcts_indels_for_cls[$cl] is the total number of insertion/deletion events in the input indel history resulting in the $cl th cluster (i.e., gapped segment) in the input MSA,
#	
#	@{$indel_hst_for_cls[$cl]} = (\%br2indels, \%node2blwise_gstat) is for the input indel history resulting in the $cl th cluster (i.e., gapped segment) in the input MSA, where 
#		%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#		@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved), and where
#		%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
#		  with $blwise_gapstats[$bl] is the gap-state ($char_pres/abs) of the $bl th block at the node.
#
#	$char_pres/abs is the character representing the "presence"/"absence" state.
#
#	$beg/end_msa is the beginning/ending coordinates of the input MSA (in the 0-based full-closed convention).
#
#	@set_all_nodes stores the IDs of all nodes in the input phylogenetic tree.
#
#
# OUTPUT: \%node2sw_gstats, where
#	%node2sw_gstats = ($node_id => \@sw_gstats, ...) for all nodes, where
#	$sw_gstats[$s] is the gap state of the $s th site on the node $node_id,
#	which is totally consistent with the input indel history.
#
sub mk_node2sw_gstats (\@$$$$\@) {

	my ($info_indel_hst, $char_abs, $char_pres, $beg_msa, $end_msa, $set_all_nodes) = @_;

	my ($set_cls_gblks, $totcts_indels_for_cls, $indel_hst_for_cls) = @{$info_indel_hst};
	my $ct_cls = @{$set_cls_gblks};


	##my $beg_lm_cls = $set_cls_gblks->[0][0][2]; # A Fatal BUG Fixed on 2020/03/29.
	my $beg_lm_cls = ($ct_cls > 0) ? $set_cls_gblks->[0][0][2] : $end_msa +1;


		# (Preliminary) Initialize the output hash table using the left-most gapless segment. #

	my %node2sw_gstats;
	foreach my $node (@{$set_all_nodes}) { # Preliminary outer foreach-loop (over all nodes). #
	    my @sw_gstats = ();
	    for (my $s = $beg_msa; $s < $beg_lm_cls; $s++) { $sw_gstats[$s] = $char_pres; }
	    $node2sw_gstats{$node} = \@sw_gstats;
	}				# End of the preliminary outer foreach-loop (over all nodes). #

	for (my $cl = 0; $cl < $ct_cls; $cl++) { # Main outer for-loop (over gapped segments). #

	    my $gblks_in_cl = $set_cls_gblks->[$cl];   
	    my $ct_gblks = @{$gblks_in_cl};

	    my $end_cl = $gblks_in_cl->[$ct_gblks-1][3];
	    my $beg_next_cl = ($cl < $ct_cls-1) ? $set_cls_gblks->[$cl+1][0][2] : $end_msa + 1; 

	    my ($br2indels, $node2bw_gstats) = @{$indel_hst_for_cls->[$cl]};

	    foreach my $node (@{$set_all_nodes}) { # Middle foreach-loop (over all nodes). #

		my $sw_gstats = $node2sw_gstats{$node};
		my $bw_gstats = $node2bw_gstats->{$node};

			# Site-wise gap states for the gapped segment. #
		for (my $bl = 0; $bl<$ct_gblks; $bl++) { # Inner for-loop (over gap-pattern blocks). #
		    my $gstat = $bw_gstats->[$bl];
		    my ($beg_bl, $end_bl) = @{$gblks_in_cl->[$bl]}[2,3];
		    $end_bl++;
		    for (my $s = $beg_bl; $s < $end_bl; $s++) { $sw_gstats->[$s] = $gstat; }
		}				# End of the inner for-loop (over gap-pattern blocks). #

			# Site-wise gap states for the right-flanking gapless segment. #
		for (my $s = $end_cl+1; $s < $beg_next_cl; $s++) { $sw_gstats->[$s] = $char_pres; }

	    }				# End of the middle foreach-loop (over all nodes). #

	}				# End of the main outer for-loop (over gapped segments). #

	return \%node2sw_gstats;

}	# End of the subroutine 'mk_node2sw_gstats'. #


	### The MAIN Subroutines of this Perl module. ###

# detect_purge_cands:
#	This subroutine detects candidates of purge errors via a sliding window analysis.
#
#	[CREATION STARTED on Jan 9, 2018, ENDED on Jan dd, 2018.]
#
# [ INPUT: ($min_wdsize, $max_wdsize, $thrsh_pv, $if_modpv, $char_abs, $char_pres, $topnode_id, \%node2pa, \%node2ntfreqs, \%br2tr_mtrx, \%br2set_sw_jnt_probs, \%node2set_sw_cond_probs, \@set_clm_probs, \@info_psm_hst0, $beg_msa, $end_msa, \%br_attr, \%br2equiv), where ] ... OBSOLATE as of Jan 13 (Sat), 2018. #
# INPUT: ($min_wdsize, $max_wdsize, $thrsh_pv, $thrsh_pv0, $if_modpv, $char_abs, $char_pres, $topnode_id, \%node2pa, \%node2ntfreqs, \%br2tr_mtrx, \%br2set_sw_jnt_probs, \%node2set_sw_cond_probs, \@set_clm_probs, \@info_psm_hst0, $beg_msa, $end_msa, \%br_attr, \%br2equiv), where
#
#	$min/max_wdsize specifies the lower/upper-bound of the window size (for the detection of purge candidates);
#	$thrsh_pv is the P-value threshold to determine the purge candidates (based on the substitution model);
#	$thrsh_pv0 is the P-value threshold to determine the purge candidates (based on the null random-match model);
#	$if_modpv = 1 if you use the modified P-value, = 0 if you use the normal P-value;
#	$char_abs/pres is the character that represents an "absence"/"presence" state.
#
#	%node2pa = ($node_id => $parent_id, ...) for the nodes except the root.
#
#	%node2ntfreqs = ($node_id => \@ntfreqs, ...), where
#	$ntfreqs[$i] is the frequency of $NUCLEOTIDES[$i] at the node, $node_id.
#
#	@{$br2tr_mtrx{$br_id}} is the transition matrix
#	that describes the transition probabilities between both ends of the branch $br_id.
#	More specifically,
#	$br2tr_mtrx{$br_id}->[$i][$j] is the probability that the node $br_id shows $NUCLEOTIDES[$j]
#	given that the parent node of $br_id shows $NUCLEOTIDES[$i].
#
#	%br2set_sw_jnt_probs = ($br_id => \@set_sw_jnt_probs, ...) for ALL branches, where 
#	$set_sw_jnt_probs[$s]->[$i] is the probability that
#	all the extant sequences on the "ancestral side" of $br_id show the residue states that are actually found in the $s th site,
#	in addition to that $NUCLEOTIDES[$i] occupies the $s th site of the UPPER-end node of $br_id.
#	(It is an output of 'calc_joint_prob_ancresstats_and_ancstat2' in 'MyLikelihoods_respattern_msa_dna.pm'.)
#
#	%node2set_sw_cond_probs = ($node_id => \@set_sw_cond_probs, ...) for ALL nodes, where
#	$set_sw_cond_probs[$s]->[$i] is the probability that
#	all the descendant sequences of $node_id show the residue states that are actually found in the $s th site,
#	given that $NUCLEOTIDES[$i] occupies the $s th site of the ancestral node, $node_id.
#	But $set_sw_cond_probs[$s] = [] (empty) if $node_id refers to an external node and if the node has its $s th site occupied by a gap.
#	(It is an output of 'calc_prob_descresstats_cond_ancstat2' in 'MyLikelihoods_respattern_msa_dna.pm'.)
#
#	$set_clm_probs[$c] is the occurrence probability of the $c th column
#	under the given substitution model.
#	(It is an output of 'calc_clm_probs' in 'MyLikelihoods_respattern_msa_dna.pm'.)
#
#	@info_psm_hst0 = (\@set_cls_gblocks, \@totcts0_indels_for_cls, \@psm_hst0_for_cls),
#	which is an output of 'cal_init_psm_cands_for_sgl_msa' in 'MyTreeMap_indels_spt_odr2.pm',
#	gives information on the Dollo parsimonious indel history giving rise to the input MSA gap configuration, where
#	
#	@{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end) is for the $gb th block in the $cl th cluster (i.e., gapped segment) in the input MSA, with
#		$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#		$len_gb being the length (in terms of #{columns}) of the block,
#		and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention),
#	
#	$totcts0_indels_for_cls[$cl] is the total number of insertion/deletion events in the Dollo parsimonious indel history resulting in the $cl th cluster (i.e., gapped segment) in the input MSA,
#	
#	@{$psm_hst0_for_cls[$cl]} = (\%br2indels0, \%node2blwise_gstat0) is for the Dollo parsimonious indel history resulting in the $cl th cluster (i.e., gapped segment) in the input MSA, where 
#		%br2indels0 = ($branch_id => \@list_indels0, ...) is the map of insertions/deletions onto branches, with
#		@{$list_indels0[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved), and where
#		%node2blwise_gstat0 = ($node_id => \@blwise_gapstats0, ...) is the map of the gap states onto the nodes,
#		  with $blwise_gapstats0[$bl] is the gap-state ($char_pres/abs) of the $bl th block at the node.
#
#	$beg/end_msa is the beginning/ending coordinates of the input MSA (in the 0-based full-closed convention).
#
#	%br_attr = ($br_id => \@attr, ...) for all branches, where $attr[$INDX_LENGTH] is the branch length.
#
#	%br2equiv = ($br_id => $equiv_br_id, ...) for those branches with equivalents.
#
# OUTPUT: \%br2purge_cands, where
#
#	%br2purge_cands = ($br_id => \@list_purge_cands, ...), where
#	@{$list_purge_cands[$pc]} = ($beg_pc, $end_pc)
#	provide the beginning and ending coordinates (in the 0-based half-open convention)
#	of the $pc th purge candidate (from left to right) along the branch $br_id.
#
sub detect_purge_cands ($$$$$$$$\%\%\%\%\%\@\@$$\%\%) {


	my ($min_wdsize, $max_wdsize, $thrsh_pv, $thrsh_pv0, $if_modpv, $char_abs, $char_pres, 
	   $topnode_id, $node2pa, $node2ntfreqs, $br2tr_mtrx, $br2set_sw_jnt_probs, $node2set_sw_cond_probs,
	    $set_clm_probs, $info_psm_hst0, $beg_msa, $end_msa, $br_attr, $br2equiv) = @_;

   (print STDOUT "< 'detect_purge_cands' ($min_wdsize <= {window size} <= $max_wdsize; absence/presence: '$char_abs'/'$char_pres'; {P-value threshold}_sbstmodel = $thrsh_pv; {P-value threshold}_random = $thrsh_pv0; ", '$if_modpv = ', "$if_modpv; $beg_msa <= {MSA coordinate} <= $end_msa) >\n\n") if ($TEST_LEVEL>0);

	my @set_all_nodes = sort {$a <=> $b} keys %{$node2pa};
	push @set_all_nodes, $topnode_id;
	my @set_all_branches = sort {$a <=> $b} keys %{$br_attr};

   if ($TEST_LEVEL>2) {
	print STDOUT "[ Secondary inputs ]\n\n";
	print STDOUT '@set_all_nodes = (', join (',', @set_all_nodes), ").\n\n";
	print STDOUT '@set_all_branches = (', join (', ', @set_all_branches), ").\n\n\n";
   }

		# Perform 'calc_br2cw_expcts_sbsts_ave' & 'calc_br2cw_obscts_sbsts', 
		# using the outputs of 'calc_prob_descresstats_cond_ancstat2',
		# 'calc_joint_prob_ancresstats_and_ancstat2' & 'calc_clm_probs'.
		# (IMPORTED from the subroutine, 'calc_br2cw_expcts_obscts_sbsts' on Jan 10, 2018.)


			# (Pre-1) Compute the expected numbers of substitutional differences per site. #

	my ($br2set_cw_expcts_sbsts_ave, $br2set_cw_expcts_sbsts_ue, $br2set_cw_expcts_sbsts_le)
	    = calc_br2cw_expcts_sbsts_ave (%{$br2tr_mtrx}, %{$br2set_sw_jnt_probs}, %{$node2set_sw_cond_probs}, %{$node2ntfreqs}, %{$node2pa});
	unless (defined $br2set_cw_expcts_sbsts_ave) {
	    print STDOUT "[ERROR 'detect_purge_cands'] The subroutine 'calc_br2cw_expcts_sbsts_ave' FAILED: $! !\n\n";
	    return undef;
	}
   (print STDOUT " --- (Pre-1) Subroutine 'calc_br2cw_expcts_sbsts_ave' ended ...\n\n") if ($TEST_LEVEL>1);

   if ($TEST_LEVEL>3) {
	print STDOUT "[ Content of ", '%{$br2set_cw_expcts_sbsts_ave}', ", which is the main output of 'calc_br2cw_expcts_sbsts_ave' ]\n\n";

	print STDOUT join ("\t", 'Column\Branch', @set_all_branches), "\n\n";

	for (my $c = $beg_msa; $c <= $end_msa; $c++) { # Outer for-loop (over columns). #
	    my @cw_expcts_along_brs = ();
	    foreach my $br (@set_all_branches) {
		my $cw_expct = $br2set_cw_expcts_sbsts_ave->{$br}[$c];
		(defined $cw_expct) or ($cw_expct = 'undef');
		push @cw_expcts_along_brs, $cw_expct;
	    }
	    print STDOUT join ("\t", $c, @cw_expcts_along_brs), "\n";

	} # End of outer for-loop (over columns). #

	print STDOUT "\n\n";

=Alternative idea
	foreach my $br (@set_all_branches) {
	    my $set_cw_expcts = $br2set_cw_expcts_sbsts_ave->{$br};
	    my $ct_clms = @{$set_cw_expcts};
	    print STDOUT "{ Branch ID: $br (w/ $ct_clms columns) }\n\n";
	    print STDOUT join ("\t", 'Column', 'Exp#{substitutions}_ave'), "\n\n";
	    for (my $c = 0; $c < $ct_clms; $c++) { print STDOUT join ("\t", $c, $set_cw_expcts->[$c]), "\n"; }
	    print STDOUT "\n\n";
	}
=cut
   } # END of "if ($TEST_LEVEL>3) {...}"

			# (Pre-2) Compute the "observed" numbers of substitutional differences per site. #

	my $br2set_cw_obscts_sbsts = calc_br2cw_obscts_sbsts (%{$br2tr_mtrx}, %{$br2set_sw_jnt_probs}, %{$node2set_sw_cond_probs}, @{$set_clm_probs}) ;
	unless (defined $br2set_cw_obscts_sbsts) {
	    print STDOUT "[ERROR 'detect_purge_cands'] The subroutine 'calc_br2cw_obscts_sbsts' FAILED: $! !\n\n";
	    return undef;
	}
   (print STDOUT " --- (Pre-2) Subroutine 'calc_br2cw_obscts_sbsts' ended ...\n\n") if ($TEST_LEVEL>1);

   if ($TEST_LEVEL>3) {
	print STDOUT "[ Content of ", '%{$br2set_cw_obscts_sbsts}', ", which is the output of 'calc_br2cw_obscts_sbsts' ]\n\n";

	print STDOUT join ("\t", 'Column\Branch', @set_all_branches), "\n\n";

	for (my $c = $beg_msa; $c <= $end_msa; $c++) { # Outer for-loop (over columns). #
	    my @cw_obscts_along_brs = ();
	    foreach my $br (@set_all_branches) {
		my $cw_obsct = $br2set_cw_obscts_sbsts->{$br}[$c];
		(defined $cw_obsct) or ($cw_obsct = 'undef');
		push @cw_obscts_along_brs, $cw_obsct;
	    }
	    print STDOUT join ("\t", $c, @cw_obscts_along_brs), "\n";

	} # End of outer for-loop (over columns). #

	print STDOUT "\n\n";

=Alternative idea
	foreach my $br (@set_all_branches) {
	   my $set_cw_obscts = $br2set_cw_obscts_sbsts->{$br};
	   my $ct_clms = @{$set_cw_obscts};
	   print STDOUT "{ Branch ID: $br (w/ $ct_clms columns) }\n\n";
	   print STDOUT join ("\t", 'Column', 'Obs#{substitutions}'), "\n\n";
	   for (my $c=0; $c < $ct_clms; $c++) { print STDOUT join ("\t", $c, $set_cw_obscts->[$c]), "\n"; }
	   print STDOUT "\n\n";
	}
=cut
   } # END of "if ($TEST_LEVEL>3) {...}"


		# END of "(IMPORTED from the subroutine, 'calc_br2cw_expcts_obscts_sbsts' on Jan 10, 2018.)"

			# (Pre-trailing) Sum the expected counts over the branches, and also sum the 'observed' counts over the branches. #
	my ($cw_sumexpcts_sbsts, $cw_sumobscts_sbsts);
	if ($if_modpv) {
	    $cw_sumexpcts_sbsts = sum_arrays_over_branches (%{$br2set_cw_expcts_sbsts_ave});
	    unless (defined $cw_sumexpcts_sbsts) {
		print STDOUT "[ERROR 'detect_purge_cands'] The subroutine 'sum_arrays_over_branches' FAILED for expected counts of substitutions: $! !\n\n";
		return undef;
	    }
	    $cw_sumobscts_sbsts = sum_arrays_over_branches (%{$br2set_cw_obscts_sbsts});
	    unless (defined $cw_sumobscts_sbsts) {
		print STDOUT "[ERROR 'detect_purge_cands'] The subroutine 'sum_arrays_over_branches' FAILED for 'observed' counts of substitutions: $! !\n\n";
		return undef;
	    }
   (print STDOUT " --- (Pre-trailing) Subtoutine 'sum_arrays_over_branches' ended to calculate total expected & 'observed' counts of substitutions ...\n\n") if ($TEST_LEVEL>1);

   if ($TEST_LEVEL>3) {
	print STDOUT '[ Contents of @{$cw_sumexpcts_sbsts} & @{$cw_sumobscts_sbsts} ]', "\n\n";
	print STDOUT join ("\t", 'Column', 'Total[Exp#{sbsts}]', 'Total[Obs#{sbsts}]'), "\n\n";
	for (my $c = $beg_msa; $c <= $end_msa; $c++) {
	    my $sumexpct = $cw_sumexpcts_sbsts->[$c];
	    (defined $sumexpct) or ($sumexpct = 'undef');
	    my $sumobsct = $cw_sumobscts_sbsts->[$c];
	    (defined $sumobsct) or ($sumobsct = 'undef');
	    print STDOUT join ("\t", $c, $sumexpct, $sumobsct), "\n";
	}
	print STDOUT "\n\n";
   } # END  of "if ($TEST_LEVEL>3) {...}"

	} # END of "if ($if_modpv) {...}"


			# (Pre-3) Prepare the site-wise gap states (of the Dollo-parsimonious indel history) at every node. #

	# %{$node2sw_gstats0} = ($node_id => \@sw_gstats0, ...) for all nodes, where
	# $sw_gstats0[$s] is the gap state of the $s th site on the node $node_id
	# (usually according to the Dollo-parsimony principle).
	#
	my $node2sw_gstats0 = mk_node2sw_gstats (@{$info_psm_hst0}, $char_abs, $char_pres, $beg_msa, $end_msa, @set_all_nodes);
	unless (defined $node2sw_gstats0) {
	    print STDOUT "[ERROR 'detect_purge_cands'] The subroutine 'mk_node2sw_gstats' FAILED: $! !\n\n";
	    return undef;
	}
   (print STDOUT " --- (Pre-3) Subroutine 'mk_node2sw_gstats' ended ...\n\n") if ($TEST_LEVEL>1);

   if ($TEST_LEVEL>3) {
	print STDOUT "[ Content of ", '%{$Node2sw_gstats0}', ", which is the output of 'mk_node2sw_gstats' ]\n\n";
	print STDOUT 'Site\Node', "\t", join (', ', @set_all_nodes), "\n\n";
	for (my $s = $beg_msa; $s <= $end_msa; $s++) { # Outer for-loop (over sites). #
	    my @sw_gstats_on_nodes = ();
	    foreach my $node (@set_all_nodes) {
		my $gstat = $node2sw_gstats0->{$node}[$s];
		(defined $gstat) or ($gstat = '?');
		push @sw_gstats_on_nodes, $gstat;
	    }
	    print STDOUT $s, "\t", join (', ', @sw_gstats_on_nodes), "\n";

	}				# End of the outer for-loop (over sites). #

	print STDOUT "\n\n";

   } # END of "if ($TEST_LEVEL>3) {...}"


			# (MAIN) Sliding-window analysis to detect the candidates of purge-type errors. #
			# (Created in refercence to ll.2835-2916 of 'examine_purge_ctrls.2.pl'.) #

   (print STDOUT "[ Main Process: Sinding-window analysis to detect 'purge candidates' ]\n\n") if ($TEST_LEVEL>1);

	my $rflank_site_msa = $end_msa+1;
	my %br2purge_cands;

	foreach my $br (@set_all_branches) { # Main outer foreach-loop (over all branches) #

	    my $eq_br = $br2equiv->{$br};
	    my $brlen = $br_attr->{$br}[$INDX_LENGTH];
	    if ((defined $eq_br) and ($brlen == 0)) { next; } # Skip the null artificial branch. #

	    my $node_ue = $node2pa->{$br};
	    my $sw_gstats0_le = $node2sw_gstats0->{$br};
	    my $sw_gstats0_ue = $node2sw_gstats0->{$node_ue};

			# ADDED on Jan 13 (Sat), 2018. #
	    my $ntfreqs_le = $node2ntfreqs->{$br};
	    my $ntfreqs_ue = $node2ntfreqs->{$node_ue};
	    my $ct_nt = @{$ntfreqs_le};
	    my $r_mch0 = 0;
	    for (my $n = 0; $n <$ct_nt; $n++) { $r_mch0 += $ntfreqs_le->[$n] * $ntfreqs_ue->[$n]; }
			# END of "ADDED on Jan 13 (Sat), 2018." #

   (print STDOUT ' ',  "{ Branch ID: $br (upper-end: $node_ue ; rate{random-match} = $r_mch0) }\n\n") if ($TEST_LEVEL>2); # MODIFIED on Jan 13 (Sat), 2018. # 
   ##(print STDOUT ' ',  "{ Branch ID: $br (upper-end: $node_ue) }\n\n") if ($TEST_LEVEL>2);

	    my @joint_purge_cands = (); # element = [$beg, $end] of a candidate region. 

	    for (my $wsize = $min_wdsize; $wsize <= $max_wdsize; $wsize++) { # Middle for-loop (over the window sizes). #

   (print STDOUT ' ' x 3, "{ Window size: $wsize }", "\n\n") if ($TEST_LEVEL>3);

	        my @purge_cands = (); # element = [$beg, $end] of a candidate region. 
		my @window = ();
		my $if_lf_filled = 0;

		for (my $s = $beg_msa; $s < $rflank_site_msa; $s++) { # Inner for-loop (over the sites). #
		    my $gstat0_le = $sw_gstats0_le->[$s];
		    my $gstat0_ue = $sw_gstats0_ue->[$s];
		    if ($gstat0_le eq $char_abs) {
			if ($gstat0_ue eq $char_abs) {
			    next; # Just skip the site. #
			} else {
			    @window = (); # Dump the growing window, and re-start to grow the window. #
			    $if_lf_filled = 0;
			    next;
			}
		    } elsif ($gstat0_ue eq $char_abs) {
			@window = (); # Dump the growing window, and re-start to grow the window. #
			$if_lf_filled = 0;
			next;
		    }

		    unless ($if_lf_filled) { # If the left-flanking site is NOT filled with "presence", make it the "left-flanking site" of the subsequent window. #
			$if_lf_filled = 1;
			next;
		    }

		    push @window, $s;

		    (@window == $wsize) or next;

			# The window is fully grown. #
			# => Check the "right-flanking" site. #

		    my $if_rf_filled = 0;
		    for (my $s2 = $s+1; $s2 < $rflank_site_msa; $s2++) { # Inner-inner for-loop (over the sites on the right of $s). #
			my $gstat0_le2 = $sw_gstats0_le->[$s2];
			my $gstat0_ue2 = $sw_gstats0_ue->[$s2];
			next if (($gstat0_le2 eq $char_abs) and ($gstat0_ue2 eq $char_abs));
			($if_rf_filled = 1) if (($gstat0_le2 ne $char_abs) and ($gstat0_ue2 ne $char_abs));
			last;
		    }					# End of the inner-inner for-loop (over the sites on the right of $s). #

		    if ($if_rf_filled == 0) { # If the right-flanking site is NOT filled with "presence", dump the window.
			@window = 0;
			$if_lf_filled = 0;
			next;
		    }

			# Now, the window is flanked (on both sides) by sites filled with "presence". #
			# => Calculate the P-value on the residue pattern. #

   (print STDOUT ' ' x 3, '| Window: (', join (', ', @window), ') |', "\n\n") if ($TEST_LEVEL>5);

		    my ($beg_wd, $end_wd) = @window[0,$wsize-1];

		    my $pv;
		    my $d_obs; #  ADDED on Jan 13 (Sat), 2018. #

		    if ($if_modpv) {
			my ($size_wd, $d_exp_wd_br, $d_obs_wd_br, $sumd_exp_wd, $sumd_obs_wd, $d_mexp_wd_br, $modpv_wd_br)
			 = calc_modpv_obscts_sbsts_wd_br (@window, $br, %{$br2set_cw_expcts_sbsts_ave}, %{$br2set_cw_obscts_sbsts}, @{$cw_sumexpcts_sbsts}, @{$cw_sumobscts_sbsts});
		        unless (defined $modpv_wd_br) {
			    print STDOUT "[ERROR 'detect_purge_cands'] The subroutine 'calc_modpv_obscts_sbsts_wd_br' FAILED along the branch '$br' and on the window of size $wsize ([beg, end] = [$beg_wd, $end_wd]) !\n";
			    return undef;
		        }
   (print STDOUT ' ' x 3, " --- Subroutine 'calc_modpv_obscts_sbsts_wd_br' ended (w/", '$modpv_wd_br = ', "$modpv_wd_br) ...\n\n") if ($TEST_LEVEL >5);

			$pv = $modpv_wd_br;
			$d_obs = $d_obs_wd_br; # ADDED on Jan 13 (Sat), 2018. #

		    } else {
		        my ($size_wd, $d_exp, $d_obs2, $pv_wd_br) = calc_pv_obscts_sbsts_wd_br (@window, $br, %{$br2set_cw_expcts_sbsts_ave}, %{$br2set_cw_obscts_sbsts});
		        unless (defined $pv_wd_br) {
			    print STDOUT "[ERROR 'detect_purge_cands'] The subroutine 'calc_pv_obscts_sbsts_wd_br' FAILED along the branch '$br' and on the window of size $wsize ([beg, end] = [$beg_wd, $end_wd]) !\n";
			    return undef;
		        }
   (print STDOUT ' ' x 3, " --- Subroutine 'calc_pv_obscts_sbsts_wd_br' ended (w/", '$pv_wd_br = ', "$pv_wd_br) ...\n\n") if ($TEST_LEVEL >5);

			$pv = $pv_wd_br;
			$d_obs = $d_obs2; # ADDED on Jan 13 (Sat), 2018. #
		    }

				# MODIFIED on Jan 13 (Sat), 2018. #

		    unless ($pv <= $thrsh_pv) { # The number of differences is not significantly more than expected from the substitution model.
			shift @window; # A BUG fixing measure!!
			next;
		    }

		    ## my $ct_mch_obs = $wsize - $d_obs;

		    my $int_d_obs = int ($d_obs);
		    my $frac_d_obs = $d_obs - $int_d_obs;

		    my $pv01 = binomial_cum_via_betai ($wsize, $int_d_obs, 1 - $r_mch0); # Prob (X <= $int_d_obs)
		    my $pv02 = ($int_d_obs >= $wsize) ? 1 : binomial_cum_via_betai ($wsize, $int_d_obs +1, 1 - $r_mch0); # Prob (X <= $int_d_obs + 1)

		    my $pv0 = (1 - $frac_d_obs) * $pv01 + $frac_d_obs * $pv02; # The probability wht $d_obs or less "observed" differences, under the binomial distribution with $wsize trials and the random mismatch rate (1-$r_mch0).

		    if ($pv0 < $thrsh_pv0) { # The number of differences is significantly less than expected from random.
		    ##if ($pv < $thrsh_pv0) { # A BUG was found!! => Fixed....}
			shift @window; # A BUG fixing measure!!
			next;
		    }

		    push @purge_cands, [$beg_wd, $end_wd, $pv, $pv0];

		    #if ($pv <= $thrsh_pv) {
		    #	push @purge_cands, [$beg_wd, $end_wd];
		    #}

				# END of "MODIFIED on Jan 13 (Sat), 2018." #

		    shift @window; # Remove the leftmost site from the window.
		}					# End of the inner for-loop (over the sites). #

   if ($TEST_LEVEL>=5) {
	print STDOUT ' ' x 3, "{ List of 'purge candidate' windows }\n\n";

	print STDOUT ' ' x 3, join ("\t", 'No.', 'beg_wd', 'end_wd', 'pv', 'pv0'), "\n\n"; # MODIFIED on Jan 13 (Sat), 2018. #
	##print STDOUT ' ' x 3, join ("\t", 'No.', 'beg_wd', 'end_wd'), "\n\n";

	my $ct_cands = @purge_cands; 
	for (my $n = 0; $n <$ct_cands; $n++) { print STDOUT ' ' x 3, join ("\t", $n, @{$purge_cands[$n]}), "\n"; }
	print STDOUT "\n\n";
   } # END of "if ($TEST_LEVEL>=5) {...}"

		push @joint_purge_cands, @purge_cands;

	    }							# End of the middle for-loop (over the window sizes). #

			# (Trailing) Constructing a non-redundant list of the candidate regions. #

	    my $mgd_purge_cands = (@joint_purge_cands<2) ? \@joint_purge_cands : merge_integer_intervals (@joint_purge_cands, 0, 1);
	    unless (defined $mgd_purge_cands) {
		print STDOUT "[ERROR 'detect_purge_cands'] The subtroutine 'merge_integer_intervals' FAILED for the set of purge candidates along the branch '$br': $! !\n\n";
		return undef;
	    }

   (print STDOUT ' ', " --- Subroutine 'merge_integer_intervals' ended ...\n\n") if ($TEST_LEVEL>3);

   if ($TEST_LEVEL>3) {
	my $ct_mgd = @{$mgd_purge_cands};
	print STDOUT ' ', "[ List of non-redundant 'purge candidate' regions (#{regions} = $ct_mgd) ]\n\n";
	print STDOUT ' ', join ("\t", 'No', 'beg_rg', 'end_rg'), "\n\n";
	for (my $n = 0; $n <$ct_mgd; $n++) { print STDOUT ' ' , join ("\t", $n, @{$mgd_purge_cands->[$n]}), "\n"; }
	print STDOUT "\n\n";

   } # END of "if ($TEST_LEVEL>3) {...}"

	    $br2purge_cands{$br} = $mgd_purge_cands;

	}			# End of the main outer foreach-loop (over all branches). #

   (print STDOUT "\n\n< 'detect_purge_cands' Ended />\n\n") if ($TEST_LEVEL>0);

	return \%br2purge_cands;

}	# END of the subroutine 'detect_purge_cands'. #


# merge_sitesets:
#	This satellite subroutine of 'detect_purge_cands2'
#	merges the sets of sites if they overlap or are immediately adjacent.
#	[CREATED in creation of 'detect_purge_cands2' on 2019/06/15.]
#	[MAJOR REVISION in DEBUGGING on 2019/06/16.]
#
# INPUT: (\@set_sitesets), where
#
#	@{$set_sitesets[$ss]} stores the sites that constitutes the $ss th site-set.
#
# OUTPUT: (\@set_mgd_sets), where
#
#	@{$set_mgd_sets[$ms]} stores the sites that constitutes the $ms th merged set.
#
sub merge_sitesets (\@) {

	my ($set_sitesets) = @_;

	my $ct_sitesets = @{$set_sitesets};

	if ($ct_sitesets < 2) { return $set_sitesets; }

	my @std_set_sitesets = sort {$a->[0] <=> $b->[0]} @{$set_sitesets} ;

	my @cp_1st_set = @{$std_set_sitesets[0]};

		# MODIFIED in DEBUGGING (on 2019/06/16). (1) #
	my @set_mgd_sets = (\@cp_1st_set);
=DISABLED in DEBUGGING (on 2019/06/16). (1)
	my $prev_siteset = \@cp_1st_set;
	my ($beg_prev, $end_prev) = @cp_1st_set[0, $#cp_1st_set];
	my @set_mgd_sets = ($prev_siteset);
=cut # END of "DISABLED in DEBUGGING (on 2019/06/16). (1)"
		# END of "MODIFIED in DEBUGGING (on 2019/06/16). (1)" #


	for (my $ss=1; $ss <$ct_sitesets; $ss++) { # Main for-loop (over the input site-sets). #

	    my $curr_siteset = $std_set_sitesets[$ss];
	    my ($beg_curr, $end_curr) = @{$curr_siteset}[0, $#{$curr_siteset}];

			# ADDED in DEBUGGING (on 2019/06/16). (1) #
	    my $prev_siteset = $set_mgd_sets[$#set_mgd_sets];
	    my ($beg_prev, $end_prev) = @{$prev_siteset}[0, $#{$prev_siteset}];
			# END of "ADDED in DEBUGGING (on 2019/06/16). (1)" #

	    if ($end_prev+1 < $beg_curr) {
			# The currend and previous sets are separated. #

		my @cp_curr_siteset = @{$curr_siteset};
		push @set_mgd_sets, \@cp_curr_siteset;

		##($beg_prev, $end_prev) = ($beg_curr, $end_curr); # DISABLED in DEBUGGING (on 2019/06/16). (2)

	    } else {
			# The current and previous sets overlap or are immediately adjacent. #

		my $new_curr_siteset = union_nmarray (@{$prev_siteset}, @{$curr_siteset});

		
		# MODIFIED in DEBUGGING (on 2019/06/16). (2) #
		$set_mgd_sets[$#set_mgd_sets] = $new_curr_siteset; # Update the previous (i.e., last) siteset.
=DISABLED in DEBUGGING (on 2019/06/16). (3)
		$prev_siteset = $new_curr_siteset;
		($beg_prev, $end_prev) = @{$prev_siteset}[0, $#{$prev_siteset}];
=cut # END of "DISABLED in DEBUGGING (on 2019/06/16). (3)"
		# END of "MODIFIED in DEBUGGING (on 2019/06/16). (2)" #

	    } # End of "if ($end_prev+1 < $beg_curr) {...} else {...}".

	}    # End of the main for-loop (over the input site-sets). #

	return \@set_mgd_sets;

}	# End of the subroutine, 'merge_sitesets'. #


# assign_sgn_wds_to_regions:
#	This satellite subroutine of 'detect_purge_cands2'
#	assigns significant windows to the accommodating regions.
#	[CREATED on 2019/06/16.]
#
# INPUTS: (\@set_info_sgn_wds, \@set_sites_in_sgn_wds, \@set_regions), where
#
#	($beg_wd, $end_wd, $pv, $pv0) = @{$set_info_sgn_wds[$swd]} provides information on the $swd th significant window (arranged in the order they were found, NOT in the order of significance);
#	@{$set_sites_in_sgn_wds[$swd]} stores the constituent sites of the $swd th significant window;
#
#	($beg_rg, $end_rg) = @{$set_regions[$rg]} are the beginning and end of the $rg th region;
#
# OUTPUTS: \@sgn_wds_in_rgs, where
#
#	(\@sites_in_wd, $pv, $pv0) = @{$sgn_wds_in_rgs[$rg][$k]} provides information on the significant window assigned to the $rg th region in the $k th place, where
#	@sites_in_wd stores the constituent sites of the window;
#
sub assign_sgn_wds_to_regions (\@\@\@) {

	my ($set_info_sgn_wds, $set_sites_in_sgn_wds, $set_regions) = @_;

	my $ct_swds = @{$set_info_sgn_wds};
	my $ct_rgs = @{$set_regions};
	my @indices_rlv = (0 .. $ct_rgs-1);

		# Initialize the output. #
	my @sgn_wds_in_rgs = ();
	for (my $rg=0; $rg<$ct_rgs; $rg++) { $sgn_wds_in_rgs[$rg] = []; }

	for (my $swd=0; $swd < $ct_swds; $swd++) { # Main for-loop (over the significant windows). #

	    my ($beg_wd, $end_wd, $pv, $pv0) = @{$set_info_sgn_wds->[$swd]};
	    my $sites_in_wd = $set_sites_in_sgn_wds->[$swd];

	    my ($fl_exist, $indx2_qu, $indx_qu) = binsearch_num ($beg_wd, @{$set_regions}, 0, @indices_rlv);
	    unless (defined $fl_exist) {
		print STDOUT "[ERROR 'assign_sgn_wds_to_regions'] The subroutine 'binsearch_num' FAILED for the $swd th significant window: $! !\n\n";
		return undef;
	    }

	    if ($indx_qu < 0) {
		print STDOUT "[ERROR 'assign_sgn_wds_to_regions'] IMPOSSIBLE: The $swd th significant window (beg, end) = ($beg_wd, $end_wd) is on the left of ALL the regions: $! !\n\n";
		return undef;
	    }

	    my ($beg_rg, $end_rg) = @{$set_regions->[$indx_qu]};
	    if ($end_rg < $end_wd) {
		print STDOUT "[ERROR 'assign_sgn_wds_to_regions'] IMPOSSIBLE: The $indx_qu th region (with (beg, end) = ($beg_rg, $end_rg)) canNOT completely ACCOMMODATE the $swd th significant window (beg, end) = ($beg_wd, $end_wd): $! !\n\n";
		return undef;
	    }

	    push @{$sgn_wds_in_rgs[$indx_qu]}, [$sites_in_wd, $pv, $pv0];

	} # End of the main for-loop (over the significant windows). #

	return \@sgn_wds_in_rgs;

}	# End of the subroutine, 'assign_sgn_wds_to_regions'. #


# detect_purge_cands2:
#	This is a MODIFIED version of 'detect_purge_cands'.
#	There are TWO MAJOR MODIFICATIONS:
#	(1) Unlike the predecessor, this subroutine no longer examines
#	   whether or not the flanking column(s) of the window holds "presence" states on both ends of the branch;
#	(2) The output of this subroutine also contains the two P-values for each merged candidate.
#	[CREATED on 2019/06/15; DEBUGGED on 2019/06/16.]
#
#	[REVISED on 2016/06/16]:
#	Now the subroutine also returns the significant windows in each purge-candidate region.
#
# detect_purge_cands:
#	This subroutine detects candidates of purge errors via a sliding window analysis.
#
#	[CREATION STARTED on Jan 9, 2018, ENDED on Jan dd, 2018.]
#
# [ INPUT: ($min_wdsize, $max_wdsize, $thrsh_pv, $if_modpv, $char_abs, $char_pres, $topnode_id, \%node2pa, \%node2ntfreqs, \%br2tr_mtrx, \%br2set_sw_jnt_probs, \%node2set_sw_cond_probs, \@set_clm_probs, \@info_psm_hst0, $beg_msa, $end_msa, \%br_attr, \%br2equiv), where ] ... OBSOLATE as of Jan 13 (Sat), 2018. #
# INPUT: ($min_wdsize, $max_wdsize, $thrsh_pv, $thrsh_pv0, $if_modpv, $char_abs, $char_pres, $topnode_id, \%node2pa, \%node2ntfreqs, \%br2tr_mtrx, \%br2set_sw_jnt_probs, \%node2set_sw_cond_probs, \@set_clm_probs, \@info_psm_hst0, $beg_msa, $end_msa, \%br_attr, \%br2equiv), where
#
#	$min/max_wdsize specifies the lower/upper-bound of the window size (for the detection of purge candidates);
#	$thrsh_pv is the P-value threshold to determine the purge candidates (based on the substitution model);
#	$thrsh_pv0 is the P-value threshold to determine the purge candidates (based on the null random-match model);
#	$if_modpv = 1 if you use the modified P-value, = 0 if you use the normal P-value;
#	$char_abs/pres is the character that represents an "absence"/"presence" state.
#
#	%node2pa = ($node_id => $parent_id, ...) for the nodes except the root.
#
#	%node2ntfreqs = ($node_id => \@ntfreqs, ...), where
#	$ntfreqs[$i] is the frequency of $NUCLEOTIDES[$i] at the node, $node_id.
#
#	@{$br2tr_mtrx{$br_id}} is the transition matrix
#	that describes the transition probabilities between both ends of the branch $br_id.
#	More specifically,
#	$br2tr_mtrx{$br_id}->[$i][$j] is the probability that the node $br_id shows $NUCLEOTIDES[$j]
#	given that the parent node of $br_id shows $NUCLEOTIDES[$i].
#
#	%br2set_sw_jnt_probs = ($br_id => \@set_sw_jnt_probs, ...) for ALL branches, where 
#	$set_sw_jnt_probs[$s]->[$i] is the probability that
#	all the extant sequences on the "ancestral side" of $br_id show the residue states that are actually found in the $s th site,
#	in addition to that $NUCLEOTIDES[$i] occupies the $s th site of the UPPER-end node of $br_id.
#	(It is an output of 'calc_joint_prob_ancresstats_and_ancstat2' in 'MyLikelihoods_respattern_msa_dna.pm'.)
#
#	%node2set_sw_cond_probs = ($node_id => \@set_sw_cond_probs, ...) for ALL nodes, where
#	$set_sw_cond_probs[$s]->[$i] is the probability that
#	all the descendant sequences of $node_id show the residue states that are actually found in the $s th site,
#	given that $NUCLEOTIDES[$i] occupies the $s th site of the ancestral node, $node_id.
#	But $set_sw_cond_probs[$s] = [] (empty) if $node_id refers to an external node and if the node has its $s th site occupied by a gap.
#	(It is an output of 'calc_prob_descresstats_cond_ancstat2' in 'MyLikelihoods_respattern_msa_dna.pm'.)
#
#	$set_clm_probs[$c] is the occurrence probability of the $c th column
#	under the given substitution model.
#	(It is an output of 'calc_clm_probs' in 'MyLikelihoods_respattern_msa_dna.pm'.)
#
#	@info_psm_hst0 = (\@set_cls_gblocks, \@totcts0_indels_for_cls, \@psm_hst0_for_cls),
#	which is an output of 'cal_init_psm_cands_for_sgl_msa' in 'MyTreeMap_indels_spt_odr2.pm',
#	gives information on the Dollo parsimonious indel history giving rise to the input MSA gap configuration, where
#	
#	@{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end) is for the $gb th block in the $cl th cluster (i.e., gapped segment) in the input MSA, with
#		$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#		$len_gb being the length (in terms of #{columns}) of the block,
#		and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention),
#	
#	$totcts0_indels_for_cls[$cl] is the total number of insertion/deletion events in the Dollo parsimonious indel history resulting in the $cl th cluster (i.e., gapped segment) in the input MSA,
#	
#	@{$psm_hst0_for_cls[$cl]} = (\%br2indels0, \%node2blwise_gstat0) is for the Dollo parsimonious indel history resulting in the $cl th cluster (i.e., gapped segment) in the input MSA, where 
#		%br2indels0 = ($branch_id => \@list_indels0, ...) is the map of insertions/deletions onto branches, with
#		@{$list_indels0[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved), and where
#		%node2blwise_gstat0 = ($node_id => \@blwise_gapstats0, ...) is the map of the gap states onto the nodes,
#		  with $blwise_gapstats0[$bl] is the gap-state ($char_pres/abs) of the $bl th block at the node.
#
#	$beg/end_msa is the beginning/ending coordinates of the input MSA (in the 0-based full-closed convention).
#
#	%br_attr = ($br_id => \@attr, ...) for all branches, where $attr[$INDX_LENGTH] is the branch length.
#
#	%br2equiv = ($br_id => $equiv_br_id, ...) for those branches with equivalents.
#
# [ OUTPUT: (\%br2purge_cands), where ] ... for the predecessor, 'detect_purge_cands'.
# [ OUTPUT: (\%br2purge_cands2, \%br2purge_cands), where ] ... OBSOLETE as of 2019/06/16.
# OUTPUT: (\%br2purge_cands2, \%br2purge_cands, \%br2sgn_wds_in_pcands), where
#
#		# Added in creation of 'detect_purge_cands2' (on 2019/06/15). (0) #
#	%br2purge_cands2 = ($br_id => \@list_purge_cands2, ...), where
#	@{$list_purge_cands2[$pc]} = (\@sites_in_pc, $pv, $pv0) gives information on the $pc th purge candidate,
#	with @sites_in_pc storing the sites constituting the region,
#		$pv and $pv0 being the P-values of deviation from the substutiton model and of deviation from the random match model, respectively;
#		# End of "Added in creation of 'detect_purge_cands2' (on 2019/06/15). (0)" #
#
#	%br2purge_cands = ($br_id => \@list_purge_cands, ...), where
#	@{$list_purge_cands[$pc]} = ($beg_pc, $end_pc)
#	provide the beginning and ending coordinates (in the 0-based half-open convention)
#	of the $pc th purge candidate (from left to right) along the branch $br_id;
#
#		# Added in revision (on 2019/06/16). #
#	%br2sgn_wds_in_pcands = ($br_id => \@list_sgn_wds_in_pcands, ...), where
#	@{$list_sgn_wds_in_pcands[$pc]->[$swd]} = (\@sites_in_wd, $pv, $pv0) gives information on the $swd th significant window in the $pc th purge candidate,
#	with @sites_in_wd storing the sites in the window,
#		and $pv and $pv0 being the P-values as explained above.
#		# End of "Added in revision (on 2019/06/16)." #
#
sub detect_purge_cands2 ($$$$$$$$\%\%\%\%\%\@\@$$\%\%) {

	my ($min_wdsize, $max_wdsize, $thrsh_pv, $thrsh_pv0, $if_modpv, $char_abs, $char_pres, 
	   $topnode_id, $node2pa, $node2ntfreqs, $br2tr_mtrx, $br2set_sw_jnt_probs, $node2set_sw_cond_probs,
	    $set_clm_probs, $info_psm_hst0, $beg_msa, $end_msa, $br_attr, $br2equiv) = @_;

   (print STDOUT "< 'detect_purge_cands2' ($min_wdsize <= {window size} <= $max_wdsize; absence/presence: '$char_abs'/'$char_pres'; {P-value threshold}_sbstmodel = $thrsh_pv; {P-value threshold}_random = $thrsh_pv0; ", '$if_modpv = ', "$if_modpv; $beg_msa <= {MSA coordinate} <= $end_msa) >\n\n") if ($TEST_LEVEL>0);

	my @set_all_nodes = sort {$a <=> $b} keys %{$node2pa};
	push @set_all_nodes, $topnode_id;
	my @set_all_branches = sort {$a <=> $b} keys %{$br_attr};

   if ($TEST_LEVEL>2) {
	print STDOUT "[ Secondary inputs ]\n\n";
	print STDOUT '@set_all_nodes = (', join (',', @set_all_nodes), ").\n\n";
	print STDOUT '@set_all_branches = (', join (', ', @set_all_branches), ").\n\n\n";
   }

		# Perform 'calc_br2cw_expcts_sbsts_ave' & 'calc_br2cw_obscts_sbsts', 
		# using the outputs of 'calc_prob_descresstats_cond_ancstat2',
		# 'calc_joint_prob_ancresstats_and_ancstat2' & 'calc_clm_probs'.
		# (IMPORTED from the subroutine, 'calc_br2cw_expcts_obscts_sbsts' on Jan 10, 2018.)


			# (Pre-1) Compute the expected numbers of substitutional differences per site. #

	my ($br2set_cw_expcts_sbsts_ave, $br2set_cw_expcts_sbsts_ue, $br2set_cw_expcts_sbsts_le)
	    = calc_br2cw_expcts_sbsts_ave (%{$br2tr_mtrx}, %{$br2set_sw_jnt_probs}, %{$node2set_sw_cond_probs}, %{$node2ntfreqs}, %{$node2pa});
	unless (defined $br2set_cw_expcts_sbsts_ave) {
	    print STDOUT "[ERROR 'detect_purge_cands2'] The subroutine 'calc_br2cw_expcts_sbsts_ave' FAILED: $! !\n\n";
	    return (undef, undef, undef);
	}
   (print STDOUT " --- (Pre-1) Subroutine 'calc_br2cw_expcts_sbsts_ave' ended ...\n\n") if ($TEST_LEVEL>1);

   if ($TEST_LEVEL>3) {
	print STDOUT "[ Content of ", '%{$br2set_cw_expcts_sbsts_ave}', ", which is the main output of 'calc_br2cw_expcts_sbsts_ave' ]\n\n";

	print STDOUT join ("\t", 'Column\Branch', @set_all_branches), "\n\n";

	for (my $c = $beg_msa; $c <= $end_msa; $c++) { # Outer for-loop (over columns). #
	    my @cw_expcts_along_brs = ();
	    foreach my $br (@set_all_branches) {
		my $cw_expct = $br2set_cw_expcts_sbsts_ave->{$br}[$c];
		(defined $cw_expct) or ($cw_expct = 'undef');
		push @cw_expcts_along_brs, $cw_expct;
	    }
	    print STDOUT join ("\t", $c, @cw_expcts_along_brs), "\n";

	} # End of outer for-loop (over columns). #

	print STDOUT "\n\n";

=Alternative idea
	foreach my $br (@set_all_branches) {
	    my $set_cw_expcts = $br2set_cw_expcts_sbsts_ave->{$br};
	    my $ct_clms = @{$set_cw_expcts};
	    print STDOUT "{ Branch ID: $br (w/ $ct_clms columns) }\n\n";
	    print STDOUT join ("\t", 'Column', 'Exp#{substitutions}_ave'), "\n\n";
	    for (my $c = 0; $c < $ct_clms; $c++) { print STDOUT join ("\t", $c, $set_cw_expcts->[$c]), "\n"; }
	    print STDOUT "\n\n";
	}
=cut
   } # END of "if ($TEST_LEVEL>3) {...}"

			# (Pre-2) Compute the "observed" numbers of substitutional differences per site. #

	my $br2set_cw_obscts_sbsts = calc_br2cw_obscts_sbsts (%{$br2tr_mtrx}, %{$br2set_sw_jnt_probs}, %{$node2set_sw_cond_probs}, @{$set_clm_probs}) ;
	unless (defined $br2set_cw_obscts_sbsts) {
	    print STDOUT "[ERROR 'detect_purge_cands2'] The subroutine 'calc_br2cw_obscts_sbsts' FAILED: $! !\n\n";
	    return (undef, undef, undef);
	}
   (print STDOUT " --- (Pre-2) Subroutine 'calc_br2cw_obscts_sbsts' ended ...\n\n") if ($TEST_LEVEL>1);

   if ($TEST_LEVEL>3) {
	print STDOUT "[ Content of ", '%{$br2set_cw_obscts_sbsts}', ", which is the output of 'calc_br2cw_obscts_sbsts' ]\n\n";

	print STDOUT join ("\t", 'Column\Branch', @set_all_branches), "\n\n";

	for (my $c = $beg_msa; $c <= $end_msa; $c++) { # Outer for-loop (over columns). #
	    my @cw_obscts_along_brs = ();
	    foreach my $br (@set_all_branches) {
		my $cw_obsct = $br2set_cw_obscts_sbsts->{$br}[$c];
		(defined $cw_obsct) or ($cw_obsct = 'undef');
		push @cw_obscts_along_brs, $cw_obsct;
	    }
	    print STDOUT join ("\t", $c, @cw_obscts_along_brs), "\n";

	} # End of outer for-loop (over columns). #

	print STDOUT "\n\n";

=Alternative idea
	foreach my $br (@set_all_branches) {
	   my $set_cw_obscts = $br2set_cw_obscts_sbsts->{$br};
	   my $ct_clms = @{$set_cw_obscts};
	   print STDOUT "{ Branch ID: $br (w/ $ct_clms columns) }\n\n";
	   print STDOUT join ("\t", 'Column', 'Obs#{substitutions}'), "\n\n";
	   for (my $c=0; $c < $ct_clms; $c++) { print STDOUT join ("\t", $c, $set_cw_obscts->[$c]), "\n"; }
	   print STDOUT "\n\n";
	}
=cut
   } # END of "if ($TEST_LEVEL>3) {...}"


		# END of "(IMPORTED from the subroutine, 'calc_br2cw_expcts_obscts_sbsts' on Jan 10, 2018.)"

			# (Pre-trailing) Sum the expected counts over the branches, and also sum the 'observed' counts over the branches. #
	my ($cw_sumexpcts_sbsts, $cw_sumobscts_sbsts);
	if ($if_modpv) {
	    $cw_sumexpcts_sbsts = sum_arrays_over_branches (%{$br2set_cw_expcts_sbsts_ave});
	    unless (defined $cw_sumexpcts_sbsts) {
		print STDOUT "[ERROR 'detect_purge_cands2'] The subroutine 'sum_arrays_over_branches' FAILED for expected counts of substitutions: $! !\n\n";
		return (undef, undef, undef);
	    }
	    $cw_sumobscts_sbsts = sum_arrays_over_branches (%{$br2set_cw_obscts_sbsts});
	    unless (defined $cw_sumobscts_sbsts) {
		print STDOUT "[ERROR 'detect_purge_cands2'] The subroutine 'sum_arrays_over_branches' FAILED for 'observed' counts of substitutions: $! !\n\n";
		return (undef, undef, undef);
	    }
   (print STDOUT " --- (Pre-trailing) Subtoutine 'sum_arrays_over_branches' ended to calculate total expected & 'observed' counts of substitutions ...\n\n") if ($TEST_LEVEL>1);

   if ($TEST_LEVEL>3) {
	print STDOUT '[ Contents of @{$cw_sumexpcts_sbsts} & @{$cw_sumobscts_sbsts} ]', "\n\n";
	print STDOUT join ("\t", 'Column', 'Total[Exp#{sbsts}]', 'Total[Obs#{sbsts}]'), "\n\n";
	for (my $c = $beg_msa; $c <= $end_msa; $c++) {
	    my $sumexpct = $cw_sumexpcts_sbsts->[$c];
	    (defined $sumexpct) or ($sumexpct = 'undef');
	    my $sumobsct = $cw_sumobscts_sbsts->[$c];
	    (defined $sumobsct) or ($sumobsct = 'undef');
	    print STDOUT join ("\t", $c, $sumexpct, $sumobsct), "\n";
	}
	print STDOUT "\n\n";
   } # END  of "if ($TEST_LEVEL>3) {...}"

	} # END of "if ($if_modpv) {...}"


			# (Pre-3) Prepare the site-wise gap states (of the Dollo-parsimonious indel history) at every node. #

	# %{$node2sw_gstats0} = ($node_id => \@sw_gstats0, ...) for all nodes, where
	# $sw_gstats0[$s] is the gap state of the $s th site on the node $node_id
	# (usually according to the Dollo-parsimony principle).
	#
	my $node2sw_gstats0 = mk_node2sw_gstats (@{$info_psm_hst0}, $char_abs, $char_pres, $beg_msa, $end_msa, @set_all_nodes);
	unless (defined $node2sw_gstats0) {
	    print STDOUT "[ERROR 'detect_purge_cands2'] The subroutine 'mk_node2sw_gstats' FAILED: $! !\n\n";
	    return (undef, undef, undef);
	}
   (print STDOUT " --- (Pre-3) Subroutine 'mk_node2sw_gstats' ended ...\n\n") if ($TEST_LEVEL>1);

   if ($TEST_LEVEL>3) {
	print STDOUT "[ Content of ", '%{$Node2sw_gstats0}', ", which is the output of 'mk_node2sw_gstats' ]\n\n";
	print STDOUT 'Site\Node', "\t", join (', ', @set_all_nodes), "\n\n";
	for (my $s = $beg_msa; $s <= $end_msa; $s++) { # Outer for-loop (over sites). #
	    my @sw_gstats_on_nodes = ();
	    foreach my $node (@set_all_nodes) {
		my $gstat = $node2sw_gstats0->{$node}[$s];
		(defined $gstat) or ($gstat = '?');
		push @sw_gstats_on_nodes, $gstat;
	    }
	    print STDOUT $s, "\t", join (', ', @sw_gstats_on_nodes), "\n";

	}				# End of the outer for-loop (over sites). #

	print STDOUT "\n\n";

   } # END of "if ($TEST_LEVEL>3) {...}"


			# (MAIN) Sliding-window analysis to detect the candidates of purge-type errors. #
			# (Created in refercence to ll.2835-2916 of 'examine_purge_ctrls.2.pl'.) #

   (print STDOUT "[ Main Process: Sinding-window analysis to detect 'purge candidates' ]\n\n") if ($TEST_LEVEL>1);

	my $rflank_site_msa = $end_msa+1;
	my %br2purge_cands;
	my %br2purge_cands2; # ADDED in creation of 'detect_purge_cands2' (on 2019/06/15). (1) #
	my %br2sgn_wds_in_pcands; # Added in revision (on 2019/06/16). (1)

	foreach my $br (@set_all_branches) { # Main outer foreach-loop (over all branches) #

	    my $eq_br = $br2equiv->{$br};
	    my $brlen = $br_attr->{$br}[$INDX_LENGTH];
	    if ((defined $eq_br) and ($brlen == 0)) { next; } # Skip the null artificial branch. #

	    my $node_ue = $node2pa->{$br};
	    my $sw_gstats0_le = $node2sw_gstats0->{$br};
	    my $sw_gstats0_ue = $node2sw_gstats0->{$node_ue};

			# ADDED on Jan 13 (Sat), 2018. #
	    my $ntfreqs_le = $node2ntfreqs->{$br};
	    my $ntfreqs_ue = $node2ntfreqs->{$node_ue};
	    my $ct_nt = @{$ntfreqs_le};
	    my $r_mch0 = 0;
	    for (my $n = 0; $n <$ct_nt; $n++) { $r_mch0 += $ntfreqs_le->[$n] * $ntfreqs_ue->[$n]; }
			# END of "ADDED on Jan 13 (Sat), 2018." #

   (print STDOUT ' ',  "{ Branch ID: $br (upper-end: $node_ue ; rate{random-match} = $r_mch0) }\n\n") if ($TEST_LEVEL>2); # MODIFIED on Jan 13 (Sat), 2018. # 
   ##(print STDOUT ' ',  "{ Branch ID: $br (upper-end: $node_ue) }\n\n") if ($TEST_LEVEL>2);

	    my @joint_purge_cands = (); # element = [$beg, $end] of a candidate region. 
	    my @joint_purge_cands2 = (); # element = \@sites_in_purge_cand

	    for (my $wsize = $min_wdsize; $wsize <= $max_wdsize; $wsize++) { # Middle for-loop (over the window sizes). #

   (print STDOUT ' ' x 3, "{ Window size: $wsize }", "\n\n") if ($TEST_LEVEL>3);

	        my @purge_cands = (); # element = [$beg, $end] of a candidate region. 
	        my @purge_cands2 = (); # element = \@cp_window of a candidate region. (Added on 2019/06/15). #
		my @window = ();
		my $if_lf_filled = 0;

		for (my $s = $beg_msa; $s < $rflank_site_msa; $s++) { # Inner for-loop (over the sites). #
		    my $gstat0_le = $sw_gstats0_le->[$s];
		    my $gstat0_ue = $sw_gstats0_ue->[$s];
		    if ($gstat0_le eq $char_abs) {
			if ($gstat0_ue eq $char_abs) {
			    next; # Just skip the site. #
			} else {
			    @window = (); # Dump the growing window, and re-start to grow the window. #
			    $if_lf_filled = 0;
			    next;
			}
		    } elsif ($gstat0_ue eq $char_abs) {
			@window = (); # Dump the growing window, and re-start to grow the window. #
			$if_lf_filled = 0;
			next;
		    }

		    unless ($if_lf_filled) { # If the left-flanking site is NOT filled with "presence", make it the "left-flanking site" of the subsequent window. #
			$if_lf_filled = 1;
			## next; DISABLED in creation of 'detect_purge_cands2' (on 2019/06/15). (1)
		    }

		    push @window, $s;

		    (@window == $wsize) or next;

			# The window is fully grown. #
			# => Check the "right-flanking" site. #

=DISABLED in creation of 'detect_purge_cands2' (on 2019/06/15). (2)
		    my $if_rf_filled = 0;
		    for (my $s2 = $s+1; $s2 < $rflank_site_msa; $s2++) { # Inner-inner for-loop (over the sites on the right of $s). #
			my $gstat0_le2 = $sw_gstats0_le->[$s2];
			my $gstat0_ue2 = $sw_gstats0_ue->[$s2];
			next if (($gstat0_le2 eq $char_abs) and ($gstat0_ue2 eq $char_abs));
			($if_rf_filled = 1) if (($gstat0_le2 ne $char_abs) and ($gstat0_ue2 ne $char_abs));
			last;
		    }					# End of the inner-inner for-loop (over the sites on the right of $s). #

		    if ($if_rf_filled == 0) { # If the right-flanking site is NOT filled with "presence", dump the window.
			@window = 0;
			$if_lf_filled = 0;
			next;
		    }
=cut # END of "DISABLED in creation of 'detect_purge_cands2' (on 2019/06/15). (2)"


			# Now, the window is flanked (on both sides) by sites filled with "presence". #
			# => Calculate the P-value on the residue pattern. #

   (print STDOUT ' ' x 3, '| Window: (', join (', ', @window), ') |', "\n\n") if ($TEST_LEVEL>5);

		    my ($beg_wd, $end_wd) = @window[0,$wsize-1];

		    my $pv;
		    my $d_obs; #  ADDED on Jan 13 (Sat), 2018. #

		    if ($if_modpv) {
			my ($size_wd, $d_exp_wd_br, $d_obs_wd_br, $sumd_exp_wd, $sumd_obs_wd, $d_mexp_wd_br, $modpv_wd_br)
			 = calc_modpv_obscts_sbsts_wd_br (@window, $br, %{$br2set_cw_expcts_sbsts_ave}, %{$br2set_cw_obscts_sbsts}, @{$cw_sumexpcts_sbsts}, @{$cw_sumobscts_sbsts});
		        unless (defined $modpv_wd_br) {
			    print STDOUT "[ERROR 'detect_purge_cands2'] The subroutine 'calc_modpv_obscts_sbsts_wd_br' FAILED along the branch '$br' and on the window of size $wsize ([beg, end] = [$beg_wd, $end_wd]) !\n";
			    return (undef, undef, undef);
		        }
   (print STDOUT ' ' x 3, " --- Subroutine 'calc_modpv_obscts_sbsts_wd_br' ended (w/", '$modpv_wd_br = ', "$modpv_wd_br) ...\n\n") if ($TEST_LEVEL >5);

			$pv = $modpv_wd_br;
			$d_obs = $d_obs_wd_br; # ADDED on Jan 13 (Sat), 2018. #

		    } else {
		        my ($size_wd, $d_exp, $d_obs2, $pv_wd_br) = calc_pv_obscts_sbsts_wd_br (@window, $br, %{$br2set_cw_expcts_sbsts_ave}, %{$br2set_cw_obscts_sbsts});
		        unless (defined $pv_wd_br) {
			    print STDOUT "[ERROR 'detect_purge_cands2'] The subroutine 'calc_pv_obscts_sbsts_wd_br' FAILED along the branch '$br' and on the window of size $wsize ([beg, end] = [$beg_wd, $end_wd]) !\n";
			    return (undef, undef, undef);
		        }
   (print STDOUT ' ' x 3, " --- Subroutine 'calc_pv_obscts_sbsts_wd_br' ended (w/", '$pv_wd_br = ', "$pv_wd_br) ...\n\n") if ($TEST_LEVEL >5);

			$pv = $pv_wd_br;
			$d_obs = $d_obs2; # ADDED on Jan 13 (Sat), 2018. #
		    }

				# MODIFIED on Jan 13 (Sat), 2018. #

		    unless ($pv <= $thrsh_pv) { # The number of differences is not significantly more than expected from the substitution model.
			shift @window; # A BUG fixing measure!!
			next;
		    }

		    ## my $ct_mch_obs = $wsize - $d_obs;

		    my $int_d_obs = int ($d_obs);
		    my $frac_d_obs = $d_obs - $int_d_obs;

		    my $pv01 = binomial_cum_via_betai ($wsize, $int_d_obs, 1 - $r_mch0); # Prob (X <= $int_d_obs)
		    my $pv02 = ($int_d_obs >= $wsize) ? 1 : binomial_cum_via_betai ($wsize, $int_d_obs +1, 1 - $r_mch0); # Prob (X <= $int_d_obs + 1)

		    my $pv0 = (1 - $frac_d_obs) * $pv01 + $frac_d_obs * $pv02; # The probability wht $d_obs or less "observed" differences, under the binomial distribution with $wsize trials and the random mismatch rate (1-$r_mch0).

		    if ($pv0 < $thrsh_pv0) { # The number of differences is significantly less than expected from random.
		    ##if ($pv < $thrsh_pv0) { # A BUG was found!! => Fixed....}
			shift @window; # A BUG fixing measure!!
			next;
		    }

		    push @purge_cands, [$beg_wd, $end_wd, $pv, $pv0];

		    #if ($pv <= $thrsh_pv) {
		    #	push @purge_cands, [$beg_wd, $end_wd];
		    #}

				# END of "MODIFIED on Jan 13 (Sat), 2018." #

			# ADDED in creation of 'detect_purge_cands2' (on 2019/06/15). (2) #
		    my @cp_window = @window;
		    push @purge_cands2, \@cp_window;
			# End of "ADDED in creation of 'detect_purge_cands2' (on 2019/06/15). (2)" #

		    shift @window; # Remove the leftmost site from the window.
		}					# End of the inner for-loop (over the sites). #

   if ($TEST_LEVEL>=5) {
	print STDOUT ' ' x 3, "{ List of 'purge candidate' windows }\n\n";

	print STDOUT ' ' x 3, join ("\t", 'No.', 'beg_wd', 'end_wd', 'pv', 'pv0'), "\n\n"; # MODIFIED on Jan 13 (Sat), 2018. #
	##print STDOUT ' ' x 3, join ("\t", 'No.', 'beg_wd', 'end_wd'), "\n\n";

	my $ct_cands = @purge_cands; 
	for (my $n = 0; $n <$ct_cands; $n++) { print STDOUT ' ' x 3, join ("\t", $n, @{$purge_cands[$n]}), "\n"; }
	print STDOUT "\n\n";
   } # END of "if ($TEST_LEVEL>=5) {...}"

		push @joint_purge_cands, @purge_cands;
		push @joint_purge_cands2, @purge_cands2;

	    }	  # End of the middle for-loop (over the window sizes). #

			# (Trailing) Constructing a non-redundant list of the candidate regions. #

	    my $mgd_purge_cands = (@joint_purge_cands<2) ? \@joint_purge_cands : merge_integer_intervals (@joint_purge_cands, 0, 1);
	    unless (defined $mgd_purge_cands) {
		print STDOUT "[ERROR 'detect_purge_cands2'] The subtroutine 'merge_integer_intervals' FAILED for the set of purge candidates along the branch '$br': $! !\n\n";
		return (undef, undef, undef);
	    }

   (print STDOUT ' ', " --- Subroutine 'merge_integer_intervals' ended ...\n\n") if ($TEST_LEVEL>3);

	    if (@joint_purge_cands == 1) { # Added on 2019/06/16.
		my @bds_cand = @{$joint_purge_cands[0]}[0, 1];
		$mgd_purge_cands = [\@bds_cand];
	    }

	    my $ct_mgd = @{$mgd_purge_cands};

=MOVED with modification to below the "ADDED in creation of 'detect_purge_cands2' (on 2019/06/15). (3)"
   if ($TEST_LEVEL>3) {
	print STDOUT ' ', "[ List of non-redundant 'purge candidate' regions (#{regions} = $ct_mgd) ]\n\n";
	print STDOUT ' ', join ("\t", 'No', 'beg_rg', 'end_rg'), "\n\n";
	for (my $n = 0; $n <$ct_mgd; $n++) { print STDOUT ' ' , join ("\t", $n, @{$mgd_purge_cands->[$n]}), "\n"; }
	print STDOUT "\n\n";

   } # END of "if ($TEST_LEVEL>3) {...}"
=cut # END of "MOVED with modification to below the "ADDED in creation of 'detect_purge_cands2' (on 2019/06/15). (3)""

	    $br2purge_cands{$br} = $mgd_purge_cands;


			## ADDED in creation of 'detect_purge_cands2' (on 2019/06/15). (3) ##


	    my $raw_mgd_purge_cands2 = (@joint_purge_cands2 <2) ? \@joint_purge_cands2 : merge_sitesets (@joint_purge_cands2);
	    unless (defined $raw_mgd_purge_cands2) {
		print STDOUT "[ERROR 'detect_purge_cands2'] The subtroutine 'merge_sitesets' FAILED for the set of purge candidates along the branch '$br': $! !\n\n";
		return (undef, undef, undef);
	    }
	    my $ct_mgd2 = @{$raw_mgd_purge_cands2};
	    unless ($ct_mgd == $ct_mgd2) {
		print STDOUT "[CAUTION 'detect_purge_cands2'] The size (=$ct_mgd2) of the output of 'merge_sitesets' CLASHES with that (=$ct_mgd) of the output of 'merge_integer_intervals' for the purge candidates along the branch '$br': $! !\n\n";
	    }
   (print STDOUT ' ', " --- Subroutine 'merge_sitesets' ended ...\n\n") if ($TEST_LEVEL>3);

	    my @mgd_purge_cands2 = ();
	    foreach my $mpcand2 (@{$raw_mgd_purge_cands2}) { # Trailing middle foreach-loop (over merged purge-candidates). 
		    my $wsize = @{$mpcand2};
		
		    my $pv;
		    my $d_obs;

		    if ($if_modpv) {
			my ($size_wd, $d_exp_wd_br, $d_obs_wd_br, $sumd_exp_wd, $sumd_obs_wd, $d_mexp_wd_br, $modpv_wd_br)
			 = calc_modpv_obscts_sbsts_wd_br (@{$mpcand2}, $br, %{$br2set_cw_expcts_sbsts_ave}, %{$br2set_cw_obscts_sbsts}, @{$cw_sumexpcts_sbsts}, @{$cw_sumobscts_sbsts});
		        unless (defined $modpv_wd_br) {
			    print STDOUT "[ERROR 'detect_purge_cands2'] The subroutine 'calc_modpv_obscts_sbsts_wd_br' FAILED along the branch '$br' and on the merged purge-candidate, (", join (',',  @{$mpcand2}), ") (w/ size=$wsize) !\n";
			    return (undef, undef, undef);
		        }
   (print STDOUT ' ' x 3, " --- Subroutine 'calc_modpv_obscts_sbsts_wd_br' ended (w/", '$modpv_wd_br = ', "$modpv_wd_br) ...\n\n") if ($TEST_LEVEL >5);

			$pv = $modpv_wd_br;
			$d_obs = $d_obs_wd_br;

		    } else {
		        my ($size_wd, $d_exp, $d_obs2, $pv_wd_br) = calc_pv_obscts_sbsts_wd_br (@{$mpcand2}, $br, %{$br2set_cw_expcts_sbsts_ave}, %{$br2set_cw_obscts_sbsts});
		        unless (defined $pv_wd_br) {
			    print STDOUT "[ERROR 'detect_purge_cands2'] The subroutine 'calc_pv_obscts_sbsts_wd_br' FAILED along the branch '$br' and on the merged purge-candidate, (", join (',', @{$mpcand2}), ") (w/ size=$wsize) !\n";
			    return (undef, undef, undef);
		        }
   (print STDOUT ' ' x 3, " --- Subroutine 'calc_pv_obscts_sbsts_wd_br' ended (w/", '$pv_wd_br = ', "$pv_wd_br) ...\n\n") if ($TEST_LEVEL >5);

			$pv = $pv_wd_br;
			$d_obs = $d_obs2;
		    }

		    my $int_d_obs = int ($d_obs);
		    my $frac_d_obs = $d_obs - $int_d_obs;

		    my $pv01 = binomial_cum_via_betai ($wsize, $int_d_obs, 1 - $r_mch0); # Prob (X <= $int_d_obs)
		    my $pv02 = ($int_d_obs >= $wsize) ? 1 : binomial_cum_via_betai ($wsize, $int_d_obs +1, 1 - $r_mch0); # Prob (X <= $int_d_obs + 1)

		    my $pv0 = (1 - $frac_d_obs) * $pv01 + $frac_d_obs * $pv02; # The probability wht $d_obs or less "observed" differences, under the binomial distribution with $wsize trials and the random mismatch rate (1-$r_mch0).

		    push @mgd_purge_cands2, [$mpcand2, $pv, $pv0];

			# @{$list_purge_cands2[$pc]} = (\@sites_in_pc, $pv, $pv0) gives information on the $pc th purge candidate,

	    } # End of the trailing middle foreach-loop (over merged purge-candidates).
	    
	    $br2purge_cands2{$br} = \@mgd_purge_cands2;

			## END of "ADDED in creation of 'detect_purge_cands2' (on 2019/06/15). (3)" ##

			## ADDED in revision (on 2019/06/16). (2) ##

	    my $sgn_wds_in_pcands = (@joint_purge_cands>0) ? assign_sgn_wds_to_regions (@joint_purge_cands, @joint_purge_cands2, @{$mgd_purge_cands}) : [];
	    unless (defined $sgn_wds_in_pcands) {
		print STDOUT "[ERROR 'detect_purge_cands2'] The subtroutine 'assign_sgn_wds_to_regions' FAILED for the set of purge candidates along the branch '$br': $! !\n\n";
		return (undef, undef, undef);
	    }
	    $br2sgn_wds_in_pcands{$br} = $sgn_wds_in_pcands;

#	%br2sgn_wds_in_pcands = ($br_id => \@list_sgn_wds_in_pcands, ...), where
#	@{$list_sgn_wds_in_pcands[$pc]->[$swd]} = (\@sites_in_wd, $pv, $pv0) gives information on the $swd th significant window in the $pc th purge candidate,


			## END of "ADDED in revision (on 2019/06/16). (2)" ##

			## MOVED with modification from above the "ADDED in creation of 'detect_purge_cands2' (on 2019/06/15). (3)" (on 2019/06/16) ##
   if ($TEST_LEVEL>3) {
	print STDOUT ' ', "[ List of non-redundant 'purge candidate' regions (#{regions} = $ct_mgd) ]\n\n";
	print STDOUT ' ', join ("\t", 'No', 'beg_rg', 'end_rg', '(sites,in,the,region)', 'pv', 'pv0', '{significant window included ([sites,in,wd]/pv_wd/pv0_wd)}_0', '{...}_1', '...'), "\n\n"; # Further modified in revision (on 2019/06/16).
	##print STDOUT ' ', join ("\t", 'No', 'beg_rg', 'end_rg', '(sites,in,the,region)', 'pv', 'pv0'), "\n\n";
	####print STDOUT ' ', join ("\t", 'No', 'beg_rg', 'end_rg'), "\n\n"; # for the predecessor, 'detect_purge_cand'.
	for (my $n = 0; $n <$ct_mgd; $n++) {
	    my ($sites, $pv, $pv0) = @{$mgd_purge_cands2[$n]};
	    my $cnct_sites = join (',', @{$sites});

			# Further MODIFIED in revision (on 2019/06/16). #
	    my @out_sgn_wds = ();
	    foreach my $info_swd (@{$sgn_wds_in_pcands->[$n]}) { # Inner foreach-loop (over significant windows in the region).
		my ($sites_in_wd, $pv_wd, $pv0_wd) = @{$info_swd};
		my $cnct_sites_in_wd = join (',', @{$sites_in_wd});
		push @out_sgn_wds, "[$cnct_sites_in_wd]/$pv_wd/$pv0_wd";
	    } # End of the inner foreach-loop (over significant windows in the region).
	    print STDOUT ' ' , join ("\t", $n, @{$mgd_purge_cands->[$n]}, "($cnct_sites)", $pv, $pv0, @out_sgn_wds), "\n";
	    ##print STDOUT ' ' , join ("\t", $n, @{$mgd_purge_cands->[$n]}, "($cnct_sites)", $pv, $pv0), "\n";
			# End of "Further MODIFIED in revision (on 2019/06/16)." #
	    ####print STDOUT ' ' , join ("\t", $n, @{$mgd_purge_cands->[$n]}), "\n"; # for the predecessor, 'detect_purge_cand'.
	}
	print STDOUT "\n\n";

   } # END of "if ($TEST_LEVEL>3) {...}"
			## END of "MOVED with modification from above the "ADDED in creation of 'detect_purge_cands2' (on 2019/06/15). (3)" (on 2019/06/16)" ##

	}			# End of the main outer foreach-loop (over all branches). #

   (print STDOUT "\n\n< 'detect_purge_cands2' Ended />\n\n") if ($TEST_LEVEL>0);

	return (\%br2purge_cands2, \%br2purge_cands, \%br2sgn_wds_in_pcands); # Further MODIFIED in revision (on 2019/06/16). #
	##return (\%br2purge_cands2, \%br2purge_cands); # MODIFIED in creation of 'detect_purge_cands2' (on 2019/06/15). #
	####return \%br2purge_cands;

}	# END of the subroutine 'detect_purge_cands2'. #



1;		# a "spell" which makes the module "use"-able and/or "require"-able #


