package MyFile_manipu;
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.5: Copyright (C) 2019 Kiyoshi Ezawa
#	Version 0.6: Copyright (C) 2020 Kiyoshi Ezawa
#	Version 0.7: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyFile_manipu.pm: 
#	This Perl package, which originally belonged to the 'DENSERM_P' package,
#	was REVISED along with the creation of 'ANEX_P.ver0.7',
#	and was MOVED to '../MyPerlModules_ANEX/".
#
#	[MAJOR REVISION on 2020/04/30]:
#	Some subroutines can now handle symbolic links as well,
#	instead of just normal files as before the revision.
#
###################################################################################
#	DENSERM_P: A package to DEtect Negative SElection on Recurrent Mutations
#			(almost exclusively via Perl modules and scripts).
#	Version 0.3: Copyright (C) 2012 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyFile_manipu.pm: This module consists of miscellaneous subroutines
#			which manipulates files and/or directories.
#

use strict;
use warnings;

use Exporter;

our @ISA  = qw(Exporter);

our @EXPORT_OK = qw(
	enlist_files search_grand_daughters exhaust_files peek_into_directory
	fetch_fields fetch_chunk open_input_file open_input_file2 open_output_file
	list_sessions list_sessions_ext
	fetch_entries_into_array fetch_entries_into_hash
	construct_label2indx const_indices_from_labels tfm_labels2indices tfm_labels2indices_gen
	readin_matrix extend_matrix
		);



# enlist_files: This subroutine looks into a directory and makes a list of files
#		which satisfy the specified conditions.
#
# CONDITIONS: 	[ $type = 'd' for a directory, '-' for a normal file; ] ... before revision on 2020/04/30.
# 		$type = 'd' for a directory, '-' for a normal file, 'l' for a symbolic link; # Modified on 2020/04/30.
#		$prefix = '' if unspecified;
#		$suffix = '' if unspecified;
#
sub enlist_files (\$$$$) {

	my ($rf_indir, $type, $prefix, $suffix) = @_;

=for DEBUGGING the subroutine
	print STDERR "\n< Subroutine 'enlist_files' >\n\n"; 
=cut
	my $type_length = length $type;
	my $pre_length = length $prefix;
	my $suf_length = length $suffix;

	local $/ = "\n";
	my $fh;
	open ($fh, "ls -l ${$rf_indir} |" ) or die "\nCannot folk or look into ${$rf_indir}: $!\n";

	my @list = ();
	while (my $line = <$fh>) {

=for DEBUGGING the subroutine
		print STDERR "Input line:  $line\n";
=cut

		#($line =~ m/^$type/o ) or next;
		( ($type_length < length $line) and ($type eq substr ($line, 0, $type_length) )  ) or next;

		my ($filename) = ($line =~ m/(\S+)$/ );	# assume that the filename does not contain white letters #
		my $namelength = length $filename;

=for DEBUGGING the subroutine
		print STDERR "Filename:  $filename .\n";
=cut

				# screening by the prefix #
		unless ($prefix eq '') {
			($namelength < $pre_length) and next;
			my $this_prefix = substr($filename, 0, $pre_length);
			($prefix eq $this_prefix) or next; 
		}

				# screening by the suffix;
				# this method allows the suffix to be more general than usual. #
		unless ($suffix eq '') {	# this method allows 
			($namelength < $suf_length) and next;
			my $this_suffix = substr($filename, -$suf_length);
			($suffix eq $this_suffix) or next;
		}

		push @list, $filename;
		
	}
	close ($fh);

=for DEBUGGING the subroutine
	print STDERR "\n --- End of the subroutine 'enlist_files' ...\n\n"; 
=cut

	return \@list;

}			# end of the subroutine enlist_files #


# search_grand_daughters: This subroutine receives an directory,
#	and looks for its 'ground daughter' files with specified characters.
#	The specifiable charachters conform to that of 'enlist_files'.
#
# INPUT: In accord to that of 'enlist_files'.
#
# OUTPUT: \@list_grand_daughters, where
#		$list_grand_daughters[m] = "daughter_name/grand_daughter_name".
#
sub search_grand_daughters (\$$$$$$) {

	my ($grand_father, $pref_d, $suff_d, $type_gd, $pref_gd, $suff_gd) = @_;

	my $list_daughters = enlist_files (${$grand_father}, 'dr', $pref_d, $suff_d);
	unless (@{$list_daughters}) {
		print STDERR "NO daughter directories of the directory ${$grand_father} to look for its grend daughters.\n";
		return undef;
	}

	my @grand_daughters = ();

	for my $daughter (@{$list_daughters}) {

		my $dir = "${$grand_father}/$daughter";

		my $daughters_of_d = enlist_files ($dir, $type_gd, $pref_gd, $suff_gd);

		my $ct_dd = @{$daughters_of_d};
		for (my $i=0; $i< $ct_dd; $i++) {
			push @grand_daughters, ( $daughter . '/' . $daughters_of_d->[$i] ) ;
		}
		
	}

	return \@grand_daughters;

}		# end of the subroutine search_grand_daughters #


# exhaust_files:
#	This subroutine exhaustively searches the filesystem under a directory
#	for files satisfying specified conditions.
#	The subroutine uses the 'find' command.
#
# INPUT: (\$directory, $prefix, $suffix, \$additional_conditions, $fl_trimdir), where
#	$additional_conditions conforms to the grammer of the 'find' command.
#	$fl_trimdir = 1 if you want to remove directories preceeding the file name,
#		    = 0 (or undef or '') if you want to keep the entire pathname.
#
# OUTPUT: \@list_files.
#
sub exhaust_files (\$$$\$$) {

	my ($dir, $prefix, $suffix, $add_condi, $fl_trimdir) = @_;
	local $/ = "\n";

	my $fh;
	open ($fh, "find ${$dir} -name \"${prefix}\*${suffix}\" ${$add_condi} |")
		or die "\nCannot fork or look into the file system under '${$dir}': $!\n";

	my @list_files = ();
	while (my $in = <$fh> ) {
	    $in =~ s/\s+$//;
	    ($in =~ m/\S+/) or next;
	    $fl_trimdir and ( ($in) = ($in =~ m/([^\/\s]+)$/ ) );
	     push @list_files, $in;
	}

	return \@list_files;

}			# end of the subroutine 'exhaust_files' #


# peek_into_directory: This subroutine sees if the specified directory has the desired information.
#
sub peek_into_directory (\$$) {

	my ($rf_indir, $inbase) = @_;

	local $/ = "\n";
	open (TMP, " ls -l ${$rf_indir} | ") or die "\nCannot fork or look into ${$rf_indir}: $!\n";
	my $info = '';
	while (my $line = <TMP>) {

		if ($line =~  m/\s$inbase$/o ) {
			$info = $line;
			last ;
		}
	}
	close (TMP);

	unless ($info) {

		print STDERR "There is no $inbase in ${$rf_indir} \!\n";
		return 0;
	}


	my  @fields = split ' ', $info;
	unless ($fields[4]) {

		print STDERR "No information content in ${$rf_indir}/$inbase \!\n";
		return 0;
	}

	return 1;
}				# end of the subroutine peek_into_directory #


# fetch_fields: This subroutine reads in a file which is in the form of
#		a white character-separated table with one entry in one line,
#		and constructs a hash:
#		(cluster_ID => \@sub_fields),
#		where @sub_fields consists of fields for each line whose indices specified by the 2nd argument.
#		if @{$indices} is empty, it stores all fields. 
#
sub fetch_fields (\$\%\@) {

	my ($infile, $cnum2sbflds, $indices) = @_;
	local $/ = "\n";
	my $fh;

	open ($fh, "< ${$infile}" ) or open ($fh, "gunzip -c ${$infile}.gz |" )
		or die "\nCannot fork or open ${$infile}(.gz): $!\n";
	my $header = <$fh>;

	my $last_index = @{$indices} ?  $indices->[$#{$indices}] : undef;
	
	while (my $in = <$fh>) {

		#my ($cnum) = ($in =~ m/^(\d[\d_]+)/ ) ;
		#(defined $cnum) or next;

		my @fields = split /\s+/, $in;
		my $cnum = $fields[0];	# regard the 1st field as the cluster_ID

		unless (defined $last_index) {
			(defined $cnum) and ($cnum2sbflds->{$cnum} = \@fields) ; 
			next;
		}
		(@fields > $last_index) or next;

		my @subfields = ();
		foreach my $indx (@{$indices}) { push  @subfields, $fields[$indx]; }
		$cnum2sbflds->{$cnum} = \@subfields;
	}
	close ($fh);

	return \$header;

}		# end of the subroutine fetch_fields #


# fetch_chunk: This subroutine receives a chunk of information, separated by blank lines.
#		and consisting of multiple lines in general.
#		Then it returns a 2-dimensional hash:
#		(\@info_line1, \@info_line2, ...),
#		where @info_linem consists of all fields in the  line m.
#		If the flag instructs, it removes the 1st line from the list
#		and returns its reference.
#
sub fetch_chunk (\$$) {

	my ($in, $fl_title) = @_;

	my @lines = split /\n/, ${$in};
	my $title = ($fl_title and @lines) ? (shift @lines) : '';

	my @output  = ();

	while (my $line = shift @lines) {

		my @fields = split /\s+/, $line;
		@fields and (push @output , \@fields) ;
	}

	return (\@output, \$title);

}		# end of the subroutine fetch_chunk #


# open_input_file: This subroutine receives a filename, opens the file as an input,
#			and  return the filehandle.
#
# REVISED on Sunday, July 2, 2006:
#	The subroutine now behaves in a more reasonable manner.
#
sub open_input_file (\$) {

	my ($infile) =  @_;
	my $fh;
	(my $infile_name = ${$infile}) =~ s/\.gz$//;
	##open ($fh, "< $infile_name" ) or open ($fh, "gunzip -c $infile_name.gz |" )
	##	or die "\nCannot fork or open $infile_name(.gz): $!\n";
	unless ( open ($fh, "< $infile_name" ) or open ($fh, "gunzip -c $infile_name.gz |" ) ) {
	    print STDERR "[ERROR 'open_input_file'] Cannot fork or open $infile_name(.gz): $!\n";
	    return undef;
	}

	return $fh;

}	# end of the subroutine open_input_file #


# open_input_file2:
#	This is a revised version of 'open_input_file'
#	that is supposed to behave in a more reasonable manner.
#
# open_input_file:
#	This subroutine receives a filename, opens the file as an input,
#	and  return the filehandle.
#
sub open_input_file2 (\$) {

	my ($infile) =  @_;
	my $fh;

	if (${$infile} =~ m/\.gz$/ ) { # Likely to be a gzipped file #
	    ##if (( -f ${$infile} ) and ( -r ${$infile} )) {
	    if ( (( -f ${$infile} ) or  ( -l ${$infile} )) and ( -r ${$infile} )) { # MODIFIED on 2020/04/30. (1) #

		unless (open ($fh, "gunzip -c ${$infile} |") ) {
		   print STDERR "[ERROR 'open_input_file2'] Cannot fork or open ${$infile}: $!\n";
		   return undef;
		}
	    } else {
		print STDERR "[WARNING 'open_input_file2'] CanNOT FIND the input file '${$infile}'. I will try the name without the '.gz' suffix ...\n";
		(my $substitute = ${$infile}) =~ s/\.gz$//;
		##if (( -f $substitute ) and ( -r $substitute )) {
		if ( (( -f $substitute ) or ( -l $substitute)) and ( -r $substitute )) { # Modified on 2020/04/30. (2) #

		   unless (open ($fh, "< $substitute") ) {
			print STDERR "[ERROR 'open_input_file2'] Cannot open '$substitute': $!\n";
			return undef;
		   }
		} else {
		   print STDERR "[ERROR 'open_input_file2'] CanNOT FIND a file named '$substitute' either !!\n";
		   return undef;
		}
	    }
	} else { # the input name itself suggests an ungzipped file #

	    ##if (( -f ${$infile} ) and ( -r ${$infile} )) {
	    if ( (( -f ${$infile} ) or ( -l ${$infile} ))  and ( -r ${$infile} )) { # Modified on 2020/04/30. (3) #

		unless (open ($fh, "< ${$infile}") ) {
		    print STDERR "[ERROR 'open_input_file2'] Cannot open '${$infile}': $!\n";
		    return undef;
		}

	    ##} elsif (( -f "${$infile}.gz" ) and ( -r "${$infile}.gz" )) {
	    } elsif ( (( -f "${$infile}.gz" ) or ( -l "${$infile}.gz" )) and ( -r "${$infile}.gz" )) { # Modified on 2020/04/30. (3+1) #

		unless (open ($fh, "gunzip -c ${$infile}.gz |") ) {
		   print STDERR "[ERROR 'open_input_file2'] Cannot fork or open ${$infile}.gz: $!\n";
		   return undef;
		}
	    } else {
		print STDERR "[ERROR 'open_input_file2'] CanNOT FIND the input file ${$infile}(.gz): $!\n";
		return undef;
	    }
	}

	return $fh;

}	# end of the subroutine open_input_file2 #


# open_output_file: This subroutine receives a filename, opens the file as an output,
#			and returns the filehandle.
#
sub open_output_file (\$$) {

	my ($outfile, $fl_compress) = @_;
	(my $outfile_name = ${$outfile}) =~ s/\.gz$//;
	my $fh;
	if ($fl_compress) {	# compress the output # 
	    ##open ($fh, "| gzip > $outfile_name.gz" )
	    ##	or  die "\nCannot fork or open $outfile_name.gz: $!\n";
	    unless ( open ($fh, "| gzip > $outfile_name.gz" ) ) {
		print STDERR "[ERROR 'open_output_file'] Cannot fork or open $outfile_name.gz: $!\n";
		return undef;
	    }

	} else {	# leave the output uncompressed #
	    ##open ($fh, "> $outfile_name" ) or die "\nCannot open $outfile_name: $!\n";
	    unless ( open ($fh, "> $outfile_name" ) ) {
		print STDERR "\nCannot open $outfile_name: $!\n";
		return undef;
	    }
	} 

	return $fh;
}


# list_sessions: This subroutine looks into a directory,
#		 and returns the session numbers.
#		It is originally made for summarizing the results
#		obtained via GridEngine.
#
sub list_sessions (\$$$) {

	my ($indir, $pref, $suff) = @_;
 
	my $infiles = enlist_files (${$indir}, '-', $pref, $suff);

	my @sessions = ();
	my $len_pref = length $pref;
	my $len_suff = length $suff;

	while (my $filename = shift @{$infiles} ) {

		my $session = $filename;
		substr ($session, 0, $len_pref, '' );
		substr ($session, - $len_suff, $len_suff, '');
		##$session =~ s/^_+//;
		$session =~ s/^[_\.]+//;
		$session =~ s/\.$//;
		push @sessions, $session;
	}

	return \@sessions;

}			# end of the subroutine list_sessions #


# list_sessions_ext:
#	This subroutine is a extended version of 'list_sessions'.
#	It now can retrieve directory numbers as well,
#	and can specify the conditions (on the permission) in a finer manner.
#
# CONDITIONS: 	[ $type = 'd' for a directory, '-' for a normal file; ] ... for before the revision on 2020/04/30.
#		$type = 'd' for a directory, '-' for a normal file, 'l' for a symbolic link; # Modified on 2020/04/30.
#		$prefix = '' if unspecified;
#		$suffix = '' if unspecified;
#	
# list_sessions:
#	This subroutine looks into a directory,
#	and returns the session numbers.
#	It is originally made for summarizing the results
#	obtained via GridEngine.
#
sub list_sessions_ext (\$$$$) {

	my ($indir, $pref, $suff, $type) = @_;
 
	my $infiles = enlist_files (${$indir}, $type, $pref, $suff);

=for DEBUGGING the subroutine
	print STDERR "\n< Subroutine 'list_sessions_ext' >\n\n";
	print STDERR "Input files:\n", join ("\n", @{$infiles}), " .\n\n";
=cut

	my @sessions = ();
	my $len_pref = length $pref;
	my $len_suff = length $suff;

	while (my $filename = shift @{$infiles} ) {

		my $session = $filename;
		substr ($session, 0, $len_pref, '' );
		substr ($session, - $len_suff, $len_suff, '');
		##$session =~ s/^_+//;
		$session =~ s/^[_\.]+//;
		$session =~ s/\.$//;
		push @sessions, $session;
	}

	return \@sessions;

}			# end of the subroutine list_sessions_ext #


# fetch_entries_into_array: This subroutine reads in a list of descriptions, whose format is:
#		^{entry_name(or ID)}  description${separtor}$,
#		and returns an ARRAY OF ARRAYS:
#		([entry_name (or ID), \"the whole line" or \"description"], ...).
#
# INPUT: (\$infile, \@out_array, $fl_desc, $fl_header, $separator), where
#	$fl_desc = 1 when you want to extract only the description parts.
#		 = 0 (or undef or '') when you  don't mind storing the whole line.
#	$fl_header =1 when the list has a header prepended,
#		  = 0 otherwise.
#
sub fetch_entries_into_array (\$\@$$$) {

	my ($infile, $outlist, $fl_desc, $fl_header, $separator) = @_;
	local $/ = $separator;
	my $fh = open_input_file (${$infile}) ; 

	my $header = $fl_header ?  <$fh> : '' ;

	while (my $in = <$fh> ) {

		$in =~ s/^\s+//;
		my ($id, $desc) = split /\s+/, $in, 2;
		(defined $desc) or next;

		my @entry = $fl_desc ? ($id, \$desc) : ($id, \$in);
		push @{$outlist}, \@entry;

	}
	close ($fh);

	return \$header;

}		# end of the subroutine fetch_entries_into_array #


# fetch_entries_into_hash: This subroutine reads in a list of descriptions, whose format is:
#		^{entry_name(or ID)}  description...${separator}$ ,
#		and returns  a HASH:
#		(entry_name (or ID) => \"the whole line of the entry" or \"description", 
#		...).
#
# INPUT: (\$infile, \%out_hash, $fl_desc, $fl_header, $separator), where
#	$fl_desc = 1 when you want to extract only the description parts.
#		 = 0 (or undef or '') when you  don't mind storing the whole line.
#	$fl_header = 1 when the list has a header prepended,
#		  = 0 otherwise.
#
sub fetch_entries_into_hash (\$\%$$) {

	my ($infile, $outlist, $fl_desc, $separator) = @_;
	local $/ = $separator;
	my $fh = open_input_file (${$infile}) ; 

	while (my $in = <$fh> ) {

		$in =~ s/^\s+//;
		my ($id, $desc) = split /\s+/, $in, 2;
		(defined $desc) or next;

		my @entry = $fl_desc ? ($id, \$desc) : ($id, \$in);
		$outlist->{$id} = \@entry;
	}
	close ($fh);

}		# end of the subroutine fetch_entries_into_hash #

# construct_label2indx:
#	This subroutine constructs a hash table:
#	%label2indx = (the label of a field => the corresponding index, ...).
#
sub construct_label2indx (\$) {

	my ($header) = @_;

	my @labels_fields = split /\s+/, ${$header};

	my %label2indx;

	for (my $k =0; $k < @labels_fields; $k++) {
		$label2indx{$labels_fields[$k]} = $k;
	}

	return \%label2indx;

}		# end of the subroutine construct_label2indx #


# const_indices_from_labels:
#	This subroutine constructs an array of indices for necessary fields
#	by using a hash table %label2indx = (a label => the corresponding index, ...),
#	and a list of labels.
#
sub const_indices_from_labels (\%\@) {

	my ($label2indx, $labels) = @_;

	my @indices = ();

	foreach my $label (@{$labels}) {

	    my $indx = $label2indx->{$label};
	    unless (defined $indx) {
		print STDERR "NO INDEX corresponding to the label '$label'.\n";
		return undef;
	    }
	    push @indices, $indx;
	}

	return \@indices ;

}			# end of the subroutine const_indices_from_labels #


# tfm_labels2indices:
#	This subroutine transforms a set of labels into
#	a set of corresponding indices,
#	with the help of an object header line.
#	(This is in effect a serial combination of the subroutines
#	'construct_label2indx' and 'construct_indices_from_labels'.
#	If you prefer them to this subroutine, then use them.)
#
# INPUT: (\@labels, \$header), where
#	@labels stores the labels whose field indices you want;
#	$header is the header in question.
#
# OUTPUT: \@indices, which stores the field indices you want.
#	 = undef if the process failed.
#
sub tfm_labels2indices (\@\$) {

	my ($labels, $header) = @_;

	my %tags2indx; # ($field_tag => $indx, ...)
	my @field_tags = split /\s+/, ${$header};
	my $ct_fld = @field_tags;
	for (my $i=0; $i<$ct_fld; $i++) { $tags2indx{$field_tags[$i]} = $i; }

	my @indices = ();
	foreach my $label (@{$labels}) { # The 2nd foreach-loop #
	    my $indx = $tags2indx{$label};
	    unless (defined $indx) {
		print STDERR "[ERROR 'tfm_labels2indices'] CanNOT FIND the label '$label' in the input list of field tags:\n";
		print STDERR "  '", join ('  ', @field_tags), "' .\n\n";
		return undef;
	    }
	    push @indices, $indx;
	}			# End of thd 2nd foreach-loop #

	return  \@indices;

}		## End of the subroutine 'tfm_labels2indices' ##


# tfm_labels2indices_gen:
#	This revised version of the subroutine 'tfm_labels2indices'
#	can use any field separator, which is specified by an input argument.
#
# tfm_labels2indices:
#	This subroutine transforms a set of labels into
#	a set of corresponding indices,
#	with the help of an object header line.
#	(This is in effect a serial combination of the subroutines
#	'construct_label2indx' and 'construct_indices_from_labels'.
#	If you prefer them to this subroutine, then use them.)
#
# INPUT: (\@labels, \$header, $fld_sep), where
#	@labels stores the labels whose field indices you want;
#	$header is the header in question;
#	$fld_sep is the pre-compiled regular expression for
#	the specified field separator.
#
# OUTPUT: \@indices, which stores the field indices you want.
#	 = undef if the process failed.
#
sub tfm_labels2indices_gen (\@\$$) {

	my ($labels, $header, $fld_sep) = @_;

	my %tags2indx; # ($field_tag => $indx, ...)

	   ## The only major change in revision to 'tfm_labels2indices_gen' (on Dec 25 (Thu), 2008) ##
	my @field_tags = split /$fld_sep/, ${$header};
	(@field_tags>0) and ($field_tags[$#field_tags] =~ s/\n+$//);
	##my @field_tags = split /\s+/, ${$header};

	my $ct_fld = @field_tags;
	for (my $i=0; $i<$ct_fld; $i++) { $tags2indx{$field_tags[$i]} = $i; }

	my @indices = ();
	foreach my $label (@{$labels}) { # The 2nd foreach-loop #
	    my $indx = $tags2indx{$label};
	    unless (defined $indx) {
		print STDERR "[ERROR 'tfm_labels2indices_gen'] CanNOT FIND the label '$label' in the input list of field tags:\n";
		print STDERR "  '", join ('  ', @field_tags), "' .\n\n";
		return undef;
	    }
	    push @indices, $indx;
	}			# End of thd 2nd foreach-loop #

	return  \@indices;

}		## End of the subroutine 'tfm_labels2indices_gen' ##


# readin_matrix:
#	This subroutine reads in a matrix of the form:
#
#              {tag}_column1 ...
#{tag}_row1    {entry}_row1_column1 ...
# ...           ....
#
# 	and stores the matrix in a hash:
#	%matrix = ('{tag}_rowk:{tag}_columnl' => {entry}_rowk_columnl, ...),
#	':' is replaceable to any delimiter of your choice.
#
# INPUT: (\$input_file, $delimiter, $head_skip)
#	where $head_skip is the heading part of lines you want to skip reading
#
# OUTPUT: (\%matrix, \@tags_row, \@tags_column)
#
sub readin_matrix (\$$$) {

	my ($infile, $delimiter, $head_skip ) = @_;
	local $/ = "\n";

	my $len_skip = length ($head_skip);

	my $fh = open_input_file (${$infile}) ;

	my %matrix;
	my @tags_row = my @tags_column = ();
	my $ct_columns = 0;

	while (my $in = <$fh>) {

	    ($len_skip <= length ($in) ) and ($head_skip eq substr ($in, 0, $len_skip) ) and next;

	    my @fields = split /\s+/, $in;
	   (@fields > 1) or next;  # skip rows without any field but a tag

	    my $tag_row = shift @fields;
	   if ($tag_row eq '') {
		if (@tags_column < 1) {
		# Regard the first appearing row of this type as a set of column tags
		   @tags_column = @fields;
		   $ct_columns = @tags_column;
		}
		# Ignore the second and later rows #

		next;
	   }

	   push @tags_row, $tag_row;

	   my $ub = (scalar (@fields) <  $ct_columns) ? scalar (@fields) : $ct_columns;

	   for (my $k=0; $k<$ub; $k++) { 
		my $entry = $fields[$k];
		my $tag_column = $tags_column[$k];
		my $key = $tag_row . $delimiter . $tag_column;
		$matrix{$key} = $entry;
	   }

	}
	close ($fh);

	return (\%matrix, \@tags_row, \@tags_column);

}			# end of the subroutine readin_matrix #


# extend_matrix:
#	This subroutine 'extend's the matrix fetched by the subroutine 'readin_matrix'.
#	It assumes that the input matrix is square.
#
# INPUT: (\%matrix, \@tags_column, $delimiter, \@tags_ext, \@tags_src), where
#	%matrix = ($tags_column[$k] . $delimiter . $tags_column[$l] => $element, ...);
#	@tags_ext stores the tags that are to be newly added;
#	$tags_src[$m] gives the source matrix elements for the tag $tags_ext[$m].
#
# OUTPUT: 1 if the process succeeded,
#	 0 if it failed.
#
sub extend_matrix (\%\@$\@\@) {

	my ($matrix, $tags_column, $delimiter, $tags_ext, $tags_src) = @_;

	my $ct_ext = @{$tags_ext};
	unless (@{$tags_src} == $ct_ext) {
	    my $ct_src = @{$tags_src};
	    print STDERR "[ERROR 'extend_matrix'] ILLIGAL INPUT: #{extended} = $ct_ext vs #{sources} = $ct_src !\n\n";
	    return 0;
	}

	for (my $m=0; $m <$ct_ext; $m++) { # The outer for-loop #
	    my $ext = $tags_ext->[$m];
	    my $src = $tags_src->[$m];

		# Add rows and columns #
	    foreach my $tag (@{$tags_column}) { # The 1st inner  for-loop #
		my ($rext_tag, $cext_tag) = ($ext . $delimiter . $tag, $tag . $delimiter . $ext);
		my ($rsrc_tag, $csrc_tag) = ($src . $delimiter . $tag, $tag . $delimiter . $ext);
		my $src_ele = (defined $matrix->{$rsrc_tag}) ?
			$matrix->{$rsrc_tag} : $matrix->{$csrc_tag} ;
		unless (defined $src_ele) {
		    print STDERR "[WARNING 'extend_matrix'] MISSING MATRIX ELEMENT for the source pair '$rsrc_tag' (or its mirror image '$csrc_tag') !\n\n";
		    next;
		}

		$matrix->{$rext_tag} = $matrix->{$cext_tag} = $src_ele;
	    }			# End of the 1st inner for-loop #

		# Fill the bottom-right corner #
	    for (my $n=0; $n<$ct_ext; $n++) { # The 2nd inner for-loop #
		my $ext2 = $tags_ext->[$n];
		my $src2 = $tags_src->[$n];
		my ($ext_pair1, $ext_pair2) = ($ext . $delimiter . $ext2, $ext2 . $delimiter . $ext);
		my ($src_pair1, $src_pair2) = ($src . $delimiter . $src2, $src2 . $delimiter . $src);
		my $src_pele = (defined $matrix->{$src_pair1}) ?
			$matrix->{$src_pair1} : $matrix->{$src_pair2} ;
		unless (defined $src_pele) {
		    print STDERR "[WARNING 'extend_matrix'] MISSING MATRIX ELEMENT for the source pair '$src_pair1' (or its mirror image '$src_pair2') !\n\n";
		    next;
		}

		$matrix->{$ext_pair1} = $matrix->{$ext_pair2} = $src_pele ;
	    }			# End of the 2nd inner for-loop #

	}		# End of the outer for-loop #

	return 1;

}		# End of the subroutine 'extend_matrix' #


1;		# a 'spell which makes the module "use"-able and/or "require"-able #



