package MyLikelihoods_respattern_msa_dna;
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyLikelihoods_respattern_msa_dna.pm:
#	This package is a collection of subroutines
#	to calculate the likelihood of the residue pattern of a given MSA (of DNA sequences)
#	and its ingredients.
#	[CREATION STARTED on Apr 29 (Fri), 2016, ENDED on May 2 (Mon), 2016.]
#

use strict;
use warnings;

use MyTaxonomy qw(
	if_ancestor fetch_ancestry calculate_abs_depth
	fetch_ext_offsprings_nr fetch_all_offsprings_nr
	);

use MyTreeManipu_II qw( );
	# extract_sgltree_from_file output_tree 

use MyTreeManipu_III qw( make_node2abs_depth );

=set aside for potential use
use MyTreeManipu_IV qw( bi2multifurcated_top construct_neat_subtree );
use MyPAML_parser qw( retrieve_aa_freqs prepare_paml_ctrl );
=cut

use Exporter;

our @ISA =qw(Exporter);

our @EXPORT_OK = qw( 
	calc_transition_mtrx_TN
	calc_br2tr_mtrx_TN
	calc_node2ntfreqs_td
	init_prob_descresstats_cond_ancstat2
	calc_prob_descresstats_cond_ancstat2
	calc_joint_prob_ancresstats_and_ancstat2
	calc_clm_probs
	br_calc_sbst_prob_sgl_clm
	);
	## mk_node2set_sw_gapstats
	##calc_prob_descresstats


	### Default values of the Parameters ###

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;

our @NUCLEOTIDES = qw( A G C T ); # Unless otherwise stated, nucleotide MSA scores will be calculated according to this order (or index assignment) of the nucleotide bases.
our $CT_NUCLS = @NUCLEOTIDES;

our %NUCL2INDX; # = ($nucleotide => {index in @NUCLEOTIDES}, ...)
for (my $k=0; $k<$CT_NUCLS; $k++) { $NUCL2INDX{$NUCLEOTIDES[$k]} = $k; }

our $LOG_ZERO = 'log(ZERO)';


	## The following parameters specify the verboseness of the test output.
	## (0, silent; 1, very modest; 2, modest; 3, medium; 4, verbose; 5, very verbose; 6, extremely verbose.)

	# For 'calc_prob_descresstats_cond_ancstat2'. #
our $TEST_LEVEL_cpdca = 0;
##our $TEST_LEVEL_cpdca = 6;

	# For 'calc_joint_prob_ancresstats_and_ancstat2'. #
our $TEST_LEVEL_cjpaaa = 0;
##our $TEST_LEVEL_cjpaaa = 6;


### 5 important hash tables which describe a tree ###
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.



# calc_transition_mtrx_TN:
#	[NOTE: This subroutine was imported from the package, "ComplLiMment_P.ver0.6.1.3," on Apr 29 (Fri), 2016.]
#	This subroutine calculates (the integrated version of) the transition matrix
#	for a branch of a given length (measured in the number of expected substitutions per site),
#	in the Tamura-Nei nucleotide substitution model.
#	(CREATED on Aug 12 (Tue), 2014)
#
# INPUT: (\@in_equil_freqs, $a1, $a2, $b, $dist), where
#	$in_equil_freqs[$b] is the equilibrium frequency of the nucleotide $NUCLEOTIDES[$b];
#	$a1 and $a2 are the rate parameters for the transitional substitutions between A and G, and between C and T, respectively;
#	$b is the rate parameter for the transversional substitutions;
#
#	More specifically, we have the Tamura-Nei instantaneous substitution matrix as:
#	[M^{TN}_{$i,$j}] =	| ---           $a1 * pi_G	$b * pi_C	$b * pi_T  |
#				| $a1 * pi_A	---       	$b * pi_C	$b * pi_T  |
#				| $b * pi_A	$b * pi_G	---      	$a2 * pi_T |
#				| $b * pi_A	$b * pi_G	$a2 * pi_C	---        |
#	Here, M^{TN}_{$i,$j} represents the substitution rate from $NUCLEOTIDES[i] to $NUCLEOTIDES[$j].
#	I omitted the diagonal elements, which are always specified so that each row sums up to give zero.
#
#	$dist is the distance between the both ends of a branch,
#	measured in terms of the expected number of substitutions per nucleotide site.
#
# OUTPUT: \@tr_matrix_tn, where
#	$tr_matrix_tn[$i]->[$j] is the element of the Tamura-Nei transition matrix,
#	which gives the probability that $NUCLEOTIDES[$j] is found on one end of the branch,
#	on condition that $NUCLEOTIDES[$i] is found on the other end. 
#	(That is, the conditional probability of the transition from the $i th to the $j th nucleotide.)
#
sub calc_transition_mtrx_TN (\@$$$$) {

	my ($in_equil_freqs, $a1, $a2, $b, $dist) = @_;

	my ($pi_A, $pi_G, $pi_C, $pi_T) = @{$in_equil_freqs}; # Base frequencies.

	my $pi_R = $pi_A + $pi_G; # Purine frequency.
	my $pi_Y = $pi_C + $pi_T;  # Pyrimidine frequency.

	my ($tpi_A, $tpi_G) = ($pi_R>0) ? ($pi_A/$pi_R, $pi_G/$pi_R) : (1/2, 1/2); # Relative frequencies between purines.

	unless ($pi_R>0) {
	    print STDOUT "\n[WARNING 'calc_transition_mtrx_TN'] The purine frequency is ZERO ... I arbitrarily set ", '($tpi_A, $tpi_G) = ', "($tpi_A, $tpi_G) ...\n\n";
	}

	my ($tpi_C, $tpi_T) = ($pi_Y>0) ? ($pi_C/$pi_Y, $pi_T/$pi_Y) : (1/2, 1/2); # Relative frequencies between pyrimidines.

	unless ($pi_Y>0) {
	    print STDOUT "\n[WARNING 'calc_transition_mtrx_TN'] The pyrimidine frequency is ZERO ... I arbitrarily set ", '($tpi_C, $tpi_T) = ', "($tpi_C, $tpi_T) ...\n\n";
	}

		# Column vectors. #
	my @cvec1 = (1, 1, 1, 1);
	my @cvec2 = ($tpi_G, -$tpi_A, 0, 0);
	my @cvec3 = (0, 0, $tpi_T, -$tpi_C);
	my @cvec4 = ($pi_Y, $pi_Y, -$pi_R, -$pi_R);

		# Row vectors. #
	my @rvec1 = ($pi_A, $pi_G, $pi_C, $pi_T);
	my @rvec2 = (1, -1, 0, 0);
	my @rvec3 = (0, 0, 1, -1);
	my @rvec4 = ($tpi_A, $tpi_G, -$tpi_C, -$tpi_T);

		# Exponential factors. #
	my $n_fac = 2*($a1 * $pi_A * $pi_G + $a2 * $pi_C * $pi_T + $b * $pi_R * $pi_Y); # Normalization factor.
	my $exp2 = exp ( - ($pi_R * $a1 + $pi_Y * $b) * $dist /$n_fac );
	my $exp3 = exp ( - ($pi_Y * $a2 + $pi_R * $b) * $dist /$n_fac );
	my $exp4 = exp ( - $b * $dist /$n_fac );

	my @tr_matrix_tn = ();

	for (my $i=0; $i< $CT_NUCLS; $i++) { # Main outer for-loop (over initial nucleotides). #

	    $tr_matrix_tn[$i] = []; # Initialize the row.

	    for (my $j=0; $j<$CT_NUCLS; $j++) { # Main inner for-loop (over final nucleotides). #

		$tr_matrix_tn[$i]->[$j] =
		   $cvec1[$i] * $rvec1[$j] + $exp2 * $cvec2[$i] * $rvec2[$j] 
		   + $exp3 * $cvec3[$i] * $rvec3[$j] + $exp4 * $cvec4[$i] * $rvec4[$j] ;

	    }				# End of the main inner for-loop (over final nucleotides). #

	}			# End of the main outer for-loop (over initial nucleotides).

	return \@tr_matrix_tn ;

}	# End of the subroutine 'calc_transition_mtrx_TN'. #


# calc_br2tr_mtrx_TN:
#	[NOTE: This subroutine was imported from the package, "ComplLiMment_P.ver0.6.1.3," on Apr 29 (Fri), 2016.]
#	This subroutine calculates (the integrated version of) the transition matrices
#	in the Tamura-Nei nucleotide substitution model
#	for tree branches, each with a given length (measured in the number of expected substitutions per site),
#	(CREATED on Aug 12 (Tue), 2014)
#
#	For more details on each transition matrix, see the heading of 'calc_transition_mtrx_TN'.
#
# INPUT: (\%br_attr, \@in_equil_freqs, $a1, $a2, $b), where
#	%br_attr = ($branch_id => \@attr, ...), with
#	$attr[$INDX_LENGTH] giving the branch length (measured in the expected number of base substitutions per site;
#
#	$in_equil_freqs[$b] is the equilibrium frequency of the nucleotide $NUCLEOTIDES[$b];
#	$a1 and $a2 are the rate parameters for the transitional substitutions between A and G, and between C and T, respectively;
#	$b is the rate parameter for the transversional substitutions.
#
# OUTPUT: \%br2tr_mtrx_tn, where
#	@{$br2tr_mtrx_tn{$br_id}} is the transition matrix in the Tamura-Nei model 
#	that describes the transition probabilities between both ends of the branch $br_id.
#	More specifically,
#	$br2tr_mtrx_tn{$br_id}->[$i][$j] is the probability that the node $br_id shows $NUCLEOTIDE[$j]
#	given that the parent of $br_id shows $NUCLEOTIDE[$i].
#
sub calc_br2tr_mtrx_TN (\%\@$$$) {

	my ($br_attr, $in_equil_freqs, $a1, $a2, $b) = @_;

	my %br2tr_mtrx_tn;

	foreach my $br_id (keys %{$br_attr}) { # Main foreach-loop (over branches). #

	    my $attr = $br_attr->{$br_id};
	    my $brlen = $attr->[$INDX_LENGTH];

	    my $tr_matrix_tn = calc_transition_mtrx_TN (@{$in_equil_freqs}, $a1, $a2, $b, $brlen);
	    unless (defined $tr_matrix_tn) {
		print STDOUT "[ERROR 'calc_br2tr_mtrx_TN'] 'calc_transition_mtrx_TN' FAILED on the branch $br_id : $! !\n\n";
		return undef;
	    }

	    $br2tr_mtrx_tn{$br_id} = $tr_matrix_tn;

	}			# End of the main foreach-loop (over branches). #

	return \%br2tr_mtrx_tn;

}	# End of the subroutine 'calc_br2tr_mtrx_TN'. #


# calc_node2ntfreqs_td:
#	This subroutine calculates the residue frequencies at all nodes in a top-down manner.
#	(CREATED on May 2 (Mon), 2016)
#
#	[Formulas]
#	Given the residue frequencies at the root, { p(\omega, n^{R}) }_{\omega in \Omega},
#	the residue frequencies at the other nodes are calculated iteratively, as follows:
#
#	p(\omega, n^D(b)) = Sum_{\omega'} p(\omega', n^A(b)) * P(\omega' -> \omega; b).
#	Here, n^A(b) and n^D(b) are the "ancestral" and "descendant" nodes,
#	which are at the upper-end and the lower-end, respectively, of the branch b.
#	P(\omega' -> \omega; b) is the transition probability along the branch b
#	that the state at n^D(b) is \omega conditioned on that the state at n^A(b) is \omega'.
#
# INPUT: (\@ntfreqs_root, \%br2tr_mtrx, \%node2ch, $topnode_id), where
#
#	$ntfreqs_root[$i] is the frequency of $NUCLEOTIDES[$i] at the root;
#
#	@{$br2tr_mtrx{$br_id}} is the transition matrix
#	that describes the transition probabilities between both ends of the branch $br_id.
#	More specifically,
#	$br2tr_mtrx{$br_id}->[$i][$j] is the probability that the node $br_id shows $NUCLEOTIDES[$j]
#	given that the parent node of $br_id shows $NUCLEOTIDES[$i];
#
#	%node2ch = ($node_id => \@child_nodes, ...) for the interior nodes of the tree;
#
#	$topnode_id is the ID of the top-node (i.e., the root).
#
# OUTPUT: \%node2ntfreqs, where
#	%node2ntfreqs = ($node_id => \@ntfreqs, ...), where
#	$ntfreqs[$i] is the frequency of $NUCLEOTIDES[$i] at the node, $node_id.
#
sub calc_node2ntfreqs_td (\@\%\%$) {

	my ($ntfreqs_root, $br2tr_mtrx, $node2ch, $topnode_id) = @_;

	my @set_pa_sbjct_branches = ($topnode_id);

	my %node2ntfreqs = ($topnode_id => $ntfreqs_root); # Initialize the output. #


	while (@set_pa_sbjct_branches>0) { # The main outermost while-loop (continues until the subjects run out). #

	   my @newset = ();

	   foreach my $pa_sbj (@set_pa_sbjct_branches) { # The main outer foreach-loop (over the subject parents). #

		my $set_sbjcts = $node2ch->{$pa_sbj};
		((defined $set_sbjcts) and (@{$set_sbjcts}>0)) or next; # Skip if there is no subject branches. #

		push @newset, @{$set_sbjcts}; # "Increment" the new set of subject parents. #

		my $ntfreqs_pa = $node2ntfreqs{$pa_sbj};
		unless ((defined $ntfreqs_pa) and (@{$ntfreqs_pa}>0)) {
		    print STDOUT "[ERROR 'calc_node2ntfreqs_td'] NO NUCLEOTIDE FREQUENCY is assigned to the parent node (ID: $pa_sbj): $! !\n\n";
		    return undef;
		}

		foreach my $node_id (@{$set_sbjcts}) { # Main middle foreach-loop (over the subject nodes). #

		    my $tr_mtrx = $br2tr_mtrx->{$node_id};

		    my @ntfreqs = ();

		    for (my $i=0; $i<$CT_NUCLS; $i++) { # Main inner for-loop (over nucleotides). #
			my $freq = 0;
			for (my $j=0; $j<$CT_NUCLS; $j++) { $freq += $ntfreqs_pa->[$j] * $tr_mtrx->[$j][$i]; }
			$ntfreqs[$i] = $freq;
		    }				# End of the main inner for-loop (over nucleotides). #

		    $node2ntfreqs{$node_id} = \@ntfreqs;

		}			# End of the main middle foreach-loop (over the subject nodes). #

	   }					# End of the main outer foreach-loop (over the subject parents). #
	   @set_pa_sbjct_branches = @newset;

	}		# END of the main outermost while-loop (continues until the subjects run out.) #

	return \%node2ntfreqs;

}	# End of the subroutine 'calc_node2ntfreqs_td'. #

	### RESTART FROM HERE ###

# init_prob_descresstats_cond_ancstat2:
#	This satellite subroutine of 'calc_prob_descresstats_cond_ancstat2' initializes the main outputs of
#	the master subroutine, %node2set_sw_cond_probs and %node2set_sw_extcond_probs, so that
#
#	$node2set_sw_cond_probs{$node}->[$s][$i] = 1 if the $s th site of the sequence at the external node $node is occupied by $NUCLEOTIDES[$i],
#						= 0 if occupied by aother nucleotide, and
#	$node2set_sw_cond_probs{$node}->[$s] = [] (empty) if the site is occupied by a gap;
#
#	and so that
#
#	$node2set_sw_extcond_probs{$node}->[$s][$i] = Sum_{$j} $br2tr_mtrx{$br_id}->[$i][$j] * $node2set_sw_cond_probs{$node}->[$s][$j]
#						    if the $s th site of the sequence at the external node $node is occupied by a base,
#						    = 1 if the site is occupied by a gap.
#
#	This subroutine leaves the input MSA intact.
#	It was created by modifying the subroutine, "init_prob_descresstats_cond_ancstat,"
#	in the package, "ComplLiMment_P.ver0.6.1.3."
#	[CREATED on Apr 29 (Fri), 2016.]
#
# init_prob_descresstats_cond_ancstat:
#	This satellite subroutine of 'calc_prob_descresstats_cond_ancstat' initializes the main output of
#	the master subroutine, %node2set_sw_cond_probs, so that
#
#	$node2set_sw_cond_probs{$node}->[$s][$i] = 1 if the $s th site of the sequence at the exterior node $node is occupied by $NUCLEOTIDES[$i],
#						= 0 if occupied by aother nucleotide, and
#	$node2set_sw_cond_probs{$node}->[$s] = [] (empty) if the site is occupied by a gap.
#
#	This subroutine leaves the input MSA intact.
#
#	(CREATED on Sep 19 (Fri), 2014, as a subroutine in the package, "ComplLiMment_P.ver0.6.1.3.")
#
# INPUT: (\%node2set_sw_cond_probs, \%node2set_sw_extcond_probs, \%br2tr_mtrx, \@in_msa, \@seqnames, \%name2id, $gap), where
#
#	%node2set_sw_cond_probs is one of the two main outputs of 'calc_prob_descresstats_cond_ancstat2', which this subroutine will initialize;
#	%node2set_sw_extcond_probs is the other main output of 'calc_prob_descresstats_cond_ancstat2', which this subroutine will also initialize;
#
#	$br2tr_mtrx{$br_id}->[$i][$j] is the probability that the node $br_id shows $NUCLEOTIDE[$j]
#	given that the parent of $br_id shows $NUCLEOTIDE[$i];
#
#	${$in_msa[$k]} is the $k th sequence embedded in the input MSA;
#	$seqnames[$k] is the name (or ID) of the $k th sequence in the input MSA;
#	%name2id = ($node_name => $node_id, ...) for the exterior nodes and the top node (i.e., root) of the tree;
#
#	$gap is the character representing a gap.
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
sub init_prob_descresstats_cond_ancstat2 (\%\%\%\@\@\%$) {

	my ($node2set_sw_cond_probs, $node2set_sw_extcond_probs, $br2tr_mtrx, $in_msa, $seqnames, $name2id, $gap) = @_;

	my $ct_seqs = @{$in_msa};

	unless ($ct_seqs>0) {
	    print STDOUT "[ERROR 'init_prob_descresstats_cond_ancstat2'] ILLEGAL INPUT: The input MSA contains NO Sequences: $! !\n\n";
	    return 0;
	}
	unless (@{$seqnames} == $ct_seqs) {
	    my $ct_seqs2 = @{$seqnames};
	    print STDOUT "[ERROR 'init_prob_descresstats_cond_ancstat2'] ILLEGAL INPUT: The number of sequences in the input MSA (=$ct_seqs) CLASHES the number of sequence names (=$ct_seqs2): $! !\n\n";
	    return 0;
	}
	$gap = uc $gap;

	my $ct_sites = length ( ${$in_msa->[0]} );

	for (my $k=0; $k<$ct_seqs; $k++) { # Main outer for-loop (over sequences in the MSA). #

	    my $seqname = $seqnames->[$k];
	    my $node_id = $name2id->{$seqname};
	    unless (defined $node_id) {
		print STDOUT "[ERROR 'init_prob_descresstats_cond_ancstat2'] ILLEGAL INPUT: The $k th sequence in the MSA (name: $seqname) does NOT correspond to ANY EXTERIOR NODE: $! !\n\n";
		return 0;
	    }

	    my $cp_seq = uc ${$in_msa->[$k]};  # Copy the sequence. #

	    my @set_sw_cond_probs = ();
	    my @set_sw_extcond_probs = (); # ADDED in creation of 'init_prob_descresstats_cond_ancstat2'.

	    for (my $s=0; $s<$ct_sites; $s++) { # Main inner for-loop (over sites). #

		my $char = substr ($cp_seq, 0, 1, '');

		if ($char eq $gap) {
		    $set_sw_cond_probs[$s] = []; # Assign an empty array. #

			# Assign an array consisting only of 1's.
			# (ADDED in creation of 'init_prob_descresstats_cond_ancstat2'.) #
		    my @extcond_probs = ();
		    for (my $i=0; $i<$CT_NUCLS; $i++) { $extcond_probs[$i] = 1; }
		    $set_sw_extcond_probs[$s] = \@extcond_probs;

		    next;
		}

		my $indx = $NUCL2INDX{$char};
		unless (defined $indx) {
		    print STDOUT "[ERROR 'init_prob_descresstats_cond_ancstat2'] ILLEGAL INPUT: The $s th site of the $k th sequence (name: $seqname) is OCCUPIED by an UNEXPECTED CHARACTER, '$char': $! !\n\n";
		    return 0;
		}

		my @probs = ();
		for (my $c=0; $c<$CT_NUCLS; $c++) { $probs[$c] = ($c == $indx) ? 1 : 0; }

		$set_sw_cond_probs[$s] = \@probs;

			# ADDED in creation of 'init_prob_descresstats_cond_ancstat2'. #

		my $tr_mtrx = $br2tr_mtrx->{$node_id};
		unless (defined $tr_mtrx) {
		    print STDOUT "[ERROR 'init_prob_descresstats_cond_ancstat2'] ILLEGAL INPUT: The branch ID $node_id, which leads to the $k th sequence (name: $seqname), is NOT ASSIGNED a transition matrix: $! !\n\n";
		    return 0;
		}

		my @extprobs = ();
		for (my $c=0; $c<$CT_NUCLS; $c++) {
		   my $tr_vec = $tr_mtrx->[$c];
		   $extprobs[$c] = $tr_vec->[$indx];
		}

		$set_sw_extcond_probs[$s] = \@extprobs;

			# END of "ADDED in creation of 'init_prob_descresstats_cond_ancstat2'." #

	    }			# End of the main inner for-loop (over sites). #

	    $node2set_sw_cond_probs->{$node_id} = \@set_sw_cond_probs;

	    $node2set_sw_extcond_probs->{$node_id} = \@set_sw_extcond_probs; # ADDED in creation of 'init_prob_descresstats_cond_ancstat2'. #

	}			# End of the main outer for-loop (over sequences in the MSA). #

	return 1; # The subroutine successfully ended. #

}	# End of the subroutine 'init_prob_descresstats_cond_ancstat'. #



# calc_prob_descresstats_cond_ancstat2:
#	This subroutine calculates probabilities of the residue states of all the descendants
#	of an ancestral node, conditioned on the residue state of that ancestral node,
#	as well as probabilities of the same residue states conditioned on the residue state of the parent of the ancestral node,
#	both of every site in an input MSA.
#	The subroutine leaves the input MSA intact.
#	It was created by modifying the subroutine, "calc_prob_descresstats_cond_ancstat," in the package, "ComplLiMment_P.ver0.6.1.3."
#	[CREATION STARTED on Apr 29 (Fri), 2016, ENDED Apr 30 (Sat), 2016.]
#	[REVISED on 2019/06/28]: The [Formula] was corrected, so that it will no longer be confusing.
#
# calc_prob_descresstats_cond_ancstat:
#	This subroutine calculates probabilities of the residue states of all the descendants
#	of an ancestral node, conditional on the residue state of that ancestral node,
#	of every site in an input MSA.
#	The subroutine leaves the input MSA intact.
#	(CREATED on Sep 19 (Fri), 2014, as a subroutine in the package, "ComplLiMment_P.ver0.6.1.3.")
#
#	[Formula]
#	The probability, p(n)_i = Pr [{s(d)}_n | s(n) = N_i ],
#	that the descendant sequences of node n, {d}_n, show the states, {s(d)}_n, at the s-th site,
#	given that the s-th site at node n is nucleotide N_i, can be calculated by the following formula:
#
#	p(n)_i = Prod_{m in ch(n)}  ep(m)_i,
#	where
#	   [ ep(m)_i =: Sum_{j} { t(n;m)_{i,j} * p(m)_j  = Pr [ {s(d)}_m | s(pa(m)) = N_i ] ] ... OBSOLETE as of 2019/06/28, because it is somewhat confusing.
#	   ep(m)_i =: Sum_{j} { t(pa(m);m)_{i,j} * p(m)_j  = Pr [ {s(d)}_m | s(pa(m)) = N_i ] # CORRECTED on 2019/06/28.
#	   is the probability that the same descendant residue states resulted, given that
#	    the state at the parent node, pa(m), is nucleotide N_i.
#
#	In the above equations, t(n;m)_{i,j} is the substitution transition probability that
#	node m has nucleotide N_j, on condition that node n has nucleotide N_i.
#
#	[NOTE] This subroutine itself can handle multifurcated trees as well,
#		although PRANK analyses cannot, in general.
#
# INPUT: (\@in_msa, \@seqnames, \%name2id, \%node2ch, \%br2tr_mtrx, $gap), where
#	${$in_msa[$k]} is the $k th sequence embedded in the input MSA;
#	$seqnames[$k] is the name (or ID) of the $k th sequence in the input MSA;
#	%name2id = ($node_name => $node_id, ...) for the exterior nodes and the top node (i.e., root) of the tree;
#	%node2ch = ($node_id => \@child_nodes, ...) for the interior nodes of the tree;
#
#	@{$br2tr_mtrx{$br_id}} is the transition matrix
#	that describes the transition probabilities between both ends of the branch $br_id.
#	More specifically,
#	$br2tr_mtrx{$br_id}->[$i][$j] is the probability that the node $br_id shows $NUCLEOTIDES[$j]
#	given that the parent node of $br_id shows $NUCLEOTIDES[$i].
#
#	$gap is the character representing a gap.
#
# OUTPUT: (\%node2set_sw_cond_probs, \%node2set_sw_extcond_probs), where
#	%node2set_sw_cond_probs = ($node_id => \@set_sw_cond_probs, ...) for ALL nodes, where
#	$set_sw_cond_probs[$s]->[$i] is the probability that
#	all the descendant sequences of $node_id show the residue states that are actually found in the $s th site,
#	given that $NUCLEOTIDES[$i] occupies the $s th site of the ancestral node, $node_id.
#
#	But $set_sw_cond_probs[$s] = [] (empty) if $node_id refers to an external node and if the node has its $s th site occupied by a gap.
#
#	%node2set_sw_extcond_probs = ($node_id => \@set_sw_extcond_probs, ...) for ALL nodes except the root, where
#	$set_sw_extcond_probs[$s]->[$i] is the probability of the same residue configuration was $set_sw_cond_probs[$s]->[$i],
#	but conditioned on that $NUCLEOTIDES[$i] occupies the $s th site of the PARENT OF the ancestral node, $node_id.
#
sub calc_prob_descresstats_cond_ancstat2 (\@\@\%\%\%$) {

	my ($in_msa, $seqnames, $name2id, $node2ch, $br2tr_mtrx, $gap) = @_;

	$gap = uc $gap;

	my $ct_seqs = @{$in_msa};
	my $ct_sites = length (${$in_msa->[0]});

	my $topnode_id = $name2id->{$TOP_NODE};

   (print STDOUT "\n< Subroutine 'calc_prob_descresstats_cond_ancstat2'  (#{sequences} = $ct_seqs; #{sites} = $ct_sites; topnode ID = $topnode_id; Gap = '$gap') >\n\n") if ($TEST_LEVEL_cpdca>0);

		# Preliminary (I): Initialize the output hash. #

	my %node2set_sw_cond_probs;
	my %node2set_sw_extcond_probs; # ADDED in creation of 'calc_prob_descresstats_cond_ancstat2'. #

	unless ( init_prob_descresstats_cond_ancstat2 (%node2set_sw_cond_probs, %node2set_sw_extcond_probs, %{$br2tr_mtrx}, @{$in_msa}, @{$seqnames}, %{$name2id}, $gap) ) {
	    print STDOUT "[ERROR 'calc_prob_descresstats_cond_ancstat2'] Subroutine 'init_prob_descresstats_cond_ancstat2' FAILED : $! !\n\n";
	    return undef;
	}
   (print STDOUT "\n ... 'init_prob_descresstats_cond_ancstat2' Ended ...\n\n") if ($TEST_LEVEL_cpdca >1);


   if ($TEST_LEVEL_cpdca>3) {

	print STDOUT "\n[ Initialized Output (1), '", '%node2set_sw_cond_probs', "' ]\n\n";

	foreach my $node (sort {$a <=> $b} keys %node2set_sw_cond_probs) { # 1st outer testing foreach-loop (over exterior nodes). #

	    my $set_sw_cond_probs = $node2set_sw_cond_probs{$node};

	    print STDOUT "( Conditional probabilities for Node ID: $node )\n\n";
	    print STDOUT join ("\t", 'Site_no.', @NUCLEOTIDES), "\n\n";

	    for (my $s=0; $s<$ct_sites; $s++) { print STDOUT join ("\t", $s, @{$set_sw_cond_probs->[$s]}), "\n"; }
	    print STDOUT "\n\n";

	}			# End of the 1st outer testing foreach-loop (over exterior nodes). #


	print STDOUT "\n[ Initialized Output (2), '", '%node2set_sw_extcond_probs', "' ]\n\n";

	foreach my $node (sort {$a <=> $b} keys %node2set_sw_extcond_probs) { # 2nd outer testing foreach-loop (over exterior nodes). #

	    my $set_sw_extcond_probs = $node2set_sw_extcond_probs{$node};

	    print STDOUT "( Extended conditional probabilities for Branch ID: $node )\n\n";
	    print STDOUT join ("\t", 'Site_no.', @NUCLEOTIDES), "\n\n";

	    for (my $s=0; $s<$ct_sites; $s++) { print STDOUT join ("\t", $s, @{$set_sw_extcond_probs->[$s]}), "\n"; }
	    print STDOUT "\n\n";

	}			# End of the 2nd outer testing foreach-loop (over exterior nodes). #

   }	# END of "if ($TEST_LEVEL_cpdca>2) {...}"


		# Preliminary (II): Assign depths to the nodes. #

	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
	unless (defined $node2depth) {
	    print STDOUT "[ERROR 'calc_prob_descresstats_cond_ancstat'] Subroutine 'make_node2abs_depth' FAILED : $! !\n\n";
	    return undef;
	}
   (print STDOUT "\n ... 'make_node2abs_depth' Ended ...\n\n") if ($TEST_LEVEL_cpdca >1);

   if ($TEST_LEVEL_cpdca>2) {
	print STDOUT "\n[ Depths assigned to the nodes ]\n\n";
	print STDOUT join ("\t", 'Node_ID', 'Depth'), "\n\n";
	foreach my $node (sort {$a <=> $b} keys %{$node2depth}) { print STDOUT join ("\t", $node, $node2depth->{$node}), "\n"; }
	print STDOUT "\n\n";
   }


		## MAIN PROCESS: Calculate the conditional probabilities in a BOTTOM-UP manner. ##

   (print STDOUT "\n{{ MAIN PROCESS: Calculate the conditional probabilities from bottom (leaves) to top (root)... }}\n\n") if ($TEST_LEVEL_cpdca>0);


	foreach my $node_id (sort { $node2depth->{$b} <=> $node2depth->{$a} } keys %{$node2ch}) { # Main outer foreach-loop (over interior nodes). #

	    my $depth = $node2depth->{$node_id};

    (print STDOUT "{ Node ID: $node_id (depth = $depth) }\n\n") if ($TEST_LEVEL_cpdca>1);

	    my $child_nodes = $node2ch->{$node_id};
	    my $ct_children = @{$child_nodes};

   (print STDOUT "There are $ct_children child nodes: ", join (', ', @{$child_nodes}), "\n\n") if ($TEST_LEVEL_cpdca>2) ;

=For the predecessor 'calc_prob_descresstats_cond_ancstats'
	    my @setset_sw_ch_cond_probs = ();
	    my @set_tr_mtrices = ();
	    foreach my $ch (@{$child_nodes}) { # Middle preliminary foreach-loop (over child nodes). #

		my $set_sw_ch_cond_probs = $node2set_sw_cond_probs{$ch};
		unless (defined $set_sw_ch_cond_probs) {
		    print STDOUT "[ERROR 'calc_prob_descresstats_cond_ancstats'] NO SET of conditional PROBABILITIES was ASSIGNED TO the child '$ch' of the node '$node_id': $! !\n\n";
		    return undef;
		}
		push @setset_sw_ch_cond_probs, $set_sw_ch_cond_probs;

		my $tr_mtrx = $br2tr_mtrx->{$ch};
		unless (defined $tr_mtrx) {
		    print STDOUT "[ERROR 'calc_prob_descresstats_cond_ancstats'] NO TRANSITION MATRIX was ASSIGNED TO the branch from the node '$node_id' to its child '$ch': $! !\n\n";
		    return undef;
		}
		push @set_tr_mtrices, $tr_mtrx;

	    }		# End of the middle preliminary foreach-loop (over child nodes). #

   if ($TEST_LEVEL_cpdca>3) {

	print STDOUT "[ Transition matrices along the branches from the node '$node_id' ]\n\n";
	for (my $c=0; $c<$ct_children; $c++) {

	    my $ch_id = $child_nodes->[$c];
	    my $tr_mtrx = $set_tr_mtrices[$c];

	    print STDOUT "| Down to the $c th child (ID: $ch_id) |\n\n";
	    print STDOUT join ("\t", 'From\To', @NUCLEOTIDES), "\n\n";
	    for (my $i=0; $i<$CT_NUCLS; $i++) { print STDOUT join ("\t", $NUCLEOTIDES[$i], @{$tr_mtrx->[$i]}), "\n"; } 
	    print STDOUT "\n\n";
	}

   }
=cut # END of "For the predecessor 'calc_prob_descresstats_cond_ancstats'."

		# ADDED in creation of 'calc_prob_descresstats_cond_ancstats2'. #

	    my $tr_mtrx = $br2tr_mtrx->{$node_id};

	    unless ((defined $tr_mtrx) and (@{$tr_mtrx}>0)) { # ADDED on 2019/03/01.
		unless ($node_id == $topnode_id) {
		    print STDOUT "[ERROR 'calc_prob_descresstats_cond_ancstats2'] A transition matrix is NOT ASSIGNED TO the branch '$node_id': $! !\n\n";
		    return undef;
		}
	    }

   if (($TEST_LEVEL_cpdca>3) and (defined $tr_mtrx) and (@{$tr_mtrx}>0)) {

	print STDOUT "[ Transition matrix along the branch: '$node_id' ]\n\n";
	print STDOUT join ("\t", 'From\To', @NUCLEOTIDES), "\n\n";
	for (my $i=0; $i<$CT_NUCLS; $i++) { print STDOUT join ("\t", $NUCLEOTIDES[$i], @{$tr_mtrx->[$i]}), "\n"; } 
	print STDOUT "\n\n";
   }

	    my @setset_sw_ch_extcond_probs = ();

	    foreach my $ch (@{$child_nodes}) { # Middle preliminary foreach-loop (over child nodes). #

		my $set_sw_ch_extcond_probs = $node2set_sw_extcond_probs{$ch};
		unless (defined $set_sw_ch_extcond_probs) {
		    print STDOUT "[ERROR 'calc_prob_descresstats_cond_ancstats2'] NO SET of extended conditional PROBABILITIES was ASSIGNED TO the child '$ch' of the node '$node_id': $! !\n\n";
		    return undef;
		}
		push @setset_sw_ch_extcond_probs, $set_sw_ch_extcond_probs;
	    }				# END of the middle preliminary foreach-loop (over child nodes). #


		# END of "ADDED in creation of 'calc_prob_descresstats_cond_ancstats2'." #


   (print STDOUT "[ Site-wise calculations ... ]\n\n") if ($TEST_LEVEL_cpdca>2) ;

	    my @set_sw_cond_probs = ();
	    my @set_sw_extcond_probs = (); # ADDED in creation of 'calc_prob_descresstats_cond_ancstats2'. #

	    for (my $s=0; $s<$ct_sites; $s++) { # Main middle for-loop (over sites). #


   (print STDOUT ' ' x 2,  "|| The $s th site ||\n\n") if ($TEST_LEVEL_cpdca>3) ;

			# Initialize the conditional probabilities at the site. # 

		my @cond_probs = ();
		for (my $i=0; $i<$CT_NUCLS; $i++) { $cond_probs[$i] = 1; }

		for (my $c=0; $c<$ct_children; $c++) { # Main inner-middle for-loop (over child nodes). #

		    my $ch_id = $child_nodes->[$c];

   (print STDOUT ' ' x 4, "| The $c th child (ID: $ch_id) |\n\n") if ($TEST_LEVEL_cpdca>3);


		    my $ch_extcond_probs = $setset_sw_ch_extcond_probs[$c]->[$s];

		    for (my $i=0; $i<$CT_NUCLS; $i++) { # Main inner for-loop (over potential residue states at the subject node). #

			my $ch_spec_prob = $ch_extcond_probs->[$i];

			$cond_probs[$i] *= $ch_spec_prob;

   (print STDOUT ' ' x 6, join ("\t", $NUCLEOTIDES[$i], $ch_spec_prob), "\n") if ($TEST_LEVEL_cpdca>5); 

		    }				# End of the main inner for-loop (over potential residue states at the subject node).
   (print STDOUT "\n") if ($TEST_LEVEL_cpdca>5);
   (print STDOUT "\n") if ($TEST_LEVEL_cpdca>=5);

		}			# END of the main inner-middle for-loop (over child nodes). #

=For the predecessor 'calc_prob_descresstats_cond_ancstats'.
		my $ct_empty = 0;

		for (my $c=0; $c<$ct_children; $c++) { # Main inner-middle for-loop (over child nodes). #

		    my $ch_id = $child_nodes->[$c];

   (print STDOUT ' ' x 4, "| The $c th child (ID: $ch_id) |\n\n") if ($TEST_LEVEL_cpdca>3);


		    my $ch_cond_probs = $setset_sw_ch_cond_probs[$c]->[$s];
		    unless (@{$ch_cond_probs}>0) { # If the probabilities conditional on the child states are NOT assigned.

   (print STDOUT ' ' x 4, "... Calculation was SKIPPED because the child-conditional probabilities are UNDEFINED ...\n\n") if ($TEST_LEVEL_cpdca>=5);

			$ct_empty++;
			next;
		    }

		    my $tr_mtrx = $set_tr_mtrices[$c];

   if ($TEST_LEVEL_cpdca>=5) {
	print STDOUT ' ' x 4, "{Descendant state probabilities conditional on Res($ch_id) = (", join (',', @NUCLEOTIDES), ")} = (", join (', ', @{$ch_cond_probs}), ") .\n\n";

	print STDOUT ' ' x 4, "( Transition matrix from $node_id to $ch_id )\n\n";
	print STDOUT ' ' x 4, join ("\t", 'From\To', @NUCLEOTIDES), "\n\n";
	for (my $i=0; $i<$CT_NUCLS; $i++) { print ' ' x 4, join ("\t", $NUCLEOTIDES[$i], @{$tr_mtrx->[$i]}), "\n"; }
	print STDOUT "\n\n";
   }

   (print STDOUT ' ' x 4, "{ Calculation of conditional probabilities of the residue states of descendant sequences of the child node '$ch_id', conditional on the state of the node '$node_id' }\n\n") if ($TEST_LEVEL_cpdca>=5);

   (print STDOUT ' ' x 6, join ("\t", "Res($node_id)", 'Cond_Prob(child_spec_descendants)' ), "\n\n") if ($TEST_LEVEL_cpdca>5);

		    for (my $i=0; $i<$CT_NUCLS; $i++) { # Main inner for-loop (over potential residue states at the subject node). #
				# Here, the formula in the header will be calculated. #

			my $row_tr_probs = $tr_mtrx->[$i];

			my $ch_spec_prob = 0; # Probability factor specific to the $c th child. It corresponds to a factor in the product over children.
			for (my $j=0; $j<$CT_NUCLS; $j++) { $ch_spec_prob += $row_tr_probs->[$j] * $ch_cond_probs->[$j]; }

			$cond_probs[$i] *= $ch_spec_prob;

   (print STDOUT ' ' x 6, join ("\t", $NUCLEOTIDES[$i], $ch_spec_prob), "\n") if ($TEST_LEVEL_cpdca>5); 

		    }				# End of the main inner for-loop (over potential residue states at the subject node).
   (print STDOUT "\n") if ($TEST_LEVEL_cpdca>5);
   (print STDOUT "\n") if ($TEST_LEVEL_cpdca>=5);

		}			# End of the main inner-middle for-loop (over child nodes). #

		if ($ct_empty == $ct_children) { @cond_probs = (); } # If all the descendant sequences are occupied by gaps, conditional probabilities cannot be calculated.

=cut # END of "For the predecessor 'calc_prob_descresstats_cond_ancstats'." #

		$set_sw_cond_probs[$s] = \@cond_probs;

   if ($TEST_LEVEL_cpdca>=5) {
	print STDOUT ' ' x 4, "[ The resulting probabilies of residue states of descendants of the node '$node_id' conditional on its residue state ]\n\n";
	if (@cond_probs>0) {
	    print STDOUT ' ' x 4, join ("\t", "Res($node_id)", 'Cond_Prob(all_descendants)'), "\n\n";
	    for (my $i=0; $i<$CT_NUCLS; $i++) { print STDOUT ' ' x 4, join ("\t", $NUCLEOTIDES[$i], $cond_probs[$i]), "\n"; }
	    print STDOUT "\n\n";
	} else {
	    print STDOUT ' ' x 4, "... The probabilities are UNDEFINED because all descendants are occupied by gaps at the site ...\n\n\n";
	}
   }

			## ADDED in creation of 'calc_prob_descresstats_cond_ancstats2'. ##

		unless ((defined $tr_mtrx) and (@{$tr_mtrx}>0)) {
		    next; # If $tr_mtrx is not well-defined, the node should be the root. #
		}

		my @extcond_probs = ();
		for (my $i=0; $i<$CT_NUCLS; $i++) { # 2nd main inner-middle for-loop (over potential residue states at the parent node of the subject branch). #

		    my $tr_vec = $tr_mtrx->[$i];
		    my $extcond_prob = 0;
		    for (my $j=0; $j<$CT_NUCLS; $j++) { $extcond_prob += $tr_vec->[$j] * $cond_probs[$j]; }
		    $extcond_probs[$i] = $extcond_prob; # ADDED to remove a bug (on Apr 30 (Sat), 2016). # 

		}			# End of the 2nd main inner-middle for-loop (over potential residue states at the parent node of the subject branch). #
		
		$set_sw_extcond_probs[$s] = \@extcond_probs;

   if ($TEST_LEVEL_cpdca>5) {
	print STDOUT ' ' x 4, "[ The resulting probabilies of residue states of descendants of the node '$node_id' conditional on the residue state of its parent node ]\n\n";

	print STDOUT ' ' x 4, join ("\t", "Res(Pa($node_id))", 'Cond_Prob(all_descendants)'), "\n\n";
	for (my $i=0; $i<$CT_NUCLS; $i++) { print STDOUT ' ' x 4, join ("\t", $NUCLEOTIDES[$i], $extcond_probs[$i]), "\n"; }
	print STDOUT "\n\n";
   }

			## END of "ADDED in creation of 'calc_prob_descresstats_cond_ancstats2'." ##


	    }			# End of the main middle for-loop (over sites). #

	    $node2set_sw_cond_probs{$node_id} = \@set_sw_cond_probs;

	    if ((defined $tr_mtrx) and (@{$tr_mtrx}>0)) {
		$node2set_sw_extcond_probs{$node_id} = \@set_sw_extcond_probs;
	    }

	}			# End of the main outer foreach-loop (over interior nodes). #

   (print STDOUT "\n< End of subroutine 'calc_prob_descresstats_cond_ancstat2' />\n\n") if ($TEST_LEVEL_cpdca>0);

	return (\%node2set_sw_cond_probs, \%node2set_sw_extcond_probs);
	##return \%node2set_sw_cond_probs; # For the predecessor 'calc_prob_descresstats_cond_ancstats'. #

}	# End of the subroutine 'calc_prob_descresstats_cond_ancstat2'. #


# calc_joint_prob_ancresstats_and_ancstat2:
#	This subroutine calculates the joint probabilities of the residue states of all the extant sequences
#	on the "ancestral" side of a branch, accompanied by the residue states of the branch's upper-end node.
#	It also calculates the joint probabilities of the same residue states 
#	accompanied by the residue state of the branch's lower-end node.
#	Both types of the joint probabilities are calculated at every site in an input MSA.
#	The subroutine leaves the input MSA intact.
#	[CREATION STARTED on May 1st (Sun), 2016, ENDED on ...]
#	[REVISED on 2019/06/28]: The formula was CORRECTED, so that it will no longer be confusing.
#
#	[Formula]
#	The probability, p_U(b)_i = Pr [{s(a)}_b , s(ue(b)) = N_i ],
#	that the extant sequences on the "ancestral" side of branch b, {a}_b, show the states, {s(a)}_b, at the s-th site,
#	[ in addition to that the s-th site at the upper-end, ue(b), of the branch b is nucleotide N_i, can be calculated by the following formula: ] ... OBSOLETE as of 2019/06/28, because it is confusing.
#	in addition to that the state, s(ue(b)), of the s-th site at the upper-end, ue(b), of the branch b is nucleotide N_i, can be calculated by the following formula: # CORRECTED on 2019/06/28.
#
#	[ p_U(b)_i = ep_U(b)_i * Prod_{m in ch(ue(b)), m ne le(b)}  ep(m)_i, ] ... OBSOLETE as of 2019/06/28, because it is WRONG.
#	p_U(b)_i = ep_U(pa(b))_i * Prod_{m in ch(ue(b)), m ne le(b)}  ep(m)_i, # CORRECTED on 2019/06/28.
#	where
#	   ep(m)_i =: Pr [ {s(d)}_m | s(pa(m)) = N_i ]
#	   [ is the probability that the same descendant residue states resulted, given that 
#	   the state at the parent node, pa(m), is nucleotide N_i.  ] ... OBSOLETE as of 2019/06/28, because it does not make sense without refering to the [Formula] of 'calc_prob_descresstats_cond_ancstat'.
#	   is the probability that the residue states, {s(d)}_m, of the descendants of m (={d}_m)  resulted,
#	   given that the state, s(pa(m)), at the parent node (=pa(m)) is nucleotide N_i.
#
#	[Note: If ue(b) is the root, ep_U(pa(b))_i in the above formula is replaced with
#		Pr [s(ue(b)) = N_i],
#	that is, the probability that the root has the nucleotide N_i at the s-th site.] # Added on 2019/06/28.
#
#	The probability, ep_U(b)_i = Pr [ {s(a)}_b, s(le(b)) = N_i ],
#	[ that the same extant sequences resulted, in addition to that the s-th side at the lower-end, le(b), of the branch b is nucleitide N_i, can be calculated as: ] ... OBSOLETE as of 2019/06/28, because it is somewhat unclear.
#	that the same states of the extant sequences, {s(a)}_b,  resulted, in addition to that the state, s(le(b)), of the s-th site at the lower-end of the branch b (=le(b)) is nucleitide N_i, can be calculated as: # CORRECTED on 2019/06/28.
#
#	[ ep_U(b)_i = Sum_{j} t(ue(b); le(b))_{i,j} * p_U(b)_j . ] ... OBSOLETE as of 2019/06/28, because it is WRONG.
#	ep_U(b)_i = Sum_{j} p_U(b)_j * t(ue(b); le(b))_{j,i} . # CORRECTED on 2019/06/28. 
#
#	[ Here, t(n;m)_{i,j} is the substitution transition probability that
#	node m has nucleotide N_j, on condition that node n has nucleotide N_i. ] ... OBSOLETE as of 2019/06/28, because its index assignment is somewhat confusing.
#	Here, t(n;m)_{j,i} is the substitution transition probability that
#	node m has nucleotide N_i, on condition that node n has nucleotide N_j. # CORRECTED on 2019/06/28, so that the explanation will be much clearer.
#
#	[NOTE] This subroutine itself can handle multifurcated trees as well,
#		although PRANK analyses cannot, in general.
#
# INPUT: (\@in_msa, \@seqnames, \%name2id, \%node2ch, \%br2tr_mtrx, $gap,
#		\@ntfreqs_root, \%node2set_sw_extcond_probs), where
#	${$in_msa[$k]} is the $k th sequence embedded in the input MSA;
#	$seqnames[$k] is the name (or ID) of the $k th sequence in the input MSA;
#	%name2id = ($node_name => $node_id, ...) for the exterior nodes and the top node (i.e., root) of the tree;
#	%node2ch = ($node_id => \@child_nodes, ...) for the interior nodes of the tree;
#
#	@{$br2tr_mtrx{$br_id}} is the transition matrix
#	that describes the transition probabilities between both ends of the branch $br_id.
#	More specifically,
#	$br2tr_mtrx{$br_id}->[$i][$j] is the probability that the node $br_id shows $NUCLEOTIDES[$j]
#	given that the parent node of $br_id shows $NUCLEOTIDES[$i].
#
#	$gap is the character representing a gap;
#
#	$ntfreqs_root[$i] is the frequency of $NUCLEOTIDES[$i] at the root;
#
#	[ %node2set_sw_cond_probs = ($node_id => \@set_sw_cond_probs, ...) for ALL nodes, where
#	$set_sw_cond_probs[$s]->[$i] is the probability that
#	all the descendant sequences of $node_id show the residue states that are actually found in the $s th site,
#	given that $NUCLEOTIDES[$i] occupies the $s th site of the ancestral node, $node_id.
#
#	But $set_sw_cond_probs[$s] = [] (empty) if $node_id refers to an external node and if the node has its $s th site occupied by a gap. ] --- This hash is unnecessary!!
#
#	%node2set_sw_extcond_probs = ($node_id => \@set_sw_extcond_probs, ...) for ALL nodes except the root, where
#	$set_sw_extcond_probs[$s]->[$i] is the probability of the same residue configuration was $set_sw_cond_probs[$s]->[$i],
#	but conditioned on that $NUCLEOTIDES[$i] occupies the $s th site of the PARENT OF the ancestral node, $node_id.
#
#
# OUTPUT: (\%br2set_sw_jnt_probs, \%br2set_sw_extjnt_probs), where
#
#	%br2set_sw_jnt_probs = ($br_id => \@set_sw_jnt_probs, ...) for ALL branches, where 
#	$set_sw_jnt_probs[$s]->[$i] is the probability that
#	all the extant sequences on the "ancestral side" of $br_id show the residue states that are actually found in the $s th site,
#	in addition to that $NUCLEOTIDES[$i] occupies the $s th site of the UPPER-end node of $br_id.
#
#	%br2set_sw_extjnt_probs = ($br_id => \@set_sw_extjnt_probs, ...) for ALL branches, where 
#	$set_sw_extjnt_probs[$s]->[$i] is the probability very similar to $set_sw_jnt_probs[$s]->[$i],
#	with the only difference that the UPPER-end node of $br_id is now replaced with its LOWER-end node.
#
sub calc_joint_prob_ancresstats_and_ancstat2 (\@\@\%\%\%$\@\%) {

	my ($in_msa, $seqnames, $name2id, $node2ch, $br2tr_mtrx, $gap,
		$ntfreqs_root, $node2set_sw_extcond_probs) = @_;

	$gap = uc $gap;

	my $ct_seqs = @{$in_msa};
	my $ct_sites = length (${$in_msa->[0]});

	my $topnode_id = $name2id->{$TOP_NODE};

   (print STDOUT "\n< Subroutine 'calc_joint_prob_ancresstats_and_ancstat2'  (#{sequences} = $ct_seqs; #{sites} = $ct_sites; topnode ID = $topnode_id; Gap = '$gap') >\n\n") if ($TEST_LEVEL_cjpaaa>0);

	my @set_pa_sbjct_branches = ($topnode_id); # The set of parents of the subject-branches. #
	##my @set_pa_sbjct_branches = (); # The set of parents of the subject-branches. # A BUG fixed on May 2 (Mon), 2016. #

	my %br2set_sw_jnt_probs;
	my %br2set_sw_extjnt_probs;


		## MAIN PROCESS: Calculate the conditional probabilities in a TOP-DOWN manner. ##

   (print STDOUT "\n{{ MAIN PROCESS: Calculate the conditional probabilities from top (root) to bottom (leaves)... }}\n\n") if ($TEST_LEVEL_cjpaaa>0);

	while (@set_pa_sbjct_branches>0) { # The main outermost while-loop (continues until the subjects run out). #

	   my @newset = ();

	   foreach my $pa_sbj (@set_pa_sbjct_branches) { # The main outer foreach-loop (over the subject parents). #

		my $set_sbjcts = $node2ch->{$pa_sbj};
		((defined $set_sbjcts) and (@{$set_sbjcts}>0)) or next; # Skip if there is no subject branches. #

		push @newset, @{$set_sbjcts}; # "Increment" the new set of subject parents. #

		my $set_sw_extjnt_probs_pa = $br2set_sw_extjnt_probs{$pa_sbj}; # The set of "extended" joint-probabilities along the parent branch. #

		foreach my $br_id (@{$set_sbjcts}) { # Main outer-middle foreach-loop (over the subject branches). #

		    my $tr_mtrx = $br2tr_mtrx->{$br_id};

		    my @setset_sw_extcond_probs_sbl = ();
		    foreach my $sbl_br (@{$set_sbjcts}) { # Preliminary for-loop (over sibling branches). #
			($sbl_br eq $br_id) and next; # Skip the subject branch. #
			my $set_sw_extcond_probs_sbl = $node2set_sw_extcond_probs->{$sbl_br};
			unless (defined $set_sw_extcond_probs_sbl) {
			     print STDOUT "[ERROR 'calc_joint_prob_ancresstats_and_ancstat2'] ILLEGAL INPUT: The sibling branch (ID: $sbl_br) of the branch (ID: $br_id) has NO SET of extended conditional probabilities: $! !\n\n";
			     return (undef, undef);
			}
			push @setset_sw_extcond_probs_sbl, $set_sw_extcond_probs_sbl;
		    }					# End of the preliminary for-loop (over sibling branches). #

		    my @set_sw_jnt_probs = my @set_sw_extjnt_probs = ();

		    for (my $s=0; $s<$ct_sites; $s++) { # Main inner-middle for-loop (over the sites). #

			my $extjnt_probs_pa = (defined $set_sw_extjnt_probs_pa) ? $set_sw_extjnt_probs_pa->[$s] : $ntfreqs_root; # If the set is undefined, the "parent" must be the  root node. #
			
				# Initialize the joint probabilities. #
			my @jnt_probs = @{$extjnt_probs_pa};

			foreach my $set_sw_extcond_probs_sbl (@setset_sw_extcond_probs_sbl) { # 1st inner foreach-loop (over sibling branches). #
			    my $extcond_probs_sbl = $set_sw_extcond_probs_sbl->[$s];

			    for (my $i=0; $i<$CT_NUCLS; $i++) { # Inner-inner for-loop (over nucleotides). #
				$jnt_probs[$i] *= $extcond_probs_sbl->[$i];
			    }			# End of the inner-inner for-loop (over nucleotides). #

			}			# End of the 1st inner foreach-loop (over sibling branches). #

			$set_sw_jnt_probs[$s] = \@jnt_probs;

			my @extjnt_probs = ();

			for (my $i=0; $i<$CT_NUCLS; $i++) { # 2nd inner for-loop (over nucleotides). #

			    my $extjnt_prob = 0;

			    for (my $j=0; $j<$CT_NUCLS; $j++) { $extjnt_prob += $jnt_probs[$j] * $tr_mtrx->[$j][$i]; }

			    $extjnt_probs[$i] = $extjnt_prob;
			}			# End of the 2nd inner for-loop (over nucleotides). #

			$set_sw_extjnt_probs[$s] = \@extjnt_probs;

		    }				# End of the main inner-middle for-loop (over the sites). #

		    $br2set_sw_jnt_probs{$br_id} = \@set_sw_jnt_probs;
		    $br2set_sw_extjnt_probs{$br_id} = \@set_sw_extjnt_probs;

		}			# End of the main outer-middle foreach-loop (over the subject branches). #

	   }					# End of the main outer foreach-loop (over the subject parents). #

	   @set_pa_sbjct_branches = @newset;

	}		# END of the main outermost while-loop (continues until the subjects run out.) #


   (print STDOUT "\n< End of subroutine 'calc_joint_prob_ancresstats_and_ancstat2' />\n\n") if ($TEST_LEVEL_cjpaaa>0);

	return (\%br2set_sw_jnt_probs, \%br2set_sw_extjnt_probs);

}		# End of the subroutine 'calc_joint_prob_ancresstats_and_ancstat2'. #

# calc_clm_probs:
#	This subroutine calculates the occurrence probabilities of the MSA columns
#	under the substitution model under consideration,
#	using the output of 'calc_prob_descresstats_cond_ancstat2'.
#	[CREATED on May 3 (Tue), 2016.]
#
# INPUT: (\@ntfreqs_root, \%node2set_sw_cond_probs, $topnode_id), where
#
#	$ntfreqs_root[$i] is the frequency of $NUCLEOTIDES[$i] at the root.
#
#	%node2set_sw_cond_probs = ($node_id => \@set_sw_cond_probs, ...) for ALL nodes, where
#	$set_sw_cond_probs[$s]->[$i] is the probability that
#	all the descendant sequences of $node_id show the residue states that are actually found in the $s th site,
#	given that $NUCLEOTIDES[$i] occupies the $s th site of the ancestral node, $node_id.
#
#	But $set_sw_cond_probs[$s] = [] (empty) if $node_id refers to an external node and if the node has its $s th site occupied by a gap.
#
#	$topnode_id is the ID of the top node (i.e., the root).
#
# OUTPUT: \@set_clm_probs, where
#	$set_clm_probs[$s] is the occurrence probability of the $s th column
#	under the substitution model in question.
#
sub calc_clm_probs (\@\%$) {

	my ($ntfreqs_root, $node2set_sw_cond_probs, $topnode_id) = @_;

	my $set_sw_cond_probs_root = $node2set_sw_cond_probs->{$topnode_id};

	my $ct_sites = @{$set_sw_cond_probs_root};

	my @set_clm_probs = ();

	for (my $s=0; $s<$ct_sites; $s++) { # Main outer for-loop (over sites). #

	    my $cond_probs_root = $set_sw_cond_probs_root->[$s];

	    my $clm_prob = 0;
	    for (my $i=0; $i<$CT_NUCLS; $i++) { $clm_prob += $ntfreqs_root->[$i] * $cond_probs_root->[$i]; }

	    $set_clm_probs[$s] = $clm_prob;

	}			# End of the main outer for-loop (over sites). #

	return \@set_clm_probs;

}		# End of the subroutine 'calc_clm_probs'. #


# br_calc_sbst_prob_sgl_clm:
#	This subroutine briefly computes the probability
#	that the residue pattern of a single column of an MSA
#	results under a given substitution model,
#	by taking advantage of an output of the subroutine 'calc_prob_descresstats_cond_ancstat2' above.
#	[CREATED on 2019/02/23.]
#
# INPUTS: (\@exnode_ids_in_msa, $char_abs, \@column, $topnode_id, \%node2ch, \%node2depth, \@base_freqs_root, \%node2set_sw_extcond_probs0, \%br2tr_mtrx), where
#
#	$exnode_ids_in_msa[$k] is the ID of the "external" node occupying the $k th row of the input local MSA;
#
#	$char_abs is the character that represents a gap, or an "absence" state;
#
#	@column = specifies a column in the input local MSA. more precisely,
#	$column[$k] = $char_abs if the $k th row of the $c th column is occupied by a gap (or the "absence" state),
#	otherwise, it gives the (numerical) index specifying the column of the input alignment that accommodates the corresponding site of the $k th "sequence";
#
#	$topnode_id is the ID of the top-node (i.e., root) of the input tree;
#	%node2ch = ($node_id => \@children, ...) for the internal nodes;
#	%node2depth = ($node_id => $depth, ...) for ALL nodes;
#	[\%node2depth is an output of, e.g., 'make_node2abs_depth' in "MyTreeManipu_III.pm".]
#
#	$base_freqs_root[$i] is the frequency of the base, $NUCLEOTIDES[$i], at the root;
#
#	%node2set_sw_extcond_probs0 = ($node_id => \@set_sw_extcond_probs, ...) for the input MSA, where
#       $set_sw_extcond_probs0[$s]->[$i] is the probability that
#       all the descendant sequences of $node_id show the residue states that are actually found in the $s th site of the input MSA,
#       CONDITIONED ON that $NUCLEOTIDES[$i] occupies the $s th site of the PARENT OF the ancestral node, $node_id;
#	(It is an output of, e.g., 'calc_prob_descresstats_cond_ancstat2' in "MyLikelihoods_respattern_msa_dna.pm".)
#
#	%br2tr_mtrx = ($br => \@tr_mtrx, ...) gives the substitution transition matrices for the branches, where
#	$tr_mtrx[$i]->[$j] gives the probability that the lower-end of $br is occupied by $NUCLEOTIDES[$j],
#	given that the upper-end of $br is occupied by $NUCLEOTIDES[$i];
#	(It could be created, e.g., by 'calc_br2tr_mtrx_TN' of "MyLikelihoods_respattern_msa_dna.pm".)
#
# OUTPUT: $sbst_prob_sgl_clm, which is the probability you want.
#
sub br_calc_sbst_prob_sgl_clm (\@$\@$\%\%\@\%\%) {

	my ($exnode_ids_in_msa, $char_abs, $column, $topnode_id, $node2ch, $node2depth, $base_freqs_root, $node2set_sw_excond_probs0, $br2tr_mtrx) = @_;

	my $ct_rows = @{$column};

		# Initialize the auxiliary hashes. #

	my @null_probs = ();
	for (my $i =0 ; $i <$CT_NUCLS; $i++) { $null_probs[$i] = 1; }

	my %node2sgl_sw_cond_probs;
	my %node2sgl_sw_excond_probs;

	for (my $k = 0; $k < $ct_rows; $k++) { # Preliminary for-loop (over the MSA rows).

	    my $node = $exnode_ids_in_msa->[$k];
	    my $site = $column->[$k];
	    if ($site eq $char_abs) {
		$node2sgl_sw_excond_probs{$node} = \@null_probs;
	    } else {
		my $set_sw_excond_probs0 = $node2set_sw_excond_probs0->{$node};
		$node2sgl_sw_excond_probs{$node} = $set_sw_excond_probs0->[$site];
	    }
	} # End of the preliminary for-loop (over the MSA rows).


		# MAIN PROCESS: Compute the probabilities in a bottom-up manner. #

	foreach my $node (sort {$node2depth->{$b} <=> $node2depth->{$a} } keys %{$node2ch}) { # Main foreach-loop (over the internal nodes). #

	    my $children = $node2ch->{$node};

	    my @cond_probs = @null_probs;
	    foreach my $ch (@{$children}) { # The 1st middle foreach-loop (over the child nodes).
		my $excond_probs_ch = $node2sgl_sw_excond_probs{$ch};
		unless (defined $excond_probs_ch) {
		    print STDERR "[ERROR 'br_calc_sbst_prob_sgl_clm'] The node $ch, which is a child of $node, does NOT HAVE extended conditional probabilities: $! !!\n\n";
		    return undef;
		}

		for (my $i =0; $i < $CT_NUCLS; $i++) { # Inner for-loop (over possible bases).
		    $cond_probs[$i] *= $excond_probs_ch->[$i];
		} # End of the inner for-loop (over possible bases).

	    } # End of the 1st middle foreach-loop (over the child nodes).

	    $node2sgl_sw_cond_probs{$node} = \@cond_probs;

	    my $tr_mtrx = $br2tr_mtrx->{$node};

	    unless (defined $tr_mtrx) { next; } # This should happen only for the root. #

	    my @excond_probs = ();

	    for (my $i = 0; $i < $CT_NUCLS; $i++) { # The 2nd middle for-loop (over possible bases at the upper-end). #
		my $tr_vec = $tr_mtrx->[$i];

		my $excond_prob = 0;
		for (my $j=0; $j < $CT_NUCLS; $j++) { # Inner for-loop (over possible bases).
		    $excond_prob += $tr_vec->[$j] * $cond_probs[$j];
		} # End of the inner for-loop (over possible bases).

		$excond_probs[$i] = $excond_prob;

	    }  # End of the 2nd middle for-loop (over possible bases at the upper-end).

	    $node2sgl_sw_excond_probs{$node} = \@excond_probs;

	}  # End of the main foreach-loop (over the internal nodes). #

		# FINAL PROCESS: Compute the probability for the column. #

	my $cond_probs_root = $node2sgl_sw_cond_probs{$topnode_id};

	my $sbst_prob_sgl_clm = 0;

	for (my $i=0; $i < $CT_NUCLS; $i++) { # Finishing for-loop (over possible bases).
	    $sbst_prob_sgl_clm += $base_freqs_root->[$i]  * $cond_probs_root->[$i];
	} # End of the finishing for-loop (over possible bases).

	return $sbst_prob_sgl_clm;

}	# END of the subroutine 'br_calc_sbst_prob_sgl_clm'. #



1;		# a "spell" which makes the module "use"-able and/or "require"-able #


