package MyMapAlnShear;
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.5: Copyright (C) 2019? Kiyoshi Ezawa
#	Version 0.6: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyMapAlnShear.pm:
#	This package is a collection of subroutines
#	necessary for creating the Alignment-Shear Map,
#	which indicates where alignment-shears (i.e., mis-alignments viewed in terms of blocks of residues)
#	occur on the 2D map of sites (columns) x branches (rows).
#	[CREATION Started on 2020/03/05, ENDED on 2020/03/dd.]
#


use strict;
use warnings;


=Set aside for potential use. (1)

use MyTaxonomy qw(
	if_ancestor fetch_ancestry calculate_abs_depth
	fetch_ext_offsprings_nr fetch_all_offsprings_nr
	);

use MyTreeManipu_II qw( );
	# extract_sgltree_from_file output_tree 

use MyTreeManipu_III qw( make_node2abs_depth );

use MyLikelihoods_respattern_msa_dna qw(
	calc_node2ntfreqs_td
	calc_prob_descresstats_cond_ancstat2
	calc_joint_prob_ancresstats_and_ancstat2
	calc_clm_probs
	);

use MyTreeManipu_IV qw( bi2multifurcated_top construct_neat_subtree );
use MyPAML_parser qw( retrieve_aa_freqs prepare_paml_ctrl );

use MyProbDistri qw(
	binomial_cum_via_betai
	);

use MyArrayManipu qw(
	merge_integer_intervals
	);

	# Added in creation of 'detect_purge_cands2' (on 2019/06/15). #
use MyArrayManipu_num qw(
	union_nmarray
	binsearch_num
	); # Added 'binsearch_num' in revision of 'detect_purge_cands2' (on 2019/06/16).
	# End of "Added in creation of 'detect_purge_cands2' (on 2019/06/15)." #

=cut # End of "Set aside for potential use. (1)"

		# (IMPORTED from "MySRD_Map.pm" on 2020/03/09) #
use MyDetect_purge_cands qw(
	mk_node2sw_gstats
	);
	# sum_arrays_over_branches
	# calc_br2cw_expcts_sbsts_ave
	# calc_br2cw_obscts_sbsts
	# calc_br2cw_expcts_obscts_sbsts
		# END of "(IMPORTED from "MySRD_Map.pm" on 2020/03/09)" #


use Exporter;

our @ISA =qw(Exporter);

our @EXPORT_OK = qw( 
	mk_map_alnshear
	mk_map_alnshear_clmbased
	print_map_alnshear
	);

sub mk_map_alnshear (\@$$$$\%\@\@\%\@$$\%\%); # Added the stub on 2020/03/10.
sub mk_map_alnshear_clmbased (\@$$$$\%\@\@\%\@$$\%\%); # ADDED on 2020/03/21.
sub print_map_alnshear (\*\@\%\@$$$$); # Added the stub on 2020/03/10.


	### Default values of the Parameters ###

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;

=Set aside for potential use. (2)
our @NUCLEOTIDES = @MyLikelihoods_respattern_msa_dna::NUCLEOTIDES; # Unless otherwise stated, nucleotide MSA scores will be calculated according to this order (or index assignment) of the nucleotide bases.
our $CT_NUCLS = @NUCLEOTIDES;

our %NUCL2INDX; # = ($nucleotide => {index in @NUCLEOTIDES}, ...)
for (my $k=0; $k<$CT_NUCLS; $k++) { $NUCL2INDX{$NUCLEOTIDES[$k]} = $k; }
=cut # End of "Set aside for potential use. (2)"


	## The following parameters specify the verboseness of the test output.
	## (0, silent; 1, very modest; 2, modest; 3, medium; 4, verbose; 5, very verbose; 6, extremely verbose.)

	# For 'mk_map_alnshear'. #
our $TEST_LEVEL_mk = 0;
#our $TEST_LEVEL_mk = 5;
##our $TEST_LEVEL_mk = 6;

	# For 'print_map_alnshear'. #
our $TEST_LEVEL_pr = 0;
#our $TEST_LEVEL_pr = 5;
##our $TEST_LEVEL_pr = 6;



### 5 important hash tables which describe a tree ###
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.
#


		##### DEFINITIONS of the Subroutines. #####


# mk_map_alnshear:
#	This subroutine receives a set of "mini-class"es of position-shifts,
#	as well as some supplementary information,
#	then it creates an Alignment-Shear Map.
#	The script was created by somewhat modifying 'mk_srd_map' (in "MySRD_Map.pm").
#	[CREATION STARTED on 2020/03/08, ENDED on 2020/03/09.]
#
# mk_srd_map (in "MySRD_Map.pm"):
#	This subroutine receives some variables characterizing a phylogenetic tree,
#	and some outputs of the likelihood analysis (under a given substitution model) 
#	performed on a given multiple sequence alignment (MSA),
#	then it  creates a Substitutional Residue-Difference Map (abbr. "SRD Map").
#	The script was created by taking advantage of some portions of
#	'detect_purge_cands2' (in "MyDetect_purge_cands.pm").
#	[CREATION STARTED on 2020/03/05, ENDED on 2020/03/05.]
#	
# [ INPUT: ($if_observed, @br_ids_for_display, $char_abs, $char_pres, $topnode_id, \%node2pa, \%node2ntfreqs, \%br2tr_mtrx, \%br2set_sw_jnt_probs, \%node2set_sw_cond_probs, \@set_clm_probs, \@info_psm_hst0, $beg_msa, $end_msa, \%br_attr, \%br2equiv), where ] ... for the predecessor, 'mk_srd_map'. #
#
# INPUTS: (@br_ids_for_display, $char_abs, $char_pres, $char_noshear, $topnode_id, \%node2pa, 
#	\@set_miniclasses_shift_respos, \@set_sep_patterns_for_miniclasses, \%miniclass2trivial,
#	\@info_psm_hst0, $beg_msa, $end_msa, \%br_attr, \%br2equiv), where
#
#	@br_ids_for_display stores ALL branch IDs
#	(, each of which is actually the ID of the node at the lower-end of the branch,)
#	in the order to be displayed as the rows of the SRD Map.
#
#	$char_abs/$char_pres is the character that represents an "absence"/"presence" state;
#	$char_noshear is the character indicating that the position (= site x branch) has NO shear. 
#
#	$topnode_id is the ID of the top-node (i.e., root);
#	%node2pa = ($node_id => $parent_id, ...) for the nodes except the root.
#
#	@{$set_miniclasses_shift_respos[$mc]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
#	is for the $mc th mini-class of the run of shifts of residue positions, where
#	$shift is the value of shift,
#	$beg/end_site is the beginning/ending site (0-based full-closed) of the region with the shift value $shift,
#	@{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;
#
#	@{$set_sep_patterns_for_miniclasses[$mc]->[$pt]} = (\%node2parsstat, \%br2change, $ct_changes) is for the $pt th pattern separating the sequences affected by the $mc th mini-class of position shifts, where
#	%node2parsstat = ($node_id => [$state, $ct_pres_ext], ...),
#	%br2change = ($branch_id => [$state_upper_end, $state_lower_end], ...), and
#	$ct_changes is the number of "changes" necessary for separating the affected sequences from the rest;
#
#	%miniclass2trivial = ($indx_miniclass => 1, ...) for trivial mini-classes;
#
#	@info_psm_hst0 = (\@set_cls_gblocks, \@totcts0_indels_for_cls, \@psm_hst0_for_cls),
#	which is an output of 'cal_init_psm_cands_for_sgl_msa' in 'MyTreeMap_indels_spt_odr2.pm',
#	gives information on the Dollo parsimonious indel history giving rise to the input MSA gap configuration, where
#	
#	@{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end) is for the $gb th block in the $cl th cluster (i.e., gapped segment) in the input MSA, with
#		$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#		$len_gb being the length (in terms of #{columns}) of the block,
#		and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention),
#	
#	$totcts0_indels_for_cls[$cl] is the total number of insertion/deletion events in the Dollo parsimonious indel history resulting in the $cl th cluster (i.e., gapped segment) in the input MSA,
#	
#	@{$psm_hst0_for_cls[$cl]} = (\%br2indels0, \%node2blwise_gstat0) is for the Dollo parsimonious indel history resulting in the $cl th cluster (i.e., gapped segment) in the input MSA, where 
#		%br2indels0 = ($branch_id => \@list_indels0, ...) is the map of insertions/deletions onto branches, with
#		@{$list_indels0[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved), and where
#		%node2blwise_gstat0 = ($node_id => \@blwise_gapstats0, ...) is the map of the gap states onto the nodes,
#		  with $blwise_gapstats0[$bl] is the gap-state ($char_pres/abs) of the $bl th block at the node.
#
#	$beg/end_msa is the beginning/ending coordinates of the input MSA (in the 0-based full-closed convention).
#
#	%br_attr = ($br_id => \@attr, ...) for all branches, where $attr[$INDX_LENGTH] is the branch length.
#
#	%br2equiv = ($br_id => $equiv_br_id, ...) for those branches with equivalents.
#
# OUTPUTS: (\@map_alnshear, \%br2linmap_alnshear), where
#
#	$map_alnshear[$n]->[$s] = the Alignment-Shear status of the position specified by the site $s and the branch,  $br_ids_for_display[$n];
#	The "Alignment-Shear status" is:
#		 $char_abs if either end (or both ends) of the branch has (have) $char_abs in the initial gap-state,
#		 join (',', {list of mini-classes whose boundaries includ the position}),
#		 $char_noshear otherwise.
# 
#
#	%br2linmap_alnshear = ($br_id => \@linmap_alnshear, ...) for ALL branches, where
#	$linmap_alnshear[$s] is the Alignment-Shear status at the $s th site (more precisely, in the $s th column) of the input MSA, along the branch, $br_id.
#
sub mk_map_alnshear (\@$$$$\%\@\@\%\@$$\%\%) {

	my ($br_ids_for_display, $char_abs, $char_pres, $char_noshear, $topnode_id, $node2pa, 
	$set_miniclasses_shift_respos, $set_sep_patterns_for_miniclasses, $miniclass2trivial,
	$info_psm_hst0, $beg_msa, $end_msa, $br_attr, $br2equiv) = @_;


   (print STDOUT "< 'mk_map_alnshear' (\$char_abs = '$char_abs'; \$char_noshear = '$char_noshear'; $beg_msa <= {MSA coordinate} <= $end_msa) >\n\n") if ($TEST_LEVEL_mk>0);

	my $ct_brs = @{$br_ids_for_display};
	my $ct_brs2 = scalar (keys %{$br_attr});

	unless ($ct_brs == $ct_brs2) {
	    print STDOUT "[ERROR 'print_srd_map'] The number of branches to be displayed (=$ct_brs) CLASHES with the number of branches in ", '%{$br_attr}', " (=$ct_brs2) : $! !\n\n";
	    return (undef, undef);
	}

	my @set_all_nodes = (@{$br_ids_for_display}, $topnode_id);

   if ($TEST_LEVEL_mk >2) {
	print STDOUT "[ Secondary inputs ]\n\n";
	print STDOUT '@{$br_ids_for_display} = (', join (', ', @{$br_ids_for_display}), ") .\n\n";
	print STDOUT '@set_all_nodes = (', join (',', @set_all_nodes), ").\n\n\n";
   }


			# (Pre-1) Prepare the site-wise gap states (of the Dollo-parsimonious indel history) at every node. #

	# %{$node2sw_gstats0} = ($node_id => \@sw_gstats0, ...) for all nodes, where
	# $sw_gstats0[$s] is the gap state of the $s th site on the node $node_id
	# (usually according to the Dollo-parsimony principle).
	#
	my $node2sw_gstats0 = mk_node2sw_gstats (@{$info_psm_hst0}, $char_abs, $char_pres, $beg_msa, $end_msa, @set_all_nodes);
	unless (defined $node2sw_gstats0) {
	    print STDOUT "[ERROR 'mk_map_alnshear'] The subroutine 'mk_node2sw_gstats' FAILED: $! !\n\n";
	    return (undef, undef, undef);
	}
   (print STDOUT " --- (Pre-1) Subroutine 'mk_node2sw_gstats' ended ...\n\n") if ($TEST_LEVEL_mk >1);

   if ($TEST_LEVEL_mk >3) {

	print STDOUT "[ Content of ", '%{$Node2sw_gstats0}', ", which is the output of 'mk_node2sw_gstats' ]\n\n";
	print STDOUT 'Site\Node', "\t", join (', ', @set_all_nodes), "\n\n";
	for (my $s = $beg_msa; $s <= $end_msa; $s++) { # Outer for-loop (over sites). #
	    my @sw_gstats_on_nodes = ();
	    foreach my $node (@set_all_nodes) {
		my $gstat = $node2sw_gstats0->{$node}[$s];
		(defined $gstat) or ($gstat = '?');
		push @sw_gstats_on_nodes, $gstat;
	    }
	    print STDOUT $s, "\t", join (', ', @sw_gstats_on_nodes), "\n";

	}				# End of the outer for-loop (over sites). #

	print STDOUT "\n\n";

   } # END of "if ($TEST_LEVEL_mk >3) {...}"


		# Main Process 1: Initialize %br2linmap_alnshear. #

   (print STDOUT "[ Main Process 1: Initialize \%br2linmap_alnshear. ]\n\n") if ($TEST_LEVEL_mk >1);

	my %br2linmap_alnshear;

	foreach my $br (@{$br_ids_for_display}) { # 1st Main outer foreach-loop (over all branches) #

=For the predecessor, 'mk_srd_map'.
	    my $set_sw_srds = $br2sw_srds{$br};

	    ##my $eq_br = $br2equiv->{$br};
	    my $brlen = $br_attr->{$br}[$INDX_LENGTH];
	    ##if ((defined $eq_br) and ($brlen == 0)) {
	    if ($brlen == 0) {
			# For a null artificial branch,
			#  set all relevant entries to be $char_abs.

		for (my $s = $beg_msa; $s <= $end_msa; $s++) { $set_sw_srds->[$s] = $char_abs; }

		next;
	    }
=cut # END of "For the predecessor, 'mk_srd_map'."


	    my @linmap_alnshear = ();

	    my $node_ue = $node2pa->{$br};
	    my $sw_gstats0_le = $node2sw_gstats0->{$br};
	    my $sw_gstats0_ue = $node2sw_gstats0->{$node_ue};

	    for (my $s = $beg_msa; $s <= $end_msa; $s++) {  # Middle for-loop (over relevant sites). #

		my $gstat0_le = $sw_gstats0_le->[$s];
		my $gstat0_ue = $sw_gstats0_ue->[$s];

		$linmap_alnshear[$s] = (($gstat0_le eq $char_abs) or ($gstat0_ue eq $char_abs)) ? $char_abs : $char_noshear ;

	    } # End of the middle for-loop (over relevant sites). #

	    $br2linmap_alnshear{$br} = \@linmap_alnshear;

	}  # End of the 1st Main outer foreach-loop (over all branches) #


		# Main Process 2: Compute %br2linmap_alnshear. #

   (print STDOUT "[ Main Process 2: Compute \%br2linmap_alnshear. ]\n\n") if ($TEST_LEVEL_mk >1);

		# $set_miniclasses_shift_respos, $set_sep_patterns_for_miniclasses, $miniclass2trivial,

	my $ct_mcls = @{$set_miniclasses_shift_respos};

	for (my $mc = 0; $mc < $ct_mcls; $mc++) { # 2nd Main outer for-loop (over mini-classes of position-shifts) #

	    if ($miniclass2trivial->{$mc}) { next; }

	    my ($shift, $beg_site, $end_site, $affected_regions, $seqs_affected)
		 = @{$set_miniclasses_shift_respos->[$mc]};

	    my $set_sep_patterns = $set_sep_patterns_for_miniclasses->[$mc];

	    my %br2sbjct = ();
	    for (my $pt = 1; $pt < 2; $pt++) { # 1st middle for-loop (over separating patterns) # Regarding the range of $pt, see 'split_unnatural_miniclasses' (in "MyMalnComparison.pm"). #

		my ($node2parsstat, $br2change, $ct_changes) = @{$set_sep_patterns->[$pt]};

		foreach my $br (keys %{$br2change}) { $br2sbjct{$br} = 1; }

	    } # End of the 1st middle for-loop (over separating patterns) #

	    my @sbj_branches = sort {$a <=> $b} keys %br2sbjct;

	    foreach my $br (@sbj_branches) { # 2nd middle for-loop (over branches involved in the alignment shear). #

		my $linmap_alnshear = $br2linmap_alnshear{$br};

		foreach my $region (@{$affected_regions}) { # Inner for-loop (over affected regions). #
		   my ($beg_rg, $end_rg) = @{$region};

		   for (my $s = $beg_rg; $s <= $end_rg; $s++) { # Innermost for-loop (over sites). #

			my $alnshear = $linmap_alnshear->[$s];

			if ($alnshear eq $char_abs) { 
			    next; # Skip if the state is "absence". #

			} elsif ($alnshear eq  $char_noshear) {
			    $linmap_alnshear->[$s] = $mc;

			} else {
			    $linmap_alnshear->[$s] .= ",$mc";
			}

		   } # End of the innermost for-loop (over sites). #

		} # End of the inner for-loop (over affected regions). #

	    } # 2nd middle for-loop (over branches involved in the alignment shear). #

	} # End of the 2nd Main outer for-loop (over mini-classes of position-shifts) #

   if ($TEST_LEVEL_mk >3) {
	print STDOUT "[ Final Content of ", '%br2linmap_alnshear', ": ]\n\n";

	print STDOUT join ("\t", 'Column\Branch', @{$br_ids_for_display}), "\n\n";

	for (my $c = $beg_msa; $c <= $end_msa; $c++) { # Outer for-loop (over columns). #

	    my @clm_alnshearstats = ();
	    foreach my $br (@{$br_ids_for_display}) {
		my $alnshearstat = $br2linmap_alnshear{$br}->[$c];
		(defined $alnshearstat) or ($alnshearstat = 'undef');
		push @clm_alnshearstats, $alnshearstat;
	    }
	    print STDOUT join ("\t", $c, @clm_alnshearstats), "\n";

	} # End of outer for-loop (over columns). #

	print STDOUT "\n\n";

=Alternative idea
	foreach my $br (@{$br_ids_for_display}) {
	    my $sw_srds = $br2sw_srds{$br};
	    print STDOUT "{ Branch ID: $br }\n\n";
	    print STDOUT join ("\t", 'Column(i.e.,site)', '#{srds}'), "\n\n";
	    for (my $s = $beg_msa; $s < $end_msa; $s++) { print STDOUT join ("\t", $s, $sw_srds->[$s]), "\n"; }
	    print STDOUT "\n\n";
	}
=cut

   } # END of "if ($TEST_LEVEL_mk >3) {...}"


		# Main Process 3: Compute @map_alnshear. #

   (print STDOUT "[ Main Process 3: Compute \@map_alnshear. ]\n\n") if ($TEST_LEVEL_mk >1);

	my @map_alnshear = ();

	for (my $i=0; $i <$ct_brs; $i++) {

	    my $br_id = $br_ids_for_display->[$i];

	    my @linmap_alnshear = @{$br2linmap_alnshear{$br_id}}; # Here, an array, which is a value of %br2linmap_alnshear, is copied deliberately, to avoid inadvertent manipulation of @srd_map after it is output.

	    $map_alnshear[$i] = \@linmap_alnshear;
	}

   (print STDOUT "\n\n< 'mk_map_alnshear' Ended />\n\n") if ($TEST_LEVEL_mk >0);

	return  (\@map_alnshear, \%br2linmap_alnshear);

}	# End of the subroutine, 'mk_map_alnshear'. #


# mk_map_alnshear_clmbased:
#	This is a REVISED version of the predecessor subroutine, 'mk_map_alnshear'.
#	The only difference is that
#	this subroutine receives a set of "mini-class"es in the format of the output of 'list_minicls_shift_respos_clm_based' (in "MyPosShiftMap.pm"),
#	whereas the predecessor receives the set in the format of the output of '(br_)list_classes_shift_respos' & 'dpsm_infer_sep_patterns_for_classes' (both in "MyMalnComparison.pm").
#	[CREATED on 2020/03/21.]
#
# mk_map_alnshear:
#	This subroutine receives a set of "mini-class"es of position-shifts,
#	as well as some supplementary information,
#	then it creates an Alignment-Shear Map.
#	The script was created by somewhat modifying 'mk_srd_map' (in "MySRD_Map.pm").
#	[CREATION STARTED on 2020/03/08, ENDED on 2020/03/09.]
#
# mk_srd_map (in "MySRD_Map.pm"):
#	This subroutine receives some variables characterizing a phylogenetic tree,
#	and some outputs of the likelihood analysis (under a given substitution model) 
#	performed on a given multiple sequence alignment (MSA),
#	then it  creates a Substitutional Residue-Difference Map (abbr. "SRD Map").
#	The script was created by taking advantage of some portions of
#	'detect_purge_cands2' (in "MyDetect_purge_cands.pm").
#	[CREATION STARTED on 2020/03/05, ENDED on 2020/03/05.]
#	
# [ INPUT: ($if_observed, @br_ids_for_display, $char_abs, $char_pres, $topnode_id, \%node2pa, \%node2ntfreqs, \%br2tr_mtrx, \%br2set_sw_jnt_probs, \%node2set_sw_cond_probs, \@set_clm_probs, \@info_psm_hst0, $beg_msa, $end_msa, \%br_attr, \%br2equiv), where ] ... for the predecessor, 'mk_srd_map'. #
#
# INPUTS: (@br_ids_for_display, $char_abs, $char_pres, $char_noshear, $topnode_id, \%node2pa, 
#	\@set_miniclasses_shift_respos, \@set_sep_patterns_for_miniclasses, \%miniclass2trivial,
#	\@info_psm_hst0, $beg_msa, $end_msa, \%br_attr, \%br2equiv), where
#
#	@br_ids_for_display stores ALL branch IDs
#	(, each of which is actually the ID of the node at the lower-end of the branch,)
#	in the order to be displayed as the rows of the SRD Map.
#
#	$char_abs/$char_pres is the character that represents an "absence"/"presence" state;
#	$char_noshear is the character indicating that the position (= site x branch) has NO shear. 
#
#	$topnode_id is the ID of the top-node (i.e., root);
#	%node2pa = ($node_id => $parent_id, ...) for the nodes except the root.
#
#	@{$set_miniclasses_shift_respos[$mc]} = ($shift, $beg_site, $end_site, \@affected_regions, \@seqs_affected)
#	is for the $mc th mini-class of the run of shifts of residue positions, where
#	$shift is the value of shift,
#	$beg/end_site is the beginning/ending site (0-based full-closed) of the region with the shift value $shift,
#	@{$affected_regions[$rg]} = ($beg_rg, $end_rg) is for the $rg th region affected, with $beg/end_rg being its beginning/ending coordinate (in the 0-based full-closed convention),
#	@seqs_affected stores the (0-based) indices of the sequences affected by (i.e., accommodating) such a run of shifts;
#	[Typically, \@set_miniclasses_shift_respos is an element of the main output of 'list_minicls_shift_respos_clm_based' (in "MyPosShiftMap.pm").]
#
#
#	[ @{$set_sep_patterns_for_miniclasses[$mc]->[$pt]} = (\%node2parsstat, \%br2change, $ct_changes) is for the $pt th pattern separating the sequences affected by the $mc th mini-class of position shifts, where ] ... for the predecessor, 'mk_map_alnshear'.
#
#	@{$set_sep_patterns_for_miniclasses[$mc]} = (\%node2parsstat, \%br2change, $ct_changes) is for the $mc th mini-class of position shifts, where # MODIFIED in creation of 'mk_map_alnshear_clmbased' (on 2020/03/21). #
#
#	%node2parsstat = ($node_id => [$state, $ct_pres_ext], ...),
#	%br2change = ($branch_id => [$state_upper_end, $state_lower_end], ...), and
#	$ct_changes is the number of "changes" necessary for separating the affected sequences from the rest;
#	[Typically, \@set_sep_patterns_for_miniclasses is an element of the main output of 'list_minicls_shift_respos_clm_based' (in "MyPosShiftMap.pm").]
#
#	%miniclass2trivial = ($indx_miniclass => 1, ...) for trivial mini-classes;
#
#	@info_psm_hst0 = (\@set_cls_gblocks, \@totcts0_indels_for_cls, \@psm_hst0_for_cls),
#	which is an output of 'cal_init_psm_cands_for_sgl_msa' in 'MyTreeMap_indels_spt_odr2.pm',
#	gives information on the Dollo parsimonious indel history giving rise to the input MSA gap configuration, where
#	
#	@{$set_cls_gblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end) is for the $gb th block in the $cl th cluster (i.e., gapped segment) in the input MSA, with
#		$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#		$len_gb being the length (in terms of #{columns}) of the block,
#		and $start/$end being the starting/ending coordinate of the block (in the 0-based full-closed convention),
#	
#	$totcts0_indels_for_cls[$cl] is the total number of insertion/deletion events in the Dollo parsimonious indel history resulting in the $cl th cluster (i.e., gapped segment) in the input MSA,
#	
#	@{$psm_hst0_for_cls[$cl]} = (\%br2indels0, \%node2blwise_gstat0) is for the Dollo parsimonious indel history resulting in the $cl th cluster (i.e., gapped segment) in the input MSA, where 
#		%br2indels0 = ($branch_id => \@list_indels0, ...) is the map of insertions/deletions onto branches, with
#		@{$list_indels0[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved), and where
#		%node2blwise_gstat0 = ($node_id => \@blwise_gapstats0, ...) is the map of the gap states onto the nodes,
#		  with $blwise_gapstats0[$bl] is the gap-state ($char_pres/abs) of the $bl th block at the node.
#
#	$beg/end_msa is the beginning/ending coordinates of the input MSA (in the 0-based full-closed convention).
#
#	%br_attr = ($br_id => \@attr, ...) for all branches, where $attr[$INDX_LENGTH] is the branch length.
#
#	%br2equiv = ($br_id => $equiv_br_id, ...) for those branches with equivalents.
#
# OUTPUTS: (\@map_alnshear, \%br2linmap_alnshear), where
#
#	$map_alnshear[$n]->[$s] = the Alignment-Shear status of the position specified by the site $s and the branch,  $br_ids_for_display[$n];
#	The "Alignment-Shear status" is:
#		 $char_abs if either end (or both ends) of the branch has (have) $char_abs in the initial gap-state,
#		 join (',', {list of mini-classes whose boundaries includ the position}),
#		 $char_noshear otherwise.
# 
#
#	%br2linmap_alnshear = ($br_id => \@linmap_alnshear, ...) for ALL branches, where
#	$linmap_alnshear[$s] is the Alignment-Shear status at the $s th site (more precisely, in the $s th column) of the input MSA, along the branch, $br_id.
#
# [ sub mk_map_alnshear (\@$$$$\%\@\@\%\@$$\%\%) { ] ... for the predecessor, 'mk_map_alnshear'.
#
sub mk_map_alnshear_clmbased (\@$$$$\%\@\@\%\@$$\%\%) {

	my ($br_ids_for_display, $char_abs, $char_pres, $char_noshear, $topnode_id, $node2pa, 
	$set_miniclasses_shift_respos, $set_sep_patterns_for_miniclasses, $miniclass2trivial,
	$info_psm_hst0, $beg_msa, $end_msa, $br_attr, $br2equiv) = @_;


   (print STDOUT "< 'mk_map_alnshear_clmbased' (\$char_abs = '$char_abs'; \$char_noshear = '$char_noshear'; $beg_msa <= {MSA coordinate} <= $end_msa) >\n\n") if ($TEST_LEVEL_mk>0);

	my $ct_brs = @{$br_ids_for_display};
	my $ct_brs2 = scalar (keys %{$br_attr});

	unless ($ct_brs == $ct_brs2) {
	    print STDOUT "[ERROR 'print_srd_map'] The number of branches to be displayed (=$ct_brs) CLASHES with the number of branches in ", '%{$br_attr}', " (=$ct_brs2) : $! !\n\n";
	    return (undef, undef);
	}

	my @set_all_nodes = (@{$br_ids_for_display}, $topnode_id);

   if ($TEST_LEVEL_mk >2) {
	print STDOUT "[ Secondary inputs ]\n\n";
	print STDOUT '@{$br_ids_for_display} = (', join (', ', @{$br_ids_for_display}), ") .\n\n";
	print STDOUT '@set_all_nodes = (', join (',', @set_all_nodes), ").\n\n\n";
   }


			# (Pre-1) Prepare the site-wise gap states (of the Dollo-parsimonious indel history) at every node. #

	# %{$node2sw_gstats0} = ($node_id => \@sw_gstats0, ...) for all nodes, where
	# $sw_gstats0[$s] is the gap state of the $s th site on the node $node_id
	# (usually according to the Dollo-parsimony principle).
	#
	my $node2sw_gstats0 = mk_node2sw_gstats (@{$info_psm_hst0}, $char_abs, $char_pres, $beg_msa, $end_msa, @set_all_nodes);
	unless (defined $node2sw_gstats0) {
	    print STDOUT "[ERROR 'mk_map_alnshear_clmbased'] The subroutine 'mk_node2sw_gstats' FAILED: $! !\n\n";
	    return (undef, undef, undef);
	}
   (print STDOUT " --- (Pre-1) Subroutine 'mk_node2sw_gstats' ended ...\n\n") if ($TEST_LEVEL_mk >1);

   if ($TEST_LEVEL_mk >3) {

	print STDOUT "[ Content of ", '%{$Node2sw_gstats0}', ", which is the output of 'mk_node2sw_gstats' ]\n\n";
	print STDOUT 'Site\Node', "\t", join (', ', @set_all_nodes), "\n\n";
	for (my $s = $beg_msa; $s <= $end_msa; $s++) { # Outer for-loop (over sites). #
	    my @sw_gstats_on_nodes = ();
	    foreach my $node (@set_all_nodes) {
		my $gstat = $node2sw_gstats0->{$node}[$s];
		(defined $gstat) or ($gstat = '?');
		push @sw_gstats_on_nodes, $gstat;
	    }
	    print STDOUT $s, "\t", join (', ', @sw_gstats_on_nodes), "\n";

	}				# End of the outer for-loop (over sites). #

	print STDOUT "\n\n";

   } # END of "if ($TEST_LEVEL_mk >3) {...}"


		# Main Process 1: Initialize %br2linmap_alnshear. #

   (print STDOUT "[ Main Process 1: Initialize \%br2linmap_alnshear. ]\n\n") if ($TEST_LEVEL_mk >1);

	my %br2linmap_alnshear;

	foreach my $br (@{$br_ids_for_display}) { # 1st Main outer foreach-loop (over all branches) #

=For the predecessor, 'mk_srd_map'.
	    my $set_sw_srds = $br2sw_srds{$br};

	    ##my $eq_br = $br2equiv->{$br};
	    my $brlen = $br_attr->{$br}[$INDX_LENGTH];
	    ##if ((defined $eq_br) and ($brlen == 0)) {
	    if ($brlen == 0) {
			# For a null artificial branch,
			#  set all relevant entries to be $char_abs.

		for (my $s = $beg_msa; $s <= $end_msa; $s++) { $set_sw_srds->[$s] = $char_abs; }

		next;
	    }
=cut # END of "For the predecessor, 'mk_srd_map'."


	    my @linmap_alnshear = ();

	    my $node_ue = $node2pa->{$br};
	    my $sw_gstats0_le = $node2sw_gstats0->{$br};
	    my $sw_gstats0_ue = $node2sw_gstats0->{$node_ue};

	    for (my $s = $beg_msa; $s <= $end_msa; $s++) {  # Middle for-loop (over relevant sites). #

		my $gstat0_le = $sw_gstats0_le->[$s];
		my $gstat0_ue = $sw_gstats0_ue->[$s];

		$linmap_alnshear[$s] = (($gstat0_le eq $char_abs) or ($gstat0_ue eq $char_abs)) ? $char_abs : $char_noshear ;

	    } # End of the middle for-loop (over relevant sites). #

	    $br2linmap_alnshear{$br} = \@linmap_alnshear;

	}  # End of the 1st Main outer foreach-loop (over all branches) #


		# Main Process 2: Compute %br2linmap_alnshear. #

   (print STDOUT "[ Main Process 2: Compute \%br2linmap_alnshear. ]\n\n") if ($TEST_LEVEL_mk >1);

		# $set_miniclasses_shift_respos, $set_sep_patterns_for_miniclasses, $miniclass2trivial,

	my $ct_mcls = @{$set_miniclasses_shift_respos};

	for (my $mc = 0; $mc < $ct_mcls; $mc++) { # 2nd Main outer for-loop (over mini-classes of position-shifts) #

	    if ($miniclass2trivial->{$mc}) { next; }

	    my ($shift, $beg_site, $end_site, $affected_regions, $seqs_affected)
		 = @{$set_miniclasses_shift_respos->[$mc]};

	    my $set_sep_patterns = $set_sep_patterns_for_miniclasses->[$mc];

	    my %br2sbjct = ();

			# MODIFIED in creation of 'mk_map_alnshear_clmbased' (on 2020/03/21). #

	    my ($node2parsstat, $br2change, $ct_changes) = @{$set_sep_patterns};

	    foreach my $br (keys %{$br2change}) { $br2sbjct{$br} = 1; }

=For the predecessor, 'mk_map_alnshear'.
	    for (my $pt = 1; $pt < 2; $pt++) { # 1st middle for-loop (over separating patterns) # Regarding the range of $pt, see 'split_unnatural_miniclasses' (in "MyMalnComparison.pm"). #

		my ($node2parsstat, $br2change, $ct_changes) = @{$set_sep_patterns->[$pt]};

		foreach my $br (keys %{$br2change}) { $br2sbjct{$br} = 1; }

	    } # End of the 1st middle for-loop (over separating patterns) #
=cut # End of "For the predecessor, 'mk_map_alnshear'."

			# END of "MODIFIED in creation of 'mk_map_alnshear_clmbased' (on 2020/03/21)." #

	    my @sbj_branches = sort {$a <=> $b} keys %br2sbjct;

	    foreach my $br (@sbj_branches) { # 2nd middle for-loop (over branches involved in the alignment shear). #

		my $linmap_alnshear = $br2linmap_alnshear{$br};

		foreach my $region (@{$affected_regions}) { # Inner for-loop (over affected regions). #
		   my ($beg_rg, $end_rg) = @{$region};

		   for (my $s = $beg_rg; $s <= $end_rg; $s++) { # Innermost for-loop (over sites). #

			my $alnshear = $linmap_alnshear->[$s];

			if ($alnshear eq $char_abs) { 
			    next; # Skip if the state is "absence". #

			} elsif ($alnshear eq  $char_noshear) {
			    $linmap_alnshear->[$s] = $mc;

			} else {
			    $linmap_alnshear->[$s] .= ",$mc";
			}

		   } # End of the innermost for-loop (over sites). #

		} # End of the inner for-loop (over affected regions). #

	    } # 2nd middle for-loop (over branches involved in the alignment shear). #

	} # End of the 2nd Main outer for-loop (over mini-classes of position-shifts) #

   if ($TEST_LEVEL_mk >3) {
	print STDOUT "[ Final Content of ", '%br2linmap_alnshear', ": ]\n\n";

	print STDOUT join ("\t", 'Column\Branch', @{$br_ids_for_display}), "\n\n";

	for (my $c = $beg_msa; $c <= $end_msa; $c++) { # Outer for-loop (over columns). #

	    my @clm_alnshearstats = ();
	    foreach my $br (@{$br_ids_for_display}) {
		my $alnshearstat = $br2linmap_alnshear{$br}->[$c];
		(defined $alnshearstat) or ($alnshearstat = 'undef');
		push @clm_alnshearstats, $alnshearstat;
	    }
	    print STDOUT join ("\t", $c, @clm_alnshearstats), "\n";

	} # End of outer for-loop (over columns). #

	print STDOUT "\n\n";

=Alternative idea
	foreach my $br (@{$br_ids_for_display}) {
	    my $sw_srds = $br2sw_srds{$br};
	    print STDOUT "{ Branch ID: $br }\n\n";
	    print STDOUT join ("\t", 'Column(i.e.,site)', '#{srds}'), "\n\n";
	    for (my $s = $beg_msa; $s < $end_msa; $s++) { print STDOUT join ("\t", $s, $sw_srds->[$s]), "\n"; }
	    print STDOUT "\n\n";
	}
=cut

   } # END of "if ($TEST_LEVEL_mk >3) {...}"


		# Main Process 3: Compute @map_alnshear. #

   (print STDOUT "[ Main Process 3: Compute \@map_alnshear. ]\n\n") if ($TEST_LEVEL_mk >1);

	my @map_alnshear = ();

	for (my $i=0; $i <$ct_brs; $i++) {

	    my $br_id = $br_ids_for_display->[$i];

	    my @linmap_alnshear = @{$br2linmap_alnshear{$br_id}}; # Here, an array, which is a value of %br2linmap_alnshear, is copied deliberately, to avoid inadvertent manipulation of @srd_map after it is output.

	    $map_alnshear[$i] = \@linmap_alnshear;
	}

   (print STDOUT "\n\n< 'mk_map_alnshear_clmbased' Ended />\n\n") if ($TEST_LEVEL_mk >0);

	return  (\@map_alnshear, \%br2linmap_alnshear);

}	# End of the subroutine, 'mk_map_alnshear_clmbased'. #


# print_map_alnshear:
#	This subroutine prints out an Alignment-Shear Map
#	(, which was typically created via 'mk_map_alnshear',)
#	by formatting it so that the map will be decently visible.
#	This subroutine was created by modifying the subroutine, 'print_srd_map' (in "MySRD_Map.pm").
#	[CREATION STARTED on 2020/03/08, ENDED on 2020/03/dd.]
#
# print_srd_map (in "MySRD_Map.pm"):
#	This subroutine prints out a Substitutional Residue-Difference Map
#	(, which was typically created via 'mk_srd_map',)
#	by formatting it so that the map will be decently visible.
#	This subroutine was created by refering to the subroutine, 'pr_map_shifts_respos',
#	in the script, 'classify_msa_errors_via_mblks.alpha2.pl',
#	in the directory, "$HOME/Ezawa_works_UH/Main_Project/Secondary_Projects/OpenSource/Exported/ComplLiMment_P/Sample_Scripts/".
#	[CREATION STARTED on 2020/03/05, ENDED on 2020/03/06.]
#
# pr_map_shifts_respos:
#	This subroutine prints out a position-shift map
#	by formatting it so that the map will be decently visible.
#	[CREATED on Oct 15 (Thu), 2015.]
#
# [ INPUTS: (\*${$fh}, \@sequence_ids, \@map_shifts_respos, $char_gap, $dis_len_shift), where ] ... for the predecessor, 'pr_map_shifts_respos'.
#
# [ INPUTS: (\*${$fh}, \@br_ids_for_display, \%node_id2name, \@srd_map, $beg_msa, $end_msa, $char_abs, $dis_len_srd, $ct_eff_dgts), where ] ... for the predecessor, 'print_srd_map'.
#
# INPUTS: (\*${$fh}, \@br_ids_for_display, \%node_id2name, \@map_alnshear, $beg_msa, $end_msa, $dis_len_asmap, $min_wdth_fld_asmap), where
#
#	@br_ids_for_display stores ALL branch IDs
#	(, each of which is actually the ID of the node at the lower-end of the branch,)
#	in the order to be displayed as the rows of the SRD Map.
#
#	%node_id2name = ($node_ID => $node_name, ...)  for all NAMED nodes (NOT including top_node_ID => $TOP_NODE).
#
#	$map_alnshear[$n]->[$s] = the alignment-shear status at the site $s and along the branch,  $br_ids_for_display[$n];
#
#	$beg/end_msa is the beginning/ending coordinates of the input MSA (in the 0-based full-closed convention).
#
#	[ $char_abs is the character that represents an "absence" state. ] ... for the predecessor, 'print_srd_map'.
#
#	$dis_len_asmap (positive integer) is the maximum length of the run of alignment-shear statuses to be displayed in a line.
#
#	$min_wdth_fld_asmap (positive integer) is the minimum width of each field (including the spacer-part) of the Alignment-Shear Map.
#
#
# OUTPUTS: 1 if the subroutine succeeded,
#	   0 if it failed.
#
sub print_map_alnshear (\*\@\%\@$$$$) {

	my ($fh, $br_ids_for_display, $node_id2name, $map_alnshear, $beg_msa, $end_msa, $dis_len_asmap, $min_wdth_fld_asmap) = @_;

	my $ct_brs = @{$br_ids_for_display};
	unless ($ct_brs >0) {
	    print STDOUT "[INFORMATION 'print_map_alnshear'] NO SEQUENCES in the alignment !\n\n";
	    return 1;
	}

	unless (@{$map_alnshear} == $ct_brs) {
	    print STDOUT "[ERROR 'print_map_alnshear'] The number of rows in the map (=", scalar (@{$map_alnshear}), ") CLASHES with the number of branches to be displayed (=$ct_brs): $! !\n\n";
	    return 0;
	}

	my $in_array_len = @{$map_alnshear->[0]}; 
	unless ($in_array_len > $end_msa) {
	    print STDOUT "[ERROR 'print_map_alnshear'] The length of the input array, ", '@{$map_alnshear->[0]}', " is NOT LARGER THAN \$end_msa (=$end_msa): $! !\n\n";
	    return 0;
	}


		# Create a set of row-labels. #

	my @row_labels = ();
	my $maxlen_rlbl = 0;
	for (my $n=0; $n <$ct_brs; $n++) {
	    my $br_id  = $br_ids_for_display->[$n];
	    my $node_name = $node_id2name->{$br_id};
	    my $rlabel = (defined $node_name) ? $node_name : $br_id ;

	    $row_labels[$n] = $rlabel;

	    my $len = length ($rlabel);
	    if ($maxlen_rlbl < $len) { $maxlen_rlbl = $len; }
	}

	####my $wdth_rlbl = ($maxlen_rlbl > 6) ? $maxlen_rlbl : 6;
	##my $wdth_rlbl = ($maxlen_rlbl > length ('Branch/Site') ) ? $maxlen_rlbl : length ('Branch/Site') ; # MODIFIED on 2020/03/11. #
	my $wdth_rlbl = ($maxlen_rlbl > length ('Branch\Site') ) ? $maxlen_rlbl : length ('Branch\Site') ; # => CORRECTED on 2020/03/11. #

=For the predecessor, 'print_srd_map'. (1)
		# Set the maginification ratio. #

	my $magratio = 1;
	for (my $i = 0; $i < $ct_eff_dgts; $i++) { $magratio *= 10; }

	my $wdth_fld = ($ct_eff_dgts + 3 > length ($end_msa) +2 ) ? $ct_eff_dgts + 3 : length ($end_msa) + 2;
=cut # End of "For the predecessor, 'print_srd_map'. (1)"


		# Set the field width. #

	my $wdth_fld = ($min_wdth_fld_asmap < length ($end_msa) + 2) ? $min_wdth_fld_asmap : length ($end_msa) + 2 ;

	for (my $n=0; $n<$ct_brs; $n++) {  # Preliminary for-loop (over branches to be displayed). #

	    my $linmap_alnshear = $map_alnshear->[$n];

	    for (my $s = $beg_msa; $s <= $end_msa; $s++) { # Inner for-loop (over sites to be displayed). #
		unless (defined $linmap_alnshear->[$s]) { next; }
		my $wdth_ass = length ($linmap_alnshear->[$s]);
		if ($wdth_fld < $wdth_ass +1) { $wdth_fld = $wdth_ass +1; }
	    } # Inner for-loop (over sites to be displayed). #

	} # Preliminary for-loop (over branches to be displayed). #


		# Set the $quotient #

	my $map_len = $end_msa - $beg_msa + 1;

	my $quotient = int (($map_len -1) / $dis_len_asmap);

	#    sprintf ("%-${maxlen_rlbl}s", $row_labels[$n]);


	for (my $r=0; $r<=$quotient; $r++) {  # 1st outer for-loop (over segmented displayes). #

	    my $beg_dis = $r * $dis_len_asmap;
	    my $end_dis = ($r == $quotient) ? $map_len - 1 : ($r+1) * $dis_len_asmap - 1 ;
	    $beg_dis += $beg_msa;
	    $end_dis += $beg_msa;

	    ##my $format = "%-10s\t" . "%5d" x (($r==$quotient) ? $map_len - $quotient * $dis_len_srd : $dis_len_srd) . "\n"; # For the predecessor, 'pr_map_shifts_respos'. #
	    my $format_clbl = "%-${wdth_rlbl}s\t" . "%${wdth_fld}d" x (($r==$quotient) ? $map_len - $quotient * $dis_len_asmap : $dis_len_asmap) . "\n";
	    ##my $format = "%-${wdth_rlbl}s\t" . "%${wdth_fld}s" x (($r==$quotient) ? $map_len - $quotient * $dis_len_asmap : $dis_len_asmap) . "\n";

	    ######printf $fh $format, 'Seq_ID', ($beg_dis .. $end_dis);
	    ####printf $fh $format_clbl, 'Seq_ID', ($beg_dis .. $end_dis); # CORRECTED on 2020/03/09.
	    ##printf $fh $format_clbl, 'Branch/Site', ($beg_dis .. $end_dis); # MODIFIED on 2020/03/11.
	    printf $fh $format_clbl, 'Branch\Site', ($beg_dis .. $end_dis); # => CORRECTED on 2020/03/11.
	    print $fh "\n";

	    for (my $k=0; $k<$ct_brs; $k++) {  # Middle for-loop (over branches to be displayed). #

		##my $out_string = sprintf ("%-${wdth_rlbl}s\t", $br_ids_for_display->[$k]);
		my $out_string = sprintf ("%-${wdth_rlbl}s\t", $row_labels[$k]); # CORRECTED on 2020/03/09. #

		my $linmap_alnshear = $map_alnshear->[$k];
		my $ct_sites = @{$linmap_alnshear};

		my $end_dis2 = ($ct_sites-1 < $end_dis) ? $ct_sites-1 : $end_dis;

		for (my $s=$beg_dis; $s<= $end_dis2; $s++) { # Inner for-loop (over sites). #

		   my $alnshearstat = $linmap_alnshear->[$s];

		   $out_string .= sprintf ("%${wdth_fld}s", $alnshearstat) ;

=For the predecessor, 'print_srd_map'. (2)
		   my $srd = $row_srds->[$s];
		   my $out_srd;
		   if ($srd eq $char_abs) {
			$out_srd = $srd;
		   } else {
			$out_srd = int ($magratio * $srd + 1/2) / $magratio;
			if ($out_srd > 0) { $out_srd =~ s/^0//; }
		   }
		   ##$out_string .= ($shift eq $char_gap) ? sprintf ("%5s", "'$char_gap'") : sprintf ("%5d", $shift) ; # For the predecessor, 'pr_map_shifts_respos'. #
		   $out_string .= sprintf ("%${wdth_fld}s", $out_srd) ;
=cut # End of "For the predecessor, 'print_srd_map'. (2)"

		}				# End of the inner for-loop (over sites). #

		print $fh $out_string, "\n";
	    
	    }			# End of the middle for-loop (over branches to be displayed). #
	    print $fh "\n\n\n";

	}			# End of the 1st outer for-loop (over segmented displayes). #

	return 1;

}	# End of the subroutine, 'print_map_alnshear'. #


1;		# a "spell" which makes the module "use"-able and/or "require"-able # (ADDED on 2020/03/10.)!! #


