package MyTreeMap_II_xgap;
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#	Version 0.5: Copyright (C) 2019? Kiyoshi Ezawa
#	Version 0.6: Copyright (C) 2020 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyTreeMap_II_xgap.pm:
#	This package is a collection of subroutines
#	to BASICALLY perform Sankoff's parsimony algorithm,
#	but DEAL WITH GAPS via Farris' Dollo parsimony principle.
#	The package was created by modifying the predecessor, "MyTreeMap_II.pm" (in "../MyPerlModules_DENSERM/"),
#	and with the assist of the subroutine, 'map_parsim_states_dollo_1clm',
#	in "MyTreeMap_D.pm" (in "../MyPerlModules_LOLIPOG/").
#	[CREATION Started on 2020/03/13, ENDED on 2020/03/dd.]
#
###################################################################################
#	DENSERM_P: A package to DEtect Negative SElection on Recurrent Mutations
#			(almost exclusively via Perl modules and scripts).
#	Version 0.3: Copyright (C) 2012 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyTreeMap_II.pm:
#	This package performs more refined algorithms than 'MyTreeMap.pm',
#	to map various properties, or their changes,
#	onto the nodes or the branches of a phylogenetic tree.
#	(CREATED on June 16 (Wed), 2010)
#	(REVISED on Nov 27 (Sun), 2011)
#	The subroutine 'min_penalties_sankoff_bo' now excludes from later consideration
#	the characters on each node that will NEVER be included in any parsiminious states.
#
#	Some subroutines in this module and "MyTreeMap.pm" are based on the algorithms or methods described in:
#	"Inferring Phylogenetics",
#	written by Felsenstein J (Sinauer Associates, Inc.; Sunderland, MA, USA; 2004).
#
# MyTreeMap.pm:
#	This package maps various properties, or their changes,
#	 onto the nodes or the branches
#	of a (molecular) phylogenetic tree.
#	It mainly uses the parsimony principle,
#	but sometimes uses other methods, such as the maximum likelihood.

use strict;
use warnings;
use MyArrayManipu qw( diff_array union_array nr_array);
use MyArrayManipu_num qw( diff_nmarray );

use MyTaxonomy qw(
	fetch_ancestry 
	fetch_last_common_ancestor_ov2
	fetch_common_ancestors
		);
	## fetch_ext_offsprings fetch_all_offsprings
	## fetch_ext_offsprings_br

use MyTreeManipu_II qw(
		);
	## input_tree_file read_tree_file br_output_tree
	##make_node2children
	## input_subtree copy_tree
	## output_tree draw_subtree topo_subtree
	## shift_topnode move_topnode segregate_subset_nodes
	## sum_branch_lengths intra_lineage_distance inter_node_distance
	## fetch_boot_series extract_hidden_bootstrap
	## resolve_multipartite

use MyTreeManipu_III qw(
	make_node2abs_depth
	);

	# ADDED in creation of 'MyTreeMap_II_xgap.pm' (on 2020/03/13). (USE) #
use MyTreeMap_D qw(
	map_parsim_states_dollo_1clm	
	);
	# End of "ADDED in creation of 'MyTreeMap_II_xgap.pm' (on 2020/03/13). (USE)" #

use Exporter;

our @ISA = qw(Exporter);

		# MODIFIED in creation of 'MyTreeMap_II_xgap.pm' (on 2020/03/13). (USE) #
our @EXPORT_OK = qw(
	min_penalties_sankoff_xgap_bo
	enum_node2parsstat_sankoff_xgap_td
	map_parsim_states_sankoff_xgap_clm
	map_parsim_states_sankoff_xgap_1clm
	locate_parsim_changes_xgap
	);
	# [Set aside for later use:]
	# map_parsim_states_sankoff_msa
	# map_parsim_gapstates_sankoff
	# [End of "Set aside for later use:"]

		# END of "MODIFIED in creation of 'MyTreeMap_II_xgap.pm' (on 2020/03/13). (USE)" #

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
##our $INDX_LENGTH = 0;
##our $INDX_BOOT = 1;
##our $START_GROUP = '(';	# the symbol marking the start of the grouping of nodes.
##our $END_GROUP = ')';  # the symbol marking the end of the grouping of nodes.

	## These two indices specify where the information is stored
	## in %nodee2pars_stats = ($node_id => \@stat_info, ...) #
our $INDX_STAT = 0; # The state (or the character) mapped onto the node.
our $INDX_SBTOT_PEN = 1; # The subtotal of penalties on the subtree under $node_id. 


	## The following parameters specify the verboseness of the test output.
	## (0, silent; 1, very modest; 2, modest; 3, medium; 4, verbose; 5, very verbose; 6, extremely verbose.)

	# For 'map_parsim_states_sankoff_xgap_clm'. #
our $TEST_LEVEL_mpssxc = 0;
##our $TEST_LEVEL_mpssxc = 5;


	### 5 important hash tables that are used in this Module
	### in order to describe a tree
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.


	### DEFINITIONS of the SUBROUTINES ###

# min_penalties_sankoff_xgap_bo:
#	This subroutine was created by modifying 'min_penalties_sankoff_bo'.
#	The difference from the predecessor is that
#	this new subroutine deals with gaps via Farris' Dollo parsimony principle,
#	and, if the subroutine encounters a node with an "absence" state,
#	it regards the node as "missing information" and EXCLUDES it from Sankoff's parsimony analysis.
#	[CREATION STARTED on 2020/03/13, ENDED on 2020/03/13.]
#
# min_penalties_sankoff_bo:
#	This satellite subroutine of 'map_parsim_states_sankoff_clm'
#	assigns minimum penalties on the subtree 
#	under the node in question under the condition that
#	the node  has particular characters.
#	(REVISED on Nov 27 (Sun), 2011)
#	A process is introduced to remove characters at each node that will NEVER be included
#	in any parsimonious states as soon as possible.
#
# [ INPUT: (\%node2children, \%node2depth, \%exnode2charindx, \@pen_mtrx), where ] ... for the predecessor, 'min_penalties_sankoff_bo'.
#
# INPUT: (\%node2children, \%node2depth, \%exnode2charindx, \@pen_mtrx, $char_abs,\%node2dparsstat), where
#
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2depth = ($node_id => $depth, ...) for ALL nodes, with
#	$depth defined as the number of nodes from the top node (inclusive) to themselves (inclusive);
#	%exnode2charindx = ($node_id => $indx of the character at the node, ...)
#	for the exterior nodes. The $indx conforms to @pen_mtrx;
#	$pen_mtrx[$i]->[$j] is the penalty imposed on the change
#	from the $i th character to the $j th character;
#
#		# (ADDED in creation of 'min_penalties_sankoff_xgap_bo' on 2020/03/13.)
#	$char_abs is the character representing the "absence" state (i.e., gap);
#
#	%node2dparsstat = ($node_id => \@state_info, ...) for the 
#	Dollo parsimonious scenario of the gap configuration of the clumn mapped onto the nodes,
#	with
#	@state_info = ($stat, $ct_pres_ext), where
#	$stat (= $char_abs/pres) is the gap-status on the node,
#	$ct_pres_ext = #{child nodes with #{exterior descendants with state ne $char_abs}>0} + #{parent with state eq $char_pres};
#		# End of "(ADDED in creation of 'min_penalties_sankoff_xgap_bo' on 2020/03/13.)"
#
# OUTPUT: \%node2min_penalties, where
#	[ %node2min_penalties = ($node_id => \@min_penalties, ...) for ALL nodes, ] ... for the predecessor, 'min_penalties_sankoff_bo'.
#	%node2min_penalties = ($node_id => \@min_penalties, ...) for ALL nodes with "presence" states, # MODIFIED in creation of 'min_penalties_sankoff_xgap_bo'.
#
#	with $min_penalties[$i] being the minimum summed penalty
#	on the subtree under $node_id under the condition that the node has
#	the $i th character in the matrix @pen_mtrx.
#
# [ sub min_penalties_sankoff_bo (\%\%\%\@) {
#	my ($node2ch, $node2depth, $exnode2charindx, $pen_mtrx) = @_; ] ... for the predecessor, 'min_penalties_sankoff_bo'.
#
sub min_penalties_sankoff_xgap_bo (\%\%\%\@$\%) {

	my ($node2ch, $node2depth, $exnode2charindx, $pen_mtrx, $char_abs, $node2dparsstat) = @_;

	my $ct_chars = @{$pen_mtrx};

	my %node2min_penalties;

	my @nodes = sort {$node2depth->{$b} <=> $node2depth->{$a} } keys %{$node2depth};
	foreach my $node (@nodes) { # Main outer foreach-loop #

			# (ADDED in creation of 'min_penalties_sankoff_xgap_bo' on 2020/03/13.) (1) #
	    my $gstate_info = $node2dparsstat->{$node};
	    my $gstat = $gstate_info->[0];
	    if ($gstat eq $char_abs) { next; } # Skip if the node has an "absence" state.
			# End of "(ADDED in creation of 'min_penalties_sankoff_xgap_bo' on 2020/03/13.) (1)" #

	    my $charindx = $exnode2charindx->{$node};
	    if (defined $charindx) { # An exterior node #
		my @min_penalties = ();
		$min_penalties[$charindx] = 0;
		$node2min_penalties{$node} = \@min_penalties;
		next;
	    }

	    my $children = $node2ch->{$node};
	    unless (defined $children) {
		print STDERR "[ERROR 'min_penalties_sankoff_xgap_bo'] The 'interior' node '$node'  has NO CHILDREN: $! !\n";
		return undef;
	    }

	    my @min_penalties = ();
	    my $minmin_penalty; # ADDED to simplify the downstream analyses on Nov 27 (Sun), 2011. #
	    for (my $i=0; $i<$ct_chars; $i++) { # Mid-outer for-loop (over the possible characters on the node) #
		my $min_penalty = 0;
		my $pen_row = $pen_mtrx->[$i];
		foreach my $ch (@{$children}) { # Mid-inner for-loop (over the children) #

			# (ADDED in creation of 'min_penalties_sankoff_xgap_bo' on 2020/03/13.) (2) #
		    my $gstate_info_ch = $node2dparsstat->{$ch};
		    my $gstat_ch = $gstate_info_ch->[0];
		    if ($gstat_ch eq $char_abs) { next; } # Skip if the child node has an "absence" state.
			# End of "(ADDED in creation of 'min_penalties_sankoff_xgap_bo' on 2020/03/13.) (2)" #

		    my $min_pens_ch = $node2min_penalties{$ch};
		    unless (defined $min_pens_ch) {
			print STDERR "[ERROR 'min_penalties_sankoff_xgap_bo'] The child node '$ch' of the node '$node' has NO INFORMATION on the MINIMUM PENALTIES: $! !\n";
			return undef; 
		    }

		    my $sb_min_pen;
		    for (my $j=0; $j<$ct_chars; $j++) { # Inner for-loop (over the possible characters on the child) #
			my $pen = $min_pens_ch->[$j];
			(defined $pen) or next;
			$pen += $pen_row->[$j];

			((defined $sb_min_pen) and ($sb_min_pen <= $pen)) and next;
			$sb_min_pen = $pen;
		    }				# End of the inner for-loop (over the possible characters on the child) #
		    $min_penalty += $sb_min_pen;

		}		# End of the mid-inner for-loop (over the children) #

		$min_penalties[$i] = $min_penalty;
		(!(defined $minmin_penalty) or ($minmin_penalty > $min_penalty)) and ($minmin_penalty = $min_penalty); # ADDED to simplify the downstream analyses on Nov 27 (Sun), 2011. #

	    }			# End of the mid-outer for-loop (over the possible characters on the node) #


=DISABLED, hoping that the BUG is Fixed (on 2020/03/21).

	    	# ADDED to simplify the downstream analyses
		# by removing characters that canNOT be included in any parsimonious states
		# (on Nov 27 (Sun), 2011). #

	    for (my $i=0; $i<$ct_chars; $i++) { # Trailing mid-outer for-loop (over possible characters on the node (switches to the parent node in the middle)) #
		my $cond_min_pen = $min_penalties[$i];
		((defined $cond_min_pen) and ($cond_min_pen> $minmin_penalty)) or next;
		my $pen_row = $pen_mtrx->[$i];
		my $if_ok = 1;
		for (my $j=0; $j<$ct_chars; $j++) { # Trailing inner for-loop (over possible characters on the node in question) #
		    my $alt_minpen = $min_penalties[$j];
		    (defined $alt_minpen) or next;
		    $alt_minpen += $pen_row->[$j];
		    if ($cond_min_pen > $alt_minpen) {
			$if_ok = 0;
			last;
		    }
		}		# End of the trailng inner for-loop (over possible characters on the node in question) #

		unless ($if_ok) { # Remove characters that will NEVER be included in parsiminious states #
		   $min_penalties[$i] = undef;
##=for TESTING the new function
##	print STDERR "[WARNING 'min_penalties_sankoff_xgap_bo'] The $i th character was removed on the node '$node' ... \n";
##=cut
		}
	    }		# End of the trailing mid-outer for-loop (over possible characters on the node (switches to the parent node in the middle)) #
=cut # END of "DISABLED, hoping that the BUG is Fixed (on 2020/03/21)."

	    $node2min_penalties{$node} = \@min_penalties;

	}			# End of the main outer foreach-loop #

	return \%node2min_penalties;

}		# End of the subroutine 'min_penalties_sankoff_xgap_bo' #


# enum_node2parsstat_sankoff_xgap_td:
#	This subroutine was created by modifying 'enum_node2parsstat_sankoff_td'.
#	The difference from the predecessor is that
#	this new subroutine deals with gaps via Farris' Dollo parsimony principle,
#	and, if the subroutine encounters a node with an "absence" state,
#	it regards the node as "missing information" and EXCLUDES it from Sankoff's parsimony analysis.
#	[CREATION STARTED on 2020/03/13, ENDED on 2020/03/13.]
#
#	[A BUG fixed on 2020/03/21]:
#	Before this fixation, the subroutine ERRONEOUSLY returned the set of %node2parsstat's
#	that do NOT specify the states of some nodes if there is a node whose parent has an "absence" state.
#	Now, such a BUG was fixed.
#
#	[A BUG fixed on 2020/03/22]:
#	Before this fixation, the subroutine could NOT initialize the set of \%node2parsstat's
#	when the root's parsimonious state is an "absence" state.
#	Now, such a BUG was fixed.
#
# enum_node2parsstat_sankoff_td:
#	This satellite subroutine of 'map_parsim_states_sankoff_clm'
#	enumerates the maps from the nodes to the parsimonious states
#	in a top-down manner.
#
#	DEBUGGED on Oct 13 (Wed), 2010:
#	When there are multiple maps of parsimonious states,
#	the old way of copying a map just copied the references of arrays,
#	which could cause errors when the output is manipulated later.
#
# [ INPUT: (\%node2pa, \%node2depth, \%node2min_penalties, \@pen_mtrx, $fl_full_mode), where ] ... for the predecessor, 'enum_node2parsstat_sankoff_td'.
#
# INPUT: (\%node2pa, \%node2depth, \%node2min_penalties, \@pen_mtrx, $char_abs, \%node2dparsstat, $fl_full_mode), where
#
#	%node2parent = ($node_id => $parent_id,...) for the non-top nodes;
#	%node2depth = ($node_id => $depth, ...) for ALL nodes, with
#	$depth defined as the number of nodes from the top node (inclusive) to themselves (inclusive);
#	%node2min_penalties = ($node_id => \@min_penalties, ...), with
#	$min_penalties[$i] being the minimum summed penalty on the subtree
#	under $node_id under the condition that  the node has the $i th character;
#	$pen_mtrx[$i]->[$j] is the penalty imposed on the change
#	from the $i th character to the $j th character;
#
#		# (ADDED in creation of 'enum_node2parsstat_sankoff_xgap_td' on 2020/03/13.)
#	$char_abs is the character representing the "absence" state (i.e., gap);
#
#	%node2dparsstat = ($node_id => \@state_info, ...) for the 
#	Dollo parsimonious scenario of the gap configuration of the clumn mapped onto the nodes,
#	with
#	@state_info = ($stat, $ct_pres_ext), where
#	$stat (= $char_abs/pres) is the gap-status on the node,
#	$ct_pres_ext = #{child nodes with #{exterior descendants with state ne $char_abs}>0} + #{parent with state eq $char_pres};
#		# End of "(ADDED in creation of 'enum_node2parsstat_sankoff_xgap_td' on 2020/03/13.)"
#
#	$fl_full_mode = 1 if you want the full mode,
#			= 0 if the brief mode is enough.
#
# OUTPUT: \@set_node2parsstat, where
#
#	%{$set_node2parsstat[$opt]} = ($node_id => \@state_info, ...) for the $opt th
#	option of the parsimonious states of the site (or column) mapped onto the nodes,
#	with
#	@state_info = ($char_indx, $sum_penal), where
#	$char_indx is the index of a parsimonious character on the node,
#	(BUT = $char_abs if the node has an "absence" state (... ADDED in creation of 'enum_node2parsstat_sankoff_xgap_td' on 2020/03/13),
#
#	$sum_penal is the minimum summation of the penalties across the subtree under the node,
#	(BUT is undef if $char_indx = $char_abs (... ADDED in creation of 'enum_node2parsstat_sankoff_xgap_td' on 2020/03/13)).
#
#	If $fl_full_mode = 1, the information on ALL the nodes is given.
#	If $fl_full_mode = 0, ONLY the information on the TOP node  is given.
#
# [ sub enum_node2parsstat_sankoff_td (\%\%\%\@$) {
#	my ($node2pa, $node2depth, $node2min_penalties, $pen_mtrx, $fl_full_mode) = @_; ] ... for the predecessor, 'enum_node2parsstat_sankoff_td'.
#
sub enum_node2parsstat_sankoff_xgap_td (\%\%\%\@$\%$) {

	my ($node2pa, $node2depth, $node2min_penalties, $pen_mtrx, $char_abs, $node2dparsstat, $fl_full_mode) = @_;

	my $ct_chars = @{$pen_mtrx};

	my @nodes = sort {$node2depth->{$a} <=> $node2depth->{$b}} keys %{$node2depth};

	my @set_node2parsstat = ();

	foreach my $node (@nodes) { # The main outermost foreach-loop (over all nodes) #

			# (ADDED in creation of 'enum_node2parsstat_sankoff_xgap_td' on 2020/03/13.) (1) #
	    my $gstate_info = $node2dparsstat->{$node};
	    my $gstat = $gstate_info->[0];

			# REVISED in DEBUGGING on 2020/03/22. #
	    if ($gstat eq $char_abs) {

		if (defined $node2pa->{$node}) {

			# The node is NOT the root. => Add the node to the existing %{$node2parsstat}'s. #

		    foreach my $node2parsstat (@set_node2parsstat) {
			$node2parsstat->{$node} = [$char_abs];
		    }

		} else {
			# The node is the root. => Initialize a %node2parsstat. #

		    my %node2parsstat = ($node => [$char_abs]);
		    push @set_node2parsstat, \%node2parsstat;
		}

		next;
	    }

=A BUGGY BLOCK (found on 2020/03/22).
	    if ($gstat eq $char_abs) {
	        foreach my $node2parsstat (@set_node2parsstat) {
		    $node2parsstat->{$node} = [$char_abs];
		}
		next;
	    }
=cut # END of "A BUGGY BLOCK (found on 2020/03/22)."
			# END of "REVISED in DEBUGGING on 2020/03/22." #


			# End of "(ADDED in creation of 'enum_node2parsstat_sankoff_xgap_td' on 2020/03/13.) (1)" #

	    my $cond_min_penalties = $node2min_penalties->{$node}; 
	    unless (defined $cond_min_penalties) {
		print STDERR "[ERROR 'enum_node2parsstat_sankoff_xgap_td'] The node '$node' has NO CONDITIONAL MINIMUM PENALTIES: $! !\n";
		return undef;
	    }

			# (ADDED in creation of 'enum_node2parsstat_sankoff_xgap_td' on 2020/03/13.) (2) #
	    my $parent = $node2pa->{$node};
	    my $gstat_pa;
	    if (defined $parent) {
		my $gstate_info_pa = $node2dparsstat->{$parent};
		$gstat_pa = $gstate_info_pa->[0];
	    }
			# End of "(ADDED in creation of 'enum_node2parsstat_sankoff_xgap_td' on 2020/03/13.) (2)" #

	    if (!(defined $parent) or ($gstat_pa eq $char_abs)) { # The Big if #

		# Either the ABSOLUTE top node or the top of the sub-tree with "presence" states.
		# [ --> Initialize the @set_node2parsstat # ] ... A BUG fixed on 2020/03/21.
		#
		#		(REVISED to Fix the BUG on 2020/03/21.)
		# --> if (!(defined $parent)} {
		#	Initialize the @set_node2parsstat
		#     } else {
		#	Add the 'parsimonious' states at the current node to the current set of @set_node2parsstat.
		#     }
		#		END of "(REVISED to Fix the BUG on 2020/03/21.)"

=For the predecessor, 'enum_node2parsstat_sankoff_td'. (Beg)
	    unless (defined $parent) { # The Big unless #
		# The top node --> Initialize the @set_node2parsstat #
=cut # END of "For the predecessor, 'enum_node2parsstat_sankoff_td'. (Beg)"

		my $min_penalty;
		my @pars_chars = ();
		for (my $j=0; $j<$ct_chars; $j++) { # The 1st middle for-loop #

		    my $cond_min_pen = $cond_min_penalties->[$j];

		    (defined $cond_min_pen) or next;

		    unless (defined $min_penalty) {
			$min_penalty = $cond_min_pen;
			@pars_chars = ($j);
			next;
		    }
		    ($min_penalty < $cond_min_pen) and next;

		    if ($min_penalty > $cond_min_pen) {
			$min_penalty = $cond_min_pen;
			@pars_chars = ($j);
		    } else {
			push @pars_chars, $j;
		    }
		}		# End of the 1st middle for-loop #

		if (!(defined $parent)) { # ADDED the condition to Fix a BUG (on 2020/03/21). (BEG) #

			# Initialize the @set_node2parsstat. #

		    foreach my $char (@pars_chars) { # THe 2nd middle for-loop #
		        my %node2parsstat = ($node => [$char, $min_penalty]);
		        push @set_node2parsstat, \%node2parsstat;
		    }		# End of the 2nd middle for-loop #

		    if ($fl_full_mode == 1) {
		        next;
		    } else {
		        return \@set_node2parsstat; # Brief version #
		    }

	 	} else { # ADDED the condition to Fix a BUG (on 2020/03/21). (MID) #

			# ADD the "parsimonious" states to the existing %{$node2parsstat}. #
			# (ADDED to Fix a BUG (on 2020/03/21).) #

		    my @new_set_node2parsstat = ();

		    my $counter = 0;
		    foreach my $node2parsstat (@set_node2parsstat) { # The 3rd middle foreach-loop. 

			foreach my $char (@pars_chars) { # Inner foreach-loop. #

			  my %new_node2parsstat;

			  if ($counter>0) {
			    %new_node2parsstat = ($node => [$char, $min_penalty]);

			    foreach my $node2 (keys %{$node2parsstat}) {  # Innermost foreach-loop. #
				my @parsstat = @{$node2parsstat->{$node2}};
				$new_node2parsstat{$node2} = \@parsstat;
			    } # End of the innermost foreach-loop. #

			  } else {
			     %new_node2parsstat = %{$node2parsstat};
			     $new_node2parsstat{$node} = [$char, $min_penalty];
			  }

			  push @new_set_node2parsstat, \%new_node2parsstat;
			  $counter++;

			} # End of the inner foreach-loop. #
		    } # End of the 3rd middle foreach-loop. #

		    @set_node2parsstat = @new_set_node2parsstat;

		    next;

			# End of "(ADDED to Fix a BUG (on 2020/03/21).)" #

		} # ADDED the condition to Fix a BUG (on 2020/03/21). (END) #

=For the predecessor, 'enum_node2parsstat_sankoff_td'. (End)
	    } 			# End of the Big unless #
=cut # End of "For the predecessor, 'enum_node2parsstat_sankoff_td'. (End)"

	    } 			# End of the Big if #

	    my @new_set_node2parsstat = ();

		# Let the parent node determine the state of the node #

	    foreach my $node2parsstat (@set_node2parsstat) { # Outer for-loop #

		my $parsstat_pa = $node2parsstat->{$parent};
		unless (defined $parsstat_pa) {
		    print STDERR "[ERROR 'enum_node2parsstat_sankoff_xgap_td'] The parent node '$parent' of the node '$node' has NO INFORMATION on the PARSIMONIOUS STATE: $! !\n";
		    return undef;
		}
		my $pars_char_pa = $parsstat_pa->[0];
		my $pen_row = $pen_mtrx->[$pars_char_pa];
		my $min_penalty;
		my @pars_chars = ();
		for (my $j=0; $j<$ct_chars; $j++) { # The 3rd middle for-loop #

		    my $cond_min_pen = $cond_min_penalties->[$j];

		    (defined $cond_min_pen) or next;

		    $cond_min_pen += $pen_row->[$j];

		    unless (defined $min_penalty) {
			$min_penalty = $cond_min_pen;
			@pars_chars = ($j);
			next;
		    }
		    ($min_penalty < $cond_min_pen) and next;

		    if ($min_penalty > $cond_min_pen) {
			$min_penalty = $cond_min_pen;
			@pars_chars = ($j);
		    } else {
			push @pars_chars, $j;
		    }
		}			# End of the 3rd middle for-loop #

		my $counter = 0;
		foreach my $char (@pars_chars) { # The 4th middle for-loop #

			## Corrected a Potentially FATAL BUG on Oct 13 (Wed), 2010 ##
		    my %new_node2parsstat;
		    if ($counter>0) {
			foreach my $node (keys %{$node2parsstat}) {
			    my @parsstat = @{$node2parsstat->{$node}};
			    $new_node2parsstat{$node} = \@parsstat;
			}
		    } else {
			%new_node2parsstat = %{$node2parsstat};
		    }
		    ##my %new_node2parsstat = %{$node2parsstat};

			# A BUG fixed on June 16 (Wed), 2010 #
		    my $cond_min_pen = $cond_min_penalties->[$char];
		    $new_node2parsstat{$node} = [$char, $cond_min_pen];
		    ##$new_node2parsstat{$node} = [$char, $min_penalty];

		    push @new_set_node2parsstat, \%new_node2parsstat;
		    $counter++;
		}		# End of the 4th middle for-loop #

	    }			# End of the outer for-loop (over %node2parsstat) #

		# Update the set of parsimonious states #

	    @set_node2parsstat = @new_set_node2parsstat;

	}		# End of the main outermost foreach-loop (over all nodes) #

	return \@set_node2parsstat;

}	# End of the subroutine 'enum_node2parsstat_sankoff_xgap_td' #


# map_parsim_states_sankoff_xgap_clm:
#	This satellite subroutine of 'map_parsim_states_sankoff_xgap_1clm'
#	performs Sankoff's parsimony algorithm on a column of MSA,
#	while dealing with gaps via Farris' Dollo parsimony principle,
#	and find (a) map(s) of states onto nodes that are the most parsimonious
#	under the given penalty matrix.
#	If the subroutine encounters a node with an "absence" state,
#	it regards the node as "missing information" and EXCLUDES it from Sankoff's parsimony analysis.
#	The subroutine was created by modifying the predecessor, 'map_parsim_states_sankoff_clm'.
#	[CREATION STARTED on 2020/03/13, ENDED on 2020/03/13.]
#
# map_parsim_states_sankoff_clm:
#	This satellite subroutine of 'map_parsim_states_sankoff_msa'
#	performs Sankoff's parsimony algorithm on a column of MSA,
#	and find (a) map(s) of states onto nodes that are the most parsimonious
#	under the given penalty matrix.
#
# [ INPUT: (\%node2children, \%node2parent, \%node2depth, \%exnode2charindx, \@pen_mtrx, $fl_full_mode), where ] ... for the predecessor, 'map_parsim_states_sankoff_clm'.
#
# INPUT: (\%node2children, \%node2parent, \%node2depth, \%exnode2charindx, \@pen_mtrx, $char_abs, \%node2dparsstat, $fl_full_mode), where
#
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2parent = ($node_id => $parent_id,...) for the non-top nodes;
#	%node2depth = ($node_id => $depth, ...) for ALL nodes, with
#	$depth defined as the number of nodes from the top node (inclusive) to themselves (inclusive);
#	%exnode2charindx = ($node_id => $indx of the character at the node, ...)
#	for the exterior nodes. The $indx conforms to @pen_mtrx;
#	$pen_mtrx[$i]->[$j] is the penalty imposed on the change
#	from the $i th character to the $j th character (, BOTH states BELONG TO the "presence" state);
#
#		# (ADDED in creation of 'map_parsim_states_sankoff_xgap_clm' on 2020/03/13.)
#	$char_abs is the character representing the "absence" state (i.e., gap);
#
#	%node2dparsstat = ($node_id => \@state_info, ...) for the 
#	Dollo parsimonious scenario of the gap configuration of the clumn mapped onto the nodes,
#	with
#	@state_info = ($stat, $ct_pres_ext), where
#	$stat (= $char_abs/pres) is the gap-status on the node,
#	$ct_pres_ext = #{child nodes with #{exterior descendants with state ne $char_abs}>0} + #{parent with state eq $char_pres};
#		# End of "(ADDED in creation of 'map_parsim_states_sankoff_xgap_clm' on 2020/03/13.)"
#
#	$fl_full_mode = 1 if you want the full mode,
#			= 0 if the brief mode is enough.
#
#
# OUTPUT: \@set_node2parsstat, where
#
#	%{$set_node2parsstat[$opt]} = ($node_id => \@state_info, ...) for the $opt th
#	option of the parsimonious states of the site (or column) mapped onto the nodes,
#	with
#	[ @state_info = ($char, $sum_penal), where 
#	$char is the character on the node, ] ... WRONG
#	@state_info = ($indx_char, $sum_penal), where
#	$indx_char is the index of the character on the node (but = $char_abs if the node has the "absence" state), # CORRECTED on 2020/03/13.
#
#	[ $sum_penal is the minimum summation of the penalties across the subtree under the node. ] ... for the predecessor, 'map_parsim_states_sankoff_clm'.
#	$sum_penal is the minimum summation of the penalties across the subtree under the node,
#	BUT is undef if ($char eq $char_abs). # MODIFIED in creation of 'map_parsim_states_sankoff_xgap_clm' (on 2020/03/13).
#
#	If $fl_full_mode = 1, the information on ALL the nodes is given.
#	If $fl_full_mode = 0, ONLY the information on the TOP node  is given.
#
# [ sub map_parsim_states_sankoff_clm (\%\%\%\%\@$) {
#	my ($node2ch, $node2pa, $node2depth, $exnode2charindx, $pen_mtrx, $fl_full_mode) = @_; ] ... for the predecessor, 'map_parsim_states_sankoff_clm'.
#
sub map_parsim_states_sankoff_xgap_clm (\%\%\%\%\@$\%$) {

	my ($node2ch, $node2pa, $node2depth, $exnode2charindx, $pen_mtrx, $char_abs, $node2dparsstat, $fl_full_mode) = @_;


	# my $set_node2parsstat =  map_parsim_states_sankoff_xgap_clm (%{$node2ch}, %{$node2pa}, %{$node2depth}, %exnode2charindx, @{$pen_mtrx}, $char_abs, %{$node2dparsstat}, $fl_full_mode) ; # (MODIFIED in creation of 'map_parsim_states_sankoff_xgap_1clm' on 2020/03/13). (2) #



##=for DEBUGGING the subroutine
   (print STDOUT "\n< Subroutine 'map_parsim_states_sankoff_xgap_clm' >\n\n") if ($TEST_LEVEL_mpssxc>0);

   if ($TEST_LEVEL_mpssxc>2) { 

	print STDOUT '%{$exnode2charindx} = (', "\n";
	foreach my $node (sort {$a <=> $b} keys %{$exnode2charindx}) {
	    print STDOUT $node, '  =>  ', $exnode2charindx->{$node}, ",\n";
	}
	print STDOUT ")\n\n";
	print STDOUT '@{$pen_mtrx} = (', "\n";
	foreach my $row (@{$pen_mtrx}) { print STDOUT "\t\[", join (', ', @{$row}), "\]\,\n"; }
	print STDOUT "\t)\n\n";

   } # End of "if ($TEST_LEVEL_mpssxc>2) { ...}"
##=cut

		## Assign conditional minimum penalties to the nodes
		## in a bottom-up manner

	# [ %{$node2min_penalties} = ($node_id => \@min_penals, ...) for all nodes, where ] ... for the predecessor, 'map_parsim_states_sankoff_clm'.
	#
	# %{$node2min_penalties} = ($node_id => \@min_penals, ...) for all nodes with "presence" states, where
	# $min_penals[$i] is the minimum summed penalty
	# imposed on the subtree under the node in question
	# under the condition that the node has the $i th character.
	#
	## my $node2min_penalties = min_penalties_sankoff_bo (%{$node2ch}, %{$node2depth}, %{$exnode2charindx}, @{$pen_mtrx});
	my $node2min_penalties = min_penalties_sankoff_xgap_bo (%{$node2ch}, %{$node2depth}, %{$exnode2charindx}, @{$pen_mtrx}, $char_abs, %{$node2dparsstat}); # MODIFIED in creation of 'map_parsim_states_sankoff_xgap_clm' (on 2020/03/13). (1)
	unless (defined $node2min_penalties) {
	    print STDERR "[ERROR 'map_parsim_states_sankoff_xgap_clm'] The subroutine 'min_penalties_sankoff_xgap_bo' FAILED: $! !\n";
	    return undef;
	}

##=for DEBUGGING the subroutine
   if ($TEST_LEVEL_mpssxc>2) { 

	print STDOUT "\n{ Output of 'min_penalties_sankoff_bo' }\n\n"; 
	print STDOUT '%{$node2min_penalties} = ( # Node_ID => [minimin, penalties, conditional, on, the, state]', "\n";
	foreach my $node (sort {$a <=> $b} keys %{$node2min_penalties}) {
	    my $min_penalties = $node2min_penalties->{$node};
	    my @out_min_penalties = ();
	    for (my $c=0; $c <@{$pen_mtrx}; $c++) {
		my $min_pen = $min_penalties->[$c];
		unless (defined $min_pen) { $min_pen = 'undef'; }
		$out_min_penalties[$c] = $min_pen;
	    }
	    ##print STDOUT "\t", $node, '  =>   [', join (', ', @{$node2min_penalties->{$node}}), '],', "\n";
	    print STDOUT "\t", $node, '  =>   [', join (', ', @out_min_penalties) , '],', "\n";
	}
	print STDOUT "\t)\n\n";
   } # End of "if ($TEST_LEVEL_mpssxc>2) { ...}"
##=cut

		## Enumerate the maps from the nodes to the parsimonious states
		## in a top-down manner

	##my $set_node2parsstat = enum_node2parsstat_sankoff_td (%{$node2pa}, %{$node2depth}, %{$node2min_penalties}, @{$pen_mtrx}, $fl_full_mode);
	my $set_node2parsstat = enum_node2parsstat_sankoff_xgap_td (%{$node2pa}, %{$node2depth}, %{$node2min_penalties}, @{$pen_mtrx}, $char_abs, %{$node2dparsstat}, $fl_full_mode); # MODIFIED in creation of 'map_parsim_states_sankoff_xgap_clm' (on 2020/03/13). (2)
	unless (defined $set_node2parsstat) {
	    print STDERR "[ERROR 'map_parsim_states_sankoff_xgap_clm'] The subroutine 'enum_node2parsstat_sankoff_xgap_td' FAILED: $! !\n";
	    return undef;
	}
##=for DEBUGGING the subroutine
   if ($TEST_LEVEL_mpssxc>2) { 

	my $ct_sets = @{$set_node2parsstat};
	print STDOUT "\n{ Output of 'enum_node2parsstat_sankoff_td' ($ct_sets maps of states} }\n\n";
	for (my $m = 0; $m <$ct_sets; $m++) {
	    my $node2parsstat = $set_node2parsstat->[$m];
	    print STDOUT "( The $m th map )\n\n";
	    print STDOUT join ("\t", '#Node_ID', 'State', 'sum_penal'), "\n\n";
	    foreach my $node (sort {$a <=> $b} keys %{$node2parsstat}) {
		print STDOUT join ("\t", $node, @{$node2parsstat->{$node}}), "\n";
	    }
	    print STDOUT "\n\n";
	}
   } # End of "if ($TEST_LEVEL_mpssxc>2) { ...}"
##=cut

   (print STDOUT  "< Subroutine 'map_parsim_states_sankoff_xgap_cml' ENDED />\n\n") if ($TEST_LEVEL_mpssxc>0);
=for DEBUGGING the subroutine
	print STDERR " --- 'map_parsim_states_sankoff_cml' ENDED ---\n\n";
=cut

	return $set_node2parsstat;

}	# End of the subroutine 'map_parsim_states_sankoff_xgap_clm' # 


=Set Aside for Later Use. (1)

# map_parsim_states_sankoff_msa:
#	This subroutine performs Sankoff's parsimony algorithm on a MSA,
#	and serially find a set of parsimonious maps of states onto nodes for
#	each of the columns in the input MSA.
#
#	NOTE that this subroutine eats up the input MSA.
#
# INPUT: ($topnode_id, \%node2children, \%node2parent, \%name2id, \@seqnames, \@msa, \@charlist, \@pen_mtrx, $fl_full_mode), where
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2parent = ($node_id => $parent_id, ...) for the non-top nodes;
#	%name2id = ($seqname => $node_id, ...) for the exterior nodes and the top node;
#	$seqnames[$k] is the name of the $k th sequence in the input MSA;
#	${$msa[$k]} is the $k th sequence embedded in the input MSA;
#	@charlist lists the possible characters;
#	$pen_mtrx[$i]->[$j] gives the penalty for the change from $charlist[$i] to $charlist[$j];
#	$fl_full_mode = 1 if you want the full mode,
#			= 0 if the brief mode is enough.
#
# OUTPUT: \@setset_node2parsstat, where
#	%{$setset_node2parsstat[$l]->[$opt]} = ($node_id => \@state_info, ...) for the $opt th
#	option of the parsimonious states of the $l th site mapped onto the nodes,
#	with
#	@state_info = ($char, $sum_penal), where
#	$char is the character on the node,
#	$sum_penal is the minimum summation of the penalties across the subtree under the node.
#	If $fl_full_mode = 1, the information on ALL the nodes is given.
#	If $fl_full_mode = 0, ONLY the information on the TOP node  is given.
#
sub map_parsim_states_sankoff_msa ($\%\%\%\@\@\@\@$) {

	my ($topnode_id, $node2ch, $node2pa, $name2id, $seqnames, $msa, $charlist, $pen_mtrx, $fl_full_mode) = @_;

	my $ct_seqs = @{$seqnames};
	unless ($ct_seqs>0) {
	    print STDERR "[ERROR 'map_parsim_states_sankoff_msa'] NO SEQUENCE in the input MSA: $! !\n";
	    return undef;
	}
	unless ($ct_seqs == @{$msa}) {
	    print STDERR "[ERROR 'map_parsim_states_sankoff_msa'] The number of sequences (", scalar (@{$msa}), ") in the MSA CLASHES with the number of sequence names ($ct_seqs): $! !\n";
	    return undef;
	}
	my $ct_clms = length ${$msa->[0]};
	($ct_clms>0) or (return []);

	my $ct_chars = @{$charlist};
	unless ($ct_chars == @{$pen_mtrx}) {
	    print STDERR "[ERROR 'map_parsim_states_sankoff_msa'] The number of rows in the penalty matrix (", scalar (@{$pen_mtrx}), ") CLASHES with the number of characters ($ct_chars): $! !\n";
	    return undef;
	}

		# Prepare auxiliary arrays/hashes #

	# %{$node2depth} = ($node_id => #{nodes from the top node to itself}, ...);
	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
	unless (defined $node2depth) {
	    print STDERR "[ERROR 'map_parsim_states_sankoff_msa'] The subroutine 'make_node2abs_depth' FAILED : $! !\n";
	    return undef;
	}

	my %char2indx; # ($character => $index in @{$charlist}, ...)
	for (my $i = 0; $i<$ct_chars; $i++) { $char2indx{$charlist->[$i]} = $i; }

	my @node_ids; # $node_ids[$k] is the exterior node ID corresponding to the $k th sequence in @{$msa}
	for (my $k=0; $k<$ct_seqs; $k++) { # The 2nd preliminary for-loop #
	    my $name = $seqnames->[$k];
	    my $node_id = $name2id->{$name};
	    unless (defined $node_id) {
		print STDERR "[ERROR 'map_parsim_states_sankoff_msa'] NODE ID is UNDEFINED for the $k th sequence (name: $name): $! !\n";
		return undef;
	    }
	    $node_ids[$k] = $node_id;
	}			# End of the 2nd preliminary for-loop #

		## Main Process ##

	my @setset_node2parsstat = ();

	for (my $l=0; $l<$ct_clms; $l++) { # The main outer for-loop #

	    my %exnode2charindx; # ($node_id => $index of the character, ...) for the exterior nodes
	    my @char_in_clm = (); ## For a debugging purpose ##
	    for (my $k=0; $k<$ct_seqs; $k++) { # The 1st inner for-loop #

		my $node_id = $node_ids[$k];

		my $char = substr (${$msa->[$k]}, 0, 1, '');
		my $indx = $char2indx{$char};
		unless (defined $indx) {
		    print STDERR "[ERROR 'map_parsim_states_sankoff_msa'] INDEX is UNDEFINED for the character '$char' in the $l th site of the $k th sequence: $! !\n";
		    return undef;
		}
		$exnode2charindx{$node_id} = $indx;
		push @char_in_clm, $char; ## For debugging ##
	    }				# End of the 1st inner for-loop #

	    my $set_node2parsstat =  map_parsim_states_sankoff_clm (%{$node2ch}, %{$node2pa}, %{$node2depth}, %exnode2charindx, @{$pen_mtrx}, $fl_full_mode) ;
	    unless (defined $set_node2parsstat) {
		print STDERR "[ERROR 'map_parsim_states_sankoff_msa'] 'map_marsim_states_sankoff_msa' FAILED on the $l th site: $! !\n";
		return undef;
	    }

		## Translate the character index back to the character ##

	    foreach my $node2parsstat (@{$set_node2parsstat}) { # The 2nd inner foreach-loop #
		foreach my $nodeid (keys %{$node2parsstat}) { # The inner-most foreach-loop #
		    my $parsstat = $node2parsstat->{$nodeid};
		    my $charindx = $parsstat->[0];
		    my $char = $charlist->[$charindx];
		    unless (defined $char) {
			print STDERR "[ERROR 'map_parsim_states_sankoff_msa'] NO CHARACTER CORRESPONDING to the character index '$charindx' for the node ID $nodeid at the $l th column: $! !\n";
			return undef;
		    }
		    $parsstat->[0] = $char;
		}			# End of the inner-most foreach-loop #
	    }			# End of the 2nd inner foreach-loop #

	    push @setset_node2parsstat, $set_node2parsstat;

	}			# End of the main outer for-loop #

	return \@setset_node2parsstat;

}	# End of the subroutine 'map_parsim_states_sankoff_msa' #

=cut # END of "Set Aside for Later Use. (1)"


# map_parsim_states_sankoff_xgap_1clm:
#	This subroutine deals with a gap (or an "absence" state) 
#	under the Dollo parsimony principle,
#	and if it encounters a node with an "absence" state,
#	it regards the node as "missing information" and EXCLUDES it from Sankoff's parsimony analysis.
#	[CREATION STATED on 2020/03/13, ENDED on 2020/03/13.]
#
#	[DEBUGGED on 2020/03/22]:
#	REMOVED the UNNECESSARY CONDITION on the length of the character.
#
# map_parsim_states_sankoff_1clm:
#	This variant of 'map_parsim_states_sankoff_msa' specializes in
#	handling just a single column in a MSA,
#	namely comparing just a single trait across taxa.
#
# map_parsim_states_sankoff_msa:
#	This subroutine performs Sankoff's parsimony algorithm on a MSA,
#	and serially find a set of parsimonious maps of states onto nodes for
#	each of the columns in the input MSA.
#
#	NOTE that this subroutine eats up the input MSA.
#
# [ INPUT: ($topnode_id, \%node2children, \%node2parent, \%name2id, \@seqnames, \@clm, \@charlist, \@pen_mtrx, $fl_full_mode), where ] ... for the predecessor, 'map_parsim_states_sankoff_1clm'.
#
# INPUT: ($topnode_id, \%node2children, \%node2parent, \%name2id, \@seqnames, \@clm, $char_abs, $char_pres, \@charlist, \@pen_mtrx, $fl_full_mode), where
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2parent = ($node_id => $parent_id, ...) for the non-top nodes;
#	%name2id = ($seqname => $node_id, ...) for the exterior nodes and the top node;
#	$seqnames[$k] is the name of the $k th sequence in the input MSA;
#	$clm[$k] is the character of the $k th sequence (or taxon);
#
#	$char_abs/$char_pres is a character representing the "absence"/"presence" state; # ADDED in creation of 'map_parsim_states_sankoff_xgap_1clm'.
#
#	[ @charlist lists the possible characters; ] ... for the predecessor , 'map_parsim_states_sankoff_1clm'. 
#	@charlist lists the possible characters IN THE "PRESENCE" STATE; # MODIFIED in creation of 'map_parsim_states_sankoff_xgap_1clm'.
#
#	$pen_mtrx[$i]->[$j] gives the penalty for the change from $charlist[$i] to $charlist[$j];
#	$fl_full_mode = 1 if you want the full mode,
#			= 0 if the brief mode is enough.
#
# OUTPUT: \@set_node2parsstat, where
#
#	%{$set_node2parsstat[$opt]} = ($node_id => \@state_info, ...) for the $opt th
#	option of the parsimonious states of the clumn mapped onto the nodes,
#	with
#	@state_info = ($char, $sum_penal), where
#	$char is the character on the node,
#
#	[ $sum_penal is the minimum summation of the penalties across the subtree under the node. ] ... for the predecessor, 'map_parsim_states_sankoff_1clm'.
#	$sum_penal is the minimum summation of the penalties across the subtree under the node,
#	BUT is undef if $char = $char_abs. # MODIFIED in creation of 'map_parsim_states_sankoff_xgap_1clm'.
#
#	If $fl_full_mode = 1, the information on ALL the nodes is given.
#	If $fl_full_mode = 0, ONLY the information on the TOP node  is given.
#
# [ sub map_parsim_states_sankoff_1clm ($\%\%\%\@\@\@\@$) {
#	my ($topnode_id, $node2ch, $node2pa, $name2id, $seqnames, $clm, $charlist, $pen_mtrx, $fl_full_mode) = @_; ] ... for the predecessor, 'map_parsim_states_sankoff_1clm'.
#
sub map_parsim_states_sankoff_xgap_1clm ($\%\%\%\@\@$$\@\@$) {

	my ($topnode_id, $node2ch, $node2pa, $name2id, $seqnames, $clm, $char_abs, $char_pres, $charlist, $pen_mtrx, $fl_full_mode) = @_;

	my $ct_seqs = @{$seqnames};
	unless ($ct_seqs>0) {
	    print STDERR "[ERROR 'map_parsim_states_sankoff_xgap_1clm'] NO SEQUENCE in the input MSA: $! !\n";
	    return undef;
	}
	unless ($ct_seqs == @{$clm}) {
	    print STDERR "[ERROR 'map_parsim_states_sankoff_xgap_1clm'] The number of sequences (", scalar (@{$clm}), ") in the column CLASHES with the number of sequence names ($ct_seqs): $! !\n";
	    return undef;
	}

=DISABLED in DEBUGGING because they are NOT NECESSARY (on 2020/03/22).
	my $len_char  = length $clm->[0];
	unless ($len_char>0) {
	    print STDERR "[ERROR 'map_parsim_states_sankoff_xgap_1clm'] We do NOT HANDLE NULL CHARACTERS: $! !\n";
	    return undef;
	}

		# (MODIFIED in creation of 'map_parsim_states_sankoff_xgap_1clm' on 2020/03/13). (1) #
	unless ($len_char <2) {
	    print STDERR "[WARNING 'map_parsim_states_sankoff_xgap_1clm'] ILLEGAL INPUT: The character seems be of length $len_char, instead of one: $! !\n";
	    return undef;
	}
##=For the predecessor, 'map_parsim_states_sankoff_1clm'.
##	if ($len_char>1) {
##	    print STDERR "[WARNING 'map_parsim_states_sankoff_1clm'] The character seems be of length $len_char, instead of one, but I will go on anyway ...\n\n";
##	}
##=cut
		# End of "(MODIFIED in creation of 'map_parsim_states_sankoff_xgap_1clm' on 2020/03/13). (1)" #
=cut # END of "DISABLED in DEBUGGING because they are NOT NECESSARY (on 2020/03/22)."

	my $ct_chars = @{$charlist};
	unless ($ct_chars == @{$pen_mtrx}) {
	    print STDERR "[ERROR 'map_parsim_states_sankoff_xgap_1clm'] The number of rows in the penalty matrix (", scalar (@{$pen_mtrx}), ") CLASHES with the number of characters ($ct_chars): $! !\n";
	    return undef;
	}

		# Prepare auxiliary arrays/hashes #

	# %{$node2depth} = ($node_id => #{nodes from the top node to itself}, ...);
	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
	unless (defined $node2depth) {
	    print STDERR "[ERROR 'map_parsim_states_sankoff_xgap_1clm'] The subroutine 'make_node2abs_depth' FAILED : $! !\n";
	    return undef;
	}

	my %char2indx; # ($character => $index in @{$charlist}, ...)
	for (my $i = 0; $i<$ct_chars; $i++) { $char2indx{$charlist->[$i]} = $i; }

		# (ADDED in creation of 'map_parsim_states_sankoff_xgap_1clm' on 2020/03/13). (1) #
	$char2indx{$char_abs} = $char_abs; # Assign the "index", $char_abs, to a gap (i.e., an "absence" state).
		# End of "(ADDED in creation of 'map_parsim_states_sankoff_xgap_1clm' on 2020/03/13). (1)" #

	my @node_ids; # $node_ids[$k] is the exterior node ID corresponding to the $k th sequence in @{$msa}
	for (my $k=0; $k<$ct_seqs; $k++) { # The 2nd preliminary for-loop #
	    my $name = $seqnames->[$k];
	    my $node_id = $name2id->{$name};
	    unless (defined $node_id) {
		print STDERR "[ERROR 'map_parsim_states_sankoff_xgap_1clm'] NODE ID is UNDEFINED for the $k th sequence (name: $name): $! !\n";
		return undef;
	    }
	    $node_ids[$k] = $node_id;
	}			# End of the 2nd preliminary for-loop #


		## Sub-Main Process: Perform the FULL Dollo-parsimony analysis on the gap configuration of the column. ##
		# (ADDED in creation of 'map_parsim_states_sankoff_xgap_1clm' on 2020/03/13). (2) #

	my ($node2dparsstat, $br2change_dpars) 
	   = map_parsim_states_dollo_1clm ($topnode_id, %{$node2ch}, %{$node2pa}, %{$name2id}, @{$seqnames}, @{$clm}, $char_abs, $char_pres, 1);
	unless (defined $br2change_dpars) {
	    print STDERR "[ERROR 'map_parsim_states_sankoff_xgap_1clm'] The subroutine 'map_parsim_states_dollo_1clm' FAILED : $! !\n";
	    return undef;
	}

# INPUT: ($topnode_id, \%node2children, \%node2parent, \%name2id, \@seqnames, \@clm, $char_abs, $char_pres, $fl_full_mode), where

		# END of "(ADDED in creation of 'map_parsim_states_sankoff_xgap_1clm' on 2020/03/13). (2)" #


		## Main Process ##

	my %exnode2charindx; # ($node_id => $index of the character, ...) for the exterior nodes
	for (my $k=0; $k<$ct_seqs; $k++) { # The 1st main for-loop #

	    my $node_id = $node_ids[$k];

	    my $char = $clm->[$k];


	    my $indx = $char2indx{$char};
	    unless (defined $indx) {
	        print STDERR "[ERROR 'map_parsim_states_sankoff_xgap_1clm'] INDEX is UNDEFINED for the character '$char' of the $k th sequence: $! !\n";
	        return undef;
	    }
	    $exnode2charindx{$node_id} = $indx;
	}				# End of the 1st main for-loop #

	## my $set_node2parsstat =  map_parsim_states_sankoff_clm (%{$node2ch}, %{$node2pa}, %{$node2depth}, %exnode2charindx, @{$pen_mtrx}, $fl_full_mode) ;
	my $set_node2parsstat =  map_parsim_states_sankoff_xgap_clm (%{$node2ch}, %{$node2pa}, %{$node2depth}, %exnode2charindx, @{$pen_mtrx}, $char_abs, %{$node2dparsstat}, $fl_full_mode) ; # (MODIFIED in creation of 'map_parsim_states_sankoff_xgap_1clm' on 2020/03/13). (2) #
	unless (defined $set_node2parsstat) {
	    print STDERR "[ERROR 'map_parsim_states_sankoff_xgap_1clm'] 'map_marsim_states_sankoff_xgap_clm' FAILED: $! !\n";
	    return undef;
	}


		## Translate the character index back to the character ##

	foreach my $node2parsstat (@{$set_node2parsstat}) { # The 2nd main foreach-loop #
	    foreach my $nodeid (keys %{$node2parsstat}) { # The inner foreach-loop #
		my $parsstat = $node2parsstat->{$nodeid};
		my $charindx = $parsstat->[0];

		##my $char = $charlist->[$charindx];
		my $char = ($charindx eq $char_abs) ? $char_abs : $charlist->[$charindx]; # (MODIFIED in creation of 'map_parsim_states_sankoff_xgap_1clm' on 2020/03/13). (3) #

		unless (defined $char) {
		    print STDERR "[ERROR 'map_parsim_states_sankoff_xgap_1clm] NO CHARACTER CORRESPONDING to the character index '$charindx' for the node ID $nodeid: $! !\n";
		    return undef;
		}
		$parsstat->[0] = $char;
	    }			# End of the inner foreach-loop #
	}			# End of the 2nd main foreach-loop #

	return $set_node2parsstat;

}	# End of the subroutine 'map_parsim_states_sankoff_xgap_1clm' #


=Set Aside for Later Use. (2)

# map_parsim_gapstates_sankoff:
#	This subroutine maps parsimonious gap states on the tree
#	by first using Sankoff's parsimony on a set of dimorphic character states (present/absent),
#	then by excluding those character states containing more than one insertion, if you want.
#	We could regard @chars_in_clm as gap information on a block
#	(or a set of columns) of the same gap pattern, if desired.
#
#	NOTE that this could end up in NO parsimonious states.
#	In that case, it would be better to apply Farris's (1977) Dollo parsimony method.
#
#	REVISED on June 16 (Thu), 2010:
#	The subroutine now uses 'map_parsim_states_sankoff_1clm'
#	instead of 'map_parsim_states_sankoff_msa'.
#
# INPUT: ($topnode_id, \%node2children, \%node2parent, \%name2id, \@seqnames, \@chars_in_clm, $gap, $ngap, $fl_full_mode, $fl_rmv_dblins), where
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2parent = ($node_id => $parent_id, ...) for the non-top nodes;
#	%name2id = ($seqname => $node_id, ...) for the exterior nodes and the top node;
#	$seqnames[$k] is the name of the $k th sequence in the input MSA;
#	$chars_in_clm[$k] is the character that the $k th sequence has in the column in question;
#	$gap and $ngap are the codes to specify a gap and a non-gap residue, respectively.
#	NOTE that $gap must be used in @chars_in_clm as well;
#	$fl_full_mode = 1 if you want the full mode,
#			= 0 if the brief mode is enough;
#	$fl_rmv_dblins = 1 if you want to remove those states containing more than one insertion,
#	               = 0 if you don't mind retaining sub states.
#	NOTE that $fl_rmv_dblins = 1 takes effect only if $fl_full_mode = 1.
#
# OUTPUT: \@set_node2pars_gapstat, where
#	%{$set_node2pars_gapstat[$opt]} = ($node_id => [$gap/$ngap, $sbct_indels], ...)
#	 for the $opt th option of the parsimonious gap states mapped onto the nodes,
#	with $sbct_indels is the number of insertions/deletions in the subtree
#	under $node_id.
#	If $fl_full_mode = 1, the information on ALL the nodes is given.
#	If $fl_full_mode = 0, ONLY the information on the TOP node  is given.
#
sub map_parsim_gapstates_sankoff ($\%\%\%\@\@$$$$) {

	my ($topnode_id, $node2ch, $node2pa, $name2id, $seqnames,
	$chars_in_clm, $gap, $ngap, $fl_full_mode, $fl_rmv_dblins) = @_;

	my $ct_seqs = @{$seqnames};
	unless (@{$chars_in_clm}== $ct_seqs) {
	    print STDERR "[ERROR 'map_parsim_gapstates_sankoff'] ILLEGAL INPUT: THe number of characters in the column (", scalar (@{$chars_in_clm}), ") CLASHES with that of sequences ($ct_seqs): $! !\n";
	    return undef;
	}
		# Prepare additional input data #

	my @rdcd_clm = (); # REVISED on June 17 (Thu), 2010 #
	##my @rdcd_msa = ();
	foreach my $char (@{$chars_in_clm}) { # 1st preliminary foreach-loop #
	    my $code = ($char eq $gap) ? $gap : $ngap;

		# REVISED on June 17 (Thu), 2010 #
	    push @rdcd_clm, $code;
	    ##push @rdcd_msa, \$code;
	}			# End of the 1st preliminary foreach-loop #
	my @charlist = ($ngap, $gap);
	my @gap_pen = (
		[0, 1],
		[1, 0]
		);


		## REVISED on June 17 (Thu), 2010 ##
	my $set_node2pars_gapstat = map_parsim_states_sankoff_1clm ( $topnode_id, %{$node2ch}, %{$node2pa}, %{$name2id}, @{$seqnames}, @rdcd_clm, @charlist, @gap_pen, $fl_full_mode);
	unless (defined $set_node2pars_gapstat) {
	    print STDERR "[ERROR 'map_parsim_gapstates_sankoff'] 'map_parsim_states_sankoff_1clm' FAILED on the reduced MSA: $! !\n";
	    return undef;
	}
	##my $setset_node2pars_gapstat = map_parsim_states_sankoff_msa ( $topnode_id, %{$node2ch}, %{$node2pa}, %{$name2id}, @{$seqnames}, @rdcd_msa, @charlist, @gap_pen, $fl_full_mode);
	##unless (defined $setset_node2pars_gapstat) {
	##    print STDERR "[ERROR 'map_parsim_gapstates_sankoff'] 'map_parsim_states_sankoff_msa' FAILED on the reduced MSA: $! !\n";
	##    return undef;
	##}
	##unless (@{$setset_node2pars_gapstat}==1) {
	##    print STDERR "[ERROR 'map_parsim_gapstates_sankoff'] Somehow, 'map_parsim_states_sankoff_msa' has output parsimonious states on ", scalar (@{$setset_node2pars_gapstat}), " sites, instead of those on a single site: $! !\n";
	##    return undef;
	##}
	##my $set_node2pars_gapstat = $setset_node2pars_gapstat->[0];

		# If you want, return the raw output #

		# Revised on June 17 (Thu), 2010 # 
	($fl_full_mode and $fl_rmv_dblins) or (return $set_node2pars_gapstat);
	##($fl_rmv_dblins) or (return $set_node2pars_gapstat);

		# If you want,
		# remove states containing more than one insertion.

	my @rfnd_set_node2pars_gapstat = ();
	foreach my $node2gapstat (@{$set_node2pars_gapstat}) { # Finishing foreach-loop #
	    my $ct_ins = 0;
	    foreach my $node (keys %{$node2pa}) { # Inner foreach-loop #
		my $parent = $node2pa->{$node};
		(defined $parent) or next;

		    # An insertion is defined as:
		    # $gap at a parent -> $ngap at a child

		my $gapstat_ch = $node2gapstat->{$node};
		($gapstat_ch->[0] eq $ngap) or next;

		my $gapstat_pa = $node2gapstat->{$parent};
		($gapstat_pa->[0] eq $gap) and ($ct_ins++);
	    }				# End of the inner foreach-loop #
	    ($ct_ins>1) and next;

	     push @rfnd_set_node2pars_gapstat, $node2gapstat;

	}			# End of the finishing foreach-loop #

	return \@rfnd_set_node2pars_gapstat;

}	# End of the subroutine 'map_parsim_gapstates_sankoff' #

=cut # END of "Set Aside for Later Use. (2)"


# locate_parsim_changes_xgap:
#	This subroutine was created by modifying 'locate_parsim_changes'.
#	The only difference from the predecessor is that it has an option to ignore the changes from or to gaps.
#	[CREATED on 2020/03/13.]
#
# locate_parsim_changes:
#	This subroutine receives a map of states (or characters) onto nodes,
#	and locates the branches that exhibit changes
#	under the assumption of parsimony.
#
# [ INPUT: (\%node2pa, \%node2stat, $indx_stat), where ] ... for the predecessor, 'locate_parsim_changes'.
#
# INPUT: (\%node2pa, \%node2stat, $indx_stat, $char_abs, $if_ignore_gaps), where
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the top node;
#	%node2stat = ($node_id => $stat_info, ...) for all nodes, with
#	$stat_info->[$indx_stat] if $indx_stat is defined,
#	or $stat_info if $indx_stat = undef, being the state (character)
#	mapped onto the node $node_id.
#
#		# (ADDED in creation of 'locate_parsim_changes_xgap' on 2020/03/13.) #
#	$char_abs is a character representing an "absence" state (i.e., gap);
#	$if_ignore_gaps = 1 if you want to ignore changes from or to gaps,
#			= 0 if you want to include such changes as well.
#		# End of "(ADDED in creation of 'locate_parsim_changes_xgap' on 2020/03/13.)" #
#
# OUTPUT: (\%br2change, $ct_changes), where
#
#	%br2change = ($branch_id => [$stat_parent, $stat_node], ...), where
#	$branch_id is actually identical to the ID of the node immediately downstream of the branch.
#	$stat_node is the state (or character) of the node, and
#	$stat_parent is the state (or character) of the parent node.
#
# [ sub locate_parsim_changes (\%\%$) {
#	my ($node2pa, $node2stat, $indx_stat) = @_; ] ... for the predecessor, 'locate_parsim_changes'.
#
sub locate_parsim_changes_xgap (\%\%$$$) {

	my ($node2pa, $node2stat, $indx_stat, $char_abs, $if_ignore_gaps) = @_;

	my $ct_changes = 0;
	my %br2change;

	foreach my $node (keys %{$node2pa}) {
	    my $parent = $node2pa->{$node};
	    (defined $parent) or next;

	    my $stat_info = $node2stat->{$node};
	    unless (defined $stat_info) {
		print STDERR "[ERROR 'locate_parsim_changes_xgap'] NO STATE INFORMATION on the node '$node': $! !\n";
		return (undef, undef);
	    }
	    my $pa_stat_info = $node2stat->{$parent};
	    unless (defined $pa_stat_info) {
		print STDERR "[ERROR 'locate_parsim_changes_xgap'] NO STATE INFORMATION on the parent '$parent' of the node '$node': $! !\n";
		return (undef, undef);
	    }
	    my $stat_node = (defined $indx_stat) ? $stat_info->[$indx_stat] : $stat_info ;
	    my $stat_pa = (defined $indx_stat) ? $pa_stat_info->[$indx_stat] : $pa_stat_info ;
	    ($stat_node eq $stat_pa) and next;

			# (ADDED in creation of 'locate_parsim_changes_xgap' on 2020/03/13.) #
	    if ($if_ignore_gaps) {
		if (($stat_node eq $char_abs) or ($stat_pa eq $char_abs)) { next; }
	    }
			# End of "(ADDED in creation of 'locate_parsim_changes_xgap' on 2020/03/13.)" #

	    $br2change{$node} = [$stat_pa, $stat_node];
	    $ct_changes++;
	}

	return (\%br2change, $ct_changes);

}	# End of the subroutine 'locate_parsim_changes_xgap' #


1;	# a "spell" which makes the module "use"-able and/or "require"-able #


