package MyTreeMap_indels_ML_hs_finer;
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyTreeMap_indels_ML_hs_finer.pm:
#	This package does essentially the same job as 'MyTreeMap_indels_ML_hs.pm'
#	in terms of a "finer" format, in which the multiplication factor of
#	the first-approximate probability under an indel evolutionary model
#	will be calculated FOR each isolated gap-block bounded
#	by an internal node where the Dollo ancestral state is completely occupied by residues
#	and by a pair of sites where all 'descendants' of the internal node is completely occupied by residues, instead of for each gapped segment as the predecessor ('MyTreeMap_indels_ML_hs.pm') did.
#	[CREATION STARTED on Feb 4 (Sat), 2017, ENDED on Feb xx (www), 2017.]
#
#	[MODIFICATION STARTED on Dec 30 (Sat), 2017, ENDED on Jan dd (www), 2018. ]
#
#	This subroutine now adds a correction to the contribution from each "indel history"
#	involving either the deletions of contiguous sub-sequences along the two child branches of a bifurcated root,
#	or an insertion along a parent branch (with only one child branch)
#	followed by a deletion of a sub-sequence immediately next to the inserted sub-sequence
#	along the child branch.
#
# MyTreeMap_indels_ML_hs.pm:
#	This revised package does the same jobs as the predecessor 'MyTreeMap_indels_ML.pm'.
#	MAJOR DIFFERENCES from the predecessor are the following:
#
#	(1) It only considers the "homology structures" of the pairwise alignments (PWAs) (between the sequences at a node and its parent),
#	which are building blocks of each multiple sequence alignment (MSA) accompanied by a given set of ancestral sequence states.
#	[ For the homology structure, see Lunter et al. (2005) BMC Bioinformatics 6:83.]
#
#	In consequence, it gave up distinguishing among an insertion followed by a left-overlapping deletion,
#	an insertion followed by a right-overlapping deletion, and a deletion followed by an insertion in between the sites flanking the deletion.
#	It also regards it an error if a run of inserted sites are split by a run of deleted sites, or vise versa, in an ingredient PWA.
#
#	(2) It uses the pre-computed set of multiplication factors, instead of computing each factor on the fly.
#
#	AS IN THE PREDECESSOR, the calculations are based on Dawg's indel model,
#	whose indel rate parameters are space- and time-homogeneous.
#
#	[CREATION STARTED on May 14 (Thu), 2015, ENDED on May xx (yyy), 2015.]
#
# MyTreeMap_indels_ML.pm:
#	This package calculates an approximate log-likelihood of
#	the gap pattern of an input MSA,
#	using the set of parsimonious indel scenarios realizing the gap pattern.
#
#	[ REVISION from 'MyTreeMap_indels_ML_alpha.pm' ]
#	Whenever possible, 
#	this version uses the logarithms of the (relative) frequencies of indel events,
#	to prevent the subroutines from crashing unexpectedly.
#	
#	(CREATION STARTED on June ?? (???), 2012, ENDED on July ?? (???), 2012)
#
#	[ REVISED on Mar 2 (Sun), 2014]
#	The subroutine 'apprx1_tot_lnlk_gpattern_msa' can now continue its analysis
#	even if some of the gap-block clusters have empty sets for parsimonious scenarios,
#	which indicates that there were some abnormalities,
#	while skipping the calculation of likelihoods for such clusters
#	(and setting their likelihood 'N/A'). 
#
#	[ REVISED on Mar 6 (Thu), 2014 ]
#	The subroutine 'apprx1_tot_lnlk_gpattern_msa' can now continue its analysis
#	even if its satellite subroutine, 'cal_lnlk1_indel_scn_cls', returns undefined outputs,
#	while registering the likelihood in question as 'undef'.
#
#	[ REVISED on Mar 10 (Mon), 2014 ]
#	Accompanying the revision of 'MyTreeMap_indels.pm',
#	the subroutine 'apprx1_tot_lnlk_gpattern_msa' was slightly modified
#	in order to handle exceptional cases which probably involves (a) long gap(s).
#

use strict;
use warnings;

use MyArrayManipu_num qw( diff_nmarray log_sum_exp );
	## union_nmarray

use MyTreeManipu_II qw(
		);
	## input_tree_file read_tree_file br_output_tree
	##make_node2children
	## input_subtree copy_tree
	## output_tree draw_subtree topo_subtree
	## shift_topnode move_topnode segregate_subset_nodes
	## sum_branch_lengths intra_lineage_distance inter_node_distance
	## fetch_boot_series extract_hidden_bootstrap
	## resolve_multipartite

=set aside for later use (IMPORTED from "MyTreeMap_indels_ML_hs.pm") (I)

use MyArrayManipu qw( diff_array union_array nr_array);

use MyMalnIO qw( pr_ODEN_input );

use MyTaxonomy qw(
	fetch_ancestry 
	fetch_last_common_ancestor_ov2
	fetch_common_ancestors
		);
	## fetch_ext_offsprings fetch_all_offsprings
	## fetch_ext_offsprings_br

use MyTreeManipu_III qw(
	make_node2abs_depth make_node2ct_exoffs
	);

use MyTreeMap_II qw(
	locate_parsim_changes
	);
use MyTreeMap_D qw(
	map_parsim_states_dollo_msa
	);
	# ct_presext_dollo_bu map_node2parsstat_dollo_td
	# map_parsim_states_dollo_clm
	# map_parsim_states_dollo_1clm

=cut # END of "set aside for later use (IMPORTED from "MyTreeMap_indels_ML_hs.pm") (I)"

	# ADDED in creation of 'MyTreeMap_indels_ML_hs_finer.pm'. #
use MyTreeMap_indels_spt_odr_finer qw( );

use MyTreeMap_indels_ML_hs qw(
	mk_node2set_empty_blks2

	search_paired_indels2
	search_leftmost_events2
	search_rightmost_events2
	cal_lnlk1_sgl_indel_cls2
	cal_lnlk1_pair_indel_cls2
	);
		# ADDED 'mk_node2set_empty_blks2' in conjunction with IMPORTING the modifications of 'apprx1_tot_lnlk_gpattern_msa2_finer' to 'apprx1_tot_lnlk_gpattern_msa2' in "../MyPerlModules_LOLIPOG/MyTreeMap_indels_ML_hs.pm" (on 2019/04/19). #
	# END of "ADDED in creation of 'MyTreeMap_indels_ML_hs_finer.pm'." #


use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	search_biroot_uninodes
	cal_corr_fac_from_biroot
	cal_corr_fac_from_uninode
	cal_corr_fac_lnmfac1_indel_hst
	calc_lnmfac_gc_partit_nw_for_cl
	cal_sumint_diff_xrate_in_idb
	cal_lnmfac1_gpdseg_br_in_idb2
	cal_lnmfac1_indel_hst_in_idb2
	apprx1_tot_lnlk_gpattern_msa2_finer
	);
	# mk_node2set_empty_blks
	# mk_node2set_empty_blks2 # MOVED to "../MyPerlModules_LOLIPOG/MyTreeMap_indels_ML_hs.pm" (on 2019/04/19).

	## log_freq_indel_len
	## cal_log_normal_fac_distri
	## cal_log_normal_facs_indel_distri

	## cal_delta_exp_del_5p2

	## log_dbl_sum_freq_indel
	## log_compos_freq_indel_len
	## dbl_tinteg_exp_indel

	## search_paired_indels2
	## search_leftmost_events2
	## search_rightmost_events2
	## cal_lnlk1_sgl_indel_cls2
	## cal_lnlk1_pair_indel_cls2
	## cal_lnlk1_indel_br_cls2
	## cal_lnlk1_indel_scn_cls2
	## apprx1_tot_lnlk_gpattern_msa2



	# This specifies the level of tests on the main subroutines,
	# (0, Nothing; 1, brief; 2, a bit detailed; 3, quite detailed; 5, very detailed)

##our $TEST_LEVEL = 1;
##our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 3;
our $TEST_LEVEL = 5;

	# For the main satellite subroutine, 'calc_lnmfac_gc_partit_nw_for_cl' 
our $TEST_LEVEL1a = 5;

	# For the main satellite subroutine, 'cal_lnmfac1_indel_hst_in_idb2'. #
our $TEST_LEVEL1b = 5;

	# For the satellite-satellite subroutine, 'cal_sumint_diff_xrate_in_idb'. #
our $TEST_LEVEL_csidx = 0;
#our $TEST_LEVEL_csidx = 5;

	# For the satellite-satellite subroutine, 'cal_lnmfac1_gpdseg_br_in_idb2'. #
our $TEST_LEVEL_clmbidb = 0;
#our $TEST_LEVEL_clmbidb = 5;


	# For the portions of the subroutines involved in the calculation of correction factors
	# (e.g., 'cal_corr_fac_from_biroot', 'cal_corr_fac_from_uninode', 'cal_corr_fac_lnmfac1_indel_hst').
#our $TEST_LEVEL_corr = 0;
our $TEST_LEVEL_corr = 6;

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
##our $INDX_LENGTH = 0;
##our $INDX_BOOT = 1;
##our $START_GROUP = '(';	# the symbol marking the start of the grouping of nodes.
##our $END_GROUP = ')';  # the symbol marking the end of the grouping of nodes.


=set aside for later use (IMPORTED from "MyTreeMap_indels_ML_hs.pm") (II)

	## These two indices specify where the information is stored
	## in %node2pars_stats = ($node_id => \@stat_info, ...) #
	#
our $INDX_STAT = 0; # The state (or the character) mapped onto the node.
our $INDX_SBTOT_PEN = 1; # The subtotal of penalties on the subtree under $node_id. 

	# This parameter is important to decide whether we should employ
	# the direct summation over indel lengths or its complement,
	# when calculating a 'composite' frequency.

our $MID_INDEL_LEN = 20; # An indel longer than this is regarded as 'long' => direct summation will be calculated.
## our $MID_INDEL_LEN = 10;

=cut # END of "set aside for later use (IMPORTED from "MyTreeMap_indels_ML_hs.pm") (II)"

	## This variable sets an upper-bound for the gap-length.
	# IF a gap-block cluster contains a run of gaps longer than this variable,
	# the subroutine 'enum_psm_indel_scns_for_msa' skips further analyses on the cluster.
	# The variable will be used in the new subroutine, 'scan_gbcls_for_long_gaps'.
	# (ADDED on Mar 10 (Mon), 2014.)
	#
	# NOTE: It should be synchronized with the synonymous global variable in the module 'MyTreeMap_indels.pm'.
	#
our $UPPER_BOUND_ANAL_GAPLEN = $MyTreeMap_indels_spt_odr_finer::UPPER_BOUND_ANAL_GAPLEN;
##our $UPPER_BOUND_ANAL_GAPLEN = 150;

	# This variable sets an upper-bound for the length of the gap-block cluster itself.
	# (ADDED on Mar 11 (Tue), 2014.):w
	#
	# NOTE: It should be synchronized with the synonymous global variable in the module 'MyTreeMap_indels.pm'.
	#
our $UPPER_BOUND_ANAL_LEN_GBCLS = $MyTreeMap_indels_spt_odr_finer::UPPER_BOUND_ANAL_LEN_GBCLS;
##our $UPPER_BOUND_ANAL_LEN_GBCLS = 200;


	### 5 important hash tables that are used in this Module
	### in order to describe a tree
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.


	### DEFINITIONS of the SUBROUTINES ###

	## Subroutines necessary for the corrections of multiplication factors
	## for the "insertion followed by deletion" type indel histories and their relatives.


# if_all_absent:
#	This satellite subroutine of 'find_delpairs' (& 'find_insdelpairs')
#	examines whether a specified set of contiguous blocks 
#	are all occupied by the 'absence' state at all specified nodes.
#	[CREATED on Jan 1 (Mon), 2018]
#	[A FATAL BUG FIXED on 2019/04/18.]
#
# INPUT: ($indx_beg, $indx_end, \@set_nodes, \%node2bw_gstats, $char_abs), where
#
#	$indx_beg/end is the index of the beginning/end of the specified set of contiguous blocks;
#	@set_nodes lists the specified nodes;
#
#	%node2bw_gstats_in_idb = ($node_id => \@bw_gstats_in_idb, ...) is the map of gap states within the indel block onto the
#	nodes, with $blwise_gapstats[$bl] being $char_pres/abs for the $bl th block (in the cluster) at the node $node_id (, or $char_pres if the gap-pattern block (at the node) does not belong to the indel block);
#
#	$char_abs is the character representing an "absence" state.
#
# OUTPUT: $if_all_absent = 1 if all specified blocks are in the 'absent' state at all specified nodes,
#			 = 0 otherwise.
#
sub if_all_absent ($$\@\%$) {

	my ($indx_beg, $indx_end, $set_nodes, $node2bw_gstats, $char_abs) = @_;

	$indx_end++;

	foreach my $node (@{$set_nodes}) { # Outer foreach-loop (over specified nodes). #

	    my $bw_gstats = $node2bw_gstats->{$node};

	    for (my $b = $indx_beg; $b < $indx_end; $b++) { # Inner for-loop (over specified blocks). #
		($bw_gstats->[$b] ne $char_abs) and (return 0); # A FATAL BUG fixed on 2019/04/18!!
		##($bw_gstats->[$b] eq $char_abs) and (return 0);
	    }					# End of the inner for-loop (over specified blocks). #

	}				# End of the outer foreach-loop (over specified nodes). #

	return 1;

}		# End of 'if_all_absent'. #


# search_biroot_uninodes:
#	This satellite subroutine of 'apprx1_tot_lnlk_gpattern_msa2_finer'
#	searches a phylogenetic tree for a bifurcated root
#	and uni-furcated nodes (other than the root).
#	[CREATED in MODIFICATION of "MyTreeMap_indels_ML_hs_finer.pm" on Dec 31 (Sun), 2017.]
#
# INPUT: ($topnode_id, \%node2pa, \%node2ch), where
#
#	$topnode_id is the ID of the top-node (i.e., root);
#	%node2pa = ($node_id => $parent_id, ...) for the non-root nodes;
#	%node2ch = ($node_id => \@child_node_ids, ...) for the internal nodes;
#
# OUTPUT: (\@info_biroot, \@list_uninodes), where
#
#	@info_biroot = ($id_root, $id_child1, $id_child2) for the bifurcated root,
#		      = () if the root is NOT bifurcated;
#
#	@{$list_uninodes[$u]} = ($id_parent, $id_uninode, $id_child) for the $u th unifurcated node (other than the root),
#	$id_uninode, whose parent and the sole child are $id_parent and $id_child, respectively.
#
sub search_biroot_uninodes ($\%\%) {

	my ($topnode_id, $node2pa, $node2ch) = @_;

	my @info_biroot = ();
	my $ch_root = $node2ch->{$topnode_id};
	if (@{$ch_root} == 2) {
	    @info_biroot = ($topnode_id, @{$ch_root});
	}

	my @list_uninodes = ();

	foreach my $node (sort {$a <=> $b} keys %{$node2ch}) { # Main foreach-loop (over internal nodes) #

		# A BUG fixed on Jan 4 (Thu), 2018. #
	    ($node == $topnode_id) and next;
	    ##($node == $topnode_id) or next;

	    my $ch = $node2ch->{$node};
	    (defined $ch) or next;
	    (@{$ch} == 1) or next;
	    my $id_child = $ch->[0];
	    my $id_parent = $node2pa->{$node};

	    push @list_uninodes, [$id_parent, $node, $id_child];

	}	# End of the main foreach-loop (over internal nodes). #

	return (\@info_biroot, \@list_uninodes);

}	# End of 'search_biroot_uninodes'. #


# find_delpairs:
#	This satellite subroutine of 'cal_corr_fac_from_biroot'
#	searches for the pairs of deletions (of contiguous sub-sequences) separated by a bifurcated root.
#
#	[CREATED in MODIFICATION of 'MyTreeMap_indels_ML_hs_finer.pm' on Jan 1 (Mon), 2018]
#
# INPUT: (\@info_biroot, \%br2indels, \%node2blwise_gstats, $char_abs), where
#
#	@{$info_biroot} = ($id_root, $id_child1, $id_child2) for the bifurcated root,
#		      = () if the root is NOT bifurcated;
#
#	Detailes on other input variables are described in the header of 'cal_corr_fac_lnmfac1_indel_hst'.
#
# OUTPUT: \@delpairs, where
#
#	@{$delpairs[$dp]} = ($info_del_ch1, $info_del_ch2] is for the $dp th deletion pair, with
#	$info_del_ch1/2 = [$stat_pa, $stat_node, \@indices_blk] is the information for the deletion along the child branch 1/2.
#	The subroutine returns undef if it fails.
#
sub find_delpairs (\@\%\%$) {

	my ($info_biroot, $br2indels, $node2bw_gstats, $char_abs) = @_;

	my ($id_root, $id_ch1, $id_ch2) = @{$info_biroot};
	my $indels_ch1 = $br2indels->{$id_ch1};
	my $indels_ch2 = $br2indels->{$id_ch2};

	unless ((defined $indels_ch1) and (@{$indels_ch1}>0) and (defined $indels_ch2) and (@{$indels_ch2}>0)) {
	    return []; # A pair of deletions can NEVER be formed.
	}

#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})

	my @dels_ch1 = my @dels_ch2 = ();

	my $ct_indels_ch1 = @{$indels_ch1};
	my $ct_indels_ch2 = @{$indels_ch2};

	for (my $e=0; $e <$ct_indels_ch1; $e++) { # 1st preliminary for-loop. #
		my $indel = $indels_ch1->[$e];
		((@{$indel}>0) and ($indel->[0] ne $char_abs)) and (push @dels_ch1, $indel); 
	}				# End of the 1st preliminary for-loop. #

	unless (@dels_ch1>0) {
	    return []; # A pair of deletions can NEVER be formed.
	}

	for (my $e=0; $e <$ct_indels_ch2; $e++) { # 2nd preliminary for-loop. #
		my $indel = $indels_ch2->[$e];
		((@{$indel}>0) and ($indel->[0] ne $char_abs)) and (push @dels_ch2, $indel); 
	}				# End of the 2nd preliminary for-loop. #

	unless (@dels_ch2>0) {
	    return []; # A pair of deletions can NEVER be formed.
	}

	my @std_dels_ch1 = sort {$a->[2][0] <=> $b->[2][0]} @dels_ch1;
	my @std_dels_ch2 = sort {$a->[2][0] <=> $b->[2][0]} @dels_ch2;

	my @delpairs = ();

	my $del1 = shift @std_dels_ch1;
	my $del2 = shift @std_dels_ch2;

	while ((defined $del1) and (defined $del2)) { # Main outer while-loop (as long as we have pairs to examine). #
	    my $indices_blk1 = $del1->[2];
	    my $indices_blk2 = $del2->[2];
	    my ($first_blk1, $last_blk1) = @{$indices_blk1}[0, $#{$indices_blk1}];
	    my ($first_blk2, $last_blk2) = @{$indices_blk2}[0, $#{$indices_blk2}];

	    if ($last_blk1 < $last_blk2) {

		if ($last_blk1 + 1 == $first_blk2) {
		    push @delpairs, [$del1, $del2];
		} elsif ($last_blk1 < $first_blk2) {
		    my $if_all_absent = if_all_absent ($last_blk1 + 1, $first_blk2-1, @{$info_biroot}, %{$node2bw_gstats}, $char_abs);
		    $if_all_absent and (push @delpairs, [$del1, $del2]);
		}

		$del1 = shift @std_dels_ch1; # Update the first deletion. #

	    } elsif ($last_blk2 < $last_blk1) {

		if ($last_blk2 + 1 == $first_blk1) {
		    push @delpairs, [$del1, $del2];
		} elsif ($last_blk2 < $first_blk1) {
		    my $if_all_absent = if_all_absent ($last_blk2 + 1, $first_blk1-1, @{$info_biroot}, %{$node2bw_gstats}, $char_abs);
		    $if_all_absent and (push @delpairs, [$del1, $del2]);
		}

		$del2 = shift @std_dels_ch2; # Update the second deletion. #

	    } else {
		# This case is extremely inconceivable. If it occurs, however, update both deletions.
		$del1 = shift @std_dels_ch1;
		$del2 = shift @std_dels_ch2;
	    }

	}				# End of the main outer while-loop (as long as we have pairs to examine). #

	return \@delpairs;

}		# End of 'find_delpairs'. #


# cal_corr_fac_from_biroot:
#	This satellite subroutine of 'cal_corr_fac_lnmfac1_indel_hst' computes
#	the correction factor contributed from a pair of deletions separated by a bifurcated root.
#
#	[CREATED in MODIFICATION of 'MyTreeMap_indels_ML_hs_finer.pm', from Dec 31 (Sun), 2017 till Jan 2 (Tue), 2018.]
#
#	[NOTE] This initial version of this subroutine deals with the deletion pair at either end of the MSA
#	in the same way as the pair in the bulk.
#
#
# INPUT: (\@info_biroot, \@cls_gpb, \%br2atrr, \%br2indels, \%node2blwise_gstats, $char_abs, $totrate_ins, $totrate_del, \@params_boundaries, \%br2log_mfacs), where
#
#	@{$info_biroot} = ($id_root, $id_child1, $id_child2) for the bifurcated root,
#		      = () if the root is NOT bifurcated;
#
#	Detailes on other input variables are described in the header of 'cal_corr_fac_lnmfac1_indel_hst'.
#
# OUTPUT: $corr_fac1,
#	which is the logarithm of the multiplicative correlation factor for the contribution 
#	of the pair of deletions separated by the bifurcated root.
#	It is undefined if the subroutine failed.
#
sub cal_corr_fac_from_biroot (\@\@\%\%\%$$$\@\%) {

	my ($info_biroot, $cls_gpblks, $br2attr, $br2indels, $node2bw_gstats, $char_abs, $totrate_ins, $totrate_del, $params_bds, $br2log_mfacs) = @_;

	my $ct_gpblks = @{$cls_gpblks};

	my ($id_root, $id_ch1, $id_ch2) = @{$info_biroot};

	my $delpairs = find_delpairs (@{$info_biroot}, %{$br2indels}, %{$node2bw_gstats}, $char_abs);
	unless (defined $delpairs) {
	    print STDOUT "[ERROR 'cal_corr_fac_from_biroot'] The subroutine 'find_delpairs' FAILED: $! !\n\n";
	    return undef;
	}
	unless (@{$delpairs}>0) {
	    return 0; # A pair of deletions can NEVER be formed, thus no correction is necessary.
	}
	my $ct_delpairs = @{$delpairs};

   (print STDOUT ' ' x 5, "< 'cal_corr_fac_from_biroot'   (ID{root} = $id_root; ID{child node 1} = $id_ch1; ID{child node 2} = $id_ch2; #{delpairs} = $ct_delpairs) >\n\n") if ($TEST_LEVEL_corr>2);

	my ($cutoff_ins, $cutoff_del, $if_5p_end, $if_3p_end) = @{$params_bds};

	my $len_br1 = $br2attr->{$id_ch1}[$INDX_LENGTH];
	my $len_br2 = $br2attr->{$id_ch2}[$INDX_LENGTH];

	my $exponent_per_site = ($totrate_ins + $totrate_del) * ($len_br1 + $len_br2);

   if ($TEST_LEVEL_corr>3) {
	print STDOUT ' ' x 5, "[ Main inputs and secondary input ]\n\n";
	print STDOUT ' ' x 5, '$cutoff_del = ', $cutoff_del, ",\n";
	print STDOUT ' ' x 5, '$totrate_ins = ', $totrate_ins, ",\n";
	print STDOUT ' ' x 5, '$totrate_del = ', $totrate_del, ",\n";
	print STDOUT ' ' x 5, '$len_br1 = ', $len_br1, ",\n\n";
	print STDOUT ' ' x 5, '$len_br2 = ', $len_br2, ",\n\n";
	print STDOUT ' ' x 5, '$exponent_per_site (= ($totrate_ins + $totrate_del) * ($len_br1 + $len_br2)) = ', $exponent_per_site, "\n\n";
   }

	my $log_mfacs_del1 = $br2log_mfacs->{$id_ch1}[1];
	my $log_mfacs_del2 = $br2log_mfacs->{$id_ch2}[1];

#	%br2log_mfacs = ($branch_id => \@setset_log_mfacs, ...) assigns the multiplication factors into each branch, with
#	@setset_log_mfacs = (\@log_mfacs_ins, \@log_mfacs_del, \@log_mfacs_pair, \@log_mfacs_ins_bd, \@log_mfacs_del_bd, \@log_mfacs_pair_bd), with
#	$log_mfacs_ins/del[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertion/deletion of $l sites in the bulk of the MSA,

	my $corr_fac1 = 0;

   (print STDOUT ' ' x5, "[ Calculating correction factors from $ct_delpairs deletion pairs ]\n\n") if ($TEST_LEVEL_corr>3);

	foreach my $delpair (@{$delpairs}) { # Main outer foreach-loop (over deletion pairs). #

	    my ($info_del1, $info_del2) = @{$delpair};
	    my $blks_del1 = $info_del1->[2];
	    my $blks_del2 = $info_del2->[2];

	    my $len_del1 = my $len_del2 = 0;

	    foreach my $blk (@{$blks_del1}) { $len_del1 += $cls_gpblks->[$blk][1]; }
	    foreach my $blk (@{$blks_del2}) { $len_del2 += $cls_gpblks->[$blk][1]; }

#	@{$cls_gpblocks[$gpb]} = ($indx_gp, $len_gpb, $start, $end) is
#	for the $gpb th gap-pattern-block in the input cluster, with
#	$len_gpb being the length (in terms of #{columns}) of the block,

	    my @elements = ();

	    my $upper_bound = ($len_del1 > $len_del2) ? $cutoff_del - $len_del1 : $cutoff_del - $len_del2;
	    $upper_bound++;

   if ($TEST_LEVEL_corr>= 5) {
	print STDOUT ' ' x 5, "{deleted blocks 1} = ", join (', ', @{$blks_del1}), " (total length: $len_del1);\n";
	print STDOUT ' ' x 5, "{deleted blocks 2} = ", join (', ', @{$blks_del2}), " (total length: $len_del2);\n";
	print STDOUT ' ' x 5, "Summation over #{'invisible' residues} in [0, $upper_bound) .\n\n";
   }

	    ($upper_bound>0) or next; # Ignore the pair. #

	    for (my $x = 0; $x <$upper_bound; $x++) {
		my $element = (-$x * $exponent_per_site) + $log_mfacs_del1->[$len_del1 + $x] + $log_mfacs_del2->[$len_del2 + $x];
		push @elements, $element;

    (print STDOUT ' ' x 7, '$x = ', $x, ",\t", '$element = ', $element, ",\n") if ($TEST_LEVEL_corr>5); 
	    }

	    my $numer = (@elements==1) ? $elements[0] : log_sum_exp (@elements);
	    unless (defined $numer) {
		print STDOUT "[ERROR 'cal_corr_fac_from_biroot'] The subroutine 'log_sum_exp' FAILED: $! !\n\n";
		return undef;
	    }

	    my $corr_fac11 = ($numer + log (2)) - ($log_mfacs_del1->[$len_del1] + $log_mfacs_del2->[$len_del2]) ;

	    $corr_fac1 +=  $corr_fac11;

    if ($TEST_LEVEL_corr>= 5) {
	print STDOUT ' ' x 5, 'Log(total contribution from the delpairs with 0-', $upper_bound, ' invisible residues) (= $numer) = ', $numer, ",\n";
	print STDOUT ' ' x 5, 'Log(multiple factor from the original deletion 1) (= $log_mfacs_del1->[$len_del1]) = ', $log_mfacs_del1->[$len_del1], ",\n";
	print STDOUT ' ' x 5, 'Log(multiple factor from the original deletion 2) (= $log_mfacs_del2->[$len_del2]) = ', $log_mfacs_del2->[$len_del2], ",\n\n";
    }
    if ($TEST_LEVEL_corr > 3) {
	print STDOUT ' ' x 5, '$corr_fac11 (= ($numer + log (2)) - ($log_mfacs_del1->[$len_del1] + $log_mfacs_del2->[$len_del2])) = ', $corr_fac11, ",\n";
	print STDOUT ' ' x 5, 'Current $corr_fac1 = ', $corr_fac1, ".\n\n";
    }

	}			# End of the main outer foreach-loop (over deletion pairs). #

			### RESTART FROM HERE (2a) ###

   (print STDOUT ' ' x 5, "< 'cal_corr_fac_from_biroot'   ended  (with", '$corr_fac1 = ', "$corr_fac1) />\n\n") if ($TEST_LEVEL_corr>2);

	return $corr_fac1;

}	# End of 'cal_corr_fac_from_biroot'. #


# find_insdelpairs:
#	This satellite subroutine of 'cal_corr_fac_from_uninode'
#	searches for pairs of an insertion and a subsequent deletion (of contiguous sub-sequences) separated by a unifurcated node.
#	The subroutine was created by modifying the sister subroutine 'find_delpairs'.
#
#	[CREATED in MODIFICATION of 'MyTreeMap_indels_ML_hs_finer.pm' on Jan 2 (Tue), 2018]
#
# find_delpairs:
#	This satellite subroutine of 'cal_corr_fac_from_biroot'
#	searches for the pairs of deletions (of contiguous sub-sequences) separated by a bifurcated root.
#
#	[CREATED in MODIFICATION of 'MyTreeMap_indels_ML_hs_finer.pm' on Jan 1 (Mon), 2018]
#
# INPUT: (\@info_uninode, \%br2indels, \%node2blwise_gstats, $char_abs), where
#
#	@{$info_uninode} = ($id_parent, $id_uninode, $id_child) for a unifurcated node (other than the root),
#	$id_uninode, whose parent and the sole child are $id_parent and $id_child, respectively;
#
#	Detailes on other input variables are described in the header of 'cal_corr_fac_lnmfac1_indel_hst'.
#
# OUTPUT: \@insdelpairs, where
#
#	@{$insdelpairs[$idp]} = ($info_ins, $info_del_ch] is for the $idp th insertion-deletion pair, where
#	$info_ins = [$stat_pa, $stat_node, \@indices_blk] is the information for the insertion along the branch immediately above the unifurcated node, and
#	$info_del_ch = [$stat_pa, $stat_node, \@indices_blk] is the information for the deletion along the branch immediately below the unifurcated node.
#	The subroutine returns undef if it fails.
#
sub find_insdelpairs (\@\%\%$) {

	my ($info_uninode, $br2indels, $node2bw_gstats, $char_abs) = @_;

	my ($id_pa, $id_uninode, $id_ch) = @{$info_uninode};
	my $indels_sbj = $br2indels->{$id_uninode};
	my $indels_ch = $br2indels->{$id_ch};

	unless ((defined $indels_sbj) and (@{$indels_sbj}>0) and (defined $indels_ch) and (@{$indels_ch}>0)) {
	    return []; # A pair of an insertion and a subsequent deletion can NEVER be formed.
	}

#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})

	my @ins_sbj = my @dels_ch = ();

	my $ct_indels_sbj = @{$indels_sbj};
	my $ct_indels_ch = @{$indels_ch};

	for (my $e=0; $e <$ct_indels_sbj; $e++) { # 1st preliminary for-loop. #
		my $indel = $indels_sbj->[$e];
		((@{$indel}>0) and ($indel->[0] eq $char_abs)) and (push @ins_sbj, $indel); 
	}				# End of the 1st preliminary for-loop. #

	unless (@ins_sbj>0) {
	    return []; # A pair of an insertion and a subsequent deletion can NEVER be formed.
	}

	for (my $e=0; $e <$ct_indels_ch; $e++) { # 2nd preliminary for-loop. #
		my $indel = $indels_ch->[$e];
		((@{$indel}>0) and ($indel->[0] ne $char_abs)) and (push @dels_ch, $indel); 
	}				# End of the 2nd preliminary for-loop. #

	unless (@dels_ch>0) {
	    return []; # A pair of an insertion and a subsequent deletion can NEVER be formed.
	}

	my @std_ins_sbj = sort {$a->[2][0] <=> $b->[2][0]} @ins_sbj;
	my @std_dels_ch = sort {$a->[2][0] <=> $b->[2][0]} @dels_ch;

	my @insdelpairs = ();

	my $ins = shift @std_ins_sbj;
	my $del = shift @std_dels_ch;

	while ((defined $ins) and (defined $del)) { # Main outer while-loop (as long as we have pairs to examine). #

	    my $indices_blk1 = $ins->[2];
	    my $indices_blk2 = $del->[2];
	    my ($first_blk1, $last_blk1) = @{$indices_blk1}[0, $#{$indices_blk1}];
	    my ($first_blk2, $last_blk2) = @{$indices_blk2}[0, $#{$indices_blk2}];

	    if ($last_blk1 < $last_blk2) {

		if ($last_blk1 + 1 == $first_blk2) {
		    push @insdelpairs, [$ins, $del];
		} elsif ($last_blk1 < $first_blk2) {
		    my $if_all_absent = if_all_absent ($last_blk1 + 1, $first_blk2-1, @{$info_uninode}, %{$node2bw_gstats}, $char_abs);
		    $if_all_absent and (push @insdelpairs, [$ins, $del]);
		}

		$ins = shift @std_ins_sbj; # Update the insertion. #

	    } elsif ($last_blk2 < $last_blk1) {

		if ($last_blk2 + 1 == $first_blk1) {
		    push @insdelpairs, [$ins, $del];
		} elsif ($last_blk2 < $first_blk1) {
		    my $if_all_absent = if_all_absent ($last_blk2 + 1, $first_blk1-1, @{$info_uninode}, %{$node2bw_gstats}, $char_abs);
		    $if_all_absent and (push @insdelpairs, [$ins, $del]);
		}

		$del = shift @std_dels_ch; # Update the deletion. #

	    } else {
		# This case is extremely inconceivable. If it occurs, however, update both deletions.
		$ins = shift @std_ins_sbj;
		$del = shift @std_dels_ch;
	    }

	}				# End of the main outer while-loop (as long as we have pairs to examine). #

	return \@insdelpairs;

}		# End of 'find_insdelpairs'. #


# cal_corr_fac_from_uninode:
#	This satellite subroutine of 'cal_corr_fac_lnmfac1_indel_hst' computes
#	the correction factor contributed from a pair of an insertion and a subsequent deletion
#	separated by a unifurcated node.
#	This subroutine was created by modifying the sister subroutine, 'cal_corr_fac_from_biroot'.
#
#	[CREATED in MODIFICATION of 'MyTreeMap_indels_ML_hs_finer.pm', on Jan 2 (Tue), 2018.]
#
#	[NOTE] This initial version of this subroutine deals with the deletion pair at either end of the MSA
#	in the same way as the pair in the bulk.
#
# cal_corr_fac_from_biroot:
#	This satellite subroutine of 'cal_corr_fac_lnmfac1_indel_hst' computes
#	the correction factor contributed from a pair of deletions separated by a bifurcated root.
#
#	[CREATED in MODIFICATION of 'MyTreeMap_indels_ML_hs_finer.pm', from Dec 31 (Sun), 2017 till Jan 2 (Tue), 2018.]
#
# INPUT: (\@info_uninode, \@cls_gpb, \%br2atrr, \%node2pa, \%br2indels, \%node2blwise_gstats, $char_abs, $totrate_ins, $totrate_del, \@params_boundaries, \%br2log_mfacs), where
#
#	@{$info_uninode} = ($id_parent, $id_uninode, $id_child) for a unifurcated node (other than the root),
#	$id_uninode, whose parent and the sole child are $id_parent and $id_child, respectively;
#
#	Detailes on other input variables are described in the header of 'cal_corr_fac_lnmfac1_indel_hst'.
#
# OUTPUT: $corr_fac2,
#	which is the logarithm of the multiplicative correlation factor for the contribution 
#	of the pairs of an insertion and a subsequent deletion separated by the unifurcated node.
#	It is undefined if the subroutine failed.
#
sub cal_corr_fac_from_uninode (\@\@\%\%\%$$$\@\%) {

	my ($info_uninode, $cls_gpblks, $br2attr, $br2indels, $node2bw_gstats, $char_abs, $totrate_ins, $totrate_del, $params_bds, $br2log_mfacs) = @_;

	my $ct_gpblks = @{$cls_gpblks};

	my ($id_pa, $id_uninode, $id_ch) = @{$info_uninode};

	my $insdelpairs = find_insdelpairs (@{$info_uninode}, %{$br2indels}, %{$node2bw_gstats}, $char_abs);
	unless (defined $insdelpairs) {
	    print STDOUT "[ERROR 'cal_corr_fac_from_uninode'] The subroutine 'find_insdelpairs' FAILED: $! !\n\n";
	    return undef;
	}
	unless (@{$insdelpairs}>0) {
	    return 0; # A pair of an insertion & a subsequent deletion can NEVER be formed, thus no correction is necessary.
	}
	my $ct_insdelpairs = @{$insdelpairs};

   (print STDOUT ' ' x 5, "< 'cal_corr_fac_from_uninode'   (ID{uni-furcated node} = $id_uninode; ID{child node} = $id_ch; #{insdelpairs} = $ct_insdelpairs) >\n\n") if ($TEST_LEVEL_corr>2);

	my ($cutoff_ins, $cutoff_del, $if_5p_end, $if_3p_end) = @{$params_bds};

	my $len_br_ch = $br2attr->{$id_ch}[$INDX_LENGTH];

	my $exponent_per_site = ($totrate_ins + $totrate_del) * $len_br_ch;

   if ($TEST_LEVEL_corr>3) {
	print STDOUT ' ' x 5, "[ Main inputs and secondary input ]\n\n";
	print STDOUT ' ' x 5, '$cutoff_ins = ', $cutoff_ins, ",\n";
	print STDOUT ' ' x 5, '$cutoff_del = ', $cutoff_del, ",\n";
	print STDOUT ' ' x 5, '$totrate_ins = ', $totrate_ins, ",\n";
	print STDOUT ' ' x 5, '$totrate_del = ', $totrate_del, ",\n";
	print STDOUT ' ' x 5, '$len_br_ch = ', $len_br_ch, ",\n\n";
	print STDOUT ' ' x 5, '$exponent_per_site (= ($totrate_ins + $totrate_del) * $len_br_ch) = ', $exponent_per_site, "\n\n";
   }

	my $log_mfacs_ins_sbj = $br2log_mfacs->{$id_uninode}[0];
	my $log_mfacs_del_ch = $br2log_mfacs->{$id_ch}[1];

#	%br2log_mfacs = ($branch_id => \@setset_log_mfacs, ...) assigns the multiplication factors into each branch, with
#	@setset_log_mfacs = (\@log_mfacs_ins, \@log_mfacs_del, \@log_mfacs_pair, \@log_mfacs_ins_bd, \@log_mfacs_del_bd, \@log_mfacs_pair_bd), with
#	$log_mfacs_ins/del[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertion/deletion of $l sites in the bulk of the MSA,

	my $corr_fac2 = 0;

   (print STDOUT ' ' x5, "[ Calculating correction factors from $ct_insdelpairs insertion-deletion pairs ]\n\n") if ($TEST_LEVEL_corr>3);

	foreach my $insdelpair (@{$insdelpairs}) { # Main outer foreach-loop (over insertion-deletion pairs). #

	    my ($info_ins, $info_del) = @{$insdelpair};
	    my $blks_ins = $info_ins->[2];
	    my $blks_del = $info_del->[2];

	    my $len_ins = my $len_del = 0;

	    foreach my $blk (@{$blks_ins}) { $len_ins += $cls_gpblks->[$blk][1]; }
	    foreach my $blk (@{$blks_del}) { $len_del += $cls_gpblks->[$blk][1]; }

#	@{$cls_gpblocks[$gpb]} = ($indx_gp, $len_gpb, $start, $end) is
#	for the $gpb th gap-pattern-block in the input cluster, with
#	$len_gpb being the length (in terms of #{columns}) of the block,

	    my @elements = ();

	    my $upper_bound1 = $cutoff_ins - $len_ins;
	    my $upper_bound2 = $cutoff_del - $len_del;
	    my $upper_bound = ($upper_bound1 < $upper_bound2) ? $upper_bound1 : $upper_bound2;
	    $upper_bound++;

   if ($TEST_LEVEL_corr>= 5) {
	print STDOUT ' ' x 5, "{inserted blocks} = ", join (', ', @{$blks_ins}), " (total length: $len_ins);\n";
	print STDOUT ' ' x 5, "{deleted blocks} = ", join (', ', @{$blks_del}), " (total length: $len_del);\n";
	print STDOUT ' ' x 5, "Summation over #{'invisible' residues} in [0, $upper_bound) .\n\n";
   }

	    ($upper_bound>0) or next; # Ignore the pair. #

	    for (my $x = 0; $x <$upper_bound; $x++) {
		my $element = (-$x * $exponent_per_site) + $log_mfacs_ins_sbj->[$len_ins + $x] + $log_mfacs_del_ch->[$len_del + $x];
		push @elements, $element;

    (print STDOUT ' ' x 7, '$x = ', $x, ",\t", '$element = ', $element, ",\n") if ($TEST_LEVEL_corr>5); 
	    }

	    my $numer = (@elements==1) ? $elements[0] : log_sum_exp (@elements);
	    unless (defined $numer) {
		print STDOUT "[ERROR 'cal_corr_fac_from_uninode'] The subroutine 'log_sum_exp' FAILED: $! !\n\n";
		return undef;
	    }


	    my $corr_fac22 = ($numer + log (2)) - ($log_mfacs_ins_sbj->[$len_ins] + $log_mfacs_del_ch->[$len_del]) ;

	    $corr_fac2 +=  $corr_fac22;

    if ($TEST_LEVEL_corr>= 5) {
	print STDOUT ' ' x 5, 'Log(total contribution from the insdelpairs with 0-', $upper_bound, ' invisible residues) (= $numer) = ', $numer, ",\n";
	print STDOUT ' ' x 5, 'Log(multiple factor from the original insertion) (= $log_mfacs_ins_sbj->[$len_ins]) = ', $log_mfacs_ins_sbj->[$len_ins], ",\n";
	print STDOUT ' ' x 5, 'Log(multiple factor from the original deletion) (= $log_mfacs_del_ch->[$len_del]) = ', $log_mfacs_del_ch->[$len_del], ",\n\n";
    }
    if ($TEST_LEVEL_corr > 3) {
	print STDOUT ' ' x 5, '$corr_fac22 (= ($numer + log (2)) - ($log_mfacs_ins_sbj->[$len_ins] + $log_mfacs_del_ch->[$len_del])) = ', $corr_fac22, ",\n";
	print STDOUT ' ' x 5, 'Current $corr_fac2 = ', $corr_fac2, ".\n\n";
    }

	}			# End of the main outer foreach-loop (over insertion-deletion pairs). #

			### RESTART FROM HERE (2b) ###

   (print STDOUT ' ' x 5, "< 'cal_corr_fac_from_uninode'   ended  (with", '$corr_fac2 = ', "$corr_fac2) />\n\n") if ($TEST_LEVEL_corr>2);

	return $corr_fac2;

}	# End of 'cal_corr_fac_from_uninode'. #


# cal_corr_fac_lnmfac1_indel_hst:
#	This satellite subroutine of 'apprx1_tot_lnlk_gpattern_msa2_finer' 
#	computes the correction factor for the multiplication factor contributed from an indel history 
#	involving at least a pair of an insertion and a subsequent deletion separated by a uni-furcated node,
#	and/or a pair of deletions separated by a bifurcated root.
#
#	[NOTE] This initial version of this subroutine deals with the indel history at either end of the MSA
#	in the same way as dealing with the history in the bulk.
#
#	[CREATED in MODIFICATION of 'MyTreeMap_indels_ML_hs_finer.pm', from Dec 31 (Sun), 2017 till ...]
#
# INPUT: (\@info_biroot, \@list_uninodes, \@cls_gpb, \%br2atrr, \%node2pa, \%br2indels, \%node2blwise_gstats, \%node2empty_gpbs0, $char_abs, $totrate_ins, $totrate_del, \@params_boundaries, \%br2log_mfacs, $delta_log_prior_rootseq), where
#
#	@{$info_biroot} = ($id_root, $id_child1, $id_child2) for the bifurcated root,
#		      = () if the root is NOT bifurcated;
#	@{$list_uninodes[$u]} = ($id_parent, $id_uninode, $id_child) for the $u th unifurcated node (other than the root),
#	$id_uninode, whose parent and the sole child are $id_parent and $id_child, respectively;
#
#	@{$cls_gpblocks[$gpb]} = ($indx_gp, $len_gpb, $start, $end) is
#	for the $gpb th gap-pattern-block in the input cluster, with
#	$indx_gp specifying the gap pattern,
#	$len_gpb being the length (in terms of #{columns}) of the block,
#	and $start/$end being the starting/ending coordinate of the block
#	in the 0-based convention;
#
#	%br2attr = ($br_id => \@attr, ...) for the branches, with
#	$br_id being the ID of the node at the lower-end of the branch, and
#	$attr[$INDX_LENGTH] giving the length of the branch (in terms of #{substitutions}/site);
#
#	%node2pa = ($node_id => $parent_node_id, ...) for all the nodes except the top node;
#	$len_msa is the length of the input MSA;
#	$char_abs is the character (or symbol) representing the 'absence' state;
#
#	%br2indels_in_idb = ($branch_id => \@list_indels, ...) is the map of insertions/deletions within the indel block onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})
#	of the blocks involved in the indel event;
#
#	%node2bw_gstats_in_idb = ($node_id => \@bw_gstats_in_idb, ...) is the map of gap states within the indel block onto the
#	nodes, with $blwise_gapstats[$bl] being $char_pres/abs for the $bl th block (in the cluster) at the node $node_id (, or $char_pres if the gap-pattern block (at the node) does not belong to the indel block);
#
#	%node2empty_gpbs0 = ($node_id => \@empty_gpbs0, ...) for the nodes involved in the indel block,
#	with @empty_gpbs0 listing the indices of the gap-pattern-blocks that are empty at the node; # ADDED in creation of 'cal_lnmfac1_indel_hst_in_idb2'. #
#
#	$totrate_ins/del being the total rate of insertions/deletions (per site) (relative to the substitution rate),
#
#	@params_boundaries = ($cutoff_ins, $cutoff_del, $if_5p_end, $if_3p_end), with
#
#	$cutoff_ins/del being the cut-off length for the insertions/deletions that is common along the MSA,
#
#	$if_5p/3p_end = 1 if the gap-block cluster extends into the 5'-/3'-end (=0 otherwise).
#
#	%br2log_mfacs = ($branch_id => \@setset_log_mfacs, ...) assigns the multiplication factors into each branch, with
#	@setset_log_mfacs = (\@log_mfacs_ins, \@log_mfacs_del, \@log_mfacs_pair, \@log_mfacs_ins_bd, \@log_mfacs_del_bd, \@log_mfacs_pair_bd), with
#	$log_mfacs_ins/del[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertion/deletion of $l sites in the bulk of the MSA,
#	$log_mfacs_ins/del_bd[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertions/deletions of $l sites on either boundary of the MSA,
#	$log_mfacs_pair[$l_A]->[$l_D] ($l_A <= $cutoff_del and $l_D <= $cutoff_ins) being the logarithm of the multiplication factor for the pair of an (effective) insertion of $l_D sites and an (effective) deletion of $l_A sites in the bulk of the MSA,
#	$log_mfacs_pair_bd[$l_A]->[$l_D] ($l_A <= $cutoff_del and $l_D <= $cutoff_ins) being the logarithm of the multiplication factor for the pair of an (effective) insertion of $l_D sites and an (effective) deletion of $l_A sites on either boundary of the MSA.
#
#	$delta_log_prior_rootseq_in_idb is the increment of the logarithm of the multiplicative contribution from the indel block to the prior probability of the root sequence state. # ADDED in creation of 'cal_lnmfac1_indel_hst_in_idb2'. #
#
#
# OUTPUT: $corr_fac, 
#	which is the logarithm of the multiplicative correlation factor for the multiplication factor
#	contributed from this indel history.
#	It is undefined if the subroutine failed.
#
sub cal_corr_fac_lnmfac1_indel_hst (\@\@\@\%\%\%\%\%$$$\@\%$) {

	my ($info_biroot, $list_uninodes, $cls_gpblks, $br2attr, $node2pa, $br2indels, $node2bw_gstats, $node2empty_gpbs0, $char_abs, $totrate_ins, $totrate_del, $params_bds, $br2log_mfacs, $delta_log_prior_rootseq_in_idb) = @_;

	##my $ct_gpblks = @{$cls_gpblks};
	my $ct_uninodes = @{$list_uninodes};

    (print STDOUT ' ' x 3, "< Subroutine 'cal_corr_fac_lnmfac1_indel_hst' (#{uni-furcated nodes} = $ct_uninodes) >\n\n") if (($TEST_LEVEL_corr>0) and ((@{$info_biroot}>0) or ($ct_uninodes>0))) ;

	my $corr_fac = 0;

	if (@{$info_biroot}>0) {
	    my $corr_fac1 = cal_corr_fac_from_biroot (@{$info_biroot}, @{$cls_gpblks}, %{$br2attr}, %{$br2indels}, %{$node2bw_gstats}, $char_abs, $totrate_ins, $totrate_del, @{$params_bds}, %{$br2log_mfacs});

	    unless (defined $corr_fac1) {
		print STDOUT "[ERROR 'cal_corr_fac_lnmfac1_indel_hst'] The subroutine 'cal_corr_fac_from_biroot' FAILED: $! !\n\n";
		return undef;
	    }

	    $corr_fac += $corr_fac1;

   if ($TEST_LEVEL_corr>1) {
	print STDOUT ' ' x 3, "{Correction factor from the bifurcated root} = $corr_fac1,\n";
	print STDOUT ' ' x 3, 'Current $corr_fac = ', $corr_fac, ".\n\n";
   } 
	} # END of "if (@{$info_biroot}>0) {...}"


	for (my $u=0; $u<$ct_uninodes; $u++) { # Main for-loop (over uni-furcated nodes). #

	    my $info_uninode = $list_uninodes->[$u];

	    my $corr_fac2 = cal_corr_fac_from_uninode (@{$info_uninode}, @{$cls_gpblks}, %{$br2attr}, %{$br2indels}, %{$node2bw_gstats}, $char_abs, $totrate_ins, $totrate_del, @{$params_bds}, %{$br2log_mfacs});

	    unless (defined $corr_fac2) {
		print STDOUT "[ERROR 'cal_corr_fac_lnmfac1_indel_hst'] The subroutine 'cal_corr_fac_from_uninode' FAILED for the $u th uni-furcated node: $! !\n\n";
		return undef;
	    }
	    $corr_fac += $corr_fac2;

   if ($TEST_LEVEL_corr>1) {
	print STDOUT ' ' x 3, "{Correction factor from the $u th uni-furcated node} = $corr_fac2,\n";
	print STDOUT ' ' x 3, 'Current $corr_fac = ', $corr_fac, ".\n\n";
   }

	}		# End of the main for-loop (over uni-furcated nodes). #


    (print STDOUT "\n", ' ' x 3, "< Subroutine 'cal_corr_fac_lnmfac1_indel_hst' Ended (with ", '$corr_fac = ', "$corr_fac) />\n\n") if (($TEST_LEVEL_corr>0) and ((@{$info_biroot}>0) or ($ct_uninodes>0))) ;

	return $corr_fac;

}	# End of 'cal_corr_fac_lnmfac1_indel_hst'. #


	### Subroutines Necessary for STEP 5 in "workflow_for_indel_histories&multiplication_factors.rtf". ###


	# Satellite Subroutines of 'apprx1_tot_lnlk_gpattern_msa2_finer'. #

# calc_lnmfac_gc_partit_nw_for_cl:
#	This satellite subroutine of 'apprx1_tot_lnlk_gpattern_msa2_finer' calculates
#	the logarithm of the multiplication factor (of the MSA probability) contributed from
#	the "partitioning network" (defined by the Dollo parsimonious indel history) within a given gapped segment.
#
#	[DEfinition]
#	$lnmfac_gc_partit_nw = log ( Prior [partitioning network] ) - \Sum_{$br} { \int_{t(upper_end)}^{t(lower_end)} dt \delta R_{X}^{ID} (s_0^A (b) , s_0^{Root}, t)[gapped segment] }.
#	Here, \delta R_{X}^{ID} (s1, s2, t) =: R_{X}^{ID} (s1, t) - R_{X}^{ID} (s2, t)
#	is the difference of the exit rate (R_{X}^{ID}) between state s1 and state s2.
#	Because we are currently using Dawg's indel model, whose indel rates are space- and time-homogeneous,
#	we have \delta R_{X}^{ID} (s1, s2, t) = \lambda_{ID} * (Length (s1) - Length (s2) ),
#	where \lambda_{ID} is the total indel rate (per site).
#
#
# INPUT: ($topnode_id, \%node2len_ancseq0_in_cl, \%br2attr, \%node2pa, $totrate_indel, $ln_prior_partit_nw), where
#
#	$topnode_id is the ID of the top node (i.e., root);
#	%node2len_ancseq0_in_cl = ($node_id => $length_ancseq0, ...) for the internal nodes,
#	 with $length_ancseq0 being the length of the 'reference' ancestral sequence (determined by the Dollo parsimony principle) within the given gapped segment;
#
#	%br2attr = ($br_id => \@attr, ...) for the branches, with
#	$br_id being the ID of the node at the lower-end of the branch, and
#	$attr[$INDX_LENGTH] giving the length of the branch (in terms of #{substitutions}/site);
#
#	%node2pa = ($node_id => $parent_node_id, ...) for all the nodes except the top node;
#
#	$totrate_indel is the total indel rate (per site);
#
#	$ln_prior_partit_nw is the logarithm of the prior probability of the partitioning network in question.
#
# OUTPUT: $lnmfac_gc_partit_nw, which is the logarithm of the multiplication factor contributed from the partitioning network in question.
#
#
sub calc_lnmfac_gc_partit_nw_for_cl ($\%\%\%$$) {

	my ($topnode_id, $node2len_ancseq0_in_cl, $br2attr, $node2pa, $totrate_indel, $ln_prior_partit_nw) = @_;

    (print STDOUT "\n < Subroutine 'calc_lnmfac_gc_partit_nw_for_cl' (topnode ID: $topnode_id) >\n\n") if ($TEST_LEVEL1a>0);

    if ($TEST_LEVEL1a>2) {
	print STDOUT "\n [ Input Parameters ]\n\n";
	print STDOUT ' $totrate_indel = ',  "$totrate_indel ,\n";
	print STDOUT ' $ln_prior_partit_nw = ', "$ln_prior_partit_nw .\n\n";
	print STDOUT " ( Information on the branches in the tree: )\n\n";
	print STDOUT ' ' , join ("\t", 'Br_ID', 'Upper_end', 'Br_len'), "\n";
	foreach my $br (sort {$a <=> $b} keys %{$br2attr}) {
	    my $attr = $br2attr->{$br};
	    my $br_len = (defined $attr) ? $attr->[$INDX_LENGTH] : 'undef';
	    my $ue = $node2pa->{$br};
	    (defined $ue) or ($ue = 'undef');
	    print STDOUT ' ' , join ("\t", $br, $ue, $br_len), "\n";
	} 
	print STDOUT "\n\n";

	print STDOUT " ( Lengths of the Dollo parsimonious ancestral states within the gapped segment )\n\n";
 	print STDOUT ' ' , join ("\t", 'Node_ID', 'Len(ancseq0)_in_cl'), "\n\n";
	foreach my $node (sort {$a <=> $b} keys %{$node2len_ancseq0_in_cl}) {
	    print STDOUT ' ' , join ("\t", $node, $node2len_ancseq0_in_cl->{$node}), "\n";
	}
	print STDOUT "\n\n";

    } # End of "if ($TEST_LEVEL1a>2) {...}"


   (print STDOUT " [[  Calculation of \$clwise_lnmfac_gc_partit_nw'  ]]\n\n") if ($TEST_LEVEL1a>1);

		# Initialize the output. #

	my $clwise_lnmfac_gc_partit_nw = $ln_prior_partit_nw;

   (print STDOUT ' Initial $clwise_lnmfac_gc_partit_nw = ', $clwise_lnmfac_gc_partit_nw, " .\n\n") if ($TEST_LEVEL1a>2);

	my $len_rootseq0 = $node2len_ancseq0_in_cl->{$topnode_id};
	unless (defined $len_rootseq0) {
	    print STDOUT "[ERROR 'calc_lnmfac_gc_partit_nw_for_cl'] The root (ID: $topnode_id) has UNDEFINED LENGTH in the relevant cluster: $! !\n\n";
	    return undef;
	}

    (print STDOUT ' Len(rootseq0)_in_cl = ', $len_rootseq0, " .\n\n") if ($TEST_LEVEL1a>2);

    (print STDOUT " [ Main loop to calculate the contributions from branches ]\n\n") if ($TEST_LEVEL1a>1);

	foreach my $br (keys %{$br2attr}) { # Main foreach-loop (over branches). #

	    my $ancnode = $node2pa->{$br};
	    unless (defined $ancnode) {
		print STDOUT "[ERROR 'calc_lnmfac_gc_partit_nw_for_cl'] The branch (ID: $br) has UNDEFINED ANCESTRAL NODE: $! !\n\n";
		return undef;
	    }

    (print STDOUT " { Branch ID: $br }\n\n") if ($TEST_LEVEL1a>1);

	    ($ancnode == $topnode_id) and next; # In this case, this branch will obviously contribute zero.

	    my $len_ancseq0 = $node2len_ancseq0_in_cl->{$ancnode};
	    unless (defined $len_ancseq0) {
		print STDOUT "[ERROR 'calc_lnmfac_gc_partit_nw_for_cl'] The ancestral node (ID: $ancnode) of the branch (ID: $br) has UNDEFINED LENGTH in the relevant cluster: $! !\n\n";
		return undef;
	    }

	    my $attr = $br2attr->{$br};
	    unless (defined $attr) {
		print STDOUT "[ERROR 'calc_lnmfac_gc_partit_nw_for_cl'] The branch (ID: $br) has NO DEFINED ATTRIBUTES: $! !\n\n";
		return undef;

	    }
	    my $br_len = $attr->[$INDX_LENGTH];

	    $clwise_lnmfac_gc_partit_nw -= $totrate_indel * $br_len * ($len_ancseq0 - $len_rootseq0);

    if ($TEST_LEVEL1a>2) {
	print STDOUT " | Interim result |\n\n";
	print STDOUT "  Branch length = $br_len ,\n";
	print STDOUT '  Len(ancseq0)_in_cl = ', $len_ancseq0, " ,\n";
	print STDOUT '  Current $clwise_lnmfac_gc_partit_nw = ', $clwise_lnmfac_gc_partit_nw, " .\n\n";
    }

	}			# End of the main foreach-loop (over branches). #

    (print STDOUT ' Final $clwise_lnmfac_gc_partit_nw = ', $clwise_lnmfac_gc_partit_nw, " .\n\n") if ($TEST_LEVEL1a>1);

    (print STDOUT "\n --- End of the subroutine, 'calc_lnmfac_gc_partit_nw_for_cl' ---\n\n") if ($TEST_LEVEL1a>0);

	return $clwise_lnmfac_gc_partit_nw;

}	# End of the subroutine 'calc_lnmfac_gc_partit_nw_for_cl'. #


# mk_node2set_empty_blks:
#	This satellite subroutine of 'apprx1_tot_lnlk_gpattern_msa2_finer'
#	creates a hash tablel associating subject nodes to the set of gap-pattern-blocks that are empty in the specified ancestral gap states at the nodes.
#	(If the node has no empty gap-pattern-block, it will NOT be recorded in the hash table.)
#
#	[CREATED on June 8 (Thu), 2017.]
#
# INPUT: (\@sbj_nodes, \%node2bw_gstats, $char_abs), where
#
#	@sbj_nodes stores the IDs of subject nodes;
#	%node2bw_gstats = ($node_id => \@bw_gstates, ...), where @bw_gtates stores the block-wise gap-states at the node;
#	$char_abs is the character representing the "absence" of residues (i.e., gap).
#
# OUTPUT: \%node2empty_blks, where
#
#	%node2empty_blks = ($node_id => \@empty_blks, ...) for subject nodes with at least one empty block,
#	where @empty_blks stores the IDs of empty blocks.
#
sub mk_node2set_empty_blks (\@\%$) {

	my ($sbj_nodes, $node2bw_gstats, $char_abs) = @_;

	my %node2empty_blks;

	foreach my $node (@{$sbj_nodes}) { # Main foreach-loop (over subject nodes). #

	    my $bw_gstats = $node2bw_gstats->{$node};
	    unless (defined $bw_gstats) {
		print STDOUT "[ERROR 'mk_node2set_empty_blks'] The node (ID: $node) has UNDEFINED block-wise gap-states: $! !\n\n";
		return undef;
	    }

	    my @set_empty_blks = ();
	    my $ct_blks = @{$bw_gstats};
	    for (my $bl=0; $bl<$ct_blks; $bl++) { # Inner for-loop (over blocks). #
		if ($bw_gstats->[$bl] eq $char_abs) { push @set_empty_blks, $bl; }
	    }				# End of the inner for-loop (over blocks). #

	    if (@set_empty_blks>0) { $node2empty_blks{$node} = \@set_empty_blks; }

	}			# End of the main foreach-loop (over subject nodes). #

	return \%node2empty_blks;

}	# End of the subroutine 'mk_node2set_empty_blks'. #


=MOVED to "../MyPerlModules_LOLIPOG/MyTreeMap_indels_ML_hs.pm" in conjunction with IMPORTING modifications of 'apprx1_tot_lnlk_gpattern_msa2_finer' to 'apprx1_tot_lnlk_gpattern_msa2'.

# mk_node2set_empty_blks2:
#	This modified version of 'mk_node2set_empty_blks'
#	records also nodes with NO empty block, provided that it is the lower-end of a branch
#	experiencing (an) indel event(s) in the input indel history.
#	[CRATED on June 8 (Thu), 2017.]
#
# mk_node2set_empty_blks:
#	This satellite subroutine of 'apprx1_tot_lnlk_gpattern_msa2_finer'
#	creates a hash tablel associating subject nodes to the set of gap-pattern-blocks that are empty in the specified ancestral gap states at the nodes.
#	(If the node has no empty gap-pattern-block, it will NOT be recorded in the hash table.)
#
#	[CREATED on June 8 (Thu), 2017.]
#
# INPUT: (\@sbj_nodes, \%node2bw_gstats, \%br2indels, $char_abs), where
#
#	@sbj_nodes stores the IDs of subject nodes;
#	%node2bw_gstats = ($node_id => \@bw_gstates, ...), where @bw_gtates stores the block-wise gap-states at the node;
#	%br2indels = ($branch_id => \@list_indels, ...) for branches experiencing indels.
#	$char_abs is the character representing the "absence" of residues (i.e., gap).
#
# OUTPUT: \%node2empty_blks, where
#
#	%node2empty_blks = ($node_id => \@empty_blks, ...) for subject nodes with at least one empty block (or lower-end nodes of the indel-experiencing branches),
#	where @empty_blks stores the IDs of empty blocks.
#
sub mk_node2set_empty_blks2 (\@\%\%$) {

	my ($sbj_nodes, $node2bw_gstats, $br2indels, $char_abs) = @_;

	my %node2empty_blks;

	foreach my $node (@{$sbj_nodes}) { # Main foreach-loop (over subject nodes). #

	    my $bw_gstats = $node2bw_gstats->{$node};
	    unless (defined $bw_gstats) {
		print STDOUT "[ERROR 'mk_node2set_empty_blks'] The node (ID: $node) has UNDEFINED block-wise gap-states: $! !\n\n";
		return undef;
	    }

	    my @set_empty_blks = ();
	    my $ct_blks = @{$bw_gstats};
	    for (my $bl=0; $bl<$ct_blks; $bl++) { # Inner for-loop (over blocks). #
		if ($bw_gstats->[$bl] eq $char_abs) { push @set_empty_blks, $bl; }
	    }				# End of the inner for-loop (over blocks). #

		# MODIFIED in creation of 'mk_node2set_empty_blks2'. #
		# => Further MODIFIED on June 10 (Sat), 2017. #

	    my $indels = $br2indels->{$node};
	    if (@set_empty_blks>0) {
		$node2empty_blks{$node} = \@set_empty_blks;
	    } elsif ((defined $indels) and (@{$indels}>0)) {
		my $if_indels = 0;
		foreach my $indel (@{$indels}) {
		    if (@{$indel}>0) {
			$if_indels = 1;
			last;
		    }
		}
		$if_indels and ($node2empty_blks{$node} = \@set_empty_blks);
	    }
	    ##if ((@set_empty_blks>0) or ((defined $indels) and (@{$indels}>0)) ) { $node2empty_blks{$node} = \@set_empty_blks; }

	}			# End of the main foreach-loop (over subject nodes). #

	return \%node2empty_blks;

}	# End of the subroutine 'mk_node2set_empty_blks2'. #

=cut # END of "MOVED to "../MyPerlModules_LOLIPOG/MyTreeMap_indels_ML_hs.pm" in conjunction with IMPORTING modifications of 'apprx1_tot_lnlk_gpattern_msa2_finer' to 'apprx1_tot_lnlk_gpattern_msa2'."



# cal_sumint_diff_xrate_in_idb: 
#	This satellite subroutine of 'cal_lnmfac1_indel_hst_in_idb2' calculates the
#	summation of the integrated difference of exit rate coming from a specified indel block, as follows:
#
#	SumInt_diff_Xrate = Sum_{b: branches} \int_{upper_end(b)}^{lower_end(b)} dt \delta R^{ID}_{X} (s(upper_end), s_0(upper_end), t)[indel block] ,
#	where \delta R^{ID}_{X} (s(upper_end), s_0(upper_end))[indel block] is the difference in the exit rate between the gap-state and the reference gap-state (conforming to Dollo parsimony principle) originated from the indel block.
#	(See Eq.(SSA-1.13) in "suppl_addendum.xxx.pdf" for more details.)
#
#	It forms an important constituent of the log-multiplication factor of each indel history (or, more precisely, each set of ancestral gap states) within the indel block.
#
#	[CREATED on June 9th (Fri), 2017.]
#
#	NOTE: Because we assume a space- and time-homogeneous indel history here, we have:
#	\delta R^{ID}_{X} (s(upper_end), s_0(upper_end)) [indel block]
#	   = total_indel_rate (per site) * { Length (s(upper_end)) - Length (s_0(upper_end)) }[indel block].
#
#	Because the Dollo-parsimonious gap states have the minimum residue-occupied sites,
#	the above \delta R^{ID}_{X} is always non-negative,
#	which means it MUST BE ZERO if s_0(upper_end) has NO residue-occupied sites within the indel block.
#
# INPUT: (\@cls_gpblocks, \%node2empty_gpbs0, \%br2attr, \%node2pa, \%node2bw_gstats_in_idb, $char_abs, $totrate_indel), where
#
#	@{$cls_gpblocks[$gpb]} = ($indx_gp, $len_gpb, $start, $end) is
#	for the $gpb th gap-pattern-block in the input gapped segment, with
#	$indx_gp specifying the gap pattern,
#	$len_gpb being the length (in terms of #{columns}) of the block,
#	and $start/$end being the starting/ending coordinate of the block
#	in the 0-based convention;
#
#	%node2empty_gpbs0 = ($node_id => \@empty_gpbs0, ...) for the nodes involved in the indel block,
#	with @empty_gpbs0 listing the indices of the gap-pattern-blocks that are empty at the node; # ADDED in creation of 'cal_lnmfac1_indel_hst_in_idb2'. #
#	%br2attr = ($br_id => \@attr, ...) for the branches, with
#	$br_id being the ID of the node at the lower-end of the branch, and
#	$attr[$INDX_LENGTH] giving the length of the branch (in terms of #{substitutions}/site);
#
#	%node2pa = ($node_id => $parent_node_id, ...) for all the nodes except the top node;
#	$len_msa is the length of the input MSA;
#	$char_abs is the character (or symbol) representing the 'absence' state;
#
#	%br2indels_in_idb = ($branch_id => \@list_indels, ...) is the map of insertions/deletions within the indel block onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})
#	of the blocks involved in the indel event;
#
#	%node2bw_gstats_in_idb = ($node_id => \@bw_gstats_in_idb, ...) is the map of gap states within the indel block onto the
#	nodes, with $blwise_gapstats[$bl] being $char_pres/abs for the $bl th block (in the cluster) at the node $node_id (, or $char_pres if the gap-pattern block (at the node) does not belong to the indel block);
#
#	$totrate_indel is the total rate of insertions & deletions (per site) (relative to the substitution rate) ;
#
# OUTPUT: $sumint_diff_xrate_in_idb, which is the desired summation of the integrated difference of the exit rate.
#
sub cal_sumint_diff_xrate_in_idb (\@\%\%\%\%$$) {

	my ($cls_gpb, $node2empty_gpbs0, $br2attr, $node2pa, $node2bw_gstats_in_idb, $char_abs, $totrate_indel) = @_;

    (print STDOUT ' ' x 3, "<< Subroutine 'cal_sumint_diff_xrate_in_idb' >>\n\n") if ($TEST_LEVEL_csidx > 0);

	my $sumint_diff_xrate_in_idb = 0;

	foreach my $rlv_br (sort {$a <=> $b} keys %{$node2empty_gpbs0}) { # Main foreach-loop (over relevant nodes). #
	    my $ue_node = $node2pa->{$rlv_br};
	    (defined $ue_node) or next; # Skip if it does NOT have the parent(, and thus it is NOT the lower-end of a branch). #

	    my $empty_gpbs0_ue = $node2empty_gpbs0->{$ue_node};
	    ((defined $empty_gpbs0_ue) and (@{$empty_gpbs0_ue}>0)) or next; # Skip if the Dollo-parsimonious state at the upper-end has NO empty gap-pattern-block.

	    my $attr = $br2attr->{$rlv_br};
	    (defined $attr) or next; # Skip if the branch attribute is NOT defined. #
	    my $br_len = $attr->[$INDX_LENGTH];

   (print STDOUT ' ' x 3, "< Branch ID: $rlv_br  (length = $br_len\; upper-end node: $ue_node) >\n\n") if ($TEST_LEVEL_csidx > 1);

	    my $bw_gstats_ue = $node2bw_gstats_in_idb->{$ue_node};

	    my $diff_len = 0;

    (print STDOUT ' ' x 3, join ("\t", 'Indx_gpb', 'diff_len_gpb', 'Tot(diff_len)_curr'), "\n\n") if ($TEST_LEVEL_csidx>3);

	    foreach my $gpb (@{$empty_gpbs0_ue}) { # Inner foreach-loop (over gap-pattern blocks which are empty in the Dollo parsimonious gap state at the upper-end). #

		($bw_gstats_ue->[$gpb] eq $char_abs) and next; # Skip if it is also empty in the subject gap state(at the upper-end).

		my $len_gpb = $cls_gpb->[$gpb][1];

		$diff_len += $len_gpb;

    (print STDOUT ' ' x 3, join ("\t", $gpb, $len_gpb, $diff_len), "\n") if ($TEST_LEVEL_csidx>3);

	    }  				# End of the inner foreach-loop (over gap-pattern blocks which are empty in the Dollo parsimonious gap state at the upper-end). #

    (print STDOUT "\n") if ($TEST_LEVEL_csidx>3);

	    ($diff_len > 0) or next;

	    $sumint_diff_xrate_in_idb += $totrate_indel * $br_len * $diff_len;

    if ($TEST_LEVEL_csidx > 2) {
	print STDOUT ' ' x 3, "Total(diff_len)_final = $diff_len ,\n";
	print STDOUT ' ' x 3, "Sum_Int_diff_Xrate_current = $sumint_diff_xrate_in_idb .\n\n\n";
    }

	}							# End of the main foreach-loop (over relevant nodes). #

    (print STDOUT ' ' x 3, "<< End of the subroutine 'cal_sumint_diff_xrate_in_idb' (Sum_Int_diff_Xrate_final = $sumint_diff_xrate_in_idb) />>\n\n") if ($TEST_LEVEL_csidx>0);

	return $sumint_diff_xrate_in_idb;

}	# End of the subroutine 'cal_sumint_diff_xrate_in_idb'. #


# cal_lnmfac1_gpdseg_br_in_idb2:
#	This satellite subroutine of 'cal_lnmfac1_indel_hst_in_idb2' calculates
#	the summation of the logarithms of the multiplication factors
#	contributed from the gapped segments along a specified branch within a specified indel block.
#	It was created by considerably modifying the predecessor subroutine, 'cal_lnlk1_indel_br_cls2'.
#
#	[CREATION STARTED on June 9 (Fri), 2017, ENDED on June dd (www), 2017.]
#
#
# cal_lnlk1_indel_br_cls2:
#	This subroutine was created by modifying the predecessor, 'cal_lnlk1_indel_br_cls',
#	so that it will fit its master subroutine, 'cal_lnlk1_indel_scn_cls2'.
#	Basically, the subroutine does the same jobs as its predecessor.
#	There are, however, 2 MAJOR DIFFERENCES:
#
#	(1) It only considers the "homology structures" of the pairwise alignments (PWAs) (between the sequences at a node and its parent),
#	which are building blocks of each multiple sequence alignment (MSA) accompanied by a given set of ancestral sequence states.
#	[ For the homology structure, see Lunter et al. (2005) BMC Bioinformatics 6:83.]
#	In consequence, it gave up distinguishing among an insertion followed by a left-overlapping deletion,
#	an insertion followed by a right-overlapping deletion, and a deletion followed by an insertion in between the sites flanking the deletion.
#	It also regards it an error if a run of inserted sites are split by a run of deleted sites, or vise versa, in an ingredient PWA.
#
#	(2) It uses the pre-computed set of multiplication factors, instead of computing each factor on the fly.
#
#	[CREATION STARTED on May 15 (Fri), 2015, ENDED on May xx (yyy), 2015.]
#
#
# cal_lnlk1_indel_br_cls:
#	This satellite subroutine of 'cal_lnlk1_indel_scn_cls'
#	calculates the 1st approximation of the log-likelihood of
#	indel events along a specified branch in a specified gap-block cluster.
#
#
# INPUT: (\@cls_gpblocks, $br_len, \@indels_br, \@blwise_gstats_ue, \@blwise_gstats_le, $char_abs, $totrate_ins, $totrate_del, \@params_boundaries, \@setset_log_mfacs), where
#
#	@{$cls_gpblocks[$gpb]} = ($indx_gp, $len_gpb, $start, $end) is
#	for the $gpb th gap-pattern block in the input gapped segment, with
#	$indx_gp specifying the gap pattern,
#	$len_gpb being the length (in terms of #{columns}) of the block,
#	and $start/$end being the starting/ending coordinate of the block
#	in the 0-based convention;
#
#	$br_len is the length of the specified branch (in the expected number of substitutions per site);
#
#	@{$indels_br[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the specified branch, with @indices_blk listing the indices (in @cls_gpblocks)
#	of the blocks involved in the indel event;
#
#	$blwise_gapstats_ue/le[$bl] is the gap state for the $bl th gap-pattern block (in the specified gapped segment) at the upper-/lower-end node of the specified branch;
#
#	$char_abs is the character (or symbol) representing the 'absence' state;
#
#	$totrate_ins/del is the total rate of insertions/deletions (relative to the substitution rate);
#
#	@params_boundaries = ($cutoff_ins, $cutoff_del, $if_5p_end, $if_3p_end), with
#
#	$cutoff_ins/del being the cut-off length for the insertions/deletions that is common along the MSA,
#
#	$if_5p/3p_end = 1 if the gap-block cluster (i.e., gapped segment) extends into the 5'-/3'-end (=0 otherwise);
#
#	@setset_log_mfacs = (\@log_mfacs_ins, \@log_mfacs_del, \@log_mfacs_pair, \@log_mfacs_ins_bd, \@log_mfacs_del_bd, \@log_mfacs_pair_bd), with
#	$log_mfacs_ins/del[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertion/deletion of $l sites in the bulk of the MSA,
#	$log_mfacs_ins/del_bd[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertions/deletions of $l sites on either boundary of the MSA,
#	$log_mfacs_pair[$l_A]->[$l_D] ($l_A <= $cutoff_del and $l_D <= $cutoff_ins) being the logarithm of the multiplication factor for the pair of an (effective) insertion of $l_D sites and an (effective) deletion of $l_A sites in the bulk of the MSA,
#	$log_mfacs_pair_bd[$l_A]->[$l_D] ($l_A <= $cutoff_del and $l_D <= $cutoff_ins) being the logarithm of the multiplication factor for the pair of an (effective) insertion of $l_D sites and an (effective) deletion of $l_A sites on either boundary of the MSA.
#
#
# [ OUTPUT: ($lnlk1_indel_br_cls, $ct_smpl), where ] --- for the predecessor, 'cal_lnlk1_indel_br_cls2'. 
#
# OUTPUT: ($lnmfac1_gpdseg_br_in_idb, $ct_smpl), where
#
#	$lnmfac1_gpdseg_br_in_idb is the 1st approximation of the summation of the logarithms of the multiplication factors
#	from the gapped segments along the specified branch in the specified indel block;
#
#	$ct_smpl is the number of simplifications that had to be done during the course of this subroutine.
# 
#
##sub cal_lnlk1_indel_br_cls2 (\@$\@\@\@$$$\@\@) {...}
#
sub cal_lnmfac1_gpdseg_br_in_idb2 (\@$\@\@\@$$$\@\@) {

	my ($cls_gpblocks, $br_len, $indels_br, $blwise_gstats_ue, $blwise_gstats_le, $char_abs,
		$totrate_ins, $totrate_del, $params_bds, $setset_log_mfacs) = @_;

	    ## ($lnmfac1_gpdseg_br_in_idb, $ct_smpl) = cal_lnmfac1_gpdseg_br_in_idb2 (@{$cls_gpblks}, $br_len, @indels_br, @{$blwise_gstats_ue}, @{$blwise_gstats_le}, $char_abs, $totrate_ins, $totrate_del, @{$params_bds}, @{$setset_log_mfacs});

    (print STDOUT ' ' x 2, "< Subroutine 'cal_lnmfac1_gpdseg_br_in_idb2' (branch length = $br_len; character for 'absence' = $char_abs) >\n\n") if ($TEST_LEVEL_clmbidb>0);

		# MODIFIED in creation of 'cal_lnlk1_indel_br_cls2'. #

	my ($cutoff_ins, $cutoff_del, $if_5p_end, $if_3p_end) = @{$params_bds};
	my ($log_mfacs_ins, $log_mfacs_del, $log_mfacs_pair, $log_mfacs_ins_bd, $log_mfacs_del_bd, $log_mfacs_pair_bd) = @{$setset_log_mfacs};

	my $ct_gpblks = @{$cls_gpblocks};
	my $ct_indels = @{$indels_br};

    if ($TEST_LEVEL_clmbidb>2) {
	print STDOUT ' ' x 2, "[ Raw Input variables (#{gap-pattern-blocks} = $ct_gpblks; #{indels} = $ct_indels) ]\n\n";

	print STDOUT ' ' x 2, '@{$cls_gpblocks} = ( # [$indx_gp, $len_gpb, $start, $end]', "\n";
	foreach my $gpblock (@{$cls_gpblocks}) {
	    print STDOUT ' ' x 2, '                 [', join (', ', @{$gpblock}), "],\n";
	}
	print STDOUT ' ' x 2, '                  );', "\n\n";

	print STDOUT ' ' x 2, '@{$indels_br} = ( # [$stat_before, $stat_after, indices/involved/blocks]', "\n";
	foreach my $event (@{$indels_br}) {
	    my ($stat_bf, $stat_af, $indices_blks) = @{$event};
	    my $cnct_indices_blks = join ('/', @{$indices_blks});
	    print STDOUT ' ' x 2, '               [', join (', ', $stat_bf, $stat_af, $cnct_indices_blks), "],\n";
	}
	print STDOUT ' ' x 2, '                );', "\n\n";

	print STDOUT ' ' x 2, 'Contents of @{$blwise_gstats_ue} and @{$blwise_gstats_le}:', "\n\n";
	print STDOUT ' ' x 2, join ("\t", 'Indx_blk', 'gstat_upper', 'gstat_lower'), "\n\n";
	for (my $i=0; $i<$ct_gpblks; $i++) { print STDOUT ' ' x 2, join ("\t", $i, $blwise_gstats_ue->[$i], $blwise_gstats_le->[$i]), "\n"; }
	print STDOUT "\n\n";


		# MODIFIED in creation of 'cal_lnlk1_indel_br_cls2'. #

	print STDOUT ' ' x 2, '($totrate_ins, $totrate_del) = ', "($totrate_ins, $totrate_del) ,\n";
	print STDOUT ' ' x 2, '($cutoff_ins, $cutoff_del, $if_5p_end, $if_3p_end)', "\n";
	print STDOUT ' ' x 2, '= (', join (', ', @{$params_bds}), ") .\n\n";

    }
		## Initialize the output ## MODIFIED in creation of 'cal_lnmfac1_gpdseg_br_in_idb2'. ##

	my $lnmfac1_gpdseg_br_in_idb = 0;

=DISABLED in creation of 'cal_lnmfac1_gpdseg_br_in_idb2'. (I)

		## Initialize the output ##

    (print STDOUT ' ' x 2, '[ Initialize the output variable: $lnlk1_indel_br_cls ]', "\n\n") if ($TEST_LEVEL1a>1);

	##my $lnlk1_indel_br_cls = 0;
	my $totlen_init_blks = 0;

	for (my $bl=0; $bl<$ct_gblks; $bl++) { # 1st preliminary for-loop (over gap blocks) #

	    ($blwise_gstats_ue->[$bl] eq $char_abs) and next;

    (print STDOUT ' ' x 2, "( The $bl th block has a 'presence' state at the upper-end )\n\n") if ($TEST_LEVEL1a>2);

	    my ($indx_gp, $len_bl, $start, $end) = @{$cls_gblocks->[$bl]};
	    ##$lnlk1_indel_br_cls -= $len_bl;
	    $totlen_init_blks += $len_bl;

    if ($TEST_LEVEL1a>3) {
	print STDOUT ' ' x 2, "Block length = $len_bl ,\n";
	print STDOUT ' ' x 2, "Total initial block length so far = $totlen_init_blks .\n\n";
    }

	}			# End of the 1st preliminary for-loop (over gap blocks) #

	##$lnlk1_indel_br_cls *= $br_len * ($totrate_ins + $totrate_del);
	my $lnlk1_indel_br_cls = - $br_len * ($totrate_ins + $totrate_del) * $totlen_init_blks;

    if ($TEST_LEVEL1a>3) {
	print STDOUT ' ' x 2, "Total initial block length (final) = $totlen_init_blks ,\n";
	print STDOUT ' ' x 2, 'Branch length = ', $br_len, " ,\n";
	print STDOUT ' ' x 2, '$totrate_ins + $totrate_del = ', $totrate_ins + $totrate_del, " .\n\n";
    }
    (print STDOUT ' ' x 2, 'Initial $lnlk1_indel_br_cls = ', $lnlk1_indel_br_cls, " .\n\n") if ($TEST_LEVEL1a>2);

=cut # END of "DISABLED in creation of 'cal_lnmfac1_gpdseg_br_in_idb2'. (I)"


	unless (@{$indels_br}) {
		# NO indel event is assigned to the branch. # Normally, this will NOT happen. #
	    return ($lnmfac1_gpdseg_br_in_idb, 0);
	}


    (print STDOUT ' ' x 2, "[ Preliminary Process (i): Sort out $ct_indels indel 'events' into single and paired 'events' ]\n\n") if ($TEST_LEVEL1a>1);

		# ADDED in creation of 'cal_lnlk1_indel_br_cls2'. #

	my @std_indels_br = sort {$a->[2][0] <=> $b->[2][0]} @{$indels_br} ;

    if ($TEST_LEVEL_clmbidb>3) {
	print STDOUT ' ' x 2, '[ @{$indels_br} was SORTED !! ]', "\n\n";

	print STDOUT ' ' x 2, '@std_indels_br = ( # [$stat_before, $stat_after, indices/involved/blocks]', "\n";
	for (my $e=0; $e<@std_indels_br; $e++) {

	    my $event  = $std_indels_br[$e];
	    my ($stat_bf, $stat_af, $indices_blks) = @{$event};
	    my $cnct_indices_blks = join ('/', @{$indices_blks});
	    print STDOUT ' ' x 2, "$e th:",  '          [', join (', ', $stat_bf, $stat_af, $cnct_indices_blks), "],\n";
	}
	print STDOUT ' ' x 2, '                );', "\n\n";
    }


		# 1st MAJOR MODIFICATION in creation of 'cal_lnlk1_indel_br_cls2'. #

	# @{$paired_indels[$pr]} = ($indx_ins, $indx_del)
	# specifies the $pr th pair of an 'insertion' and a 'deletion',
	# with $indx_ins/del being the index in @indels_br.
	#
	my ($sgl_indels, $paired_indels) = search_paired_indels2 (@std_indels_br, @{$blwise_gstats_ue}, @{$blwise_gstats_le}, $char_abs);

	unless (defined $sgl_indels) {
	    print STDOUT "[ERROR 'cal_lnmfac1_gpdseg_br_in_idb2'] The subroutine 'search_paired_indels2' FAILED: $! !\n";
	    return (undef, undef);
	}
	my $ct_sgls = @{$sgl_indels};
	my $ct_pairs = @{$paired_indels};

    (print STDOUT "\n", ' ' x 2, "... 'search_paired_indels2' Ended, finding $ct_sgls single 'events' and $ct_pairs pairs of 'events'...\n\n") if ($TEST_LEVEL_clmbidb>1);

    if ($TEST_LEVEL_clmbidb>2) {
	print STDOUT ' ' x 2, '@{$sgl_indels} = (', join (', ', @{$sgl_indels}), ");\n\n";
	print STDOUT ' ' x 2, '@{$paired_indels} = ( # [$indx_ins, $indx_del, $indices/overlapping/blocks]', "\n";
	foreach my $pair (@{$paired_indels}) {
	    my ($indx_ins, $indx_del, $indices_ovlp_blks) = @{$pair};

		# MODIFIED for debugging on June 16 (Fri), 2017. #
	    print STDOUT ' ' x 2, '                    [', join (', ', $indx_ins, $indx_del), '],', "\n";
	    ##my $cnct_indices = join ('/', @{$indices_ovlp_blks}); 
	    ##print STDOUT ' ' x 2, '                    [', join (', ', $indx_ins, $indx_del, $cnct_indices), '],', "\n";
	}
	print STDOUT ' ' x 2, '                    );', "\n\n";
    }


		# NEWLY INTRODUCED in revision on July 22 (Sun), 2012. #

	my $indx_indel2if_rm = {};
	my $indx_indel2if_lm = {};

			# 2nd MAJOR MODIFICATION in creation of 'cal_lnlk1_indel_br_cls'. #
	if ($if_5p_end) {
	    $indx_indel2if_lm = search_leftmost_events2 (@{$blwise_gstats_ue}, @{$blwise_gstats_le}, @std_indels_br, $char_abs, @{$sgl_indels}, @{$paired_indels});

	    unless (defined $indx_indel2if_lm) {
		print STDOUT "[ERROR 'cal_lnmfac1_gpdseg_br_in_idb2'] The subroutine 'search_leftmost_events2' FAILED: $! !\n";
		return (undef, undef);
	    }
    (print STDOUT "\n", ' ' x 2, "... 'search_leftmost_events2' Ended...\n\n") if ($TEST_LEVEL1a>1);

    if ($TEST_LEVEL_clmbidb>2) {
	print STDOUT "\n", ' ' x 2, '[ Contents of %{$indx_indel2if_lm} ]', "\n\n";
	print STDOUT ' ' x 2, "Indices of 'leftmost' indel 'events':  ", join (', ', sort {$a <=> $b} keys %{$indx_indel2if_lm}), " .\n\n";
    }
	}

			# 3rd MAJOR MODIFICATION in creation of 'cal_lnlk1_indel_br_cls'. #
	if ($if_3p_end) {
	    $indx_indel2if_rm = search_rightmost_events2 (@{$blwise_gstats_ue}, @{$blwise_gstats_le}, @std_indels_br, $char_abs, @{$sgl_indels}, @{$paired_indels});

	    unless (defined $indx_indel2if_rm) {
		print STDOUT "[ERROR 'cal_lnmfac1_gpdseg_br_in_idb2'] The subroutine 'search_rightmost_events2' FAILED: $! !\n";
		return (undef, undef);
	    }

    (print STDOUT "\n", ' ' x 2, "... 'search_rightmost_events2' Ended...\n\n") if ($TEST_LEVEL1a>1);
    if ($TEST_LEVEL_clmbidb>2) {
	print STDOUT "\n", ' ' x 2, '[ Contents of %{$indx_indel_if_rm} ]', "\n\n";
	print STDOUT ' ' x 2, "Indices of 'rightmost' indel 'events': ", join (', ', sort {$a <=> $b} keys %{$indx_indel2if_rm}), " .\n\n";
    }
	}

	my $ct_smpl = 0;

		## Main Process 1: calculate the contributions from single indel events. ##

    (print STDOUT "\n", ' ' x 2, "[ Main Process 1: Contributions from $ct_sgls single indel 'events' ]\n\n") if ($TEST_LEVEL_clmbidb>1);

	for (my $sg = 0; $sg < $ct_sgls; $sg++) { # 1st main outer for-loop (over single indel events) #

	    my $indx_id = $sgl_indels->[$sg];

		# MODIFIED in creation of 'cal_lnlk1_indel_br_cls2'. #
	    my $id_event = $std_indels_br[$indx_id];
	    ##my $id_event = $indels_br->[$indx_id];

		# ADDED accompanying the introduction of 'search_left/rightmost_events' (on July 22 (Sun), 2012) #
	    my $if_lm = $indx_indel2if_lm->{$indx_id};
	    my $if_rm = $indx_indel2if_rm->{$indx_id};

    (print STDOUT ' ' x 2, "{ The $sg th event (indel index = $indx_id) }\n\n") if ($TEST_LEVEL_clmbidb>2);
    if ($TEST_LEVEL_clmbidb>3) {
	my ($stat_bf, $stat_af, $indices_blks) = @{$id_event};
	my $cnct_indices_blks = join ('/', @{$indices_blks});
	print STDOUT ' ' x 2, '($stat_before, $stat_after, indices/involved/blocks) = ', "($stat_bf, $stat_af, $cnct_indices_blks)\n\n";
	print STDOUT ' ' x 2, 'Leftmost  ? = ', ($if_lm ? 'Yes' : 'No'), " ,\n"; 
	print STDOUT ' ' x 2, 'Rightmost ? = ', ($if_rm ? 'Yes' : 'No'), " .\n\n";
    }

		# 4th MAJOR MODIFICATION in creation of 'cal_lnlk1_indel_br_cls2'. #

	    my ($lnmfac1_sgl_event_cls, $fl_smpl) = cal_lnlk1_sgl_indel_cls2 (@{$id_event}, @{$cls_gpblocks}, $br_len, $char_abs, $totrate_ins, $totrate_del, @{$params_bds}, @{$setset_log_mfacs}, $if_lm, $if_rm);

	    unless (defined $lnmfac1_sgl_event_cls) {
		print STDOUT "[ERROR 'cal_lnmfac1_gpdseg_br_in_idb2'] The subroutine 'cal_lnlk1_sgl_indel_cls2' FAILED on the $sg th single indel 'event' (= the $indx_id th among all the input 'events'): $! !\n";
		return (undef, undef);
	    }
    (print STDOUT ' ' x 2, " ... 'cal_lnlk1_sgl_indel_cls' Ended ", '($lnmfac1_sgl_event_cls = ', $lnmfac1_sgl_event_cls, ") ...\n\n") if ($TEST_LEVEL_clmbidb>2);
	    if ($fl_smpl) {
		print STDOUT "[WARNING 'cal_lnmfac1_gpdseg_br_in_idb2'] The subroutine 'cal_lnlk1_sgl_indel_cls' was SIMPLIFIED on the $sg th single indel event (= the $indx_id th among all the input events): $! !\n";
		$ct_smpl++ ;
	    }

	    $lnmfac1_gpdseg_br_in_idb += $lnmfac1_sgl_event_cls;
	    ##$lnlk1_indel_br_cls += $lnlk1_sgl_event_cls;

    (print STDOUT ' ' x 2, 'Current $lnmfac1_gpdseg_br_in_idb = ', $lnmfac1_gpdseg_br_in_idb, " .\n\n") if ($TEST_LEVEL_clmbidb>3);


	}			# End of the 1st main outer for-loop (over single indel events) #


    if ($TEST_LEVEL_clmbidb>2) {
	print STDOUT ' ' x 2, 'Interim $lnmfac1_gpdseg_br_in_idb = ', $lnmfac1_gpdseg_br_in_idb, " ,\n";
	print STDOUT ' ' x 2, 'Interim $ct_smpl = ', $ct_smpl, " .\n\n";
    }

		## Main Process 2: calculate the contributions from paired indel 'events'. ##

    (print STDOUT "\n", ' ' x 2, "[ Main Process 2: Contributions from $ct_pairs paired indel 'events' ]\n\n") if ($TEST_LEVEL_clmbidb>1);

	for (my $pr=0; $pr < $ct_pairs; $pr++) { # 2nd main outer for-loop (over paired indel events) #

	    # @{$info_pair} = ($indx_ins, $indx_del)
	    # with $indx_ins/del being the index in @indels_br.
	    my $info_pair = $paired_indels->[$pr];
	    my ($indx_id1, $indx_id2) = @{$info_pair};

    if ($TEST_LEVEL_clmbidb>2) {
        print STDOUT ' ' x 2, "{ The $pr th pair (insertion index = $indx_id1; deletion index = $indx_id2) }\n\n";
    }
    if ($TEST_LEVEL_clmbidb>3) {
	my ($stat_bf_ins, $stat_af_ins, $indices_blks_ins) = @{$indels_br->[$indx_id1]};
	my $cnct_indices_blks_ins = join ('/', @{$indices_blks_ins});
	my ($stat_bf_del, $stat_af_del, $indices_blks_del) = @{$indels_br->[$indx_id2]};
	my $cnct_indices_blks_del = join ('/', @{$indices_blks_del});
	print STDOUT ' ' x 2, join ("\t", 'Ins/Del  ', 'stat_before', 'stat_after', 'indices/involved/blocks'), "\n\n";
	print STDOUT ' ' x 2, join ("\t", 'Insertion', $stat_bf_ins, $stat_af_ins, $cnct_indices_blks_ins), "\n";
	print STDOUT ' ' x 2, join ("\t", 'Deletion ', $stat_bf_del, $stat_af_del, $cnct_indices_blks_del), "\n";
	print STDOUT "\n\n";
    }

			# 5th MAJOR MODIFICATION in creation of 'cal_lnlk1_indel_br_cls'. #

	    my ($lnmfac1_pr_id_cls, $fl_smpl) = cal_lnlk1_pair_indel_cls2 (@{$info_pair}, @std_indels_br, @{$cls_gpblocks}, $br_len, $char_abs, $totrate_ins, $totrate_del, @{$params_bds}, @{$setset_log_mfacs},  %{$indx_indel2if_lm}, %{$indx_indel2if_rm});

	    unless (defined $lnmfac1_pr_id_cls) {
		print STDOUT "[ERROR 'cal_lnmfac1_gpdseg_br_in_idb2'] The subroutine 'cal_lnlk1_pair_indel_cls2' FAILED on the $pr th pair of indel 'events' (= the $indx_id1 th and the $indx_id2 th among all the input 'events'): $! !\n";
		return (undef, undef);
	    }
    (print STDOUT ' ' x 2, " ... 'cal_lnlk1_pair_indel_cls2' Ended ", '($lnmfac1_pr_id_cls = ', $lnmfac1_pr_id_cls, ") ...\n\n") if ($TEST_LEVEL_clmbidb>2);
	    if ($fl_smpl) {
		print STDOUT "[WARNING 'cal_lnmfac1_gpdseg_br_in_idb2'] The subroutine 'cal_lnlk1_pair_indel_cls2' was SIMPLIFIED on the $pr th pair of indel 'events' (= the $indx_id1 th and the $indx_id2 th among all the input 'events'): $! !\n";
		$ct_smpl++ ;
	    }

	    $lnmfac1_gpdseg_br_in_idb += $lnmfac1_pr_id_cls;

    (print STDOUT ' ' x 2, 'Current $lnmfac1_gpdseg_br_in_idb = ', $lnmfac1_gpdseg_br_in_idb, " .\n\n") if ($TEST_LEVEL_clmbidb>3);

	}				# End of the 2nd main outer for-loop (over paired indel events) #


    if ($TEST_LEVEL_clmbidb>2) {
	print STDOUT ' ' x 2, 'Final $lnmfac1_gpdseg_br_in_idb = ', $lnmfac1_gpdseg_br_in_idb, " ,\n"; 
	print STDOUT ' ' x 2, 'Final $ct_smpl = ', $ct_smpl, " .\n\n";
    }

    (print STDOUT "\n", ' ' x 2, "--- Subroutine 'cal_lnmfac1_gpdseg_br_in_idb2' Ended ---\n\n\n") if ($TEST_LEVEL_clmbidb>0);

	return ( $lnmfac1_gpdseg_br_in_idb, $ct_smpl);
	##return ($lnlk1_indel_br_cls, $ct_smpl);

}		# End of the subroutine 'cal_lnmfac1_gpdseg_br_in_idb2'. #



# cal_lnmfac1_indel_hst_in_idb2:
#	This satellite subroutine of 'apprx1_tot_lnlk_gpattern_msa2_finer' calculates the logarithm of the multiplication factor contributed by an indel block.
#	It was created by considerably modifying the predecessor 'cal_lnlk1_indel_scn_cls2' in 'MyTreeMap_indels_ML_hs.pm'.
#
#	[CREATION STARTED on June 8 (Thu), 2017, ENDED on June dd (www), 2017.]
#
#	[Underlying Formula]
#	The main output is:
#	log (multiplication_factor from the indel block)
#	= log {prior probability contribution from the root sequence portion within the indel block}  - SumInt_diff_Xrate + Sum_{\gamma in gapped segments along branches belonging to the indel block} log (multiplication factor (for the ancestor-descendant alignment) from \gamma} 
#
#	SumInt_diff_Xrate = Sum_{b: branches} \int_{upper_end(b)}^{lower_end(b)} dt \delta R^{ID}_{X} (s(upper_end), s_0(upper_end), t)[indel block] ,
#	where \delta R^{ID}_{X} (s(upper_end), s_0(upper_end))[indel block] is the difference in the exit rate between the gap-state and the reference gap-state (conforming to Dollo parsimony principle) originated from the indel block.
#	(See Eqs.(SSA-1.13,14) in "suppl_addendum.xxx.pdf" for more rigorous expression.)
#	
#	
# cal_lnlk1_indel_scn_cls2:
#	This subroutine was created by modifying the predecessor, 'cal_lnlk1_indel_scn_cls',
#	in order to fit its master subroutine, 'apprx1_tot_lnlk_gpattern_msa2'.
#	Basically, the subroutine does the same jobs as its predecessor.
#	There are, however, 2 MAJOR DIFFERENCES:
#
#	(1) It only considers the "homology structures" of the pairwise alignments (PWAs) (between the sequences at a node and its parent),
#	which are building blocks of each multiple sequence alignment (MSA) accompanied by a given set of ancestral sequence states.
#	[ For the homology structure, see Lunter et al. (2005) BMC Bioinformatics 6:83.]
#	In consequence, it gave up distinguishing among an insertion followed by a left-overlapping deletion,
#	an insertion followed by a right-overlapping deletion, and a deletion followed by an insertion in between the sites flanking the deletion.
#	It also regards it an error if a run of inserted sites are split by a run of deleted sites, or vise versa, in an ingredient PWA.
#
#	(2) It uses the pre-computed set of multiplication factors, instead of computing each factor on the fly.
#
#	[CREATION STARTED on May 15 (Fri), 2015, ENDED on May xx (yyy), 2015.]
#
# 
# cal_lnlk1_indel_scn_cls:
#	This satellite subroutine of 'apprx1_tot_lnlk_gpattern_msa' calculates
#	the first approximation of the log-likelihood of an indel scenario
#	in a specified gap-block cluster on a tree.
#
#
# [ INPUT: (\@cls_gblocks, \%br2attr, \%node2pa, \%bl2indels, \%node2blwise_gstats, $char_abs, $totrate_ins, $totrate_del, \@params_boundaries, \%br2log_mfacs), where ] --- For the predecessor, 'cal_lnlk1_indel_scn_cls2'.
#
# INPUT: (\@cls_gpblocks, \%br2attr, \%node2pa, \%br2indels_in_idb, \%node2bw_gstats_in_idb, \%node2empty_gpbs0, $char_abs, $totrate_ins, $totrate_del, \@params_boundaries, \%br2log_mfacs, $delta_log_prior_rootseq_in_idb), where
#
#	@{$cls_gpblocks[$gpb]} = ($indx_gp, $len_gpb, $start, $end) is
#	for the $gpb th gap-pattern-block in the input cluster, with
#	$indx_gp specifying the gap pattern,
#	$len_gpb being the length (in terms of #{columns}) of the block,
#	and $start/$end being the starting/ending coordinate of the block
#	in the 0-based convention;
#
#	%br2attr = ($br_id => \@attr, ...) for the branches, with
#	$br_id being the ID of the node at the lower-end of the branch, and
#	$attr[$INDX_LENGTH] giving the length of the branch (in terms of #{substitutions}/site);
#
#	%node2pa = ($node_id => $parent_node_id, ...) for all the nodes except the top node;
#	$len_msa is the length of the input MSA;
#	$char_abs is the character (or symbol) representing the 'absence' state;
#
#	%br2indels_in_idb = ($branch_id => \@list_indels, ...) is the map of insertions/deletions within the indel block onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})
#	of the blocks involved in the indel event;
#
#	%node2bw_gstats_in_idb = ($node_id => \@bw_gstats_in_idb, ...) is the map of gap states within the indel block onto the
#	nodes, with $blwise_gapstats[$bl] being $char_pres/abs for the $bl th block (in the cluster) at the node $node_id (, or $char_pres if the gap-pattern block (at the node) does not belong to the indel block);
#
#	%node2empty_gpbs0 = ($node_id => \@empty_gpbs0, ...) for the nodes involved in the indel block,
#	with @empty_gpbs0 listing the indices of the gap-pattern-blocks that are empty at the node; # ADDED in creation of 'cal_lnmfac1_indel_hst_in_idb2'. #
#
#	$totrate_ins/del being the total rate of insertions/deletions (per site) (relative to the substitution rate),
#
#	@params_boundaries = ($cutoff_ins, $cutoff_del, $if_5p_end, $if_3p_end), with
#
#	$cutoff_ins/del being the cut-off length for the insertions/deletions that is common along the MSA,
#
#	$if_5p/3p_end = 1 if the gap-block cluster extends into the 5'-/3'-end (=0 otherwise).
#
#	%br2log_mfacs = ($branch_id => \@setset_log_mfacs, ...) assigns the multiplication factors into each branch, with
#	@setset_log_mfacs = (\@log_mfacs_ins, \@log_mfacs_del, \@log_mfacs_pair, \@log_mfacs_ins_bd, \@log_mfacs_del_bd, \@log_mfacs_pair_bd), with
#	$log_mfacs_ins/del[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertion/deletion of $l sites in the bulk of the MSA,
#	$log_mfacs_ins/del_bd[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertions/deletions of $l sites on either boundary of the MSA,
#	$log_mfacs_pair[$l_A]->[$l_D] ($l_A <= $cutoff_del and $l_D <= $cutoff_ins) being the logarithm of the multiplication factor for the pair of an (effective) insertion of $l_D sites and an (effective) deletion of $l_A sites in the bulk of the MSA,
#	$log_mfacs_pair_bd[$l_A]->[$l_D] ($l_A <= $cutoff_del and $l_D <= $cutoff_ins) being the logarithm of the multiplication factor for the pair of an (effective) insertion of $l_D sites and an (effective) deletion of $l_A sites on either boundary of the MSA.
#
#	$delta_log_prior_rootseq_in_idb is the increment of the logarithm of the multiplicative contribution from the indel block to the prior probability of the root sequence state. # ADDED in creation of 'cal_lnmfac1_indel_hst_in_idb2'. #
#
# [ OUTPUT: ($lnlk1_indel_scn_cls, \%br2lnlk1_indel_cls), where ] --- for the predecessor 'cal_lnlk1_indel_scn_cls2'. 
#
# OUTPUT: ($lnmfac1_indel_hst_in_idb, \%br2lnmfac1_gpdseg_in_idb, $sumint_diff_xrate_in_idb), where
#
#	$lnmfac1_indel_hst_in_idb is the 1st approximation of the logarithm of the multiplication factor of the input
#	indel history (or, more precisely, ancestral gap states) in the specified indel block;
#
#	%br2lnmfac1_gpdseg_in_idb = ($br_id => $lnmfac1_gpdseg_br_in_idb, ...) is the 1st approximation of the
#	logarithm of the multiplication factor of the input indel history (more precisely, set of ancestral gap states) along the branch $br_id in the specified indel block.
#
#	$sumint_diff_xrate_in_idb is the summation of the time-integrated exit rate differnces (between the ancestral states and their Dollo parsimonious counterparts) within the specified indel block. # ADDED in creation of 'cal_lnmfac1_indel_hst_in_idb2'. #
#
##sub cal_lnlk1_indel_scn_cls2 (\@\%\%\%\%$$$\@\%) {
##	my ($cls_gblks, $br2attr, $node2pa, $br2indels, $node2blwise_gstats, $char_abs, $totrate_ins, $totrate_del, $params_bds, $br2log_mfacs) = @_;
#
sub cal_lnmfac1_indel_hst_in_idb2 (\@\%\%\%\%\%$$$\@\%$) {

	my ($cls_gpblks, $br2attr, $node2pa, $br2indels_in_idb, $node2bw_gstats_in_idb, $node2empty_gpbs0, $char_abs, $totrate_ins, $totrate_del, $params_bds, $br2log_mfacs, $delta_log_prior_rootseq_in_idb) = @_;

		### my ($lnmfac1_indel_hst_in_idb, $br2lnmfac1_gpdseg_in_idb, $sumint_diff_xrate_in_idb) 
		### = cal_lnmfac1_indel_hst_in_idb2 (@{$cls_gpb}, %{$br2attr}, %{$node2pa}, %{$br2indels_in_idb}, %{$node2blwise_gstats_in_idb}, %{$node2empty_gpbs0}, $char_abs, $totrate_ins, $totrate_del, @params_boundaries, %{$br2log_mfacs}, $delta_log_prior_rootsesq_in_idb);
		###

	my $ct_gpblks = @{$cls_gpblks};
	my @br_ids = sort {$a <=> $b} keys %{$br2attr};
	my $ct_branches = @br_ids;


		# ADDED in creation of 'cal_lnmfac1_indel_hst_in_idb2' (on June 9 (Fri), 2017). #
	my @relevant_nodes = sort {$a <=> $b} keys %{$node2empty_gpbs0};
	my $ct_rlv_nodes = @relevant_nodes;

	my @rlv_brs = sort {$a <=> $b} keys %{$br2indels_in_idb};
	my $ct_rlv_brs = @rlv_brs;

		# END of "ADDED in creation of 'cal_lnmfac1_indel_hst_in_idb2' (on June 9 (Fri), 2017)." #

    (print STDOUT "\n < Subroutine 'cal_lnmfac1_indel_hst_in_idb2' ('Absence' = $char_abs\; #{gap-pattern-blocks in the gapped segment} = $ct_gpblks\; #{branches in the tree} = $ct_branches\; #{relevant_nodes} = $ct_rlv_nodes\; #{indel-experiencing branches} = $ct_rlv_brs) >\n\n") if ($TEST_LEVEL1b>0);

    if ($TEST_LEVEL1b>2) {
	print STDOUT "\n [ Input Parameters ]\n\n";

	print STDOUT ' ($totrate_ins, $totrate_del) = ', "($totrate_ins, $totrate_del);\n";
	print STDOUT ' ($cutoff_ins, $cutoff_del, $if_5p_end, $if_3p_end) = (', join (', ', @{$params_bds}), ");\n\n";

	print STDOUT ' [ Content of @{$cls_gpblks} ]', "\n\n";

	print STDOUT ' ' , join ("\t", 'Indx_blk', 'indx_gpattern', 'length_blk', 'start', 'end'), "\n\n";
	for (my $g=0; $g<$ct_gpblks; $g++) { print STDOUT ' ' , join ("\t", $g, @{$cls_gpblks->[$g]}), "\n"; }
	print STDOUT "\n";

    }

    if ($TEST_LEVEL1b>3) {
	my @nodes = sort {$a <=> $b} keys %{$node2bw_gstats_in_idb};
	my $ct_nodes = @nodes;
	print STDOUT "\n [ There are $ct_nodes nodes ]\n\n";
	print STDOUT ' ' , join ("\t", 'Node_ID', "'{gap-states on the blocks}'"), "\n\n";
	foreach my $node (@nodes) {
	    my $cnct_blwise_gstats = join ('', @{$node2bw_gstats_in_idb->{$node}});
	    print STDOUT ' ' , join ("\t", $node, "'$cnct_blwise_gstats'"), "\n";
	}
	print STDOUT "\n";

	print STDOUT "\n [  There are $ct_branches branches ]\n\n";
	print STDOUT ' ', join ("\t", 'Branch_ID', 'parent_ID', 'branch_length', '#{indels}', '[stat_ue/stat_le/indices,involved,blks]_event1', '[...]_event2', '...'), "\n\n";

	foreach my $br (@br_ids) {
	    my $br_len = $br2attr->{$br}[$INDX_LENGTH];
	    my $pa = $node2pa->{$br};
	    my $list_indels = $br2indels_in_idb->{$br};
	    (defined $list_indels) or ($list_indels = []);
	    my $ct_indels = @{$list_indels};
	    my @out_indels = ();
	    foreach my $indel (@{$list_indels}) {
		my ($stat_ue, $stat_le, $indices_blk) = @{$indel};
		my $cnct_indices_blk = join (',', @{$indices_blk});
		my $out_indel = join ('/', $stat_ue, $stat_le, $cnct_indices_blk);
		push @out_indels, "[$out_indel]";
	    }
	    print STDOUT ' ' , join ("\t", $br, $pa, $br_len, $ct_indels, @out_indels), "\n";
	}
	print STDOUT "\n";

	# @{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})

		# ADDED in creation of 'cal_lnmfac1_indel_hst_in_idb2' (on June 9 (Fri), 2017). #

	print STDOUT "\n [  The following $ct_rlv_brs branches experienced indels: ]\n\n";
	print STDOUT ' ', join (", ", @rlv_brs), "\n\n";

	print STDOUT "\n [  There are $ct_rlv_nodes relevant nodes ]\n\n";
	print STDOUT ' ', join ("\t", 'Node_ID', '[list,empty,gap-pattern-blocks]'), "\n\n";

	foreach my $node (@relevant_nodes) {
		my $cnct_empty_gpbs0 = join (',', @{$node2empty_gpbs0->{$node}});
		print STDOUT ' ' , join ("\t", $node, "[$cnct_empty_gpbs0]"), "\n";
	}
	print STDOUT "\n\n";

	## @relevant_nodes = sort {$a <=> $b} keys %{$node2empty_gpbs0};

    }  # END of "if ($TEST_LEVEL1b>3) {...}"


    (print STDOUT "\n [[  Calculation of \$lnmfac1_indel_hst_in_idb  ]]\n\n") if ($TEST_LEVEL1b>1);

	my $lnmfac1_indel_hst_in_idb = $delta_log_prior_rootseq_in_idb;
	my %br2lnmfac1_gpdseg_in_idb;
	##my $lnlk1_indel_scn_cls = 0;
	##my %br2lnlk1_indel_cls;

    (print STDOUT "\n", ' Initial $lnmfac1_indel_hst_in_idb = ', $lnmfac1_indel_hst_in_idb, " .\n\n") if ($TEST_LEVEL1b>2);

	my $totrate_indel = $totrate_ins + $totrate_del;
	my $sumint_diff_xrate_in_idb = cal_sumint_diff_xrate_in_idb (@{$cls_gpblks}, %{$node2empty_gpbs0}, %{$br2attr}, %{$node2pa}, %{$node2bw_gstats_in_idb}, $char_abs, $totrate_indel);
	unless (defined $sumint_diff_xrate_in_idb) {
	    print STDOUT "[ERROR 'cal_lnmfac1_indel_hst_in_idb2'] The subroutine 'cal_sumint_diff_xrate_in_idb' FAILED: $! !\n\n";
	    return (undef, undef, undef);
	}
    (print STDOUT " ... Subroutine 'cal_sumint_diff_xrate_in_idb' Ended  (with ", '$sumint_diff_xrate_in_idb = ', $sumint_diff_xrate_in_idb, ") ...\n\n") if ($TEST_LEVEL1b>1);

	$lnmfac1_indel_hst_in_idb -= $sumint_diff_xrate_in_idb;

    (print STDOUT "\n", ' Current $lnmfac1_indel_hst_in_idb = ', $lnmfac1_indel_hst_in_idb, " .\n\n") if ($TEST_LEVEL1b>2);

    (print STDOUT "\n [ Main loop to calculate the contributions from indels ]\n\n") if ($TEST_LEVEL1b>1);


	foreach my $br (@rlv_brs) { # Main foreach-loop (over indel-experiencing branches) # MODIFIED in creation of 'cal_lnmfac1_indel_hst_in_idb2' (on June 9 (Fri), 2017). #
	##foreach my $br (@br_ids) { # Main foreach-loop (over branches) #


=MOVED to below the re-collection of indel events.
	    my $attr = $br2attr->{$br};
	    my $br_len = $attr->[$INDX_LENGTH];
	    my ($node_le, $node_ue) = ($br, $node2pa->{$br});
=cut # END of "MOVED to below the re-collection of indel events."

		# Indel events along the branch #

	    my $indels_br = $br2indels_in_idb->{$br}; # MODIFIED in creation of '
	    ##my $indels_br = $br2indels->{$br};
	    my @indels_br = ();
	    if (defined $indels_br) {
			# A measure to prevent a bug in the upstream processes. #
		foreach my $event (@{$indels_br}) { (@{$event}>0) and (push @indels_br, $event); }
	    }
	    my $ct_indels = @indels_br;

	    ($ct_indels>0) or next; # Branches with NO indel will contribute NO multiplication factor.

		# MOVED from above the re-collection of indel events. #
	    my $attr = $br2attr->{$br};
	    my $br_len = $attr->[$INDX_LENGTH];
	    my ($node_le, $node_ue) = ($br, $node2pa->{$br});
		# END of "MOVED from above the re-collection of indel events." #


    (print STDOUT "\n { Branch ID: $br (length = $br_len, upper-end = $node_ue, lower-end = $node_le, #{indel events} = $ct_indels) }\n\n") if ($TEST_LEVEL1b>1);

		# Gap states on the lower-end and upper-end nodes. #

	    my $blwise_gstats_le = $node2bw_gstats_in_idb->{$node_le};
	    my $blwise_gstats_ue = $node2bw_gstats_in_idb->{$node_ue};

    if ($TEST_LEVEL1b>2) {
	print STDOUT "\n ( Gap states on the lower- and upper-ends )\n\n";
	print STDOUT ' ' , join ("\t", 'Block_ID', 'state_upper', 'state_lower'), "\n\n";
	for (my $b=0; $b<$ct_gpblks; $b++) { print STDOUT ' ' , join ("\t", $b, $blwise_gstats_ue->[$b], $blwise_gstats_le->[$b]), "\n"; }

	print STDOUT "\n\n ",  "( There are $ct_indels indel events )\n\n";
	print STDOUT ' ' , join ("\t", 'Indel_indx', 'stat_bf', 'stat_af', 'indices,involved,blocks'), "\n\n";

	for (my $id=0; $id<$ct_indels; $id++) {
	    my ($stat_bf, $stat_af, $indices_involved) = @{$indels_br[$id]};
	    my $cnct_indices = join (',', @{$indices_involved});
	    print STDOUT ' ' , join ("\t", $id, $stat_bf, $stat_af, $cnct_indices), "\n";
	}
	print STDOUT "\n\n";

	# @{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})
	
    } # END of "if ($TEST_LEVEL1b>2) {...}"


		# Logarithmic multiplication factors along the branch. (ADDED in creation of 'cal_lnlk1_indel_scn_cls2'.) #

	   my $setset_log_mfacs = $br2log_mfacs->{$br};


		# [ Calculate the log-likelihood of the input indel scenario along a specified branch.] --- for the predecessor 'cal_lnlk1_indel_scn_cls2'. #
		# => Calculate the logarithm of the multiplication factor contributed from the gapped-segments aloing the specified indel-experiencing branch. (MODIFIED in creation of 'cal_lnmfac1_indel_hst_in_idb2' on June 9 (Fri), 2017.) #

	    my ($lnmfac1_gpdseg_br_in_idb, $ct_smpl) = cal_lnmfac1_gpdseg_br_in_idb2 (@{$cls_gpblks}, $br_len, @indels_br, @{$blwise_gstats_ue}, @{$blwise_gstats_le}, $char_abs, $totrate_ins, $totrate_del, @{$params_bds}, @{$setset_log_mfacs});
	    ## my ($lnlk1_indel_br_cls, $ct_smpl) = cal_lnlk1_indel_br_cls2 (@{$cls_gblks}, $br_len, @{$indels_br}, @{$blwise_gstats_ue}, @{$blwise_gstats_le}, $char_abs, $totrate_ins, $totrate_del, @{$params_bds}, @{$setset_log_mfacs});


	    unless (defined $lnmfac1_gpdseg_br_in_idb) {
		print STDOUT "[ERROR 'cal_lnmfac1_indel_hst_in_idb2'] The subroutine 'cal_lnmfac1_gpdseg_br_in_idb2' FAILED along the branch $br: $! !\n";
		return (undef, undef);
	    }
    (print STDOUT "\n", ' ' x 3, "--- Subroutine 'cal_lnmfac1_gpdseg_br_in_idb2' Ended ...\n\n") if ($TEST_LEVEL1b>1);

	    if ($ct_smpl>0) {
		print STDOUT "[WARNING 'cal_lnmfac1_indel_hst_in_idb2'] The subroutine 'cal_lnmfac1_gpdseg_br_in_idb2' experienced $ct_smpl simplifications ...\n\n";
	    }

		# MODIFIED in creation of 'cal_lnmfac1_indel_hst_in_idb2' (on June 9 (Fri), 2017). #

	    $lnmfac1_indel_hst_in_idb += $lnmfac1_gpdseg_br_in_idb;
	    $br2lnmfac1_gpdseg_in_idb{$br} = $lnmfac1_gpdseg_br_in_idb;
	    ##$lnlk1_indel_scn_cls += $lnlk1_indel_br_cls;
	    ##$br2lnlk1_indel_cls{$br} = $lnlk1_indel_br_cls;

    if ($TEST_LEVEL1b>2) {
	print STDOUT "\n ( Outputs of 'cal_lnmfac1_gpdseg_br_in_idb2' )\n\n"; 
	print STDOUT ' $lnmfac1_gpdseg_br_in_idb = ', $lnmfac1_gpdseg_br_in_idb, ",\n";
	print STDOUT ' $ct_smpl = ', $ct_smpl, " .\n\n";
	print STDOUT ' Current $lnmfac1_indel_hst_in_idb = ', $lnmfac1_indel_hst_in_idb, " .\n\n";
    }

	##}				# End of the main foreach-loop (over branches) #
	}				# End of the main foreach-loop (over indel-experiencing branches) # MODIFIED in creation of 'cal_lnmfac1_indel_hst_in_idb2'. #

    (print STDOUT "\n", ' Final $lnmfac1_indel_hst_in_idb = ' , $lnmfac1_indel_hst_in_idb, " .\n\n") if ($TEST_LEVEL1b>0);

    (print STDOUT "\n --- End of the subroutine 'cal_lnmfac1_indel_hst_in_idb2' ---\n\n") if ($TEST_LEVEL1b>0);

	return ($lnmfac1_indel_hst_in_idb, \%br2lnmfac1_gpdseg_in_idb, $sumint_diff_xrate_in_idb); # MODIFIED in creation of 'cal_lnmfac1_indel_hst_in_idb2'. #
	##return ($lnlk1_indel_scn_cls, \%br2lnlk1_indel_cls);

}		# End of the subroutine 'cal_lnmfac1_indel_hst_in_idb2'. #



	## The Main Subroutine, 'apprx1_tot_lnlk_gpattern_msa2_finer'. ##

# apprx1_tot_lnlk_gpattern_msa2_finer:
#	This subroutine calculate the first approximation of the multiplication factors each of which
#	is contributed from each effectively independent indel block in the input MSA.
#	(The subroutine comprizes STEP 5 in "workflow_for_indel_histories&multiplication_factors.rtf".)
#	It was created by considerably modifying the subroutine, 'apprx1_tot_lnlk_gpattern_msa2', in "MyTreeMap_indels_ML_hs.pm".
#
#	[CREATION STARTED on Feb 25 (Sat), 2017, ENDED on Mmm dd (www), 2017.]
#
#	NOTE added as of June 6th (Tue), 2017:
#	   Like its predecessors, "apprx1_tot_lnlk_gpattern_msa" and "apprx1_tot_lnlk_gpattern_msa2",
# 	   this subroutine ASSUMES that the indel evolutionary model is space- and time-homogeneous,
#	   and that the prior probabilities of the sequences follow a uniform distribution
#	   (, and thus priors will NOT come into the actual calculation of the log multiplication factors).
#
#
#	[MODIFICATION STARTED on Dec 30 (Sat), 2017, ENDED on Mmm dd (www), yyyy. ]
#
#	This subroutine now adds a correction to the contribution from each "indel history"
#	involving either the deletions of contiguous sub-sequences along the two child branches of a bifurcated root,
#	or an insertion along a parent branch (with only one child branch)
#	followed by a deletion of a sub-sequence immediately next to the inserted sub-sequence
#	along the child branch.
#
#	[Minor MODIFICATION on 2019/04/19]:
#	The output now contains the list of correction factors for relevant indel histories.
#	
# [ apprx1_tot_lnlk_gpattern_msa2:
#	This subroutine does the same jobs as its predecessor, 'apprx1_tot_lnlk_gpattern_msa'.
#	The 2 MAJOR DIFFERENCES are the following:
#
#	(1) It only considers the "homology structures" of the pairwise alignments (PWAs) (between the sequences at a node and its parent),
#	which are building blocks of each multiple sequence alignment (MSA) accompanied by a given set of ancestral sequence states.
#	[ For the homology structure, see Lunter et al. (2005) BMC Bioinformatics 6:83.]
#	In consequence, it gave up distinguishing among an insertion followed by a left-overlapping deletion,
#	an insertion followed by a right-overlapping deletion, and a deletion followed by an insertion in between the sites flanking the deletion.
#	It also regards it an error if a run of inserted sites are split by a run of deleted sites, or vise versa, in an ingredient PWA.
#
#	(2) It uses the pre-computed set of multiplication factors, instead of computing each factor on the fly. ] ... Predecessor ...
#	{See the subroutine, "apprx1_tot_lnlk_gpattern_msa2", in "MyTreeMap_indels_ML_hs.pm" for details.}
#
#
# [ INPUT: (\%br2attr, \%node2pa, $len_msa, $char_abs, \@set_cls_gpblocks,  \@sets_eff_indep_indel_blocks, \@setssets_psm_hsts_in_indel_blks, $totrate_ins, $totrate_del, $cutoff_ins, $cutoff_del, $delta_del, \%br2log_mfacs), where ] ... OLD version (, which was effective until June 6 (Tue), 2017.)
#
# INPUT: ($topnode_id, \%br2attr, \%node2pa, \%node2ch, $len_msa, $char_abs, \@set_cls_gpblocks,  \@psm_hsts0_for_cls,  \@sets_psm_hsts0_in_indel_blks, \@setssets_psm_hsts_in_indel_blks, $totrate_ins, $totrate_del, $cutoff_ins, $cutoff_del, $delta_del, \%br2log_mfacs), where
#
#	$topnode_id is the ID of the top node (i.e., root) of the tree;
#
#	%br2attr = ($br_id => \@attr, ...) for the branches, with
#	$br_id being the ID of the node at the lower-end of the branch, and
#	$attr[$INDX_LENGTH] giving the length of the branch (in terms of #{substitutions}/site);
#
#	%node2pa = ($node_id => $parent_node_id, ...) for all the nodes except the top node;
#	%node2ch = ($node_id => \@child_node_ids, ...) for all the internal nodes;
#	$len_msa is the length of the input MSA;
#	$char_abs is the character (or symbol) representing the 'absence' state;
#
#	@{$set_cls_gblocks[$cl]->[$gpb]} = ($indx_gp, $len_gpb, $start, $end) is
#	for the $gpb th gap-pattern block in the $cl th cluster, with
#	$indx_gp specifying the gap pattern,
#	$len_gpb being the length (in terms of #{columns}) of the gap-pattern block,
#	and $start/$end being the starting/ending coordinate of the gap-pattern block
#	in the 0-based convention;
#
#	[ @{$sets_eff_indep_indel_blocks[$cl]->[$eiidb]} = ($br, $up_or_lw, $start, $end, \@invlvd_gpbs, \@invlvd_gbs, \@invlvd_sbs) [<= Is this format really general enough to work on any types of indel blocks?]
#	gives information on the $eiidb th effectively independent indel block in the $cl th gapped segment.
#
#	$br is the ID of the branch delimiting the indel block,
#	$up_or_lw = 'U'/'L' if the upper/lower-end of $br is the indel block, 
#	$start/$end is the start/end coordinate of the boundary,
#	@invlvd_gpbs lists the gap-pattern-blocks involved,
#	@invlvd_gbs lists the gap-blocks involved,
#	@invlvd_sbs is the sequence-blocks involved; ] ... for OLD VERSION (, which was effective until June 6th (Tue), 2017.)
#
#		[NEWLY ADDED on June 6th (Tue), 2017.]
#
#	@{$psm_hsts0_for_cls[$cl]} = (\%br2indels0, \%node2blwise_gstat0) is for the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the input MSA, where 
#	%br2indels0 = ($branch_id => \@list_indels0, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels0[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved), and where
#	%node2blwise_gstat0 = ($node_id => \@blwise_gapstats0, ...) is the map of the gap states onto the nodes,
#	 with $blwise_gapstats0[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node;
#
#	@{$sets_psm_hsts0_in_indel_blks[$cl]->[$eiidb]} = (\%br2indels0, \%node2blwise_gstats0) is for the initial candidate indel history realizing the $eiidb th effectively indepenent indel block in the $cl th cluster (i.e., gapped segment);
#
#		[END of "NEWLY ADDED on June 6th (Tue), 2017."]
#
#	@{$setssets_psm_hsts_in_indel_blks[$cl]->[$eiidb]->[$ph]} = (\%br2indels, \%node2blwise_gapstat)
#	is for the $ph th of the parsimonious histories realizing the $eiidb th effectively independent indel block in the $cl th cluster, where
#
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_gpb) describing the $id th indel event at the branch,
#	with @indices_gpb listing the indices (in @{$set_cls_gpblocks[$cl]}) of the gap-pattern blocks involved in the indel event,
#
#	%node2blwise_gapstat = ($node_id => \@blwise_gapstats, ...) is the map of gap states onto the interior nodes, 
#	with $blwise_gapstats[$gpb] being $char_pres/abs for the $gpb th gap-pattern block (in the cluster) at the node $node_id;
#
#	NOTE that @{$setsets_psm_hsts_in_indel_blks[$cl]->[$eiidb]->[$ph]} = () indicates that the lieklihood of the parsimonious scenario cannot be calculated due to some abnormality.
#
#	[ @{$set_psm_scns_for_cls[$cl]->[$ps]} = (\%br2indels, \%node2blwise_gapstat) is for the $ps th of the parsimonious scenarios realizing the $cl th cluster, where
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_blk) describing the $id th indel event at the branch, with @indices_blk listing the indices (in @blocks_in_cls = @{$set_cls_gblocks[$cl]})
#	of the blocks involved in the indel event,
#	%node2blwise_gapstat = ($node_id => \@blwise_gapstats, ...) is the map of gap states onto the
#	interior nodes, with $blwise_gapstats[$bl] being $char_pres/abs for the $bl th block (in the cluster) at the node $node_id;
#
#	NOTE that @{$set_psm_scns_for_cls[$cl]->[$ps]} = () indicates that the likelihood of the parsimonious scenario cannot be calculated due to some abnormality. (ADDED on Mar 2 (Sun), 2014) ] ... For the Predecessor, "apprx1_tot_lnlk_gpattern_msa2" ...
#
#	$totrate_ins/del is the total rate of insertions/deletions (relative to the substitution rate),
#
#	$cutoff_ins/del is the cut-off length for insertions/deletions, which is common along the MSA;
#
#	$delta_del = Sum_{$l=1}^{$cutoff_del}($l-1)*distr_del ($l)/$nfac_del,
#	which can be calculated via 'cal_delta_exp_del_5p', is the excessive exponential factor
#	due to deletions extending out of the MSA boundaries;
#
#	%br2log_mfacs = ($branch_id => \@setset_log_mfacs, ...) assigns the multiplication factors into each branch, with
#	@setset_log_mfacs = (\@log_mfacs_ins, \@log_mfacs_del, \@log_mfacs_pair, \@log_mfacs_ins_bd, \@log_mfacs_del_bd, \@log_mfacs_pair_bd), with
#	$log_mfacs_ins/del[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertion/deletion of $l sites in the bulk of the MSA,
#	$log_mfacs_ins/del_bd[$l] ($l <= $cutoff_ins/del) being the logarithm of the multiplication factor for the single (effective) insertions/deletions of $l sites on either boundary of the MSA,
#	$log_mfacs_pair[$l_A]->[$l_D] ($l_A <= $cutoff_del and $l_D <= $cutoff_ins) being the logarithm of the multiplication factor for the pair of an (effective) insertion of $l_D sites and an (effective) deletion of $l_A sites in the bulk of the MSA,
#	$log_mfacs_pair_bd[$l_A]->[$l_D] ($l_A <= $cutoff_del and $l_D <= $cutoff_ins) being the logarithm of the multiplication factor for the pair of an (effective) insertion of $l_D sites and an (effective) deletion of $l_A sites on either boundary of the MSA.
#
#
# [ OUTPUT: ($tot_lnlk_gc_msa, $shared_fac1p, $shared_fac2p, \@set_clwise_tot_lnmfac_gc, \@set_clwise_lnmfac_gc_partit_nw, \@setsets_tot_lnmfac_gc_indel_blk, 
#	    $tot_div_tree, $totlen_ng, $totlen_rootseq0, \@len_rootseq0_in_cls, \@set_node2len_ancseq0_in_cls, \@setsetssets_lnmfac_gc_indel_blk_hst), among which  ] ... OBSOLETE as of 2019/04/19.
#
# OUTPUT: ($tot_lnlk_gc_msa, $shared_fac1p, $shared_fac2p, \@set_clwise_tot_lnmfac_gc, \@set_clwise_lnmfac_gc_partit_nw, \@setsets_tot_lnmfac_gc_indel_blk, 
#	    $tot_div_tree, $totlen_ng, $totlen_rootseq0, \@len_rootseq0_in_cls, \@set_node2len_ancseq0_in_cls, \@setsetssets_lnmfac_gc_indel_blk_hst, \@setsets_indel_hst2corr_fac), among which 
#
#	MAIN Outputs are ($tot_lnlk_gc_msa, $shared_fac1p, $shared_fac2p, \@set_clwise_tot_lnmfac_gc, \@setsets_tot_lnmfac_gc_indel_blk), where
#
#	$tot_lnlk_gc_msa is the total log-likelihood of the gap-configuration of the input MSA in the 1st approximation
#	(= 'N/A' if any of the parsimonious histories contain such abnormalities that hamper the calculation of the likel
#ihoods);
#
#	$shared_fac1p = - $tot_div_tree * ($totrate_del * $delta_del_5p + $totrate_ins);
#	$shared_fac2p = - ($totrate_ins+$totrate_del) * $tot_div_tree * $totlen_rootseq0;
#
#	$set_clwise_tot_lnmfac_gc[$i] is the total log-multiplication factor from the gap configuration in the ($i+1) th gapped segment, 
#	which is: $set_clwise_lnmfac_gc_partit_nw[$i] + Sum_{$eiidb} $setsets_tot_lnmfac_gc_indel_blk[$i]->[$eiidb]
#	(= 'N/A' if any of the parsimonious histories in the cluster contain such abnormalities that hamper the calculation of the multiplication factors);
#
#	$set_clwise_lnmfac_gc_partit_nw[$i] is the log-multiplication factor from the gap-configuration of the "partitioning" network (, which is determined exclusively from the Dollo parsiomonious ancestral gap-states,) of the ($i+1) th gapped segment;
#
#	$setsets_tot_lnmfac_gc_indel_blk[$i]->[$eiidb] is the total log-multiplication factor from the gap configuration in the $eiidb th effectively independent indel block of the ($i+1) th gapped segment,
#	which is log (Sum_{$hst} exp ($setsetsets_lnmfac_gc_indel_blk_hst[$i]->[$eiidb]->[$hst]))
#	(= 'N/A' if any of the parsimonious histories in the cluster contain such abnormalities that hamper the calculation of the multiplication factors);
#
#
#	AUXILIARY Outputs are ($tot_div_tree, $totlen_ng, $totlen_rootseq0, \@len_rootseq0_in_cls, \@set_node2len_ancseq0_in_cls, \@setsetsets_lnmfac_gc_indel_blk_hst), where
#
#	$tot_div_tree is the total divergence of the tree, which is the summation of the branch lengths over branches in the entire tree;
#
#	$totlen_ng is the total number of gapless columns (i.e., the total length over the gapless segments);
#
#	$totlen_rootseq0 is the total length of the "reference" root sequence across the MSA.
#	As a "reference" root sequence, this subroutine uses the root sequence under the Dollo parsimonious indel history;
#
#	$len_rootseq0_in_cls[$i] is the length of the "reference" root sequence in the ($i+1) th gapped segment;
#
#	%{$set_node2len_ancseq0_in_cls[$i]} = ($node_id => $len_ancseq0_in_cls, ...) gives the lengths of the "reference" ancestral states at node $node_id in the ($i+1) th gapped segment.
#	As a "reference" set of ancestral states, this subroutine uses the ancestral states in the Dollo parsimonious indel history;
#
#	$setsetssets_lnmfac_gc_indel_blk_hst[$i]->[$eiidb]->[$hst] is the log-multiplication factor from the $hst th indel history
#	that parsimoniously give the $eiidb th (effectively independent) indel block in the ($i+1) th gapped segment
#	(= 'N/A' if the parsimonious history contains such abnormalities that hamper the calculation of the likelihood ).
#
#	%{$setsets_rlv_indel_hst2corr_fac[$i]->[$eiidb]} = ($indx_indel_hst => $corr_fac) for all indel histories  with NON-ZERO correction factors in the $eiidb th (effectively independent) indel block in the ($i+1) th gapped segments; # ADDED on 2019/04/19.
#
#
#	[NOTE1] Although the document may count the indel blocks as $eiidb = 1, 2, ... (1-based),
#            the program implementation here counts the indel blocks as $eiidb = 0, 1, 2, ... (0-based).
#
#	[NOTE2] Two different series of indices, referred to as $cl and $i, are used to specify the gapped segments.
#               By examining the predecessor subroutine, "apprx1_tot_lnlk_gpattern_msa2" in "MyTreeMap_indels_ML_hs.pm",
#		I CONFIRMED that the "$cl th gapped segment" is actually identical to the "($i+1) th gapped segment" (when $cl = $i).
#		(That is, {$i +1} count the gapped segments in a 1-based way, and {$cl} count them in a 0-based way.)
#
sub apprx1_tot_lnlk_gpattern_msa2_finer ($\%\%\%$$\@\@\@\@$$$$$\%) {

	my ($topnode_id, $br2attr, $node2pa, $node2ch, $len_msa, $char_abs, $set_cls_gpblocks,  $psm_hsts0_for_cls, $sets_psm_hsts0_in_indel_blks,
	   $setssets_psm_hsts_in_indel_blks, $totrate_ins, $totrate_del, $cutoff_ins, $cutoff_del, $delta_del, $br2log_mfacs) = @_;

	my $ct_cls = @{$set_cls_gpblocks};
	my @set_int_nodes = (sort {$a <=> $b} keys %{$node2ch});
	my @set_all_nodes = (sort {$a <=> $b} ($topnode_id, keys %{$node2pa}) ); # ADDED on June 8 (Thu), 2017. #


	### MODIFY the Predecessor, "apprx1_tot_lnlk_gpattern_msa2", in "MyTreeMap_indels_ML_hs.pm". ###

    (print STDOUT "\n<< Subroutine 'apprx1_tot_lnlk_gpattern_msa2_finer' (top_node_ID = $topnode_id, 'Absence' = $char_abs, #{gapped segments} = $ct_cls, length(MSA) = $len_msa) >>\n\n") if ($TEST_LEVEL>0);

    if ($TEST_LEVEL>2) {

	print STDOUT "[[ Primary Input Data ]]\n\n";

	print STDOUT '($cutoff_ins, $cutoff_del) = ', "($cutoff_ins, $cutoff_del) .\n\n";
	print STDOUT '($totrate_ins, $totrate_del) = ', "($totrate_ins, $totrate_del) ,\n";
	print STDOUT '$delta_del = ', $delta_del, " .\n";
	
	print STDOUT "\n\n";
    } # END of "if ($TEST_LEVEL>2) {...}"

    if ($TEST_LEVEL>3) {
	print STDOUT "[ Input tree (top node: $topnode_id) ]\n\n";
	print STDOUT join ("\t", 'Branch_ID', 'parent_ID', 'branch_length'), "\n\n";
	foreach my $br (sort {$a <=> $b} keys %{$br2attr}) {
	     my $len = $br2attr->{$br}[$INDX_LENGTH];
	     my $pa_id = $node2pa->{$br};
	     print STDOUT join ("\t", $br, $pa_id, $len), "\n";
	}
	print STDOUT "\n\n";

	print STDOUT "Internal nodes: ", join (', ', @set_int_nodes), " .\n\n\n"; # ADDED in creation of 'apprx1_tot_lnlk_gpattern_msa2_finer'. #
	print STDOUT "ALL nodes: ", join (', ', @set_all_nodes), " .\n\n\n"; # ADDED in creation of 'apprx1_tot_lnlk_gpattern_msa2_finer'. #

	print STDOUT "[ There are $ct_cls gapped segments ]\n\n";
	for (my $cl=0; $cl<$ct_cls; $cl++) {

	    my $cls_gpblks = $set_cls_gpblocks->[$cl];
	    my $ct_gpblks = @{$cls_gpblks};

	    print STDOUT "{ The $cl th gapped segment (#{blocks} = $ct_gpblks) }\n\n";
	    print STDOUT join ("\t", 'Indx_blk', 'indx_gpattern', 'length_blk', 'start', 'end'), "\n\n";

	    for (my $bl=0; $bl<$ct_gpblks; $bl++) { print STDOUT join ("\t", $bl, @{$cls_gpblks->[$bl]}), "\n"; }
	    print STDOUT "\n\n";
	}

	# @{$set_cls_gpblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end) is

    } # END of "if ($TEST_LEVEL>3) {...}"


		## Examine whether the tree contains a bifurcated root and/or (a) uni-furcated node(s). ##
		## (ADDED in MODIFICATION from Dec 30, 2017 till Jan 2, 2018.)

	# @{$info_biroot} = ($id_root, $id_child1, $id_child2) for the bifurcated root,
	#	      = () if the root is NOT bifurcated;
	# @{$list_uninodes[$u]} = ($id_parent, $id_uninode, $id_child) for the $u th unifurcated node (other than the root),
	# $id_uninode, whose parent and the sole child are $id_parent and $id_child, respectively.
	#
	my ($info_biroot, $list_uninodes) = search_biroot_uninodes ($topnode_id, %{$node2pa}, %{$node2ch});
	unless ((defined $info_biroot) and (defined $list_uninodes)) {
	    print STDOUT "[ERROR 'apprx1_tot_lnlk_gpattern_msa2_finer'] The subroutine 'search_biroot_uninodes' FAILED: $! !\n";
	    return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	    
	}

    if ($TEST_LEVEL_corr>1) {
	print STDOUT "[ Output of 'search_biroot_uninodes' ]\n\n";
	if (@{$info_biroot}>0) {
	    print STDOUT 'Bifurcated root: (Root, $child1, $child2) = (', join (', ', @{$info_biroot}), ").\n\n\n";
	}
	my $ct_uninodes = @{$list_uninodes};
	if ($ct_uninodes>0) {
	    print STDOUT "{ There are $ct_uninodes uni-furcated nodes }\n\n";
	    print STDOUT join ("\t", 'No.', 'ID_parent', 'ID_uninode', 'ID_child'), "\n\n";
	    for (my $u = 0; $u < $ct_uninodes; $u++) {
		print STDOUT join ("\t", $u, @{$list_uninodes->[$u]}), "\n";
	    }
	    print STDOUT "\n\n";
	}
    }
			### RESTART FROM HERE (1) ###

		## END of "(ADDED in MODIFICATION from Dec 30, 2017 till Jan 2, 2018.)"


		## Prepare auxiliary variables. ##

    if ($TEST_LEVEL>3) {
	print STDOUT '[ Calculating $tot_div_tree ]', "\n\n";
	print STDOUT join ("\t", 'Branch_ID', 'br_len', '$tot_div_tree(current)'), "\n\n";
    }

	my $tot_div_tree = 0;
	foreach my $br (keys %{$br2attr}) { # 1st preliminary foreach-loop #
	    my $attr = $br2attr->{$br};
	    unless (defined $attr) {
		print STDOUT "[ERROR 'apprx1_tot_lnlk_gpattern_msa2_finer'] The branch '$br' has NO ATTRIBUTE : $! !\n";
		return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	    }
	    my $br_len = $attr->[$INDX_LENGTH];
	    $tot_div_tree += $br_len;

    (print STDOUT join ("\t", $br, $br_len, $tot_div_tree), "\n") if ($TEST_LEVEL>3);

	}			# End of the 1st preliminary foreach-loop #

    if ($TEST_LEVEL>3) {
	print STDOUT "\n\n";
	print STDOUT '[ Calculating $totct_gpsites ]', "\n\n";
	print STDOUT join ("\t", 'Cluster_ID', 'beg_cls', 'end_cls', 'len_cls', '$totct_gpsites(current)'), "\n\n";
    }

	my $totct_gpsites = 0;
	for (my $cl=0; $cl<$ct_cls; $cl++) { # 2nd preliminary for-loop #
	    #  @{$set_cls_gpblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end)
	    my $cls_gpb = $set_cls_gpblocks->[$cl];
	    my $first_gpb = $cls_gpb->[0];
	    my $last_gpb = $cls_gpb->[$#{$cls_gpb}];
	    my $beg_cls = $first_gpb->[2];
	    my $end_cls = $last_gpb->[3];
	    my $len_cls = $end_cls - $beg_cls + 1;
	    $totct_gpsites += $len_cls;

    (print STDOUT join ("\t", $cl, $beg_cls, $end_cls, $len_cls, $totct_gpsites), "\n") if ($TEST_LEVEL>3);

	}			# End of the 2nd preliminary for-loop #

    (print STDOUT "\n\n") if ($TEST_LEVEL>3);

	my $totlen_ng = $len_msa - $totct_gpsites;


		## NEWLY ADDED in creation of 'apprx1_tot_lnlk_gpattern_msa2_finer' (I) (on June 7th (Wed), 2017). ##

    if ($TEST_LEVEL>3) {
	print STDOUT "\n";
	print STDOUT '[ Calculating $totlen_rootseq0 and @set_node2len_ancseq0_in_cls ]', "\n\n";
	print STDOUT join ("\t", 'Cluster_ID', 'beg_cls', 'end_cls', 'len_cls', '$totct_gpsites(current)'), "\n\n";
    }

	my $totlen_rootseq0 = $totlen_ng;
	my @len_rootseq0_in_cls = my @set_node2len_ancseq0_in_cls = ();

	for (my $cl=0; $cl<$ct_cls; $cl++) { # 3rd preliminary for-loop (over gapped segments). #

	    my $gpbs_in_cls = $set_cls_gpblocks->[$cl];
	    my $ct_gpbs = @{$gpbs_in_cls};

	    my $node2bw_gstats0 = $psm_hsts0_for_cls->[$cl][1];

#	@{$psm_hsts0_for_cls[$cl]} = (\%br2indels0, \%node2blwise_gstat0) is for the initial candidate of the parsimonious scenario resulting in the $cl th cluster (i.e., gapped segment) in the input MSA, where 
#	%node2blwise_gstat0 = ($node_id => \@blwise_gapstats0, ...) is the map of the gap states onto the nodes,
#	 with $blwise_gapstats0[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node;

	    my %node2len_ancseq0_in_cl;

	    foreach my $node (@set_int_nodes) { # Middle foreach-loop (over internal nodes). #

		my $bw_gstats0 = $node2bw_gstats0->{$node};

		my $len_ancseq0_in_cl = 0;

		for (my $gpb = 0; $gpb < $ct_gpbs; $gpb++) { # Inner for-loop (over gap-pattern blocks). #

		    if ($bw_gstats0->[$gpb] ne $char_abs) { # The block is filled with residues. #

			my $len_gpb = $gpbs_in_cls->[$gpb][1];
	    			#  @{$set_cls_gpblocks[$cl]->[$gb]} = ($indx_gp, $len_gb, $start, $end)

			$len_ancseq0_in_cl += $len_gpb;
		    }
		}				# End of the inner for-loop (over gap-pattern blocks). #

		$node2len_ancseq0_in_cl{$node} = $len_ancseq0_in_cl;

		if ($node == $topnode_id) {
		    $totlen_rootseq0 += $len_ancseq0_in_cl;
		    $len_rootseq0_in_cls[$cl] = $len_ancseq0_in_cl;
		}

	    }				# End of the middle foreach-loop (over internal nodes). #

	    $set_node2len_ancseq0_in_cls[$cl] = \%node2len_ancseq0_in_cl;

	}			# End of the 3rd preliminary for-loop (over gapped segments). #

    (print STDOUT "\n\n") if ($TEST_LEVEL>3);


		## END of "NEWLY ADDED in creation of 'apprx1_tot_lnlk_gpattern_msa2_finer' (I) (on June 7th (Wed), 2017)." ##


		# MODIFIED in creation of 'apprx1_tot_lnlk_gpattern_msa2_finer'. #
	my $totrate_indel = $totrate_ins + $totrate_del;
	my $shared_fac1p = - $tot_div_tree * ($totrate_del * $delta_del + $totrate_ins);
	my $shared_fac2p = - $totrate_indel * $tot_div_tree * $totlen_rootseq0;
	##my $shared_fac1 = - $tot_div_tree * ($totrate_del * $delta_del + $totrate_ins);
	##my $shared_fac2 = - ($totrate_ins+$totrate_del) * $tot_div_tree * $totlen_ng;

    if ($TEST_LEVEL>2) {
	print STDOUT "[[ Secondary Input Data ]]\n\n";
	print STDOUT '$tot_div_tree = ', $tot_div_tree, " .\n\n";
	print STDOUT '$len_msa = ', $len_msa, " ,\n";
	print STDOUT '$totct_gpsites = ', $totct_gpsites, " ,\n";
	print STDOUT '$totlen_ng = ', $totlen_ng, " .\n";
	print STDOUT '$totlen_rootseq0 = ', $totlen_rootseq0, " .\n\n";

		# MODIFIED in creation of 'apprx1_tot_lnlk_gpattern_msa2_finer'. #
	print STDOUT '$totrate_indel (= $totrate_ins + $totrate_del) = ', $totrate_indel, " .\n";
	print STDOUT '$shared_fac1p (= - $tot_div_tree * ($totrate_del * $delta_del + $totrate_ins) = ', $shared_fac1p, " ,\n";
	print STDOUT '$shared_fac2p (= - $totrate_indel * $tot_div_tree * $totlen_rootseq0) = ' , $shared_fac2p, " .\n\n";
	##print STDOUT '$shared_fac1 (= - $tot_div_tree * ($totrate_del * $delta_del + $totrate_ins) = ', $shared_fac1, " ,\n";
	##print STDOUT '$shared_fac2 (= - ($totrate_ins + $totrate_del) * $tot_div_tree * $totlen_ng) = ' , $shared_fac2, " .\n\n";

	if ($TEST_LEVEL>3) {
	    print STDOUT "\n", '[ (Auxiliary:) Content of @set_node2len_ancseq0_in_cls ]', "\n\n";
	    print STDOUT join ("\t", 'Indx_cl', @set_int_nodes), "\n\n";
	    for (my $cl = 0; $cl < $ct_cls; $cl++) {
		my $node2len_ancseq0_in_cl = $set_node2len_ancseq0_in_cls[$cl];
		my @lens_ancseq0_in_cl = ();
		foreach my $node (@set_int_nodes) { push @lens_ancseq0_in_cl, $node2len_ancseq0_in_cl->{$node}; }
		print STDOUT join ("\t", $cl, @lens_ancseq0_in_cl), "\n";
	    }
	    print STDOUT "\n\n";
	} # END of "if ($TEST_LEVEL>3) {...}"

    } # END of "if ($TEST_LEVEL>2) {...}"


	   ## MAIN PROCESS to calculate log-likelihood for each gap-block cluster. ##

    (print STDOUT "\n[[ Main Process ]]\n\n") if ($TEST_LEVEL>1);


	my $tot_lnlk_gc_msa = $shared_fac1p + $shared_fac2p;
	my @set_clwise_tot_lnmfac_gc = ();
	my @set_clwise_lnmfac_gc_partit_nw = my @setsets_tot_lnmfac_gc_indel_blk = ();
	my @setsetssets_lnmfac_gc_indel_blk_hst = ();

	my @setsets_rlv_indel_hst2corr_fac = (); # ADDED in minor modification (on 2019/04/19).

    (print STDOUT 'Initial $tot_lnlk_gc_msa (= $shared_fac1p + $shared_fac2p) = ', $tot_lnlk_gc_msa, " .\n\n") if ($TEST_LEVEL>2);

		## NEWLY ADDED in creation of 'apprx1_tot_lnlk_gpattern_msa2_finer' (II) (on June 7th (Wed), 2017). ##


	for (my $cl=0; $cl<$ct_cls; $cl++) { # Main outer for-loop (over gap-block clusters, i.e., gapped segments) #

	    #  @{$set_cls_gpblocks[$cl]->[$gpb]} = ($indx_gp, $len_gb, $start, $end)

	    my $cls_gpb = $set_cls_gpblocks->[$cl];
	    my $ct_gpblocks = @{$cls_gpb};

    (print STDOUT "\n[ The $cl th gapped segment (#{gap-pattern-blocks} = $ct_gpblocks) ]\n\n") if ($TEST_LEVEL>1);

	    my $first_gpb = $cls_gpb->[0];
	    my $last_gpb = $cls_gpb->[$ct_gpblocks-1];
	    my $beg_cls = $first_gpb->[2];
	    my $end_cls = $last_gpb->[3];

	    my $if_5p_end = ($beg_cls == 0) ? 1 : 0;
	    my $if_3p_end = ($end_cls == $len_msa-1) ? 1 : 0;

		# MODIFIED in creation of 'apprx1_tot_lnlk_gpattern_msa2'. #
	    my @params_boundaries = ($cutoff_ins, $cutoff_del, $if_5p_end, $if_3p_end);
	    ##my @params_boundaries = ($cutoff_5p, $cutoff_3p, $opt_bd_indels, $if_5p_end, $if_3p_end);


		# MODIFIED in creation of 'apprx1_tot_lnlk_gpattern_msa2_finer' (I) (on June 7 (Wed), 2017). #

	    my ($br2indels0_for_cls, $node2bw_gstats0_for_cls) = @{$psm_hsts0_for_cls->[$cl]};
	    my $set_psm_hsts0_in_indel_blks = $sets_psm_hsts0_in_indel_blks->[$cl];
	    my $sets_psm_hsts_in_indel_blks = $setssets_psm_hsts_in_indel_blks->[$cl];
	    my $ct_indel_blks = @{$sets_psm_hsts_in_indel_blks};
	    ##my $psm_hsts_in_cls = $sets_psm_hsts_for_cls->[$cl];
	    ##my $ct_psm_hsts = @{$psm_hsts_in_cls};

	    my @set_tot_lnmfac_gc_indel_blk = my @setsets_lnmfac_gc_indel_blk_hst = ();
	    ####my $clwise_tot_lnlk1_gp = 0;
	    ##my @set_lnlk_gp_hsts = ();

	## return ($tot_lnlk_gc_msa, $shared_fac1p, $shared_fac2p, \@set_clwise_tot_lnmfac_gc, \@set_clwise_lnmfac_gc_partit_nw, \@setsets_tot_lnmfac_gc_indel_blk,
	##	$tot_div_tree, $totlen_ng, $totlen_rootseq0, \@len_rootseq0_in_cls, \@set_node2len_ancseq0_in_cls, \@setsetssets_lnmfac_gc_indel_blk_hst);

		# END of "MODIFIED in creation of 'apprx1_tot_lnlk_gpattern_msa2_finer' (I) (on June 7 (Wed), 2017)." #

    if ($TEST_LEVEL>2) {
	print STDOUT '($beg_cls, $end_cls) = ', "($beg_cls, $end_cls) .\n\n";
	print STDOUT '($if_5p_end, $if_3p_end) = ', "($if_5p_end, $if_3p_end) .\n\n";

		# MODIFIED in creation of 'apprx1_tot_lnlk_gpattern_msa2_finer'. #
	print STDOUT '#{indel blocks} = ', $ct_indel_blks, " .\n\n";
	##print STDOUT '#{parsimonious histories} = ', $ct_psm_hsts, " .\n\n";
    }


		## Handle an exceptional case, which probably contains some long gaps. ##
		## (ADDED accompanying the revision of 'MyTreeMap_indels.pm' on Mar 10 (Mon), 2014.) ##
		## (=> Further MODIFIED accompanying the further revision of 'MyTreeMap_indels.pm' on Mar 11 (Tue), 2014.)
		## (=> Further MODIFIED in creation of 'apprx1_tot_lnlk_gpattern_msa2_finer' (on June 7 (Wed), 2017).)


	    my $len_cls = $end_cls - $beg_cls + 1;
	    unless ( !( ($ct_indel_blks == 0) and (($ct_gpblocks> 1) or ($len_cls > $UPPER_BOUND_ANAL_GAPLEN) or ($len_cls > $UPPER_BOUND_ANAL_LEN_GBCLS) ) ) ) {
	    ##unless ( !( ($ct_psm_hsts == 0) and (($ct_gblocks> 1) or ($len_cls > $UPPER_BOUND_ANAL_GAPLEN) or ($len_cls > $UPPER_BOUND_ANAL_LEN_GBCLS) ) ) ) {

		print STDOUT "[CAUTION 'apprx1_tot_lnlk_gpattern_msa2_finer'] The $cl th gapped segment (#{gap-pattern-blocks} = $ct_gpblocks \; region = \[$beg_cls\, $end_cls\] \; length = $len_cls) has an EMPTY SET of parsimonious histories for indel blocks, probably because it contains (a) long gap(s): $! !\n\n";

			# MODIFIED in creation of 'apprx1_tot_lnlk_gpattern_msa2_finer' (on June 7 (Wed), 2017). #
		$tot_lnlk_gc_msa = 'N/A';
		$set_clwise_lnmfac_gc_partit_nw[$cl] = 'N/A';
		$setsets_tot_lnmfac_gc_indel_blk[$cl] = [];
		$setsetssets_lnmfac_gc_indel_blk_hst[$cl] = [];
		$set_clwise_tot_lnmfac_gc[$cl] = 'N/A';
		##$tot_lnlk_gp = 'N/A';
		##$set_clwise_tot_lnlk_gp[$cl] = 'N/A';
		##$setset_lnlk_gp_cl_hst[$cl] = [];

		next;
	    }

	    my $node2len_ancseq0_in_cl = $set_node2len_ancseq0_in_cls[$cl];
	    my $ln_prior_partit_nw_in_cl = 0; # Because we assume a UNIFORM distribution of state priors.

	    my $clwise_lnmfac_gc_partit_nw = calc_lnmfac_gc_partit_nw_for_cl ($topnode_id, %{$node2len_ancseq0_in_cl}, %{$br2attr}, %{$node2pa}, $totrate_indel, $ln_prior_partit_nw_in_cl);
	    unless (defined $clwise_lnmfac_gc_partit_nw) {
		print STDOUT "[ERROR 'apprx1_tot_lnlk_gpattern_msa2_finer'] The subroutine 'calc_lnmfac_gc_partit_nw_for_cls' FAILED on the $cl th gapped segment (#{gap-pattern-blocks} = $ct_gpblocks \; region = \[$beg_cls\, $end_cls\] \; length = $len_cls): $! !\n\n";
		return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	    }

	    $set_clwise_lnmfac_gc_partit_nw[$cl] = $clwise_lnmfac_gc_partit_nw;
	    my $clwise_tot_lnmfac1_gc = $clwise_lnmfac_gc_partit_nw; # ADDED on June 10 (Sat), 2017. #

    if ($TEST_LEVEL>3) {
	print STDOUT "\n{ Initializing \$clwise_tot_lnmfac1_gc }\n\n";
	print STDOUT '$clwise_tot_lnmfac1_gc (initial) (= $clwise_lnmfac_gc_partit_nw) = ', $clwise_tot_lnmfac1_gc, " .\n\n";
    }

		# ADDED in minor modification (on 2019/04/19). (1a) #
	    my @set_rlv_indel_hst2corr_fac = ();
	    $setsets_rlv_indel_hst2corr_fac[$cl] = \@set_rlv_indel_hst2corr_fac;
		# END of "ADDED in minor modification (on 2019/04/19). (1a)" #


	    for (my $idb = 0; $idb < $ct_indel_blks; $idb++) { # Main middle for-loop (over indel blocks). #

		my $psm_hst0_in_idb = $set_psm_hsts0_in_indel_blks->[$idb];
		my ($br2indels0_in_idb, $node2bw_gstats0_in_idb) = @{$psm_hst0_in_idb};
			# @{$sets_psm_hsts0_in_indel_blks[$cl]->[$eiidb]} = (\%br2indels0, \%node2blwise_gstats0) is for the initial candidate indel history realizing the $eiidb th effectively indepenent indel block in the $cl th cluster (i.e., gapped segment);

		my $set_psm_hsts_in_idb = $sets_psm_hsts_in_indel_blks->[$idb];
		my $ct_psm_hsts = @{$set_psm_hsts_in_idb};

     (print STDOUT "{{ The $idb th indel block (#{parsimonious histories} = $ct_psm_hsts) }}\n\n") if ($TEST_LEVEL>2);

			# First, prepare the set of gap-pattern-blocks that are empty in the Dollo parsimonious gap states (at each ancestral node). #

			# %{$node2empty_gpbs0} = ($node_id => \@empty_gpbs0, ...) for the relevant nodes, where @empty_gpbs0 stores indices of the gap-pattern-blocks that are empty in the Dollo parsimonious ancestral gap state at the node.
			#
		my $node2empty_gpbs0 = mk_node2set_empty_blks2 (@set_all_nodes, %{$node2bw_gstats0_in_idb}, %{$br2indels0_in_idb}, $char_abs); # MODIFIED on June 8 (Thu), 2017, in order to specify the indel block unambiguously.
		##my $node2empty_gpbs0 = mk_node2set_empty_blks (@set_int_nodes, %{$node2bw_gstats0_in_idb}, $char_abs);
		unless (defined $node2empty_gpbs0) {
		    print STDOUT "[ERROR 'apprx1_tot_lnlk_gpattern_msa2_finer'] The subroutine 'mk_node2set_empty_blks2' FAILED on the $idb th indel block in the $cl th gapped segment: $! !\n\n";
		    return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
		}

    if ($TEST_LEVEL>=5) {

	print STDOUT ' ' x 3, '[ Contents of %{$node2empty_gpbs0} ]', "\n\n";
	print STDOUT ' ' x 3, join ("\t", 'Node_ID', '[list,empty,gap-pattern-blocks0]'), "\n\n";

	foreach my $node (sort {$a <=> $b} keys %{$node2empty_gpbs0}) {
	    my $cnct_list_empty_gpbs0 = join (',', @{$node2empty_gpbs0->{$node}});
	    print STDOUT ' ' x 3, join ("\t", $node, "[$cnct_list_empty_gpbs0]"), "\n";
	}
	print STDOUT "\n\n";

    } # END of "if ($TEST_LEVEL>=5) {...}"


		my @set_lnmfac_gc_idb_hsts = ();

		my $ct_anom_psns0 = my $ct_anom_psns = 0;

			# ADDED in minor modification (on 2019/04/19) (1b) #
		my %rlv_indel_hst2corr_fac;
		$set_rlv_indel_hst2corr_fac[$idb] = \%rlv_indel_hst2corr_fac;
			# END of "ADDED in minor modification (on 2019/04/19) (1b)" #


		for (my $ps =0; $ps<$ct_psm_hsts; $ps++) {  # Main inner for-loop (over parsimonious indel histories in the indel block) #

		    unless (@{$set_psm_hsts_in_idb->[$ps]}>0) {

			# The scenario information is empty. #
			# (ADDED in revision on Mar 2 (Sun), 2014). #
			# (==> IMPORTED with modifications from 'apprx1_tot_lnlk_gpattern_msa2' in 'MyTreeMap_indels_ML_hs.pm' in creation of 'apprx1_tot_lnlk_gpattern_msa2_finer' (on June 8 (Thu), 2017).)

			$ct_anom_psns++;
			$set_lnmfac_gc_idb_hsts[$ps] = 'N/A';
			next;
		    }

		    my ($br2indels_in_idb, $node2blwise_gstats_in_idb) = @{$set_psm_hsts_in_idb->[$ps]};


    (print STDOUT "{ The $ps th indel history }\n\n") if ($TEST_LEVEL>3);
    ##(print STDOUT "{ The $ps th indel scenario }\n\n") if ($TEST_LEVEL>2);

		    my $delta_log_prior_rootseq_in_idb = 0; # Because we ASSUME a space-homogeneous model. (ADDED in creation of 'apprx1_tot_lnlk_gpattern_msa2_finer' (on June 8 (Thu), 2017). #

			# MODIFIED in creation of 'apprx1_tot_lnlk_gpattern_msa2_finer' (on June 8 (Thu), 2017). #
		    my ($lnmfac1_indel_hst_in_idb, $br2lnmfac1_gpdseg_in_idb, $sumint_diff_xrate_in_idb) = cal_lnmfac1_indel_hst_in_idb2 (@{$cls_gpb}, %{$br2attr}, %{$node2pa}, %{$br2indels_in_idb}, %{$node2blwise_gstats_in_idb}, %{$node2empty_gpbs0}, $char_abs, $totrate_ins, $totrate_del, @params_boundaries, %{$br2log_mfacs}, $delta_log_prior_rootseq_in_idb);
		    ##my ($lnlk1_indel_hst_cls, $br2lnlk1_indel_cls) = cal_lnlk1_indel_hst_cls2 (@{$cls_gb}, %{$br2attr}, %{$node2pa}, %{$br2indels}, %{$node2blwise_gstats}, $char_abs, $totrate_ins, $totrate_del, @params_boundaries, %{$br2log_mfacs});


		    unless (defined $lnmfac1_indel_hst_in_idb) {

			print STDOUT "[CAUTION 'apprx1_tot_lnlk_gpattern_msa2_finer'] The subroutine 'cal_lnmfac1_indel_hst_in_idb2' FAILED on the $ps th parsimonious indel history in the $idb th indel block of the $cl th gapped segment: $! !\n\n";
			$ct_anom_psns0++;
			$set_lnmfac_gc_idb_hsts[$ps] = 'undef';
			next;
		    }

    (print STDOUT " --- The subroutine 'cal_lnmfac1_indel_hst_in_idb2' ENDED (with ", '$lnmfac1_indel_hst_in_idb = ', $lnmfac1_indel_hst_in_idb, ") ...\n\n") if ($TEST_LEVEL>1);

			### If the tree contains a bifurcated root and/or (a) uni-furcated node(s),
			### calculate the correction factor. 
			### (ADDED in MODIFICATION from Dec 30, 2017 till Jan 3 , 2018.)

		if ((@{$info_biroot}>0) or (@{$list_uninodes}>0)) {
		    my $corr_fac = cal_corr_fac_lnmfac1_indel_hst (@{$info_biroot}, @{$list_uninodes},
				@{$cls_gpb}, %{$br2attr}, %{$node2pa}, %{$br2indels_in_idb}, %{$node2blwise_gstats_in_idb},
				 %{$node2empty_gpbs0}, $char_abs, $totrate_ins, $totrate_del, @params_boundaries,
				 %{$br2log_mfacs}, $delta_log_prior_rootseq_in_idb);

	# @{$info_biroot} = ($id_root, $id_child1, $id_child2) for the bifurcated root,
	#	      = () if the root is NOT bifurcated;
	# @{$list_uninodes[$u]} = ($id_parent, $id_uninode, $id_child) for the $u th unifurcated node (other than the root),
	# $id_uninode, whose parent and the sole child are $id_parent and $id_child, respectively.


			### RESTART FROM HERE (3) ###

		    unless (defined $corr_fac) {
			print STDOUT "[CAUTION 'apprx1_tot_lnlk_gpattern_msa2_finer'] The subroutine 'cal_corr_fac_lnmfac1_indel_hst' FAILED on the $ps th parsimonious indel history in the $idb th indel block of the $cl th gapped segment: $! !\n\n";
			$ct_anom_psns0++;
			$set_lnmfac_gc_idb_hsts[$ps] = 'undef';
			next;
		    }

    (print STDOUT " --- The subroutine 'cal_corr_fac_lnmfac1_indel_hst' ENDED (with ", '$corr_fac = ', $corr_fac, ") ...\n\n") if ($TEST_LEVEL>1);

		    if ($corr_fac != 0) {
			$lnmfac1_indel_hst_in_idb += $corr_fac;

			$rlv_indel_hst2corr_fac{$ps} = $corr_fac; # ADDED in minor modification (on 2019/04/19) (1c) #

    (print STDOUT '($cl, $idb, $ps, $corr_fac, $lnmfac1_indel_hst_in_idb) = ', "($cl, $idb, $ps, $corr_fac, $lnmfac1_indel_hst_in_idb) .\n\n") if ($TEST_LEVEL_corr>3); # ADDED in minor modification (on 2019/04/19) (1d) #

		    }
		} # END of "if ((@{$info_biroot}>0) or (@{$list_uninodes}>0)) {...}"

			### END of "(ADDED in MODIFICATION from Dec 30, 2017 till Jan 3, 2018.)"


		    $set_lnmfac_gc_idb_hsts[$ps] = $lnmfac1_indel_hst_in_idb;

		}		# End of the main inner for-loop (over parsimonious indel histories in the indel block) #


			# MODIFIED to summarize the results for an indel block, in creation of 'apprx1_tot_lnlk_gpattern_msa2_finer' (from June 7 (Wed) - June 10 (Sat), 2017). #

		my $tot_lnmfac1_gc_indel_blk = ($ct_anom_psns>0) ? 'N/A' : (($ct_anom_psns0>0) ? 'undef' : (($ct_psm_hsts == 1) ? $set_lnmfac_gc_idb_hsts[0] : log_sum_exp (@set_lnmfac_gc_idb_hsts))) ;
	        ## my $clwise_tot_lnlk1_gp = ($ct_anom_psns>0) ? 'N/A' : (($ct_anom_psns0>0) ? 'undef' : (($ct_psm_scns == 1) ? $set_lnlk_gp_scns[0] : log_sum_exp (@set_lnlk_gp_scns))) ;


		if ($ct_anom_psns>0) {
		    $clwise_tot_lnmfac1_gc = 'N/A';
		    ##$tot_lnlk_gp = 'N/A';
		} elsif ($ct_anom_psns0>0) {
		    ($clwise_tot_lnmfac1_gc eq 'N//A') or ($clwise_tot_lnmfac1_gc = 'undef');
		    ##$tot_lnlk_gp = 'undef';
		} else {
		    (($clwise_tot_lnmfac1_gc eq 'N/A') or ($clwise_tot_lnmfac1_gc eq 'undef')) or ($clwise_tot_lnmfac1_gc += $tot_lnmfac1_gc_indel_blk) ;
		    ##(($tot_lnlk_gp eq 'N/A') or ($tot_lnlk_gp eq 'undef')) or ($tot_lnlk_gp += $clwise_tot_lnlk1_gp);
		}
		$set_tot_lnmfac_gc_indel_blk[$idb] = $tot_lnmfac1_gc_indel_blk;
		$setsets_lnmfac_gc_indel_blk_hst[$idb] = \@set_lnmfac_gc_idb_hsts;
		##$set_clwise_tot_lnlk_gp[$cl] = $clwise_tot_lnlk1_gp;
		##$setset_lnlk_gp_cl_hst[$cl] = \@set_lnlk_gp_hsts;

    if ($TEST_LEVEL>3) {
	print STDOUT "\n{ Final results for the $idb th indel block }\n\n";
	print STDOUT '$tot_lnmfac1_gc_indel_blk  = ', $tot_lnmfac1_gc_indel_blk, " .\n\n";
	print STDOUT '$clwise_tot_lnmfac1_gc (current) = ', $clwise_tot_lnmfac1_gc, " .\n\n\n";
    }

		    	# END of "MODIFIED to summarize the results for an indel block, in creation of 'apprx1_tot_lnlk_gpattern_msa2_finer' (from June 7 (Wed) - June 10 (Sat), 2017)." #

=Just a reminder
	    my ($br2indels0_for_cls, $node2bw_gstats0_for_cls) = @{$psm_hsts0_for_cls->[$cl]};
	    my $set_psm_hsts0_in_indel_blks = $sets_psm_hsts0_in_indel_blks->[$cl];
	    my $sets_psm_hsts_in_indel_blks = $setssets_psm_hsts_in_indel_blks->[$cl];
	    my $ct_indel_blks = @{$sets_psm_hsts_in_indel_blks};


		my $set_psm_hsts_in_idb = $sets_psm_hsts_in_indel_blks->[$idb];
		my $ct_psm_hsts = @{$set_psm_hsts_in_idb};

	my $tot_lnlk_gc_msa = $shared_fac1p + $shared_fac2p;

	    my $clwise_tot_lnmfac1_gc = $clwise_lnmfac_gc_partit_nw; # ADDED on June 10 (Sat), 2017. #
	    my @set_tot_lnmfac_gc_indel_blk = my @setsets_lnmfac_gc_indel_blk_hst = ();

=cut # END of "Just a reminder"

	    }		# End of the main middle for-loop (over indel blocks). #


			# NEWLY ADDED in creation of 'apprx1_tot_lnlk_gpattern_msa2_finer' (xx) (on June 7 (Wed), 2017). #
	    $setsets_tot_lnmfac_gc_indel_blk[$cl] = \@set_tot_lnmfac_gc_indel_blk;
	    $setsetssets_lnmfac_gc_indel_blk_hst[$cl] = \@setsets_lnmfac_gc_indel_blk_hst;
	    $set_clwise_tot_lnmfac_gc[$cl] = $clwise_tot_lnmfac1_gc;

	    if ($clwise_tot_lnmfac1_gc eq 'N/A') {  # ADDED on June 10 (Sat), 2017. #
		$tot_lnlk_gc_msa = 'N/A';
	    } elsif ($clwise_tot_lnmfac1_gc eq 'undef') {
		($tot_lnlk_gc_msa eq 'N/A') or ($tot_lnlk_gc_msa = 'undef');
	    } else {
		(($tot_lnlk_gc_msa eq 'N/A') or ($tot_lnlk_gc_msa eq 'undef')) or ($tot_lnlk_gc_msa += $clwise_tot_lnmfac1_gc);
	    } # END of "if () { # ADDED on June 10 (Sat), 2017. # ...} elsif () {...} else {...}"

    if ($TEST_LEVEL>2) {
	print STDOUT "\n{{ Final results for the $cl th cluster }}\n\n";
	print STDOUT '$clwise_tot_lnmfac1_gc = ', $clwise_tot_lnmfac1_gc, " ,\n";
	print STDOUT '$tot_lnlk_gc_msa (current) = ', $tot_lnlk_gc_msa, " .\n\n";
    }
			# END of "NEWLY ADDED in creation of 'apprx1_tot_lnlk_gpattern_msa2_finer' (xx) (on June 7 (Wed), 2017)." #

	}			# End of the main outer for-loop (over gap-block clusters, i.e., gapped segments) #


    if ($TEST_LEVEL>2) {
	print STDOUT "\n[ Final Result of the Main Process ]\n\n";
	print STDOUT '$tot_lnlk_gc_msa = ', $tot_lnlk_gc_msa, " .\n\n";
    }

    (print STDOUT "\n=== END of the Subroutine 'apprx1_tot_lnlk_gpattern_msa2_finer' ===\n\n") if ($TEST_LEVEL>0);

		# Minor MODIFICATION (on 2019/04/19) (2) #
	return ($tot_lnlk_gc_msa, $shared_fac1p, $shared_fac2p, \@set_clwise_tot_lnmfac_gc, \@set_clwise_lnmfac_gc_partit_nw, \@setsets_tot_lnmfac_gc_indel_blk,
		$tot_div_tree, $totlen_ng, $totlen_rootseq0, \@len_rootseq0_in_cls, \@set_node2len_ancseq0_in_cls, \@setsetssets_lnmfac_gc_indel_blk_hst, \@setsets_rlv_indel_hst2corr_fac);
	##return ($tot_lnlk_gc_msa, $shared_fac1p, $shared_fac2p, \@set_clwise_tot_lnmfac_gc, \@set_clwise_lnmfac_gc_partit_nw, \@setsets_tot_lnmfac_gc_indel_blk,
	##	$tot_div_tree, $totlen_ng, $totlen_rootseq0, \@len_rootseq0_in_cls, \@set_node2len_ancseq0_in_cls, \@setsetssets_lnmfac_gc_indel_blk_hst);
		# END of "Minor MODIFICATION (on 2019/04/19) (2)" #

}	# END of the subroutine 'apprx1_tot_lnlk_gpattern_msa2_finer'. # 


1;	# a "spell" which makes the module "use"-able and/or "require"-able #


