package MyTreeMap_indels_spt_odr_finer;
#
###################################################################################
#	
#	ANEX_P (Alignment Neighborhood Explorer, type P):
#	    A package including a program to "proofread" an input multiple sequence alignment taking account of
#	    stochasticity, evolutionary consistency, and biological realism.
#	   (Written almost exclusively in Perl.)
#
#	Version 0.3: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
#
###################################################################################
#
# MyTreeMap_indels_spt_odr_finer.pm:
#	This package does essentially the same job as 'MyTreeMap_indels_spt_odr.pm'
#	in terms of a "finer" format, in which parsimonious indel histories
#	will be searched for WIHTIN each isolated gap-block bounded
#	by an internal node where the Dollo ancestral state is completely occupied by residues
#	and by a pair of sites where all 'descendants' of the internal node is completely occupied by residues, instead of within each gapped segment as the predecessor ('MyTreeMap_indels_spt_odr.pm') did.
#	[CREATION STARTED on Feb 3 (Fri), 2017, ENDED on Feb xx (www), 2017.]
#
#
# MyTreeMap_indels_spt_odr.pm:
#	This modified version of the package 'MyTreeMap_indels.pm' searches for
#	the most parsimonious indel scenarios while respecting the spatial order
#	between the gap-blocks, even when they appear to be swappable with each other,
#	as long as each of them has 'presence' states in the (temporal) 'downstream' of the other.
#	(CREATION STARTED on Mar 18 (Tue), 2014, ENDED on Mar 20 (Thu), 2014)
#
#	REVISED on Mar 22 (Sat) & 23 (Sun), 2014:
#	Now, the subroutine, 'incorporate_null_block', was overhauled by introducing a new satellite subroutine,
#	'search_flanking_blocks_spt_odr2', and became much closer to being perfect.
#
#	REVISED on Mar 24 (Mon), 2014:
#	Now, the subroutine, 'recede_absence_spt_odr', was revised by incorporating the subroutine,
#	'search_flanking_blocks_spt_odr2', and became closer to being perfect.
#
# MyTreeMap_indels.pm:
#	This package makes a set of maps of the most parsimonious insertions/deletions
#	scenarios onto an input sequence tree.
#	The main algorithm is based on Farris's (1977) Dollo parsimony principle.
#	(CREATION STARTED on May 3 (Thu), 2012, ENDED on May ?? (???), 2012)
#
#	REVISED on Oct 24 (Wed), 2012:
#	To make the algorithm closer to perfection,
#	the subroutin "recede_absence" was modified so that it will utilize
#	flanking blocks in broad sense as well as those in strict sense.
#
#	REVISED on Nov 13 (Wed), 2013:
#	The module can now correctly handle trees containing two-branch nodes.
#
#	REVISED on Nov 15 (Fri), 2013:
#	An optional flag (global variable) was added to enable the subroutine 'enum_psm_indel_scns_in_cls_blk'
#	to ignore null columns when necessary.
#	[ This may be useful when we deal with a 'culled' MSA that resulted from removing some sequences
#	from an alignmentof a larger number of sequences reconstructed via an aligning program. ]
#
#	REVISESD on Mar 8 (Sat) & 9 (Sun), 2014:
#	A new subroutine, 'enum_psm_scns_gen_null_blk', was created.
#	It enables the subroutine 'enum_psm_indel_scns_in_cls_blks' to enumerate the parsimonious indel scenarios
#	that resulted in "NULL-cluster", which consists only of gaps.
#
#	REVISED on Mar 10 (Mon), 2014:
#	The package now introduces a global variable, $UPPER_BOUND_ANAL_GAPLEN.
#	If a gap-block cluster contains a run of gaps longer than this variable,
#	the subroutine 'enum_psm_indel_scns_for_msa' skips further analyses of the cluster.
#	Accompanying the introduction of this global variable,
#	a new subroutine 'scan_gbcls_for_long_gaps' was created.
#
#	REVISED on Mar 11 (Tue), 2014:
#	TEMPORARILY Modified the subroutine 'incorporate_new_event' to TENTATIVELY cope with
#	the rare cases where the new 'deletion' event encounters BOTH left- and right-neighboring events.
#
#	(I WILL SERIOUSLY ADDRESS THIS ISSUE LATER. ... Probably, a solution would be to leave 'deletion' events unmarged when they are intervened by the 'presence' state in any 'downstream' nodes.)
#
#
#	REVISED on Mar 11 (Tue), 2014 (II):
#	The subroutine now introduces an upper-bound, $UPPER_BOUND_ANAL_LEN_GBCLS = 200,
#	for the cluster-length itself as well.
# 
# MyTreeMap_D.pm:
#	This package maps states and their changes
#	onto the nodes and the branches, respectively, of a phylogenetic tree,
#	according to Farris's (1977) Dollo parsimony principle.
#	In a nutshell, Farris's (1977) Dollo parsimony seeks for the evolutionary scenario
#	that minimizes the number of changes, UNDER THE CONSTRAINT that
#	the 'creation' of a state occurs at most once in each column.
#	(CREATION STARTED on Apr 19 (Thu), 2012, ENDED on Apr ?? (???), 2012)
#
# MyTreeMap_II.pm:
#	This package performs more refined algorithms than 'MyTreeMap.pm',
#	to map various properties, or their changes,
#	onto the nodes or the branches of a phylogenetic tree.
#	(CREATED on June 16 (Wed), 2010)
#	(REVISED on Nov 27 (Sun), 2011)
#	The subroutine 'min_penalties_sankoff_bo' now excludes from later consideration
#	the characters on each node that will NEVER be included in any parsiminious states.
#
# MyTreeMap.pm:
#	This package maps various properties, or their changes,
#	 onto the nodes or the branches
#	of a (molecular) phylogenetic tree.
#	It mainly uses the parsimony principle,
#	but sometimes uses other methods, such as the maximum likelihood.

use strict;
use warnings;


use MyArrayManipu qw( diff_array union_array nr_array lin_search multiple_sort union_array_of_arrays );
use MyArrayManipu_num qw( diff_nmarray union_nmarray nr_nmarray binsearch_num );

use MyTreeManipu_II qw(
		);
	## input_tree_file read_tree_file br_output_tree
	##make_node2children
	## input_subtree copy_tree
	## output_tree draw_subtree topo_subtree
	## shift_topnode move_topnode segregate_subset_nodes
	## sum_branch_lengths intra_lineage_distance inter_node_distance
	## fetch_boot_series extract_hidden_bootstrap
	## resolve_multipartite

use MyTaxonomy qw(
	fetch_ancestry 
	fetch_common_ancestors
	fetch_all_offsprings_nr
		);
	# fetch_last_common_ancestor_ov2
	# fetch_ext_offsprings fetch_all_offsprings
	# fetch_ext_offsprings_br

use MyTreeManipu_III qw(
	make_node2abs_depth
	make_node2ct_exoffs
        );

use MyTreeMap_indels_sppl qw(
	counts_indels
	create_br2eqbr
	create_br2nn
	);

use MyTreeMap_indels_spt_odr qw(
	recede_absence_spt_odr
	cleanup_br2indels
	scan_gbcls_for_long_gaps
	create_list_to_be_examined
	);

use MyTreeMap_indels_spt_odr2 qw( cal_init_psm_cands_for_sgl_msa );

=set aside for later use (IMPORTED from 'MyTreeMap_indels_spt_odr.pm')

use MyMalnIO qw( pr_ODEN_input );

use MyTreeManipu_III qw(
	make_node2abs_depth make_node2ct_exoffs
	);

use MyTreeMap_II qw(
	locate_parsim_changes
	);
use MyTreeMap_D qw(
	map_parsim_states_dollo_msa
	);
	# ct_presext_dollo_bu map_node2parsstat_dollo_td
	# map_parsim_states_dollo_clm
	# map_parsim_states_dollo_1clm

=cut # END of "set aside for later use (IMPORTED from 'MyTreeMap_indels_spt_odr.pm')"


use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	partit_into_gblocks_seqblocks
	if_idb_includes_root
	partit_into_eff_indep_indel_blocks
	find_invlvd_in_eff_indep_indel_blocks
	mk_node2blwise_gstat
	mk_br2indels

	enum_psm_indel_hsts_in_eff_indp_indel_blk
	enum_psm_indel_hsts_in_cls_blks_finer
	enum_psm_indel_hsts_for_msa_finer_PRE
	enum_psm_indel_hsts_for_msa_finer_1ST_HF
	enum_psm_indel_hsts_for_msa_finer_2ND_HF
	enum_psm_indel_hsts_for_msa_finer
	);
	## define_off_limits
	## enum_psm_indel_scns_in_eff_indp_indel_blk ... RENAMED by replacing '_scns_' with '_hsts_' on May 11 (Thu), 2017.


		# Module-specific subroutines #
sub exam_pw_bd_dependence (\@\@\%$\%\%$\%) ;
sub pw_merge_indel_blocks (\@\@) ;



	# This specifies the level of tests on the subroutine, 'enum_psm_indel_hsts_for_msa_finer'. #
	# (0: Nothing, 1: brief, 2: a bit detailed, 3: quite detailed, 5: very detailed)
our $TEST_LEVEL = 1;
##our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 3;
##our $TEST_LEVEL = 5;

	# For 'enum_psm_indel_hsts_for_msa_finer_PRE'. #
our $TEST_LEVEL_PRE = $TEST_LEVEL;
#$TEST_LEVEL_PRE = 1/2; # For the briefest non-null output. #

	# For 'enum_psm_indel_hsts_for_msa_finer_1ST_HF'. #
our $TEST_LEVEL_1ST = $TEST_LEVEL;
#$TEST_LEVEL_1ST = 1/2; # For the briefest non-null output. #

	# For 'enum_psm_indel_hsts_for_msa_finer_2ND_HF'. #
our $TEST_LEVEL_2ND = $TEST_LEVEL;
#$TEST_LEVEL_2ND = 1/2; # For the briefest non-null output. #

	# For 'enum_psm_indel_hsts_in_eff_indp_indel_blk'. #
our $TEST_LEVEL1 = 1;
##our $TEST_LEVEL1 = 5;

	# This specifies the level of tests on other important satellite subroutines.
our $TEST_LEVEL2 = 0;
##our $TEST_LEVEL2 = 1;
##our $TEST_LEVEL2 = 2;
##our $TEST_LEVEL2 = 5;


our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
##our $INDX_LENGTH = 0;
##our $INDX_BOOT = 1;
##our $START_GROUP = '(';	# the symbol marking the start of the grouping of nodes.
##our $END_GROUP = ')';  # the symbol marking the end of the grouping of nodes.

	## A flag that determines whether the subroutine 'enum_psm_indel_scns_in_cls_blks'
	## should leave null columns untouched or not.
	## = 0 when handle null columns, regarding them as resulting from indel events within the phylogeny of aligned sequences (useful when handling simulated MSAs),
	## = 1 when ignore (i.e., leave untouched) null columns, regarding them as resulting from the removal of sequences
	##	that hold residues in the columns (useful when handling 'culled' MSAs made from MSAs reconstructed via aligners).
	## (ADDED on Nov 15 (Fri), 2013)
	#
our $FL_IGNORE_NULL_COLUMNS = 1;

	## A flag that determines whether or not you should conjure the subroutine 'enum_psm_scns_gen_null_blk',
	## which enumerates parsimonious scenarios generating null-clusters (namely, gap-block clusters consisting only of null columns).
	## =1 if you want to, 
	## = 0 if you don't want to.
	# (ADDED on Mar 8 (Sat), 2014.)
	#
our $FL_HANDLE_NULL_CLS = 0;
($FL_IGNORE_NULL_COLUMNS) and ($FL_HANDLE_NULL_CLS  = 0); # When ignoring null-columns, null-clusters must always be ignored as well.

	## This variable sets an upper-bound for the gap-length.
	# IF a gap-block cluster contains a run of gaps longer than this variable,
	# the subroutine 'enum_psm_indel_scns_for_msa' skips further analyses on the cluster.
	# The variable will be used in the new subroutine, 'scan_gbcls_for_long_gaps'.
	# (ADDED on Mar 10 (Mon), 2014.)
	#
our $UPPER_BOUND_ANAL_GAPLEN = 150;

	## This variable sets an upper-bound for the length of the gap-block cluster itself.
	## (ADDED on Mar 11 (Tue), 2014.):w
our $UPPER_BOUND_ANAL_LEN_GBCLS = 200;

	## These two indices specify where the information is stored
	## in %nodee2pars_stats = ($node_id => \@stat_info, ...) #
	#
our $INDX_STAT = 0; # The state (or the character) mapped onto the node.
our $INDX_SBTOT_PEN = 1; # The subtotal of penalties on the subtree under $node_id. 


	### 5 important hash tables that are used in this Module
	### in order to describe a tree
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.


	### DEFINITIONS of the SUBROUTINES ###


# define_off_limits:
#	This subroutine defines "off-limit" zones of gaps and indels in a gapped segment,
#	using the Dollo parsimonious ancestral states within the gapped segment.
#	(It comprizes STEP 3 (A) in the 'workflow_for_indel_histories&multiplication_factors.rtf'.)
#
#	[CREATION STARTED on Feb 24 (Fri), 2017, ENDED on mmm dd (www), 2017.]
#
# INPUT: (\@gpblocks_in_cls, \%br2indels, \%node2blwise_gstat, $topnode_id, \%node2pa, \%node2ch, $char_abs), where
#
#	@{$gpblocks_in_cls[$gpb]} = (($indx_gp, $len_gpb, $start, $end) is for the $gpb th gap-pattern block in the subject gapped segment, with
#	$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#	$len_gpb being the length (in terms of #{columns}) of the gap-pattern block,
#	and $start/$end being the starting/ending coordinate of the gap-pattern block (in the 0-based full-closed convention);
#
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved);
#
#	%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
#	with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node;
#
#	$topnode_id is the ID of the top-node (i.e., root) of the tree;
#
#	%node2pa = ($node_id => $parent_node_id, ...) for all the non-root nodes;
#	%node2ch = ($node_id => \@children, ...) for all internal nodes, with @children listing the IDs of all child nodes;
#	$char_abs is the character specifying the absence of the residue at the site. (Usually, it's the same as the 'gap' character.)
#
# OUTPUT: \@list_off_limits, where
#	
#	@{$list_off_limits[$ol]} = ($br, $up_or_lw, $start, $end, \@invlvd_gpbs) defines the $ol th off-limit boundary in the gapped segment in question.
#	$br is the ID of the branch delimiting the off-limit region,
#	$up_or_lw = 'U'/'L' if the upper/lower-end of $br is off-limit, 
#	$start/$end is the start/end coordinate of the boundary,
#	@invlvd_gpbs lists the gap-pattern-blocks involved. 
#
sub define_off_limits (\@\%\%$\%\%$) {

	my ($gpblocks_in_cls, $br2indels, $node2blwise_gstat, $topnode_id, $node2pa, $node2ch, $char_abs) = @_;

	my @list_off_limits = ();

	return \@list_off_limits;

}	# END of the subroutine 'define_off_limits'. #


# partit_into_gblocks_seqblocks:
#	This subroutine partitions each gapped segment into 'gap-block's and 'sequence-block's,
#	according to the Dollo parsimonious indel history.
#	(It comprizes STEP 3 (A) in the 'workflow_for_indel_histories&multiplication_factors.rtf'.)
#
#	[DEFINITIONS]
#	(1) A "gap-block" is defined by each indel event in the Dollo parsimonious indel history,
#	as a block of gaps extending from one end of the branch where the indel event occur.
#	(Because the Dollo parsimonious indel history is always phylogenetically consistent with the MSA,
#	the block should never be interfered by any other indel event.)
#
#	(2) A "sequence-block" is defined ONLY IF the bounding indel events satisfy the following two conditions:
#	  (i) There is a single bounding event (called the 'primary event') whose set of involved blocks
#		(a) consists solely of contiguous blocks,
#		and (b) completely includes (but does not equal) the set of blocks involved in each other bounding event.
#	  (ii) the primary event has NO nearest-neighboring indel event.
#
#	[CREATION STARTED on Feb 25 (Sat), 2017, ENDED on mmm dd (www), 2017.]
#
# INPUT: (\@gpblocks_in_cls, \%br2indels, \%node2blwise_gstat, $topnode_id, \%node2pa, \%node2ch, $char_abs, \%br2eq, \%br2nn), where
#
#	@{$gpblocks_in_cls[$gpb]} = ($indx_gp, $len_gpb, $start, $end) is for the $gpb th gap-pattern block in the subject gapped segment, with
#	$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#	$len_gpb being the length (in terms of #{columns}) of the gap-pattern block,
#	and $start/$end being the starting/ending coordinate of the gap-pattern block (in the 0-based full-closed convention);
#
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved);
#
#	%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
#	with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node;
#
#	$topnode_id is the ID of the top-node (i.e., root) of the tree;
#
#	%node2pa = ($node_id => $parent_node_id, ...) for all the non-root nodes;
#	%node2ch = ($node_id => \@children, ...) for all internal nodes, with @children listing the IDs of all child nodes;
#	$char_abs is the character specifying the absence of the residue at the site (Usually, it's the same as the 'gap' character);
#	%br2eq = ($br_id => $equivalent_br_id, ...) for branches with equivalents;
#	%br2nn = ($br_id => {$id_nearest_neighbor => 1, ...}, ...).
#
# OUTPUT: (\@list_gblocks, \@list_seqblocks), where
#
#	@{$list_gblocks[$gb]} = ($br, $up_or_lw, $start, $end, \@invlvd_gpbs, \@nesting) gives information on the $gb th gap-block.
#	$br is the ID of the branch delimiting the gap-block,
#	$up_or_lw = 'U'/'L' if the gap-block extends on the upper/lower side of the branch,
#	$start/$end is the start/end coordinate of the gap-block,
#	@invlvd_gpbs lists the gap-pattern blocks involved,
#	@nesting lists the sequence-blocks nesting the gap-block.
#
#	@{$list_seqblocks[$sb]} = ($br, $up_or_lw, $start, $end, \@invlvd_gpbs, \@nested) gives information on the $sb th sequence-block.
#	$br is the ID of the primary branch delimiting the sequence-block,
#	$up_or_lw = 'U'/'L' if the sequence block extends on the upper/lower side of the branch,
#	$start/$end is the start/end coordinate of the sequence-block,
#	@invlvd_gpbs lists the gap-pattern blocks involved,
#	@nested lists the gap-blocks nested in the sequence-block.
#
sub partit_into_gblocks_seqblocks (\@\%\%$\%\%$\%\%) {

	my ($gpblocks_in_cls, $br2indels, $node2blwise_gstat, $topnode_id, $node2pa, $node2ch, $char_abs, $br2eq, $br2nn) = @_;


		# First, prepare a preliminary list of gap-blocks. #

	my @prelist_gblocks = (); # element = [$br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs];

	foreach my $br (sort {$a <=> $b} keys %{$br2indels}) { # Preliminary foreach-loop (over branches where indels occur). #

	    my $indels = $br2indels->{$br};

	    foreach my $indel (@{$indels}) { # Inner foreach-loop (over indels along $br). #

		my ($state_ue, $state_le, $invlvd_gpbs) = @{$indel};
		(defined $state_ue) or next; # Skip a null event. #
		my $up_or_lw = ($state_ue eq $char_abs) ? 'U' : 'L' ;
		my ($lm_gpb, $rm_gpb) = @{$invlvd_gpbs}[0, $#{$invlvd_gpbs}];

		push @prelist_gblocks, [$br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs];

	    }				# End of the inner foreach-loop (over indels along $br). #

	}					# End of the preliminary foreach-loop (over branches where indels occur). #

	my @indices_srt = (2, 3, 0); my @methods_srt = ('+', '+', '+');

	my $std_prelist_gblocks = multiple_sort (@prelist_gblocks, @indices_srt, @methods_srt);

		# Then, create the output list of gap-blocks.

	my @list_gblocks = ();

	my $ct_gblocks = @{$std_prelist_gblocks};
	for (my $gb = 0; $gb < $ct_gblocks; $gb++) { # 1st main foreach-loop (over gap-blocks). #

	    my $pre_gblk = $std_prelist_gblocks->[$gb]; 

	    my ($br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs) = @{$pre_gblk};
	    my $start = $gpblocks_in_cls->[$lm_gpb][2];
	    my $end   = $gpblocks_in_cls->[$rm_gpb][3];
			# @{$gpblocks_in_cls[$gpb]} = ($indx_gp, $len_gpb, $start, $end)

	    my @nesting = ();
	    push @list_gblocks, [$br, $up_or_lw, $start, $end, $invlvd_gpbs, \@nesting];

	    ## push @{$pre_gblk}, $gb;

	}				# End of the 1st main foreach-loop (over gap-blocks). #

		# Finally, create the list of sequence blocks. #

	my @list_seqblocks = ();

	for (my $gb=0; $gb < $ct_gblocks; $gb++) { # 2nd main for-loop (over gap-blocks). #

	    my $pre_gblk = $std_prelist_gblocks->[$gb]; 

	    my ($br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs) = @{$pre_gblk};

	    if (scalar (@{$invlvd_gpbs}) < $rm_gpb - $lm_gpb + 1) {
		next; # Skip if the subject does NOT consist solely of contiguous gap-pattern blocks. # [...This condition might be removed in the future...]
	    }

	    my $eq_br = $br2eq->{$br};
	    my $br2if_nn = $br2nn->{$br};

	    my $if_overlapping = 0; # = 1 if there is a gap-block that is overlapping, but neither nested nor nesting, the subject gap-block. #
	    my $if_full_overlapping = 0; # = 1 if there is a gap-block that is fully overlapping the subject gap-block. #
	    my $if_nesting = 0; # = 1 if there is a gap-block that is nesting the subject gap-block. #
	    my $if_nn_br = 0; # = 1 if there is a nested gap-block whose delimiting branch is either identical, equivalent, or nearest-neighboring the branch delimiting the subject gap-block.
	    my @nested = (); # The list of indices of gap-blocks that are nested in the subject gap-block.

	    for (my $gb2=0; $gb2 < $ct_gblocks; $gb2++) { # Main inner for-loop (over gap-blocks). #

		($gb2 == $gb) and next; # Skip the subject itself. #

		my $pre_gblk2 = $std_prelist_gblocks->[$gb2]; 
		my ($br2, $up_or_lw2, $lm_gpb2, $rm_gpb2, $invlvd_gpbs2) = @{$pre_gblk2};

		(($rm_gpb < $lm_gpb2) or ($rm_gpb2 < $lm_gpb)) and next; # Skip if the gap-blocks do NOT overlap. #

		my ($common, $ex1, $ex2) = diff_nmarray (@{$invlvd_gpbs}, @{$invlvd_gpbs2});

		(@{$common}>0) or next; # Skip if the gap-blocks do NOT overlap. #

		if (@{$ex1}>0) {
		    if (@{$ex2}>0) {
			$if_overlapping = 1;
			last;
		    } else {
			if ($br == $br2 or ((defined $eq_br) and ($eq_br == $br2)) or (defined $br2if_nn->{$br2})) {
			    $if_nn_br = 1; # I'm not sure if this condition is really necessary or not... #
			    last;
			} else {
			    push @nested, $gb2;
			}
		    }
		} elsif (@{$ex2}>0) {
		    $if_nesting = 1;
		    last;
		} else {
		    $if_full_overlapping = 1;
		    last;
		}

	    }					# End of the main inner for-loop (over gap-blocks). #

	    if ($if_overlapping or $if_full_overlapping or $if_nesting or $if_nn_br) {
		next; # In this case, the gap-block canNOT provide a primary boundary of a sequence-block. #
	    }

			## Create an entry of the sequence-block. ##

		## ($br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs) = @{$pre_gblk};
	    my $start = $gpblocks_in_cls->[$lm_gpb][2];
	    my $end   = $gpblocks_in_cls->[$rm_gpb][3];

	    $up_or_lw = ($up_or_lw eq 'L') ? 'U' : 'L'; # Flip the flag, from that of a gap-block to that of a sequence-blok. #

	    my $indx_sb = @list_seqblocks;
	    push @list_seqblocks, [$br, $up_or_lw, $start, $end, $invlvd_gpbs, \@nested];

		# @{$list_seqblocks[$sb]} = ($br, $up_or_lw, $start, $end, \@invlvd_gpbs, \@nested) gives information on the $sb th sequence-block.

			## Fill-in the nesting-nested relationships in @list_gblocks. ##

	    foreach my $gb (@nested) { # Trailing inner for-loop (over gap-blocks nested in the subject sequence-block). #

		my ($br2, $up_or_lw2, $start2, $end2, $invlvd_gpbs2, $nesting2) = @{$list_gblocks[$gb]};

		push @{$nesting2}, $indx_sb;
	    }			# End of the trailing inner for-loop (over gap-blocks nested in the subject sequence-block). #
	}				# End of the 2nd main for-loop (over gap-blocks). #


	return (\@list_gblocks, \@list_seqblocks);

}	# END of the subroutine 'partit_into_gblocks_seqblocks'. #


# exam_pw_bd_dependence:
#	This satellite subroutine of 'partit_into_eff_indep_indel_blocks' 
#	examines whether two boundaries are "dependent" of each other or not.
#	The boundaries are considered as "independent" of each other if
#	(1) they do NOT overlap or mutually nearest-neighbor horizontally
#	(more precisely, in this case, the boundaries canNOT be CLEARLY judged as "mutually dependent"),
#	or if
#	(2) they are separated by a node with a Dollo-parsimonious state
#	whose relevant sites (or gap-pattern blocks) are ALL occupied (simultaneously) by "presence".
#	(The "relevant" gap-pattern blocks are those shared by the two boundaries
#	and those in one boundary whose nearest-neighbors exist in the other boundary.)
#	(In case (2), the boundaries are CLEARLY independent of each other.)
#
#	[CREATEION STARTED on Mar 2 (Thu), 2017, ENDED on Mar 4 (Sat), 2017]
#
# INPUT: (\@boundary1, \@boundary2, \%node2blwise_gstat, $topnode_id, \%node2pa, \%node2ch, $char_abs, \%br2equiv), where
#
#	@boundary1/2 = ($br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs) defines the 1st/2nd boundary, with
#	$up_or_lw = 'U'/'L' if the upper/lower-end of $br is the indel block (occupied by gaps),
#	@invlvd_gpbs storing the indices of gap-pattern blocks involved in the boundary;
#
#	%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
#	with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node;
#
#	$topnode_id is the ID of the top-node (i.e., root) of the tree;
#
#	%node2pa = ($node_id => $parent_node_id, ...) for all the non-root nodes;
#	%node2ch = ($node_id => \@children, ...) for all internal nodes, with @children listing the IDs of all child nodes;
#	$char_abs is the character specifying the absence of the residue at the site. (Usually, it's the same as the 'gap' character.)
#	%br2equiv = ($br_id => $equiv_br_id, ...) for branches that have equivalents.
#
# OUTPUT: $if_pw_dependent, which is
#	1 if the two boundaries are dependent on each other,
#	0 if they are independent of each other.	       
#
sub exam_pw_bd_dependence (\@\@\%$\%\%$\%) {

	my ($boundary1, $boundary2, $node2blwise_gstat, $topnode_id, $node2pa, $node2ch, $char_abs, $br2equiv) = @_;

	my ($br1, $up_or_lw1, $lm_gpb1, $rm_gpb1, $invlvd_gpbs1) = @{$boundary1};
	my ($br2, $up_or_lw2, $lm_gpb2, $rm_gpb2, $invlvd_gpbs2) = @{$boundary2};

	if ( ($rm_gpb1 + 1 < $lm_gpb2) or ($rm_gpb2 + 1 < $lm_gpb1) ) {

		# If the two boundaries NEITHER overlap NOR neighbor each other,
		# they are regarded as "independent" of each other.
		# (Actually, NOT CLEARLY judged as "dependent" on each other.)

	    return 0;
	}

	## my ($ovlping, $exclusive1, $exclusive2) = diff_nmarray (@{$invlvd_gpbs1}, @{$invlvd_gpbs2}) ; # Turned out to be unnecessary. #

		# Search for overlapping gap-pattern blocks & nearest-neighboring pairs of gap-pattern blocks,
		# one in the 1st boundary and the other in the 2nd boundary.

	my @ovlping = ();
	my @nn_pairs = (); # element = [$gpb_in_1, $gpb_in_2]
	my $indx_num = undef;
	my @indices_rlv = (0 .. $#{$invlvd_gpbs2}) ;

	foreach my $gpb_in_1 (@{$invlvd_gpbs1}) { # 1st foreach-loop (over gap-pattern blocks involved in the 1st boundary). #

	    my ($if_exist, $indx2_qu, $indx_qu) = binsearch_num ($gpb_in_1, @{$invlvd_gpbs2}, $indx_num, @indices_rlv);

	    if ($if_exist) {
		push @ovlping, $gpb_in_1;
	    }

	    if ( ($if_exist and ($indx_qu>0)) or (!($if_exist) and ($indx_qu >= 0)) ) { # Examine the left-neighbor #
		my $gpb_in_2 = $if_exist ? $invlvd_gpbs2->[$indx_qu-1] : $invlvd_gpbs2->[$indx_qu] ;
		if ($gpb_in_2 + 1 == $gpb_in_1) {
		    push @nn_pairs, [$gpb_in_1, $gpb_in_2];
		}
	    }
	    if ($indx_qu < $#{$invlvd_gpbs2}) { # Examine the right-neighbor #
		my $gpb_in_2 = $invlvd_gpbs2->[$indx_qu + 1];
		if ($gpb_in_1 + 1 == $gpb_in_2) {
		    push @nn_pairs, [$gpb_in_1, $gpb_in_2];
		}
	    }
	}				# END of the 1st foreach-loop (over gap-pattern blocks involved in the 1st boundary). #

	unless ((@ovlping>0) or (@nn_pairs>0)) {
		# If the two boundaries NEITHER overlap NOR neighbor each other,
		# they are regarded as "independent" of each other.
		# (Actually, NOT CLEARLY judged as "dependent" on each other.)

	    return 0;
	}

		# If the boundaries share the same branch, the boundaries ARE unconditionally dependent of each other. #

	my $equiv_br1 = $br2equiv->{$br1};
	if ( ($br1 == $br2) or ((defined $equiv_br1) and ($equiv_br1 == $br2)) ) { return 1; }

		### Examine the phylogenetic relationship between the lower-ends of the boundary branches. ###

	my ($common_ancestors, $div1, $div2) = fetch_common_ancestors ($br1, $br2, %{$node2pa});

	my $last_common_ancestor = $common_ancestors->[$#{$common_ancestors}];


		# If either of the boundaries is of 'insertion' type (and is NOT upstream of the other boundary),
		# the boundaries ARE dependent of each other. #

	if ( ((@{$div1}>0) and ($up_or_lw1 eq 'U')) or ((@{$div2}>0) and ($up_or_lw2 eq 'U')) ) { return 1; }


		# Create a set of relevant gap-pattern blocks. #

	my @rlv_blks = @ovlping;
	foreach my $pair (@nn_pairs) { push @rlv_blks, @{$pair}; }

	my $nr_rlv_blks = nr_nmarray (@rlv_blks);
	my $ct_rlv_blks = @{$nr_rlv_blks};


		### Finally, examine the existence of a node separating the two boundaries. ###

	pop @{$div1}; pop @{$div2}; # Remove the lower-ends of the boundary-branches themselves
				    # if they are included in the lineage-specific sets of ancestral nodes. #

	my $if_dependent = 1;

	foreach my $node ($last_common_ancestor, @{$div1}, @{$div2}) { # 2nd foreach-loop (over intervening nodes). #

	    my $blwise_gapstats = $node2blwise_gstat->{$node};

	    my $if_gap = 0;
	    foreach my $gpb (@{$nr_rlv_blks}) { # Inner foreach-loop (over relevant gap-pattern blocks). #
		if ($blwise_gapstats->[$gpb] eq $char_abs) {
		    $if_gap = 1;
		    last;
		}
	    }				# End of the inner foreach-loop (over relevant gap-pattern blocks). #

	    unless ($if_gap) {
		# NO relevant block is occupied with a gap at this node,
		# which means that the node separates the two boundaries. #

		$if_dependent = 0;
		last;
	    }

	}					# End of the 2nd foreach-loop (over intervening nodes). #

	return $if_dependent ;

}	# END of subroutine 'exam_pw_bd_dependence'. #


# pw_merge_indel_blocks:
#	This satellite subroutine of 'partit_into_eff_indep_indel_blocks' merges a pair of indel blocks.
#	[CREATION STARTED on Mar 4 (Sat), 2017, ENDED on Mmm dd (www), 2017]
#	
# INPUT: (\@indel_block1, \@indel_block2), where
#
#	@indel_block1/2 = ($lm_gpb1/2, $rm_gpb1/2, \@set_delimiting_bds1/2) gives information on the 1st/2nd indel block to be merged, with
#	$lm/rm_gpb1/2 being the leftmost/rightmost gap-pattern block,
#	@set_delimiting_bds1/2 storing the (not necessary complete) delimiting boundaries of the 1st/2nd indel block,
#	whose element is formatted as [$br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs]
#
#
# OUTPUT: \@mgd_indel_block, where
#	@mgd_indel_block = ($lm_gpb_mgd, $rm_gpb_mgd, \@set_delimiting_bds_mgd), where
#	$lm_gpb_mgd = Min{$lm_gpb1, $lm_gpb2},
#	$rm_gpb_mgd = Max{$rm_gpb1, $rm_gpb2},
#	@set_delimiting_bds_mgd is the union of @set_delimiting_bds1 and @set_delimiting_bds2.
#
sub pw_merge_indel_blocks (\@\@) {

	my ($indel_block1, $indel_block2) = @_;

	my ($lm_gpb1, $rm_gpb1, $set_delimiting_bds1) = @{$indel_block1};
	my ($lm_gpb2, $rm_gpb2, $set_delimiting_bds2) = @{$indel_block2};

	my $lm_gpb_mgd = ($lm_gpb1 < $lm_gpb2) ? $lm_gpb1 : $lm_gpb2 ;
	my $rm_gpb_mgd = ($rm_gpb1 > $rm_gpb2) ? $rm_gpb1 : $rm_gpb2 ;

	my @indices_srt = (2, 3, 0, 1); my @methods_srt = ('+', '+', '+', 'a');

	my $set_delimiting_bds_mgd = union_array_of_arrays (@{$set_delimiting_bds1}, @{$set_delimiting_bds2}, @indices_srt, @methods_srt);

	my @mgd_indel_block = ($lm_gpb_mgd, $rm_gpb_mgd, $set_delimiting_bds_mgd);

	return \@mgd_indel_block;

}	# END of the subroutine 'pw_merge_indel_blocks'. #
 

# if_idb_includes_root:
#	This satellite subroutine of 'partit_into_eff_indep_indel_blocks' examines whether
#	the input indel block includes the root (i.e., top node) or not.
#
#	[Theory]
#	If the indel history and the alignment are phylogenetically consistent with each other,
#	all sites "above" an insertion must be occupied with gaps (or "absence" marks),
#	thus, if the block is bounded by at least one insertion,
#	it MUST include a root.
#	On the other hand, if the block is bounded solely by deletions,
#	it MUST NOT include a root.
#
#	[CREATED on March 7 (Tue), 2017]
#
# INPUT: (\@set_delimiting_bds), where
#
#	@set_delimiting_bds stores information on the boundaries delimiting the indel block, with
#	@{$set_delimiting_bds[$bd]} = ($br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs) being for the $bd th boundary, with
#	$up_or_lw = 'U'/'L' if the upper/lower-end of the branch ($br) is the indel block, 
#
# OUTPUT: $if_incl_root, which is
#	1 if the indel block includes root,
#	0 otherwise.
#
sub if_idb_includes_root (\@) {

	my ($set_delimiting_bds) = @_;

	my $if_incl_root = 0;

	my $ct_bds = @{$set_delimiting_bds};
	for (my $bd=0; $bd < $ct_bds; $bd++) { # Main for-loop (over delimiting boundaries). #

	    my $delim_bd = $set_delimiting_bds->[$bd];
	    my ($br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs) = @{$delim_bd};

	    if ($up_or_lw eq 'U') { # The boundary is an insertion. #
		$if_incl_root = 1;
		last;
	    }

	}			# END of the main for-loop (over delimiting boundaries). #

	return $if_incl_root;

}	# END of the subroutine 'if_idb_includes_root'. # 


# partit_into_eff_indep_indel_blocks:
#	This subroutine partitions each gapped segment into effectively independent indel blocks,
#	each of which accommodates a set of effectively independent indel histories, as well as
#	a "null" block, which accommodates no indel history,
#	according to the inputs and outputs of subroutines, 'define_off_limits' and 'partit_into_gblocks_seqblocks'.
#	(It comprizes STEP 3 (B) in the 'workflow_for_indel_histories&multiplication_factors.rtf'.)
#	
#	[CREATION STARTED on Feb 25 (Sat), 2017, ENDED on mmm dd (www), 2017.]
#
# [ INPUT: (\@list_off_limits, \@list_gblocks, \@list_seqblocks,
#	\@gpblocks_in_cls, \%br2indels, \%node2blwise_gstat, $topnode_id, \%node2pa, \%node2ch, $char_abs), where ] ... OBSOLETE
#
# INPUT: (\@gpblocks_in_cls, \%br2indels, \%node2blwise_gstat, $topnode_id, \%node2pa, \%node2ch, $char_abs, \%br2equiv), where
#
#	[
#	@{$list_off_limits[$ol]} = ($br, $up_or_lw, $start, $end, \@invlvd_gpbs) defines the $ol th off-limit boundary in the gapped segment in question.
#	$br is the ID of the branch delimiting the off-limit region,
#	$up_or_lw = 'U'/'L' if the upper/lower-end of $br is off-limit, 
#	$start/$end is the start/end coordinate of the boundary,
#	@invlvd_gpbs lists the gap-pattern-blocks involved. 
#
#	@{$list_gblocks{$gb]} = ($br, $start, $end, \@invlvd_gpbs, \@nesting) gives information on the $gb th gap-block.
#	$br is the ID of the branch delimiting the gap-block, $start/$end is the start/end coordinate of the gap-block,
#	@invlvd_gpbs lists the gap-pattern blocks involved, @nesting lists the sequence-blocks nesting the gap-block.
#
#	@{$list_seqblocks[$sb]} = ($br, $start, $end, \@invlvd_gpbs, \@nested) gives information on the $sb th sequence-block.
#	$br is the ID of the branch delimiting the sequence-block, $start/$end is the start/end coordinate of the sequence-block,
#	@invlvd_gpbs lists the gap-pattern blocks involved, @nested lists the gap-blocks nested in the sequence-block.
#	] ... NO LONGER NECESSARY FOR THIS SUBROUTINE ...
#
#	@{$gpblocks_in_cls[$gpb]} = ($indx_gp, $len_gpb, $start, $end) is for the $gpb th gap-pattern block in the subject gapped segment, with
#	$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#	$len_gpb being the length (in terms of #{columns}) of the gap-pattern block,
#	and $start/$end being the starting/ending coordinate of the gap-pattern block (in the 0-based full-closed convention);
#
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved);
#
#	%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
#	with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node;
#
#	$topnode_id is the ID of the top-node (i.e., root) of the tree;
#
#	%node2pa = ($node_id => $parent_node_id, ...) for all the non-root nodes;
#	%node2ch = ($node_id => \@children, ...) for all internal nodes, with @children listing the IDs of all child nodes;
#	$char_abs is the character specifying the absence of the residue at the site. (Usually, it's the same as the 'gap' character.)
#	%br2equiv = ($branch_ID => $equivalent_branch_id, ...) for branches with equivalents.
#
# OUTPUT: (\@eff_indep_indel_blocks, \@indices_incl_root), where
#
#	[ @{$eff_indep_indel_blocks[$eiidb]} = ($br, $up_or_lw, $start, $end, \@invlvd_gpbs, \@invlvd_gbs, \@invlvd_sbs) ] ... OBSOLETE because this format is NOT general enough to accommodate all types of indel blocks....
#
#	@{$eff_indep_indel_blocks[$eiidb]} = (\@set_delimiting_bds, $start, $end, \@union_invlvd_gpbs, \@invlvd_gbs, \@invlvd_sbs)
#	gives information on the $eiidb th effectively independent indel block.
#
#	@set_delimiting_bds stores information on all boundaries delimiting the indel block,
#	with each element being of the form [$br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs]
#	$br is the ID of the branch delimiting the indel block,
#	$up_or_lw = 'U'/'L' if the upper/lower-end of $br is the indel block, 
#	$lm/rm_gpb is the leftmost/rightmost gap-pattern block involved,
#	@invlvd_gpbs lists the gap-pattern-blocks involved in this boundary.
#
#	$start/$end is the start/end coordinate of the boundary,
#	@union_invlvd_gpbs lists all the gap-pattern-blocks involved in this indel block,
#	@invlvd_gbs lists the gap-blocks involved (EMPTY at this stage),
#	@invlvd_sbs is the sequence-blocks involved (EMPTY at this stage).
#
#	@indices_incl_root stores the indices of the effectively independent indel blocks that include root (top node).
#	It is empty if no effectively independent indel block includes root.
#		... It may be better to relegate this function to another subroutine ... 
#
# [ sub partit_into_eff_indep_indel_blocks (\@\@\@\@\%\%$\%\%$\%) {
#
#	my ($list_off_limits, $list_gpblocks, $list_seqblocks,
#	    $gpblocks_in_cls, $br2indels, $node2blwise_gstat, $topnode_id, $node2pa, $node2ch, $char_abs, $br2equiv) = @_; } ... OBSOLETE]
#
sub partit_into_eff_indep_indel_blocks (\@\%\%$\%\%$\%) {
 
	my ($gpblocks_in_cls, $br2indels, $node2blwise_gstat, $topnode_id, $node2pa, $node2ch, $char_abs, $br2equiv) = @_;

	my $ct_gpblocks = @{$gpblocks_in_cls};


		# Create a list of boundaries (each defined by an indel event), sorted from left to right along the MSA. #

	my @list_boundaries = (); # element = [$br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs] # $up_or_lw = 'U'/'L' if the upper/lower-end of $br is the indel block. #

	foreach my $br (sort keys %{$br2indels}) { # Preliminary outer foreach-loop (over branches along which indels occurred). #

	    my $indels_along_br = $br2indels->{$br};

	    foreach my $indel (@{$indels_along_br}) { # Inner foreach-loop (over indels along $br). #

		(@{$indel}>0) or next; # Skip an empty entry.
		my ($state_ue, $state_le, $indices_blk) = @{$indel};

		my ($lm_gpb, $rm_gpb) = @{$indices_blk}[0, $#{$indices_blk}];
		my $invlvd_gpbs = $indices_blk;

		my $up_or_lw = ($state_ue eq $char_abs) ? 'U' : 'L' ;

		push @list_boundaries, [$br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs];

	    }				# End of the inner foreach-loop (over indels along $br). #

	}				# End of the preliminary outer foreach-loop (over branches along which indels occurred). #

	my @indices_srt = (2, 3); my @methods_srt = ('+', '+');

	my $std_list_boundaries = multiple_sort (@list_boundaries, @indices_srt, @methods_srt);

	my $totct_boundaries = @{$std_list_boundaries};

		# Construct effectively independent indel blocks by examining the boundaries from left to right. #

	my @preset_eff_indep_indel_blocks = (); # element = [$lm_gpb, $rm_gpb, \@set_delimiting_bds] # @set_delimiting_bds stores elements of @list_boundaries that are used for constructing the indel block. #


	for (my $bd = 0; $bd<$totct_boundaries; $bd++) { # Main outer for-loop (over boundaries to be used). #

	    my $boundary_sbj = $std_list_boundaries->[$bd];

	    my ($br_sbj, $up_or_lw_sbj, $lm_gpb_sbj, $rm_gpb_sbj, $invlvd_gpbs_sbj) = @{$boundary_sbj};

		# First, examine whether or not any pre-defined indel block can incorporate the subject boundary. #

	    my @candidate_pre_blks = ();

	    my $ct_pre_blks = @preset_eff_indep_indel_blocks;

	    for (my $pre=0; $pre< $ct_pre_blks; $pre++) { # Middle for-loop (over pre-defined indel blocks). #

		my $info_pre_blk = $preset_eff_indep_indel_blocks[$pre];

		my ($lm_gpb_pre, $rm_gpb_pre, $set_delimiting_bds_pre) = @{$info_pre_blk};

		if ( ($rm_gpb_pre + 1 < $lm_gpb_sbj) or ($rm_gpb_sbj + 1 < $lm_gpb_sbj) ) {

			# If the pre-set indel block in question neither overlap nor nearest-neighbor the subject boundary,
			# then, the former will not incorporate the latter, at least at this stage. 

		    next;
		}

			# Examine whether the pre-set indel block can inforporate the boundary. #

		my $if_dependent = 0;

		my $ct_delim_bds_pre = @{$set_delimiting_bds_pre};

		for (my $dlbd = 0; $dlbd < $ct_delim_bds_pre; $dlbd++) { # Inner for-loop (over delimiting boudaries). #

		    my $delim_bd_pre = $set_delimiting_bds_pre->[$dlbd];

		    my $if_pw_dependent = exam_pw_bd_dependence (@{$boundary_sbj}, @{$delim_bd_pre}, %{$node2blwise_gstat}, $topnode_id, %{$node2pa}, %{$node2ch}, $char_abs, %{$br2equiv});
		    unless (defined $if_pw_dependent) {
			print STDOUT "[ERROR 'partit_into_eff_indep_indel_blocks'] The subroutine 'exam_pw_bd_dependence' FAILED between the subject boundary (the $bd th) and the $dlbd th boundary of the $pre th pre-set indel block: $! !\n\n";
			return (undef, undef);
		    }

		    if ($if_pw_dependent) {
			$if_dependent = 1;
			last;
		    }

		}						# End of inner for-loop (over delimiting boundaries). #

		if ($if_dependent) {
		    push @candidate_pre_blks, $pre;
		}

	    }				# End of the middle for-loop (over pre-defined indel blocks). #


	    my $new_eff_indep_indel_block = [$lm_gpb_sbj, $rm_gpb_sbj, [$boundary_sbj] ]; # A new block defined solely by the subject boundary. #

	    unless (@candidate_pre_blks >0) {

			# If no pre-defined block can incorporate the subject boundary,
			# then, store the new block defined solely by the subject boundary.

		push @preset_eff_indep_indel_blocks, $new_eff_indep_indel_block;

		next;
	    }

			# Create a new "larger" block by incorporating the new boundary into some pre-defined blocks. #

	    my $mgd_indel_blk = $preset_eff_indep_indel_blocks[ $candidate_pre_blks[0] ];

	    for (my $i=1; $i<@candidate_pre_blks; $i++) { # 2nd middle for-loop (over candidate pre-defined indel blocks). #

		my $indx_cand_blk = $candidate_pre_blks[$i];

		$mgd_indel_blk = pw_merge_indel_blocks (@{$mgd_indel_blk}, @{$preset_eff_indep_indel_blocks[$indx_cand_blk]});
		unless (defined $mgd_indel_blk) {
		    print STDOUT "[ERROR 'partit_into_eff_indep_indel_blocks'] The subroutine 'pw_merge_indel_blocks' FAILED when creating a new indel block when merging the $indx_cand_blk th block with the 'laerge' block to incorporate the $bd th boundary: $! !\n\n";
		    return (undef, undef);
		}

	    }					# End of the 2nd middle for-loop (over candidate pre-defined indel blocks). #

	    $new_eff_indep_indel_block = pw_merge_indel_blocks (@{$mgd_indel_blk}, @{$new_eff_indep_indel_block});
	    unless (defined $new_eff_indep_indel_block) {
		print STDOUT "[ERROR 'partit_into_eff_indep_indel_blocks'] The subroutine 'pw_merge_indel_blocks' FAILED when incorporating the $bd th boundary to the 'union' of indel blocks that can incorporate the boundary: $! !\n\n";
		return (undef, undef); 
	    }


			# Create a new set of preset indel blocks by removing the candidate blocks and adding the newly created block. #

	    my @indices_pre_blocks = (0 .. $ct_pre_blks-1);
	    my ($common, $has_to_be_empty, $remaining) = diff_nmarray (@candidate_pre_blks, @indices_pre_blocks);

	    my @newset_preset_blks = @preset_eff_indep_indel_blocks[@{$remaining}];

	    push @newset_preset_blks, $new_eff_indep_indel_block;

		# Update the set of preset indel blocks. #

	    @preset_eff_indep_indel_blocks = @newset_preset_blks;

	}				# End of the main outer for-loop (over boundaries to be used). #

		## To finish, "Refine" the preset blocks to create the output set of effectively independent indel blocks. ##

	## my @preset_eff_indep_indel_blocks = (); # element = [$lm_gpb, $rm_gpb, \@set_delimiting_bds] # @set_delimiting_bds stores elements of @list_boundaries that are used for constructing the indel block. #
	## @list_boundaries = (); # element = [$br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs] # $up_or_lw = 'U'/'L' if the upper/lower-end of $br is the indel block. #

	my @indices_srt2 = (0, 1); my @methods_srt2 = ('+', '+');
	my $std_preset_eff_indep_indel_blocks = multiple_sort (@preset_eff_indep_indel_blocks, @indices_srt2, @methods_srt2);

	my $ct_eff_indep_indel_blks = @{$std_preset_eff_indep_indel_blocks};
	my @eff_indep_indel_blocks = ();
	my @indices_incl_root = ();

	for (my $eiidb=0; $eiidb < $ct_eff_indep_indel_blks; $eiidb++) { # Trailing outer foreach-loop (over RAW effectively independent indel blocks). # 

	    my $preset_block  = $std_preset_eff_indep_indel_blocks->[$eiidb];

	    my ($lmm_gpb, $rmm_gpb, $set_delimiting_bds) = @{$preset_block};

	    my $union_invlvd_gpbs = [];

	    foreach my $delim_bd (@{$set_delimiting_bds}) { # Inner foreach-loop (over delimiting boundaries). #

		my ($br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs) = @{$delim_bd};

		$union_invlvd_gpbs = union_nmarray (@{$union_invlvd_gpbs}, @{$invlvd_gpbs});

	    }						# End of the inner foreach-loop (over delimiting boundaries). #

	    my $start =$gpblocks_in_cls->[$lmm_gpb][2];
	    my $end = $gpblocks_in_cls->[$rmm_gpb][3];
		# @{$gpblocks_in_cls[$gpb]} = ($indx_gp, $len_gpb, $start, $end) is for the $gpb th gap-pattern block in the subject gapped segment

	    my @invlvd_gbs = my @invlvd_sbs = ();

	    push @eff_indep_indel_blocks, [$set_delimiting_bds, $start, $end, $union_invlvd_gpbs, \@invlvd_gbs, \@invlvd_sbs];

		# @{$eff_indep_indel_blocks[$eiidb]} = (\@set_delimiting_bds, $start, $end, \@union_invlvd_gpbs, \@invlvd_gbs, \@invlvd_sbs)
		# gives information on the $eiidb th effectively independent indel block.
		#
		# @set_delimiting_bds stores information on all boundaries delimiting  the indel block,
		# with each element being of the form [$br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs]

	    my $if_incl_root = if_idb_includes_root (@{$set_delimiting_bds});
	    unless (defined $if_incl_root) {
		print STDOUT "[ERROR 'partit_into_eff_indep_indel_blocks'] The subroutine 'if_idb_includes_root' FAILED on the $eiidb th effectively independent indel block: $! !\n\n";
		return (undef, undef);
	    }
	    if ($if_incl_root) { push @indices_incl_root, $eiidb; }

	}						# END of the trailing outer foreach-loop (over RAW effectively independent indel blocks). #

	return (\@eff_indep_indel_blocks, \@indices_incl_root) ;

}	# END of the subroutine 'partit_into_eff_indep_indel_blocks'. #


# find_invlvd_in_eff_indep_indel_blocks:
#	This subroutine examines what gap-blocks and sequence-blocks are involved in each of the effectively independent indel blocks.
#	(It comprizes STEP 3 (C) in the 'workflow_for_indel_histories&multiplication_factors.rtf'.)
#
#	[DEFINITIONS]
#	(1) A gap-block is regarded as "involved in" an indel block, if and only if
#	   the delimiting event of the former is one of the delimiting indel events of the latter.
#
#	(2) A sequence-block is regarded as "involved in" an indel block, if and only if
#	   one of the delimiting boundaries of the former (including those of the gap-blocks nested in it)
#	   is also one of the delimiting boundaries of the latter.
#
#	[CREATION STARTED on Mar 1 (Wed), 2017, ENDED on Mmm dd (www), 2017.]
#
#
# INPUT: (\@eff_indep_indel_blocks,  \@list_gblocks, \@list_seqblocks, \@gpblocks_in_cls,
#		$top_node_id, \%node2pa, \%node2ch), where
#
#	[ @{$eff_indep_indel_blocks[$eiidb]} = ($br, $up_or_lw, $start, $end, \@invlvd_gpbs, \@invlvd_gbs, \@invlvd_sbs) ] ... OBSOLETE because this format is NOT general enough to accommodate all types of indel blocks....
#
#	@{$eff_indep_indel_blocks[$eiidb]} = (\@set_delimiting_bds, $start, $end, \@union_invlvd_gpbs, \@invlvd_gbs, \@invlvd_sbs)
#	gives information on the $eiidb th effectively independent indel block.
#
#	@set_delimiting_bds stores information on all boundaries delimiting the indel block,
#	with each element being of the form [$br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs]
#	$br is the ID of the branch delimiting the indel block,
#	$up_or_lw = 'U'/'L' if the upper/lower-end of $br is the indel block, 
#	$lm/rm_gpb is the leftmost/rightmost gap-pattern block involved,
#	@invlvd_gpbs lists the gap-pattern-blocks involved in this boundary.
#
#	$start/$end is the start/end coordinate of the boundary,
#	@union_invlvd_gpbs lists all the gap-pattern-blocks involved in this indel block,
#	@invlvd_gbs lists the gap-blocks involved (EMPTY at this stage),
#	@invlvd_sbs is the sequence-blocks involved (EMPTY at this stage).
#
#	@{$list_gblocks[$gb]} = ($br, $up_or_lw, $start, $end, \@invlvd_gpbs, \@nesting) gives information on the $gb th gap-block.
#	$br is the ID of the branch delimiting the gap-block,
#	$up_or_lw = 'U'/'L' if the gap-block extends on the upper/lower side of the branch,
#	$start/$end is the start/end coordinate of the gap-block,
#	@invlvd_gpbs lists the gap-pattern blocks involved,
#	@nesting lists the sequence-blocks nesting the gap-block.
#
#	@{$list_seqblocks[$sb]} = ($br, $up_or_lw, $start, $end, \@invlvd_gpbs, \@nested) gives information on the $sb th sequence-block.
#	$br is the ID of the primary branch delimiting the sequence-block,
#	$up_or_lw = 'U'/'L' if the sequence block extends on the upper/lower side of the branch,
#	$start/$end is the start/end coordinate of the sequence-block,
#	@invlvd_gpbs lists the gap-pattern blocks involved,
#	@nested lists the gap-blocks nested in the sequence-block.
#
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved);
#
#	$topnode_id is the ID of the top-node (i.e., root) of the tree;
#
#	%node2pa = ($node_id => $parent_node_id, ...) for all the non-root nodes;
#	%node2ch = ($node_id => \@children, ...) for all internal nodes, with @children listing the IDs of all child nodes.
#
#
# OUTPUT: 1 if the subroutine ended successfully,
#	  0 if it failed.
#
#	* This subroutine fills in @invlvd_gbs and @invlvd_sbs in each element of @eff_indep_indel_blocks.
#
sub find_invlvd_in_eff_indep_indel_blocks (\@\@\@\@$\%\%) {

	my ($eff_indep_indel_blocks, $list_gblocks, $list_seqblocks, 
		$gpblocks_in_cls, $top_node_id, $node2pa, $node2ch) = @_; 

	my $ct_eiidbs = @{$eff_indep_indel_blocks};
	my $ct_gblocks = @{$list_gblocks};
	my $ct_seqblocks = @{$list_seqblocks};
	my $ct_gpblocks = @{$gpblocks_in_cls};

		# Preliminary process 1: Create an association from a gap-block to its index. #
	my %gblock2indx; # = ("$br;$up_or_lw;list,invlvd,gpbs" => $index_gb, ...)

	for (my $gb = 0; $gb < $ct_gblocks; $gb++) { # 1st preliminary for-loop (over gap-blocks). #

	    my ($br, $up_or_lw, $start, $end, $invlvd_gpbs, $nesting) = @{$list_gblocks->[$gb]};

	    my $cnct_invlvd_gpbs = join (',', @{$invlvd_gpbs});
	    my $gb_identifier = join (';', $br, $up_or_lw, $cnct_invlvd_gpbs);

	    $gblock2indx{$gb_identifier} = $gb;

	}				# End of the 1st preliminary for-loop (over gap-blocks). #

		# Preliminary process 2: Create an association from a sequence-block to its index. #
	my %seqblock_bd2indx; # = ("$br;$up_or_lw;list,invlvd,gpbs" => $indx_sb, ...) # NOTE that $up_or_lw of a sequence block is the OPPOSITE of that of a gap-block. #
			# (This hash table must work because each gap-block can be nested in only one sequence-block, if at all.) #

	for (my $sb = 0; $sb < $ct_seqblocks; $sb++) { # 2nd preliminary for-loop (over sequence-blocks). #

	    my ($br, $up_or_lw, $start, $end, $invlvd_gpbs, $nested) = @{$list_seqblocks->[$sb]};

	    my $cnct_invlvd_gpbs = join (',', @{$invlvd_gpbs});
	    my $sb_bd_identifier = join (';', $br, $up_or_lw, $cnct_invlvd_gpbs);
	    $seqblock_bd2indx{$sb_bd_identifier} = $sb;

	    foreach my $gb2 (@{$nested}) { # Inner foreach-loop (over gap-blocks nested in $sb). #

		my ($br2, $up_or_lw2, $start2, $end2, $invlvd_gpbs2, $nesting2) = @{$list_gblocks->[$gb2]};
		$up_or_lw2 = ($up_or_lw2 eq 'L') ? 'U' : 'L'; # Flip from the flag of a gap-block to the flag of a sequence-block. #
		my $cnct_invlvd_gpbs2 = join (',', @{$invlvd_gpbs2});
		my $sb_bd_identifier2 = join (';', $br2, $up_or_lw2, $cnct_invlvd_gpbs2);

		$seqblock_bd2indx{$sb_bd_identifier2} = $sb;

	    }			# End of the inner foreach-loop (over gap-blocks nested in $sb). #

	}				# End of the 2nd preliminary for-loop (over sequence-blocks). #


		# Main Process: Find the gap-blocks and sequence-blocks involved in each indel block. #

	for (my $idb = 0; $idb < $ct_eiidbs; $idb++) {  # Main for-loop (over indel blocks). #

	    my $indel_block = $eff_indep_indel_blocks->[$idb];
	    my ($set_delimiting_bds, $start, $end, $union_invlvd_gpbs, $invlvd_gbs, $invlvd_sbs) = @{$indel_block};

	    my @invlvd_gbs = ();
	    my @invlvd_sbs = ();

	    my $ct_delimiting_bds = @{$set_delimiting_bds};

	    for (my $db = 0; $db< $ct_delimiting_bds; $db++) { # Inner for-loop (over delimiting boundaries). #

		my $delim_bds = $set_delimiting_bds->[$db];
		my ($br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs) = @{$delim_bds};
			#
			# @set_delimiting_bds stores information on all boundaries delimiting the indel block,
			# with each element being of the form [$br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs]

		my $cnct_invlvd_gpbs = join (',', @{$invlvd_gpbs});
		my $gb_identifier = join (';', $br, $up_or_lw, $cnct_invlvd_gpbs);
		my $indx_gb = $gblock2indx{$gb_identifier};
		if (defined $indx_gb) {
		    push @invlvd_gbs, $indx_gb;
		} else {
		    print STDOUT "[WARNING 'find_invlvd_in_eff_indep_indel_blocks'] NO GAP-BLOCK INDEX is associated with the $db th boundary of the $idb th indel block (identifier: '$gb_identifier'): $! !\n\n";
		}

		my $up_or_lw2 = ($up_or_lw eq 'L') ? 'U' : 'L' ;
		my $sb_bd_identifier = join (';', $br, $up_or_lw2, $cnct_invlvd_gpbs);
		my $indx_sb = $seqblock_bd2indx{$sb_bd_identifier};
		if (defined $indx_sb) {
		    push @invlvd_sbs, $indx_sb;
		}

	    }					# End of the inner for-loop (over delimiting boundaries). #

		# Updated on Mar 22nd (Wed), 2017. #
	    my $nr_invlvd_gbs = (@invlvd_gbs <2) ? \@invlvd_gbs : nr_nmarray (@invlvd_gbs) ;
	    my $nr_invlvd_sbs = (@invlvd_sbs <2) ? \@invlvd_sbs : nr_nmarray (@invlvd_sbs) ;
	    $indel_block->[4] = $nr_invlvd_gbs;
	    $indel_block->[5] = $nr_invlvd_sbs;
	    ##$indel_block->[4] = \@invlvd_gbs;
	    ##$indel_block->[5] = \@invlvd_sbs;

	}				# End of the main for-loop (over indel blocks). #


	return 1;

}	# END of the subroutine 'find_invlvd_in_eff_indep_indel_blocks'. #



# mk_node2blwise_gstat:
#	This subroutine receives an element of the main output of 'partit_into_eff_indep_indel_blocks',
#	and creates an association table from each node to a gap-state corresponding to each indel block.
#
#	[THEORY]
#	Assuming that the indel history and the MSA are phylogenetically consistent with each other,
#	the inserted blocks are all occupied by "absence" at all the nodes "above" the insertion event,
#	and the deleted blocks are all occupied by "absence" at all the nodes "below" the deletion event.
#
#	[CREATION STARTED on March 14 (Tue), 2017, ENDED on March dd (Tue), 2017.]
#
# INPUT: (\@indel_block, $ct_gpblocks_in_cls, \@set_all_nodes, \%node2ch, $char_abs, $char_pres), where
#
#	@indel_block = (\@set_delimiting_bds, $start, $end, \@union_invlvd_gpbs, \@invlvd_gbs, \@invlvd_sbs)
#	gives information on the input indel block.
#
#	@set_delimiting_bds stores information on all boundaries delimiting the indel block,
#	with each element being of the form [$br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs]
#	$br is the ID of the branch delimiting the indel block,
#	$up_or_lw = 'U'/'L' if the upper/lower-end of $br is the indel block, 
#	$lm/rm_gpb is the leftmost/rightmost gap-pattern block involved,
#	@invlvd_gpbs lists the gap-pattern-blocks involved in this boundary.
#
#	$start/$end is the start/end coordinate of the boundary,
#	@union_invlvd_gpbs lists all the gap-pattern-blocks involved in this indel block,
#	@invlvd_gbs lists the gap-blocks involved (EMPTY at this stage),
#	@invlvd_sbs is the sequence-blocks involved (EMPTY at this stage).
#
#	[ @{$gpblocks_in_cls[$gpb]} = ($indx_gp, $len_gpb, $start, $end) is for the $gpb th gap-pattern block in the subject gapped segment, with
#	$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#	$len_gpb being the length (in terms of #{columns}) of the gap-pattern block,
#	and $start/$end being the starting/ending coordinate of the gap-pattern block (in the 0-based full-closed convention). ] ... Probably, UNNECESSARY ...
#	$ct_gpblocks_in_cls is the number of gap-pattern blocks in the gapped segment.
#
#	@set_all_nodes stores the IDs of all nodes;
#	%node2ch = ($node_id => \@children, ...) for all internal nodes;
#	$char_abs/pres is the character representing the "absence"/"presence" of a sub-sequence fully occupying the gap-pattern block.
#
# OUTPUT: \%node2blwise_gstat, where
#
#	%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
#	with $blwise_gapstats[$gpb] is the gap-state ($char_abs/pres) of the $gpb th gap-pattern block at the node.
#
sub mk_node2blwise_gstat (\@$\@\%$$) {

	my ($indel_block, $ct_gpbs_in_cls, $set_all_nodes, $node2ch, $char_abs, $char_pres) = @_;

		# (Preliminary Process) Initialize the output hash. #

	my @all_pres = ();
	for (1 .. $ct_gpbs_in_cls) { push @all_pres, $char_pres; }

	my %node2blwise_gstat;
	foreach my $node (@{$set_all_nodes}) { # Preliminary foreach-loop (over all nodes). #
	    my @cp = @all_pres;
	    $node2blwise_gstat{$node} = \@cp;
	}				# End of the preliminary foreach-loop (over all nodes). #


		## (Main Process) Examine all the boundaries delimiting the indel block. ##

	my ($set_delimiting_bds, $start, $end, $union_invlvd_gpbs, $invlvd_gbs, $invlvd_sbs) = @{$indel_block};
	my $ct_dlm_bds = @{$set_delimiting_bds};

	for (my $db=0; $db < $ct_dlm_bds; $db++) { # Main for-loop (over delimiting boundaries). #

	    my ($br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs) = @{$set_delimiting_bds->[$db]};

	    my $all_offsprings = fetch_all_offsprings_nr ($br, %{$node2ch});

	    my $nodes_absence;
	    if ($up_or_lw eq 'U') {
		# The indel block is on the upper-end of the branch. #

		my ($all_offsprings2, $null, $all_upstreams) = diff_nmarray (@{$all_offsprings}, @{$set_all_nodes});
		$nodes_absence = $all_upstreams;
	    } else {
		# The indel block is on the lower-end of the branch. #

		$nodes_absence = $all_offsprings;
	    }

			# In relevant blocks at relevant nodes,
			# flip the "presence" states to the "absence" states.

	    foreach my $node (@{$nodes_absence}) { # Middle foreach-loop (over nodes with "absence" states). #

		my $blwise_gstat = $node2blwise_gstat{$node};
		foreach my $gpb (@{$invlvd_gpbs}) { $blwise_gstat->[$gpb] = $char_abs; }

	    }				# END of the middle foreach-loop (over nodes with "absence" states). #

	}				# End of the main for-loop (over delimiting boundaries). #

	return \%node2blwise_gstat;

}	# END of the subroutine 'mk_node2blwise_gstat'. #


# mk_br2indels:
#	This subroutine receives an element of the main output of "partit_into_eff_indep_indel_blocks",
#	and create an association from each branch to a set of indels.
#
#	[CREATED on May 10th (Wed), 2017.]
#
# INPUT: (\@indel_block, $char_abs, $char_pres), where
#
#	@indel_block = (\@set_delimiting_bds, $start, $end, \@union_invlvd_gpbs, \@invlvd_gbs, \@invlvd_sbs)
#	gives information on the input indel block.
#
#	@set_delimiting_bds stores information on all boundaries delimiting the indel block,
#	with each element being of the form [$br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs]
#	$br is the ID of the branch delimiting the indel block,
#	$up_or_lw = 'U'/'L' if the upper/lower-end of $br is the indel block, 
#	$lm/rm_gpb is the leftmost/rightmost gap-pattern block involved,
#	@invlvd_gpbs lists the gap-pattern-blocks involved in this boundary.
#
#	$start/$end is the start/end coordinate of the boundary,
#	@union_invlvd_gpbs lists all the gap-pattern-blocks involved in this indel block,
#	@invlvd_gbs lists the gap-blocks involved (EMPTY at this stage),
#	@invlvd_sbs is the sequence-blocks involved (EMPTY at this stage).
#
#	$char_abs/pres is a character indicating that a residue is absent/present at the site (block).
#
# OUTPUT: \%br2indels, where
#
#	%br2indels = ($branch_id => \@indels, ...) is the map of the indel events onto branches,
#	with @{$indels[$ev]} = ($state_ue, $state_le, \@invlvd_gbs).
#
sub mk_br2indels (\@$$) {

	my ($indel_block, $char_abs, $char_pres) = @_;

	my ($set_delimiting_bds, $start, $end, $union_invlvd_gpbs, $invlvd_gbs, $invlvd_sbs) = @{$indel_block};
	my $ct_dlm_bds = @{$set_delimiting_bds};

	my %br2indels;

	for (my $db=0; $db < $ct_dlm_bds; $db++) { # Main for-loop (over delimiting boundaries). #

	    my ($br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs) = @{$set_delimiting_bds->[$db]};

	    my $indels = $br2indels{$br};
	    (defined $indels) or ($indels = $br2indels{$br} = []);

	    my ($state_ue, $state_le) = ($up_or_lw eq 'U') ? ($char_abs, $char_pres) : ($char_pres, $char_abs) ;

		# $up_or_lw = ($state_ue eq $char_abs) ? 'U' : 'L' ;

	    push @{$indels}, [$state_ue, $state_le, $invlvd_gpbs];

	}				# END of the main for-loop (over delimiting boundaries). #

	return \%br2indels;

}	# End of the subroutine 'mk_br2indels'. #


	### Hereafter, the subroutines necessary for STEP 4 are defined. ###


=This old attempt became OBSOLETE as of May 10th (Wed), 2017, because a simpler version started to be created BELOW.

# enum_psm_indel_scns_in_eff_indp_indel_blk: [... RENAMED as "enum_psm_indel_hsts_in_eff_indp_indel_blk" on May 11 (Thu), 2017.] 
#
#	This subroutine attempts to enumerate parsimoniousindel histories within a subject effectively independent indel block.
#	(It comprizes STEP 4 in the 'workflow_for_indel_histories&multiplication_factors.rtf'.)
#
#	It was created by MODIFYING the subroutine, 'enum_psm_indel_in_cls_blks', in the package "MyTreeMap_indels_spt_odr.pm".
#
#	[CREATION STARTED on Feb 25 (Sat), 2017, ENDED on mmm dd (www), 2017.]
#
# [ enum_psm_indel_scns_in_cls_blks:
#	This subroutine enumerates the most parsimonious scenarios of insertions/deletions
#	in a cluster of contiguous blocks, each consisting of sites with an identical presence/absence pattern. ] ... Predecessor ...
#	{For more details, see "MyTreeMap_indels_spt_odr.pm".}
#
#
# INPUT: ($top_node_id, \%node2pa, \%node2ch, \%node2ct_exoffs, $char_pres, $char_abs, \@blocks_in_cls, 
#		$totcts_indels_for_cls0, \%br2indels0, \%node2blwise_gstat0, \@eff_indep_indel_blocks, $sbj_eiidb), where
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the top node;
#	%node2ch = ($node_id => \@child_node_ids, ...) for all interior nodes;
#	%node2ct_exoffs = ($node_id => #{exterior offsprings}, ...) for all nodes;
#	$char_pres/abs is the character representing the 'presence'/'absence' state;
#
#	@blocks_in_cls represents a gapped segment, i.e., a cluster of contiguous blocks (gap-pattern blocks),
#	each consisting of sites with a particular gap-pattern, where
#	[$indx_gp = $blocks_in_cls[$b]->[$indx_indx_gp] (or $blocks_in_cls[$b] itself if $indx_indx_gp is undefined) is the index specifying the gap pattern in @set_node2parsstat and @set_br2change] ... OBSOLETE ;
#
#	($totcts_indels_for_cls0, \%br2indels0, \%node2blwise_gstat0) define the initial candidate of the parsimonious indel history in the subject gapped segment (created by concatenating the Dollo parsimonious indel histories for the gap-pattern blocks). More precisely,
#
#	$totcts_indels_for_cls0 is the total count of indel events in the initial candiate indel history,
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of gap-pattern blocks involved), and
#	%node2blwise_gstat = ($node_id => \@blwise_gapstats, ...) is the map of the gap states onto the nodes,
#	 with $blwise_gapstats[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th gap-pattern block at the node;
#
#	@{$eff_indep_indel_blocks[$eiidb]} = ($br, $up_or_lw, $start, $end, \@invlvd_gpbs, \@invlvd_gbs, \@invlvd_sbs) [<= Is this format really general enough to work on any types of indel blocks?]
#	gives information on the $eiidb th effectively independent indel block.
#	$br is the ID of the branch delimiting the indel block,
#	$up_or_lw = 'U'/'L' if the upper/lower-end of $br is the indel block, 
#	$start/$end is the start/end coordinate of the boundary,
#	@invlvd_gpbs lists the gap-pattern-blocks involved,
#	@invlvd_gbs lists the gap-blocks involved,
#	@invlvd_sbs is the sequence-blocks involved.
#
#	$sbj_eiidb is the index of the subject effectively independent indel bock (in @eff_indep_indel_blocks).
#
# OUTPUT: ($min_totct_indels_in_indel_blk, \@set_psm_hsts_in_indel_blk), where
#
#	$min_totct_indels_in_indel_blk is the total number of insertion/deletion events in the most parsimonious history(ies) confined in the subject effectively independent indel block;
#
#	@{$set_psm_hsts_in_indel_blk[$ph]} = (\%br2indels, \%node2blwise_gapstat) is for the $ph th parsimonious history within the subject indel block, where
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_gpb) describing the $id th indel event at the branch,
#	with @indices_gpb listing the indices (in @blocks_in_cls) of the gap-pattern blocks involved in the indel event,
#	%node2blwise_gapstat = ($node_id => \@blwise_gapstats, ...) is the map of gap states onto the interior nodes, 
#	If $fl_full = 0, %node2gapstat is given only at the top node.
#
sub enum_psm_indel_scns_in_eff_indp_indel_blk ($\%\%\%$$\@$\%\%\@$) {

	my ($top_node_id, $node2pa, $node2ch, $node2ct_exoffs, $char_pres, $char_abs, $blocks_in_cls, 
		$totcts_indels_for_cls0, $br2indels0, $node2blwise_gstat0, $eff_indep_indel_blocks, $sbj_eiidb) = @_;


	## MODIFY the subroutine, "enum_psm_indel_in_cls_blks", in "MyTreeMap_indels_spt_odr.pm". ##

	my $min_totct_indels_in_indel_blk = 0;
	my @set_psm_hsts_in_indel_blk = ();


	return ($min_totct_indels_in_indel_blk, \@set_psm_hsts_in_indel_blk);

}	# END of the subroutine 'enum_psm_indel_scns_in_eff_indp_indel_blk'. # 

=cut # END of "This old attempt became OBSOLETE as of May 10th (Wed), 2017, because a simpler version started to be created BELOW."


# enum_psm_indel_hsts_in_eff_indp_indel_blk:
#	This subroutine attempts to enumerate parsimoniousindel histories within a subject effectively independent indel block.
#	(It comprizes STEP 4 in the 'workflow_for_indel_histories&multiplication_factors.rtf'.)
#
#	It was created by MODIFYING the subroutine, 'enum_psm_indel_scns_in_cls_blks', in the package "MyTreeMap_indels_spt_odr.pm".
#	NOTE: As opposed to its predecessor, 'enum_psm_indel_in_cls_blks' in "MyTreeMap_indels_spt_odr.pm", this subroutine assumes that the subject gapped segment contains NO blank sites (columns), each of which contains no residue.
#
#	[CREATION STARTED on May 10 (Wed), 2017, ENDED on mmm dd (www), 2017.]
#
# [ enum_psm_indel_scns_in_cls_blks:
#	This subroutine enumerates the most parsimonious scenarios of insertions/deletions
#	in a cluster of contiguous blocks, each consisting of sites with an identical presence/absence pattern. ] ... Predecessor ...
#	{For more details, see "MyTreeMap_indels_spt_odr.pm".}
#
#
# INPUT: ($topnode_id, \%node2pa, \%node2ch, \%node2ct_exoffs, $char_pres, $char_abs, \@blocks_in_cls, 
#		$totct_indels_for_eiidb0, \%br2indels_for_eiidb0, \%node2blwise_gstat_for_eiidb0), where
#
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the top node;
#	%node2ch = ($node_id => \@child_node_ids, ...) for all interior nodes;
#	%node2ct_exoffs = ($node_id => #{exterior offsprings}, ...) for all nodes;
#	$char_pres/abs is the character representing the 'presence'/'absence' state;
#
#	@blocks_in_cls represents a gapped segment, i.e., a cluster of contiguous blocks (gap-pattern blocks),
#	each consisting of sites with a particular gap-pattern;
#
#	($totct_indels_for_eiidb0, \%br2indels_for_eiidb0, \%node2blwise_gstat_for_eiidb0) define the initial candidate of the parsimonious indel history in the subject effectively independent indel block (eiidb) in the subject gapped segment (created by concatenating the Dollo parsimonious indel histories for the gap-pattern blocks, AND by selecting only those indel events which belong to the subject eiidb). More precisely,
#
#	$totct_indels_for_eiidb0 is the total count of indel events in the initial candiate indel history,
#	%br2indels_for_eiidb0 = ($branch_id => \@list_indels0, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels0[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of gap-pattern blocks involved), and
#	%node2blwise_gstat_for_eiidb0 = ($node_id => \@blwise_gapstats0, ...) is the map of the gap states onto the nodes,
#	 with $blwise_gapstats0[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th gap-pattern block at the node.
#
#	(These 3 ingredients can be prepared via 'counts_indels' (in "MyTreeMap_indels_sppl.pm"), 'mk_br2indels', and 'mk_node2blwise_gstat', respectively.)
#
#
# OUTPUT: ($min_totct_indels_in_indel_blk, \@set_psm_hsts_in_indel_blk), where
#
#	$min_totct_indels_in_indel_blk is the total number of insertion/deletion events in the most parsimonious history(ies) confined in the subject effectively independent indel block;
#
#	@{$set_psm_hsts_in_indel_blk[$ph]} = (\%br2indels, \%node2blwise_gapstat) is for the $ph th parsimonious history within the subject indel block, where
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_gpb) describing the $id th indel event at the branch,
#	with @indices_gpb listing the indices (in @blocks_in_cls) of the gap-pattern blocks involved in the indel event,
#	%node2blwise_gapstat = ($node_id => \@blwise_gapstats, ...) is the map of gap states onto the interior nodes, 
#	If $fl_full = 0, %node2gapstat is given only at the top node.
#
sub enum_psm_indel_hsts_in_eff_indp_indel_blk ($\%\%\%$$\@$\%\%) {

	my ($topnode_id, $node2pa, $node2ch, $node2ct_exoffs, $char_pres, $char_abs, $blocks_in_cls, 
		$totct_indels_for_eiidb0, $br2indels_for_eiidb0, $node2blwise_gstat_for_eiidb0) = @_;

		# Introduce shorthand names. #
	my $totct_indels0 = $totct_indels_for_eiidb0;
	my $br2indels0 = $br2indels_for_eiidb0;
	my $node2blwise_gstat0 = $node2blwise_gstat_for_eiidb0;

	my $ct_seqs = $node2ct_exoffs->{$topnode_id};

	## MODIFY the subroutine, "enum_psm_indel_in_cls_blks", in "MyTreeMap_indels_spt_odr.pm". ##

    (print STDOUT " << Subroutine 'enum_psm_indel_hsts_in_eff_indp_indel_blk' (#{sequences} = $ct_seqs) >>\n\n") if ($TEST_LEVEL1>0); 
    if ($TEST_LEVEL1>3) {
	print STDOUT "[ Input data ]\n\n";
	print STDOUT '$topnode_id = ', $topnode_id, ".\n";
	print STDOUT '$char_pres/$char_abs = ', "$char_pres/$char_abs.\n"; 
	print STDOUT "\n\n";

	print STDOUT '$totct_indels_for_eiidb0 = ', $totct_indels_for_eiidb0, ".\n\n";

	print STDOUT 'Content of %{$node2blwise_gstat_for_eiidb0}:', "\n\n";
	print STDOUT join ("\t", 'Node_ID', "'initial_block-wise_gap_state(concatenated)'"), "\n\n";
	foreach my $node (sort {$a <=> $b} keys %{$node2blwise_gstat_for_eiidb0}) {
	    my $cnct_gstat0 = join ('', @{$node2blwise_gstat_for_eiidb0->{$node}});
	    print STDOUT join ("\t", $node, "'$cnct_gstat0'"), "\n";
	}
	print STDOUT "\n";

	print STDOUT 'Content of %{$br2indels_for_eiidb0}:', "\n\n";
	print STDOUT join ("\t", 'Branch_ID', '#{exteior_offsprings}', 'State_ue/State_le/indices,in,@{$blocks_in_cls}', '...'), "\n\n";
	foreach my $br (sort {$a <=> $b} keys %{$br2indels_for_eiidb0}) {
	    my $ct_exoffs = $node2ct_exoffs->{$br};
	    my @out_list_indels0 = ();
	    foreach my $indel (@{$br2indels_for_eiidb0->{$br}}) {
		my ($stat_ue, $stat_le, $indices_blk0) = @{$indel};
		my $cnct_indices_blk0 = join (',', @{$indices_blk0});
		push @out_list_indels0, join ('/', $stat_ue, $stat_le, $cnct_indices_blk0);
	    }
	    print STDOUT join ("\t", $br, $ct_exoffs, @out_list_indels0), "\n";
	}
	print STDOUT "\n\n";
    }

		# Create a set of all nodes.
		# (ADDED in creation of 'MyTreeMap_indels_spt_odr.pm'.) #
	my @set_all_nodes = sort keys %{$node2pa};
	unshift @set_all_nodes, $topnode_id;


		# Handle an exceptional case,
		# where the gap-block cluster consists only of a null-block (i.e., a blank block). #
		# (ADDED in revision on Mar 8 (Sat), 2014.) 
		# ==> OMITTED on May 11 (Thu), 2017, when creating "MyTreeMap_indels_spt_odr_finer.pm", because this probably will NOT be necessary here. #


		# Create a list of indel events to be examined #

	# @{$list_to_be_examined->[$k][$i]}
	# = ($branch_id, index in the list @{$br2indels{$branch_id}})
	# is for the $i th indel event that result in $k exterior nodes with $char_abs (for 0<$k<$ct_seqs).
	# @{$list_indels_to_be_examined[0]} should always be empty.
	#
		# MODIFIED on Nov 13 (Wed), 2013, to enable the subroutine to correctly handle two-branch nodes. #
	my $list_indels_to_be_examined0 = create_list_to_be_examined (%{$node2ch}, %{$node2ct_exoffs}, $ct_seqs, $char_abs, %{$br2indels0});
	##my $list_indels_to_be_examined0 = create_list_to_be_examined (%{$node2ct_exoffs}, $ct_seqs, $char_abs, %{$br2indels0});
	unless (defined $list_indels_to_be_examined0) {
	    print STDOUT "[ERROR 'enum_psm_indel_scns_in_cls_blks'] The subroutine 'create_list_to_be_examined' FAILED: $! !\n";
	    return (undef, undef);
	}
    (print STDOUT "\n --- Subroutine 'create_list_to_be_examined' Ended ...\n\n") if ($TEST_LEVEL1>1);

    if ($TEST_LEVEL1>2) {
	my $size_list_to_be_exmd = @{$list_indels_to_be_examined0};
	print STDOUT "[ Output of 'create_list_to_be_examined' (Size_list = $size_list_to_be_exmd) ]\n\n";
	print STDOUT 'Content of @{$list_indels_to_be_examined0}:', "\n\n";
	print STDOUT join ("\t", '#{Exterior $char_abs\'s resulting from the event}', 'indx/$branch_id/index_in_@{$br2indels_for_eiidb0{$branch_id}}', '...'), "\n\n";
	for (my $k=0; $k<$size_list_to_be_exmd; $k++) {
	    my $indels = $list_indels_to_be_examined0->[$k];
	    my $ct_indels=@{$indels};
	    my @out_list_indels = ();
	    for (my $i=0; $i<$ct_indels; $i++) {
		my ($br_id, $indx_indel) = @{$indels->[$i]};
		push @out_list_indels, join ('/', $i, $br_id, $indx_indel);
	    }
	    print STDOUT join ("\t", $k, @out_list_indels), "\n";
	}
	print STDOUT "\n\n";
    }

		# Initialize the variables/arrays to be used for the main process. #

	my $min_totct_indels0 = undef;
	my @set_psm_hsts_in_eiidb0 = ();
	my @stack_scns_to_be_examined = ([$totct_indels0, $node2blwise_gstat0, $br2indels0, $list_indels_to_be_examined0]);

		# Now the main process begins. #
    (print STDOUT "\n{{ Beginning of the Main Process }}\n\n") if ($TEST_LEVEL1>1);
     my $iter_no = 0;

	while (my $scn_to_be_examined = shift @stack_scns_to_be_examined) { # Main outermost while-loop (over scenarios to be examined) #

	    my ($totct_indels, $node2blwise_gstat, $br2indels, $indels_to_be_examined) = @{$scn_to_be_examined};
	    my $lclmin_totct_indels = $totct_indels;

	if ($TEST_LEVEL1>1) {
	    $iter_no++;
	    my $ct_stacked = @stack_scns_to_be_examined;
	    print STDOUT "\n{{ The $iter_no th Iteration:  #{Stacked scenarios} = $ct_stacked\; Initial Min(Total#{indels}) = $lclmin_totct_indels }}\n\n";
	}

	    ##while (@{$indels_to_be_examined}>0) { # Main outer while-loop  #
	    ##while (@{$indels_to_be_examined}>2) { # Main outer while-loop (over the set of indel events resulting in the largest number of exterior nodes with $char_abs) #
	    while (@{$indels_to_be_examined}>1) { # Main outer while-loop (over the set of indel events that could possibly be receded). ## MODIFIED on Nov 13 (Wed), 2013. ##
		# @{$indels_to_be_examined->[0]} should always be empty,
		# [@{$indels_to_be_examined->[1]} contain only deletions along exteior branches (, which therefore cannot recede any further). ] ---  This logic does not hold any longer (as of Nov 13 (Wed), 2013).

		my $indx_to_be_examined = $#{$indels_to_be_examined};
		my $list_indels = $indels_to_be_examined->[$indx_to_be_examined];

		while (@{$list_indels}>0) { # Main middle while-loop (over the indels to be examined) # 

		    my $indx_exmd_indel = $#{$list_indels};
	if ($TEST_LEVEL1>1) {
	    (print STDOUT "\n{ To be examined: The $indx_exmd_indel th indel event of the $indx_to_be_examined th set of events. }\n\n") if ($TEST_LEVEL1>1);
	}

			## Newly ADDED on Sep 22, 2012, to make the subroutine more robust against unexpected upstream behaviors. ##
		    my $examined_event = $list_indels->[$indx_exmd_indel];
		    unless ((defined $examined_event) and (@{$examined_event}>0)) {
			print STDOUT "\n[WARNING 'enum_psm_indel_scns_in_cls_blks'] The $indx_exmd_indel th indel event of the $indx_to_be_examined th set of events, which should have been examined now, is UNDEFINED or EMPTY! So I will remove it...\n\n";
			pop @{$list_indels};
			next;
		    }

			# (MODIFIED in creation of 'MyTreeMap_indels_spt_odr.pm'.) #
		    my ($diff_totct_indels, $new_node2blwise_gstat, $new_br2indels, $new_indels_to_be_examined) = recede_absence_spt_odr (%{$node2blwise_gstat}, %{$br2indels}, @{$indels_to_be_examined}, $char_abs, $char_pres, %{$node2pa}, %{$node2ch}, @set_all_nodes, %{$node2ct_exoffs}, $ct_seqs);
		    ##my ($diff_totct_indels, $new_node2blwise_gstat, $new_br2indels, $new_indels_to_be_examined) = recede_absence (%{$node2blwise_gstat}, %{$br2indels}, @{$indels_to_be_examined}, $char_abs, $char_pres, %{$node2pa}, %{$node2ch}, %{$node2ct_exoffs}, $ct_seqs);
		    unless (defined $diff_totct_indels) {
			print STDOUT "[ERROR 'enum_psm_indel_scns_in_cls_blks'] The subroutine 'recede_absence_spt_odr' FAILED : $! !\n";
			return (undef, undef);
		    }
	(print STDOUT "\n --- Subroutine 'recede_absence_spt_odr' Ended ...\n\n") if ($TEST_LEVEL1>1);

		    if ($diff_totct_indels>0) {
			# Just dismiss the new scenario (, which may often be undefined). #
	(print STDOUT '$diff_totct_indels = ', $diff_totct_indels, " => Dismiss the new scenario.\n\n") if ($TEST_LEVEL1>1);
			next;
		    } elsif ($diff_totct_indels<0) { 
			# Replace the old scenario with the new scenario. #
	(print STDOUT '$diff_totct_indels = ', $diff_totct_indels, " => Replace the old scenario(s) with the new scenario.\n\n") if ($TEST_LEVEL1>1);

			$lclmin_totct_indels += $diff_totct_indels;
			$node2blwise_gstat = $new_node2blwise_gstat;
			$br2indels = $new_br2indels;
			$indels_to_be_examined = $new_indels_to_be_examined;

				## A BUG fixed (on Sep 23 (Sun), 2012). ##
			$indx_to_be_examined = $#{$indels_to_be_examined};
			$list_indels = $indels_to_be_examined->[$indx_to_be_examined];

		    } elsif (defined $new_node2blwise_gstat) {
			# Add the new scenario to the stack of scenarios to be examined. #
			##push @{$new_indels_to_be_examined}, $list_indels;
	(print STDOUT '$diff_totct_indels = ', $diff_totct_indels, " => Append the new scenario to the stack of scenarios to be examined.\n\n") if ($TEST_LEVEL1>1);
			push @stack_scns_to_be_examined, [$lclmin_totct_indels, $new_node2blwise_gstat, $new_br2indels, $new_indels_to_be_examined];
		    } else {
			# Do nothing, because the 'absence' state in question cannot recede any further. #
	(print STDOUT '$diff_totct_indels = ', $diff_totct_indels, " (AND", '%{$new_node2blwise_gstat}', " is UNDEFINED) => STOP receding the 'absence' in the event just examined.\n\n") if ($TEST_LEVEL1>1);
			next;
		    }

		}	# End of the main middle while-loop (over the indels to be examined) #

	    (print STDOUT "\n +++ End of the $indx_to_be_examined th set of events. +++\n\n") if ($TEST_LEVEL1>1);

	    pop @{$indels_to_be_examined}; # Remove the last entry, which has been emptied. #

	    } # End of the main outer while-loop (over the set of indel events resulting in the largest number of exterior nodes with $char_abs) #

	(print STDOUT "\n ~~~ The Set of parsimonious scenarios will be renewed ~~~\n\n") if ($TEST_LEVEL1>1);

	    my @cand_psm_scn = ($br2indels, $node2blwise_gstat); # A candidate parsimonious scenario. #

	    unless (defined $min_totct_indels0) {
		# Initialize the set of global parsimonious scenarios. #
	(print STDOUT 'INITIALIZE the set of parsimonious scenarios (with $min_totct_indels0 = ', $lclmin_totct_indels, ").\n\n") if ($TEST_LEVEL1>1);

		@set_psm_hsts_in_eiidb0 = (\@cand_psm_scn);
		$min_totct_indels0 = $lclmin_totct_indels;
		next;
	    }

		# Compare the local minimum with the current minimum. #

	    if ($lclmin_totct_indels < $min_totct_indels0) {
		# Replace the set of parsimonious scenarios. #
	(print STDOUT 'REPLACE the OLD parsimonious scenario(s) with the NEW one (with $min_totct_indels0 = ', $lclmin_totct_indels, ").\n\n") if ($TEST_LEVEL1>1);
		
		@set_psm_hsts_in_eiidb0 = (\@cand_psm_scn);
		$min_totct_indels0 = $lclmin_totct_indels;

	    } elsif ($lclmin_totct_indels > $min_totct_indels0) {
		# Just discard the candidate parsimonious scenario. #

	(print STDOUT 'DISMISS the NEW parsimonious scenario (with $lclmin_totct_indels = ', $lclmin_totct_indels, ").\n\n") if ($TEST_LEVEL1>1);

		next;

	    } else {
		# Append the candidate parsimonious scenario. (Keep the total#indels as it is.) #
	(print STDOUT 'APPEND the NEW parsimonious scenario to the OLD set of scenario(s) (with $min_totct_indels0 = ', $lclmin_totct_indels, ").\n\n") if ($TEST_LEVEL1>1);

		push @set_psm_hsts_in_eiidb0, \@cand_psm_scn;
	    }

	}	# End of the main outermost while-loop (over scenarios to be examined) #


    (print STDOUT "\n = The Main Process Ended =\n\n") if ($TEST_LEVEL1>0);

    if ($TEST_LEVEL1>1) {
	print STDOUT "Min(Tot#{indels}) = $min_totct_indels0.\n\n";
    }
    if ($TEST_LEVEL1>2) {
	my $ct_psm_hsts_in_eiidb0 = @set_psm_hsts_in_eiidb0;
	print STDOUT "[ Interim output (Set of $ct_psm_hsts_in_eiidb0 parsimonious scenarios) ]\n\n";
	##print STDOUT "[ Interim output (Set of $ct_psm_scns_in_cls0 parsimonious scenarios before incorporating 'null' blocks) ]\n\n";
	print STDOUT 'Content of @set_psm_hsts_in_eiidb0:', "\n\n";

	for (my $ps=0; $ps<$ct_psm_hsts_in_eiidb0; $ps++) { # Outermost testing for-loop #
	    print STDOUT "{ The $ps th scenario }\n\n";

	    my ($br2indels, $node2blwise_gstat) = @{$set_psm_hsts_in_eiidb0[$ps]};

	    print STDOUT 'Content of %{$node2blwise_gstat}:', "\n\n";
	    print STDOUT join ("\t", 'Node_ID', "'initial_block-wise_gap_state(concatenated)'"), "\n\n";
	    foreach my $node (sort {$a <=> $b} keys %{$node2blwise_gstat}) {
		my $cnct_gstat = join ('', @{$node2blwise_gstat->{$node}});
		print STDOUT join ("\t", $node, "'$cnct_gstat'"), "\n";
	    }
	    print STDOUT "\n";

	    print STDOUT 'Content of %{$br2indels}:', "\n\n";
	    print STDOUT join ("\t", 'Branch_ID', '#{exteior_offsprings}', 'State_ue/State_le/indices,in,@{$blocks_in_cls}', '...'), "\n\n";
	    foreach my $br (sort {$a <=> $b} keys %{$br2indels}) {
		my $ct_exoffs = $node2ct_exoffs->{$br};
		my @out_list_indels = ();
		foreach my $indel (@{$br2indels->{$br}}) {
		    (@{$indel}>0) or next; # Skip an empty entry. (ADDED to fix a BUG on May 13 (Sun), 2012) #
		    my ($stat_ue, $stat_le, $indices_blk) = @{$indel};
		    my $cnct_indices_blk = join (',', @{$indices_blk});
		    push @out_list_indels, join ('/', $stat_ue, $stat_le, $cnct_indices_blk);
		}
		print STDOUT join ("\t", $br, $ct_exoffs, @out_list_indels), "\n";
	    }
	    print STDOUT "\n\n";

	}			# End of the outermost testing for-loop #

    } # END of "if ($TEST_LEVEL1>2) {...}"



		# Incorporate the "null" blocks, each of which consists only of $char_abs. #
		# => This function was DISABLED in effect on May 11 (Thu), 2017, when creating "MyTreeMap_indels_spt_odr_finer.pm".

	$FL_IGNORE_NULL_COLUMNS = 1; # Ignore null columns even if there exist ones. 
	my $ct_blank_blks = 0; # Basically, there must be NO "null" blocks in the input MSA.

	if ($FL_IGNORE_NULL_COLUMNS) { ## (This "if (...) {...} else {...}" block was ADDED in revision on Nov 15 (Fri), 2013). ##

    (print STDOUT " = Although the MSA has $ct_blank_blks 'null' blocks, we will just leave them untouched, because the flag, ", '$FL_IGNORE_NULL_COLUMNS ', "(=$FL_IGNORE_NULL_COLUMNS), tells me to do so. =\n\n") if ($TEST_LEVEL1>0);
    (print STDOUT "   --- But I will anyway clean up the ", scalar (@set_psm_hsts_in_eiidb0), " raw parsimonious scenarios obtained so far ...\n\n\n") if ($TEST_LEVEL1>1);

	} else {

    (print STDOUT " = NOW, let us incorporate the $ct_blank_blks 'null' blocks =\n\n") if ($TEST_LEVEL1>0);
    if ($TEST_LEVEL1>1) {
	my $list_blank_blks = []; # ADDED in creation of 'enum_psm_indel_hsts_in_eff_indp_indel_blk'.
	print STDOUT '"Null" blocks = (', join (', ', @{$list_blank_blks}), ").\n\n";
	print STDOUT '#{Parsimonious scenarios to accomodate the null blocks} = ', scalar (@set_psm_hsts_in_eiidb0), ".\n";
	print STDOUT "\n\n";
    }

	} ## END of "if ($FL_IGNORE_NULL_COLUMNS) {...} else {...}".


    my $indx_psm_scn0 = 0;

	my @set_psm_hsts_in_eiidb = ();
	##my @set_psm_scns_in_cls = ();
	my $min_totct_indels = undef;

	while (my $psm_scn0 = shift @set_psm_hsts_in_eiidb0) { # Second main outer while-loop (over 'initial' parsimonious scenarios) #

	(print STDOUT "\n[[ The $indx_psm_scn0 th 'initial' parsimonious scenario (#{indels} = $min_totct_indels0) ]]\n\n") if ($TEST_LEVEL1>1);

	    unless ( cleanup_br2indels (%{$psm_scn0->[0]}) ) {
		print STDOUT "[ERROR 'enum_psm_indel_scns_in_cls_blks'] The subroutine 'cleanup_br2indels' FAILED on the $indx_psm_scn0 th scenario: $! !\n";
		return (undef, undef);
	    }
	(print STDOUT "\n --- Subroutine 'cleanup_br2indels' ended ...\n\n")  if ($TEST_LEVEL1>1);


	    if ($FL_IGNORE_NULL_COLUMNS) { ## Skip the process to incorporate null blocks (ADDED in revision on Nov 15 (Fri), 2013). ##
	$indx_psm_scn0++;
		push @set_psm_hsts_in_eiidb, $psm_scn0;
		(defined $min_totct_indels) or ($min_totct_indels = $min_totct_indels0);
		next;
	    }

	    if ($ct_blank_blks==0) {
	(print STDOUT " ... APPEND the 'initial' scenario to the final set of parsimonious scenarios, bcause there are NO 'null' blocks to be incorporated ...\n\n") if ($TEST_LEVEL1>1);
	$indx_psm_scn0++;
		push @set_psm_hsts_in_eiidb, $psm_scn0;
		(defined $min_totct_indels) or ($min_totct_indels = $min_totct_indels0);
		next;
	    }

		# (The Main Part to Incorporate Indels to Create Null a Block)
		# ==> OMITTED when creating "MyTreeMap_indels_spt_odr_finer.pm", because the input MSAs are supposed to contain NO "null" blocks. #

	}	# End of the second main outer while-loop (over 'initial' parsimonious scenarios) #

    ##(print STDOUT "\n = Finished icorporating the 'null' blocks =\n\n") if ($TEST_LEVEL1>0);
    (print STDOUT "\n = Finished cleaning up the indel histories =\n\n") if ($TEST_LEVEL1>0);

    (print STDOUT "\n << Subroutine 'enum_psm_indel_hsts_in_eff_indp_indel_blk' Ended />>\n\n") if ($TEST_LEVEL1>0);


	my $min_totct_indels_in_eiidb = $min_totct_indels;

	return ($min_totct_indels_in_eiidb, \@set_psm_hsts_in_eiidb);

}	# END of the subroutine 'enum_psm_indel_hsts_in_eff_indp_indel_blk'. # 



	# The InterMediate Subroutine, which works on each gapped segment (i.e., each cluster of gap-pattern blocks). #


# enum_psm_indel_hsts_in_cls_blks_finer:
#	This subroutine enumerates the most parsimonious histories of insertions/deletions
#	in each of effectively independent indel blocks that belong to a single gapped segment (i.e., cluster of gap-pattern blocks).
#	The subroutine therefore is a "counterpart" of 'enum_psm_indel_scns_in_cls_blks' in "MyTreeMap_indels_spt_odr.pm"
#	in terms of philosophy & function.
#	It was created by inheriting useful parts from 'enum_psm_indel_scns_in_cls_blks'.
#
#	[CREATION STARTED on May 11th (Thu), 2017, ENDED on May 25th (Wed), 2017.]
#
# [ enum_psm_indel_scns_in_cls_blks:
#	This subroutine enumerates the most parsimonious scenarios of insertions/deletions
#	in a cluster of contiguous blocks, each consisting of sites with an identical presence/absence pattern.
#
#	REVISED on May 13 (Sun), 2012:
#	Made the format for the input set of indel blocks a bit more general,
#	so make the subroutine more flexible.
#	(This revision accompanied the revision of 'create_initial_set_indels'.)
#
#	REVISED on Nov 13 (Wed), 2013:
#	Now, the subroutine can correctly handle trees with two-branch nodes.
#	(This revision accompanied the revisions of 'create_list_to_be_examined' and 'recede_absence'.)
#
#	REVISED on Mar 8 (Sat), 2013:
#	Now, a new satellite subroutine, 'enum_psm_scns_gen_null_blk', was introduced, which can correctly enumerates
#	the parsimonious indel scenarios that generates a cluster consisting only of a single null-block.
#  ] ... Information on the Predecessor.
#
# INPUT: ($top_node_id, \@set_all_nodes, \%node2pa, \%node2ch, \%node2ct_exoffs, $char_pres, $char_abs, \@gpblocks_in_cls, 
#	\%br2indels0, \%node2blwise_gstats0, \%br2eq, \%br2nn), where
#
#	$topnode_id is the ID of the top-node (i.e., root) of the tree;
#	@set_all_nodes stores the IDs of all nodes;
#	%node2pa = ($node_id => $parent_id, ...) for all nodes except the top node;
#	%node2ch = ($node_id => \@child_node_ids, ...) for all interior nodes;
#
#	%node2ct_exoffs = ($node_id => #{exterior offsprings}, ...) for all nodes;
#
#	$char_pres/abs is the character representing the 'presence'/'absence' state;
#
#	@gpblocks_in_cls represents a gapped segment (i.e., a cluster of contiguous blocks),
#	each block of which consists of sites with a particular gap-pattern, where
#	@{$gpblocks_in_cls[$gpb]} = ($indx_gp, $len_gpb, $start, $end) is information on the $gpb th gap-pattern block in the cluster, where
#	$indx_gp specifying the gap-pattern (stored in @{$set_gpatterns}),
#	$len_gpb being the length (in terms of #{columns}) of the gap-pattern block,
#	and $start/$end being the starting/ending coordinate of the gap-pattern block (in the 0-based full-closed convention);
#
#	%br2indels0 = ($branch_id => \@list_indels0, ...) is the map of insertions/deletions onto branches for the initial indel history, with
#	@{$list_indels0[$e]} = ($state_ue, $state_le, \@indices_blk) for the $e th event (@indices_blk storing the indices of blocks involved);
#
#	%node2blwise_gstats0 = ($node_id => \@blwise_gapstats0, ...) is the map of the gap states onto the nodes for the initial indel history,
#	with $blwise_gapstats0[$bl] is the gap-state ($CHAR_PRES/ABS) of the $bl th block at the node;
#
#	%br2eq = ($br_id => $equivalent_br_id, ...) for branches with equivalents;
#	%br2nn = ($br_id => {$id_nearest_neighbor => 1, ...}, ...).
#
#
# OUTPUT: (\@min_totcts_indels_in_indel_blks, \@sets_psm_hsts_in_indel_blks, \@psm_hsts0_in_indel_blks,
#	   \@list_gblocks, \@list_seqblocks, \@set_eff_indep_indel_blocks, \@indices_incl_root), where
#
#	$min_totcts_indels_in_indel_blks[$eiidb] is the total number of insertion/deletion events in the most parsimonious history(ies) for the $eiidb th effectively independent indel block;
#	@{$sets_psm_hsts_in_cls[$eiidb]->[$ps]} = (\%br2indels, \%node2blwise_gapstat) is for the $ps th parsimonious history for the $eiidb th effectively independent indel block;
#
#	@{$psm_hsts0_in_indl_blks[$eiidb]} = (\%br2indels0, \%node2blwise_gapstat0) is for the initial candidate history for the $eiidb th effectively independent indel block;
#
#	@list_gblocks and @list_seqblocks are the main outputs of 'partit_into_gblocks_seqblocks';
#	@set_eff_indep_indel_blocks is the main output of 'partit_into_eff_indep_indel_blocks';
#	@indices_incl_root stores the indices of the effectively independent indel blocks that include root (top node).
#	It is empty if no effectively independent indel block includes root.
#
sub enum_psm_indel_hsts_in_cls_blks_finer ($\@\%\%\%$$\@\%\%\%\%) {

	my ($topnode_id, $set_all_nodes, $node2pa, $node2ch, $node2ct_exoffs, $char_pres, $char_abs, $gpblocks_in_cls, 
	   $br2indels0, $node2blwise_gstat0, $br2eq, $br2nn) = @_;

	my $ct_seqs = $node2ct_exoffs->{$topnode_id};

	my $ct_gpblks_in_cls = @{$gpblocks_in_cls};


		# [ Step 3 (A) ] Perform 'partit_into_gblocks_seqblocks' on the input gapped segment with the initial candidate indel history. #

	my ($list_gblocks, $list_seqblocks) = partit_into_gblocks_seqblocks (@{$gpblocks_in_cls}, %{$br2indels0}, %{$node2blwise_gstat0}, $topnode_id, %{$node2pa}, %{$node2ch}, $char_abs, %{$br2eq}, %{$br2nn});
	unless (defined $list_seqblocks) {
	    print STDOUT "[ERROR 'enum_psm_indel_hsts_in_cls_blks_finer'] The subroutine 'partit_into_gblocks_seqblocks' FAILED on the input gapped segment with the initial candidate indel history: $! !\n\n";
	    return (undef, undef, undef, undef, undef, undef);
	}

		# [ Step 3 (B) ] Perform 'partit_into_eff_indep_indel_blocks' on the input gapped segment with the initial candidate indel history. #

	my ($set_eff_indep_indel_blocks, $indices_incl_root) = partit_into_eff_indep_indel_blocks (@{$gpblocks_in_cls}, %{$br2indels0}, %{$node2blwise_gstat0}, $topnode_id, %{$node2pa}, %{$node2ch}, $char_abs, %{$br2eq});
	unless (defined $indices_incl_root) {
	    print STDOUT "[ERROR 'enum_psm_indel_hsts_in_cls_blks_finer'] The subroutine 'partit_into_eff_indep_indel_blocks' FAILED on the input gapped segment with the initial candidate indel history: $! !\n\n";
	    return (undef, undef, undef, undef, undef, undef);
	}

	my $ct_eiidbs = @{$set_eff_indep_indel_blocks};

		# [ Step 3 (C) ] Perform 'find_invlvd_in_eff_indep_indel_blocks', to examine the involvement of the gap-blocksand/or sequence-blocks in the effectively independent indel blocks, 

	my $if_success3c = find_invlvd_in_eff_indep_indel_blocks (@{$set_eff_indep_indel_blocks}, @{$list_gblocks}, @{$list_seqblocks}, @{$gpblocks_in_cls}, $topnode_id, %{$node2pa}, %{$node2ch});
	unless ($if_success3c) {
	    print STDOUT "[ERROR 'enum_psm_indel_hsts_in_cls_blks_finer'] The subroutine 'find_invlvd_in_eff_indep_indel_blocks' FAILED on the input gapped segment with the initial candidate indel history: $! !\n\n";
	    return (undef, undef, undef, undef, undef, undef);
	}

		## Initialize the Main outputs. ##

	my @min_totcts_indels_in_indel_blks = my @sets_psm_hsts_in_indel_blks = ();
	my @psm_hsts0_in_indel_blks = (); # ADDED on May 27 (Sat), 2017.


		## Search for parsimonious indel histories. #

	for (my $eiidb = 0; $eiidb < $ct_eiidbs; $eiidb++) { # Main for-loop (over the effectively independent indel blocks). #

	    my $indel_block = $set_eff_indep_indel_blocks->[$eiidb];

		# [ Step 3 (D) ] Perform 'mk_node2blwise_gstat'. #

	    my $node2blwise_gstat_for_eiidb0 = mk_node2blwise_gstat (@{$indel_block}, $ct_gpblks_in_cls, @{$set_all_nodes}, %{$node2ch}, $char_abs, $char_pres);
	    unless (defined $node2blwise_gstat_for_eiidb0) {
		print STDOUT "[ERROR 'enum_psm_indel_hsts_in_cls_blks_finer'] The subroutine 'mk_node2blwise_gstat' FAILED on the $eiidb th effectively independent indel block: $! !\n\n";
		return (undef, undef, undef, undef, undef, undef);
	    } 

		# [ Step 3 (E) ] Perform 'mk_br2indels'. #

	    my $br2indels_for_eiidb0 = mk_br2indels (@{$indel_block}, $char_abs, $char_pres);
	    unless (defined $br2indels_for_eiidb0) {
		print STDOUT "[ERROR 'enum_psm_indel_hsts_in_cls_blks_finer'] The subroutine 'mk_br2indels' FAILED on the $eiidb th effectively independent indel block: $! !\n\n";
		return (undef, undef, undef, undef, undef, undef);
	    }

		# [ Auxiliary Step ] Count the number of indels in the initial candidate history. #

	    my ($ct_ins_for_eiidb0, $ct_dels_for_eiidb0) = counts_indels (%{$br2indels_for_eiidb0}, $char_abs);
	    unless (defined $ct_dels_for_eiidb0) {
		print STDOUT "[ERROR 'enum_psm_indel_hsts_in_cls_blks_finer'] The subroutine 'counts_indels' FAILED on the $eiidb th effectively independent indel block: $! !\n\n";
		return (undef, undef, undef, undef, undef, undef);
	    }

	    my $totct_indels_for_eiidb0 = $ct_ins_for_eiidb0 + $ct_dels_for_eiidb0;

			# ADDED on May 27 (Sat), 2017. #

	    my %node2blwise_gstat0;
	    foreach my $node (keys %{$node2blwise_gstat_for_eiidb0}) {
		my @cp_gstats = @{$node2blwise_gstat_for_eiidb0->{$node}};
		$node2blwise_gstat0{$node} = \@cp_gstats;
	    }
	    my %br2indels0;
	    foreach my $br (keys %{$br2indels_for_eiidb0}) {
		my @cp_indels0 = ();
		my $indels0 = $br2indels_for_eiidb0->{$br};
		foreach my $info_indel (@{$indels0}) {
		    (defined $info_indel) or next;
		    my @cp = @{$info_indel};
		    push @cp_indels0, \@cp;
		}
		$br2indels0{$br} = \@cp_indels0;
	    }

	$psm_hsts0_in_indel_blks[$eiidb] = [\%br2indels0, \%node2blwise_gstat0];
	    
			# END of "ADDED on May 27 (Sat), 2017." #


		# [ Step 4 ] Perform 'enum_psm_indel_hsts_in_eff_indp_indel_blk'. #

	    my ($min_totct_indels_in_eiidb, $set_psm_hsts_in_eiidb)
		= enum_psm_indel_hsts_in_eff_indp_indel_blk ($topnode_id, %{$node2pa}, %{$node2ch}, %{$node2ct_exoffs}, $char_pres, $char_abs, @{$gpblocks_in_cls}, $totct_indels_for_eiidb0, %{$br2indels_for_eiidb0}, %{$node2blwise_gstat_for_eiidb0});
	    unless (defined $set_psm_hsts_in_eiidb) {
		print STDOUT "[ERROR 'enum_psm_indel_hsts_in_cls_blks_finer'] The subroutine 'enum_psm_indel_hsts_in_eff_indp_indel_blk' FAILED on the $eiidb th effectively independent indel block: $! !\n\n";
		return (undef, undef, undef, undef, undef, undef);
	    }


		# Store the set of persimonious indel histories in this effectively independent indel block. #

	    $min_totcts_indels_in_indel_blks[$eiidb] = $min_totct_indels_in_eiidb;
	    $sets_psm_hsts_in_indel_blks[$eiidb] = $set_psm_hsts_in_eiidb;

	}					# END of the main for-loop (over the effectively independent indel blocks). # 


	return (\@min_totcts_indels_in_indel_blks, \@sets_psm_hsts_in_indel_blks, \@psm_hsts0_in_indel_blks,
	       $list_gblocks, $list_seqblocks, $set_eff_indep_indel_blocks, $indices_incl_root);

}	# End of the subroutine 'enum_psm_indel_hsts_in_cls_blks_finer'. #



	## The INTEGRATED Subroutine (performing STEPs 1-4 of "workflow_for_indel_histories&multiplication_factors.rtf"). ##

	# Three Main Parts of the INTEGRATED Subroutine. #

# enum_psm_indel_hsts_for_msa_finer_PRE:
#	This is a portion of the INTEGRATED subroutine, "enum_psm_indel_hsts_for_msa_finer".
#	It does some preliminary works, especially handling the sequence tree.
#
#	[CREATED on May 26 (Fri), 2017] 
#
# INPUT: ($topnode_id, \%node2ch, \%node2pa), where
#	%node2ch = ($node_id => \@child_node_ids, ...) is for the all interior nodes;
#	%node2pa = ($node_id => $parent_node_id, ...) is for all the nodes except the top node;
#
# OUTPUT: (\%node2depth, \%node2ct_exoffs, \%br2eq, \%br2nn, \@set_all_nodes), where
#
#	%node2depth = ($node_id => depth of the node, ...);
#	%node2ct_exoffs = ($node_id => #{external offsprings of the node}, ...);
#	%br2eq = ($branch_id => $equivalent_branch_id, ...) for the branches with equivalents;
#	%br2nn = ($branch_id => {$id_of_a_neighbor => 1, ...}, ...);
#	@set_all_nodes stores the IDs of all nodes.
# 
sub enum_psm_indel_hsts_for_msa_finer_PRE ($\%\%) {

	my ($topnode_id, $node2ch, $node2pa) = @_;

    (print STDOUT "<< Subroutine 'enum_psm_indel_hsts_for_msa_finer_PRE' (Top node ID: $topnode_id) >>\n\n") if ($TEST_LEVEL_PRE>1/4);

    (print STDOUT '/// [ Preliminary Step 1 ] Prepare %{$node2ct_exoffs} for the input tree ///', "\n\n") if ($TEST_LEVEL_PRE>1);

	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
	unless (defined $node2depth) {
	    print STDOUT "[ERROR 'enum_psm_indel_hsts_for_msa_for_finer_PRE'] Subroutine 'make_node2abs_depth' FAILED: $! !\n\n";
	    return (undef, undef, undef, undef, undef);
	}
    (print STDOUT "\n--- Subroutine 'make_node2abs_depth' Ended ...\n\n") if ($TEST_LEVEL_PRE>1);

	my $node2ct_exoffs = make_node2ct_exoffs (%{$node2ch}, %{$node2depth});
	unless (defined $node2ct_exoffs) {
	    print STDOUT "[ERROR 'enum_psm_indel_scns_for_msa_for_finer_PRE'] Subroutine 'make_node2ct_exoffs' FAILED: $! !\n\n";
	    return ($node2depth, undef, undef, undef, undef);
	}
    (print STDOUT "\n--- Subroutine 'make_node2ct_exoffs' Ended ...\n\n") if ($TEST_LEVEL_PRE>1);

	my @set_all_nodes = sort {$node2ct_exoffs->{$b} <=> $node2ct_exoffs->{$a}} keys %{$node2ct_exoffs} ; 

    if ($TEST_LEVEL_PRE>2) {
	print STDOUT '[ TEST P1: Output of the Preliminary Step 1 ]', "\n\n";
	print STDOUT join ("\t", 'Node_ID', 'depth', '#{ext_offsprings}'), "\n\n";
	foreach my $node (sort { $node2depth->{$a} <=> $node2depth->{$b} } keys %{$node2depth}) {  # Auxiliary foreach-loop #
	    my $depth = $node2depth->{$node};
	    my $ct_exoffs = $node2ct_exoffs->{$node};
	    print STDOUT join ("\t", $node, $depth, $ct_exoffs), "\n";
	}				# End of the auxiliary foreach-loop #
	print STDOUT "\n\n";

	print STDOUT '@set_all_nodes = (', join (', ', @set_all_nodes), ").\n\n\n";

    } # END of "if ($TEST_LEVEL_PRE>2) {...} for the test output of preliminary Step 1. #


    (print STDOUT '/// [ Preliminary Step 2 ] Prepare %{$br2eq} &  %{$br2nn} for the input tree ///', "\n\n") if ($TEST_LEVEL_PRE>1);

	my $br2eq = create_br2eqbr ($topnode_id, %{$node2ch}) ;
	unless (defined $br2eq) {
	    print STDOUT "[ERROR 'enum_psm_indel_scns_for_msa_for_finer_PRE'] Subroutine 'create_br2eqbr' FAILED: $! !\n\n";
	    return ($node2depth, $node2ct_exoffs, undef, undef, \@set_all_nodes);
	}
    (print STDOUT "\n--- Subroutine 'create_br2eqbr' Ended ...\n\n") if ($TEST_LEVEL_PRE>1);

	my $br2nn = create_br2nn (%{$node2pa}, %{$node2ch}, %{$node2pa}, %{$br2eq}) ;
	unless (defined $br2nn) {
	    print STDOUT "[ERROR 'enum_psm_indel_scns_for_msa_for_finer_PRE'] Subroutine 'create_br2nn' FAILED: $! !\n\n";
	    return ($node2depth, $node2ct_exoffs, $br2eq, undef, \@set_all_nodes);
	}
    (print STDOUT "\n--- Subroutine 'create_br2nn' Ended ...\n\n") if ($TEST_LEVEL_PRE>1);


    if ($TEST_LEVEL_PRE>2) {
	print STDOUT '[ TEST P2: Output of the Preliminary Step 2 ]', "\n\n";

	print STDOUT '%{$br2eq} = ( # branch_ID => ID_of_equivalent_branch', "\n";
	foreach my $br (sort {$a <=> $b} keys %{$br2eq}) { print STDOUT "\t", $br, ' => ', $br2eq->{$br}, ",\n"; }
	print STDOUT '            ).', "\n\n\n";

	print STDOUT '( Contents of %{$br2nn} )', "\n\n";
	print STDOUT join ("\t", 'Branch_ID', 'list,of,neighboring,branches'), "\n\n";
	foreach my $br (sort {$a <=> $b} keys %{$br2nn}) {
	    my @list_nns = sort {$a <=> $b} keys %{$br2nn->{$br}};
	    my $cnct_list_nns = join (',', @list_nns);
	    print STDOUT join ("\t", $br, $cnct_list_nns), "\n";
	}
	print STDOUT "\n\n";

    } # END of "if ($TEST_LEVEL_PRE>2) {...}" for the test output of preliminary Step 2. #


    (print STDOUT "\n<< Subroutine 'enum_psm_indel_hsts_for_msa_finer_PRE' ENDED />>\n\n") if ($TEST_LEVEL_PRE> 1/4);

	return ($node2depth, $node2ct_exoffs, $br2eq, $br2nn, \@set_all_nodes);

}	# End of the subroutine, 'enum_psm_indel_hsts_for_msa_finer_PRE' #


# enum_psm_indel_hsts_for_msa_finer_1ST_HF:
#	This is a portion of the INTEGRATED subroutine, "enum_psm_indel_hsts_for_msa_finer".
#	It does the first half of the main works (STEPS 1 & 2),
#	up to inferring the initial candidate indel history for each gapped segment.
#
#	(It actually is nothing other than a wrapper of the subroutine, 'cal_init_psm_cands_for_sgl_msa' (STEP2).)
#
#	[CREATED on May 26 (Fri), 2017] 
#
# INPUT: (\%node2ch, \%node2pa, \%name2id, \@seqnames, \@in_msa, $char_abs, $char_pres, \%node2depth, \%node2ct_exoffs), where
#
#	%node2ch = ($node_id => \@child_node_ids, ...) is for the all interior nodes;
#	%node2pa = ($node_id => $parent_node_id, ...) is for all the nodes except the top node;
#	%name2id = ($node_name => $node_id, ...) is for the exterior nodes and the top node (with $node_name = $TOP_NODE);
#	$seqnames[$k] is the name of the $k th sequence in the input MSA
#	(NOTE that these names must exactly match those used in %name2id);
#	${$in_msa[$k]} is the $k th sequence embedded in the input MSA;
#	$char_pres/$char_abs is the character (or symbol) representing the 'presence'/'absence' state;
#
#	%node2depth = ($node_id => {depth of the node}, ...);
#	%node2ct_exoffs = ($node_id => #{external offsprings}, ...).
#
# OUTPUT: (\@set_gpatterns, \@set_node2dparsstat, \@set_br2change, \@info_psm_indel_hsts0), where
#
#	\@set_gpatterns, \@set_node2dparsstat and \@set_br2change are the three components of @info_gpatterns,
#	which in turn is a piece of the output of "cal_init_psm_cands_for_sgl_msa" in "MyTreeMap_indels_spt_odr2.pm" (STEP 2);
#
#	@info_init_psm_hsts (= (\@set_cls_gpblocks, \@totcts0_indels_for_cls, \@psm_hsts0_for_cls))
#	is a piece of the output of "cal_init_psm_cands_for_sgl_msa" (STEP 2).
# 
sub enum_psm_indel_hsts_for_msa_finer_1ST_HF (\%\%\%\@\@$$\%\%) {

	my ($node2ch, $node2pa, $name2id, $seqnames, $in_msa, $char_abs, $char_pres, $node2depth, $node2ct_exoffs) = @_;


	my $topnode_id = $name2id->{$TOP_NODE};
	my $ct_seqs = @{$seqnames};
    (print STDOUT "<< Subroutine 'enum_psm_indel_hsts_for_msa_finer_1ST_HF' (Top node ID: $topnode_id; #{sequences} = $ct_seqs) >>\n\n") if ($TEST_LEVEL_1ST>1/4);

		# A sort of sanity check. #
	unless (@{$in_msa} == $ct_seqs) {
	    print STDOUT "[ERROR 'enum_psm_indel_hsts_for_msa_finer_1ST_HF'] The number of sequences in the input MSA (= ", scalar (@{$in_msa}), ") CLASHES with the number of sequence names (= $ct_seqs): $! !\n\n";
	    return (undef, undef, undef, undef);
	}


    (print STDOUT '/// [ STEP 1 ] Examine the gap pattern structure of the input MSA (& Chop the MSA into gapped and gapless segments) ///', "\n\n") if ($TEST_LEVEL_1ST>1);
    (print STDOUT '/// & [ STEP 2 ] For each gapped segment, determine the Dollo parsimonious indel history and ancestral gap states. ///', "\n\n") if ($TEST_LEVEL_1ST>1);

	my @info_gpatterns0 = ([], [], []);

	my ($info_psm_indel_hsts0, $info_gpatterns) = cal_init_psm_cands_for_sgl_msa (%{$node2ch}, %{$node2pa}, %{$node2depth}, %{$node2ct_exoffs}, %{$name2id}, @{$seqnames}, @{$in_msa}, $char_abs, $char_pres, @info_gpatterns0);
	unless (defined $info_gpatterns) {
	    print STDOUT "[ERROR 'enum_psm_indel_hsts_for_msa_finer_1ST_HF'] Subroutine 'cal_init_psm_cands_for_sgl_msa' FAILED (1): $! !\n\n";
	    return (undef, undef, undef, undef);
	}
	unless (@{$info_psm_indel_hsts0}>0) { # ADDED to fix a BUG (on June 22 (Thu), 2017). #
			# CORRECTED on 2020/03/09. #
	    ##print STDOUT "[ERROR 'enum_psm_indel_hsts_for_msa_finer_1ST_HF'] Subroutine 'cal_init_psm_cands_for_sgl_msa' FAILED (2): $! !\n\n";
	    ##return (undef, undef, undef, undef);
	    print STDOUT "[ERROR 'enum_psm_indel_hsts_for_msa_finer_1ST_HF'] Subroutine 'cal_init_psm_cands_for_sgl_msa' FAILED (2), probably because the input MSA contains (a) gap(s) (or (a) gapped segment(s)) that is (are) too long: $! !\n\n";
	    return (undef, undef, undef, []);
			# END of "CORRECTED on 2020/03/09." #
	}

	my ($set_cls_gpblocks, $totcts0_indels_for_cls, $psm_hsts0_for_cls) = @{$info_psm_indel_hsts0};
	my ($set_gpatterns, $set_node2dparsstat, $set_br2change) = @{$info_gpatterns};

	my $ct_cls = @{$set_cls_gpblocks};
	my $ct_gpatterns = @{$set_gpatterns};

			# Sorts of sanity checkes. #

	unless ($ct_cls == @{$totcts0_indels_for_cls}) {
	    print STDOUT "[ERROR 'enum_psm_indel_hsts_for_msa_finer_1ST_HF'] The number of initial total indel counts for gapped segments (= ", scalar (@{$totcts0_indels_for_cls}), ") CLASHES with that of gapped segments (= $ct_cls): $! !\n\n";
	    return (undef, undef, undef, undef);
	}
	unless ($ct_cls == @{$psm_hsts0_for_cls}) {
	    print STDOUT "[ERROR 'enum_psm_indel_hsts_for_msa_finer_1ST_HF'] The number of initial indel histories for gapped segments (= ", scalar (@{$psm_hsts0_for_cls}), ") CLASHES with that of gapped segments (= $ct_cls): $! !\n\n";
	    return (undef, undef, undef, undef);
	}
	unless ($ct_gpatterns == @{$set_node2dparsstat}) {
	    print STDOUT "[ERROR 'enum_psm_indel_hsts_for_msa_finer_1ST_HF'] The number of maps of Dollo parsimonious states onto nodes (= ", scalar (@{$set_node2dparsstat}), ") CLASHES with that of gap patterns (= $ct_gpatterns): $! !\n\n";
	    return (undef, undef, undef, undef);
	}
	unless ($ct_gpatterns == @{$set_br2change}) {
	    print STDOUT "[ERROR 'enum_psm_indel_hsts_for_msa_finer_1ST_HF'] The number of maps of Dollo parsimoinous indel events onto branches (= ", scalar (@{$set_br2change}), ") CLASHES with that of gap patterns (= $ct_gpatterns): $! !\n\n";
	    return (undef, undef, undef, undef);
	}


    if ($TEST_LEVEL_1ST>2) {
	print STDOUT "[[TEST1: Output of the STEP 1 ]]\n\n";
	
	##print STDOUT "[ The MSA representation of the gap patterns in the input MSA ]\n\n";
	##unless ( pr_ODEN_input (@{$seqnames}, @{$sec_msa}, 50, *STDOUT) ) {
	##    print STDOUT "[ERROR 'enum_psm_indel_scns_for_msa'] The subroutine 'pr_ODEN_input'  FAILED: $! !\n\n";
	##    return (undef, undef, undef, undef, undef, undef);
	##}
	print STDOUT "\n\n[ List of gap patterns (in order of their first occurrences) (#{gap-patterns} = $ct_gpatterns) ]\n\n";
	print STDOUT join ("\t", 'Indx_gp', "'gap_pattern(horizontal_representation)'"), "\n\n";

	for (my $indx_gp = 0; $indx_gp<$ct_gpatterns; $indx_gp++) { # Auxiliary for-loop #
	    my $gpattern = ${$set_gpatterns->[$indx_gp]};
	    print STDOUT join ("\t", $indx_gp, "'$gpattern'"), "\n";
	} 					# End of the auxiliary for-loop #

	print STDOUT "\n\n[ Clusters of contiguous blocks with particular gap patterns (#{clusters (i.e., gapped segments)} = $ct_cls) ]\n\n";

	for (my $cl=0; $cl<$ct_cls; $cl++) { # Outer auxiliary for-loop (over clusters) #

	    my $cls_gpblks = $set_cls_gpblocks->[$cl];
	    my $ct_gpblks = @{$cls_gpblks};

	    print STDOUT ' ' x 3, "{ The $cl th cluster (#{gap-pattern-blocks} = $ct_gpblks) }\n\n";
	    print STDOUT ' ' x 3, join ("\t", 'Block_no.', 'length', 'start_coord', 'end_coord', 'indx_gp', "'gap_pattern(horizontal_representation)'"), "\n\n";

	    for (my $bl=0; $bl<$ct_gpblks; $bl++) { # Inner for-loop (over the gap-blocks) #
		my ($indx_gp, $len_gb, $start, $end) = @{$cls_gpblks->[$bl]};

		my $gpattern = ${$set_gpatterns->[$indx_gp]};
		print STDOUT ' ' x 3, join ("\t", $bl, $len_gb, $start, $end, $indx_gp, "'$gpattern'"), "\n";
	    }			# End of the inner for-loop (over the gap-blocks) #
	    print STDOUT "\n\n";
	}		# End of the outer auxiliary for-loop (over clusters) #

    }	# END of "if ($TEST_LEVEL_1ST>2) {...}" for the test output for the STEP 1 . #

    if ($TEST_LEVEL_1ST>2) {
	print STDOUT "\n[[ TEST 2: Output of STEP 2 (1st half) (#{gap-patterns} = $ct_gpatterns) ]]\n\n";
	print STDOUT '[ Contents of @{$set_node2dparsstaat} and @{$set_br2change}) ]', "\n\n";

	for (my $indx_gp=0; $indx_gp < $ct_gpatterns; $indx_gp++) { # Outer auxiliary for-loop #
	    my $node2dparsstat = $set_node2dparsstat->[$indx_gp];
	    my $br2change = $set_br2change->[$indx_gp];

	    print STDOUT ' ' x 3, "{ The $indx_gp th gap pattern }\n\n";
	    print STDOUT ' ' x 3, '( Map of Dollo parsimonious states onto nodes )', "\n\n";
	    print STDOUT ' ' x 3, join ("\t", 'Node_ID', 'State', '#{children with #{exterior offsprings with \'presence\'}>0}'), "\n\n";
	    foreach my $node (sort {$a <=> $b} keys %{$node2dparsstat}) { print STDOUT ' ' x 3, join ("\t", $node, @{$node2dparsstat->{$node}}), "\n"; }
	    print STDOUT "\n";
	    print STDOUT ' ' x 3, '( Map of indel events onto branches )', "\n\n";
	    print STDOUT ' ' x 3, join ("\t", 'Branch_ID', 'state(upper-end)', 'state(lower-end)'), "\n\n"; 
	    foreach my $br (sort {$a <=> $b} keys %{$br2change}) { print STDOUT ' ' x 3, join ("\t", $br, @{$br2change->{$br}}), "\n"; }
	    print STDOUT "\n\n";
 
	}			# End of the outer auxiliary for-loop #

    }	# END of "if ($TEST_LEVEL_1ST>2) {...}" for the test output for STEP 2 (1st half). #


	# ($set_cls_gpblocks, $totcts0_indels_for_cls, $psm_hsts0_for_cls) = @{$info_psm_indel_hsts0};

    if ($TEST_LEVEL_1ST>2) {

	print STDOUT "\n[[ TEST 3: Output of STEP 2 (2nd half) (#{gapped segments} = $ct_cls) ]]\n\n";

	print STDOUT '[ Contents of @{$totcts0_indels_for_cls} and @{$psm_hsts0_for_cls} ]', "\n\n";

	for (my $cl=0; $cl<$ct_cls; $cl++) { # Outer auxiliary for-loop (over clusters) #

	    my $totct0_indels_for_cls = $totcts0_indels_for_cls->[$cl];

	    my ($br2indels0, $node2bw_gstats0) = @{$psm_hsts0_for_cls->[$cl]};
	    
	    print STDOUT ' ' x 3, "{ The $cl th cluster (Total#{indels}_init = $totct0_indels_for_cls) }\n\n";

	    print STDOUT ' ' x 3, "( Map of INITIAL block-wise gap states onto the nodes )\n\n";
	    print STDOUT ' ' x 3, join ("\t", 'Node_ID', "'block_wise_gap_states0(concatenated)'"), "\n\n";
	    
	    foreach my $node (sort {$a <=> $b} keys %{$node2bw_gstats0}) { # 1st inner foreach-loop (over nodes). #
		my $cnct_bw_gstats0 = join ('', @{$node2bw_gstats0->{$node}});
		print STDOUT ' ' x 3, join ("\t", $node, "'$cnct_bw_gstats0'"), "\n";
	    }				# End of the 1st inner foreach-loop (over nodes). #
	    print STDOUT "\n\n";

	    print STDOUT ' ' x 3, "( Map of INITIAL indel events onto the branches )\n\n";
	    print STDOUT ' ' x 3, join ("\t", 'Branch_ID', 'state_ue0/state_le0/indices,blocks,involved0', '...'), "\n\n";
	    foreach my $br (sort {$a <=> $b} keys %{$br2indels0}) { # 2nd inner foreach-loop (over branches). #
		my @out_indel_info = ();
		foreach my $indel_info (@{$br2indels0->{$br}}) { # Innermost foreach-loop (over indels along $br). #
		    my ($stat_ue, $stat_le, $indices_blk) = @{$indel_info};
		    (defined $stat_ue) or ($stat_ue = 'undef');
		    (defined $stat_le) or ($stat_le = 'undef');
		    my $cnct_indices_blk = (defined $indices_blk) ? join (',', @{$indices_blk}) : 'undef';
		    push @out_indel_info, join ('/', $stat_ue, $stat_le, $cnct_indices_blk);
		}			# End of the innermost foreach-loop (over indels along $br). #
		print STDOUT ' ' x 3, join ("\t", $br, @out_indel_info), "\n";
	    }				# End of the 2nd inner foreach-loop (over branches). ##
	    print STDOUT "\n\n";

	}			# End of the outer auxiliary for-loop (over clusters) #

    }   # END of "if ($TEST_LEVEL_1ST>2) {...}" for the test output for STEP 2 (2nd half). #


    (print STDOUT "\n<< Subroutine 'enum_psm_indel_hsts_for_msa_finer_1ST_HF' ENDED />>\n\n") if ($TEST_LEVEL_1ST>1/4);

	## my ($set_cls_gpblocks, $totcts0_indels_for_cls, $psm_hsts0_for_cls) = @{$info_psm_indel_hsts0};

	return ($set_gpatterns, $set_node2dparsstat, $set_br2change, $info_psm_indel_hsts0);

}	# End of the subroutine, 'enum_psm_indel_hsts_for_msa_finer_1ST_HF'. #


# enum_psm_indel_hsts_for_msa_finer_2ND_HF:
#	This is a portion of the INTEGRATED subroutine, "enum_psm_indel_hsts_for_msa_finer".
#	It does the second half of the main works (STEPS 3 & 4),
#	up to enumerating the parsimonious indel histories for each (effectively independent) indel block in each gapped segment.
#
#	(It simply performs the subroutine, 'enum_psm_indel_hsts_in_cls_blks_finer', serially on all gapped segments in the input MSA.)
#
#	[CREATED on May 26 (Fri), 2017] 
#
# INPUT: ($topnode_id, \%node2ch, \%node2pa, $char_abs, $char_pres, \%node2ct_exoffs, \%br2eq, \%br2nn, \@set_all_nodes,
#	\@set_gpatterns, \@set_node2dparsstat, \@set_br2change, \@info_psm_indel_hsts0), where
#
#	%node2ch = ($node_id => \@child_node_ids, ...) is for the all interior nodes;
#	%node2pa = ($node_id => $parent_node_id, ...) is for all the nodes except the top node;
#	$char_pres/$char_abs is the character (or symbol) representing the 'presence'/'absence' state;
#
#	%node2ct_exoffs = ($node_id => #{external offsprings}, ...);
#	%br2eq = ($branch_id => $equivalent_branch_id, ...) for branches with equivalents;
#	%br2nn = ($branch_id => {$id_of_a_neighboring_branch => 1, ...}, ...); 
#	@set_all_nodes stores the IDs of all nodes;
#
#	\@set_gpatterns, \@set_node2dparsstat and \@set_br2change are the three components of @info_gpatterns,
#	which in turn is a piece of the output of "cal_init_psm_cands_for_sgl_msa" in "MyTreeMap_indels_spt_odr2.pm" (STEP 2);
#
#	@info_init_psm_hsts0 (= (\@set_cls_gpblocks, \@totcts0_indels_for_cls, \@psm_hsts0_for_cls))
#	is a piece of the output of "cal_init_psm_cands_for_sgl_msa" (STEP 2).
#
# 
# OUTPUT: (\@sets_min_totcts_indels_in_indel_blks, \@setssets_psm_hsts_in_indel_blks, \@sets_psm_hsts0_in_indel_blks, \@set_lists_off_limits,
#	\@set_lists_gblocks, \@set_lists_seqblocks, \@sets_eff_indep_indel_blocks, \@sets_indices_incl_root), where
#
#	$sets_min_totcts_indels_in_indel_blks[$cl]->[$eiidb] is the total number of insertion/deletion events in the most parsimonious history(ies) realizing the $eiidb th effectively independent indel block in the $cl th gapped segment;
#
#	@{$setssets_psm_hsts_in_indel_blks[$cl]->[$eiidb]->[$ph]} = (\%br2indels, \%node2blwise_gapstat) is for the $ph th of the parsimonious histories realizing the $eiidb th effectively independent indel block in the $cl th cluster, where
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_gpb) describing the $id th indel event at the branch,
# with @indices_gpb listing the indices (in @{$set_cls_gpblocks[$cl]}) of the gap-pattern blocks involved in the indel event,
#	%node2blwise_gapstat = ($node_id => \@blwise_gapstats, ...) is the map of gap states onto the interior nodes, 
#	with $blwise_gapstats[$gpb] being $char_pres/abs for the $gpb th gap-pattern block (in the cluster) at the node $node_id.
#
#	@{$sets_psm_hsts0_in_indel_blks[$cl]->[$eiidb]} = (\%br2indels0, \%node2blwise_gapstat0) is for the initial candidate history realizing the $eiidb th effectively independent indel block in the $cl th cluster;
#
#	@{$set_lists_off_limits[$cl]} is the output of "define_off_limits" in this module (STEP 3A),
#	performed on the $cl th gapped segment;
#
#	@{$set_lists_gblocks[$cl]} & @{$set_lists_seqblocks[$cl]} are the output of "partit_into_gblocks_seqblocks" in this module (STEP 3A),
#	performed on the $cl th gapped segment;
#
#	@{$sets_eff_indep_indel_blocks[$cl]} is the output of "partit_into_eff_indep_indel_blocks" in this module (STEP 3B),
#	performed on the $cl th gapped segment;
#
#	@{$sets_indices_incl_root[$cl]} is also the output of "partit_into_eff_indep_indel_blocks" in this module (STEP 3B),
#	performed on the $cl th gapped segment.
#
#
sub enum_psm_indel_hsts_for_msa_finer_2ND_HF ($\%\%$$\%\%\%\@\@\@\@\@) {

	my ($topnode_id, $node2ch, $node2pa, $char_abs, $char_pres, $node2ct_exoffs, $br2eq, $br2nn, $set_all_nodes, $set_gpatterns, $set_node2dparsstat, $set_br2change, $info_psm_indel_hsts0) = @_;

	my ($set_cls_gpblocks, $totcts0_indels_for_cls, $psm_hsts0_for_cls) = @{$info_psm_indel_hsts0};

	my $ct_cls = @{$set_cls_gpblocks};

    (print STDOUT "<< Subroutine 'enum_psm_indel_hsts_for_msa_finer_2ND_HF' (Top node ID: $topnode_id; #{gapped segments} = $ct_cls) >>\n\n") if ($TEST_LEVEL_2ND>1/4);


    (print STDOUT '/// [ STEP 3 ] Using the initial candidate indel history, vertically partition each gapped segment into effectively independent indel blocks. ///', "\n\n") if ($TEST_LEVEL_2ND>1);
    (print STDOUT '/// & [ STEP 4 ] Attempt to enumerate parsimonious indel histories within each effectively independent indel block. ///', "\n\n") if ($TEST_LEVEL_2ND>1);


		# Initialize the Output Variables. #

	my @sets_min_totcts_indels_in_indel_blks = ();
	my @setssets_psm_hsts_in_indel_blks = ();
	my @sets_psm_hsts0_in_indel_blks = (); # ADDED on May 27 (Sat), 2017. #
	my @set_lists_off_limits = (); # OBSOLETE, but will be kept for potential future use. #
	my @set_lists_gblocks = ();
	my @set_lists_seqblocks = ();
	my @sets_eff_indep_indel_blocks = ();
	my @sets_indices_incl_root = ();

	for (my $cl=0; $cl<$ct_cls; $cl++) { # Main outermost for-loop (over the clusters) #

	    my $gpblocks_in_cls = my $cls_gpblks = $set_cls_gpblocks->[$cl];
	    my $ct_gpblks = @{$cls_gpblks};

    (print STDOUT ' ' x 3, "{ The $cl th cluster (#{gap-pattern-blocks} = $ct_gpblks) }\n\n") if ($TEST_LEVEL_2ND>1);

    if ($TEST_LEVEL_2ND>3) {
	print STDOUT ' ' x 3, "( Gap-pattern blocks in the cluster )\n\n";
	print STDOUT ' ' x 3, join ("\t", 'Block_no.', 'length', 'start_coord', 'end_coord', 'indx_gp', "'gap_pattern(horizontal_representation)'"), "\n\n";

	for (my $bl=0; $bl<$ct_gpblks; $bl++) { # Inner auxiliary for-loop (over the gap-pattern-blocks) #
	    my ($indx_gp, $len_gb, $start, $end) = @{$cls_gpblks->[$bl]};

	    my $gpattern = ${$set_gpatterns->[$indx_gp]};
	    print STDOUT ' ' x 3, join ("\t", $bl, $len_gb, $start, $end, $indx_gp, "'$gpattern'"), "\n";
	}			# End of the inner auxiliary for-loop (over the gap-pattern-blocks) #
	print STDOUT "\n\n";
    } # END of "if ($TEST_LEVEL_2ND>3) {...}"

		# Before advancing to the core sub-routine, 'enum_psm_indel_hsts_in_cls_blks_finer',
		# screen the gap-block cluster (i.e., gapped segment) with its length itself.
		# (ADDED in revision on Mar 11 (Tue), 2014. => IMPORTED with minor modifications from 'enum_psm_indel_scns_for_msa' on May 25 (Thu), 2017.) #

	    my $cl_start = $cls_gpblks->[0][2];
	    my $cl_end = $cls_gpblks->[$ct_gpblks - 1][3];
	    my $cl_len = $cl_end - $cl_start + 1;

	    unless ($cl_len <= $UPPER_BOUND_ANAL_LEN_GBCLS) {
		print STDOUT "[WARNING 'enum_psm_indel_hsts_for_msa_finer_2ND_HF'] The $cl th cluster is of length $cl_len, which is LONGER THAN the upper-bound, $UPPER_BOUND_ANAL_LEN_GBCLS... I will skip analyzing this cluster: $! !\n\n";

			# Assign an empty set (or undef) as a component of each output array. #
		$sets_min_totcts_indels_in_indel_blks[$cl] = [];
		$setssets_psm_hsts_in_indel_blks[$cl] = [];
		$sets_psm_hsts0_in_indel_blks[$cl] = []; # ADDED on May 27 (Sat), 2017. # 
		$set_lists_gblocks[$cl] = [];
		$set_lists_seqblocks[$cl] = undef;
		$sets_eff_indep_indel_blocks[$cl] = [];
		$sets_indices_incl_root[$cl] = undef;
		next;
	    }

		# Before advancing to the core sub-routine, 'enum_psm_indel_hsts_in_cls_blks_finer',
		# scan the gap-block cluster for long gaps.
		# (ADDED in revision on Mar 10 (Mon), 2014. => IMPORTED with minor modifications from 'enum_psm_indel_scn     s_for_msa' on May 25 (Thu), 2017.) #

	    my $list_long_gaps = scan_gbcls_for_long_gaps (@{$cls_gpblks}, @{$set_gpatterns}, $UPPER_BOUND_ANAL_GAPLEN, $char_abs);
	    unless (defined $list_long_gaps) {
		print STDOUT "[ERROR 'enum_psm_indel_hsts_for_msa_finer_2ND_HF'] The subroutine 'scan_gbcls_for_long_gaps' FAILED on the $cl th cluster: $! !\n\n";
		return (undef, undef, undef, undef, undef, undef, undef, undef);
	    }
	    unless (@{$list_long_gaps} == 0) {
		my $ct_long_gaps = @{$list_long_gaps};
		print STDOUT "[WARNING 'enum_psm_indel_hsts_for_msa_finer_2ND_HF'] The $cl th cluster contains $ct_long_gaps gaps longer than $UPPER_BOUND_ANAL_GAPLEN. So, I will skip analyzing this cluster: $! !\n\n"; 

			# Assign an empty set (or undef) as a component of each output array. #
		$sets_min_totcts_indels_in_indel_blks[$cl] = [];
		$setssets_psm_hsts_in_indel_blks[$cl] = [];
		$sets_psm_hsts0_in_indel_blks[$cl] = []; # ADDED on May 27 (Sat), 2017. #
		$set_lists_gblocks[$cl] = [];
		$set_lists_seqblocks[$cl] = undef;
		$sets_eff_indep_indel_blocks[$cl] = [];
		$sets_indices_incl_root[$cl] = undef;
		next;
	    }

			### Perform the CORE (Composite-)Subroutine, 'enum_psm_indel_hsts_in_cls_blks_finer'. ###

	    my $totct0_indels_for_cls = $totcts0_indels_for_cls->[$cl];
	    my $psm_hst0_for_cls = $psm_hsts0_for_cls->[$cl];
	    my ($br2indels0_for_cls, $node2bw_gstats0_for_cls) = @{$psm_hst0_for_cls};

# INPUT: ($top_node_id, \@set_all_nodes, \%node2pa, \%node2ch, \%node2ct_exoffs, $char_pres, $char_abs, \@gpblocks_in_cls, \%br2indels0, \%node2blwise_gstats0, \%br2eq, \%br2nn), where
#
# OUTPUT: (\@min_totcts_indels_in_indel_blks, \@sets_psm_hsts_in_indel_blks,
#	   \@list_gblocks, \@list_seqblocks, \@set_eff_indep_indel_blocks, \@indices_incl_root), where

	    my ($min_totcts_indels_in_eiidb, $sets_psm_hsts_in_eiidb, $psm_hsts0_in_eiidb, $list_gblks, $list_seqblks, $set_eiidbs, $indices_incl_root)
		 = enum_psm_indel_hsts_in_cls_blks_finer ($topnode_id, @{$set_all_nodes}, %{$node2pa}, %{$node2ch}, %{$node2ct_exoffs}, $char_pres, $char_abs, @{$gpblocks_in_cls},  %{$br2indels0_for_cls}, %{$node2bw_gstats0_for_cls}, %{$br2eq}, %{$br2nn});

	    unless (defined $sets_psm_hsts_in_eiidb) {
	 	print STDOUT "[ERROR 'enum_psm_indel_hsts_for_msa_finer_2ND_HF'] The CORE composite-subroutine 'enum_psm_indel_hsts_in_cls_blks_finer' FAILED on the $cl th cluster: $! !\n\n";
		return (undef, undef, undef, undef, undef, undef, undef, undef);
	    }
	    my $ct_eiidbs = @{$set_eiidbs};

			# Sorts of sanity checks. #
	    unless (scalar (@{$min_totcts_indels_in_eiidb} ) == $ct_eiidbs) {
		print STDOUT "[ERROR 'enum_psm_indel_hsts_for_msa_finer_2ND_HF'] For the $cl th cluster, the number of minimum total counts of indels in the indel blocks (= ", scalar (@{$min_totcts_indels_in_eiidb}), ") CLASHES with that of indel blocks (= $ct_eiidbs): $! !\n\n";
		return (undef, undef, undef, undef, undef, undef, undef, undef);
	    }
	    unless (scalar (@{$sets_psm_hsts_in_eiidb} ) == $ct_eiidbs) {
		print STDOUT "[ERROR 'enum_psm_indel_hsts_for_msa_finer_2ND_HF'] For the $cl th cluster, the number of sets of parsimonious indel histories in the indel blocks (= ", scalar (@{$sets_psm_hsts_in_eiidb}), ") CLASHES with that of indel blocks (= $ct_eiidbs): $! !\n\n";
		return (undef, undef, undef, undef, undef, undef, undef, undef);
	    }
	    unless (scalar (@{$psm_hsts0_in_eiidb} ) == $ct_eiidbs) {
		print STDOUT "[ERROR 'enum_psm_indel_hsts_for_msa_finer_2ND_HF'] For the $cl th cluster, the number of initial indel histories in the indel blocks (= ", scalar (@{$psm_hsts0_in_eiidb}), ") CLASHES with that of indel blocks (= $ct_eiidbs): $! !\n\n";
		return (undef, undef, undef, undef, undef, undef, undef, undef);
	    }

	    my $cnct_min_totcts_indels = join (',', @{$min_totcts_indels_in_eiidb});

	    my @set_cts_hsts = ();
	    for (my $k=0; $k < $ct_eiidbs; $k++) { $set_cts_hsts[$k] = scalar (@{$sets_psm_hsts_in_eiidb->[$k]}); }
	    my $cnct_cts_hsts = join (',', @set_cts_hsts);
    
    (print STDOUT "\n --- 'enum_psm_indel_hsts_in_cls_blks_finer' Ended (#{eiidbs} = $ct_eiidbs;  Set of Min#{indels} = $cnct_min_totcts_indels; Set of #{parsimonious histories} = $cnct_cts_hsts) ...\n\n") if ($TEST_LEVEL_2ND>1);

    if ($TEST_LEVEL_2ND>3) {

	print STDOUT ' ' x 3, "[ TEST 5: Main Output of 'enum_psm_indel_hsts_in_cls_blks_finer' (#{eiidbs} = $ct_eiidbs) ]\n\n";

			## ADDED on May 27 (Sat), 2017. ##

	my $ct_gbs = @{$list_gblks};
	my $ct_sqbs = @{$list_seqblks};

	print STDOUT ' ' x 3, "{ Gap-blocks (#{blocks} = $ct_gbs) }\n\n";

	print STDOUT ' ' x 3, join ("\t", 'Indx_gb', 'br', 'up_or_lw', 'start', 'end', '(list,invlvd,gpbs)', '(list,nesting,seqblocks)'), "\n\n";

	for (my $gb = 0; $gb < $ct_gbs; $gb++) { # Preliminary for-loop (over gap-blocks). #

	    my ($br, $up_or_lw, $start, $end, $invlvd_gpbs, $nesting) = @{$list_gblks->[$gb]};
	    my $cnct_invlvd_gpbs = join (',', @{$invlvd_gpbs});
	    $cnct_invlvd_gpbs = join ('', '(', $cnct_invlvd_gpbs, ')');
	    my $cnct_nesting = join (',', @{$nesting});
	    $cnct_nesting = join ('', '(', $cnct_nesting, ')');
	    print STDOUT ' ' x 3, join ("\t", $gb, $br, $up_or_lw, $start, $end, $cnct_invlvd_gpbs, $cnct_nesting), "\n" ;
	}				# End of the preliminary for-loop (over gap-blocks). #
	print STDOUT "\n\n";


	print STDOUT ' ' x 3, "{ Sequence-blocks (#{blocks} = $ct_sqbs) }\n\n";

	print STDOUT ' ' x 3, join ("\t", 'Indx_sb', 'br', 'up_or_lw', 'start', 'end', '(list,invlvd,gpbs)', '(list,nested,gblocks)'), "\n\n";

	for (my $sb = 0; $sb < $ct_sqbs; $sb++) { # Preliminary for-loop (over sequence-blocks). #

	    my ($br, $up_or_lw, $start, $end, $invlvd_gpbs, $nested) = @{$list_seqblks->[$sb]};
	    my $cnct_invlvd_gpbs = join (',', @{$invlvd_gpbs});
	    $cnct_invlvd_gpbs = join ('', '(', $cnct_invlvd_gpbs, ')');
	    my $cnct_nested = join (',', @{$nested});
	    $cnct_nested = join ('', '(', $cnct_nested, ')');
	    print STDOUT ' ' x 3, join ("\t", $sb, $br, $up_or_lw, $start, $end, $cnct_invlvd_gpbs, $cnct_nested), "\n";
	}			# End of the preliminary for-loop (over sequence-blocks). #
	print STDOUT "\n\n";

	print STDOUT ' ' x 3, "{ Indel blocks including the root:  ", join (', ', @{$indices_incl_root}), " }\n\n\n";


	print STDOUT ' ' x 3 , "{{ Information on the $ct_eiidbs Indel blocks and the Parsimonious indel histories for each of them }}\n\n\n"; # ADDED on May 27 (Sat), 2017. #

			## END of "ADDED on May 27 (Sat), 2017." ##


      for (my $k=0; $k<$ct_eiidbs; $k++) { # Auxiliary outermost for-loop (over indel blocks). #

	my $info_eiidb = $set_eiidbs->[$k];
	my $min_totct_indels = $min_totcts_indels_in_eiidb->[$k];
	my $set_psm_hsts = $sets_psm_hsts_in_eiidb->[$k];
	my $ct_hsts = @{$set_psm_hsts};

	my ($set_delimiting_bds, $start_eiidb, $end_eiidb, $union_invlvd_gpbs, $invlvd_gbs, $invlvd_sbs) = @{$info_eiidb};
	my $ct_delim_bds = @{$set_delimiting_bds};
	my $cnct_invlvd_gbs = join (',', @{$invlvd_gbs});
	my $cnct_invlvd_sbs = join (',', @{$invlvd_sbs});

	print STDOUT ' ' x 3, "|| The $k th indel block (min#{indels} = $min_totct_indels; #{parsimonious histories} = $ct_hsts; {involved gap-blocks} = ($cnct_invlvd_gbs); {involved sequence-blocks} = ($cnct_invlvd_sbs) ) ||\n\n";

	print STDOUT ' ' x 3, "{ Set of boundaries delimiting the indel block (#{boundaries} = $ct_delim_bds) }\n\n";

	print STDOUT ' ' x 3, join ("\t", 'Index', 'Br_ID', 'Up_or_Lw', 'lm_gpb', 'rm_gpb', 'involved,gap,pattern,blocks'), "\n\n";

	for (my $db = 0; $db<$ct_delim_bds; $db++) { # Preliminary outer for-loop (over delimiting boundaries). #

	    my ($br, $up_or_lw, $lm_gpb, $rm_gpb, $invlvd_gpbs) = @{$set_delimiting_bds->[$db]};
	    my $cnct_invlvd_gpbs = join (',', @{$invlvd_gpbs});

	    print STDOUT ' ' x 3, join ("\t", $db, $br, $up_or_lw, $lm_gpb, $rm_gpb, $cnct_invlvd_gpbs), "\n";

	}			# End of the preliminary outer for-loop (over delimiting boundaries). #

	print STDOUT "\n\n";

				# ADDED on May 27 (Sat), 2017. #

	print STDOUT ' ' x 3, "{ Initial candidate indel history (within the indel block) }\n\n\n";

	my ($br2indels0, $node2bw_gstats0) = @{$psm_hsts0_in_eiidb->[$k]};
	    print STDOUT ' ' x 5, "( Map of block-wise gap states onto the nodes )\n\n";
	    print STDOUT ' ' x 5, join ("\t", 'Node_ID', "'block_wise_gap_states(concatenated)'"), "\n\n";
	    foreach my $node (sort {$a <=> $b} keys %{$node2bw_gstats0}) { # 1st inner foreach-loop #
		my $cnct_bw_gstats = join ('', @{$node2bw_gstats0->{$node}});
		print STDOUT ' ' x 5, join ("\t", $node, "'$cnct_bw_gstats'"), "\n";
	    }				# End of the 1st inner foreach-loop #
	    print STDOUT "\n\n";

	    print STDOUT ' ' x 5, "( Map of indel events onto the branches )\n\n";
	    print STDOUT ' ' x 5, join ("\t", 'Branch_ID', 'state_ue/state_le/indices,blocks,involved', '...'), "\n\n";
	    foreach my $br (sort {$a <=> $b} keys %{$br2indels0}) { # 2nd inner foreach-loop #
		my @out_indel_info = ();
		foreach my $indel_info (@{$br2indels0->{$br}}) { # Innermost foreach-loop #
		    my ($stat_ue, $stat_le, $indices_blk) = @{$indel_info};
		    (defined $stat_ue) or ($stat_ue = 'undef');
		    (defined $stat_le) or ($stat_le = 'undef');
		    my $cnct_indices_blk = (defined $indices_blk) ? join (',', @{$indices_blk}) : 'undef';
		    push @out_indel_info, join ('/', $stat_ue, $stat_le, $cnct_indices_blk);
		}			# End of the innermost foreach-loop #
		print STDOUT ' ' x 5, join ("\t", $br, @out_indel_info), "\n";
	    }				# End of the 2nd inner foreach-loop #
	    print STDOUT "\n\n";

				# END of "ADDED on May 27 (Sat), 2017." #


	print STDOUT ' ' x 3, "{ Set of parsimonous histories (#{histories} = $ct_hsts) }\n\n\n";


#	@{$eff_indep_indel_blocks[$eiidb]} = (\@set_delimiting_bds, $start, $end, \@union_invlvd_gpbs, \@invlvd_gbs, \@invlvd_sbs)
#	gives information on the $eiidb th effectively independent indel block.
#
#	@set_delimiting_bds stores information on all boundaries delimiting the indel block,
#	with each element being of the form [$br, $up_or_lw, $lm_gpb, $rm_gpb, \@invlvd_gpbs]
#	$br is the ID of the branch delimiting the indel block,
#	$up_or_lw = 'U'/'L' if the upper/lower-end of $br is the indel block, 
#	$lm/rm_gpb is the leftmost/rightmost gap-pattern block involved,
#	@invlvd_gpbs lists the gap-pattern-blocks involved in this boundary.

	for (my $ps=0; $ps<$ct_hsts; $ps++) { # Axiliary outer for-loop (over parsimonious scenarios (histories)) #

	    print STDOUT ' ' x 5, "| The $ps th history |\n\n";

	    my ($br2indels, $node2bw_gstats) =  @{$set_psm_hsts->[$ps]};

	    print STDOUT ' ' x 5, "( Map of block-wise gap states onto the nodes )\n\n";
	    print STDOUT ' ' x 5, join ("\t", 'Node_ID', "'block_wise_gap_states(concatenated)'"), "\n\n";
	    foreach my $node (sort {$a <=> $b} keys %{$node2bw_gstats}) { # 1st inner foreach-loop #
		my $cnct_bw_gstats = join ('', @{$node2bw_gstats->{$node}});
		print STDOUT ' ' x 5, join ("\t", $node, "'$cnct_bw_gstats'"), "\n";
	    }				# End of the 1st inner foreach-loop #
	    print STDOUT "\n\n";

	    print STDOUT ' ' x 5, "( Map of indel events onto the branches )\n\n";
	    print STDOUT ' ' x 5, join ("\t", 'Branch_ID', 'state_ue/state_le/indices,blocks,involved', '...'), "\n\n";
	    foreach my $br (sort {$a <=> $b} keys %{$br2indels}) { # 2nd inner foreach-loop #
		my @out_indel_info = ();
		foreach my $indel_info (@{$br2indels->{$br}}) { # Innermost foreach-loop #
		    my ($stat_ue, $stat_le, $indices_blk) = @{$indel_info};
		    (defined $stat_ue) or ($stat_ue = 'undef');
		    (defined $stat_le) or ($stat_le = 'undef');
		    my $cnct_indices_blk = (defined $indices_blk) ? join (',', @{$indices_blk}) : 'undef';
		    push @out_indel_info, join ('/', $stat_ue, $stat_le, $cnct_indices_blk);
		}			# End of the innermost foreach-loop #
		print STDOUT ' ' x 5, join ("\t", $br, @out_indel_info), "\n";
	    }				# End of the 2nd inner foreach-loop #
	    print STDOUT "\n\n";

	}			# End of the auxiliary outer for-loop (over parsimonious scenarios (histories)) #
	print STDOUT "\n";

      }			# End of the auxiliary outermost for-loop (over indel blocks). #

    }  # END of "if ($TEST_LEVEL_2ND>3) {...} for the Main Output of 'enum_psm_indel_hsts_in_cls_blks_finer'. #

		$sets_min_totcts_indels_in_indel_blks[$cl] = $min_totcts_indels_in_eiidb;
		$setssets_psm_hsts_in_indel_blks[$cl] = $sets_psm_hsts_in_eiidb;
		$sets_psm_hsts0_in_indel_blks[$cl] = $psm_hsts0_in_eiidb; # ADDED on May 27 (Sat), 2017. #
		$set_lists_gblocks[$cl] = $list_gblks;
		$set_lists_seqblocks[$cl] = $list_seqblks;
		$sets_eff_indep_indel_blocks[$cl] = $set_eiidbs;
		$sets_indices_incl_root[$cl] = $indices_incl_root;

	    ## ($min_totcts_indels_in_eiidb, $sets_psm_hsts_in_eiidb, $list_gblks, $list_seqblks, $set_eiidbs, $indices_incl_root)
		### RESTART FROM HERE ###


	}			# End of the main outermost for-loop (over the clusters) #

    (print STDOUT "\n/// The STEP 3 & STEP 4 Ended ///\n\n") if ($TEST_LEVEL_2ND>1);

    (print STDOUT "\n<< Subroutine 'enum_psm_indel_hsts_for_msa_finer_2ND_HF' ENDED />>\n\n") if ($TEST_LEVEL_2ND>1/4);


	return (\@sets_min_totcts_indels_in_indel_blks, \@setssets_psm_hsts_in_indel_blks, \@sets_psm_hsts0_in_indel_blks, \@set_lists_off_limits,
		\@set_lists_gblocks, \@set_lists_seqblocks, \@sets_eff_indep_indel_blocks, \@sets_indices_incl_root);

}	# End of the subroutine, 'enum_psm_indel_hsts_for_msa_finer_2ND_HF'. #


	# The INTEGRATED Subroutine. #

# enum_psm_indel_hsts_for_msa_finer:
#	This subroutine INTEGRATEs all the processes in STEPS1-4 (of "workflow_for_indel_histories&multiplication_factors.rtf"),
#	and attempts to enumerate all parsimonious indel histories resulting in the input MSA,
#	in a FINER way than its predecessor, "enum_psm_indel_scns_for_msa" (in "MyTreeMap_indels_spt_odr.pm"), does.
#
#	NOTE that this subroutine eats up the input MSA.
#	
#	[CREATION STARTED on Feb 25th (Sat), 2017, ENDED on Mmm dd (www), 2017.]
#
# [ enum_psm_indel_scns_for_msa:
#	This is one of the main subroutines in this Perl module.
#	to enumerate all parsimonious indel scenarios
#	each of which realizes the input MSA.
#	It puts together several subroutines in this Perl module,
#	out of which 'enum_psm_indel_scns_in_cls_blks' and 'map_parsim_psm_dollo_msa'
#	play core roles,
#	and 'extract_cls_gblocks' and 'msa_rep_gpatterns'
#	perform necessary pre-processes of the input data. ] ... Predecessor...
#	{For more details, see "MyTreeMap_indels_spt_odr.pm".}
#
#
# INPUT: (\%node2ch, \%node2pa, \%name2id, \@seqnames, \@in_msa, $char_abs, $char_pres), where
#
#	%node2ch = ($node_id => \@child_node_ids, ...) is for the all interior nodes;
#	%node2pa = ($node_id => $parent_node_id, ...) is for all the nodes except the top node;
#	%name2id = ($node_name => $node_id, ...) is for the exterior nodes and the top node (with $node_name = $TOP_NODE);
#	$seqnames[$k] is the name of the $k th sequence in the input MSA
#	(NOTE that these names must exactly match those used in %name2id);
#	${$in_msa[$k]} is the $k th sequence embedded in the input MSA;
#	$char_pres/$char_abs is the character (or symbol) representing the 'presence'/'absence' state;
#
#
# OUTPUT:  (\@sets_min_totcts_indels_in_indel_blks, \@setssets_psm_hsts_in_indel_blks, \@sets_psm_hsts0_in_indel_blks,
#		\@set_gpatterns, \@set_node2dparsstat, \@set_br2change,
#		\@set_cls_gpblocks, \@info_init_psm_scn, \@set_lists_off_limits, \@set_lists_gblocks, \@set_lists_seqblocks, \@sets_eff_indep_indel_blocks, \@sets_indices_incl_root), where
#
#	Main output consists of (\@sets_min_totcts_indels_in_indel_blks, \@setsset_psm_hsts_in_indel_blks), where
#
#	$sets_min_totcts_indels_in_indel_blks[$cl]->[$eiidb] is the total number of insertion/deletion events in the most parsimonious history(ies) realizing the $eiidb th effectively independent indel block in the $cl th gapped segment;
#
#	@{$setssets_psm_hsts_in_indel_blks[$cl]->[$eiidb]->[$ph]} = (\%br2indels, \%node2blwise_gapstat) is for the $ph th of the parsimonious histories realizing the $eiidb th effectively independent indel block in the $cl th cluster, where
#	%br2indels = ($branch_id => \@list_indels, ...) is the map of insertions/deletions onto branches, with
#	@{$list_indels[$id]} = ($stat_parent, $stat_node, \@indices_gpb) describing the $id th indel event at the branch,
# with @indices_gpb listing the indices (in @{$set_cls_gpblocks[$cl]}) of the gap-pattern blocks involved in the indel event,
#	%node2blwise_gapstat = ($node_id => \@blwise_gapstats, ...) is the map of gap states onto the interior nodes, 
#	with $blwise_gapstats[$gpb] being $char_pres/abs for the $gpb th gap-pattern block (in the cluster) at the node $node_id.
#
#	@{$sets_psm_hsts0_in_indel_blks[$cl]->[$eiidb]} = (\%br2indels0, \%node2blwise_gapstat0) is for the initial candidate history realizing the $eiidb th effectively independent indel block in the $cl th cluster.
#
#
#	The first set of by-products consists of (\@set_gpatterns, \@set_node2dparsstat, \@set_br2change),
#	which are the three components of @info_gpatterns,
#	which in turn is a piece of the output of "cal_init_psm_cands_for_sgl_msa" in "MyTreeMap_indels_spt_odr2.pm" (STEP 2).
#
#	The second set of by-products consists of (\@set_cls_gpblocks, \@info_init_psm_scn, \@set_lists_off_limits, \@set_lists_gblocks, \@set_lists_seqblocks, \@sets_eff_indep_indel_blocks), where
#
#	@set_cls_gpblocks is a component of @info_init_psm_scn,
#	which in turn is a piece of the output of "cal_init_psm_cands_for_sgl_msa" (STEP 2);
#
#	@info_init_psm_hsts is a piece of the output of "cal_init_psm_cands_for_sgl_msa" (STEP 2);
#
#	@{$set_lists_off_limits[$cl]} is the output of "define_off_limits" in this module (STEP 3A),
#	performed on the $cl th gapped segment;
#
#	@{$set_lists_gblocks[$cl]} & @{$set_lists_seqblocks[$cl]} are the output of "partit_into_gblocks_seqblocks" in this module (STEP 3A),
#	performed on the $cl th gapped segment;
#
#	@{$sets_eff_indep_indel_blocks[$cl]} is the output of "partit_into_eff_indep_indel_blocks" in this module (STEP 3B),
#	performed on the $cl th gapped segment;
#
#	@{$sets_indices_incl_root[$cl]} is also the output of "partit_into_eff_indep_indel_blocks" in this module (STEP 3B),
#	performed on the $cl th gapped segment.
#
#
sub enum_psm_indel_hsts_for_msa_finer (\%\%\%\@\@$$) {
 
	my ($node2ch, $node2pa, $name2id, $seqnames, $in_msa, $char_abs, $char_pres) = @_;

	my $topnode_id = $name2id->{$TOP_NODE};
	my $ct_seqs = @{$seqnames};
	unless ($ct_seqs == scalar (@{$in_msa})) {
	    print STDOUT "[ERROR 'enum_psm_indel_hsts_for_msa_finer'] ILLEGAL INPUT: The number of sequence names (=$ct_seqs) CLASHES with the number of sequences themselves (=", scalar (@{$in_msa}), ") in the input MSA: $! !\n\n";
	    return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	}

    (print STDOUT "<<< Subroutine 'enum_psm_indel_hsts_for_msa_finer' (Top node ID: $topnode_id\; #{sequences} = $ct_seqs) >>>\n\n") if ($TEST_LEVEL>0);

			### Preliminary works ###

	my ($node2depth, $node2ct_exoffs, $br2eq, $br2nn, $set_all_nodes)
	    = enum_psm_indel_hsts_for_msa_finer_PRE ($topnode_id, %{$node2ch}, %{$node2pa});
	unless (defined $br2nn) {
	    print STDOUT "[ERROR 'enum_psm_indel_scns_for_msa_for_finer'] Subroutine 'enum_psm_indel_hsts_for_msa_finer_PRE' FAILED: $! !\n\n";
	    return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	}
    (print STDOUT "\n=== Subroutine 'enum_psm_indel_hsts_for_msa_finer_PRE' Ended :::\n\n") if ($TEST_LEVEL>1);


			### First Half (STEPS 1 & 2) ###

	my ($set_gpatterns, $set_node2dparsstat, $set_br2change, $info_psm_indel_hsts0)
	    = enum_psm_indel_hsts_for_msa_finer_1ST_HF (%{$node2ch}, %{$node2pa}, %{$name2id}, @{$seqnames}, @{$in_msa}, $char_abs, $char_pres, %{$node2depth}, %{$node2ct_exoffs});
	unless ((defined $info_psm_indel_hsts0) and (@{$info_psm_indel_hsts0}>2) ) {
	    print STDOUT "[ERROR 'enum_psm_indel_scns_for_msa_for_finer'] Subroutine 'enum_psm_indel_hsts_for_msa_finer_1ST_HF' FAILED: $! !\n\n";
	    return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	}
    (print STDOUT "\n=== Subroutine 'enum_psm_indel_hsts_for_msa_finer_1ST_HF' Ended :::\n\n") if ($TEST_LEVEL>1);

	my ($set_cls_gpblocks, $totcts0_indels_for_cls, $psm_hsts0_for_cls) = @{$info_psm_indel_hsts0};


			### Second Half (STEPS 3 & 4) ###

	my ($sets_min_totcts_indels_in_indel_blks, $setssets_psm_hsts_in_indel_blks, $sets_psm_hsts0_in_indel_blks, $set_lists_off_limits,
	    $set_lists_gblocks, $set_lists_seqblocks, $sets_eff_indep_indel_blocks, $sets_indices_incl_root)
		= enum_psm_indel_hsts_for_msa_finer_2ND_HF ($topnode_id, %{$node2ch}, %{$node2pa}, $char_abs, $char_pres, %{$node2ct_exoffs}, %{$br2eq}, %{$br2nn}, @{$set_all_nodes}, @{$set_gpatterns}, @{$set_node2dparsstat}, @{$set_br2change}, @{$info_psm_indel_hsts0});
	unless (defined $setssets_psm_hsts_in_indel_blks) {
	    print STDOUT "[ERROR 'enum_psm_indel_scns_for_msa_for_finer'] Subroutine 'enum_psm_indel_hsts_for_msa_finer_2ND_HF' FAILED: $! !\n\n";
	    return (undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef, undef);
	}
    (print STDOUT "\n=== Subroutine 'enum_psm_indel_hsts_for_msa_finer_2ND_HF' Ended :::\n\n") if ($TEST_LEVEL>1);

	### RESTART FROM HERE ###


    (print STDOUT "\n<<< Subroutine 'enum_psm_indel_hsts_for_msa_finer' ENDED />>>\n\n") if ($TEST_LEVEL>0);


	my $info_init_psm_hsts = $info_psm_indel_hsts0;
	      ## ($set_cls_gpblocks, $totcts0_indels_for_cls, $psm_hsts0_for_cls) = @{$info_psm_indel_hsts0};

# OUTPUT:  (\@sets_min_totcts_indels_in_indel_blks, \@setssets_psm_hsts_in_indel_blks, \@sets_psm_hsts0_in_indel_blks,
#		\@set_gpatterns, \@set_node2dparsstat, \@set_br2change,
#		\@set_cls_gpblocks, \@info_init_psm_scn, \@set_lists_off_limits, \@set_lists_gblocks, \@set_lists_seqblocks, \@sets_eff_indep_indel_blocks, \@sets_indices_incl_root), where

	return ($sets_min_totcts_indels_in_indel_blks, $setssets_psm_hsts_in_indel_blks, $sets_psm_hsts0_in_indel_blks,
		$set_gpatterns, $set_node2dparsstat, $set_br2change,
		$set_cls_gpblocks, $info_init_psm_hsts, $set_lists_off_limits,
		$set_lists_gblocks, $set_lists_seqblocks, $sets_eff_indep_indel_blocks, $sets_indices_incl_root);

}	# END of the subroutine, 'enum_psm_indel_hsts_for_msa_finer'. #



1;	# a "spell" which makes the module "use"-able and/or "require"-able #


