package MyProbDistri;
#
###################################################################################
#	DENSERM_P: A package to DEtect Negative SElection on Recurrent Mutations
#			(almost exclusively via Perl modules and scripts).
#	Version 0.3: Copyright (C) 2012 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyProbDistri.pm: This module calculates various probability distributions.
#
#	Some subroutines in this module are Perl-translated versions of the programs in:
#	"Numerical Recipes in C: The Art of Scientific Computing", 2nd edition,
#	written by Press WH, Teukolsky SA, Vetterling WT, Flannery BP (Cambridge Univ. Press; Cambridge, UK; 1992).
#	And some probability distributions are described in:
#	"Biometry", 3rd edition,
#	written by Sokal RR, Rohlf FJ (W.H. Freeman and Company; New York, NY, USA; 1995);
#	"Statistical Methods in Bioinformatics: An Introduction",
#	written by Ewens WJ, Grant GR (Springer-Verlag New York Inc.; New York, NY, USA; 2001).  
#

use strict;
use warnings;
use Exporter;
use MySpecFunc qw( gammln gammp gammq erff betai);
use MyBayesian qw( post_bayes_binomial post_bayes_binomial_old );

sub sum_binomial_up_2ends ($$$$);
sub sum_binomial_upwards ($$$);
sub sum_binomial_downwards ($$$);
sub sum_binomial ($$$);

our @ISA = qw(Exporter);
our @EXPORT_OK  = qw( poisson_sgl poisson_cum poisson_cum_slow
		 chi2_upper chi2_lower binomial_cum_2ends binomial_cum
		binomial_cum_via_betai pv_binom_low_dbl_fluc pv_binom_up_dbl_fluc
	pv_binom_low_dbl_fluc_old hypergeom_cum fisher_exact_margin fisher_exact_cells
	cum_gaussian mean_var_mlt_hypergeom nml_apprx_mlt_hypergeom_cum
	pv_f_distri pv_t_distri
	);


# cum_gaussian: This subroutine calculates a culumative probability under a Gausian distribution
#	of the mean $mu and the variance $sigma**2.
#
# INPUT: ($x, $mu, $sigma).
#
# OUTPUT: $p = Prob[ X < $x]_($mu, $sigma**2) = 1/sqrt($pi) * int_{-infty}^{$y}dt exp (-t*t),
#	where $y = ($x - $mu)/(sqrt(2) * $sigma).
#
sub cum_gaussian ($$$) {

	my ($x, $mu, $sigma) = @_;

	unless ($sigma > 0) {
	   print STDERR "ILLEGAL INPUT for 'cum_gaussian': (x, mu, sigma) = ($x, $mu, $sigma).\n";
	   return undef;
	}

	($x == $mu) and (return 1/2); # Just at the mid-point !

	my $y = ($x - $mu) / (sqrt (2) * $sigma); # Normalize the variable

	return (1 + erff ($y) ) /2 ;

}		# end of the subroutine 'gaussian' #


#our @EXPORT_OK  = qw( poisson_sgl poisson_cum poisson_cum_slow chi2_upper chi2_lower binomial_cum sum_binomial_up_2ends sum_binomial_upwards sum_binomial_downwards);	## a temporary revision for testing ##


# poisson_sgl: This subroutine receives the  parameter $lambda and the argument (occurence $n) for a Poisson process,
#		and returns the probability that the event occurs exactly $n times.
#
sub poisson_sgl ($$) {

	my ($lambda, $n) = @_;

	if (($lambda < 0) or ($n < 0)) {

		print STDERR "Cannot calculate the Poisson probability for (lambda,n) = ($lambda, $n) !\n";
		return -1;
	}

	if ($n != int $n) {

		my $tmp = int ($n + 1/2) ;
		print STDERR "I will approximate n= $n  to $tmp in order to calculate the Poisson probability.\n";
		$n = $tmp;
	}

	my $out = 1;

	for (my $i=1; $i <= $n; $i++) { $out *= $lambda/$i; }

	$out *= exp(-$lambda);

	return $out;

}			# end of the subroutine poisson_sgl #


# poisson_cum: This subroutine receives the parameter $lambda and the argument (upper occurence $n) for a Poisson process,
#		and returns the CUMULATIVE probability that the event occurs $n times or less.
#		The subroutine adopts a 1ST ORDER ALGORITHM, therefor it is much faster than "poisson_acm_slow" when handling a large $n.
# 
sub poisson_cum ($$) {

	my ($lambda, $n) = @_;

	if ($lambda < 0) {

		print STDERR "Cannot calculate the Poisson probability for lambda = $lambda !\n";
		return -1;
	}

	if ($n < 0) { return 0; } # the number of events is always positive or zero.

	if ($n != int $n) {

		my $tmp = int ($n + 1/2) ;
		print STDERR "I will approximate n= $n  to $tmp in order to calculate the Poisson probability.\n";
		$n = $tmp;
	}

	my $out = 1;

	for (my $i= $n; $i > 0; $i--) {		# the outer for #

		$out *= $lambda/$i;
		$out++; 
	}					# end of the outer for #

	$out *= exp( -$lambda);
	return $out;

}			# end of the subroutine poisson_cum # 


# poisson_cum_slow: This subroutine receives the parameter $lambda and the argument (upper occurence $n) for a Poisson process,
#		and returns the CUMULATIVE probability that the event occurs $n times or less.
#		The subroutine uses a QUADRATIC ALGORITHM, therefore it processes slowly the case with a large $n.
# 
sub poisson_cum_slow ($$) {

	my ($lambda, $n) = @_;

	if (($lambda < 0) or ($n < 0)) {

		print STDERR "Cannot calculate the Poisson probability for (lambda,n) = ($lambda, $n) !\n";
		return -1;
	}

	if ($n != int $n) {

		my $tmp = int ($n + 1/2) ;
		print STDERR "I will approximate n= $n  to $tmp in order to calculate the Poisson probability.\n";
		$n = $tmp;
	}

	my $out = 1;

	for (my $i=1; $i <= $n; $i++) {		# the outer for #

		my $term = 1;
		for (my $k=1; $k <= $i; $k++) {	$term *= $lambda/$k; }
		$out += $term;
	}					# end of the outer for #

	$out *= exp( -$lambda);
	return $out;

}			# end of the subroutine poisson_cum_slow # 

# chi2_upper (int $nu,float  $x):
#   This subroutine returns the probability Prob(0 <= X < $x),  where the variable X
#    complies with the chi-square distribution of DOF = $nu.
#
#   The subroutine returns undef if some (expected) error occurs.
#
sub chi2_upper ($$) {

	my ($nu, $x) = @_;

	return  gammp ($nu/2, $x/2);

}			# end of the subroutine chi2_upper #


# chi2_lower (int $nu, float $x):
# This subroutine returns the probability Prob( X > $x), where the variable X
# follows the chi-square distribution of DOF = $nu.
#
# It returns undef if some (expected) error occurs.
#
sub chi2_lower ($$) {

	my ($nu, $x) = @_;

	return gammq ($nu/2, $x/2);

}			# end of the subroutine chi2_lower #


# sum_binomial_up_2ends (int $n, int $r1, int $r2, float $p):
# This subroutine returns the probabillity Prob( $r1 <= X <= $r2), where the variable X
# is subject to the binomial distribution consisting of $n Bernoulli trials of success probability $p.
# The subroutine calculates the summation UPWARDS, from $k = $r1 to $k = $r2.
#
sub sum_binomial_up_2ends ($$$$) {

	my ($n, $r1, $r2, $p) = @_;

	if ( ($p < 0) or ($p > 1) ) { return undef; }	# illegal value for $p #
	if ($r1 > $r2) { ($r1, $r2) = ($r2, $r1); }	# rectify the order of bounds #
	if ( ($r2 < 0) or ($r1 > $n) ) { return 0; }	# the region is off the border #

	($r1 < 0) and ($r1 = 0);
	($r2 > $n) and ($r2 = $n);

	my $eps = 1E-12;	# This specifies the precision #

	my $q = 1 - $p;
	my $n1 = $n + 1;
	my $init_exp =  $r1 * log ($p) + ($n - $r1) * log ($q);
	my $r_bound1 =  ( $r1*2 < $n) ? $r1 + 1 : $n - $r1 + 1;
	for (my $i=1; $i< $r_bound1; $i++) { $init_exp += log($n1 - $i) - log($i); }
	($r1 == $r2) and (return exp ($init_exp) );

	my $r_bound2 = $r2 + 1;
	my $sum = my $term = 1;
	for (my $i= $r1 + 1; $i < $r_bound2; $i++) {

		$term *= ($n1 - $i)*$p/$i/$q;
		if ($term >1) {
			$sum /= $term;
			$sum++;
			$init_exp += log ($term);
			$term = 1;
		} else {
			$sum += $term;
		}

		($term < $eps * $sum) and last;		# we have acquired enough precision #
	}

	return exp ($init_exp + log($sum) );

}			# end of the subroutine sum_binomial_up_2ends #

# binomial_cum_2ends (int $n, int $r1, int $r2, float $p):
# This subroutine calculates the probability Prob( $r1 <= X <= $r2), where the variable X
# is subject to the binomial distribution consisting of $n Bernoulli trials with a success probability of $p.
#
sub binomial_cum_2ends ($$$$) {

	my ($n, $r1, $r2, $p) = @_;

	if ( ($p < 0) or ($p > 1) ) {
		print STDERR "Illegal input for \'binomial_cum_2ends\' (Bernoulli probability = $p)\n";
		return undef;
	}

	($r1 > $r2) and (return 0);	# null region #
	if ( ($r2 < 0) or ($r1 > $n) ) { return 0; }	# the region is off the border #
	if ( ($r1 <= 0) and ($r2 >= $n) ) { return 1; }	# the region spans the whole range. #

	if ($p == 0) {
		return ($r1 <= 0) ? 1: 0; 	# all the trials will fail. #
	} elsif ($p == 1) {
		return ($r2 >= $n) ? 1: 0;	# all the trials will succeed. #
	}

	my $m_split = int ($p * ($n + 1) );

	if ($r1 >= $m_split) {
		return sum_binomial_up_2ends ($n, $r1, $r2, $p);
	} elsif ( $r2 <= $m_split) {
		return sum_binomial_up_2ends ($n, $n - $r2, $n - $r1, 1 - $p);
	} else {
		my $s1 = sum_binomial_up_2ends ($n, $r1, $m_split -1, $p);
		my $s2 = sum_binomial_up_2ends ($n, $n - $r2, $n - $m_split, 1 - $p);
		return $s1 + $s2;
	}

}			# end of the subroutine binomial_cum_2ends #

# sum_binomial_upwards (int $n, int $r, float $p):
# This subroutine returns the probabillity Prob( X <= $r), where the variable X
# is subject to the binomial distribution consisting of $n Bernoulli trials of success probability $p.
# The subroutine calculates the summation upwards, from $k = 0 to $k = r.
#
sub sum_binomial_upwards ($$$) {

	my ($n, $r, $p) = @_;

	my $q = 1 - $p;
	my $init_exp = $n * log $q;
	($r == 0) and (return exp ($init_exp) );

	my $sum = my $term = 1;
	my $n1 = $n + 1;
	foreach my $i (1 .. $r) {

		$term *= ($n1 - $i)*$p/$i/$q;
		if ($term >1) {
			$sum /= $term;
			$sum++;
			$init_exp += log ($term);
			$term = 1;
		} else {
			$sum += $term;
		}
	}

	return exp ($init_exp + log($sum) );

}			# end of the subroutine sum_binomial_upwards #


# sum_binomial_downwards (int $n, int $r, float $p):
# This subroutine returns the probabillity Prob(0 <= X <= $r), where the variable X
# is subject to the binomial distribution consisting of $n Bernoulli trials of success probability $p.
# The subroutine calculate the summation downwards, from $k = r to $k = 0.
#
sub sum_binomial_downwards ($$$) {

	my ($n, $r, $p) = @_;

	my $q = 1 - $p;
	my $n1 = $n + 1;
	my $init_exp =  $r * log ($p) + ($n - $r) * log ($q);
	($r == 0) and (return exp ($init_exp) );
	foreach my $i (1 .. $r) { $init_exp += log($n1 - $i) - log($i); }

	my $sum = my $term = 1;
	for (my $k = $r; $k > 0; $k--) {

		$term *= $k*$q/($n1 - $k)/$p;	# theoretically, this should always be less than (or equal to ) 1 #
		$sum += $term;
	}

	return exp ($init_exp + log($sum) );

}			# end of the subroutine sum_binomial_downwards #


# sum_binomial (int $n, int $r, float $p):
# This subroutine returns the probabillity Prob(0 <= X <= $r), where the variable X
# is subject to the binomial distribution consisting of $n Bernoulli trials of success probability $p.
#
# ASSUMPTION: ($r <= $n/2 ) is assumed to hold.
#
sub sum_binomial ($$$) {

	my ($n, $r, $p ) = @_;

	($r < 0) and (return 0);
	($p == 0) and (return 1);
	($p == 1) and (return 0);
	return ($p < 1/2) ? sum_binomial_upwards($n, $r, $p) : sum_binomial_downwards($n, $r, $p) ;

}			# end of the subroutine sum_binomial #


# binomial_cum (int $n, int $r, float $p):
# This subroutine returns the probabillity Prob(0 <= X <= $r), where the variable X
# is subject to the binomial distribution consisting of $n Bernoulli trials of success probability $p.
#
sub binomial_cum ($$$) {

	my ($n, $r, $p) = @_;

	if ( ($r < 0) or ($r > $n) ) {
		print STDERR "Illegal input for \'binomial_cum\': (n, r) = ($n, $r).\n"; 
		return undef;
	} elsif ( ($p < 0) or ($p > 1) ) {
		print STDERR "Illegal input for \`binomial_cum\': p = $p.\n";
		return undef;
	}

	$n = int $n;
	$r = int $r;

	if ($r * 2 > $n) {
		my $tmp = sum_binomial($n, $n - $r -1, 1- $p);
		return 1 - $tmp;
	} else {
		return sum_binomial($n, $r, $p);
	}

}			# end of the subroutine binomial_cum #

# binomial_cum_via_betai (int $n, int $r, float $p):
#
# This subroutine returns the probability Prob(0 <= X <= $r) under the
# binomial distribution with $n samples and with the 'success' probability $p,
# by using the incomplete beta function.
#
sub binomial_cum_via_betai ($$$) {

	my ($n, $r, $p) = @_;

	if ( ($r < 0) or ($r > $n) ) {
	    print STDERR "Illegal input for \'binomial_cum_via_betai\': (n, r) = ($n, $r).\n"; 
	    return undef;
	} elsif ( ($p < 0) or ($p > 1) ) {
	    print STDERR "Illegal input for \'binomial_cum_via_betai\': p = $p.\n";
	    return undef;
	}

	($r == $n) and (return 1); 

	 return betai ($n - $r, $r + 1, 1 - $p);

}			# end of the subroutine binomial_cum_via_betai #


# pv_binom_low_dbl_fluc: This subroutine estimates the LOWER P-value for the unlikeliness
#		of observing $r2 positives out of $n2 samples,
#		given that you have already observed $r1 positives out of $n1 samples
#		taken from the same populatoin.
#		The subroutine takes account of the fluctuation in the Bernoulli
#		probability $p by using a Bayesian method ('pv_binom_dbl_fluc'),
#		and then evaluates Prob (0 <= X <= $r2)
#		under bionomial distributions each consisting of
#		$n2 independent Bernoulli trials of probability $p (variable).
#
# ASSUMPTION: The prior probability of the Bernoulli probability $p is (unnormalized):
#		Prob( $p) = 1   for $p <= $p_thrsh,
#		  = (1 - $p) ** $n1 / (1 - $p_thrsh) ** $n1   for $p > $p_thrsh .
#
sub pv_binom_low_dbl_fluc ($$$$$$) {

	my ($n1,$r1,$n2,$r2, $p_thrsh, $error) = @_;

	if ($n2 <= 0) {
	    print STDERR "Illegal input for 'pv_binom_dbl_fluc': (n2, r2) = ($n2, $r2) .\n";
	    return undef;
	}

	($r2  < 0) and (return 0 );  # 0 <= X <= $r2 canNOT happen.
	($r2 >= $n2) and (return 1);   # 0 <=  X <= $r2 always holds.

	(my $distri_p, undef) = post_bayes_binomial ($n1, $r1, $p_thrsh, $error);

	unless (defined $distri_p) {
	    print STDERR "Illegal input for 'pv_binom_dbl_fluc': (n1, r1) = ($n1, $r1) .\n";
	    return undef;
	}

	my $pvalue = 0;

	for (my $i=0; $i < @{$distri_p}; $i++) {

		my ($p_av, $post_prob) = @{$distri_p->[$i]};
		my $bin_prob = binomial_cum_via_betai ($n2, $r2, $p_av);
	        $pvalue += $post_prob * $bin_prob;
	}

	return $pvalue;

}		# end of the subroutine pv_binom_low_dbl_fluc


# pv_binom_up_dbl_fluc: This subroutine estimates the UPPER  P-value for the unlikeliness
#		of observing $r2 positives out of $n2 samples,
#		given that you have already observed $r1 positives out of $n1 samples
#		taken from the same populatoin.
#		The subroutine takes account of the fluctuation in the Bernoulli
#		probability $p by using a Bayesian method ('pv_binom_dbl_fluc'),
#		and then evaluates Prob ($r2 <= X <= $n2)
#		under bionomial distributions each consisting of
#		$n2 independent Bernoulli trials of probability $p (variable).
#
# ASSUMPTION: The prior probability of the Bernoulli probability $p is (unnormalized):
#		Prob( $p) = 1   for $p <= $p_thrsh,
#		  = (1 - $p) ** $n1 / (1 - $p_thrsh) ** $n1   for $p > $p_thrsh .
#
sub pv_binom_up_dbl_fluc ($$$$$$) {

	my ($n1,$r1,$n2,$r2, $p_thrsh, $error) = @_;

	if ( $n2 <= 0) {
	    print STDERR "Illegal input for 'pv_binom_dbl_fluc': n2 = $n2 .\n";
	    return undef;
	}

	($n2 < $r2) and (return 0);    # $r2 <= X <= $n can never happen.
	($r2 <=  0) and (return 1);	# The above inequalities always hold.

	(my $distri_p, undef) = post_bayes_binomial ($n1, $r1, $p_thrsh, $error);

	unless (defined $distri_p) {
	    print STDERR "Illegal input for 'pv_binom_dbl_fluc': (n1, r1) = ($n1, $r1) .\n";
	    return undef;
	}

	my $pvalue = 0;

	for (my $i=0; $i < @{$distri_p}; $i++) {

	    my ($p_av, $post_prob) = @{$distri_p->[$i]};
	    if ($r2 ==0) {
		$pvalue += $post_prob; # $bin_prob should  always be 1
	    } else {
		my $cmpl_bin_prob = binomial_cum_via_betai ($n2, $r2 -1, $p_av);
	        $pvalue += $post_prob * (1 - $cmpl_bin_prob);
	    }
	}

	return $pvalue;

}		# end of the subroutine pv_binom_up_dbl_fluc


#=an old version --- primarily for the debugging purpose

# pv_binom_low_dbl_fluc_old: This subroutine estimates the LOWER P-value for the unlikeliness
#		of observing $r2 positives out of $n2 samples,
#		given that you have already observed $r1 positives out of $n1 samples
#		taken from the same populatoin.
#		The subroutine takes account of the fluctuation in the Bernoulli
#		probability $p by using a Bayesian method ('pv_binom_dbl_fluc'),
#		and then evaluates Prob (0 <= X <= $r2)
#		under bionomial distributions each consisting of
#		$n2 independent Bernoulli trials of probability $p (variable).
#
# ASSUMPTION: The prior probability of the Bernoulli probability $p is (unnormalized):
#		Prob( $p) = 1   for $p <= $p_thrsh,
#		  = (1 - $p) ** $n1 / (1 - $p_thrsh) ** $n1   for $p > $p_thrsh .
#
sub pv_binom_low_dbl_fluc_old ($$$$$$) {

	my ($n1,$r1,$n2,$r2, $p_thrsh, $error) = @_;

	if ($n2 <= 0) {
	    print STDERR "Illegal input for 'pv_binom_dbl_fluc': (n2, r2) = ($n2, $r2) .\n";
	    return undef;
	}

	($r2  < 0) and (return 0 );  # 0 <= X <= $r2 canNOT happen.
	($r2 >= $n2) and (return 1);   # 0 <=  X <= $r2 always holds.

	(my $distri_p, undef) = post_bayes_binomial_old ($n1, $r1, $p_thrsh, $error);

	unless (defined $distri_p) {
	    print STDERR "Illegal input for 'pv_binom_dbl_fluc': (n1, r1) = ($n1, $r1) .\n";
	    return undef;
	}

	my $pvalue = 0;

	for (my $i=0; $i < @{$distri_p}; $i++) {

		my ($p_av, $post_prob) = @{$distri_p->[$i]};
		my $bin_prob = binomial_cum_via_betai ($n2, $r2, $p_av);
	        $pvalue += $post_prob * $bin_prob;
	}

	return $pvalue;

}		# end of the subroutine pv_binom_low_dbl_fluc_old

#=cut


# hypergeom_cum:
#	This subroutine first calculates the probability:
#	P_bd = Prob (X =x; Ntot, Ncol, Nrow) = C(Ncol, x) * C(Ntot-Ncol, Nrow-x)/ C(Ntot, Nrow)
#	under the hypergeometric probability distribution.
#	Then it calculates lower-tail and upper-tail cumulative probabilities
#	(EXCLUDING the boundary X=x !!), P_low = Prob (X < x) and P_up = Prob (X > x),
#	based on P_bd.
#
# INPUT: ($ntot, $ncol, $nrow, $x)
#
# OUTPUT: (P_low, P_bd, P_up).
#
sub hypergeom_cum ($$$$) {

	my ($ntot, $ncol, $nrow, $x)  = @_;

	unless ( ($ntot >= $ncol) and ($ntot >= $nrow) and ($ncol >= $x) and ($nrow >= $x) and ($x >=0) ) {
	    print STDERR "ILLEGAL input for 'hypergeom_cum': (Ntot, Ncol, Nrow, x)  = ($ntot, $ncol, $nrow, $x).\n";
	    return (undef, undef, undef);
	}

	($ncol > $nrow) and ( ($ncol, $nrow) = ($nrow, $ncol) );

	my $cncol = $ntot - $ncol;
	my $cnrow = $ntot - $nrow;
	my $ndiag = $ntot - $ncol - $nrow;

	if ($ndiag + $x < 0) {
	    printf STDERR "ILLEGAL input for 'hypergeom_cum': Ntot - Ncol - Nrow + x = %d.\n", $ndiag + $x;
	    return (undef, undef, undef);
	}

	my $p_bd = my $term 
	 = exp ( gammln ($ncol+1) + gammln ($cncol+1) + gammln ($nrow+1) + gammln ($cnrow+1)
	    - gammln ($ntot+1) - gammln ($x+1) - gammln ($ncol-$x+1) - gammln ($nrow-$x+1)
		- gammln ($ndiag + $x + 1) );  

	my $p_up = my $p_low = 0;

	if (2*$x > $ncol + 1) {
	    for (my $k=$x+1; $k<= $ncol; $k++) {
		$term *= ($ncol+1 -$k) * ($nrow +1  -$k) /$k /($ndiag + $k);
		$p_up += $term;
	    }
	    $p_low = 1 - $p_bd - $p_up;

	} else {
	    my $lb = ($ndiag >= 0) ? 0 : - $ndiag; 
	    for (my $k=$x-1; $k>= $lb; $k--) {
		$term *= ($k+1) * ($ndiag +$k+1) /($ncol-$k) /($nrow-$k) ;
		$p_low += $term;
	    }
	    $p_up = 1 - $p_low - $p_bd;
	}

	return ($p_low, $p_bd, $p_up );

}		# end of the subroutine hypergeom_cum #


# fisher_exact_cells:
#	This subroutine receives the values in the cells of a 2x2 contingency table,
#	and calculates P-values for the Fisher exact test.
#	It takes advantage of the symmetry that this statistical test is blessed with.
#
# INPUT: ($n11, $n12, $n21, $n22).
#
# OUTPUT: ($p_low = Prob (X <= $n11), $p_up = Prob (X >= $n11), $p_bd = Prob (X = $n11) ). 
#
sub fisher_exact_cells ($$$$) {

	my ($n11, $n12, $n21, $n22) = @_;

	unless ( ($n11 >= 0) and ($n12 >=0) and ($n21 >=0) and ($n22 >=0)) {
	    print STDERR "ILLEGAL INPUT for 'fisher_exact_cells': (n11, n12, n21, n22) = ($n11, $n12, $n21, $n22).\n";
	    return (undef, undef, undef);
	}


	($n11 > $n22) and ( ($n11, $n22) = ($n22, $n11) ); # always make $n11 <= $n22.
	($n12 > $n21) and ( ($n12, $n21) = ($n21, $n12) ) ; # always make $n12 <= $n21.

	my $ncol = $n11 + $n21;
	my $cncol = $n12 + $n22;
	my $nrow = $n11 + $n12;
	my $cnrow = $n21 + $n22;
	my $ntot = $ncol + $cncol;

	my ($p_low, $p_up, $p_bd);
	if ($n11 <= $n12) { # use $n11 as the core variable #

	    ($p_low, $p_bd, $p_up) = hypergeom_cum ($ntot, $ncol, $nrow, $n11);

	} else { # use $n12 instead, then swap $p_up with $p_low #

	    ($p_up, $p_bd, $p_low) = hypergeom_cum ($ntot, $cncol, $nrow, $n12);
	}

	unless (defined $p_up) {
	    print STDERR "'hypergeom_cum' FAILED as a subroutine of 'fisher_exact_cells' for the input: (n11, n12, n21, n22) = ($n11, $n12, $n21, $n22).\n";
	    return (undef, undef, undef);
	}

		# add boundary values #
	$p_low += $p_bd;
	$p_up += $p_bd;

	return ($p_low, $p_up, $p_bd);

}		# end of the subroutine fisher_exact_cells #


# fisher_exact_margin:
#	This subroutine conducts the Fisher exact test using
#	marginal values of a 2x2 contingency table.
#
# INPUT: ($ntot, $ncol, $nrow, $x).
#
# OUTPUT: ($p_low = Prob (X <= $x), $p_up = Prob (X >= $x), $p_bd = Prob (X = $x) ).
#
sub fisher_exact_margin ($$$$) {

	my ($ntot, $ncol, $nrow, $x) = @_;

	unless ( ($ntot >= $ncol) and ($ntot >= $nrow) and ($ncol >= $x) and ($nrow >= $x) and ($x >=0) ) {
	    print STDERR "ILLEGAL INPUT for 'fisher_exact_margin': (Ntot, Ncol, Nrow, x) = ($ntot, $ncol, $nrow, $x).\n";
	    return (undef, undef, undef);
	}

	my $n11 = $x;
	my $n12 = $nrow - $n11;
	my $n21 = $ncol - $n11;
	my $n22 = $ntot - $n11 - $n12 - $n21;

	if ($n22 < 0) { 
	    print STDERR "NEGATIVE CELL-VALUE: n22 = $n22 (Ntot=$ntot, Ncol=$ncol, Nrow=$nrow, x=$x).\n";
	    return (undef, undef, undef);
	}

	my ($p_low, $p_up, $p_bd) = fisher_exact_cells ($n11, $n12, $n21, $n22);

	unless (defined $p_low) {
	    print STDERR "'fisher_exact_cells' FAILED as a subroutine of 'fisher_exact_margin'.\n";
	    return (undef, undef, undef);
	}

	return  ($p_low, $p_up, $p_bd);

}		# end of the subroutine fisher_exact_margin #


# mean_var_mlt_hypergeom:
#	This subroutine calculates the mean and variance of a multiple hypergeometric distribution,
#	or more precisely, the distribution of the sum, m = Sum_{i=1,..,C} m_i,
#	of the number m_i of relevant elements sampled from C independent categories.
#	Here each category is assumed to follow a hypergeometric distribution:
#	P (m_i | N_i, M_i, n_i) = C(M_i, m_i) * C(N_i - M_i, n_i - m_i) / C(N_i, n_i),
#	where N_i and M_i are total number of elements and the number of relevant elements in the i-th category, respectively,
#	n_i is the number of samples out of the i-th category,
#	and C (N, M) = N!/(M!(N-M)!) is the number of possible ways that M elements are taken out of the N elements. 
#	Then the mean and the variance are given by the following formulas:
#	Mean: E(m) = Sum_{i=1,...,C} n_i * M_i /N_i ;
#	Variance: V(m) = Sum_{i=1,...,C} n_i * M_i / N_i * (1 - M_i/N_i) * (N_i - n_i)/(N_i - 1) . 
#
# INPUT: \@tbl_catewise_nums, where
#	@{$tbl_catewise_nums[$i-1]} = ($N_i, $M_i, $n_i, $m_i), with
#	$N_i is the total number of elements in the $i th category,
#	$M_i is the number of relevant elements in the $i th category,
#	$n_i is the number of sampled elements out of the $i th category,
#	$m_i is the number of relevant elements sampled out of the $i th category.
#
# OUTPUT: ($mean, $var, $tot_smpls, $tot_rlv_smpls), where
#	$mean and var are the mean and the variance, respectively, given by the above formula;
#	$tot_smpls is the total number, Sum_{i=1,...,C} n_i, of the sampled elements,
#	and $tot_rlv_smpls is the total number, Sum_{i=1,...,C} m_i, of the sampled relevant elements.
#
sub mean_var_mlt_hypergeom (\@) {

	my ($tbl_catewise_nums) = @_;

	my $ct_cate = @{$tbl_catewise_nums};

	my $mean = my $var = my $tot_smpls = my $tot_rlv_smpls = 0;

	for (my $i=0; $i<$ct_cate; $i++) { # The main foreach-loop #
	    my $cate = $tbl_catewise_nums->[$i];

	    ((defined $cate) and (@{$cate}>3)) or next;

	    my ($Ni, $Mi, $ni, $mi) =  @{$cate};

	    (($Ni>0) and ($ni>0)) or next; # Skip category not contributing to the output
	    ($Mi>0) or next;
	    if (($Mi>$Ni) or ($ni>$Ni) or ($mi>$Mi) or ($mi>$ni) or ($mi<0)) {
		print STDERR "[CAUTION 'mean_var_mlt_hypergeom'] The ", $i+1, " th category contains irregal data: (";
		print STDERR  join (', ', @{$cate}), ") !\n\n";
		next;
	    }

	    my $cate_mean = $ni * $Mi/$Ni;
	    $mean += $cate_mean;
	    $var += $cate_mean * (1 - $Mi/$Ni) * ($Ni - $ni)/($Ni - 1);
	    $tot_smpls += $ni;
	    $tot_rlv_smpls += $mi;
	}

	return ($mean, $var, $tot_smpls, $tot_rlv_smpls);

}		# End of the subroutine 'mean_var_mlt_hypergeom' #


# nml_apprx_mlt_hypergeom_cum:
#	This subroutine calculates an approximate cumulative probability Prob [ m = Sum_{i=1,...,C} m_i <= X ]
#	of the multiple hypergeometric distribution:
#	P ({m_i} | {N_i}, {M_i}, {n_i}) = Prod_{i=1,...,C} P(m_i | N_i, M_i, n_i),
#	where P (m_i | N_i, M_i, n_i) = C(M_i, m_i) * C(N_i - M_i, n_i - m_i) / C(N_i, n_i)
#	is the 'hypergeometric' probability of obtaining m_i relevant elements out of a sample of size n_i,
#	taken withOUT replacement out of the mother population of size N_i containing M_i relevant elements.
#	N_i and M_i are the numbers of all elements and relevant elements, respectively, in the i-th category,
#	n_i is the number of samples out of the i-th category,
#	and C (N, M) = N!/(M!(N-M)!) is the number of possible ways that M elements are taken out of the N elements. 
#	Here the cumulative probability Prob [ Sum_Pi=1,...,C} m_i <= X ] is approximated with the normal distribution
#	of mean E(m) and variance V(m), where
#	E(m) = Sum_{i=1,...,C} n_i * M_i /N_i , and
#	V(m) = Sum_{i=1,...,C} n_i * M_i / N_i * (1 - M_i/N_i) * (N_i - n_i)/(N_i - 1) . 
#
# INPUT: \@tbl_catewise_nums, where
#	@{$tbl_catewise_nums[$i-1]} = ($N_i, $M_i, $n_i, $m_i), with
#	$N_i is the total number of elements in the $i th category,
#	$M_i is the number of relevant elements in the $i th category,
#	$n_i is the number of sampled elements out of the $i th category,
#	$m_i is the number of relevant elements sampled out of the $i th category.
#
# OUTPUT: ($p_up, $p_dwn, $z, $mean, $var, $tot_smpls, $tot_rlv_smpls),  where
#	$mean and $var are the mean and the variance of the hypergeometric distribution,
#	$tot_smpls is the total number of sampled elements,
#	and $tot_rlv_smpls is the number of relevant samples;
#	$z = ($tot_rlv_smpls - $mean)/ sqrt ($var) is the Z-value;
#	$p_up = Prob (X >= $tot_rlv_smpls  -1/2 | Norm($mean, sqrt($var) ) if $tot_rlv_smpls>0,
#	      = 1 if $tot_rlv_smpls == 0,
#	and
#	$p_dwn = Prob (X <= $tot_rlv_smpls +1/2 | Norm($mean, sqrt($var) ) ) if $tot_rlv_smpls<$tot_smpls,
#	       = 1 if $tot_rlv_smpls == $tot_smpls,
#	are the upper-tailed and lower-tailed p-values, respectively, approximated by the normal distribution.
#
sub nml_apprx_mlt_hypergeom_cum (\@) {

	my ($tbl_catewise_nums) = @_;

	unless (@{$tbl_catewise_nums}>0) {
	    print STDERR "[CAUTION 'nml_apprx_mlt_hypergeom_cum'] The input data contains NO CATEGORY !\n";
	    return (undef, undef, undef, 0, 0, 0, 0) ;
	}

	my ($mean, $var, $tot_smpls, $tot_rlv_smpls) = mean_var_mlt_hypergeom (@{$tbl_catewise_nums}) ;

	unless ($var> 0) {
	    print STDERR  "[CAUTION 'nml_apprx_mlt_hypergeom_cum'] Because 'mean_var_mlt_hypergeom' returned ZERO for the variance, we cannot go further ...\n";
	    return (undef, undef, undef, $mean, $var, $tot_smpls, $tot_rlv_smpls );
	}

	my $sigma = sqrt ($var) ;
	my $z = ($tot_rlv_smpls - $mean ) / $sigma ;

	my $p_dwn = ($tot_rlv_smpls == $tot_smpls) ? 1 :
		cum_gaussian ($tot_rlv_smpls+1/2, $mean, $sigma) ;

	my $cp_up = ($tot_rlv_smpls == 0) ? 0 :
		cum_gaussian ($tot_rlv_smpls-1/2, $mean, $sigma) ;
	my $p_up = 1 - $cp_up;

	return ($p_up, $p_dwn, $z, $mean, $var, $tot_smpls, $tot_rlv_smpls) ;

}		# End of the subroutine 'nml_apprx_mlt_hypergeom_cum' #


# pv_f_distri:
#	This subroutine calculates a P-value of the F-distribution of the degrees of freedoms $r1, $r2:
#	$p_value = P [F > $f]_($r1, $r2), where F =: (X1/$r1) / (X2/$r2) with X1 and X2 being
#	chi-square variates of the degrees of freedom $r1 and $r2, respectively.
#	This subroutine makes full use of the formula:
#	P [F > $f]_($r1, $r2) = I_x ($r2/2, $r1/2),
#	where I_x (a, b) = 1/B(a,b) * int_0^x dt t**(a-1) * (1-t)**(b-1) is
#	an incomplete beta function, and with x = 1/(1+$r1/$r2 * $f) in this case.
#
# INPUT: ($f, $r1, $r2).
#
# OUTPUT: $p_value,
#	  or undef if there is an error.
#
sub pv_f_distri ($$$) {

	my ($f, $r1, $r2) = @_;

	if (($f <0) or ($r1 <= 0) or ($r2 <= 0)) {
	    print STDERR "[ERROR 'pv_f_distri'] IRREGAL INPUT: (f,r1,r2) = ($f, $r1, $r2) .\n";
	    return undef;
	}

	($f == 0) and (return 1); # Trivial case #

	my $a = $r2/2;
	my $b = $r1/2;
	my $x = 1/(1 + $r1 * $f /$r2);

	my $p_value = betai ($a, $b, $x);

	unless (defined $p_value) {
	    print STDERR "[ERROR 'pv_f_distri'] FAILED to conduct the subroutine 'betai' for the F-distribution parameters: (f, r1, r2) = ($f, $r1, $r2) !\n";
	    return undef;
	}

	return $p_value;

}		# End of the subroutine 'pv_f_distri' #


# pv_t_distri:
#	This subroutine calculates a (2-tailed) P-value of the T-distribution of degrees of freedom $n:
#	$p_value = P[T<-$t, $t<T]_$n, where T =: X1 / sqrt(X2/$n), with
#	X1 being a normal variate of mean 0 and stadard deviation 1,
#	and X2 being a chi-square variate of the degree of freedoms $n.
#	This subroutine makes full use of the formula:
#	P[T<-$t, $t<T]_$n = I_x ($n/2, 1/2),
#	where I_x (a, b) = 1/B(a,b) * int_0^x dt t**(a-1) * (1-t)**(b-1) is
#	an incomplete beta functionn and with x = 1/(1 + $t*$t/$n) in this case.
#
# INPUT: ($t, $n).
#
# OUTPUT: $p_value,
#	  or undef if there is an error.
#
sub pv_t_distri ($$) {

	my ($t, $n) = @_;

	if (($t<0) or ($n <=0)) {
	    print STDERR "[ERROR 'pv_t_distri'] IRREGAL INPUT: (t, n) = ($t, $n) .\n";
	    return undef;
	}

	my $a = $n/2;
	my $b = 1/2;
	my $x = 1/ (1 + $t*$t/$n);

	my $p_value = betai ($a, $b, $x);

	unless (defined $p_value) {
	    print STDERR "[ERROR 'pv_t_distri'] FAILED to conduct the subroutine 'betai' for the T-distribution parameters: (t, n) = ($t, $n) !\n";
	    return undef;
	}

	return $p_value;

}		# End of the subroutine 'pv_t_distri' #


1;	# a "spell" necessary  to make the module "use"-able and/or "require"-able #

