package MyTaxonomy;
#
###################################################################################
#	DENSERM_P: A package to DEtect Negative SElection on Recurrent Mutations
#			(almost exclusively via Perl modules and scripts).
#	Version 0.3: Copyright (C) 2012 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyTaxonomy.pm: This Module consists of subroutines which processes NCBI taxonomy tables.
#		It also contains a subroutine to process the table of 
#		correspondence between sequence IDs and taxonomy IDs (or names).
#

use strict;
use warnings;
use Exporter;
use MyArrayManipu;

use MyFile_manipu qw( open_input_file );
use MyArrayManipu qw(nr_array);
use MyArrayManipu_num qw( diff_nmarray union_nmarray nr_nmarray );

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
		find_shared_lineages interpolate_shared_lineage
		fetch_name2id fetch_id2names fetch_id2classed_names
		fetch_taxonomy_tree_bottomup fetch_taxonomy_tree_topdown fetch_taxon_ranks
		if_ancestor fetch_ancestry calculate_abs_depth
		fetch_upper_ranks fetch_best_taxon_name
		batch_fetch_best_taxon_names fetch_name_upper_ranks
		fetch_all_offsprings fetch_ext_offsprings
		fetch_ext_offsprings_nr fetch_all_offsprings_nr
		fetch_last_common_ancestor
		fetch_last_common_ancestor_ov2 fetch_common_ancestors
		fetch_intervene_nodes
		sibling_subfamilies
		fetch_seqid2taxonid
		name2id_new_lineages revise_taxonomy_tree
		make_nmhierarchy);

# find_shared_lineages:
#	This subroutine finds ancestral lineages shared by
#	multiple lineages.
#
# INPUT: \@list_lineages, where
#	@{$list_lineages[m-1]} = (top_node (ID or name), ..., bottom node (ID or name)) for the m-th lineage.
#
# OUTPUT: (\@shared_lineage, \@non_shared), where
#	@shared_lineage = (top_node, ..., last common ancestor (LCA)),
#	@{$non_shared[m-1]} = (node just below the LCA, ..., bottom node) for the m-th lineage.
#
sub find_shared_lineages (\@) {

	my ($list_lins) = @_;

	my $ct_lins = @{$list_lins};
	($ct_lins > 0) or (return ($list_lins, []) );
	($ct_lins > 1) or (return ($list_lins, [[]]) );

	my $min_depth = @{$list_lins->[0]};
	for (my $m=1; $m<$ct_lins; $m++) {
	    my $depth = @{$list_lins->[$m]};
	    ($depth < $min_depth) and ($min_depth = $depth);
	}

	my $k;
	OUTER:
	for ($k=0; $k<$min_depth; $k++) { # the outer for-loop #
	    my $ref = $list_lins->[0][$k]; # reference node
	    for (my $m=1; $m<$ct_lins; $m++) { # the inner for-loop #
		($ref eq $list_lins->[$m][$k])
			or (last OUTER); # found some discrepancy at the depth $k #
	    }			# end of the inner for-loop #
	}			# end of the outer for-loop #

	my @shared = ($k>0) ? @{$list_lins->[0]}[0 .. $k-1] : ();
	my @non_shared = ();

	for (my $m=0; $m<$ct_lins; $m++) {
	    my $lineage = $list_lins->[$m];
	    my @tmp = @{$lineage}[$k .. $#{$lineage}];
	    push @non_shared, \@tmp;
	}

	return (\@shared, \@non_shared);
	
}		# end of the subroutine find_shared_lineages #


# interpolate_shared_lineage:
#	This subroutine receives a reference lineage and a subject lineage,
#	and finds the shared lineage in terms of the reference lineage.
#
# INPUT: (\@ref_lineage, \@sb_lineage),
#	
# OUTPUT: (\@shared, \@ext_ref, \@ext_sb).
#	where \@ext_xxx contains ONLY DOWNSTREAM taxa.
#
sub interpolate_shared_lineage (\@\@) {

	my ($ref_lineage, $sb_lineage) = @_;

	my %indx_ref; # (taxon  name in reference => index in the array, ...)
	for (my $k=0; $k<@{$ref_lineage}; $k++) {
	    $indx_ref{$ref_lineage->[$k]} = $k;
	}

	my $maxdepth_ref = -1;
	my $maxdepth_sb = -1;
	for (my $k=0; $k<@{$sb_lineage}; $k++) {
	    my $taxon  = $sb_lineage->[$k];
	    my $indx = $indx_ref{$taxon};
	    ((defined $indx) and ($indx > $maxdepth_ref) ) or next;
	    $maxdepth_ref = $indx; 
	    $maxdepth_sb = $k;
	}

	if ($maxdepth_ref < 0) {

	    print STDERR "NO SHARED lineage found by 'interpolate_shared_lineage'. \n";
	    return ([], $ref_lineage, $sb_lineage);
	}

	my @shared = @{$ref_lineage}[0 .. $maxdepth_ref];
	my @ext_ref = @{$ref_lineage}[$maxdepth_ref +1 .. $#{$ref_lineage}];
	my @ext_sb = @{$sb_lineage}[$maxdepth_sb + 1 .. $#{$sb_lineage}];

	return (\@shared, \@ext_ref, \@ext_sb);

}		# end of the subroutine 'interpolate_shared_lineage' #


# fetch_name2id: This subroutine reads in a file 'names.dmp' or 'names_XXX.tbl.txt'
#		and constructs a hash table:
#		(taxon_name => (last) taxon_ID, ...),
#		as well as another hash table:
#		(taxon_name => [a list of taxon_IDs], ...)
#		which is designated for non-unique taxon names.
#
sub fetch_name2id (\$\%) {

	my ($infile, $name2id) = @_;
	local $/ = "\n";
	my $fh = open_input_file (${$infile}) ;

	my %nonunique_names;

	while (my $in = <$fh> )  {

	    ($in =~ s/\t\|\n$//) or next; # skip lines not abide by the stated format
	    my @fields = split /\t\|\t/, $in;
	    (@fields < 2) and next;
	    my ($taxid, $name_txt, $unique_name) = @fields;


	    if ( (defined $unique_name) and  ($unique_name =~ m/\S/ ) ) {

		$name2id->{$unique_name} = $taxid;
		if (defined $nonunique_names{$name_txt}) {
		    push @{$nonunique_names{$name_txt}}, $taxid;
		} else {
		    $nonunique_names{$name_txt} = [$taxid];
		}
	    } else {
	       $name2id->{$name_txt} = $taxid;
	    }
	}

	close ($fh);

	return \%nonunique_names;

}		# end of the subroutine fetch_name2id #


# fetch_id2names: This subroutine reads in a file 'names.dmp' or 'names_XXX.tbl.txt'
#		and constructs a hash table:
#		(taxon_ID => [a list of taxon names], ...).
#
sub fetch_id2names (\$\%) {

	my ($infile, $id2names) = @_;
	local $/ = "\n";
	my $fh = open_input_file (${$infile});
	my %tmp; # a temporary hash of the same format as %{$id2names}
	my %nonunique_names;

		# fetch taxon names irrespective of redundancy #
	while (my $in = <$fh> ) {

		($in =~ s/\t\|\n$//) or next; # skip lines not abide by the stated format
		my @fields = split /\t\|\t/, $in;
		(@fields < 2) and next;
		my ($taxid, $name_txt, $unique_name) = @fields;

		my @names = ( (defined $unique_name) and ($unique_name =~ m/\S/) ) ?
			($unique_name, $name_txt) : ($name_txt) ;

		if (defined $tmp{$taxid}) {
		    push @{$tmp{$taxid}} , $names[0];
		} else {
		    $tmp{$taxid} = [$names[0] ];
		}

		if (@names > 1) {
		    if (defined $nonunique_names{$taxid}) {
			push @{$nonunique_names{$taxid}}, $names[1];
		    } else {
			$nonunique_names{$taxid} = [ $names[1] ];
		    }
		}
	}

	close ($fh);

		# remove redundancy #
	foreach my $taxid (keys %tmp) {
		my $nr_names = nr_array (@{$tmp{$taxid}});
		$id2names->{$taxid} = $nr_names;
	}

	return \%nonunique_names;

}			# end of the subroutine fetch_id2names #


# fetch_id2classed_names: This subroutine reads in a file 'names.dmp' or 'names_XXX.tbl.txt'
#		and constructs a hash table:
#		(taxon_ID => \%class2names,...), where
#		%class2names = (name_class => [a set of names], ...) for a particular taxon_ID.
#
sub fetch_id2classed_names (\$\%) {

	my ($infile, $id2cl_names) = @_;
	local $/ = "\n";
	my $fh = open_input_file (${$infile});
	my %common_names;

		# fetch taxon names irrespective of redundancy #
	while (my $in = <$fh> ) {

	    ($in =~ s/\t\|\n$//) or next; # skip lines not abide by the stated format
	    my @fields = split /\t\|\t/, $in;
	    (@fields < 4) and next;

	    my ($taxid, $name_txt, $name_class) = @fields[0,1,3];

	    if (defined $id2cl_names->{$taxid} ) {
		if (defined $id2cl_names->{$taxid}{$name_class}) {
			push @{$id2cl_names->{$taxid}{$name_class}}, $name_txt;
		} else {
			$id2cl_names->{$taxid}{$name_class} =  [ $name_txt ];
		}

	    } else {
		$id2cl_names->{$taxid} = {$name_class => [ $name_txt] };
	    }

	}

	close ($fh);

	return \%common_names;

}			# end of the subroutine fetch_id2classed_names #


# fetch_taxonomy_tree_bottomup: This subroutine reads in  a table 'nodes.dmp' or a similar table
#		and constructs a hash table:
#		(tax_id => parent tax_id, ...).
#
sub fetch_taxonomy_tree_bottomup (\$\%) {

	my ($infile, $node2parent) = @_;
	local $/ = "\n";
	my $fh = open_input_file (${$infile}) ;

	while (my $in = <$fh>) {

		($in =~ s/\t\|\n$//) or next;  # skip illegal inputs

		my @fields = split /\t\|\t/, $in;
		(@fields < 2) and next;

		my ($node, $parent) = @fields[0,1];
			# skip entries with empty fields #
		if ( ($node eq '')  or ($parent eq '') ) {
			print STDERR "Found an entry with (an) empty field(s): (tax_id, parent tax_id) = ($node, $parent) .\n";
			next;
		}

		($node eq $parent)  and next;	# This should be the top node

		$node2parent->{$node} = $parent;
	}

	close ($fh);

	return 1;

}		# end of the subroutine fetch_taxonomy_tree_bottomup


# fetch_taxonomy_tree_topdown: This subroutine reads in a table 'nodes.dmp' or a similar table
#		and constructs a hash table:
#		(parent tax_id => [a list of child tax_id's ], ...)
#
sub fetch_taxonomy_tree_topdown (\$\%) {


	my ($infile, $node2children) = @_;
	local $/ = "\n";
	my $fh = open_input_file (${$infile}) ;

	while (my $in = <$fh>) {

		($in =~ s/\t\|\n$//) or next;  # skip illegal inputs

		my @fields = split /\t\|\t/, $in;
		(@fields < 2) and next;

		my ($node, $parent) = @fields[0,1];
			# skip entries with empty fields #
		if ( ($node eq '')  or ($parent eq '') ) {
			print STDERR "Found an entry with (an) empty field(s): (tax_id, parent tax_id) = ($node, $parent) .\n";
			next;
		}

		($node eq $parent)  and next;	# This should be the top node

		if (defined $node2children->{$parent}) {
			push @{$node2children->{$parent}}, $node;
		} else {
			$node2children->{$parent} = [$node];
		}
	}

	close ($fh);

	return 1;

}			# end of the subroutine fetch_taxonomy_tree_topdown #


# fetch_taxon_ranks: This subroutine reads in a table 'nodes.dmp' or a similar table
#		and constructs a hash table:
#		(taxon_id => its rank, ...).
#		For the moment, it dismisses taxon_ID's  with 'no rank'.
#
sub fetch_taxon_ranks (\$\%) {


	my ($infile, $txid2rank) = @_;
	local $/ = "\n";
	my $fh = open_input_file (${$infile}) ;

	while (my $in = <$fh>) {

		($in =~ s/\t\|\n$//) or next;  # skip illegal inputs

		my @fields = split /\t\|\t/, $in;
		(@fields < 3) and next;

		my ($txid, $rank) = @fields[0,2];

			# skip entries with empty ranks #
		if ( $txid eq '' ) {
			print STDERR "Found an entry with an empty taxon_ID .\n";
			next;
		}

		( ($rank eq '') or ($rank  eq 'no rank' ) ) and next;
		$txid2rank->{$txid} = $rank;
	}

	close ($fh);

	return 1;

}			# end of the subroutine fetch_taxon_ranks #


# if_ancestor:
#	This subroutine examines whether the 1st argument
#	is an ancestor of the 2nd argument or not.
#
# INPUT: ($node1, $node2, \%node2pa), where
#	%node2pa = ($node_id => $parent_id, ...);
#
# OUTPUT: $if_anc = 0 if $node1 eq $node2,
#	    = {distance between the nodes} if $node1 is an ancestor of $node2
#	    = -1 if $node1 is NOT an ancestor of $node2 but belongs to the same tree,
#	    = undef if $node1 does NOT belong to the same tree as $node2.
#
sub if_ancestor ($$\%) {

	my ($node1, $node2, $node2pa) = @_;

	($node1 eq $node2) and (return 0);

	my $dist = 1;
	my $anc = $node2pa->{$node2};

	while (defined $anc) {
	    ($anc eq $node1) and (return $dist);
	    $anc = $node2pa->{$anc};
	    $dist++;
	}

		# $node1 is NOT an ancestor of $node2 #

	(defined $node2pa->{$node1}) and (return -1);

	return undef;

}		# End of the subroutine 'if_ancestor' #


# fetch_ancestry: This subroutine receives the ID of a node, along with
#	an output of the subroutine 'fetch_taxonomy_tree_bottomup' in the form:
#		(node_id => parent node_id, ...),
#	and returns  a series of ancestor IDs in the form:
#	(ID of the top node, ..., ID of the parent, the input node ID).
#
sub fetch_ancestry ($\%) {

	my ($nodeid, $node2parent) = @_;

	my @ancestry = ($nodeid);
	my $pres_node = $nodeid;

	while (defined (my $predecessor = $node2parent->{$pres_node}) ) {
		unshift @ancestry, $predecessor;
		$pres_node = $predecessor;
	}

	return \@ancestry;

}			# end of the subroutine fetch_ancestry #


# calculate_abs_depth:
#	This subroutine calculates absolute depth of a node in a tree.
#
# INPUT: ($node_id, \%node2parent), where
#	%node2parent = ($node_id => $parent_id, ...);
#
# OUTPUT: $abs_depth, which is defined as
#	scalar (@ancestry) - 1.
#
sub calculate_abs_depth ($\%) {

	my ($nodeid, $node2parent) = @_;

	my $ancestry = fetch_ancestry ($nodeid, %{$node2parent}) ;

	my $abs_depth = scalar (@{$ancestry}) - 1;

	return $abs_depth;

}		# End of the subroutine 'calculate_abs_depth' #


# fetch_upper_ranks: This subroutine receives a taxon ID, 
#		and extracts the upper-level taxon-IDs with ranks (except 'no rank'),
#		then constructs a hash table:
#		(rank_name => corresponding taxon_id , ...).
#
sub fetch_upper_ranks ($\%\%) {

	my ($taxon_id, $id2parent, $id2rank) = @_;

	my %upper_ranks;

	my $upper_taxon_ids = fetch_ancestry ($taxon_id, %{$id2parent}) ;

	foreach my $upper_taxon (@{$upper_taxon_ids}) {

		my $rank = $id2rank->{$upper_taxon};
		(defined $rank) or next;
		( ($rank eq '') or ($rank eq 'no rank') ) and next;
		$upper_ranks{$rank} = $upper_taxon;
	}

	return \%upper_ranks;

}	# end of the subroutine fetch_upper_ranks #


# fetch_best_taxon_name: This subroutine receives a taxon ID and a hash table:
#		%{$id2cl_names} = (taxon_ID => {name_class => taxon_name, ...}, ...),
#		and returns taxon name which is first found in the following order:
#		scientific name => genbank common name => common name.
#
sub fetch_best_taxon_name ($\%) {

	my ($taxon_id, $id2cl_names) = @_;

	my $cl_names = $id2cl_names->{$taxon_id};

	unless (defined $cl_names) {
 	    print STDERR "NO classed names for the taxon ID $taxon_id.\n";
	    return undef;
	}

	if (defined $cl_names->{'scientific name'} ) {

	    return $cl_names->{'scientific name'}[0];
	} elsif (defined $cl_names->{'genbank common name'} ) {

	    return $cl_names->{'genbank common name'}[0];
	} elsif (defined $cl_names->{'common name'}) {

 	    return $cl_names->{'common name'}[0];	
	} else {
	    print STDERR "NO DECENTly classed names for the taxon ID $taxon_id.\n";
	    return undef;
	}

}		# end of the subroutine fetch_best_taxon_name #

# batch_fetch_best_taxon_names:
#	This subroutine just conducts 'fetch_best_taxon_name' on
#	multiple taxon IDs.
#
# INPUT: (\@taxon_ids, \%id2cl_names).
#
# OUTPUT: \@best_taxon_names, where
#	$best_taxon_names[m-1] is the best taxon name for the m-th taxon ID,
#	or the taxon ID itself if the best taxon name is NOT available.
#
sub batch_fetch_best_taxon_names (\@\%) {

	my ($taxon_ids, $id2cl_names) = @_;

	(@{$taxon_ids} > 0) or (return [] );
	my @best_taxon_names = ();

	foreach my $id (@{$taxon_ids}) {
	    my $best_name = fetch_best_taxon_name ($id, %{$id2cl_names});
	    (defined $best_name) or ($best_name = $id);
	    push @best_taxon_names, $best_name;
	}

	return \@best_taxon_names;

}		# end of the subroutine 'batch_fetch_best_taxon_names' #


# fetch_name_upper_ranks: This subroutine receives a taxon ID,
#		and extracts the upper-level taxon NAMEs with ranks (except 'no rank'),
#		then constructs a hash table:
#		(rank_name => representative taxon_name, ...).
#		It also returns the product of the intermediate process 'fetch_upper_ranks':
#		(rank_name => corresponding taxon_id, ...).
#
sub fetch_name_upper_ranks ($\%\%\%) {

	my ($taxon_id, $id2parent, $id2rank, $id2cl_names) = @_;

	my $upper_ranks = fetch_upper_ranks ($taxon_id, %{$id2parent}, %{$id2rank}) ;

	my $taxon_name = fetch_best_taxon_name ($taxon_id, %{$id2cl_names});
	unless (defined $taxon_name) {
		print STDERR "'fetch_best_taxon_name' failed for the input taxon ID $taxon_id of 'fetch_name_upper_ranks'.\n";
		$taxon_name = 'NOT_AVAILABLE';
	}

	my %named_upranks; 

	foreach my $rank (keys %{$upper_ranks}) {

	    my $rank_id = $upper_ranks->{$rank};
	    my $rank_name  = fetch_best_taxon_name ($rank_id, %{$id2cl_names});
	    unless (defined $rank_name) {
		print STDERR "\'fetch_best_taxon_name\' failed for the $rank ID $rank_id which the taxon ID $taxon_id (name: $taxon_name) belong to.\n"; 
		$rank_name = 'NOT_AVAILABLE';
	    }
	    $named_upranks{$rank} = $rank_name;
	}

	return ($upper_ranks, \%named_upranks);

}	# end of the subroutine fetch_name_upper_ranks #


sub fetch_all_offsprings ($\%); # This is necessary because the subroutine recurs.

# fetch_all_offsprings : This subroutine receives the ID of a node, as well as
#		an output of the subroutine 'fetch_ taxonomy_tree_topdown' in the form:
#		(node_id => [a list of children node_id's], ...),
#		and returns a list of ALL offsprings including itself
#		(, which is supposed to appear in the frontmost of the output array) .
#
sub fetch_all_offsprings ($\%) {

	my ($nodeid, $node2children) = @_;

	my @offsprings = ();

	if (defined $node2children->{$nodeid}) {  # an internal node #

	    foreach my $child_id (@{$node2children->{$nodeid}}) {
		my $sub_offsprings
		   = fetch_all_offsprings ($child_id, %{$node2children});
		#unshift @offsprings, @{$sub_offsprings};
		push @offsprings, @{$sub_offsprings};
	    }

	}

	unshift @offsprings, $nodeid; # put the input node itself in front of the output

	return \@offsprings;

}			# end of the subroutine fetch_all_offsprings #


sub fetch_ext_offsprings ($\%); # This is necessary because the subroutine recurs.

# fetch_ext_offsprings : This subroutine receives the ID of a node, as well as
#		an output of the subroutine 'fetch_ taxonomy_tree_topdown' in the form:
#		(node_id => [a list of children node_id's], ...),
#		and returns a list of ALL offsprings including itself
#		(, which is supposed to appear in the frontmost of the output array) .
#
sub fetch_ext_offsprings ($\%) {

	my ($nodeid, $node2children) = @_;

	my @offsprings = ();

	if (defined $node2children->{$nodeid}) {  # an internal node #

	    foreach my $child_id (@{$node2children->{$nodeid}}) {
		my $sub_offsprings
		   = fetch_ext_offsprings ($child_id, %{$node2children});
		#unshift @offsprings, @{$sub_offsprings};
		push @offsprings, @{$sub_offsprings};
	    }

	} else {	# an external node #

		unshift @offsprings, $nodeid; # put the input node itself in front of the output
	}

	return \@offsprings;

}			# end of the subroutine fetch_ext_offsprings #


# fetch_ext_offsprings_nr:
#	An improved version of 'fetch_ext_offsprings' that do not resort to
#	the recurring calls of the subroutine.
#
# INPUT: ($nodeid, \%node2children), where
#	%node2children = (ID of an iternal node => [a list of children node IDs], ...).
#
# OUTPUT: \@ext_offsprings,
#	which is a list of IDs of all external offsprings of $nodeid.
#
sub fetch_ext_offsprings_nr ($\%) {

	my ($nodeid, $node2children) = @_;

	my @subjects = ($nodeid);

	my @ext_offsprings = ();

	while (1) { # The outer while-loop #

	    my @new_sbjcts = ();
	    while (my $nodeid = shift @subjects) { # The inner while-loop #
		my $children = $node2children->{$nodeid};
		if (defined $children) { # The node is internal
		    push @new_sbjcts, @{$children};
		} else { # The node is external
		    push @ext_offsprings, $nodeid;
		}
	    }				# End of the inner while-loop #

	    (@new_sbjcts > 0)  or last; # No internal nodes to further examine.

	     @subjects = @new_sbjcts;

	}		# End of the outer while-loop #

	return \@ext_offsprings;

}		# end of the subroutine 'fetch_ext_offsprings_nr' #


# fetch_all_offsprings_nr:
#	This revised version of 'fetch_ext_offsprings_nr' fetches
#	internal nodes as well.
#
# fetch_ext_offsprings_nr:
#	An improved version of 'fetch_ext_offsprings' that do not resort to
#	the recurring calls of the subroutine.
#
# INPUT: ($nodeid, \%node2children), where
#	%node2children = (ID of an iternal node => [a list of children node IDs], ...).
#
# OUTPUT: \@ext_offsprings,
#	which is a list of IDs of all external offsprings of $nodeid.
#
sub fetch_all_offsprings_nr ($\%) {

	my ($nodeid, $node2children) = @_;

	my @subjects = ($nodeid);

	my @all_offsprings = ();

	while (1) { # The outer while-loop #

	    push @all_offsprings, @subjects;

	    my @new_sbjcts = ();
	    while (my $nodeid = shift @subjects) { # The inner while-loop #
		my $children = $node2children->{$nodeid};

		if (defined $children) { # The node is internal
		    push @new_sbjcts, @{$children};
		}
	    }				# End of the inner while-loop #

	    (@new_sbjcts > 0)  or last; # No internal nodes to further examine.

	     @subjects = @new_sbjcts;

	}		# End of the outer while-loop #

	return \@all_offsprings;

}		# end of the subroutine 'fetch_all_offsprings_nr' #


sub fetch_ancestry ($\%);

# fetch_last_common_ancestor: This subroutine receives two node IDs
#		and returns the ID of their last common ancestor.
#		It also returns the 'depth' of the last common ancestor,
#		which tells us how many steps we have to  descend in order to
#		reach the last common ancestor from the top node.
#		The larger the depth is, the closer the kinship is between the 2 nodes.
#
sub fetch_last_common_ancestor ($$\%) {

	my ($node1, $node2, $node2parent)  = @_;

	my $ancestry1 = fetch_ancestry ($node1, %{$node2parent}) ;
	unless ( (defined $ancestry1) and @{$ancestry1} ) {
		print STDERR "NO ancestry record for the the 1st node ($node1) for 'fetch_last_common_ancestor'.\n";
		return (undef, undef);
	}

	my $ancestry2 = fetch_ancestry ($node2, %{$node2parent});
	unless ( (defined $ancestry2) and @{$ancestry2} ) {
		print STDERR "NO ancestry record for the the 2nd node ($node2) for 'fetch_last_common_ancestor'.\n";
		return (undef, undef);
	}

	my $anc1 = shift @{$ancestry1};
	my $anc2 = shift @{$ancestry2};
	unless ($anc1 eq $anc2) {
		print STDERR "The two nodes ($node1, $node2) do NOT share a common ancestry.\n";
		return (undef, undef);
	}

	my $common_anc = $anc1;
	my $depth = 0;

	while (1) {

		my $anc1 = shift @{$ancestry1};
		my $anc2 = shift @{$ancestry2};

			# either of the  lineages has terminated #
		( (defined $anc1) and (defined $anc2) ) or last;

			# the two lineages begin to diverge.
		($anc1 eq $anc2) or last;

		$common_anc = $anc1; # still the ancestry is shared.
		$depth++;
	}

	return ($common_anc, $depth);

}		# end of the subroutine fetch_last_common_ancestor #


# fetch_last_common_ancestor_ov2: This subroutine receives ONE OR MORE node IDs
#		and returns the ID of their last common ancestor.
#		It also returns the 'depth' of the last common ancestor,
#		which tells us how many steps we have to  descend in order to
#		reach the last common ancestor from the top node.
#		The larger the depth is, the closer the kinship is between the 2 nodes.
#
sub fetch_last_common_ancestor_ov2 (\@\%) {

	my ($list_nodes, $node2parent)  = @_;

	my $ct_nodes = @{$list_nodes};

	if ($ct_nodes < 1) {
		print STDERR "NO nodes as an input for 'fetch_last_common_ancestor_ov2'.\n";
		return (undef, undef);
	}

	my @ancestries = ();

	for (my $i=0; $i<$ct_nodes; $i++) {
	    my $node = $list_nodes->[$i];
	    my $ancestry = fetch_ancestry ($node, %{$node2parent}) ;
	    unless ( (defined $ancestry) and @{$ancestry} ) {
		print STDERR "NO ancestry record for the the $i-th node ($node) for 'fetch_last_common_ancestor_ov2'.\n";
		next;
	    }
	    push @ancestries, $ancestry;
	}

	if  (my $ct_missing = scalar (@ancestries) - $ct_nodes) {
		printf STDERR "MISSING ancestry records ($ct_missing out of %d) for 'fetch_last_common_ancestor_ov2'.\n", scalar ( @ancestries) ;
		return (undef, undef);
		
	}

	my $anc1 = shift @{$ancestries[0]};
	my $ct_unshare = 0;
	for (my $i=1; $i<$ct_nodes; $i++) {
	    my $anc2 = shift @{$ancestries[$i]};
	    unless ($anc1 eq $anc2) {
		print STDERR "The two nodes ($list_nodes->[0], $list_nodes->[$i]) do NOT share a common ancestry.\n";
	        $ct_unshare++;
	     }
	}

	if ($ct_unshare) {
		printf  STDERR "There are $ct_unshare nodes (out of $ct_nodes) which do NOT share a common ancestry for 'fetch_last_common_ancestor_ov2'.\n";
		return (undef, undef);
	}

	my $common_anc = $anc1;
	my $depth = 0;

	OUTER:
	while (1) {

	    my $anc1 = shift @{$ancestries[0]};
	    (defined $anc1) or last; # the reference lineage has terminated

	    for (my $i=1; $i < $ct_nodes; $i++) {
		
		my $anc2 = shift @{$ancestries[$i]};

			# the subject lineage has terminated #
		(defined $anc2) or (last OUTER);

			# the two lineages begin to diverge.
		($anc1 eq $anc2) or (last OUTER);
	    }

	     $common_anc = $anc1; # All the nodes still share the ancestry.
	     $depth++;
	}

	return ($common_anc, $depth);

}		# end of the subroutine fetch_last_common_ancestor_ov2 #


# fetch_common_ancestors: This subroutine receives two node IDs
#		and returns the lineage of their common ancestors:
#		(top_node_ID, ..., ID of the last common ancestor).
#		It also returns the lineages leading to the two nodes
#		beginning with the divergence point:
# 		(ID of the first node after the divergence, ..., ID of one of the input nodes)
#
# OUTPUT: (\@common_ancestors, \@after_divergence1, \@after_divergence2)
#
sub fetch_common_ancestors ($$\%) {

	my ($node1, $node2, $node2parent)  = @_;

	my $ancestry1 = fetch_ancestry ($node1, %{$node2parent}) ;
	unless ( (defined $ancestry1) and @{$ancestry1} ) {
		print STDERR "NO ancestry record for the the 1st node ($node1) for 'fetch_common_ancestors'.\n";
		return (undef, undef, undef);
	}

	my $ancestry2 = fetch_ancestry ($node2, %{$node2parent});
	unless ( (defined $ancestry2) and @{$ancestry2} ) {
		print STDERR "NO ancestry record for the the 2nd node ($node2) for 'fetch_common_ancestor'.\n";
		return (undef, undef, undef);
	}

	my $anc1 = shift @{$ancestry1};
	my $anc2 = shift @{$ancestry2};
	unless ($anc1 eq $anc2) {
		print STDERR "The two nodes ($node1, $node2) do NOT share a common ancestry.\n";
		return (undef, undef, undef);
	}

	my @common_anc = ($anc1);

	while (1) {

		$anc1 = shift @{$ancestry1};
		$anc2 = shift @{$ancestry2};

			# either of the  lineages has terminated #
		( (defined $anc1) and (defined $anc2) ) or last;

			# the two lineages begin to diverge.
		($anc1 eq $anc2) or last;

		push @common_anc, $anc1; # still the ancestry is shared.
	}

	my @after_divergence1 = @{$ancestry1};
	my @after_divergence2 = @{$ancestry2};

	(defined $anc1) and (unshift @after_divergence1, $anc1);
	(defined $anc2) and (unshift @after_divergence2, $anc2);

	return (\@common_anc, \@after_divergence1, \@after_divergence2);

}		# end of the subroutine fetch_common_ancestors #


# fetch_intervene_nodes:
#	This subroutine extracts a series of nodes
#	that intervenes $node1 and $node2.
#
# INPUT: ($node1, $node2, \%node2pa), where
#	%node2pa = ($node_id => $parent_id, ...);
#
# OUTPUT: \@intervene_nodes, where
#	@intervene_nodes = ($neighbor_of_node1, ..., $neighbor_of_node2)
#	that connects $node1 and $node2 in the shortest path;
#	
sub fetch_intervene_nodes ($$\%) {

	my ($node1, $node2, $node2pa) = @_;

	my ($common_anc, $lin1, $lin2) = fetch_common_ancestors ($node1, $node2, %{$node2pa});
	unless (defined $common_anc) {
	    print STDERR "[ERROR 'fetch_intervene_nodes'] 'fetch_common_ancestors' FAILED: $! !\n";
	    return undef;
	}

	my $lca = pop @{$common_anc};
	unless (defined $lca) {
	    print STDERR "[ERROR 'fetch_intervene_nodes'] '$node1' and '$node2' do NOT SHARE any ANCESTRY: $! !\n";
	    return undef;
	}

	my @int_nodes = ($lca, @{$lin2});
	while (defined (my $anc1 = shift @{$lin1})) { unshift @int_nodes, $anc1; }

	pop @int_nodes; # remove $node2
	shift @int_nodes; # remove $node1

	return \@int_nodes;

}		# End of the subroutine 'fetch_intervene_nodes' #


# sibling_subfamilies:
#	This subroutine receives a pair of nodes,
#	fetches the common ancestry as well as ancestries specific to the individual nodes,
#	and finally, fetches a set of all external nodes belonging to each subtree
#	after the divergence of the two lineages.
#
# INPUT: ($node1, $node2, \%node2pa, \%node2chi), where
#	%node2pa = ($node_id => $parent_id, ...);
#	%node2chi = ($node_id => \@child_ids, ...).
#
# OUTPUT: (\@extnodes_1, \@extnodes_2, \@common_anc, \@lineage_1, \@lineage_2), where
#	@extnodes_1/2 lists the external nodes in the subtree after the divergence towards node 1/2;
#	@common_anc = (top_node_ID, ..., ID of the last common ancestor);
# 	@lineage_1/2 = (ID of the first node after the divergence towards $node1/2,
#	..., $node1/2).
#
sub sibling_subfamilies ($$\%\%) {

	my ($node1, $node2, $node2pa, $node2chi) = @_;

	my ($common_anc, $lineage_1, $lineage_2) = fetch_common_ancestors ($node1, $node2, %{$node2pa});

	my $extnodes_1 = fetch_ext_offsprings_nr ($lineage_1->[0], %{$node2chi});
	my $extnodes_2 = fetch_ext_offsprings_nr ($lineage_2->[0], %{$node2chi});

	return ($extnodes_1, $extnodes_2, $common_anc, $lineage_1, $lineage_2);

}		# End of the subroutine 'sibling_subfamilies' #


# fetch_seqid2taxonid: This subroutine reads in a table 'taxon_dad_xxxx_yyyy.txt(.gz)', which follows the format:
#
#Local_ID  DDBJ_accession  Protein_ID  order_ID  order_name  family_ID  family_name  genus_ID  genus_name  species_ID  species_name ,
#
#		where the fields are separeted from adjacent ones by tabs ("\t"s).
#		Then the subroutine constructs a hash table:
#		(local sequence ID => taxon ID, ...),
#		where 'taxon ID' is the corresponding ID of the taxon rank specified by the 2nd argument.
#
# INPUT: (\$infile, \%seqid2taxonid, $taxon_rank, \%name2indx), where
#	%seqid2taxonid is the hash where the correspondence
#	(sequence ID => taxon ID) is to be stored,
#	%name2indx is the correspondence table (taxon rank => index)
#	for your particular input files.
#	If the default setting is fine, just pass an empty hash.
#
sub fetch_seqid2taxonid (\$\%$\%) {

	my ($infile, $seqid2taxonid, $taxon_rank, $name2indx) = @_;

	my %taxon_rank2indx  = (scalar keys %{$name2indx} ) ? %{$name2indx} :
		('order' => 3, 'family' => 5, 'genus' => 7, 'species' => 9);
			
	my $taxon_indx = $taxon_rank2indx{$taxon_rank};
	unless (defined $taxon_indx) {
		print STDERR "NO INDEX corresponds to the taxon rank '$taxon_rank'.\n";
		return 0;
	}

	local $/ = "\n"; 
	my $fh = open_input_file (${$infile});

	my $header = <$fh>;
	while (my $in = <$fh> ) { 

		my @fields = split /\t/, $in;
		(@fields > $taxon_indx) or next;

		my ($seqid, $taxonid) = @fields[0, $taxon_indx];
		$seqid2taxonid->{$seqid} = $taxonid;
	}

	close ($fh);

	return 1;

}		# end of the subroutine fetch_seqid2taxonid #


use MyArrayManipu qw ( diff_array union_array);

# name2id_new_lineages: This subroutine receives an array which 
#		represents new or finer phylogenetic relationships
#		among taxa (each given by its name),
#		and replaces ONLY BOTH ENDS with the corresponding
#		taxon ID (in order to avoid unexpected conversion
#		of mediating taxa to totally unrelated taxon ID).
#		It leaves them as they are if you cannot find
#		the corresponding ID.
#
# INPUT: (\@new_lineages, \%taxon_name2id), where
#	@new_lineages = ([the end descendant ID, {a series of mediating taxa} ,
#			the end ancestor ID], ...),
#	where the ancestor (and possibly the descentant) are already incorporated
#	in the input taxonomy tree, and
#	%taxon_name2id = (taxon name => taxon ID, ...).
#
sub name2id_new_lineages (\@\%) {

	my ($new_lineages, $taxon_name2id) = @_;

	my @converted = ();
	my $ct_lin = @{$new_lineages};

	for (my $i=0; $i < $ct_lin; $i++) {
		my @lineage = @{$new_lineages->[$i]};

		my $id_descendant = $taxon_name2id->{$lineage[0]};
		(defined $id_descendant) and ($lineage[0] = $id_descendant) ;
		my $ancestor = pop @lineage;
		my $id_ancestor = $taxon_name2id->{$ancestor};
		push @lineage, ( (defined $id_ancestor) ?  $id_ancestor : $ancestor) ;
		push @converted, \@lineage;
	}

	return \@converted;

}	# end of the subroutine name2id_new_lineages #


# revise_taxonomy_tree: This subroutine receives 2 hashes
#		which together represent a taxonomy tree,
#		and an array which gives new and/or finer phylogenetical
#		relationships,
#		then it revises the input taxonomy tree.
#
# INPUT: (\%node2parent, \%node2children, \@new_lineages), where
#	 %node2parent = (node ID => parent ID, ...),
#	%node2children = (node ID => [IDs of children], ...),
#	@new_lineages = ([the end descendant ID, {a series of mediating taxa} ,
#			the end ancestor ID], ...),
#	where the ancestor (and possibly the descentant) are already incorporated
#	in the input taxonomy tree.
#
sub revise_taxonomy_tree (\%\%\@) {

	my ($node2parent, $node2children, $new_lineages) = @_;

	my %obs_node2children; # the (parent  => [children]) relationships
				# which this revison will make obsolete.

	my $ct_lin = @{$new_lineages};
	for (my $i=0; $i <$ct_lin; $i++) {
	    my $lineage = $new_lineages->[$i];
	    my $ct_ranks = @{$lineage};

	    for (my $k = 1; $k < $ct_ranks; $k++) {
		my ($child, $parent) = @{$lineage}[$k -1, $k];

			# overwrite the old relationship if there is.
		$node2parent->{$child} = $parent;

		if (defined $node2children->{$parent}) {
		    $node2children->{$parent} =
			union_array (@{$node2children->{$parent}}, @{[$child]} );
		} else {
		    $node2children->{$parent} = [ $child ];
		}
	    }

	    ($ct_ranks < 3) and next; # in order not to eleminate
				      # new relationships by mistake

	    my ($end_descendant, $end_ancestor) = @{$lineage}[0, $ct_ranks -1];

	    if (defined $obs_node2children{$end_ancestor}) {
		push @{$obs_node2children{$end_ancestor}}, $end_descendant;
	    } else {
		$obs_node2children{$end_ancestor} = [$end_descendant];
	    }
	}

		# remove the (parent -> child) relationships
		# which has become obsolete

	foreach my $parent (keys %obs_node2children) {
	    my $obs_children = $obs_node2children{$parent};
	    my $old_children = $node2children->{$parent};
	    (defined $old_children) or next; # no pre-recorded children 

	    my ($obsolete, $new_children, $not_entried )
			= diff_array (@{$old_children}, @{$obs_children});

	     $node2children->{$parent} = $new_children;
	}

	return 1;

}		# end of the subroutine revise_taxonomy_tree #


# make_nmhierarchy:
#	This subroutine makes a hierarchical classification,
#	or a tree-like structure,
#	of the subset of entries placed under a specified entry.
#
# INPUT: ($top_entry, \%entry2descendants), where
#	$top_entry is the entry that is designated as the top of the hierarchy;
#	%entry2descendants = ($entry => \@descendants, ...), with
#	@descendants listing ALL the 'descendants' of $entry ;
#	PREREQUISITE:
#	In this subroutine, input entries (and their descndants)
#	are specified by numerical indices (or IDs).
#
# OUTPUT: \%hierarchy, where
#	%hierarchy = ($entry => @children, ...) for $top_entry and its descendants,
#	with @children listing all the 'children' (or direct descendants) of $entry.
#	@children = () (empty) for the terminal entries;
#	The subroutine returns undef is the process failed.
#
sub make_nmhierarchy ($\%) {

	my ($top_entry, $entry2descendants) = @_;

	my %hierarchy;
	my @subjects = ($top_entry);
	while (@subjects>0) { # outer while-loop #

	    my @new_subjects = ();

	    foreach my $entry (@subjects) { # middle foreach-loop #
		my $descendants = $entry2descendants->{$entry};
		unless ((defined $descendants) and (@{$descendants}>0)) {
		    $hierarchy{$entry} = [];
		    next;
		}
		my $nr_indirect = [];
		foreach my $desc (@{$descendants}) { # Inner foreach-loop #
		    my $desdesc = $entry2descendants->{$desc};
		    (defined $desdesc) or next;
		     ## push @indirect, @{$desdesc};
		    $nr_indirect = union_nmarray (@{$nr_indirect}, @{$desdesc});
		}			# End of the inner foreach-loop #
		##my $nr_indirect = nr_nmarray (@indirect);
		my ($indirect2, $children, $ex2) = diff_nmarray (@{$descendants}, @{$nr_indirect});
		if (@{$ex2}>0) {
		    print STDERR "[ERROR 'make_nmhierarchy'] The following INDIRECT DESCENDANTS are NOT contained in the set of ALL DESCENDANTS of the entry '$entry' (top entry: $top_entry):  ", join ('  ', @{$ex2}), " : $! !\n";
		    return undef;
		}
		push @new_subjects, @{$children};
		$hierarchy{$entry} = $children;

	    }			# End of the middle foreach-loop #

	    my $nr_new_subjects = nr_nmarray (@new_subjects);
	    @subjects = @{$nr_new_subjects};
	}		# End of the outer while-loop #

	return \%hierarchy;

}		# End of the subroutine 'make_nmhierarchy' #


1;	# a "spell" which makes the module "use"-able and/or "require"-able #


