package MyTreeManipu_II;
#
###################################################################################
#	DENSERM_P: A package to DEtect Negative SElection on Recurrent Mutations
#			(almost exclusively via Perl modules and scripts).
#	Version 0.3: Copyright (C) 2012 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyTreeManipu_II.pm: This package can process more general types of trees,
#	(i.e. NOT limited to binary trees with trichotomous top nodes).
#	It is also consistent with the format of the taxonomy tree which are handled by
#	the module 'MyTaxonomy.pm'.
#
# MyTreeManipu.pm: This package performs basic tree manipulations.
#
#  Subroutines whose substitutes are available in this module ('MyTreeManipu_II.pm')
#	or in 'MyTaxonomy.pm':
#   our @EXPORT = qw(retrieve_top_node confine_outgroup print_unrooted);
#   our @EXPORT_OK = qw(switch_topnode  print_subtree topo_unrooted topo_subtree
#		list_sequences fetch_node_series dist_nodes
#		    );
#
# Subroutines whose substitutes are not available yet:
#   our @EXPORT_OK = ( retrieve_bootstrap  prod_bootstrap );
#

use strict;
use warnings;
use MyArrayManipu qw( diff_array union_array );
use MyFile_manipu qw( open_input_file );
use MyTaxonomy qw( fetch_ancestry fetch_last_common_ancestor_ov2 fetch_common_ancestors
		 fetch_ext_offsprings fetch_all_offsprings );

use Exporter;

our @ISA = qw(Exporter);

##sub rec_fillin_boot (\@\$);
#sub rec_prod_boot (\@\@);


our @EXPORT_OK = qw(
	extract_sgltree_from_file
	input_tree_file read_tree_file input_subtree
	input_tree_with_nodeids copy_tree
	output_tree draw_subtree
	br_output_tree br_draw_subtree topo_subtree
	br_output_tree_sep br_draw_subtree_sep
	make_node2children make_node2parent
	shift_topnode move_topnode segregate_subset_nodes
	sum_branch_lengths intra_lineage_distance inter_node_distance
	fetch_boot_series extract_hidden_bootstrap extract_hidden_bootstrap2
	resolve_multipartite examine_multifurcation
			) ;

sub input_tree_file (\$\%\%\%\%\%);
sub read_tree_file (\$\$\%\%\%\%\%);
sub draw_subtree ($\%\%\%$$);
sub br_draw_subtree ($\%\%\%$$);
sub br_draw_subtree_sep ($\%\%\%$$$);

our $TOP_NODE = 'TOP_NODE';
our $INDX_LENGTH = 0;
our $INDX_BOOT = 1;
our $START_GROUP = '(';	# the symbol marking the start of the grouping of nodes.
our $END_GROUP = ')';  # the symbol marking the end of the grouping of nodes.


### 5 important hash tables which describe a tree ###
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.


# extract_sgltree_from_file:
#	This subroutine is a sort of 'wrapper' of the subroutine 'input_tree_file',
#	and reads in a SINGLE TREE n the Newick format and stores the information.
#	If the comment-initiating 'word' is specified,
#	it removes lines starting with the word.
#
# INPUT: (\$infile, \%node2parent, \%node2children, \%name2id, \%id2name, \%branch_attributes, $comment), where
#	the 1st argument is the path to the input tree file;
#	The 2nd to 6th arguments are hashes whose formats are described above.
#
# OUTPUT: 1 if the process succeeded,
#	  0 if it failed.
#
sub extract_sgltree_from_file (\$\%\%\%\%\%$) {

	my ($infile, $node2pa, $node2chi, $name2id, $id2name, $br_attr, $comment) = @_;
	local $/ = "\n";

	my $fh = open_input_file (${$infile});
	unless (defined $fh) {
	    print STDERR "[ERROR 'extract_sgltree_from_file'] FAILED TO OPEN the input tree file '${$infile}' !\n\n";
	    return 0;
	}
	my $len_comment = length $comment;
	my $intree = '';
	while (my $in = <$fh>) { # The main while-loop #

	    ($in =~ m/\S/) or next; # skip a blank line #
	    $in =~ s/^\s+//; # Remove preceding white spaces #

		# Skip a comment line #
	   ($len_comment <= length ($in)) and
		($comment eq substr ($in, 0, $len_comment)) and next;
	   $intree .= $in;
	}			# End of the main while-loop #
	close ($fh);

=for DEBUGGING the subroutine
	print STDERR "\n< Subroutine 'extract_sgltree_from_file' >\n\n";
	print STDERR "Input tree to be parsed by 'input_tree_file':\n\n";
	print STDERR $intree;
	print STDERR "\n";
=cut

	unless ( input_tree_file ($intree, %{$node2pa}, %{$node2chi}, %{$name2id}, %{$id2name}, %{$br_attr}) ) {
	    print STDERR "[ERROR 'extract_sgltree_from_file'] FAILED to perform the subroutine 'input_tree_file' !\n\n";
	    return 0;
	}

	return 1;

}		# End of the subroutine 'extract_sgltree_from_file' #


# input_tree_file: This subroutine reads in a tree file and stores the information on a whole tree.
#		This is actually an I/O interface of the subroutine 'read_tree_file'.
#
# INPUT is supposed to start with the left parenthesis ('(') , and end with the semicolon ';',
# possibly with some comments written in alphabets,
# and children are supposed to be separated by the comma (',').
#
sub input_tree_file (\$\%\%\%\%\%) {

	my ($input, $node2parent, $node2children, $name2id, $id2name, $branch_attributes) = @_; 

	#${$input} =~ s/\s*\w*\s*\;\s*$//; # remove a superfluous string trailing the main information
	${$input} =~ s/[^\)]*\;\s*$//; # remove a superfluous string trailing the main information

	my $ct_lpar = (${$input} =~ tr/(// );
	my $ct_rpar = (${$input} =~ tr/)// );

	unless ($ct_lpar == $ct_rpar) {
	    print STDERR "[ERROR 'input_tree_file'] The number of left parentheses ($ct_lpar) CLASHES with that of right parentheses ($ct_rpar) !\n";
	    return 0;
	}

	unless (${$input} =~ s/^\s*\(\s*// ) { #  remove the left parenthesis at the beginning of the tree
		print STDERR "ILLEGAL input for 'input_tree_file' .\n";
		return 0;
	}

	my $node_id = 0;

	unless ( read_tree_file (${$input}, $node_id, %{$node2parent}, %{$node2children},
			%{$name2id}, %{$id2name}, %{$branch_attributes}) ) {

		print STDERR "'read_tree_file' failed at the node ID $node_id .\n";
		return 0;
	}

	$name2id->{$TOP_NODE} = $node_id;

	if (defined $branch_attributes->{$node_id}) {
		print STDERR "CAUTION: the top_node (ID:$node_id) has somehow information on its 'upper branch', which CANNOT exist:\n";
		print STDERR ' ' x 5 . '{' . join (', ' , @{$branch_attributes->{$node_id}} ) . "\} .\n";
		undef $branch_attributes->{$node_id};
	}

	return 1;

}		# end of the subroutine input_tree_file #


# input_subtree: This subroutine was made to retrieve the complete information from an
#	output of the subroutine 'draw_subtree'.
#
sub input_subtree (\$) {

	my ($in_subtree) = @_;
	${$in_subtree} =~ s/\s*\;\s*$//;

	my ($dist, $boot) ;

	my ($appended) = (${$in_subtree} =~ m/([^\)]+)$/ );
	if (defined $appended) {
	    my $app_len = length ($appended);
	    substr (${$in_subtree}, - $app_len, $app_len, ';'); # remove the appended information from the input
	}

	my %node2parent = my %node2children = my %name2nodeid = my %nodeid2name = my %branch_attributes = ();

	unless ( input_tree_file (${$in_subtree},
		%node2parent, %node2children, %name2nodeid, %nodeid2name, %branch_attributes) ) {

		print STDERR "'input_tree_file' failed.\n";
		return (undef, undef, undef, undef, undef);
	}

	if ( defined $appended) {

			# when -bootlabel=node
		my ($bootstrap, $br_len) = split /\:/, $appended;

			# when -bootlabel=branch
		($bootstrap eq '') and ( ($br_len, $bootstrap) = ($appended =~ m/^\:([^\[]+)\[([^\]]+)\]/ ) );

			# when the branch has no bootstrap value displayed
		(defined $br_len) or ( ($br_len  = $appended) =~ s/^\:// ) ;
		( (defined $bootstrap) and ($bootstrap eq '') ) and (undef $bootstrap);

		my $topnode = $name2nodeid{$TOP_NODE};
		$branch_attributes{$topnode} = [$br_len, $bootstrap];
	}

	return (\%node2parent, \%node2children, \%name2nodeid, \%nodeid2name, \%branch_attributes );

}		# end of the subroutine input_subtree #


# read_tree_file: This subroutine reads in (a part of) a tree-file text
#		and stores the information on the 5 hash tables describing trees.
#
#	REVISED on July 13 (Tue), 2010:
#	The subroutine now can correctly read in significance or other information
#	associated with the exterior node, which was unreadable before the revision.
#
# INPUT is supposed to start with a left parenthesis ('(') ,
#	which must have been removed before being passed to this subroutine,
#	and end with a right parenthesis followed by a semicolon ';',
# 	possibly with some comments (in alphabets) in between,
#	and children are supposed to be separated by commas (',').
#
sub read_tree_file (\$\$\%\%\%\%\%) {

	my ($input, $node_id, $node2parent, $node2children, $name2id, $id2name, $branch_attributes) = @_;

	my @children = ();

	#while (${$input} !~ m/^\)/ ) ) { # the outermost while #
	while ( !(${$input} =~ s/^\)\s*// ) ) { # the outermost while
				# --- continues untill it reaches the end of the group of nodes

	    if (${$input} =~ s/^\s*\(\s*// ) { # reading in a group of nodes #

		unless (read_tree_file (${$input}, ${$node_id}, %{$node2parent}, %{$node2children},
				%{$name2id}, %{$id2name}, %{$branch_attributes} ) ) {
			return 0; # error occurred #
		}

		push @children, ${$node_id};

	    } else { # reading in a single node (probably external) #

		my ($tmp) = (${$input} =~ m/^([^\,\)]+)/ );
		unless (defined $tmp)  {
			print STDERR "Unexpected input : '" . substr (${$input}, 0, 10) . "'...\n";
			return 0;
		}

		substr (${$input}, 0, length ($tmp) , ''); # remove $tmp from the input

		$tmp =~ s/\s+$//;

		#my ($seqname, $br_len) = ($tmp =~ m/^([^\:]+)\:([^\:]+)/ );

			## REVISED on July 13 (Tue), 2010 ##
		my ($seqname, $br_attr) = split /\:/, $tmp, 2;
		##my ($seqname, $br_len) = split /\:/, $tmp;

		(defined $seqname) or ($seqname = $tmp);

		${$node_id}++;
		push @children, ${$node_id};
		$name2id->{$seqname} = ${$node_id};
		$id2name->{${$node_id}} = $seqname;

			## ADDED on July 13 (Tue), 2010 ##
		if ((defined $br_attr) and ($br_attr ne '')) {

			# when -bootlabel=node
		    my ($bootstrap, $br_len) = split /\:/, $br_attr;

			# when -bootlabel=branch
		    ((defined $br_len) and ($br_len =~ m/\d/)) or ( ($br_len, $bootstrap) = ($br_attr =~ m/^([^\[]*)\[([^\]]+)\]/ ) );

		    if (defined $br_len) {
			($br_len eq '')  and (undef $br_len);
		    } else {
				# Only a single attribute is recorded #
			$br_len = $br_attr;
		    }
		    ((defined $bootstrap) and ($bootstrap eq '')) and (undef $bootstrap);

		    my @attr = ();
		    (defined $br_len) and ($attr[$INDX_LENGTH] = $br_len);
		    (defined $bootstrap) and ($attr[$INDX_BOOT] = $bootstrap);
		    (@attr>0) and ($branch_attributes->{${$node_id} } = \@attr);
		}
		##(defined $br_len) and ($branch_attributes->{${$node_id} } = [$br_len, undef] );

	    }

	    ${$input} =~ s/^\,\s*//; # remove the comma delimiting this group from the following one

	}		# end of the outermost while #

	##${$input} =~ s/^\)\s*//;

	${$node_id}++; # The ID of the node that has just been read in.

	$node2children->{${$node_id}} = \@children;
	foreach my $child (@children) { $node2parent->{$child} = ${$node_id}; }
		
	my ($tmp) = (${$input} =~ m/^([^\,\)]+)/ ); # information belonging to the immediate upper branch 
	if ( defined $tmp) {

		substr (${$input}, 0, length ($tmp), '' ); # remove $tmp from the input
		$tmp =~ s/\s+$//;
			# when -bootlabel=node
		my ($bootstrap, $br_len) = split /\:/, $tmp;

			# when -bootlabel=branch
		($bootstrap eq '') and ( ($br_len, $bootstrap) = ($tmp =~ m/^\:([^\[]+)\[([^\]]+)\]/ ) );

			# when the branch has no bootstrap value displayed
		(defined $br_len) or ( ($br_len  = $tmp) =~ s/^\:// ) ;
		( (defined $bootstrap) and ($bootstrap eq '') ) and (undef $bootstrap);

			## REVISED on July 13 (Tue), 2010 ##
		my @attr = ();
		(defined $br_len) and ($attr[$INDX_LENGTH] = $br_len);
		(defined $bootstrap) and ($attr[$INDX_BOOT] = $bootstrap);
		(@attr>0) and ($branch_attributes->{${$node_id}} = \@attr);
		##$branch_attributes->{${$node_id}} = [$br_len, $bootstrap];

	}

	return 1;

}		# end of the subroutine read_tree_file #


# input_tree_with_nodeids:
#	This subroutine can read in a tree, possibly accompanied by
#	node IDs used in the previous analyses.
#	It ASSUMES that each node ID is given on the immediate right of
#	a cluster representing the node.
#
# INPUT: (\$in_tree, $fl_in_nodeid, \$reserve_nodeid), where
#	$in_tree is a chunk describing a tree;
#	$fl_in_nodeid = 1 if you use node IDs accompanying the input tree,
#			= 0 if you newly assign node IDs;
#	$reserve_nodeid is the ID reserved for  nodes whose input IDs are not assigned in the input tree.
#	It will be incremented by 1 before being used.
#
# OUTPUT: (\%node2pa, \%node2chi, \%name2id, \%id2name, \%br_attr), where
#	%node2pa = ($node_ID => $parent_ID, ...);
#	%node2chi = ($node_ID => \@child_IDs, ...);
#	%name2id = ($node_name => $node_ID, ...) for the nodes bearing names
#	(usually exterior nodes and the top node);
#	%id2name = ($node_ID => $node_name, ...) for the nodes bearing names
#	EXCLUDING the top node;
#	%br_attr = ($ID of the node immediatly downstream of the branch
#		=> \@attr, ...) with
#	($br_len, $boot) = @attr[$INDX_LEN, $INDX_BOOT].
#
sub input_tree_with_nodeids (\$$\$) {

	my ($in_tree, $fl_in_nodeid, $reserve_nodeid) = @_;

	my $input_topnode_id;

=for DEBUGGING the subroutine
	print STDERR "\n< Subroutinen 'input_tree_with_nodeids' >\n\n";
	print STDERR "{ Raw input tree }\n\n";
	print STDERR ${$in_tree} , "\n";
=cut

	if ($fl_in_nodeid) {
	    ($input_topnode_id) = (${$in_tree} =~ m/\{\s*([^\s\{\}]+)\s*\}\;\s*$/);
	    unless (defined $input_topnode_id) {
		print STDERR "[CAUTION 'input_tree_with_nodeids'] FAILED to extract topnode ID from the input tree: $! !\n";
		print STDERR ${$in_tree}, "\n";
		##return (undef, undef, undef, undef, undef);
	    }
	    ${$in_tree} =~ s/[^\)]+\;\s*$/\;/;
	}

	my %node2pa; my %node2chi; my %name2id; my %id2name; my %br_attr;
	unless ( input_tree_file (${$in_tree}, %node2pa, %node2chi, %name2id, %id2name, %br_attr) ) {
	    print STDERR "[ERROR 'input_tree_with_nodeids'] The subroutine 'input_tree_file' FAILED to read in the input tree: $!\n";
	    print STDERR ${$in_tree} , "\n";
	    return (undef, undef, undef, undef, undef);
	}

=for DEBUGGING the subroutine
	print STDERR "\n{ 1st extracted information on the input tree }\n\n"; 
	print STDERR "\#Node_ID --> seq_name\n\n";
	foreach my $nodeid (sort {$a <=> $b} keys %id2name) {
	    print STDERR $nodeid, ' --> ', $id2name{$nodeid}, "\n";
	}
	print STDERR "\n\#Seq_name --> node_ID\n\n";
	foreach my $seqname (sort keys %name2id) {
	    print STDERR $seqname, ' --> ', $name2id{$seqname}, "\n";
	}
	print STDERR "\n\#Node_ID <-- parent_ID\n\n";
	foreach my $nodeid (sort {$a <=> $b} keys %node2pa) {
	    print STDERR $nodeid, ' <-- ', $node2pa{$nodeid}, "\n";
	}
	print STDERR "\n\#Node_ID --> \[IDs,of,children\]\n\n";
	foreach my $nodeid (sort {$a <=> $b} keys %node2chi) {
	    print STDERR $nodeid, ' --> [', join (',', @{$node2chi{$nodeid}}), ']', "\n";
	}
	print STDERR "\n#Dw_node_ID  --> \[branch,attributes\]\n\n";
	foreach my $nodeid (sort {$a <=> $b} keys %br_attr) {
	    print STDERR $nodeid, ' --> [', join (',', @{$br_attr{$nodeid}}), ']', "\n";
	}
	print STDERR "\n";
=cut

	$fl_in_nodeid or (return (\%node2pa, \%node2chi, \%name2id, \%id2name, \%br_attr) );

		# Replace automatically assigned node IDs
		# with input node IDs

	my $old_topnode_id = $name2id{$TOP_NODE};

	my %in_id2name; my %in_br_attr;
	my %name2in_id;
	my %old_id2in_id; # ( $old_node_ID =>  $input_node_ID, ...)

	    ## Handle the case where the input topnode ID was not extracted ##
	    ## --> MOVED from below for DEBUGGING (on Sep 1 (Mon), 2008)
	unless (defined $input_topnode_id) {
	    print STDERR "[CAUTION 'input_tree_with_nodeids'] The TOP NODE (automatically assigned ID: $old_topnode_id) does NOT HAVE the corresponding INPUT ID: $! !\n";
	    ${$reserve_nodeid}++;
	    $input_topnode_id = ${$reserve_nodeid};
	    print STDERR "  --- So I will assign the new ID '$input_topnode_id' ...\n\n";
	}

		## Made from conditional to mandatory for DEBUGGING
		## (on Sep 1 (Mon), 2008)
	###if (defined $input_topnode_id) {
	    $name2in_id{$TOP_NODE} = $input_topnode_id;
	    (defined $old_topnode_id) and ($old_id2in_id{$old_topnode_id} = $input_topnode_id);
	###}

	   # Extract input node IDs from node names #
	foreach my $old_id (keys %id2name) { # The 1st foreach-loop #
	    my $old_name = $id2name{$old_id};
	    my ($in_name, $in_id) = split /\s*\{\s*/, $old_name;
	    unless (defined $in_id) {
		print STDERR "[CAUTION 'input_tree_with_nodeids'] FAILED to extract the INPUT ID out of the node name '$old_name' (automatically assigned ID: $old_id): $! !\n";
		next;
	    }
	    $in_id =~ s/\s*\}\s*$//;
	    $name2in_id{$in_name} = $in_id;
	    $in_id2name{$in_id} = $in_name;
	    my $attr = $br_attr{$old_id};
	    (defined $attr) and ($in_br_attr{$in_id} = $attr);
	    $old_id2in_id{$old_id} = $in_id;
	}			# End of the 1st foreach-loop #

	    # Extract input node IDs from branch attributes #
	foreach my $old_id (keys %br_attr) { # The 2nd foreach-loop #
	    my $old_attr = $br_attr{$old_id};
	    my ($old_brlen, $old_boot) = @{$old_attr}[$INDX_LENGTH, $INDX_BOOT];
	    my @new_attr = ();

		# The branch has aready been handled #
	    (defined $old_id2in_id{$old_id}) and next;

	    if ((defined $old_boot) and ($old_boot =~ s/^\s*\{\s*//)) {
		my ($in_id, $in_boot) = split /\s*\}\s*/, $old_boot;
		unless (defined $in_id) {
		    print STDERR "[ERROR 'input_tree_with_nodeids'] FAILED to extract the INPUT ID out of the bootstrap value '\{$old_boot' of the node with automatically assigned ID '$old_id': $! !\n";
		    return (undef, undef, undef, undef, undef);
		}
		$old_id2in_id{$old_id} = $in_id;
		$new_attr[$INDX_LENGTH] = $old_brlen;
		((defined $in_boot) and ($in_boot =~ m/\S/)) and ($new_attr[$INDX_BOOT] = $in_boot);
		((defined $new_attr[$INDX_LENGTH]) or (defined $new_attr[$INDX_BOOT])) and ($in_br_attr{$in_id} = \@new_attr);
		next;
	    }
	    if ((defined $old_brlen) and ($old_brlen =~ s/^\s*\{\s*//)) {
		my ($in_id, $in_brlen) = split /\s*\}\s*/, $old_brlen;
		unless (defined $in_id) {
		    print STDERR "[ERROR 'input_tree_with_nodeids'] FAILED to extract the INPUT ID out of the branch length '\{$old_brlen' of the node with automatically assigned ID '$old_id': $! !\n";
		    return (undef, undef, undef, undef, undef);
		}
		$old_id2in_id{$old_id} = $in_id;
		((defined $in_brlen) and ($in_brlen =~ m/\S/)) and ($new_attr[$INDX_LENGTH] = $in_brlen);
		((defined $old_boot) and ($old_boot =~ m/\S/)) and ($new_attr[$INDX_BOOT] = $old_boot);
		((defined $new_attr[$INDX_LENGTH]) or (defined $new_attr[$INDX_BOOT])) and ($in_br_attr{$in_id} = \@new_attr);
		next;
	    }

		# Somehow, the branch has  been left over #
	    print STDERR "[ERROR 'input_tree_with_nodeids'] The INPUT ID has NOT BEEN FOUND out of EITHER bootstrap value '$old_boot' OR branch length '$old_brlen' of the node with the automatically asigned ID '$old_id': $! !\n";
	    return (undef, undef, undef, undef, undef);

	}		# End of the 2nd foreach-loop #

	    ## Handle the case where the input topnode ID was not extracted ##
	    ## --> Moved upward, and current block has been changed to:
	    ## Handle the case in which a dummy top node is suspected
	    ## (on Sep 1 (Mon), 2008)

		## Double conditions were withdrawn on Sep 1 (Mon), 2008
	##unless (defined $input_topnode_id) {
	 ##  if (defined $old_topnode_id) {
			## A BUG FIXING MEASURE on July 17 (Thu), 2008 ##
		if ((2> scalar keys %id2name) and (0 == scalar keys %br_attr)) {
		    # In this case, the 'tree' must have originally consisted of just one node #
		     %node2pa=();
		     %node2chi = ();

			# Further fixed on Sep 1 (Mon), 2008 #
		     my ($sgl_node_id) = keys %in_id2name;
			# BUG fixed on July 17 (Thu), 2008 #
		     ##($input_topnode_id) = keys %in_id2name;
		     ##($input_topnode_id) = keys %id2name;

			## ADDED on Sep 1 (Mon) 2008
		     (defined $old_topnode_id) or ($old_topnode_id = 'NA');

			# Revised on Sep 1 (Mon), 2008 #
		     print STDERR "[CAUTION 'input_tree_with_nodeids'] The tree seems to consist of a single node '$sgl_node_id', and the \"Top node\" '$input_topnode_id' appear just dummy. So I will just dismiss the latter.\n\n";
		     ##print STDERR "[CAUTION 'input_tree_with_nodeids'] The tree seems to consist of a single node '$input_topnode_id', and the \"Top node\" '$old_topnode_id' appear just dummy. So I will just dismiss the latter.\n\n";

		    $input_topnode_id = $sgl_node_id;
			## Moved from below on Sep 1 (Mon), 2008
		    $name2in_id{$TOP_NODE} = $input_topnode_id;

=MOVED UPWARD for DEBUGGING  (on Sep 1 (Mon), 2008)
		} else {
			# In this case, we consider that the top node ID was just forgotten to be displayed.
		    $input_topnode_id = $old_id2in_id{$old_topnode_id};
		    unless (defined $input_topnode_id) {
			print STDERR "[CAUTION 'input_tree_with_nodeids'] The TOP NODE (automatically assigned ID: $old_topnode_id) does NOT HAVE the corresponding INPUT ID: $! !\n";
			${$reserve_nodeid}++;
			$input_topnode_id = ${$reserve_nodeid};
			print STDERR "  --- So I will assign the new ID '$input_topnode_id' ...\n\n";
			$old_id2in_id{$old_topnode_id} = $input_topnode_id;
		    }
=cut
		}
			## Moved upward on Sep 1 (Mon), 2008
		###$name2in_id{$TOP_NODE} = $input_topnode_id;
	##   }
	##}

	my %in_nodeid2pa;
	foreach my $old_id (keys %node2pa) { # The 3rd foreach-loop #
	    my $old_pa = $node2pa{$old_id};
	    my $in_id = $old_id2in_id{$old_id};
	    unless (defined $in_id) {
		print STDERR "[ERROR 'input_tree_with_nodeids'] CanNOT FIND the INPUT ID for the node with automatically assigned ID '$old_id' (and with parent '$old_pa'): $! !\n";
		return (undef, undef, undef, undef, undef);
	    }
	    my $in_pa = $old_id2in_id{$old_pa};
	    unless (defined $in_pa) {
		print STDERR "[ERROR 'input_tree_with_nodeids'] CanNOT FIND the INPUT ID for the parent (automatically assigned ID '$old_pa') of the node '$in_id'  (automatically assigned ID '$old_id'): $! !\n";
		return (undef, undef, undef, undef, undef);
	    }
	    $in_nodeid2pa{$in_id} = $in_pa;
	}			# End of the 3rd foreach-loop #

	my %in_nodeid2chi;
	foreach my $old_id (keys %node2chi) { # The 4th foreach-loop #
	    my $old_children = $node2chi{$old_id};
	    my $in_id = $old_id2in_id{$old_id};
	    unless (defined $in_id) {
		print STDERR "[ERROR 'input_tree_with_nodeids'] CanNOT FIND the INPUT ID for the node with automatically assigned ID '$old_id' (and with children ", join (' ', @{$old_children}), "): $! !\n";
		return (undef, undef, undef, undef, undef);
	    }
	    my @in_children = ();
	    foreach my $old_chi (@{$old_children}) { # The inner foreach-loop #
		my $in_chi = $old_id2in_id{$old_chi};
		unless (defined $in_chi) {
		     print STDERR "[ERROR 'input_tree_with_nodeids'] CanNOT FIND the INPUT ID for the child (automarically assigned ID '$old_chi') of the node '$in_id' (automatically assigned ID '$old_id'): $! !\n";
		    return (undef, undef, undef, undef, undef);
		}

		push @in_children, $in_chi;
	    }				# End of the inner foreach-loop #

	    $in_nodeid2chi{$in_id} = \@in_children;
	}			# End of the 4th foreach-loop #

=for DEBUGGING the subroutine
	print STDERR "\n --- End of the subroutine 'input_tree_with_nodeids' ---\n\n";
=cut

	return (\%in_nodeid2pa, \%in_nodeid2chi, \%name2in_id, \%in_id2name, \%in_br_attr);

}		# End of the subroutin 'input_tree_with_nodeids' #


# copy_tree: This subroutine copies 4 out of 5 hashes which describes a tree.
#	This does NOT copy %id2name with the assumption that they will not change forever.
#
sub copy_tree (\%\%\%\%) {

	my ($node2parent, $node2children, $name2id, $branch_attributes) = @_;

	my %cp_node2parent = %{$node2parent};
	my %cp_node2children;
	foreach my $node (%{$node2children}) {
	    my $children = $node2children->{$node};
 	    (defined $children) or next;
	    my @cp_children = @{$children};
	    $cp_node2children{$node} = \@cp_children;
	}
	my %cp_name2id = %{$name2id};
	my %cp_br_attr;
	foreach my $node (%{$branch_attributes}) {
	    my $attributes = $branch_attributes->{$node};
	    (defined $attributes) or next;
	    my @cp_attributes = @{$attributes};
	    $cp_br_attr{$node} =  \@cp_attributes;
	}

	return (\%cp_node2parent, \%cp_node2children, \%cp_name2id, \%cp_br_attr) ;

}		# end of the subroutine copy_tree #


###=for the moment, comment out the following subroutines in order to test the script little by little

# output_tree: This subroutine outputs a part of a tree under a specified node into a filehandle.
#		It outputs the daughter nodes in the ascinding order of the sizes of subtrees under them.
#
# INPUT: ($node_id, $node2children, $id2name, $branch_attributes, $filehandle,
#	$option_bootlabel, $option_nodeid),
#	where $node_id is the ID of the node which is the top of the part of the tree being output,
#	$option_bootlabel = 1 if you dsiplay  bootstrap at the branch, =0 (or undef) at the node.
#	$option_nodeid = 1 if you display node ID, =0 (or undef) otherwise.
#
sub output_tree ($\%\%\%\*$$) {

	my ($node_id, $node2children, $id2name, $branch_attributes, $fh,
		$option_bootlabel, $option_nodeid) = @_;

	my $tree = draw_subtree ($node_id, %{$node2children}, %{$id2name},
		%{$branch_attributes}, $option_bootlabel, $option_nodeid);

		# when the node is an external node #
	(defined $node2children->{$node_id}) or ( ${$tree} =  "( ${$tree} )" ) ;


	print $fh ${$tree} . "\;\n" ;

	return 1;

}		# end of the subroutine output_tree #


# draw_subtree:
#	This subroutine outputs a part of a tree under a specified node
#	in a string of the Newick format.
#	It outputs the daughter nodes in the ascinding order of
#	the sizes of subtrees under them.
#	So the tree gets more and more complex as
#	it moves from left to right.
#
# INPUT: ($node_id, $node2children, $id2name, $branch_attributes,
#	$option_bootlabel, $option_nodeid),
#	where $node_id is the ID of the node which is the top of
#	the part of the tree being output,
#	$option_bootlabel = 1 if you display bootstrap value at the branch,
#			=0 (or undef) at the node. 
#	$option_nodeid = 1 if you display node ID, =0 (or undef) otherwise.
#
sub draw_subtree ($\%\%\%$$) {

	my ($node_id, $node2children, $id2name, $branch_attributes,
		$option_bootlabel, $option_nodeid) = @_;

	my $tree = '';

	my $children = $node2children->{$node_id};
	if (defined $children) {

	    $tree .= "( ";

	    my @children_info = (); # element = [node_ct, \$subtree];
	    foreach my $child (@{$children}) {
		my $nodes = fetch_all_offsprings ($child, %{$node2children});
		my $node_ct = (defined $nodes) ? @{$nodes} : 0;
		my $subtree =
		draw_subtree ($child, %{$node2children}, %{$id2name},
			%{$branch_attributes}, $option_bootlabel, $option_nodeid);
			push @children_info , [$node_ct, $subtree];
	    }

	    my @std_children = sort { $a->[0] <=> $b->[0] } @children_info;
	    my $ct_children = @std_children;
	    for (my $i=0; $i <$ct_children; $i++) {
		$i and ($tree .= ', ') ; 
		$tree .=  ${$std_children[$i]->[1]};
	    }

	    $tree .= " )" ;

		# a tentative revision on Nov 7 (Sun), 2004 #
	     $option_nodeid and ($tree .= "{$node_id}");
	     

	} else { # the $node_id is probably an external node #

	    my $nodename = $id2name->{$node_id};
	    unless (defined $nodename) {
		print STDERR "NO NAME for the external node ID $node_id.\n";
		$nodename = $node_id;
	    }
	    $tree .= ($option_nodeid ? ("$nodename" . "{$node_id}") : $nodename ) ;
	}

	my $upper_branch = $branch_attributes->{$node_id};

	if (defined $upper_branch) {
	    my ($br_len, $boot) = @{$upper_branch}[$INDX_LENGTH, $INDX_BOOT];
	    if (defined $br_len) {
		if ($option_bootlabel) {
		    $tree .= (defined $boot) ? (":$br_len" . "[$boot]") : ":$br_len";
		} else {
		    $tree .= (defined $boot) ? "$boot:$br_len" : ":$br_len";
		}
	    }
	}

	return \$tree;

}		# end of the subroutine draw_subtree #


# br_output_tree:
#	This subroutine outputs a part of a tree under a specified node
#	into a filehandle.
#	It outputs the daughter nodes just in the order they are stored in the
#	array @{$node2children->{$node_id}}.
#
# INPUT: ($node_id, $node2children, $id2name, $branch_attributes, $filehandle,
#	$option_bootlabel, $option_nodeid),
#	where $node_id is the ID of the node which is the top of the part of the tree being output,
#	$option_bootlabel = 1 if you dsiplay  bootstrap at the branch, =0 (or undef) at the node.
#	$option_nodeid = 1 if you display node ID, =0 (or undef) otherwise.
#
sub br_output_tree ($\%\%\%\*$$) {

	my ($node_id, $node2children, $id2name, $branch_attributes, $fh,
		$option_bootlabel, $option_nodeid) = @_;

	my $tree = br_draw_subtree ($node_id, %{$node2children}, %{$id2name},
		%{$branch_attributes}, $option_bootlabel, $option_nodeid);

		# when the node is an external node #
	(defined $node2children->{$node_id}) or ( ${$tree} =  "( ${$tree} )" ) ;

	print $fh ${$tree} . "\;\n" ;

	return 1;

}		# end of the subroutine 'br_output_tree' #


# br_draw_subtree:
#	This subroutine outputs a part of a tree under a specified node
#	in a string of the Newick format.
#	It outputs the daughter nodes in the order
#	stored in the array @{$node2children->{$node_id}}.
#
# INPUT: ($node_id, $node2children, $id2name, $branch_attributes,
#	$option_bootlabel, $option_nodeid),
#	where $node_id is the ID of the node which is the top of
#	the part of the tree being output,
#	$option_bootlabel = 1 if you display bootstrap value at the branch,
#			=0 (or undef) at the node. 
#	$option_nodeid = 1 if you display node ID, =0 (or undef) otherwise.
#
sub br_draw_subtree ($\%\%\%$$) {

	my ($node_id, $node2children, $id2name, $branch_attributes,
		$option_bootlabel, $option_nodeid) = @_;

	my $tree = '';

	my $children = $node2children->{$node_id};
	if (defined $children) {

	    $tree .= "( ";

	    my $ct_children = @{$children};
	    for (my $i=0; $i <$ct_children; $i++) {

		$i and ($tree .= ', ') ; 

		my $child = $children->[$i];
		my $subtree =
		br_draw_subtree ($child, %{$node2children}, %{$id2name},
		   %{$branch_attributes}, $option_bootlabel, $option_nodeid);
		$tree .=  ${$subtree};
	    }

	    $tree .= " )" ;

		# a tentative revision on Nov 7 (Sun), 2004 #
	     $option_nodeid and ($tree .= "{$node_id}");
	     

	} else { # the $node_id is probably an external node #

	    my $nodename = $id2name->{$node_id};
	    unless (defined $nodename) {
		print STDERR "[WANING 'br_draw_subtree'] NO NAME for the external node ID $node_id.\n";
		$nodename = $node_id;
	    }
	    $tree .= ($option_nodeid ? ("$nodename" . "{$node_id}") : $nodename ) ;
	}

	my $upper_branch = $branch_attributes->{$node_id};

	if (defined $upper_branch) {
	    my ($br_len, $boot) = @{$upper_branch}[$INDX_LENGTH, $INDX_BOOT];

		## REVISED on Thursday, Sep 21, 2006 ##
	    if ((defined $br_len) or (defined $boot)) {
		(defined $br_len) or ($br_len = '');
		if ($option_bootlabel) {
		    $tree .= (defined $boot) ? (":$br_len" . "[$boot]") : ":$br_len";
		} else {
		    $tree .= (defined $boot) ? "$boot:$br_len" : ":$br_len";
		}
	    }
	}

	return \$tree;

}		# end of the subroutine br_draw_subtree #


# br_output_tree_sep:
#	A revised version of 'br_output_tree' with the freedom of choice
#	of the 'separator' to delimit the nodes in the tree.
#
# br_output_tree:
#	This subroutine outputs a part of a tree under a specified node
#	into a filehandle.
#	It outputs the daughter nodes just in the ordered in the
#	array @{$node2children->{$node_id}}.
#
# INPUT: ($node_id, $node2children, $id2name, $branch_attributes, $filehandle,
#	$option_bootlabel, $option_nodeid, $separator),
#	where $node_id is the ID of the node which is the top of the part of the tree being output,
#	$option_bootlabel = 1 if you dsiplay  bootstrap at the branch, =0 (or undef) at the node.
#	$option_nodeid = 1 if you display node ID, =0 (or undef) otherwise.
#
sub br_output_tree_sep ($\%\%\%\*$$$) {

	my ($node_id, $node2children, $id2name, $branch_attributes, $fh,
		$option_bootlabel, $option_nodeid, $separator) = @_;

	my $tree = br_draw_subtree_sep ($node_id, %{$node2children}, %{$id2name},
		%{$branch_attributes}, $option_bootlabel, $option_nodeid, $separator);

		# when the node is an external node #
	(defined $node2children->{$node_id}) or ( ${$tree} =  "($separator" . ${$tree} . "$separator)" ) ;

	print $fh ${$tree} . "\;\n" ;

	return 1;

}		# end of the subroutine 'br_output_tree_sep' #


# br_draw_subtree_sep:
#	A revised version of 'br_draw_subtree' with a freedom of choice
#	of the 'sepatator' that delimits the nodes in the output tree.
#
# br_draw_subtree:
#	This subroutine outputs a part of a tree under a specified node
#	in a string of the Newick format.
#	It outputs the daughter nodes in the order
#	stored in the array @{$node2children->{$node_id}}.
#
# INPUT: ($node_id, $node2children, $id2name, $branch_attributes,
#	$option_bootlabel, $option_nodeid, $separator),
#	where $node_id is the ID of the node which is the top of
#	the part of the tree being output,
#	$option_bootlabel = 1 if you display bootstrap value at the branch,
#			=0 (or undef) at the node. 
#	$option_nodeid = 1 if you display node ID, =0 (or undef) otherwise.
#	$separator specifies the separator that delimits the nodes in the output tree.
#
sub br_draw_subtree_sep ($\%\%\%$$$) {

	my ($node_id, $node2children, $id2name, $branch_attributes,
		$option_bootlabel, $option_nodeid, $separator) = @_;

	my $tree = '';

	my $children = $node2children->{$node_id};
	if (defined $children) {

	    $tree .= "($separator";

	    my $ct_children = @{$children};
	    for (my $i=0; $i <$ct_children; $i++) {

		$i and ($tree .= ',' . $separator) ; 

		my $child = $children->[$i];
		my $subtree =
		br_draw_subtree_sep ($child, %{$node2children}, %{$id2name},
		   %{$branch_attributes}, $option_bootlabel, $option_nodeid, $separator);
		$tree .=  ${$subtree};
	    }

	    $tree .= "$separator)" ;

		# a tentative revision on Nov 7 (Sun), 2004 #
	     $option_nodeid and ($tree .= "{$node_id}");
	     

	} else { # the $node_id is probably an external node #

	    my $nodename = $id2name->{$node_id};
	    unless (defined $nodename) {
		print STDERR "[WANING 'br_draw_subtree_sep'] NO NAME for the external node ID $node_id.\n";
		$nodename = $node_id;
	    }
	    $tree .= ($option_nodeid ? ("$nodename" . "{$node_id}") : $nodename ) ;
	}

	my $upper_branch = $branch_attributes->{$node_id};

	if (defined $upper_branch) {
	    my ($br_len, $boot) = @{$upper_branch}[$INDX_LENGTH, $INDX_BOOT];

			## REVISED  on Thursday, Sep 21, 2006 ##
	    if ((defined $br_len) or (defined $boot)) {
		(defined $br_len) or ($br_len = '');
		if ($option_bootlabel) {
		    $tree .= (defined $boot) ? (":$br_len" . "[$boot]") : ":$br_len";
		} else {
		    $tree .= (defined $boot) ? "$boot:$br_len" : ":$br_len";
		}
	    }
	}

	return \$tree;

}		# end of the subroutine 'br_draw_subtree_sep' #


sub topo_subtree ($\%\%\%);

# topo_subtree: This subroutine returns the reference to  a string
#		which shows a simple tree in which each sequence name
#		is replaced by a  simple code indicating its source species.
#		It outputs the daughter nodes in the ascinding order of the sizes of subtrees under them.
#		So the tree gets more and more complex as it moves from left to right.
#
# INPUT: ($node_id, \%node2children, \%id2name, \%name2code),
#
sub topo_subtree ($\%\%\%) {

	my ($node_id, $node2children, $id2name, $name2code) = @_;

	my $topo = '';

	my $children = $node2children->{$node_id};
	if (defined $children) {

	    $topo .= "(";

	    my @children_info = (); # element = [node_ct, \$subtopo];
	    foreach my $child (@{$children}) {
		my $nodes = fetch_all_offsprings ($child, %{$node2children});
		my $node_ct = (defined $nodes) ? @{$nodes} : 0;
		my $subtopo =
		topo_subtree ($child, %{$node2children}, %{$id2name}, %{$name2code});
			push @children_info , [$node_ct, $subtopo];
	    }

	    my @std_children = sort { $a->[0] <=> $b->[0] } @children_info;
	    my $ct_children = @std_children;
	    for (my $i=0; $i <$ct_children; $i++) {
		$i and ($topo .= ',') ; 
		$topo .=  ${$std_children[$i]->[1]};
	    }

	    $topo .= ")" ;

	} else { # the $node_id is probably an external node #

	    my $nodename = $id2name->{$node_id};
	    unless (defined $nodename) {
		print STDERR "NO NAME for the external node ID $node_id.\n";
		$nodename = $node_id;
	    }
	    my $code  = $name2code->{$nodename};
	    unless (defined $code) {
		print STDERR "NO CODE for the external node ID $node_id.\n";
		$code = $node_id;
	    }
	    $topo .= $code;
	}

	return \$topo;

}		# end of the subroutine topo_subtree #


# make_node2children:
#	This subroutine makes %node2children from %node2parent.
#
# INPUT: \%node2parent, where
#	%node2parent = (node_ID => parent_ID, ...),
#	with top_node_ID => undef.
#
# OUTPUT: \%node2children, where
#	%node2children = (node_ID => [a list of child ID's], ...),
#	with external_node_ID => undef.
#
sub make_node2children (\%) {

	my ($node2parent) = @_;

	my %node2children;
	my $topnode;
	foreach my $node (keys %{$node2parent}) { # The main foreach-loop #
	    my $parent = $node2parent->{$node};
	    unless (defined $parent) {
		if ((defined $topnode) and ($topnode ne $node)) {
		   print STDERR "[CAUTION 'make_node2children'] The top node candidate '$node' CLASHES with the existing candidate '$topnode'!!\n";
		}
		$topnode = $node;
		next;
	    }

	    if (defined $node2children{$parent}) {
		push @{$node2children{$parent}}, $node;
	    } else {
		$node2children{$parent} = [$node];
	    }
	}				# End of the main foreach-loop #

	return \%node2children;

}		# end of the subroutine 'make_node2children' #


# make_node2parent:
#	This subroutine makes %node2parent from %node2children.
#
# INPUT: \%node2children, where
#	%node2children = (node_ID => [a list of child ID's], ...),
#	with external_node_ID => undef.
#
# OUTPUT: \%node2parent, where
#	%node2parent = (node_ID => parent_ID, ...),
#	with top_node_ID => undef.
#
sub make_node2parent (\%) {

	my ($node2children) = @_;

	my %node2parent;

	foreach my $node (keys %{$node2children}) { # The outer foreach-loop #

	    my $children = $node2children->{$node};
	    (defined $children) or next;

	     foreach my $child (@{$children}) { # The inner foreach-loop #
		if ((defined $node2parent{$child}) and ($node2parent{$child} ne $node)) {
		    print STDERR "[ERROR 'make_node2parent'] The node '$child' has CONFLICTING PARENTS:  $node vs ", $node2parent{$child}, " !\n\n";
		    return undef;
		}
		$node2parent{$child} = $node;
	     }			 # End of the inner foreach-loop #

	}				# End of the outer foreach-loop #

	return \%node2parent;

}		# end of the subroutine 'make_node2parent' #


# shift_topnode: This subroutine replaces the topnode with a specified daughter nodes,
#		and revises all the hashes except %id2name.
#
# NOTE: This version does NOT allow an external node to be the top node.
#
sub shift_topnode ($\%\%\%\%) {

	my ($new_top_node, $node2parent, $node2children, $name2id, $branch_attributes) = @_;

	my $old_top_node = $name2id->{$TOP_NODE };

	if ($old_top_node == $new_top_node) {
		print STDERR "The input of 'shift_topnode' (node ID $new_top_node) is the current top node itself !\n";
		return 0;
	}

	unless ($node2parent->{$new_top_node} == $old_top_node ) {
		print STDERR "The input of 'shift_topnode' (node ID $new_top_node) is NOT a daughter of the current top node.\n";
		return 0;
	}

	my $children_newtop = $node2children->{$new_top_node};
	my $children_oldtop = $node2children->{$old_top_node};

	unless (defined $children_newtop) {
		print STDERR "The input of 'shift_topnode' (node ID $new_top_node) is an EXTERNAL node.\n";
		return 0;
	}

		## now we actually replace the current top node with the new one.

		# add ex-old topnode into the list of daughters of the new topnode #
	##$children_newtop = union_array (@{$children_newtop}, @{[$old_top_node]});
	push @{$children_newtop}, $old_top_node; # Revised on Wednesday, March 29, 2006
	$node2children->{$new_top_node} = $children_newtop;

		# remove the new topnode from the list of daughters of the ex-topnode #
		# Revised on Wednesday, March 29, 2006 #
	my @new_children_oldtop = ();
	foreach my $child (@{$children_oldtop}) {
	    ($child == $new_top_node) or (push @new_children_oldtop, $child);
	}
	$node2children->{$old_top_node} = \@new_children_oldtop;
	##(undef, $children_oldtop, undef) = diff_array (@{$children_oldtop}, @{[$new_top_node]}); 
	##$node2children->{$old_top_node} = $children_oldtop;

		# swap the parent-child relationship  #

		## REVISED on May 7 (Mon), 2007 ##
	delete $node2parent->{$new_top_node};	# now $new_top_node is the top node
	##undef $node2parent->{$new_top_node};	# now $new_top_node is the top node
	$node2parent->{$old_top_node} = $new_top_node;

		# switch the holder of the label $TOP_NODE #
	$name2id->{$TOP_NODE} = $new_top_node;

		# transfer the location of the attributes on an top branch #
	if (defined $branch_attributes->{$new_top_node}) {
	    my @attributes = @{$branch_attributes->{$new_top_node}};
	    $branch_attributes->{$old_top_node} = \@attributes;
	    undef $branch_attributes->{$new_top_node} ;
	}

	return 1;

}		# end of the subroutine shift_topnode #


###=for the moment, comment out the following subroutines in order to test the script little by little

# move_topnode: This subroutine recieves the ID of the node which you want the top node to move to,
#		and iterates 'shift_topnode' untill the top node moves to the specified node.
#
sub move_topnode ($\%\%\%\%) {

	my ($new_top_node, $node2parent, $node2children, $name2id, $branch_attributes) = @_;

	my $old_top_node = $name2id->{$TOP_NODE };

	if ($old_top_node == $new_top_node) {
		#print STDERR "The input of 'move_topnode' (node ID $new_top_node) is the current top node itself !\n";
		return 1;
	}


	unless (defined $node2children->{$new_top_node} ) {
		print STDERR "The input of 'move_topnode' (node ID $new_top_node) is an EXTERNAL node.\n";
		return 0;
	}

	my $ancestry = fetch_ancestry ($new_top_node, %{$node2parent});
	# = ($id_topnode, ..., $new_top_node)

	if (@{$ancestry} < 2) { # It canNOT happen in a normal situation.
	     print STDERR "Illegal input for 'move_topnode' (node_ID: $new_top_node, ancestry: " .
			join (', ' , @{$ancestry}) . ") . \n";
		return 0;
	}

		## now we actually replace the current top node with the new one.

	shift @{$ancestry}; # remove the top node

	while (my $node = shift @{$ancestry}) {

	    unless ( shift_topnode ($node, %{$node2parent}, %{$node2children},
				%{$name2id}, %{$branch_attributes} ) ) {
		print STDERR "'shift_topnode' failed at an ancestor (ID: $node) of the destination (ID: $new_top_node) .\n";
		return 0;
	    }
	}		# end of the while-loop # 

	return 1;

}		# end of the subroutine move_topnode #


###=for the moment, comment out the following subroutines in order to test the script little by little

# segregate_subset_nodes: This subroutine receives a subset of node IDs
#	 and tries to separate them from as many other nodes as possible
#	by confining them under one of the branches
#	stretching out of the top node.
#
sub segregate_subset_nodes (\@\%\%\%\%) {

	my ($subset, $node2parent, $node2children, $name2id, $branch_attributes) = @_;

	unless (@{$subset}) {
		print STDERR "There is NO NODE in the input subset of 'segregate_subset_nodes', so beware !!\n";
		return 0;
	}

	my ($id_lca, $depth) = fetch_last_common_ancestor_ov2 (@{$subset}, %{$node2parent});
	unless (defined $id_lca) {
		print STDERR "'fetch_last_common_ancestor_ov2' failed for the set of nodes:\n";
		print STDERR  ' ' x 3 . '{' . join (', ' , @{$subset}) . '}' . ".\n\n";    
		return 0;
	}

	my $curr_top_node = $name2id->{$TOP_NODE};
	unless (defined $curr_top_node) {
		print STDERR "NO TOP NODE for the input tree of 'segregate_subset_nodes' !!\n";
		return 0;
	}

	my $new_top_node = $node2parent->{$id_lca};
	if (defined $new_top_node) { # the subset is already confined under a branch # 
	    ($new_top_node == $curr_top_node) and (return 1);

	     if (move_topnode ($new_top_node, %{$node2parent}, %{$node2children},
				%{$name2id}, %{$branch_attributes} ) ) {
		return 1;
	     } else {
		print STDERR "'move_topnode' failed from the node ID $curr_top_node to the node ID $new_top_node.\n";
		return 0;
	    }
	}

		# the subset is dispersed under several branches #

		# Next try using the complement set #

	my $all_extnodes = fetch_ext_offsprings ($curr_top_node, %{$node2children});
	(undef, undef, my $complement) = diff_array (@{$subset}, @{$all_extnodes});

	unless (@{$complement}) {
		print STDERR "The subset fed into 'segregate_subset_nodes' was actually the whole set of external nodes. So I'll stop here.\n";
		return 0;
	}

	($id_lca, $depth) = fetch_last_common_ancestor_ov2 (@{$complement}, %{$node2parent});
	unless (defined $id_lca) {
		print STDERR "'fetch_last_common_ancestor_ov2' failed for the set of nodes:\n";
		print STDERR  ' ' x 3 . '{' . join (', ' , @{$complement}) . '}' . ".\n\n";    
		return 0;
	}

	$new_top_node = $id_lca;
	(defined $node2children->{$new_top_node}) or ($new_top_node = $node2parent->{$new_top_node}); # avoid an external branch
	if ( (defined $new_top_node) and ($new_top_node != $curr_top_node ) ) {
		# the complement is already confined under a branch # 
	     if (move_topnode ($new_top_node, %{$node2parent}, %{$node2children},
				%{$name2id}, %{$branch_attributes} ) ) {
		return 1;
	     } else {
		print STDERR "'move_topnode' failed from the node ID $curr_top_node to the node ID $new_top_node.\n";
		return 0;
	    }
	}

		# the complement set is also dispersed #

	### if there is any child node under which there are no nodes in the subset,
	### we will take the node with the most complements.
	# --> Now revised so that we will take the internal node
	# whose set of offsprings is the largest
	# among such sets that do not overlap the subset.
	# When there are more than 1 such node,
	# we will employ whatever comes first.

	##my $top_children = $node2children->{$curr_top_node};
		# Take the set of all internal offsprings of the top-node.
	my $allnodes = fetch_all_offsprings  ($curr_top_node, %{$node2children});
	push @{$all_extnodes}, $curr_top_node;
	(undef, undef, my $all_intoffs) = diff_array (@{$all_extnodes}, @{$allnodes});

	my $candidate_node ; 
	my $max_comple_ct = 0;
	my @std_subset = sort @{$subset};

	##foreach my $node (@{$top_children}) {
	foreach my $node (@{$all_intoffs}) {
	     my $extnodes = fetch_ext_offsprings ($node , %{$node2children});
	     my ($common, $extra, $comple) = diff_array (@std_subset, @{$extnodes});
	     if (!(@{$common}) and  (@{$comple} > $max_comple_ct) ) {
		$candidate_node = $node;
		$max_comple_ct = @{$comple};
	     }
	}

	(defined $candidate_node) or (return 1);  # do nothing

		#  take its parent node if it exists #
	(defined $node2parent->{$candidate_node})
		and ($candidate_node = $node2parent->{$candidate_node});

	if ( move_topnode ($candidate_node, %{$node2parent}, %{$node2children},
			%{$name2id}, %{$branch_attributes}) ) {
		return 1;
	} else {
		print STDERR "'move_topnode' failed from the node ID $curr_top_node to the node ID $candidate_node.\n";
		return 0;
	}

} 		# end of the subroutine segregate_subset_nodes #

###=for the moment, comment out the following subroutines in order to test the script little by little

# sum_branch_lenghts:  This subroutine calculats the summation of a set of branche lengths.
#
# INPUT: (\@node_ids, \%branch_attributes, $indx_len), where
#	$branch_attributes->{node_ID}[$indx_len]
#		= {the length of the branch which is the immediate upstream of the node}
#
sub sum_branch_lenghts (\@\%$) {

	my ($node_ids, $br_attr, $indx_len) = @_;

	( (defined $indx_len) and ($indx_len >= 0) ) or ($indx_len = $INDX_LENGTH) ;
	my $ct_br = @{$node_ids};

	my $distance = 0;

	for (my $i=0; $i< $ct_br; $i++) {
	    my $nodeid = $node_ids->[$i];
	    my $br_info = $br_attr->{$nodeid};

	    unless (defined $br_info) {
		print STDERR "Illegal input for 'sum_branch_lenghts':\n";
		print STDERR "   node ID $nodeid included in the set (";
		print STDERR join ( ', ', @{$node_ids} )  . ") .\n";
		return undef;
	    }
	    my $brlen = $br_info->[$indx_len];
	    unless (defined $brlen) {
		print STDERR "UNASSIGNED branch length for the branch (node) ID $nodeid (maybe the index $indx_len is wrong).\n"; 
		return undef;
	    }

	    $distance += $brlen;
	}

	return $distance;

}		# end of the subroutine sum_branch_lenghts #


# intra_lineage_distance: This  subroutine calculates the distance between 2 nodes
#		which are known to be ON THE SAME LINEAGE.
#
# INPUT: (node_offspring, node_ancestor, \%node2parent, \%branch_attributes, $indx_len)
#
sub intra_lineage_distance ($$\%\%$) {

	my ($offspring, $ancestor, $node2parent, $br_attr, $indx_len) = @_;

	( (defined $indx_len) and ($indx_len >= 0) ) or ($indx_len = $INDX_LENGTH) ;

	my $pres_node = $offspring;
	my $distance = 0;

	while (defined (my $parent = $node2parent->{$pres_node}) ) {
				# quit when you reach the top-node

	    ($pres_node  == $ancestor) and last; # Do NOT add branch lengths above $ancestor

	    my $br_info = $br_attr->{$pres_node};
	    unless (defined $br_info) {
		print STDERR "NO BRANCH INFORMATION on the branch (node) ID $pres_node on the lineage leading to $offspring.\n";
		return undef;
	    }
	    my $brlen = $br_info->[$indx_len];
	    unless (defined $brlen) {
		print STDERR "UNASSIGNED branch length for the branch (node) ID $pres_node (maybe the index $indx_len is wrong).\n"; 
		return undef;
	    }
	    $distance += $brlen;
	    $pres_node = $parent;

	}			# end of the while-loop #

	unless ($pres_node == $ancestor) {
		print STDERR "CanNOT find the node $ancestor on the lineage leading to the node $offspring.\n";
		return undef;
	}

	return $distance;

}		# end of the subroutine intra_lineage_distance #


# inter_node_distance: This subroutine calculates the distance between 2 nodes along the
#	phylogenetic tree.
#
# OUTPUT: (the distance between node1 and node2,
#		the total branch length leading from the lca to node1, 
#		the total branch length leading from the lca to node2,
#		ID of the last common ancestor (LCA), depth of the LCA).
#
sub inter_node_distance ($$\%\%$) {

	my ($node1, $node2, $node2parent, $br_attr, $indx_len) = @_;

	my @null = (undef, undef, undef, undef, undef);
	my ($common_anc, $lin1, $lin2) = fetch_common_ancestors ($node1, $node2, %{$node2parent});
	unless ( (defined $common_anc) and @{$common_anc} ) {
	    print STDERR "'fetch_common_ancestors' failed for the node pair ($node1, $node2).\n";
	    return (undef, undef, undef, undef, undef) ;
	}

	my $depth = @{$common_anc} -1;
	my $lca = $common_anc->[$depth];

	my $len_lin1 = sum_branch_lenghts (@{$lin1}, %{$br_attr}, $indx_len);
	unless (defined $len_lin1) {
	    print STDERR "\'sum_branch_lenghts\' failed for the node $node1 -specific ancestry:";
	    print STDERR  join (' -> ', @{$lin1} ) . " .\n";
	    return  (undef, undef, undef, $lca, $depth);
	}

	my $len_lin2 = sum_branch_lenghts (@{$lin2}, %{$br_attr}, $indx_len);
	unless (defined $len_lin2) {
	    print STDERR "\'sum_branch_lenghts\' failed for the node $node2 -specific ancestry:";
	    print STDERR  join (' -> ', @{$lin2} ) . " .\n";
	    return (undef, undef, undef, $lca, $depth);
	}

	return ($len_lin1 + $len_lin2, $len_lin1, $len_lin2, $lca, $depth) ;

}		# end of the subroutine inter_node_distance #


# fetch_boot_series: This  subroutine fetches a series of bootstrap values
#		for branches between 2 nodes
#		which are known to be ON THE SAME LINEAGE.
#
# INPUT: (node_offspring, node_ancestor, \%node2parent, \%branch_attributes, $indx_boot)
# OUTPUT: (bootstrap value for the branch immediate downstream of the ancestor,
#	 ..., bootstrap value for the branch immediate upstream of the offspring or its parent).
#
sub fetch_boot_series ($$\%\%$) {

	my ($offspring, $ancestor, $node2parent, $br_attr, $indx_boot) = @_;

	( (defined $indx_boot) and ($indx_boot >= 0) ) or ($indx_boot = $INDX_BOOT) ;

	my $pres_node = $offspring;
	my @boot_series = ();

	while (defined (my $parent = $node2parent->{$pres_node}) ) {
				# quit when you reach the top-node

	    ($pres_node  == $ancestor) and last; # Do NOT collect bootstrap values above $ancestor

	    my $br_info = $br_attr->{$pres_node};
	    unless (defined $br_info) {
		print STDERR "NO BRANCH INFORMATION on the branch (node) ID $pres_node on the lineage leading to $offspring.\n";
		return undef;
	    }
	    my $boot = $br_info->[$indx_boot];
	    (defined $boot) and (unshift @boot_series, $boot);
	    $pres_node = $parent;

	}			# end of the while-loop #

	unless ($pres_node == $ancestor) {
		print STDERR "CanNOT find the node $ancestor on the lineage leading to the node $offspring.\n";
		return undef;
	}

	return \@boot_series;

}		# end of the subroutine fetch_boot_series #


# extract_hidden_bootstrap: This subroutine receives 2 children of the top-node,
#		and tries to fetch a 'hidden' bootstrap value that those 2 children
#		are separated from the remaining child.
#
# INPUT: (top_node, child1, child2, \%node2children, \%branch_attributes, $indx_boot)
#
# OUTPUT: (hidden_bootstrap_value, node ID of the remaining child)
#
# PREREQUISITE: The top-node has 3 child nodes.
#
sub extract_hidden_bootstrap ($$$\%\%$) {

	my ($top_node, $child1, $child2, $node2children, $br_attr, $indx_boot) = @_;

	( (defined $indx_boot) and ($indx_boot >= 0) ) or ($indx_boot = $INDX_BOOT);

	my $children = $node2children->{$top_node};
	unless (defined $children) {
	    print STDERR "NO CHILDREN INFORMATION on the 'TOP-NODE' $top_node.\n";
	    return (undef, undef);
	}

	foreach my $child (@{$children}) {
	    ( ($child  == $child1)  or ($child == $child2) ) and next;
	    my $br_info = $br_attr->{$child};
	    unless ($br_info) {
		print STDERR "NO BRANCH ATTRIBUTES for the 'remaining' sup-top branch (node) ID $child.\n";
		return (undef, undef);
	    }

	    my $hidden_boot = $br_info->[$indx_boot];

	    return ($hidden_boot, $child);
	}

	print STDERR "NO 'REMAINING CHILD' found for the top_node ID $top_node.\n";
	return (undef, undef);

}		# end of the subroutine extract_hidden_bootstrap #


# extract_hidden_bootstrap2:
#	This revised version of the subroutine 'extract_hidden_bootstrap'
#	receives multiple children of the top-node,
#	and tries to fetch 'hidden' bootstrap values that those
#	multiple children are separated from the remaining ones.
#	It no longer require the top node to have exactly 3 children.
#
# INPUT: (top_node, @sb_children, \%node2children, \%branch_attributes, $indx_boot)
#	@sb_children is a subset of the children of the top node.
#
# OUTPUT: \@hidden_boot_lca, where
#	@{$hidden_boot_lca[$m]}
#	= (node ID of the $m th child in the remainder,
#	  the hidden_bootstrap_value associated with it).
#
sub extract_hidden_bootstrap2 ($\@\%\%$) {

	my ($top_node, $sb_children, $node2children,
		$br_attr, $indx_boot) = @_;

	( (defined $indx_boot) and ($indx_boot >= 0) ) or ($indx_boot = $INDX_BOOT);

	my $children = $node2children->{$top_node};
	unless (defined $children) {
	    print STDERR "NO CHILDREN INFORMATION on the 'TOP-NODE' $top_node.\n";
	    return (undef, undef);
	}

	(undef, undef, my $cmpl_children) = diff_array (@{$sb_children}, @{$children});
	my @hidden_boot_lca = ();

	foreach my $child (@{$cmpl_children}) {
	    my $br_info = $br_attr->{$child};
	    unless ($br_info) {
		print STDERR "[CAUTION 'extract_hidden_bootstrap2'] NO BRANCH ATTRIBUTES for the 'remaining' sup-top branch (node) ID $child.\n";
		return undef;
	    }

	    my $hidden_boot = $br_info->[$indx_boot];

	    push @hidden_boot_lca, [$child, $hidden_boot];
	}

	return \@hidden_boot_lca;

}		# end of the subroutine extract_hidden_bootstrap2 #


# resolve_multipartite:
#	This subroutine 'resolve's a 'multipartite' node,
#	 which has more than 2 children, by introducing a mediating node
#	with the length 0 and the bootstrap value undefined.
#
# INPUT: (\%node2parent, \%node2children, %branch_attributes,
#	$multipartite_node_id, $mediating_node_id, \@subset_children),
#	where @subset_children contains children which are going to cluster together
#	after the 'resolution'.
#
sub resolve_multipartite (\%\%\%$$\@) {

	my ($node2parent, $node2children, $br_attr,
		$mltpt_node_id, $med_node_id, $sub_kids) = @_;

	my $children = $node2children->{$mltpt_node_id};

	unless  (defined $children) {
	    print STDERR "NO INFORMATION on the children of the input node $mltpt_node_id for 'resolve_multipartite'.\n";
	    return 0;
	}

	if (@{$children} < 3) {
	    print STDERR "The input node $mltpt_node_id is already resolved enough.\n";
	    return 1;
	}

	if ( (defined $node2parent->{$med_node_id})  or (defined $node2children->{$med_node_id}) ) {
	    print STDERR "The ID number $med_node_id you specified for the meidating node already exists. I'm cowardly quitting.\n";
	    return 0;
	}

	(undef, undef, my $cmpl_kids) = diff_array (@{$sub_kids}, @{$children}) ;
	(@{$sub_kids} < 2) and ( ($cmpl_kids, $sub_kids) = ($sub_kids, $cmpl_kids) );

		# re-definition of parent nodes #
	foreach my $node (@{$sub_kids}) { $node2parent->{$node} = $med_node_id; }
	$node2parent->{$med_node_id} = $mltpt_node_id;

		# re-definition of child nodes #
	$node2children->{$med_node_id} = $sub_kids ;

	my @new_children = sort (@{$cmpl_kids}, $med_node_id );
	$node2children->{$mltpt_node_id} = \@new_children;

		# attach the length 0 and the bootstrap value undef to the newly inserted branch #
		## it has to change in coordination with the input/output subroutines ##

	$br_attr->{$med_node_id} = [ sprintf ("%7.5f", 0 ) , undef];

	return 1;

}		# end of the subroutine resolve_multipartite #


# examine_multifurcation:
#	This subroutine examines a tree to look for multifurcated nodes.
#
# INPUT: (\%node2chi, $topnode_id), where
#	%node2chi = ($node_id => \@children, ...) for the interior nodes;
#	$topnode_id can be different from the genuine top node.
#	In that case, the subroutine looks for the sub-tree under the node $topnode_id.
#
# OUTPUT: \@multifurcations, which stores IDs of the multifurcating nodes.
#
sub examine_multifurcation (\%$) {

	my ($node2chi, $topnode_id) = @_;

	my @multifurcations = ();

	my @curr_nodes = ($topnode_id);

	my $fl_top = 1;
	while (@curr_nodes>0) { # The outer while-loop #
	    my @next_nodes = ();
	    while (my $nodeid = shift @curr_nodes) { # The inner while-loop #
		my $children = $node2chi->{$nodeid};
		((defined $children) and (@{$children}>0)) or next;

		(@{$children} > 2 + $fl_top) and (push @multifurcations, $nodeid);

		push @next_nodes, $children;
	    }			# End of the inner while-loop #
	    $fl_top = 0;
	    @curr_nodes = @next_nodes; # update the list of nodes #
	}		# End of the outer while-loop #

	return \@multifurcations;

}		# End of the subroutine 'examine_multifurcation' #


=for the moment, comment out the following subroutines in order to test the script little by little

=cut

1;		# a "spell" which makes the module "use"-able and/or "require"-able #

