package MyTreeManipu_IV;
#
###################################################################################
#	DENSERM_P: A package to DEtect Negative SElection on Recurrent Mutations
#			(almost exclusively via Perl modules and scripts).
#	Version 0.3: Copyright (C) 2012 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyTreeManipu_IV.pm:
#	This package manipulates trees more actively than the preceding
#	members of the family "MyTreeManipu_xxx.pm".
#	For example, it can even change the topology of a tree,
#	compare to similar trees, and so on.
#
# MyTreeManipu_II.pm:
#	This package can process more general types of trees,
#	(i.e. NOT limited to binary trees with trichotomous top nodes).
#	It is also consistent with the format of the taxonomy tree which are handled by
#	the module 'MyTaxonomy.pm'.
#

	### 5 important hash tables that are used in this Module
	### in order to describe a tree
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.

use strict;
use warnings;
use MyArrayManipu qw( diff_array union_array );
use MyFile_manipu qw( open_input_file open_output_file );

use MyTaxonomy qw(
	fetch_ext_offsprings_nr
	fetch_last_common_ancestor_ov2
		);
	## fetch_ancestry fetch_common_ancestors
 	## fetch_all_offsprings

use MyTreeManipu_II qw(
	input_tree_file read_tree_file br_output_tree br_output_tree_sep
	make_node2children move_topnode
		);
	## input_subtree copy_tree
	## output_tree draw_subtree topo_subtree
	## shift_topnode move_topnode segregate_subset_nodes
	## sum_branch_lengths intra_lineage_distance inter_node_distance
	## fetch_boot_series extract_hidden_bootstrap
	## resolve_multipartite


use Exporter;

our @ISA = qw(Exporter);

##sub rec_fillin_boot (\@\$);
#sub rec_prod_boot (\@\@);


our @EXPORT_OK = qw(
	rm_external_node
	br_contract_branch reduce_tree_by_boot
	bi2multifurcated_top bi_in_multi_out 
	locate_node_in_diffrep
	combine_2attr_on_tree  pr_intg_2attr_on_tree pr_tree_with_trans_boot
	construct_raw_subtree remove_trivial_nodes rm_simple_ancestral_lineages
	construct_neat_subtree
	construct_skeleton_subtree
		) ;


our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
##our $TOP_NODE = 'TOP_NODE';
##our $INDX_LENGTH = 0;
##our $INDX_BOOT = 1;
##our $START_GROUP = '(';	# the symbol marking the start of the grouping of nodes.
##our $END_GROUP = ')';  # the symbol marking the end of the grouping of nodes.


# rm_external_node:
#	This subroutine removes a specified external node from the input tree.
#	If the external node is at the end of a series of unary nodes,
#	the subroutine also removes the unary nodes that are going to be null.
#
# ATTENTION: This subroutine could produce trivial nodes.
#	So I recommend you to apply 'remove_trivial_nodes' after
#	performing this subroutine.
#
# INPUT: ($exnode_id, \%node2parent, \%node2children, \%name2id, \%id2name, %br_attr), where
#	$exnode_id specifies the external node to be removed;
#	%node2parent = ($node_id => $parent_id, ...);
#	%node2children = ($node_id => \@children, ...);
#	%name2id = ($name => $node_id, ...) for the external node and $TOP_NODE;
#	%id2name = ($node_id => $name, ...) for the external node;
#	%br_attr = ($id of the node immediately below the branch => \@attr, ...) with
#	($length, $boot) = @attr[$INDX_LENGTH, $INDX_BOOT].
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
sub rm_external_node ($\%\%\%\%\%) {

	my ($exnode_id, $node2parent, $node2children, $name2id, $id2name,
		$br_attr) = @_;

	if (defined $node2children->{$exnode_id}) {
	    print STDERR "[ERROR 'rm_external_node'] The input node '$exnode_id' is NOT AN EXTERNAL NODE !\n";
	    return 0;
	}

	my $exnode_name = $id2name->{$exnode_id};
	unless (defined $exnode_name) {
	    print STDERR "[CAUTION 'rm_external_node'] The input node '$exnode_id' has NO NAME!\n";
	}

	my $parent = $node2parent->{$exnode_id};
	unless (defined $parent) {
	    print STDERR "[ERROR 'rm_external_node'] NO PARENT for the external node '$exnode_name' (ID: $exnode_id) !\n";
	    return 0;
	}

	(exists $id2name->{$exnode_id}) and (delete $id2name->{$exnode_id});
	(defined $exnode_name) and (exists $name2id->{$exnode_name}) and (delete $name2id->{$exnode_name});

	my $pres = $exnode_id;
	my $siblings = $node2children->{$parent};

		# Remove a series of unary branches laeding to the external node #
	while (@{$siblings}<2) { # the while-loop #

		# Turn the parent node into an external node #
	    delete $node2parent->{$pres};
	    delete $node2children->{$parent};
	    $pres = $parent;

		# Find parent and siblings of the new external node #
	    $parent = $node2parent->{$pres};
	    unless (defined $parent) {
	         print "[WARNING 'rm_external_node'] We have reached the TOP NODE, so I will stop here.\n";
		 return 1;
	    }
	    $siblings = $node2children->{$parent};
	    unless (defined $siblings) {
		print "[WARNING 'rm_external_node'] NO SIBLING for th node '$pres}\ .n";
		return 0;
	    }
	}			# End of the while-loop #

		# Remove the 'root' of the unary series #
	delete $node2parent->{$pres};
	my @new_siblings = ();
	foreach my $sibl (@{$siblings}) { ($sibl eq $pres) or (push @new_siblings, $sibl); }

	$node2children->{$parent} = \@new_siblings;

	return 1;

}		# end of the subroutine 'rm_external_node' #


# br_contract_branch:
#	This subroutine make an internal branch collapse into a point,
#	by removing its 'lower node', and tying new 'parent-children'
#	relationships between its 'upper node' and the children of
#	the 'lower node'.
#	It only takes care of the tree topology,
#	and it doesn't care about the  attributes of the branch.
#	The internal branch is specified by the ID of the lower_node.
#
# INPUT: ($lower_node, \%node2parent, \%node2children).
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
sub br_contract_branch ($\%\%) {

	my ($lower_node, $node2parent, $node2children) = @_;

	my $upper_node = $node2parent->{$lower_node};
	unless (defined $upper_node) {
	    print STDERR "[CAUTION 'br_contract_branch'] The lower node '$lower_node' has NO PARENT (maybe its the Top node), so I cannot find the branch based on the node.\n";
	   return 0;
	}
	
	my $children = $node2children->{$lower_node};
	unless (defined $children) {
	    print STDERR "[CAUTION 'br_contract_branch'] The lower node '$lower_node' has NO CHILDREN (maybe it's an external node), so I cannot contract the branch in question.\n";
	    return 0;
	}

	my $siblings = $node2children->{$upper_node}; 
	unless (defined $siblings) {
	    print STDERR "[CAUTION 'br_contract_branch'] The upper node '$upper_node' has NO CHILDREN, this means some ANOMALY in the stored tree data.\n";
	    return 0;
	}

		# Embrace the children of the lower node
		# as new children of the upper node.
	my @new_siblings = ();
	my $ct_lower = 0;
	foreach my $inode (@{$siblings}) { # The foreach-loop #
	    if ($inode == $lower_node) {
			# replace the lower node with its children #
		$ct_lower++;
		push @new_siblings, @{$children};
	    } else {
			# leave remaining siblings intact #
		push @new_siblings, $inode;
	    }
	}			 # End of the foreach-loop #

	unless ($ct_lower == 1) {
	    print STDERR "[CAUTION 'br_contract_branch'] The children of the upper node contained the lower node '$lower_node' $ct_lower times !\n";
	    ##return 1;
	    return 0;
	}

	$node2children->{$upper_node} = \@new_siblings;

		# Replace the lower node with the upper node
		# as the new parent of the children of the former.
	foreach my $child (@{$children}) { $node2parent->{$child} = $upper_node; }

		# Sever the old relationships,
		# thus putting away the lower node. 
	## undef $node2children->{$lower_node};
	## undef $node2parent->{$lower_node};
	delete $node2children->{$lower_node};
	delete $node2parent->{$lower_node};
		## -- REVISED on May 7 (Mon), 2007 ##

	return 1;

}		# End of the subroutine 'br_contract_branch' #


# reduce_tree_by_boot:
#	This subroutine reduces a tree by collapsing the branches
#	with poor bootstrap supports.
#
# RULE: In principle, the 'poor' branches are collapsed into the upper node,
#	bootstrap values for the branch will just be discarded,
#	and 1/2 of its branch length will be added to the surrounding
#	branches. (--- may be there is a room for improvement)
#
# INPUT: (\%node2parent, \%node2children, %branch_attributes,
#			$thrsh, $indx_boot, $indx_brlen), where
#	($br_len, $boot) = @{$branch_attributes{$nodeid}}[$indx_brlen, $indx_boot]
#	is for the branch directly above the node $nodeid, and
#	the branch will be reduced to a point if $boot < $thrsh.
#
# OUTPUT: 1 if the process succeeded,
#	  0 if it failed.
#
sub reduce_tree_by_boot (\%\%\%$$$) {

	my ($node2parent, $node2children, $br_attr, $thrsh, $indx_boot, $indx_brlen) = @_;

=for debugging the subroutine
	print STDERR "\n< Subroutine 'reduce_tree_by_boot' >\n\n";
	print STDERR "THRSH_BOOT = $thrsh\n\n";
=cut

	foreach my $lower_node (keys %{$br_attr}) { # the outer foreach #
	    my $upper_node = $node2parent->{$lower_node};
	    (defined $upper_node) or next;
	    my $children = $node2children->{$lower_node};
	    (defined $children) or next;
	    my ($br_len, $boot) = @{$br_attr->{$lower_node}}[$indx_brlen, $indx_boot];
	    (defined $boot) or next;
=for debugging the script 
	    print STDERR "The branch $lower_node has $boot bootstap supports and the length = $br_len .\n\n";
=cut
	    ($boot >= $thrsh) and next;
	    my $siblings = $node2children->{$upper_node};
	    my @surround = (@{$children}, $upper_node, @{$siblings});

	    unless ( br_contract_branch ($lower_node, %{$node2parent}, %{$node2children}) ) {
		print STDERR "[CAUTION 'reduce_tree_by_boot'] The subroutine 'br_contract_branch' FAILED for the branch just above '$lower_node'.\n";
		return 0;
	    }
		# add the branch length to those of surronding branches #
	    undef $br_attr->{$lower_node}; 
	    foreach my $branch (@surround) { # the inner foreach #
		my $attr = $br_attr->{$branch};
		(defined $attr) and (defined $attr->[$indx_brlen])
			and ($attr->[$indx_brlen] += $br_len/2) ;
	    }				# end of the inner foreach #

	}				# end of the outer foreach #

	return 1;

}		# end of the subroutine 'reduce_tree_by_br_attr' #


# bi2multifurcated_top:
#	This subroutine converts a bifurcated top node into a multifurcated one
#	by 'splicing out' the branch leading tothe lower node
#	containing the least number of children among the 'branching' lower nodes.
#	If a tie situation occurs, the subroutine chooses the branch
#	that comes first.
#
# INPUT: (\%node2parent, \%node2children, \%name2id, \%br_attr, \@indices_attr), where
#	@indices_attr = ($indx_blen, $indx_boot, $tot_boot), where
#	$br_attri{$node_id}->[$indx_blen] gives the branch length if $indx_blen >=0,
#	$br_attri{$node_id}->[$indx_boot] gives the bootstrap value if $indx_boot >=0.
#	$ct_boot gives the total count of bootstrap replicates.
#	It will take the summation of the lengths of the two branches,
#	and take the product of their bootstrap values, which is afterwards
#	divided by $tot_boot (>0).
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
sub bi2multifurcated_top (\%\%\%\%\@) {

	my ($node2parent, $node2children, $name2id, $br_attr, $indices_attr) = @_;

	my $top_id = $name2id->{$TOP_NODE};
	unless (defined $top_id) {
	    print STDERR "[CAUTION 'bi2multifurcated_top']  NO ID for the TOP NODE.\n"; 
	    return 0;
	}
	my $top_children = $node2children->{$top_id};

	unless (defined $top_children) {
	    print STDERR "[CAUTION 'bi2multifurcated_top'] NO CHILDREN for the TOP NODE (ID: $top_id).\n";
	    return 0;
	}

		# It's already multifurcated,
		# so there is no need to change it.
	(@{$top_children}>2) and (return 1) ;

	my $target; # The child whose upper branch are to be shrinked
	my $minct_gkids = scalar keys %{$name2id};
	foreach my $child (@{$top_children}) {
	    my $gkids = $node2children->{$child};
	    (defined  $gkids) or next; # skip an external node
	    my $ct_gkids = @{$gkids};
	    ($ct_gkids >1)  or next; # skip a non-branching node
	    ($ct_gkids < $minct_gkids) or next; 
	    $target = $child;
	    $minct_gkids = $ct_gkids;
	}

		# It cannot be converted to a multifurcated tree.
	(defined $target) or (return 1); 

		# Convert the attributes #

	my ($indx_blen, $indx_boot, $tot_boot) = @{$indices_attr};

	my $partner ;
	foreach my $child (@{$top_children}) {
	    ($child == $target) or ($partner = $child);
	 }
	my $attr_tgt = $br_attr->{$target};
	my $attr_pt = (defined $partner) ? $br_attr->{$partner} : undef;

	if ( (defined $attr_tgt) and (defined $attr_pt) ) {
	    if ($indx_blen >=0) {
		my $tgt_blen = $attr_tgt->[$indx_blen];
		my $pt_blen = $attr_pt->[$indx_blen];
		if (defined $tgt_blen) {
		    if (defined $pt_blen) {
			$attr_pt->[$indx_blen] += $tgt_blen;
		    } else {
			# evacuate the target branch length #
			$attr_pt->[$indx_blen] = $tgt_blen;
		    }
		}
	    }
	    if (($indx_boot>=0) and ($tot_boot>1) ) {
		my $tgt_boot = $attr_tgt->[$indx_boot];
		my $pt_boot = $attr_pt->[$indx_boot];

			### For debugging ###
		### print STDERR "\nTot_boot = $tot_boot, TGT_boot = $tgt_boot, Pt_boot = $pt_boot.\n\n";
		if (defined $tgt_boot) {
		   if (defined $pt_boot) {
			my $cmp_boot = int ($tgt_boot * $pt_boot /$tot_boot);

			### For debugging ###
			### print STDERR "\nCMP_boot = $cmp_boot .\n\n"; 

			$attr_pt->[$indx_boot] = $cmp_boot;
		   } else {
			# evacuate the target bootstrap value #
			$attr_pt->[$indx_boot] = $tgt_boot;
		   }
		}
	    }

		## ADDED to avoid a confusion (on May 21 (Fri), 2010) ##
	    delete $br_attr->{$target};

		## ADDED to cope with a special situation (on May 21 (Fri), 2010) ##

	} elsif ( (defined $attr_tgt) and !(defined $attr_pt) ) {
		## Transfer all the attributes on the target branch to the partner branch ##
	    $br_attr->{$partner} = $attr_tgt;
	    delete $br_attr->{$target};
	}

		# Convert the topology #
	my $fl_sub = br_contract_branch ($target, %{$node2parent}, %{$node2children});

	unless ($fl_sub) {
	    print STDERR "[CAUTION 'bi2multifurcated_top'] The subroutine 'br_contract_branch' FAILED for the child '$target' of the top node '$top_id'.\n";
	    return 0;
	}

	return 1;

}		# End of the subroutine 'bi2multifurcated_top' #


# bi_in_multi_out:
#	This subroutine reads a tree in a file,
#	converts it to a multifurcated (or unrooted) one if it is bifurcated (or rooted),
#	then outputs it to another file.
#
# INPUT: (\$infile, \$outfile, $option_bootlabel, $option_nodeid, \@indices_attr, $separator), where
#	$option_bootlabel = 1 when displaying bootstrap values at the branches,
#			 = 0 at the nodes;
#	$option_nodeid = 1 when printing out node IDs,
#			= 0 otherwise;
#	@indices_attr = ($indx_blen, $indx_boot, $tot_boot), where
#	$br_attri{$node_id}->[$indx_blen] gives the branch length if $indx_blen >=0,
#	$br_attri{$node_id}->[$indx_boot] gives the bootstrap value if $indx_boot >=0.
#	$tot_boot gives the total count of bootstrap replicates;
#	$separator specifies the separator of the nodes in the output tree.
#	How to convert these attributes is described in 'bi2multifurcated_top'.
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
##sub bi_in_multi_out (\$\$$$\@) {
	##my ($infile, $outfile, $option_bootlabel, $option_nodeid, $indices_attr) = @_;
sub bi_in_multi_out (\$\$$$\@$) {

	my ($infile, $outfile, $option_bootlabel, $option_nodeid,
		$indices_attr, $separator) = @_;
	local $/ = "\n";

		## Store the input tree into a string ##
	my $intree = '';
	my $fh_in = open_input_file (${$infile});
	unless ($fh_in) {
	    print STDERR "[CAUTION 'bi_in_multi_out'] FAILED to OPEN the INPUT FILE '${$infile}'.\n";
	    return 0;
	}
	while (my $in = <$fh_in>) { $intree .= $in; }
	close ($fh_in);

	my %node2parent; my %node2children; my %name2id; my %id2name; my %br_attr;

	my $fl_input = input_tree_file ($intree,
		%node2parent, %node2children, %name2id, %id2name, %br_attr);

	unless ($fl_input) {
	    print STDERR "[CAUTION 'bi_in_multi_out'] The subroutine 'input_tree_file' FAILED for the input file:\n";
	    print STDERR "   '${$infile}'.\n";
	    return 0;
	}

	my $fl_convert = bi2multifurcated_top (%node2parent, %node2children, %name2id,
			%br_attr, @{$indices_attr});
	unless ($fl_convert) {
	    print STDERR "[CAUTION 'bi2multifurcated_top'] The subroutine 'bi2multifurcated_top' FAILED for the input file:\n";
	    print STDERR "   '${$infile}'.\n";
	    return 0;
	}

	my $top_id = $name2id{$TOP_NODE};
	unless (defined $top_id) {
	    print STDERR "[CAUTION 'bi2multifurcated_top'] NO ID of the TOP NODE for the input file:\n";
	    print STDERR "   '${$infile}'.\n";
	    return 0;
	}


	my $fh_out = open_output_file (${$outfile}, 0);
	##my $fl_out =  br_output_tree ($top_id, %node2children, %id2name, %br_attr,
	##		*${$fh_out}, $option_bootlabel, $option_nodeid);
	my $fl_out =  br_output_tree_sep ($top_id, %node2children, %id2name, %br_attr,
			*${$fh_out}, $option_bootlabel, $option_nodeid, $separator);
	close ($fh_out);

	unless ($fl_out) {
	    print STDERR "[CAUTION 'bi2multifurcated_top'] The subroutine 'br_output_tree' FAILED for the input file:\n";
	    print STDERR "   '${$infile}'.\n";
	    return 0;
	}

	return 1;

}		# End of the subroutine 'bi_in_multi_out' #


# locate_node_in_diffrep:
#	This subroutine searches a tree in representation 2
#	for the node that is equivalent to a node of the same tree in representation 1.
#
# INPUT: ($target_node, \%name2id1, \%id2name1, \%node2parent1, \%node2children1,
#		\%name2id2, \%id2name2, \%node2parent2, \%node2children2), where
#	$target_node is a node in the tree in the representation 1;
#	%name2id1, %id2name1, %node2parent1, %node2children1 are for the representation 1;
#	%name2id2, %id2name2, %node2parent2, %node2children2 are for the representation 2.
#
# OUTPUT: $located_node,
#	which is the counterpart in the representation 2 of $target_node in the representation 1.
#
sub locate_node_in_diffrep ($\%\%\%\%\%\%\%\%) {

	my ($tgt_node, $name2id1, $id2name1, $node2pa1, $node2ch1,
		$name2id2, $id2name2, $node2pa2, $node2ch2) = @_;

	my $children1 = $node2ch1->{$tgt_node};

	unless (defined $children1) { # $tgt_node is the external node #
	    my $tgt_name = $id2name1->{$tgt_node};
	    unless (defined $tgt_name) {
		print STDERR "[ERROR 'locate_node_in_diffrep'] NO NAME assigned to the input external node '$tgt_node' !\n";
		return undef;
	    }
	    my $located_node = $name2id2->{$tgt_name};
	    unless (defined $located_node) {
		print STDERR "[ERROR 'locate_node_in_diffrep'] NO COUNTERPART in the representation 2 of the node '$tgt_node' (name: $tgt_name) in the representation 1 !\n";
	    }
	    return $located_node;
	}

	my $topnode1 = $name2id1->{$TOP_NODE};
	my $topnode2 = $name2id2->{$TOP_NODE};

	my $all_extnodes1 = fetch_ext_offsprings_nr ($topnode1, %{$node2ch1});
	my $all_extnodes1_in_2 = [];

	foreach my $node1 (@{$all_extnodes1}) { # The 1st outer foreach-loop #
	    my $name = $id2name1->{$node1};
	    (defined $name) or next;
	    my $node2 = $name2id2->{$name};
	    (defined $node2) or next;
	    push @{$all_extnodes1_in_2}, $node2;
	}				# End of the 1st outer foreach-loop #

	my @children1_in_2 = ();
	foreach my $child (@{$children1}) { # The 2nd outer foreach-loop #
	    my $sb_extnodes1 = fetch_ext_offsprings_nr ($child, %{$node2ch1});
	    my @sb_extnodes1_in_2 = ();
	    foreach my $node1 (@{$sb_extnodes1}) { # The inner foreach-loop #
		my $name = $id2name1->{$node1};
		(defined $name) or next;
		my $node2 = $name2id2->{$name};
		(defined $node2) or next;
		push @sb_extnodes1_in_2, $node2;
	    }				# End of the inner foreach-loop #
	    my ($lca2, $depth_lca2) = fetch_last_common_ancestor_ov2 (@sb_extnodes1_in_2, %{$node2pa2});
	    (undef, undef, $all_extnodes1_in_2) = diff_array (@sb_extnodes1_in_2, @{$all_extnodes1_in_2});
	    ($lca2 eq $topnode2) or (push @children1_in_2, $lca2);
	}				# End of the 2nd outer foreach-loop #

	if ((@{$all_extnodes1_in_2}>0) and (@children1_in_2 <2)) {
	    my ($lca2, $depth_lca2) = fetch_last_common_ancestor_ov2 (@{$all_extnodes1_in_2}, %{$node2pa2});
	    ($lca2 eq $topnode2) or (push @children1_in_2, $lca2);
	}
	
	my $located_node;
	if (@children1_in_2 > 1) { # This uniquely determines the location of the counterpart #
	    ($located_node, my $depth) = fetch_last_common_ancestor_ov2 (@children1_in_2, %{$node2pa2});
	} elsif (@children1_in_2 ==1) { # This is the 2nd best measure #
	    $located_node = $node2pa2->{$children1_in_2[0]};
	}

	unless (defined $located_node) {
	    print STDERR "[ERROR 'locate_node_in_diffrep'] NO COUNTERPART in the representation 2 of the internal node '$tgt_node' in the representation 1 !\n";
	}

	return $located_node;

}		# end of the subroutine 'locate_node_in_diffrep' #


# combine_2attr_on_tree:
#	This subroutine receives two sets of information on the same tree,
#	REPRESENTED IN THE SAME WAY,
#	one equipped with a set of attributes and
#	the other with another set of attributes,
#	combines these attributes,
# 	then returns a hash storing the composite set of attributes.
#
# INPUT: (\%node2pa1, \%name2id1, %br_attr1, \@indices1,
#	     \%node2pa2, \%name2id2, %br_attr2, \@indices2), where
#	the set 1 and the set 2 represents the same tree having 
#	the same OTU names, the same top node, and the same branching at the top.
#	But they may assign the node ID numbers in a slightly different way.
#
# OUTPUT: \%cmp_attr, where
#	@{$cmp_attr{$node_id1}}
#	 = ( @{$br_attr1{$node_id1}}[@indices1], @{$br_attr2{$node_id2}}[@indices2] ),
#	with $node_id1 = $name2id1{$node_name} and $node_id2  = $name2id2{$node_name}
#	representing the same node, is the array 
#	combining the attributes from the set 1 and the set 2.
#
sub combine_2attr_on_tree (\%\%\%\@\%\%\%\@) {

	my ($node2pa1, $name2id1, $br_attr1, $indices1,
		$node2pa2, $name2id2, $br_attr2, $indices2) = @_;

	my %cmp_attr;
		# ($node_id1 => \@cmp_attr, ...) where
		# @cmp_attr = ( @{$br_attr1->{$node_id1}}[@{$indices1}],
		#	@{$br_attr2->{$node_id2}}[@{$indices2}] ).

	my %already; # ($node_id1 => 1, ...) for the nodes
		     # whose attributes have already been extracted.

		## Construct a hash for the composite attributes ##
		## I will use the list of names registered in the set 1
	OUTER_CMB:
	foreach my $node_name (keys %{$name2id1}) { # The outer foreach-loop #

	    my $node_id1 = $name2id1->{$node_name};
	    (defined $node_id1) or next;
	    my $node_id2 = $name2id2->{$node_name};

		# 'Climb' the tree from bottom to top #
	    while (defined $node_id1) { # The inner while-loop #

		(defined $already{$node_id1}) and last; # skip the node already checked.
		$already{$node_id1} = 1;

		unless (defined $node_id2) {
		    print STDERR "[WARNING 'combine_2attr_on_tree'] Found NO NODE ID in set 2 corresponding to the node '$node_name'.\n";
		    $node_id1 = $node2pa1->{$node_id1};
		    $node_id2 = $node2pa2->{$node_id2};
		    next;
		}
		my $attr1 = $br_attr1->{$node_id1};
		my $attr2 = $br_attr2->{$node_id2};
		unless ((defined $attr1) or (defined $attr2) ) {
		    $node_id1 = $node2pa1->{$node_id1};
		    $node_id2 = $node2pa2->{$node_id2};
		    next;
		}

		my @attr1 = (defined $attr1) ? @{$attr1} : ();
		my @attr2 = (defined $attr2) ? @{$attr2} : ();

		my @cmp_attr = @attr1[@{$indices1}];
		(@attr2>0) and (push @cmp_attr, @attr2[@{$indices2}] );
		$cmp_attr{$node_id1} = \@cmp_attr;

		$node_id1 = $node2pa1->{$node_id1};
		$node_id2 = $node2pa2->{$node_id2};

	    }			# End of the inner while-loop #
	}		# End of the outer foreach-loop #

	return \%cmp_attr;

}		# End of the subroutine 'combine_2attr_on_tree' #


# pr_intg_2attri_on_tree:
#	This subroutine reads in two tree files for the same tree,
#	one with bootstrap values and the other with branch lengths,
#	then prints out the tree with the two kinds of attributes integrated in it.
#	It assumes that the files represents the tree in an identical manner,
#	namely, in terms of the top node, and the mode of branching at the top.
# REVISED on Sep 8, 2006:
#	The subroutine has now become more robust in that
#	it can also handle the situation that the files store the same (unrooted) tree,
#	but represented in different ways.
#
# INPUT: (\$intree_blen_path, \$intree_boot_path, \$fh_outtree,
#	$option_bootlabel, $option_nodeid), where
#	$option_bootlabel = 1 if you display bootstrap at the branch,
#			  = 0 (or undef) if at the node.
#	$option_nodeid    = 1 if you display node IDs,
#			  = 0 (or undef) otherwise.
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
sub pr_intg_2attr_on_tree (\$\$\*$$) {

	my ($intree_blen_path, $intree_boot_path, $fh_outtree,
		$option_bootlabel, $option_nodeid) = @_;

	my $intree_blen = ` cat ${$intree_blen_path} 2>&1 ` ;
	if ($intree_blen =~ m/^cat\:\s*\S+\s*No\s+such\s+file/ ) {
	    print STDERR "[ERROR 'pr_intg_2attr_on_tree'] ";
	    print STDERR $intree_blen ;
	    print STDERR "\n";
	    return 0;
	}

	my $intree_boot = ` cat ${$intree_boot_path} 2>&1 ` ;
	if ($intree_blen =~ m/^cat\:\s*\S+\s*No\s+such\s+file/ ) {
	    print STDERR "[ERROR 'pr_intg_2attr_on_tree'] ";
	    print STDERR $intree_boot ;
	    print STDERR "\n";
	    return 0;
	}

	my %node2pa1; my %node2ch1; my %name2id1; my %id2name1; my %br_lens;
	my %node2pa2; my %node2ch2; my %name2id2; my %id2name2; my %br_boot_raw;

	my $fl_in_blen = input_tree_file ($intree_blen, %node2pa1, %node2ch1,
		%name2id1, %id2name1, %br_lens);
	unless ($fl_in_blen) {
	    print STDERR "[ERROR 'pr_intg_2attr_on_tree'] The subroutine 'input_tree_file' FAILED on the tree (with branch lengths):\n";
	    print STDERR " '${$intree_blen_path}' .\n\n";
	    return 0;
	}

	my $fl_in_boot = input_tree_file ($intree_boot, %node2pa2, %node2ch2,
		%name2id2, %id2name2, %br_boot_raw);
	unless ($fl_in_boot) {
	    print STDERR "[ERROR 'pr_intg_2attr_on_tree'] The subroutine 'input_tree_file' FAILED on the tree (with bootstrap values):\n";
	    print STDERR " '${$intree_boot_path}' .\n\n";
	    return 0;
	}

		## keep only bootstrap values on the internal nodes ##
	my %br_boot_neat; # ($nodeid => [$boot (on the upper branch) ], ...)
	foreach my $node (keys %br_boot_raw) {
	   (defined $node2ch2{$node}) and
		($br_boot_neat{$node} = [int ($br_boot_raw{$node}->[0]) ]);
	}

		## (Added on Friday, Sep 8, 2006)
		## Examine whether the top node of tree1 is equivalent to that of tree2 ##
	my $topnode_tree1 = $name2id1{$TOP_NODE};
	my $topnode_tree2 = $name2id2{$TOP_NODE};
	my $new_topnode_tree2 = locate_node_in_diffrep ($topnode_tree1,
		%name2id1, %id2name1, %node2pa1, %node2ch1,
		%name2id2, %id2name2, %node2pa2, %node2ch2);

	unless (defined $new_topnode_tree2) {
	    print STDERR "[ERROR 'pr_intg_2attr_on_tree'] The subroutine 'locate_node_in_diffrep' FAILED on the tree:\n";
	    print STDERR "    '${$intree_blen_path}' .\n\n";
	    return 0;
	}

		## (Added on Friday, Sep 8, 2006)
		## Unless the topnode of rep1 maps the topnode of rep2,
		## move the topnode of rep2 so that it gives the same representation as rep1. 
	unless ($topnode_tree2 eq $new_topnode_tree2) {
	    my $fl_move = move_topnode ($new_topnode_tree2, %node2pa2, %node2ch2, %name2id2, %br_boot_neat);
	    unless ($fl_move) {
		print STDERR "[ERROR 'pr_intg_2attr_on_tree'] The subroutine 'move_topnode' FAILED on the tree:\n";
		print STDERR "   '${$intree_blen_path}' .\n\n";
		return 0;
	    }
	}

	my @indices = (0);
	my $cmp_attr = combine_2attr_on_tree (
			%node2pa1, %name2id1, %br_lens, @indices,
			%node2pa2, %name2id2, %br_boot_neat, @indices);

	unless (defined $cmp_attr) {
	    print STDERR "[ERROR 'pr_intg_2attr_on_tree'] The subroutine 'combine_2attr_on_tree' FAILED on the tree:\n";
	    print STDERR "  '${$intree_blen_path}' .\n\n";
	    return 0;
	}

	my $top_node_id = $name2id1{$TOP_NODE};
	unless (defined $top_node_id) {
	    print STDERR "[ERROR 'pr_intg_2attr_on_tree'] Found NO ID of the TOP NODE for the tree:\n";
	    print STDERR "   '${$intree_blen_path}' .\n\n";
	    return 0;
	}

	my $fl_output = br_output_tree ($top_node_id,%node2ch1, %id2name1,
			%{$cmp_attr}, *${$fh_outtree}, $option_bootlabel, $option_nodeid);

	unless ($fl_output) {
	    print STDERR "[ERROR 'pr_intg_2attr_on_tree'] The subroutine 'pr_intg_2attr_on_tree' FAILED on the tree:\n";
	    print STDERR "   '${$intree_blen_path}' .\n\n";
	    return 0;
	}

	return 1;

}		# End of the subroutine 'pr_intg_2attr_on_tree' #


# pr_tree_with_trans_boot:
#	This subroutine reads in a files for a concensus tree,
#	with bootstrap values displayed like branch lengths,
#	then prints out the tree with bootstrap valued displayed
#	according to the format of CLUSTAL W.
#
# INPUT: (\$intree_boot_path, \*${$fh_outtree},
#	$option_bootlabel, $option_nodeid), where
#	$option_bootlabel = 1 if you display bootstrap at the branch,
#			  = 0 (or undef) if at the node.
#	$option_nodeid    = 1 if you display node IDs,
#			  = 0 (or undef) otherwise.
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
sub pr_tree_with_trans_boot (\$\*$$) {

	my ($intree_boot_path, $fh_outtree,
		$option_bootlabel, $option_nodeid) = @_;

	my $intree_boot = ` cat ${$intree_boot_path} 2>&1 ` ;
	if ($intree_boot =~ m/^cat\:\s*\S+\s*No\s+such\s+file/ ) {
	    print STDERR "[ERROR 'pr_tree_with_trans_boot'] ";
	    print STDERR $intree_boot ;
	    print STDERR "\n";
	    return 0;
	}

	my %node2pa; my %node2ch; my %name2id; my %id2name; my %br_boot_raw;

	my $fl_in_boot = input_tree_file ($intree_boot, %node2pa, %node2ch,
		%name2id, %id2name, %br_boot_raw);
	unless ($fl_in_boot) {
	    print STDERR "[ERROR 'pr_tree_with_trans_boot'] The subroutine 'input_tree_file' FAILED on the tree (with bootstrap values):\n";
	    print STDERR " '${$intree_boot_path}' .\n\n";
	    return 0;
	}

		## Keep only bootstrap values on the internal nodes ##
		## and change fields to store the bootstrap value. ##

	my %br_boot_clw; # ($nodeid => \@attr, ...) with $attr[$INDX_BOOT] being the bootstrap value.
	foreach my $node (keys %br_boot_raw) {
	   (defined $node2ch{$node}) or next;  # Skip external nodes #

	   my @attr = ();
	   $attr[$INDX_BOOT] = int ($br_boot_raw{$node}->[0]);
	   $br_boot_clw{$node} = \@attr;
	}

	my $top_node_id = $name2id{$TOP_NODE};
	unless (defined $top_node_id) {
	    print STDERR "[ERROR 'pr_tree_with_trans_boot'] Found NO ID of the TOP NODE for the tree:\n";
	    print STDERR "   '${$intree_boot_path}' .\n\n";
	    return 0;
	}

	my $fl_output = br_output_tree ($top_node_id, %node2ch, %id2name,
		%br_boot_clw, *${$fh_outtree}, $option_bootlabel, $option_nodeid);

	unless ($fl_output) {
	    print STDERR "[ERROR 'pr_tree_with_trans_boot'] The subroutine 'br_output_tree' FAILED on the tree:\n";
	    print STDERR "   '${$intree_boot_path}' .\n\n";
	    return 0;
	}

	return 1;

}		# End of the subroutine 'pr_tree_with_trans_boot' #


# construct_raw_subtree:
#	This subroutine receives a tree and a subset of its nodes,
#	then constructs a 'raw' subtree consisting of thee input nodes
#	and their ancestors, including those above the last common ancestor
#	of the input nodes.
#
# INPUT: (\%node2parent, \@sub_nodes), where
#	%node2parent = ($node_id => $parent_id, ...) for the entire tree;
#	@sub_nodes = (a list of node IDs that you want to include in the subtree).
#
# OUTPUT: (\@topnode_ids, \%sb_node2children, \%sb_node2parent), where
#	@topnode_ids stores the IDs of the top nodes of the subtree;
#	%sb_node2children = ($node_id => [a list of IDs for chilren], ...) for the subtree;
#	%sb_node2parent = ($node_id => $parent_id, ...) for the subtree.
#
sub construct_raw_subtree (\%\@) {

	my ($node2parent, $sub_nodes) = @_;

	my %sb_node2parent;
	my $topnode_ids = [];

	OUTER_RAW:
	foreach my $node (@{$sub_nodes}) { # The outer foreach-loop #
	    (defined $sb_node2parent{$node}) and next;
	    my $curr_node = $node;
	    my $parent = $node2parent->{$curr_node};

	    while (defined $parent) { # The inner while-loop #
		$sb_node2parent{$curr_node} = $parent;
		if (defined $sb_node2parent{$parent}) {
		    next OUTER_RAW;
		}
		$curr_node = $parent;
		$parent = $node2parent->{$curr_node};
	    }			# End of the inner while-loop #

	    my @tmp = ($curr_node);
	    $topnode_ids = union_array (@{$topnode_ids}, @tmp);

	}			# End of the outer foreach-loop #

	if (@{$topnode_ids}>1) {
	    print STDERR "[CAUTION 'construct_raw_subtree'] Found MULTIPLE TOPNODES: ";
	    print STDERR join (', ', @{$topnode_ids}), " .\n\n";
	}
	
	my $sb_node2children = make_node2children (%sb_node2parent);

	return ($topnode_ids, $sb_node2children, \%sb_node2parent);

}		# end of the subroutine 'construct_raw_subtree' #


# rm_simple_ancestral_lineages:
#	This subroutine removes 'ancestoral lineages',
#	 each of which starts from a top node and
#	consists only of one node for each generation.
#	It does NOT remove EXTERNAL NODES, though.
#
# INPUT: (\@topnode_ids, \%node2children, \%node2parent), where
#	@topnode_ids lists the topnode IDs;
#	%node2children = ($node_id => [a list of IDs for the children], ...);
#	%node2parent = ($node_id => $parent_id, ...).
#
# OUTPUT: \@new_topnode_ids, where
#	@new_topnode_ids list the topnode IDs AFTER the processing.
#
sub rm_simple_ancestral_lineages (\@\%\%) {

	my ($topnode_ids, $node2children, $node2parent) = @_;

	my @new_topnode_ids = ();

	foreach my $topnode (@{$topnode_ids}) { # The outer foreach-loop #

	     my $curr_node = $topnode;
	     my $children = $node2children->{$curr_node};

		# See if the current node is internal or not #
		# as well as if it has one or no child #
	     while ((defined $children) and (@{$children}<2)) { # The inner while-loop #

		unless (@{$children}>0) {
		    print STDERR "[CAUTION 'rm_simple_ancestral_lineages'] The node '$curr_node' has an EMPTY SET of CHILDREN !\n";
		    return undef;
		}
		my $child = $children->[0];

			# Remove the parent of $child, which is nothing but $curr_node,
			# because $child is now going to be the topnode.
		##(defined $node2parent->{$child}) and (undef $node2parent->{$child});
		(exists $node2parent->{$child}) and (delete $node2parent->{$child});

			# For the same reason,
			# remove $curr_node from %{$node2children} and %{$node2parent}.
		(delete $node2children->{$curr_node});
		(exists $node2parent->{$curr_node}) and (delete $node2parent->{$curr_node});

		$curr_node = $child;
		$children = $node2children->{$curr_node};

	     }			# End of the inner while-loop #

	     push @new_topnode_ids, $curr_node;

	}			# End of the outer foreach-loop #

	return \@new_topnode_ids;

}		# end of the subroutine 'rm_simple_ancestral_lineages' #


# remove_trivial_nodes:
#	This subroutine removes 'trivial nodes',
#	each of which just joins two neighboring branches.
#
# INPUT: (\%node2children, \%node2parent, \%br_attr), where
#	%node2children = ($node_id => [a list of IDs for chilren], ...) ;
#	%node2parent = ($node_id => $parent_id, ...) ;
#	($length, $boot) = @{$br_attr{$node}[$INDX_LENGTH, $INDX_BOOT]
#	are the length and the bootstrap value of the branch directly above $node.
#
# OUTPUT: \%sb_br_attr, where
#	($length, $boot) = @{$sb_br_attr{$node}[$INDX_LENGTH, $INDX_BOOT]
#	are the REPRESENTATIVE length and the bootstrap value of the branch directly above $node.
#	The RULES for the representative values are:
#	$length = the summation of the constituent branch lengths;
#	$boot  = the bootstrap value of the uppermost constituent branch.
#	REVISED on Sep 15 (Fri), 2006:
#	$boot = the MAXIMUM bootstrap value among the constituent branches,
#	because that should best represent the partition effected by the composite branch.
#
sub remove_trivial_nodes (\%\%\%) {

	my ($node2children, $node2parent, $br_attr) = @_;

	my @ext_nodes = ();
	foreach my $node (keys %{$node2parent}) { # the 1st foreach-loop #
	    (defined $node2children->{$node}) or (push @ext_nodes, $node);
	}				# end of the 1st foreach-loop #

	my %already; # ($node_id => 1, ...) for nodes that have already been inspected #
	my %sb_br_attr;

	foreach my $ex_node (@ext_nodes) { # The outer foreach-loop #

	    if (defined $br_attr->{$ex_node}) {
		my @tmp = @{$br_attr->{$ex_node}};
		$sb_br_attr{$ex_node} = \@tmp;
	    }

	    my $curr_node = $node2parent->{$ex_node};

	    while (defined $curr_node) { # The inner while-loop # 

		(defined $already{$curr_node}) and last;
		$already{$curr_node} = 1;

		my $parent = $node2parent->{$curr_node};
		(defined $parent) or last; # $curr_node is a top node,
					   # which we do not tamper with here.

		my $children = $node2children->{$curr_node};
		unless ((defined $children) and (@{$children}>0)) {
		    print STDERR "[ERROR 'remove_trivial_nodes'] The internal node '$curr_node' has NO CHILDREN !\n";
		    return undef;
		}

		if (@{$children}==1) {
			# remove $curr_node because it is trivial #

		    my $child = $children->[0];

			# Make $parent be $child's parent #
		    $node2parent->{$child} = $parent;
		    delete $node2parent->{$curr_node};

			# Make $child be one of $parent's childs #
		    my @new_children = ();
		    foreach my $sibl (@{$node2children->{$parent}}) {
			push @new_children, ($sibl eq $curr_node) ? $child : $sibl;
		    }
		    $node2children->{$parent} = \@new_children;
		    delete $node2children->{$curr_node};

			# Change attributes #
		    ##my @curr_attr = (defined $br_attr->{$curr_node}) ? ## Revised on Sep 15 (Fri), 2006
		    my @add_attr = (defined $br_attr->{$curr_node}) ?
			@{$br_attr->{$curr_node}} : ();
		    ##my @add_attr = (defined $sb_br_attr{$child}) ? ## Revised on Sep 15 (Fri), 2006
		    my @curr_attr = (defined $sb_br_attr{$child}) ?
			@{$sb_br_attr{$child}} : ();
		    my $curr_len = $curr_attr[$INDX_LENGTH];
		    my $curr_boot = (@curr_attr>$INDX_BOOT) ? $curr_attr[$INDX_BOOT] : undef;
		    my $add_len = $add_attr[$INDX_LENGTH];
		    my $add_boot = (@add_attr>$INDX_BOOT) ? $add_attr[$INDX_BOOT]: undef;

		    if (defined $add_len) {
			$curr_attr[$INDX_LENGTH] = (defined $curr_len) ?
				$curr_len + $add_len : $add_len;
		    }
		    ((defined $curr_boot) and (defined $add_boot) and ($add_boot > $curr_boot))
			and ($curr_attr[$INDX_BOOT] = $add_boot);

		    (@curr_attr>0) and ($sb_br_attr{$child} = \@curr_attr);
		} else { 
			# Use the attributes for the input tree #
		    if (defined $br_attr->{$curr_node}) {
			my @tmp = @{$br_attr->{$curr_node}};
			$sb_br_attr{$curr_node} = \@tmp;
		    }
		}

		$curr_node = $parent;
	    }			# End of the inner while-loop #

	}			# End of the outer foreach-loop #

	return \%sb_br_attr;

}		# end of the subroutine 'remove_trivial_nodes' #


# construct_neat_subtree:
#	This subroutine receives a tree and a subset of its nodes,
#	then constructs a 'TIDIED-UP' subtree,
#	which does not contain any nodes that just interrupt single branch,
#	nor the ancestral lineage leading to (but excluding)
#	the last common ancestor of the input nodes.
#
# INPUT: (\%node2parent, \@sub_nodes, \%br_attr), where
#	%node2parent = ($node_id => $parent_id, ...);
#	@sub_nodes = (a list of node IDs that you want to include in the subtree);
#	($length, $boot) = @{$br_attr{$node}[$INDX_LENGTH, $INDX_BOOT]
#	are the length and the bootstrap value of the branch directly above $node.
#
# OUTPUT: (\@sb_topnode_ids, \%sb_node2children, \%sb_node2parent, \%sb_br_attr), where
#	@sb_topnode_ids stores the IDs of the top nodes in the subtree;
#	%sb_node2children = ($node_id => [a list of IDs for chilren], ...) for the subtree;
#	%sb_node2parent = ($node_id => $parent_id, ...) for the subtree;
#	($length, $boot) = @{$sb_br_attr{$node}[$INDX_LENGTH, $INDX_BOOT]
#	are the REPRESENTATIVE length and the bootstrap value of the branch directly above $node.
#	The RULES for the representative values are:
#	$length = the summation of the constituent branch lengths;
#	$boot  = the bootstrap value of the uppermost constituent branch.
#
sub construct_neat_subtree (\%\@\%) {

	my ($node2parent, $sub_nodes, $br_attr) = @_;

	my ($tnt_sb_topnode_ids, $sb_node2children, $sb_node2parent)
		= construct_raw_subtree (%{$node2parent}, @{$sub_nodes});

	my $sb_topnode_ids = rm_simple_ancestral_lineages (@{$tnt_sb_topnode_ids}, %{$sb_node2children}, %{$sb_node2parent});

	my $sb_br_attr = remove_trivial_nodes (%{$sb_node2children}, %{$sb_node2parent}, %{$br_attr});

	return ($sb_topnode_ids, $sb_node2children, $sb_node2parent, $sb_br_attr);

}		# end of the subroutine 'construct_neat_subtree' #


# construct_skeleton_subtree:
#	This subroutine, which was created by modifying 'construct_neat_subtree',
#	constructs a 'skeleton' subtree, 
#	which consists of the input sebset of nodes and their ancestral nodes
#	up to their most recent common ancestor.
#	The 'skeleton' subtree may contain some 'trivial' nodes, each of which just joins two successive branches
#	(, namely, each of which just splits a single branch into two branches).
#	[ CREATED on Nov 7 (Thu), 2013. ]
#
# construct_neat_subtree:
#	This subroutine receives a tree and a subset of its nodes,
#	then constructs a 'TIDIED-UP' subtree,
#	which does not contain any nodes that just interrupt single branch,
#	nor the ancestral lineage leading to (but excluding)
#	the last common ancestor of the input nodes.
#
# INPUT: (\%node2parent, \@sub_nodes, \%br_attr), where
#	%node2parent = ($node_id => $parent_id, ...);
#	@sub_nodes = (a list of node IDs that you want to include in the subtree);
#	($length, $boot) = @{$br_attr{$node}[$INDX_LENGTH, $INDX_BOOT]
#	are the length and the bootstrap value of the branch directly above $node.
#
# OUTPUT: (\@sb_topnode_ids, \%sb_node2children, \%sb_node2parent, \%sb_br_attr), where
#	@sb_topnode_ids stores the IDs of the top nodes in the subtree;
#	%sb_node2children = ($node_id => [a list of IDs for chilren], ...) for the subtree;
#	%sb_node2parent = ($node_id => $parent_id, ...) for the subtree;
#	($length, $boot) = @{$sb_br_attr{$node}[$INDX_LENGTH, $INDX_BOOT]
#	are the length and the bootstrap value of the branch directly above $node,
#	which are just carried over from the input tree.
#
sub construct_skeleton_subtree (\%\@\%) {

	my ($node2parent, $sub_nodes, $br_attr) = @_;

	my ($tnt_sb_topnode_ids, $sb_node2children, $sb_node2parent)
		= construct_raw_subtree (%{$node2parent}, @{$sub_nodes});

	my $sb_topnode_ids = rm_simple_ancestral_lineages (@{$tnt_sb_topnode_ids}, %{$sb_node2children}, %{$sb_node2parent});

	##my $sb_br_attr = remove_trivial_nodes (%{$sb_node2children}, %{$sb_node2parent}, %{$br_attr}); ## For the subroutine 'construct_neat_subtree'. ##

	my %sb_br_attr;

	foreach my $br (keys %{$br_attr}) { # Main foreach-loop (over branches) #

	    (defined $sb_node2parent->{$br}) or next; # Skip if the branch is not in the skeleton tree.

	     $sb_br_attr{$br} = $br_attr->{$br};

	}		# End of the main foreach-loop (over branches) #

	return ($sb_topnode_ids, $sb_node2children, $sb_node2parent, \%sb_br_attr);

}		# end of the subroutine 'construct_skeleton_subtree' #


1;	# a "spell" which makes the module "use"-able and/or "require"-able #



