package MyTreeManipu_VI2;
#
###################################################################################
#	DENSERM_P: A package to DEtect Negative SElection on Recurrent Mutations
#			(almost exclusively via Perl modules and scripts).
#	Version 0.3: Copyright (C) 2012 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyTreeManipu_VI2.pm:
#	This module conducts more 'nerdy' tree manipulations,
#	such as the estimation of the standard estimation errors of the branch lengths,
#	and the Neighbor-Joining algorithm of the phylogenetic tree construction.
#
#	Most of the subroutines in this module are based on algorithms or methods described in:
#	"Molecular Evolution and Phylogenetics",
#	written by Nei M, Kumar S (Oxford University Press, Inc.; New York, NW, USA; 2000).
#
# MyTreeManipu_VI.pm:
#	This module conducts algebraically advanced tree manipulations,
#	such as those related to the estimate of branch lengths.
#	The first few subroutines intend to emulate the 'fitch' program
#	in the Phylip package, which conducts Fitch-Margoliash algorithm of
#	the branch length estimation (and the tree topology inference).
#
# MyTreeManipu_V.pm:
#	This package conducts sophisticated tree manipulations.
#	Specifically, it has subroutines related to phylogenetic reconciliations.
#
#
# MyTreeManipu_II.pm: This package can process more general types of trees,
#	(i.e. NOT limited to binary trees with trichotomous top nodes).
#	It is also consistent with the format of the taxonomy tree which are handled by
#	the module 'MyTaxonomy.pm'.
#
# MyTreeManipu.pm: This package performs basic tree manipulations.
#
#  Subroutines whose substitutes are available in this module ('MyTreeManipu_II.pm')
#	or in 'MyTaxonomy.pm':
#   our @EXPORT = qw(retrieve_top_node confine_outgroup print_unrooted);
#   our @EXPORT_OK = qw(switch_topnode  print_subtree topo_unrooted topo_subtree
#		list_sequences fetch_node_series dist_nodes
#		    );
#
# Subroutines whose substitutes are not available yet:
#   our @EXPORT_OK = ( retrieve_bootstrap  prod_bootstrap );
#

use strict;
use warnings;

use MyArrayManipu qw(
	diff_array
	);
 	## union_array nr_array compare_arrays
	## union_array

use MyArrayManipu_num qw(
	diff_nmarray
	);

use MyFile_manipu qw(
	open_input_file
	);

use MyLinearAlgebra qw(
	solve_lin_eq_lu invert_matrix_lu
	);

use MySpecFunc qw(
	erffc
	);

use MyTaxonomy qw(
	fetch_ext_offsprings_nr
	fetch_all_offsprings_nr
	fetch_common_ancestors
	);
	## fetch_ancestry
	## fetch_last_common_ancestor fetch_last_common_ancestor_ov2
	## fetch_all_offsprings

use MyTreeMap qw(
	fractionate_by_ndepth
	);
	## assign_gnode2sp_lca
	## fetch_splca_comps assign_gnode2sp_lca_comp

use MyTreeManipu_II qw(
	br_draw_subtree
	);
use MyTreeManipu_III qw(
	make_node2abs_depth
	);

=set aside for later use
use MyProbDistri qw(
	cum_gaussian
	);
use MyTreeManipu_II qw(
	make_node2parent shift_topnode move_topnode segregate_subset_nodes
	);
	## input_tree_file read_tree_file input_subtree copy_tree
	## output_tree draw_subtree
	## br_output_tree br_draw_subtree topo_subtree
	## make_node2children make_node2parent
	## sum_branch_lengths intra_lineage_distance inter_node_distance
	## fetch_boot_series extract_hidden_bootstrap extract_hidden_bootstrap2
	## resolve_multipartite
use MyTreeManipu_IV qw(
	construct_neat_subtree
	);
use MyTreeManipu_V qw(
	fetch_species_ancestor make_gpair2sp_lca
	examine_node_states inspect_med_branches
	reroot_moltree_heur fractionate_spe_dup_gpairs
	resolve_mlt_gnode_with_sptree resolve_multifurcation
	assign_gnode2spe_dup annotate_spe_dup
	sep_genesets_by_spnode_bu sep_genesets_by_spnode_td
	separate_into_ortho_sets collect_og_phylogene_sets
	find_closest_outgroups bch_attach_closest_ogs
	);
use MyTreeManipu_VI qw(
	readin_phylip_dists extract_phylip_dists
	coeffs_leastsq_brlen leastsq_brlen
	assign_trivial_brlen leastsq_brlen_nn
	extended_upgma_tree extended_wtd_upgma_tree
	level_sglnode_upgma var_nodelevels_ext_wgt_upgma
	approx_var_brlen_ext_wgt_upgma
	brlen_ext_wgt_upgma apprx_bootspp_by_z_test
	);
=cut

use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	coeffs_qo_leastsq_brlen qo_leastsq_brlen
	rhs_qols_var_brlen qo_leastsq_var_brlen
	neighbor_joining_tree z_values_branch_lengths
	) ;


our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;

our $MAX_SPDEPTH = 1E+6; # The maximum depth in species trees,
			# which is assigned to the common ancestor of an empty set.
			# (An actual depth cannot be that large.)

our $UL_ITMLT = 20; # An upper limit to conduct iterative multiplications
		   # for calculating a power.


=Descriptions
### 5 important hash tables which describe a tree ###
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#				 # because we can always identify the top-node by
#				# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.
#
=cut

=set aside for a potential use
# mypower:
#	This subroutine calculates power, basically using
#	iterative multiplication. 
#
# INPUT: ($base, $exponent).
#
# OUTPUT: $power = $base ** $exponent.
#
sub mypower ($$) {

	my ($base, $exponent) = @_;

	($exponent ==0) and (return 1); ## DEFINE  0 * log 0 as 1 # 
	(($base ==0) and ($exponent >0)) and (return 0);
	if ($base ==0) {
	    print STDERR "[ERROR 'mypower'] ILLEGAL INPUT: (base, exponent) = \($base, $exponent\) !\n";
	    return undef;
	}

	my $power = 1;
	if ($exponent == int $exponent) {
	    my $abs_e = ($exponent>0) ? $exponent : - $exponent;
	    if ($abs_e <= $UL_ITMLT) {
		for (1 .. $abs_e) { $power *= $base; }
		($exponent<0) and ($power = 1/$power); 
	    } else {
		my $abs_b = ($base > 0) ? $base : - $base;
		$power = $abs_b ** $exponent;
		($exponent % 2) and ($base<0) and ($power = -$power);
	    }
	} elsif ($base <0) {
	    print STDERR "[ERROR 'mypower'] ILLEGAL INPUT: (base, exponent) = \($base, $exponent\) !\n";
	    return undef;
	} else {
	   $power = $base ** $exponent;
	}

	return $power;

}		# End of the subroutine 'mypower' #
=cut


# assign_branch_indices:
#	This subroutine assigns indices for the nodes immediately below the branches
#	from 0 upwards in a sloppy manner.
#
# INPUT: (\%node2parent, %null_br), where
#	%node2parent = ($node_id => $parent_id, ...);
#	%null_br =  (ID of the node immediately below the branch => some defined value, ...)
#	for the branches whose distances are definitely zero.
#
# OUTPUT: \%br2indx, where
#	%br2indx = (ID of the node immediately below the branch => $indx, ...)
#	for the branches whose distances may be nonzero.
#
sub assign_branch_indices (\%\%) {

	my ($node2parent, $null_br) = @_;

	my $indx = 0;
	my %br2indx;
	foreach my $nodeid (sort { $a <=> $b} keys %{$node2parent}) {
	    (defined $node2parent->{$nodeid}) or next; # skip the node NOT supporting a branch
	    (defined $null_br->{$nodeid}) and next; # skip the node whose supporting branch is known to be zero in length
	    $br2indx{$nodeid} = $indx;
	    $indx++;
	}

	return \%br2indx;

}		# end of the subroutine 'assign_branch_indices' #


# coeffs_qo_leastsq_brlen:
#	This satellite subroutine of 'qo_leastsq_brlen'
#	constructs the coefficient matrix and the right-hand-side vector
#	that are necessary for the least-squares estimation of the branch lengths
#	under the QUASI-ORDINARY least-squares (qOLS) framework.
#	The subroutine was made by revising 'coeffs_leastsq_brlen' in the module 'MyTreeManipu_IV.pm'.
#	The qOLS framework specifies the weight factor as:
#	w(i,j) = 1 for the Poisson model of aa substitution,
#	       = 1/[ 1 + d(i,j)/alpha ] for the gamma substitution model with the shape parameter alpha.
#
# INPUT: (\@dists, \@nodeids, \%br2indx, \%node2parent, $alpha), where
#	$dists[$i]->[$j] is the distance between the $i th and $j th sequences
#	(with 0 <= $j < $i <= #{sequences});
#	$nodeids[$i] is the node ID of the $i th sequence;
#	%br2indx = (ID of the node immediately below the branch => $indx, ...);
#	%node2parent = ($node_id => $parent_node_id, ...);
#	$alpha (>0) is tha shape parameter of the gamma model of the amino acid substitution model,
#	$alpha = 0 specifies the Poisson substitution model.
#
# OUTPUT: (\@coeffs, \@rhs), where
#	$coeffs[$m]->[$n] is the coefficient acting on the $n th branch length
#		to form the equation for the $m th branch length;
#	$rhs[$m] is the right hand side of the equation for the $m th branch length.
#
sub coeffs_qo_leastsq_brlen (\@\@\%\%$) {

	my ($dists, $nodeids, $br2indx, $node2parent, $alpha) = @_;


	if ($alpha <0) {
	    print STDERR "[ERROR 'coeffs_qo_leastsq_brlen'] ILLEGAL INPUT: NEGATIVE POWER currently NOT ALLOWED!\n";
	    return (undef, undef);
	}

=for TESTING the subroutine
	print STDERR "\n\< Subroutine 'coeffs_qo_leastsq_brlen' \>\n\n";
=cut

	my $ct_br = scalar keys %{$br2indx};
	my $ct_nodes = @{$nodeids}; 

		# Initialize the output arrays #
	my @coeffs = my @rhs = ();
	for (1 .. $ct_br) { # The double for-loop #
	   push @rhs, 0;
	   my @row = ();
	   for (1 .. $ct_br) { push @row, 0; }
	   push @coeffs, \@row;
	}		# End of the double for-loop #

		# Now construct the coefficients #
	## my $regl = ($power == 0) ? 1 : $buffer ** $power;  # A 'regulator'

	for (my $i=1; $i<$ct_nodes; $i++) { # The outermost for-loop #
	    my $node1 = $nodeids->[$i];
	    my $row_dists = $dists->[$i];
	    for (my $j=0; $j<$i; $j++) { # The outer for-loop #
		my $node2 = $nodeids->[$j];
		my $dist = $row_dists->[$j]; # Distance

			# Reciprocal denominator
			## REVISED in creation of 'coeffs_qo_leastsq_brlen' (on May 5 (Wed), 2010) ##
		my $rec_den = ($alpha == 0) ? 1 : $alpha/($alpha + $dist); # Reciprocal denominator #
=for the predecessor
		my $pwdist = mypower ($dist, $power);
		unless (defined $pwdist) {
		    print STDERR "[ERROR 'coeffs_leastsq_brlen'] 'mypower' FAILED for the node pair \($node1, $node2\) !\n";
		    return (undef, undef);
		}
		my $rec_den = ($power == 0) ? 1 : 1/($regl + $pwdist); # Reciprocal denominator #
=cut

		my ($cm_anc, $lin1, $lin2) = fetch_common_ancestors ($node1, $node2, %{$node2parent});
		unless (defined $cm_anc) {
		    print STDERR "[ERROR 'coeffs_leastsq_brlen'] 'fetch_common_ancestors' FAILED for the node pair \($node1, $node2\) !\n";
		    return (undef, undef);
		}
=for TESTING the subroutine
		print STDERR "\n{ Node pair \($node1, $node2\) }\n\n";
		print STDERR "distance = $dist, powered_distance = $pwdist,\n";
		print STDERR "    reciprocal_denominator = $rec_den .\n\n";
		print STDERR "Branches on the path: ", join (', ', @{$lin1}, @{$lin2}), ".\n\n";
=cut
		my @path = ();
		foreach my $brid (@{$lin1}, @{$lin2}) { # The 1st inner for-loop #
		    my $br_indx = $br2indx->{$brid};
		    (defined $br_indx) or next; # If the index is NOT defined,
				# the branch should have a fixed (zero) length.
		    push @path, $br_indx;
		}				# End of the 1st inner for-loop #

		my $ct_on_path = @path;
		for (my $m=0; $m<$ct_on_path; $m++) { # The 2nd inner for-loop #
		    my $br1 = $path[$m];
		    $rhs[$br1] += $dist * $rec_den;
		    my $row = $coeffs[$br1];
		    for (my $n=0; $n<$ct_on_path; $n++) { $row->[$path[$n]] += $rec_den; }
		} 				# End of the 2nd inner for-loop #
	    }				# End of the outer for-loop #
	}			# End of the outermost for-loop #

	return (\@coeffs, \@rhs);

}		# end of the subroutine 'coeffs_qo_leastsq_brlen' #


# qo_leastsq_brlen:
#	This subroutine estimates branch lengths by means of
#	a least squares method UNDER the QUASI-ORDINARY LS (qOLS) framework.
#	In the qOLS framework, the following sum of squares is minimized:
#	Sum_{(i,j)} [Exp(i,j) - Obs(i,j)]**2 /[ 1 + Obs(i,j)/ alpha ],
#	where Obs(i,j) is the observed value of the distance between the sequences i and j;
#	Exp(i,j) = Sum_{br(m) in path(i,j)} l(m)  is the expected vale of the distance
#	under a given set {l(m)} of the branch lengths;
#	alpha (>0) is the shape parameter for the gamma model of amino acid substitutions.
#
#	The subroutine was made by revising 'leastsq_brlen' in the module 'MyTreeManipu_VI.pm'.
#
# INPUT: (\@seqnames, \@dists, \%name2id, \%node2parent, $alpha), where
#	$seqnames[$i] is the name (label) of the $i th sequence;
#	$dists[$i]->[$j] is the distance between the $i th and $j th sequences
#	(with 0 <= $j < $i < #{sequences});
#	%name2id = ($seq_name => $node_id, ...), with $seq_name has to belong to @seqnames;
#	%node2parent = ($node_id => $parent_node_id, ...);
#	$alpha is the shape parameter for the gamma model of amino acid substitutions.
#
# OUTPUT: \%br_attr, where
#	%br_attr = (ID of the node immediately below the branch => [branch length], ...).
#
sub qo_leastsq_brlen (\@\@\%\%$) {

	my ($seqnames, $dists, $name2id, $node2parent, $alpha) = @_;

	if (@{$seqnames}<3) {
	    my $ct_seqs = @{$seqnames};
	    print STDERR "[ERROR 'qo_leastsq_brlen'] The number of sequences ($ct_seqs) is NOT ENOUGH for this subroutine to work !\n";
	    return undef;
	}
	if ($alpha < 0) {
	    print STDERR "[ERROR 'qo_leastsq_brlen'] ILLEGAL INPUT: A NEGATIVE shape parameter ($alpha): $! !\n";
	    return undef;
	}

		# Create an array, while checking the consistency
		# between the distance matrix and the tree.
	my @nodeids = (); # = ($node_id, ...)
	foreach my $seqname (@{$seqnames}) {
	   my $nodeid = $name2id->{$seqname};
	   unless (defined $nodeid) {
		print STDERR "[ERROR 'qo_leastsq_brlen'] The sequence '$seqname' has NO NODE ID!\n";
		return undef;
	   }
	   push @nodeids, $nodeid
	}

	my %br_attr; # (ID of the node immediately below the branch => \@attr, ...)

	# %{$br2indx} = (ID of the node immediately below the branch => $indx, ...)
	#
	my $br2indx = assign_branch_indices (%{$node2parent}, %br_attr);

		## BUG FIXING on Monday, March 5, 2007 ##
		## In the case where all the branches are trivially zero ##
	(0 < scalar keys %{$br2indx}) or (return \%br_attr);

		## REVISED in creation of 'qo_leastsq_brlen' (on May 5 (Wed), 2010) ##
	my ($coeffs, $rhs) = coeffs_qo_leastsq_brlen (@{$dists}, @nodeids,
		%{$br2indx}, %{$node2parent}, $alpha);
	##my ($coeffs, $rhs) = coeffs_leastsq_brlen (@{$dists}, @nodeids,
	##	%{$br2indx}, %{$node2parent}, $power, $buffer);

	unless (defined $coeffs) {
	    print STDERR "[ERROR 'qo_leastsq_brlen'] 'coeffs_qo_leastsq_brlen' FAILED !\n";
	    return undef;
	}

	# @{$br_lens} = (branch lengths arranged in the order of the assigned index.
	my $br_lens = solve_lin_eq_lu (@{$coeffs}, @{$rhs});
	unless (defined $br_lens) {
	    print STDERR "[ERROR 'qo_leastsq_brlen'] 'solve_lin_eq_lu' FAILED !\n";
	    return undef;
	}

		# Cast the estimated branch lengths into the form
		# commonly used in my modules.
	foreach my $dwnode (keys %{$br2indx}) { # The last foreach-loop #
	    my $indx = $br2indx->{$dwnode};
	    (defined $indx) or next;
	    my $br_len = $br_lens->[$indx];
	    unless (defined $br_len) {
		print STDERR "[ERROR 'qo_leastsq_brlen'] The branch immediately above the node '$dwnode' has NO BRANCH LENGTH estimated !\n"; 
		return undef;
	    }
	    my @tmp = ();
	    $tmp[$INDX_LENGTH] = $br_len;
	    $br_attr{$dwnode} = \@tmp;
	}			# End of the last foreach-loop #

	return \%br_attr;

}		# end of the subroutine 'qo_leastsq_brlen' #


# rhs_qols_var_brlen:
#	This satellite subroutine of 'qo_leastsq_var_brlen'
#	calculates the matrix in the RHS of the matrix equation
#	for the variance-covariance matrix between the branch lengths,
#	UNDER the quasi-ORDINARY LEAST SQUARE (qOLS) framework.
#
#	THEORY:
#	In the LS framework, the variance-covariance matrix Cov := [Cov (l_m, l_n)] is given by
#	(C Cov C)(m,n) = Sum_{(i,j) containing br(m)} Sum_{(k,l) containing br(n)}
#		wgt (i,j) * wgt (k,l) * Cov { d(i,j) , d(k,l) },  --- (***)
#	where C = [ C(m,n) ] is the coefficient matrix that were also used in 'qo_leastsq_brlen',
#	wgt(i,j) is the weight factor,
#	and Cov { d(i,j), d(k,l) }  is the covariance between two pairwise distances
#	due to the sampling error.
#	The theoretical estimation of the covariance (under the gamma model of the shape parameter alpha) is:
#	Cov { d(i,j), d(k,l) } = (1 + d(i,j)/alpha) * (1 + d(k,l)/alpha) * (exp (d_ovlp) - 1)/#{sites},
#	with d_ovlp being the length of the path overlapping between d(i,j) and d(k,l).
#	So, in the qOLS framework, with wgt (i,j) = 1/[1 + d(i,j)/alpha],
#	each term in the summation for the rhs of (***) is simplified as:
#
#	[exp (d_ovlp ((i,j), (k,l)) ) - 1  ]/#{sites} for Poisson,
#	[ ( 1 + d_ovlp((i,j), (k,l))/alpha )**alpha - 1]/#{sites} for Gamma (a make-shift).
#
#	Thus, we can calculate the entire summation in the rhs of (***) via
#	an algorithm of O(n**4) time complexity (and O(n**2) space complexity). 
#
# INPUT: ($topnode_id, \%br2indx, \%node2pa, \%node2ch, \%br_attr, $ct_sites, $alpha), where
#	%br2indx = (ID of the node immediately below the branch => $indx in the (vector or matrix) equation, ...) for the branches;
#	%node2pa = ($node_id => $parent_node_id, ...) for the nodes except the top node ;
#	%node2ch = ($node_id => \@child_node_ids, ...) for the interior nodes;
#	%br_attr = ($node_id => \@attr, ...) for the branches each immediately upstream of the node, with $attr[$INDX_LENGTH] giving the branch length;
#	$ct_sites is the number of sites in the MSA used to estimate the pairwise distance matrix;
#	$alpha is the shape parameter of the gamma model. = 0 if the Poisson model is employed.
#
# OUTPUT: \@mtrx_rhs, where
#	$mtrx_rhs[$k]->[$l] is the matrix element in the right-hand-side
#	concerning the $k th and the $l th branch;
#
sub rhs_qols_var_brlen ($\%\%\%\%$$) {

	my ($topnode_id, $br2indx, $node2pa, $node2ch, $br_attr, $ct_sites, $alpha) = @_;


		## Initialize some arrays/hashes ##
	my @br_ids = sort {$br2indx->{$a} <=> $br2indx->{$b} } keys %{$br2indx};
	my $ct_brs = @br_ids;

	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
	my %node2ct_exoffs; # ($node_id => #{exterior offsprings}, ...)
	my $ct_exnodes = 0;
	foreach my $node (sort { $node2depth->{$b} <=> $node2depth->{$a} } keys %{$node2depth}) { # Preliminary foreach-loop #
	    my $children = $node2ch->{$node};
	    if (defined $children) {
		my $ct = 0;
		foreach my $ch (@{$children}) { $ct += $node2ct_exoffs{$ch}; }
		$node2ct_exoffs{$node} = $ct;
	    } else {
		$node2ct_exoffs{$node} = 1;
		$ct_exnodes++;
	    }
	}		# End of the preliminary foreach-loop #

	my @row_rhs = ();
	for (1 .. $ct_brs) { push @row_rhs, 0; }
	my @mtrx_rhs = ();
	for (1 .. $ct_brs) { my @cp_row = @row_rhs; push @mtrx_rhs, \@cp_row; }

=for DEBUGGING 'rhs_qols_var_brlen'
	print STDERR "\n\n<< Subroutine 'rhs_qols_var_brlen' >>\n\n";
	print STDERR "< Inputs and Auxiliary variables >\n\n";
	print STDERR "Topnode ID: $topnode_id .\n";
	print STDERR '%{$br2indx} = (', "\n";
	foreach my $br (@br_ids) { print STDERR "\t", $br, ' => ', $br2indx->{$br}, "\,\n"; }
	print STDERR "\t\).\n\n";
	print STDERR '%node2ct_exoffs = (', "\n";
	foreach my $node (sort {$a <=> $b} keys %node2ct_exoffs) { print STDERR "\t", $node, ' => ', $node2ct_exoffs{$node}, "\,\n"; }
	print STDERR "\t\).\n\n";
=cut

		## Main Process ##

	for (my $n1 = 0; $n1 < $ct_brs; $n1++) { #  Outermost for-loop #
	    ##my $br1 = $br_ids[$n1]; ### The part that lead to a bug. ==> Moved below for DEBUGGING on May 7 (Fri), 2010.

	    for (my $n2 = 0; $n2 <= $n1; $n2++) { # Outer for-loop #

	        my $br1 = $br_ids[$n1]; ### Moved from above for DEBUGGING on May 7 (Fri), 2010.
		my $br2 = $br_ids[$n2];

		# Check the phylogenetic relationship between $br1 and $br2 #
		my ($common_anc, $anc_br1_spec, $anc_br2_spec) 
			= fetch_common_ancestors ($br1, $br2, %{$node2pa});
		unless (defined $common_anc) {
		    print STDERR "[ERROR 'rhs_qols_var_brlen'] 'fetch_common_ancestors' FAILED on the branch pair ($br1, $br2): $! !\n";
		    return undef;
		}

		my %br2ct_exnodes = %node2ct_exoffs; # ($branch_ID => #{exterior nodes under the branch}, ...)
		delete $br2ct_exnodes{$topnode_id};

			# When $br1 and $br2 are on the same lineage #
		if (@{$anc_br1_spec}==0) {
			# Always make $br1 an 'offspring' of $br2 #
		    unless ($br1 == $br2) {
			($br1, $br2) = ($br2, $br1);
			($anc_br1_spec, $anc_br2_spec) = ($anc_br2_spec, $anc_br1_spec);
		    }
		}

		if (@{$anc_br2_spec} == 0) {
			# Move $br2 from common ancestory to $br1-specific ancestory
		    unshift @{$anc_br1_spec}, (pop @{$common_anc});

			# Switch the definition of 'exterior offsprings' for the common ancestral branches #
		    foreach my $br (@{$common_anc}) { (defined $br2ct_exnodes{$br}) and ($br2ct_exnodes{$br} = $ct_exnodes - $br2ct_exnodes{$br}); }
		}

			# Set of branches on the path from $br1 to $br2 #
		my @path = reverse @{$anc_br1_spec};
		push @path, @{$anc_br2_spec};
		my $len_path = 0;
		foreach my $br (@path) { # 2nd preliminary foreach-loop #
		    my $attr = $br_attr->{$br};
		    my $br_len = $attr->[$INDX_LENGTH];
		    $len_path += $br_len;
		}			# End of the 2nd preliminary foreach-loop #

=for DEBUGGING 'rhs_qols_var_brlen'
	print STDERR "\n[ For the path bounded by \$br1 \= $br1 and \$br2 \= $br2  ]\n\n";
	print STDERR '@path = (', join (', ', @path), "\)  (length = $len_path) .\n\n";
	print STDERR '@{$anc_br2_spec} = (', join (', ', @{$anc_br2_spec}), "\).\n\n";
	print STDERR '%br2ct_exnodes = (', "\n";
	foreach my $br (sort {$a <=> $b} keys %br2ct_exnodes) { print STDERR "\t", $br, ' => ', $br2ct_exnodes{$br}, "\,\n"; }
	print STDERR "\t\).\n\n";
=cut

			## Don't let paths with non-positive lengths
			## contribute to the variances ##
		($len_path > 0) or next;

		my $size_path = @path;

			# Modified on May 13 (Tue), 2010 #
		my $cov_dist = ($alpha>0) ?
			(exp ($alpha * log(1 + $len_path/$alpha)) - 1)/$ct_sites :
			(exp ($len_path) - 1) /$ct_sites;
		##my $cov_dist = (exp ($len_path) - 1) /$ct_sites;

			# Sets of branches 'coming out' of the path
			# from the 'outlets' of $br1 and $br2.
		my @out_br1 = my @out_br2 = (); # element = [$ch_id, $order_ch, \@offsprings], where @offsprings contains $ch_id itself

		if (defined $node2ch->{$br1}) {
		    foreach my $ch (@{$node2ch->{$br1}}) { # 3rd preliminary foreach-loop #
			my $offsprings = fetch_all_offsprings_nr ($ch, %{$node2ch});
			my $order_ch = $br2ct_exnodes{$ch};
			push @out_br1, [$ch, $order_ch, $offsprings];
		    }			# End of the 3rd preliminary foreach-loop #
		}

		if (@{$anc_br2_spec}>0) {
		    if (defined $node2ch->{$br2}) {
			foreach my $ch (@{$node2ch->{$br2}}) { # 4th preliminary foreach-loop #
			    my $offsprings = fetch_all_offsprings_nr ($ch, %{$node2ch});
			    my $order_ch = $br2ct_exnodes{$ch};
			    push @out_br2, [$ch, $order_ch, $offsprings];
			}		# End of the 4th preliminary foreach-loop #
		    }
		} else {
			# Regard sibling branches as 'chilren' #
		    my $pa = $node2pa->{$br2};
		    foreach my $sibl (@{$node2ch->{$pa}}) { # 5th preliminary foreach-loop #
			($sibl == $br2) and next;
			my $offsprings = fetch_all_offsprings_nr ($sibl, %{$node2ch});
			my $order_sbl = $br2ct_exnodes{$sibl};
			push @out_br2, [$sibl, $order_sbl, $offsprings];
		    }			# End of the 5th preliminary foreach-loop #
		    if (defined $br_attr->{$pa}) {
			# Add the parent branch to the list of children as well #
			my $rec_offs = fetch_all_offsprings_nr ($pa, %{$node2ch});
			(undef, undef, my $offsprings) = diff_nmarray (@{$rec_offs}, @br_ids);
			unshift @{$offsprings}, $pa;
			my $order_pa = $br2ct_exnodes{$pa};
			push @out_br2, [$pa, $order_pa, $offsprings];
		    }
		}

=for DEBUGGING 'rhs_qols_var_brlen'
	print STDERR '@out_br1 = ( # [child, |child|, [offsprings]]', "\n";
	foreach my $info_ch (@out_br1) {
		my ($ch, $od_ch, $offs) = @{$info_ch};
		my $string_offs = join (', ', @{$offs});  
		print STDERR "\t\[", join (', ', $ch, $od_ch, "\[$string_offs\]"), "\]\,\n";
	}
	print STDERR "\t\).\n\n";
	print STDERR '@out_br2 = ( # [child, |child|, [offsprings]]', "\n";
	foreach my $info_ch (@out_br2) {
		my ($ch, $od_ch, $offs) = @{$info_ch};
		my $string_offs = join (', ', @{$offs});  
		print STDERR "\t\[", join (', ', $ch, $od_ch, "\[$string_offs\]"), "\]\,\n";
	}
	print STDERR "\t\).\n\n";
=cut

			# Pre-compute numerical factors that commonly appear #
		my $ct_ch_br1 = @out_br1;
		my $ct_ch_br2 = @out_br2;
		my $numfac_br1 = my $numfac_br2 = 0;
		my $order_br1 = my $order_br2 = 0;

		for (my $i=0; $i<$ct_ch_br1; $i++) { # 6th preliminary for-loop #
		    my $order_ch1 = $out_br1[$i]->[1];
		    $order_br1 += $order_ch1;
		    for (my $j=0; $j<$i; $j++) { # Inner for-loop #
			my $order_ch2 = $out_br1[$j]->[1];
			$numfac_br1 += $order_ch1 * $order_ch2;
		    }			  # End of the inner for-loop #
		}			# End of the 6th preliminary for-loop #
		$numfac_br1 *= 2;
		($numfac_br1>0) or ($numfac_br1 = 1);

		for (my $i=0; $i<$ct_ch_br2; $i++) { # 7 th preliminary for-loop #
		    my $order_ch1 = $out_br2[$i]->[1];
		    $order_br2 += $order_ch1;
		    for (my $j=0; $j<$i; $j++) { # Inner for-loop #
			my $order_ch2 = $out_br2[$j]->[1];
			$numfac_br2 += $order_ch1 * $order_ch2;
		    }			  # End of the inner for-loop #
		}			# End of the 7 th preliminary for-loop #
		$numfac_br2 *= 2;
		($numfac_br2>0) or ($numfac_br2 = 1);

=for DEBUGGING 'rhs_qols_var_brlen'
	print STDERR "(Numerical factor of \@out_br1) = $numfac_br1 ,\n";
	print STDERR "(Numerical factor of \@out_br2) = $numfac_br2 ,\n";
	print STDERR "(Order of \@out_br1) = $order_br1 ,\n";
	print STDERR "(Order of \@out_br2) = $order_br2 .\n\n";
=cut

			## Calculate the RHS of the covariance matrix equation ##

			# Both subject branches are on the path #

		my $term1 = $numfac_br1 * $numfac_br2 * $cov_dist;

		for (my $i=0; $i<$size_path; $i++) { # 1st middle for-loop #
		    my $br3 = $path[$i];
		    my $indx3 = $br2indx->{$br3};
		    $mtrx_rhs[$indx3]->[$indx3] += $term1;

		    for (my $j=0; $j< $i; $j++) { # Inner for-loop #
			my $br4 = $path[$j];
			my $indx4 = $br2indx->{$br4};
			$mtrx_rhs[$indx3]->[$indx4] += $term1;
			$mtrx_rhs[$indx4]->[$indx3] += $term1;
		    }			# End of the inner for-loop #
		}			# End of the 1st middle for-loop #

			# One is on the path, and the other is in @out_br1.

		for (my $i = 0; $i <$ct_ch_br1; $i++) { # 2nd middle for-loop #
		    my ($ch4, $order_ch4, $offsprings4) = @{$out_br1[$i]};
		    my $sum_od_ch3 = $order_br1 - $order_ch4;

		    my $mltpl = $numfac_br2 * $sum_od_ch3 * $cov_dist;

		    foreach my $br4 (@{$offsprings4}) { # Main inner for-loop #
			my $indx4 = $br2indx->{$br4};
			my $order_br4 = $br2ct_exnodes{$br4};
			my $term = $mltpl * $order_br4;

			foreach my $br3 (@path) { # Innermost foreach-loop #
			    my $indx3 = $br2indx->{$br3};
			    $mtrx_rhs[$indx3]->[$indx4] += $term;
			    $mtrx_rhs[$indx4]->[$indx3] += $term;
			}		# End of the innermost foreach-loop #
		    }				# End of the main inner for-loop #
		}				# End of the 2nd middle for-loop #


			# One is on the path, and the other is in @out_br2.

		for (my $i = 0; $i <$ct_ch_br2; $i++) { # 3rd middle for-loop #
		    my ($ch4, $order_ch4, $offsprings4) = @{$out_br2[$i]};
		    my $sum_od_ch3 = $order_br2 - $order_ch4 ;

		    my $mltpl = $numfac_br1 * $sum_od_ch3 * $cov_dist;

		    foreach my $br4 (@{$offsprings4}) { # Main inner for-loop #
			my $indx4 = $br2indx->{$br4};
			my $order_br4 = $br2ct_exnodes{$br4};
			my $term = $mltpl * $order_br4;

			foreach my $br3 (@path) { # Innermost foreach-loop #
			    my $indx3 = $br2indx->{$br3};
			    $mtrx_rhs[$indx3]->[$indx4] += $term;
			    $mtrx_rhs[$indx4]->[$indx3] += $term;
			}		# End of the innermost foreach-loop #
		    }				# End of the main inner for-loop #
		}				# End of the 3rd middle for-loop #


			# Both are in @out_br1.

		my $mltpl = $numfac_br2 * $cov_dist;
		for (my $i=0; $i<$ct_ch_br1; $i++) { # 4th middle for-loop #
		    my $offsprings3 = $out_br1[$i]->[2];
		    for (my $j=0; $j<$i; $j++) { # Inner-middle for-loop #
			my $offsprings4 = $out_br1[$j]->[2];

			foreach my $br3 (@{$offsprings3}) { # Inner for-loop #
			    my $indx3 = $br2indx->{$br3};
			    my $order_br3 = $br2ct_exnodes{$br3};
			    foreach my $br4 (@{$offsprings4}) { # Innermost for-loop #
				my $indx4 = $br2indx->{$br4};
				my $order_br4 = $br2ct_exnodes{$br4};
				my $term = $mltpl * $order_br3 * $order_br4;
				$mtrx_rhs[$indx3]->[$indx4] += $term;
				$mtrx_rhs[$indx4]->[$indx3] += $term;
			    }			     # End of the innermost for-loop #
			}			# End of the inner for-loop #
		    }			# End of the inner-middle for-loop #
		}			# End of the 4th middle for-loop #


			# Both are in @out_br2.

		$mltpl = $numfac_br1 * $cov_dist;
		for (my $i=0; $i<$ct_ch_br2; $i++) { # 5th middle for-loop #
		    my $offsprings3 = $out_br2[$i]->[2];
		    for (my $j=0; $j<$i; $j++) { # Inner-middle for-loop #
			my $offsprings4 = $out_br2[$j]->[2];

			foreach my $br3 (@{$offsprings3}) { # Inner for-loop #
			    my $indx3 = $br2indx->{$br3};
			    my $order_br3 = $br2ct_exnodes{$br3};
			    foreach my $br4 (@{$offsprings4}) { # Innermost for-loop #
				my $indx4 = $br2indx->{$br4};
				my $order_br4 = $br2ct_exnodes{$br4};
				my $term = $mltpl * $order_br3 * $order_br4;
				$mtrx_rhs[$indx3]->[$indx4] += $term;
				$mtrx_rhs[$indx4]->[$indx3] += $term;
			    }			     # End of the innermost for-loop #
			}			# End of the inner for-loop #
		    }			# End of the inner-middle for-loop #
		}			# End of the 5th middle for-loop #


			# One is in @out_br1, and the other is in @out_br2.

		for (my $i=0; $i<$ct_ch_br1; $i++) { # 6th middle for-loop #
		    my ($ch3, $order_ch3, $offsprings3) = @{$out_br1[$i]};
		    my $sum_od_ch3 = $order_br1 - $order_ch3;

		    for (my $j=0; $j<$ct_ch_br2; $j++) { # Inner-middle for-loop #
			my ($ch4, $order_ch4, $offsprings4) = @{$out_br2[$j]};
			my $sum_od_ch4 = $order_br2- $order_ch4;
			my $mltpl = $sum_od_ch3 * $sum_od_ch4 * $cov_dist;

			foreach my $br3 (@{$offsprings3}) { # Inner for-loop #
			    my $indx3 = $br2indx->{$br3};
			    my $order_br3 = $br2ct_exnodes{$br3};
			    foreach my $br4 (@{$offsprings4}) { # Innermost for-loop #
				my $indx4 = $br2indx->{$br4};
				my $order_br4 = $br2ct_exnodes{$br4};
				my $term = $mltpl * $order_br3 * $order_br4;
				$mtrx_rhs[$indx3]->[$indx4] += $term;
				$mtrx_rhs[$indx4]->[$indx3] += $term;
			    }			    # End of the innermost for-loop #
			}			# End of the inner for-loop #
		    }			     # End of the inner-middle for-loop #
		}			# End of the 6th middle for-loop #

	    }				# End of the outer for-loop #
	}		# End of the outermost for-loop #

=for DEBUGGING 'rhs_qols_var_brlen'
	print STDERR "\n --- 'rhs_qols_var_brlen' ENDED !! ---\n\n";
=cut

	return \@mtrx_rhs;

}	# End of the subroutine 'rhs_qols_var_brlen' #


# qo_leastsq_var_brlen:
#	This subroutine estimates THE VARIANCE OF the branch lengths by means of
#	a least squares method UNDER the QUASI-ORDINARY LS (qOLS) framework.
#	In the qOLS framework, the following sum of squares is minimized:
#	Sum_{(i,j)} [Exp(i,j) - Obs(i,j)]**2 /[ 1 + Obs(i,j)/ alpha ],
#	where Obs(i,j) is the observed value of the distance between the sequences i and j;
#	Exp(i,j) = Sum_{br(m) in path(i,j)} l(m)  is the expected vale of the distance
#	under a given set {l(m)} of the branch lengths;
#	alpha (>0) is the shape parameter for the gamma model of amino acid substitutions.
#
#	The subroutine was made by revising 'leastsq_brlen' in the module 'MyTreeManipu_VI.pm'.
#
# INPUT: (\@seqnames, $ct_sites, \@dists, \%name2id, \%node2parent, \%node2children, $alpha, \%in_br_attr, $fl_ls_dists), where
#	$seqnames[$i] is the name (label) of the $i th sequence;
#	$ct_sites is the number of sites in the MSA used for constructing the distance matrix;
#	$dists[$i]->[$j] is the distance between the $i th and $j th sequences
#	(with 0 <= $j < $i < #{sequences});
#	%name2id = ($seq_name => $node_id, ...), with $seq_name has to belong to @seqnames;
#	%node2parent = ($node_id => $parent_node_id, ...);
#	%node2children = ($node_id => \@child_node_ids, ...);
#	$alpha is the shape parameter for the gamma model of amino acid substitutions;
#	%in_br_attr = ($node_id => \@attr, ...) for the nodes each immediately below a branch,
#	and $attr[$INDX_LENGTH] gives the length of the branch in question;
#	$fl_ls_dists = 0 if branch lengths from %in_br_attr are used,
#	           = 1 if branch lengths are estimated here via the qOLS method.
#
# OUTPUT: (\%var_brlen, \%br_attr), where
#	%var_brlen = (ID of the node immediately below the branch => variance_branch_length, ...);
#	%br_attr = (ID of the node immediately below the branch => [branch length], ...).
#
sub qo_leastsq_var_brlen (\@$\@\%\%\%$\%$) {

	my ($seqnames, $ct_sites, $dists, $name2id, $node2parent, $node2children, $alpha, $in_br_attr, $fl_ls_dists) = @_;

	if (@{$seqnames}<3) {
	    my $ct_seqs = @{$seqnames};
	    print STDERR "[ERROR 'qo_leastsq_var_brlen'] The number of sequences ($ct_seqs) is NOT ENOUGH for this subroutine to work !\n";
	    return (undef, undef);
	}
	if ($alpha < 0) {
	    print STDERR "[ERROR 'qo_leastsq_var_brlen'] ILLEGAL INPUT: A NEGATIVE shape parameter ($alpha): $! !\n";
	    return (undef, undef);
	}

		## ADDED in creation of 'qo_leastsq_var_brlen' (on May 5 (Wed), 2010) ##
	my $topnode_id = $name2id->{$TOP_NODE};

		# Create an array, while checking the consistency
		# between the distance matrix and the tree.
	my @nodeids = (); # = ($node_id, ...)
	foreach my $seqname (@{$seqnames}) { # 1st foreach-loop #
	   my $nodeid = $name2id->{$seqname};
	   unless (defined $nodeid) {
		print STDERR "[ERROR 'qo_leastsq_var_brlen'] The sequence '$seqname' has NO NODE ID!\n";
		return (undef, undef);
	   }
	   push @nodeids, $nodeid
	}			# End of the 1st foreach-loop #

=for DEBUGGING 'qo_leastsq_var_brlen'
	print STDERR "<< Subroutine 'qo_leastsq_var_brlen' >>\n\n";
	print STDERR "Topnode = $topnode_id .\n";
	print STDERR '@nodeids = (', join (', ', @nodeids), "\).\n\n";
=cut

	my %br_attr; # (ID of the node immediately below the branch => \@attr, ...)

	# %{$br2indx} = (ID of the node immediately below the branch => $indx, ...)
	#
	my $br2indx = assign_branch_indices (%{$node2parent}, %br_attr);

=for DEBUGGING 'qo_leastseq_var_brlen'
	print STDERR "< Output of 'assign_branch_indices' >\n\n";
	print STDERR '%{$br2indx} = (', "\n";
	foreach my $br (sort { $br2indx->{$a} <=> $br2indx->{$b} } keys %{$br2indx}) { print STDERR "\t", $br, ' => ', $br2indx->{$br}, "\,\n"; }
	print STDERR "\t\).\n\n";
=cut

		## BUG FIXING on Monday, March 5, 2007 ##
		## In the case where all the branches are trivially zero ##
	(0 < scalar keys %{$br2indx}) or (return \%br_attr);

		## REVISED in creation of 'qo_leastsq_brlen' (on May 5 (Wed), 2010) ##
	my ($coeffs, $rhs) = coeffs_qo_leastsq_brlen (@{$dists}, @nodeids,
		%{$br2indx}, %{$node2parent}, $alpha);
	##my ($coeffs, $rhs) = coeffs_leastsq_brlen (@{$dists}, @nodeids,
	##	%{$br2indx}, %{$node2parent}, $power, $buffer);

	unless (defined $coeffs) {
	    print STDERR "[ERROR 'qo_leastsq_var_brlen'] 'coeffs_qo_leastsq_brlen' FAILED !\n";
	    return (undef, undef);
	}

	my $size_vec = @{$rhs};

=for DEBUGGING 'qo_leastsq_var_brlen'
	print STDERR "< Outputs of 'coeffs_qo_leastsq_brlen' >\n\n";
	print STDERR '@{$coeffs} = (', "\n";
	foreach my $row (@{$coeffs}) { print STDERR "\t\[", join (', ', @{$row}), "\]\,\n"; }
	print STDERR "\t\).\n\n";
	print STDERR '@{$rhs} = (', join (', ', @{$rhs}), "\).\n\n";
=cut

		## REVISED in creation of 'qo_leastsq_var_brlen' (on May 5 (Wed), 2010) ##
	my $inv_coeffs = invert_matrix_lu (@{$coeffs});
	unless (defined $inv_coeffs) {
	    print STDERR "[ERROR 'qo_leastsq_var_brlen'] 'invert_matrix_lu' FAILED !\n";
	    return (undef, undef);
	}

=for DEBUGGING 'qo_leastsq_var_brlen'
	print STDERR "< Output of 'invert_matrix_lu' >\n\n";
	print STDERR '@{$inv_coeffs} = (', "\n";
	foreach my $row (@{$inv_coeffs}) { print STDERR "\t\[", join (', ', @{$row}), "\]\,\n"; }
	print STDERR "\t\).\n\n";
=cut

=for the predecessor 'qo_leastsq_brlen'
	# @{$br_lens} = (branch lengths arranged in the order of the assigned index.
	my $br_lens = solve_lin_eq_lu (@{$coeffs}, @{$rhs});
	unless (defined $br_lens) {
	    print STDERR "[ERROR 'qo_leastsq_brlen'] 'solve_lin_eq_lu' FAILED !\n";
	    return undef;
	}
=cut

	if ($fl_ls_dists) {
		# Cast the LS-estimated branch lengths into the form
		# commonly used in my modules.
	    foreach my $dwnode (keys %{$br2indx}) { # 2nd outer foreach-loop #
		my $indx = $br2indx->{$dwnode};
		(defined $indx) or next;

		  ## REVISED in creation of 'qo_leastsq_var_brlen'
		  ## (on May 5 (Wed), 2010)
		my $inv_i = $inv_coeffs->[$indx];
		my $br_len = 0;
		for (my $k=0; $k<$size_vec; $k++) { $br_len += $rhs->[$k] * $inv_i->[$k]; }
=for the predecessor 'qo_leastsq_brlen'
		my $br_len = $br_lens->[$indx];
		unless (defined $br_len) {
		    print STDERR "[ERROR 'qo_leastsq_brlen'] The branch immediately above the node '$dwnode' has NO BRANCH LENGTH estimated !\n"; 
		    return undef;
	 	}
=cut

		my @tmp = ();
		$tmp[$INDX_LENGTH] = $br_len;
		$br_attr{$dwnode} = \@tmp;
	    }			# End of the 2nd outer foreach-loop #

	} else {
	    %br_attr = %{$in_br_attr};
	}

		## Modified on May 13 (Tue), 2010 ##
	my $mtrx_rhs = rhs_qols_var_brlen ($topnode_id, %{$br2indx}, %{$node2parent}, %{$node2children}, %br_attr, $ct_sites, $alpha) ;
	##my $mtrx_rhs = rhs_qols_var_brlen ($topnode_id, %{$br2indx}, %{$node2parent}, %{$node2children}, %br_attr, $ct_sites) ;
	unless (defined $mtrx_rhs) {
	    print STDERR "[ERROR 'coeffs_qols_var_brlen'] 'rhs_qols_var_brlen' FAILED !\n";
	    return (undef, undef);
	}

=for DEBUGGING the subroutine 'qo_leastsq_var_brlen'
	print STDERR "\n< Outputs of 'rhs_qols_var_brlen' >\n\n";
	print STDERR '@{$mtrx_rhs} = (', "\n";
	foreach my $row (@{$mtrx_rhs}) { print STDERR "\t\[", join (', ', @{$row}), "\]\,\n"; }
	print STDERR "\t).\n\n";
=cut

	my %node2var_brlen;
	foreach my $dwnode (keys %{$br2indx}) { # 3rd outer foreach-loop #
	    my $indx = $br2indx->{$dwnode};
	    (defined $indx) or next;
	    my $inv_i = $inv_coeffs->[$indx];
	    my $var_brlen = 0;
	    for (my $k=0; $k<$size_vec; $k++) { # Middle for-loop #
		my $rhs_k = $mtrx_rhs->[$k];
		my $inv_ik = $inv_i->[$k];
		$var_brlen += $rhs_k->[$k] * $inv_ik * $inv_ik;
		for (my $l=0; $l < $k; $l++) { $var_brlen += 2 * $rhs_k->[$l] * $inv_i->[$l] * $inv_ik; }
	    }				# End of the middle for-loop #
	    $node2var_brlen{$dwnode} = $var_brlen;
	}			# End of the 3rd outer foreach-loop #

=for DEBUGGING the subroutine 'qo_leastsq_var_brlen'
	print STDERR "\n --- Subroutine 'qo_leastsq_var_brlen' ENDED !! ---\n\n";
=cut

	return (\%node2var_brlen, \%br_attr);

}		# end of the subroutine 'qo_leastsq_var_brlen' #



=set aside for a potential use
# assign_trivial_brlen:
#	This subroutine skims through the distance matrix and assigns  
#	ZERO LENGTHS to the branches comprizing the path between
#	the sequence pairs with NON-POSITIVE distances.
#
# INPUT: (\@dists, \@nodeids, \%node2parent, \%br_attr), where
#	$dists[$i]->[$j] is the distance between the $i th and $j th sequences
#	(with 0 <= $j < $i <= #{sequences});
#	$nodeids[$i] is the node ID of the $i th sequence;
#	%node2parent = ($node_id => $parent_node_id, ...);
#	%br_attr = (ID of the node immediately below the branch => \@attr, ...)
#	 with $attr[$INDX_LENGTH] is the branch length,
#	this hash is ASSUMED to be empty when input.
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
sub assign_trivial_brlen (\@\@\%\%) {

	my ($dists, $nodeids, $node2parent, $br_attr) = @_; 

	my $ct_seqs = @{$nodeids};

##=for TESTING the subroutine
	print STDERR "\n\< Subroutine 'assign_trivial_brlen' \>\n\n";
##=cut

	for (my $i=1; $i<$ct_seqs; $i++) { # The outer for-loop #
	    my $node1 = $nodeids->[$i];
	    my $row_dists = $dists->[$i];
	    for (my $j=0; $j < $i; $j++) { # The middle for-loop #
		my $node2 = $nodeids->[$j];
		my $dist = $row_dists->[$j];
		($dist >0) and next;

			# The pair has a non-positive distance #
		my ($cm_anc, $lin1, $lin2) = fetch_common_ancestors ($node1, $node2, %{$node2parent});
		unless (defined $cm_anc) {
		    print STDERR "[ERROR 'assign_trivial_brlen'] 'fetch_common_ancestors' FAILED for the node pair \($node1, $node2\) !\n";
		    return (undef, undef);
		}

##=for TESTING the subroutine
		print STDERR "\n{ Node pair \($node1, $node2\) }\n\n";
		print STDERR "distance = $dist, powered_distance = $pwdist,\n";
		print STDERR "    reciprocal_denominator = $rec_den .\n\n";
		print STDERR "Branches on the path: ", join (', ', @{$lin1}, @{$lin2}), ".\n\n";
##=cut

		foreach my $br_id (@{$lin1}, @{$lin2}) { # The inner foreach-loop #
		   (defined $br_attr->{$br_id}) and next;
			# Don't touch already given attributes

			# Assign zero for the branch length #
		   my @attr = ();
		   $attr[$INDX_LENGTH] = 0;
		   $br_attr->{$br_id} = \@attr;
		}			# End of the inner foreach-loop #
		
	    }			# End of the middle for-loop #
	}		# End of the outer for-loop #

	return 1;

}		# End of the subroutine 'assign_trivial_brlen' #
=cut


=set aside for potential use
# leastsq_brlen_nn:
#	This revised version of the subroutine 'leastsq_brlen'
#	basically do the same job as the predecessor,
#	except that this version KEEPS branch lengths NONNEGATIVE.
#	It forcefully assigns zero length to the branches
#	whose lengths are estimated to be negative,
#	then recalculate the remaining lengths.
#
# leastsq_brlen:
#	This subroutine estimates branch lengths by means of
#	a least square method like that used for the program 'fitch'
#	in the Phylip package.
#	The subroutine minimizes the sum of squares:
#	Sum_{(i,j)} [Exp(i,j) - Obs(i,j)]**2 /[(Obs(i,j))**p + (buffer)**p],
#	where Obs(i,j) is the observed value of the distance between the sequences i and j;
#	Exp(i,j) = Sum_{br(m) in path(i,j)} l(m)  is the expected vale of the distance
#	under a given set {l(m)} of the branch lengths;
#	p is the 'power' parameter (p=2 for Fitch & Margoliash's algorithm,
#	p=0 for the Neighbor-Joining method);
#	buffer is an 'effective minimum scale', which avoids divergence
#	even when the observed distance is extremely small.
#
# INPUT: (\@seqnames, \@dists, \%name2id, \%node2parent, $power, $buffer), where
#	$seqnames[$i] is the name (label) of the $i th sequence;
#	$dists[$i]->[$j] is the distance between the $i th and $j th sequences
#	(with 0 <= $j < $i <= #{sequences});
#	%name2id = ($seq_name => $node_id, ...), with $seq_name has to belong to @seqnames;
#	%node2parent = ($node_id => $parent_node_id, ...);
#	$power and $buffer corresponds, respectively, to 'p' and 'buffer'
#	in the above description of the least .
#
# OUTPUT: \%br_attr, where
#	%br_attr = (ID of the node immediately below the branch => [branch length], ...).
#
sub leastsq_brlen_nn (\@\@\%\%$$) {

	my ($seqnames, $dists, $name2id, $node2parent, $power, $buffer) = @_;

	if (@{$seqnames}<3) {
	    my $ct_seqs = @{$seqnames};
	    print STDERR "[ERROR 'leastsq_brlen_nn'] The number of sequences ($ct_seqs) is NOT ENOUGH for this subroutine to work !\n";
	    return undef;
	}

		# Create an array, while checking the consistency
		# between the distance matrix and the tree.
	my @nodeids = (); # = ($node_id, ...)
	foreach my $seqname (@{$seqnames}) {
	   my $nodeid = $name2id->{$seqname};
	   unless (defined $nodeid) {
		print STDERR "[ERROR 'leastsq_brlen_nn'] The sequence '$seqname' has NO NODE ID!\n";
		return undef;
	   }
	   push @nodeids, $nodeid
	}

	my %br_attr; # (ID of the node immediately below the branch => \@attr, ...)

		# First assign zero to the lengths of branches comprising
		# the paths with distance zero.
		#
	assign_trivial_brlen (@{$dists}, @nodeids, %{$node2parent}, %br_attr);

	my $round = 0;
    while (1) { # An infinite while-loop ! #
	$round++;

##=for TESTING the subroutine
	print STDERR "\n[ROUND $round]\n\n";
	print STDERR "Branches with ZERO lengths: ", join (', ', sort keys %br_attr), " .\n\n";
##=cut

	# %{$br2indx} = (ID of the node immediately below the branch => $indx, ...)
	#
	my $br2indx = assign_branch_indices (%{$node2parent}, %br_attr);

		## BUG FIXING on Monday, March 5, 2007 ##
		## In the case where all the branches are trivially zero ##
	(0 < scalar keys %{$br2indx}) or (return \%br_attr);

	my ($coeffs, $rhs) = coeffs_leastsq_brlen (@{$dists}, @nodeids,
		%{$br2indx}, %{$node2parent}, $power, $buffer);

	unless (defined $coeffs) {
	    print STDERR "[ERROR 'leastsq_brlen_nn'] 'coeffs_leastsq_brlen' FAILED  in the $round th round!\n";
	    return undef;
	}

	# @{$br_lens} = (branch lengths arranged in the order of the assigned index.
	my $br_lens = solve_lin_eq_lu (@{$coeffs}, @{$rhs});
	unless (defined $br_lens) {
	    print STDERR "[ERROR 'leastsq_brlen_nn'] 'solve_lin_eq_lu' FAILED in the $round th round!\n";
	    return undef;
	}

		# Cast the estimated branch lengths into the form
		# commonly used in my modules.

		# First, look for NEGATIVE branch lengths,
		# and, if any, set them to zero.
		#
	my $ct_nega = 0;
	foreach my $dwnode (keys %{$br2indx}) { # The last foreach-loop #
	    my $indx = $br2indx->{$dwnode};
	    (defined $indx) or next;
	    my $br_len = $br_lens->[$indx];
	    unless (defined $br_len) {
		print STDERR "[ERROR 'leastsq_brlen_nn'] The branch immediately above the node '$dwnode' has NO BRANCH LENGTH estimated (in the $round th round) !\n"; 
		return undef;
	    }
	    ($br_len <0) or next;

	    $ct_nega++;
	    my @tmp = ();
	    $tmp[$INDX_LENGTH] = 0;
	    $br_attr{$dwnode} = \@tmp;
	}			# End of the last foreach-loop #

	($ct_nega>0) and next; # If there is any negative branch lengths,
				# recalculate the remaining ones.

		# Second, stuff the remaining branch lengths #
		#
	foreach my $dwnode (keys %{$br2indx}) { # The last foreach-loop #
	    (defined $br_attr{$dwnode}) and next;
	    my $indx = $br2indx->{$dwnode};
	    (defined $indx) or next;

	    my $br_len = $br_lens->[$indx];
	    my @tmp = ();
	    $tmp[$INDX_LENGTH] = $br_len;
	    $br_attr{$dwnode} = \@tmp;
	}			# End of the last foreach-loop #

	return \%br_attr;

    }			# End of the infinite while-loop # 

}		# end of the subroutine 'leastsq_brlen_nn' #
=cut


# neighbor_joining_tree:
#	This subroutine implements the neighbor-joining (NJ) method of phylogenetic tree reconstruction.
#	The subroutine is made by revising 'extended_upgma_tree' in the module 'MyTreeManipu_VI.pm'.
#	One big difference from the predecessor is that
#	this subroutine assumes that the distances are defined for ALL the pairs of OTUs.
#
#	ALGORITHM:
#	Assume that there are m OTUs and that the distances are defined for all the pairs.
#	The original NJ method compares the following values for all possible pairs of OTUs at each iteration step:
#	S(i,j) = 1/{2*(m-2)} * Sum_{k=1,..., m; k ne i or j} (dik + djk) + 1/2 * dij + 1/(m-2) * Sum_{1 <= k < l <=m; k, l ne i or j} dkl,
#	which is the sum of all branches (estimated by an ordinary least square (OLS) method) under an almost star-phylogeny where the pair (i,j) is separated from others via a branch. Because this is rewritten as:
#	S(i,j) = 1/{2*(m-2)} * [ (m-2) * dij - R(i) - R(j) + 2*T],
#	where T = Sum_{1<= k < l <=m } dkl is independent of i or j, and
#	R(i) = Sum_{1<= k <=m ; k ne i} dik .
#	There for the method is equivalent to identifying the pair (i,j) that minimize Q(i,j) given by:
#	Q(i,j) = (m-2) * dij - R(i) - R(j).
#	The calculation and comparision of Q(i,j) are done much faster than
#	the naive calculation and comparison of S(i,j).
#
# extended_upgma_tree:
#	This subroutine implements an extended version of the
#	UPGMA (unweighted pair-group method using arithmetic averages) algorithm
#	for tree construction.
#	This extended algorithm allows a distance matrix to have undefined elements.
#
# INPUT: (\@taxa, \@pw_distances), where
#	$taxa[$i] is the ID or name of the $i+1 th taxon ($i=0, ..., #{taxa} -1);
#	$pw_distance[$i]->[$j] (0<=$j< $i) is the pairwise distance between $taxa[$i] and $taxa[$j].
#	The element $pw_distance[0] is the reference to an empty array;
#
# OUTPUT: (\%name2nodeid, \%nodeid2name, \%node2pa, \%node2chi, \%br_attr), where
#	%name2nodeid = ($taxon => $node_id, ...) for $TOP_NODE and exterior taxa (given by @taxa);
#	%nodeid2name = ($node_ID => $taxon, ...) for exterior taxa;
#	%node2pa = ($node_ID => $parent_ID, ...) for all nodes other than the top node;
#	%node2chi = ($node_ID => \@child_IDs, ...) ;
#	%br_attr = ($ID of the nearest downstream node => [$length, $bootstrap (=undef)], ...);
#
#	CONVENTION for the NODE IDs:
#	The node ID $i+1 is assigned to the exterior taxon $taxa[$i] ($i = 0, ... #{taxa}-1),
#	then each interior taxon (or node) is assigned with a node ID that begins with $ct_taxa,
#	and that is incremented by 1 as the a new taxon (=cluster) is created.
#
sub neighbor_joining_tree (\@\@) {

	my ($taxa, $pw_distances) = @_;

	my $ct_taxa = @{$taxa};

		## REVISED in creation of 'neighbor_joining_tree' (on May 7 (Fri), 2010) ##
	unless ($ct_taxa>2) {
	##unless ($ct_taxa>0) {
	    print STDERR "[ERROR 'neighor_joining_tree'] ILLEGAL INPUT: The NUMBER of TAXA ($ct_taxa) is TOO SMALL: $! !\n";
	    return (undef, undef, undef, undef, undef);
	}

	unless ($ct_taxa == @{$pw_distances}) {
	    print STDERR "\n[ERROR 'neighbor_joining_tree'] ILLEGAL INPUT: The number of ROWS (", scalar (@{$pw_distances}), ") CLASHES with the number of TAXA ($ct_taxa) : $! !\n";
	    return (undef, undef, undef, undef, undef);
	}

		# Initialize the setting #
	##my @mtrx_ct_dd_pairs = ();
	my @r = ();
	my %nodeid2name;
	for (my $i=0; $i<$ct_taxa; $i++) { # The preliminary outer for-loop #
	    $nodeid2name{$i+1} = $taxa->[$i];

	    my $ri = 0;
	    for (my $j=0; $j<$ct_taxa; $j++) { # Inner for-loop #
		($j == $i) and next;
		my $pw_dist = ($i>$j) ? $pw_distances->[$i][$j] : $pw_distances->[$j][$i];
		$ri += $pw_dist;
	    }
	    push @r, $ri;
=for the predecessor 'extended_upgma_tree'
	    my $row_pw_distances = $pw_distances->[$i];
	    my @row_ct_dd_pairs = ();
	    for (my $j = 0; $j<$i; $j++) { # Inner for-loop #
		my $pw_dist = $row_pw_distances->[$j];
		push @row_ct_dd_pairs, ((defined $pw_dist) and ($pw_dist ne $str_na)) ?  1 : 0;
	    }				# End of the inner for-loop #
	    push @mtrx_ct_dd_pairs, \@row_ct_dd_pairs;
=cut
	}			# End of the preliminary outer for-loop #

	my @sbjct_nodes = (1..$ct_taxa); # list of subject nodes in exactly the same order as in the distance matrix
	my $max_nodeid = $ct_taxa;

	my @pw_distances = @{$pw_distances};

	my %node2pa; my %node2chi; my %br_attr;

	### my %node2level; # ($node_ID => 'level' from the bottom, ...), which is an auxiliary hash

	while (1) { # The main outer while-loop #

	    $max_nodeid++; ## MOVED from below (on May 7 (Fri), 2010)

	    my $ct_sbjcts = @sbjct_nodes;

		## ADDED in creation of 'neighbor_joining_tree' on May 7 (Fri), 2010 ##
	    ($ct_sbjcts > 3) or last; ## No need to cluster a pair of OTUs any longer 

		# Search for the closest pair #

	    my @closest_pair = (); # = ($indx_node1, $indx_node2, $q_ij, $distance) where $indx_node1/2 is the index in the distance matrix.
	    ##my @closest_pair = (); # = ($indx_node1, $indx_node2, $distance) where $indx_node1/2 is the index in the distance matrix.
	    for (my $i=0; $i<$ct_sbjcts; $i++) { # The 1st middle for-loop #
		my $row_pw_distances = $pw_distances[$i];
		my $ri = $r[$i];
		for (my $j=0; $j<$i; $j++) { # inner for-loop #
		    my $pw_dist = $row_pw_distances->[$j];
		    my $rj = $r[$j];

=for the predecessor 'extended_upgma_tree'
			# Skip pairs with undefined or unavailable distances #
		    (defined $pw_dist) or next;
		    ($pw_dist eq $str_na) and next;
=cut

		    my $q_ij = ($ct_sbjcts -2) * $pw_dist - $ri - $rj;

		    unless (@closest_pair>0) { # 'Initialize' the closest pair
			@closest_pair = ($i, $j, $q_ij, $pw_dist);
			##@closest_pair = ($i, $j, $pw_dist);
			next;
		    }
		    ($q_ij < $closest_pair[2]) or next; # No need to update
		    ##($pw_dist < $closest_pair[2]) or next; # No need to update

			# Update the closest pair
		     @closest_pair = ($i, $j, $q_ij, $pw_dist);
		     ##@closest_pair = ($i, $j, $pw_dist);
		
		}			# End of the inner for-loop #
	    }				# End of the 1st middle for-loop #

	    ## (@closest_pair>0) or last; # We cannot cluster the nodes any more. #

		# Cluster the closest pair #

	    my ($i, $j, $q_ij, $pw_dist) = @closest_pair;
	    ## my ($i, $j, $pw_dist) = @closest_pair;
	    ## $max_nodeid++; ## Moved upward (on May 7 (Fri), 2010)

	    ##$node2ct_dd_pairs{$max_nodeid} = $mtrx_ct_dd_pairs[$i]->[$j];

	    ##my @clustered = @sbjct_nodes[$i,$j];
	    my @clustered = @sbjct_nodes[$j,$i]; # Just for a better visualization

	    $node2chi{$max_nodeid} = \@clustered;
	    foreach my $nodeid (@clustered) { # The 2nd middle foreach-loop #
		$node2pa{$nodeid} = $max_nodeid;
=for the predecessor 'extended_upgma_tree'
		my $br_len = my $level = $pw_dist/2;
		$node2level{$nodeid} = $level;
		my $children = $node2chi{$nodeid};
		((defined $children) and (@{$children}>0))
		    and ($br_len -= $node2level{$children->[0]});
		my @attr = ();
		$attr[$INDX_LENGTH] = $br_len;
		$br_attr{$nodeid} = \@attr;
=cut
	    }			# End of the 2nd middle foreach-loop #

		## ADDED in creation of 'neighbor_joining_tree' (on May 7 (Fri), 2010) ##

	    my ($ri, $rj) = @r[$i, $j];
	    my $br_iA = ($pw_dist + ($ri - $rj)/($ct_sbjcts -2))/2;
	    my $br_jA = ($pw_dist + ($rj - $ri)/($ct_sbjcts -2))/2;
	    foreach my $br_info ([$sbjct_nodes[$i], $br_iA], [$sbjct_nodes[$j], $br_jA]) {
		my ($br_id, $br_len) = @{$br_info};
		my @attr = ();
		$attr[$INDX_LENGTH] = $br_len;
		$br_attr{$br_id} = \@attr;
	    }
	    

=for the predecessor 'extended_upgma_tree'
	    unless ($ct_sbjcts > 2) { # We do not have the subjects to cluster any longer.
		@sbjct_nodes = ($max_nodeid);
		last;
	    }
=cut

		# Update the subject nodes &
		# matrices for distances & #{distance-defined pairs} #

	    my @new_sbjct_nodes = my @new_pw_distances = my @new_r = ();

	    my @mgd_row_pw_distances = ();
	    ## my @new_mtrx_ct_dd_pairs = (); my @mgd_row_ct_dd_pairs = (); ## For the predecessor 'extended_upgma_tree' ##

		# merged information on the $i th and $j th subjects #

	    for (my $k=0; $k< $ct_sbjcts; $k++) { # The 3rd middle for-loop #
		(($k == $i) or ($k == $j)) and next; # skip the clustered subjects #

			# squeeze pairs of remaining taxa into the new arrays #

		push @new_sbjct_nodes, $sbjct_nodes[$k];

		my $row_pw_distances = $pw_distances[$k];
		my @new_row_pw_distances = ();

=For the predecessor 'extended_upgma_tree'
		my $row_ct_dd_pairs = $mtrx_ct_dd_pairs[$k];
		my @new_row_ct_dd_pairs = ();
=cut
		for (my $l=0; $l<$k; $l++) { # The inner for-loop #
		    (($l == $i) or ($l == $j)) and next; # Skip the clustered subjects

		    push @new_row_pw_distances, $row_pw_distances->[$l];
		    ## push @new_row_ct_dd_pairs, $row_ct_dd_pairs->[$l]; ## For the predecessor 'extended_upgma_tree'
		}			# End of the inner for-loop #

		push @new_pw_distances, \@new_row_pw_distances;
		## push @new_mtrx_ct_dd_pairs, \@new_row_ct_dd_pairs; ## For the predecessor 'extended_upgma_tree'

			# pairs containing a new (merged) taxon #

		my $dist_ik = ($k>$i) ? $pw_distances[$k]->[$i] : $pw_distances[$i]->[$k];
		my $dist_jk = ($k>$j) ? $pw_distances[$k]->[$j] : $pw_distances[$j]->[$k];

###=ADDED in creation of 'neighbor_joining_tree' (on May 7 (Fri), 2010)

			## Distance to the new OTU resulting from the merger of $i and $j ##
		my $dist_Ak = ($dist_ik + $dist_jk - $pw_dist)/2;
		push @mgd_row_pw_distances, $dist_Ak;

			## R_old(k) = Sum_{0 <= l < m} d(k,l)
			## => R_new(k) = d(k,A) + Sum_{0 <= l < m; l ne i or j} d(k,l)
			## 		= R_old(k) - d(i,k) - d(j,k) + d(k,A)
		my $new_rk = $r[$k] - $dist_ik - $dist_jk + $dist_Ak;
		push @new_r, $new_rk;

###=END of 'ADDED in creation of 'neighbor_joining_tree''

=for the predecessor 'extended_upgma_tree'
		## my $ct1 = ($k > $i) ? $mtrx_ct_dd_pairs[$k]->[$i] : $mtrx_ct_dd_pairs[$i]->[$k];
		## my $ct2 = ($k>$j) ? $mtrx_ct_dd_pairs[$k]->[$j] : $mtrx_ct_dd_pairs[$j]->[$k];
		my $sum_ct = $ct1 + $ct2;
		push @mgd_row_ct_dd_pairs, $sum_ct;
		if ($sum_ct>0) {
		   my $sum_dist = ($ct1>0) ? $ct1 * $dist1 : 0;
		   ($ct2>0) and ($sum_dist += $ct2 * $dist2);
		   push @mgd_row_pw_distances, $sum_dist/$sum_ct;
		} else {
		    push @mgd_row_pw_distances, undef;
		}
=cut

	    }			# End of the 3rd middle for-loop #

		# Finish updating the arrays #

	    push @new_sbjct_nodes, $max_nodeid;
	    push @new_pw_distances, \@mgd_row_pw_distances;

	    @sbjct_nodes = @new_sbjct_nodes;
	    @pw_distances = @new_pw_distances;

		## ADDED in creation of 'neighbor_joining_tree' (on May 7 (Fri), 2010) ##
	    my $r_A = 0;
	    foreach my $d (@mgd_row_pw_distances) { $r_A += $d; }
	    push @new_r, $r_A;
	    @r = @new_r;

=for the predecessor 'extended_upgma_tree'
	    push @new_mtrx_ct_dd_pairs, \@mgd_row_ct_dd_pairs;
	    @mtrx_ct_dd_pairs = @new_mtrx_ct_dd_pairs;
=cut

	}	# End of the main outer while-loop #


		## Handle an Error.
		## (ADDED in creation of 'neighbor_joining_tree' on May 7 (Fri), 2010)
	unless (@sbjct_nodes == 3) {
	    my $ct = @sbjct_nodes;
	    print STDERR "[ERROR 'neighbor_joining_tree'] Somehow the iteration loop ended up in $ct sequences (", join (', ', @sbjct_nodes), "), whereas it should end in 3 sequences: $! !\n\n";
	    return (undef, undef, undef, undef, undef);
	}

		## Finish by creating the top node on top of the remaining subject nodes
		## and by estimating the branch lengths between the top node and the subjects.
		## (ADDED in creation of 'neighbor_joining_tree' on May 7 (Fri), 2010)

	my $topnode_id = $max_nodeid;
	$node2chi{$topnode_id} = \@sbjct_nodes;
	foreach my $node (@sbjct_nodes) { $node2pa{$node} = $topnode_id; }

	my ($dist_10, $dist_20, $dist_21) = ($pw_distances[1]->[0], $pw_distances[2]->[0], $pw_distances[2]->[1]);
	my $br0 = ($dist_10 + $dist_20 - $dist_21)/2;
	my $br1 = ($dist_10 + $dist_21 - $dist_20)/2;
	my $br2 = ($dist_20 + $dist_21 - $dist_10)/2;
	my ($node0, $node1, $node2) = @sbjct_nodes;
	foreach my $br_info ([$node0, $br0], [$node1, $br1], [$node2, $br2]) {
	    my ($br_id, $br_len) = @{$br_info};
	    my @attr = ();
	    $attr[$INDX_LENGTH] = $br_len;
	    $br_attr{$br_id} = \@attr;
	}
	

		# Create a hash %name2nodeid #
	my %name2nodeid = ($TOP_NODE => $topnode_id); 
	foreach my $nodeid (keys %nodeid2name) {
		my $name = $nodeid2name{$nodeid};
		$name2nodeid{$name} = $nodeid;
	}
	
=for the predecessor 'extended_upgma_tree'
		#  Create a hash %name2nodeid for each of the remaining subject nodes,
		# which will be the top-node of a tree made of nodes connected to it.
	my @set_name2nodeid = ();
	foreach my $topnode_id (@sbjct_nodes) { # The trailing foreach-loop #
	    my %name2nodeid = ($TOP_NODE => $topnode_id); 
	    my @curr_nodes = ($topnode_id);
	    while (@curr_nodes>0) { # The middle while-loop #
	        my @child_nodes = ();
		foreach my $nodeid (@curr_nodes) { # inner foreach-loop #

		    my $name = $nodeid2name{$nodeid};
		    (defined $name) and ($name2nodeid{$name} = $nodeid);

		    my $children = $node2chi{$nodeid};
		    (defined $children) and (push @child_nodes, @{$children});
		}			# End of the inner foreach-loop #
		@curr_nodes = @child_nodes;
	    }			# End of the middle while-loop #
	    push @set_name2nodeid, \%name2nodeid;
	}			# End of the trailing foreach-loop #
=cut

	return (\%name2nodeid, \%nodeid2name, \%node2pa, \%node2chi, \%br_attr);

}		# End of the subroutine 'neighbor_joining_tree' #


# z_values_branch_lengths:
#	This subroutine calculates the Z-values and P-values that indicates
#	the statistical significance of the branches.
#	Each P-value is given for the 1-tailed test under the null hypothesis
#	that the Z-value follows the normal distribution of mean 0 and variance 1.
#	More specifically,
#	P (Z > z) = 1/sqrt(2* pi) * Int_{z}^infty dt exp (- t*t/2)
#		  = 1/sqrt (pi) * Int_{z/sqrt(2)}^infty ds exp (- s*s)
#		  = erfc (z/sqrt(2)) /2.
#
# INPUT: (\%br_attr, $indx_len, \%brid2var_brlen, $indx_var), where
#	%br_attr = ($branch_id => $attr, ...) for all the branches,
#	with $attr->[$indx_len], or $attr itself if $indx_len is undef,
#	being the length of the branch;
#	%brid2var_brlen = ($branch_id => $attr2, ...) for all(?) the branches,
#	with $attr2->[$indx_var], or $attr2 itself if $indx_var is undef,
#	being the variance of the length of the branch;
#
# OUTPUT: \%br2signif, where
#	%br2signif = ($branch_id => [$brlen, $std_err, $z_value, $p_value], ...)
#	for the branches with both branch lengths and their variances.
#
sub z_values_branch_lengths (\%$\%$) {

	my ($br_attr, $indx_len, $brid2var_brlen, $indx_var) = @_;

###=for DEBUGGING the subroutine
	print STDERR "\n< Subroutine 'z_values_branch_lengths' >\n\n";
###=cut
	my %br2signif;
	foreach my $brid (keys %{$br_attr}) { # Main foreach-loop #
	    my $attr = $br_attr->{$brid};
	    my $brlen = (defined $indx_len) ? $attr->[$indx_len] : $attr;
	    my $attr2 = $brid2var_brlen->{$brid};
	    unless (defined $attr2) {
		print STDERR "[CAUTION 'z_values_branch_lengths'] Branch-length variance is NOT DEFINED for the branch '$brid' (length = $brlen): $! !\n";
		next;
	    }
	    my $var = (defined $indx_var) ? $attr2->[$indx_var] : $attr2;
	    my $err = ($var>0) ? sqrt ($var) : $var;
	    my $z = ($err>0) ? $brlen /$err : undef;
	    my $p_value = (defined $z) ? erffc ( $z/sqrt(2) )/2 : undef;
	    $br2signif{$brid} = [$brlen, $err, $z, $p_value];
	}		# End of the main foreach-loop #

###=for DEBUGGING the subroutine
	print STDERR "\n --- 'z_value_branch_lengths'  ENDED !! ---\n\n";
###=cut
	return \%br2signif;

}	# End of the subroutine 'z_value_branch_lengths' #


=set aside for later use
# apprx_bootspp_by_z_test:
#	This subroutine approximates bootstrap support values
#	by applying the Z-test on each branch length.
#	More precisely, it assumes that Z = (branch length} / {standard error}
#	follows the normal distribution, and approximates the bootstrap support
#	with int (#{iterations} * P(Z>0)), where P(Z>0)  is the probability that
#	that Z is positive under the normal distribution of mean 0 and variance 1.
#
# INPUT: (\%node2chi, \%br_attr, \%var_brlen, $indx_var, $fl_stderr, $ct_boot_ite), where
#	%node2chi = ($nodeid => \@children,...) ;
#	%br_attr = ($ID of the node immediately below the branch => \@attr, ...), with
#	$attr[$INDX_LENGTH] giving the branch length;
#	%var_brlen = ($ID of the node immediately below the branch => $var, ...), with
#	variance = $var->[$indx_var] (for $indx_var >=0) or $var (for $indx_var < 0);
#	If $fl_stderr = 1, the "variance" above is actually the standard error;
#	$ct_boot_ite is the count of the virtual bootstrap iterations.
#	When $ct_boot_ite = 1, the subroutine calculates the probability itself.
#
# OUTPUT: \%apprx_boot, where
#	%apprx_boot = ($ID of the node immediately below the branch => {approximate bootstrap value}, ...) ;
#
sub apprx_bootspp_by_z_test (\%\%\%$$$) {

	my ($node2chi, $br_attr, $var_brlen, $indx_var, $fl_stderr, $ct_boot_ite) = @_;

	my %apprx_boot;

	foreach my $nodeid (keys %{$var_brlen}) { # The main foreach-loop #

	    my $chi = $node2chi->{$nodeid};
	    ((defined $chi) and (@{$chi}>0)) or next; # Skip exterior nodes
	    my $var = $var_brlen->{$nodeid};
	    ($indx_var >=0) and ($var = $var->[$indx_var]);

	    my $attr = $br_attr->{$nodeid};
	    my $br_len = $attr->[$INDX_LENGTH];
	    unless (defined $br_len) {
		print STDERR "[CAUTION 'apprx_bootspp_by_z_test'] NO BRANCH LENGTH is assigned to the node '$nodeid' !\n";
		next;
	    }
	    if ($br_len == 0) {
		$apprx_boot{$nodeid} = $ct_boot_ite/2;
		next;
	    }
	    unless ($var >=0) {
		print STDERR "[CAUTION 'apprx_bootspp_by_z_test'] NEGATIVE VALUE ($var) of the ", (($fl_stderr > 0) ? "standard error" : "variance" ), " for the node '$nodeid' : $! !\n";
		next;
	    }

	    my $se = $fl_stderr ? $var : sqrt ($var);

	    my $z = $br_len / $se ;

	    my $p = cum_gaussian ($z, 0, 1);
	    my $app_boot = $p * $ct_boot_ite;
	    ($ct_boot_ite == 1) or ($app_boot = int ($app_boot)) ;

	    $apprx_boot{$nodeid} = $app_boot;
	}				# End of the main foreach-loop #

	return \%apprx_boot;

}		# End of the subroutine 'apprx_bootspp_by_z_test' #
=cut


1;	# a "spell" which makes the module "use"-able and/or "require"-able #


