package MyBin_Oper;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.3: Copyright (C) 2013 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyBin_Oper.pm:
#	This package consists of modules performing
#	a bit higher binary operations.
#
#

use strict;
use warnings;
use Exporter;
our @ISA = qw(Exporter);

##our @EXPORT = qw( diff_array union_array nr_array lin_search);

our @EXPORT_OK = qw(
	bitwise_or_numarrays
	bitwise_and_numarrays
	bitwise_xor_numarrays
	select_by_bin_value
	count_binary_states
	format_bin_pattern
	sbset2numrep_binstring
	make_monobit_pattern
	match_bin_arrays
		);

# bitwise_or_numarrays:
#	This subroutine performs the bitwise OR operation
#	on two input numerical arrays,
#	regarding them as representing two binary strings
#	under the same scheme (i.e. the same bit unit).
#
# 	THEORY: Because the bitwise operation (except the shift operations)
#	only concerns respective bits, a bitwise operation on the numerical
#	array representation is easily performed by simultaneous operations
#	on the components.
#	Because 0 | 0 equals 0, we don't have to care about the
#	actual bit size of the last component.
#
#	NOTE that this subrouine requires that the numerical representations
#	have the same size.
#
# INPUT: (\@numrep1, \@numrep2), where
#	@numrep1/2 is the numerical array representation
#	of the string 1/2 under the same bit unit to divide a string into an array.
#
# OUTPUT: \@numrep_or,
#	which is the numerical array representation
#	of the binary string: string 1 | string 2.
#
#	The subroutine returns undef if the process failed.
#
sub bitwise_or_numarrays (\@\@) {

	my ($numrep1, $numrep2) = @_;

	my $size1 = @{$numrep1};
	my $size2 = @{$numrep2};
	unless ($size1 == $size2) {
	    print STDERR "[ERROR 'bitwise_or_numarrays'] CLASHING sizes of the numerical representations: $size1 vs $size2 : $! !\n";
	    return undef;
	}

	my @numrep_or = ();
	for (my $i = 0; $i < $size1; $i++) { $numrep_or[$i] = $numrep1->[$i] | $numrep2->[$i] ; }

	return \@numrep_or;

}	# End of the subroutine 'bitwise_or_numarrays' #


# bitwise_and_numarrays:
#	This subroutine performs the bitwise AND operation
#	on two input numerical arrays,
#	regarding them as representing two binary strings
#	under the same scheme (i.e. the same bit unit).
#
# 	THEORY: Because the bitwise operation (except the shift operations)
#	only concerns respective bits, a bitwise operation on the numerical
#	array representation is easily performed by simultaneous operations
#	on the components.
#	Because 0 & 0 equals 0, we don't have to care about the
#	actual bit size of the last component.
#
#	NOTE that this subrouine requires that the numerical representations
#	have the same size.
#
# INPUT: (\@numrep1, \@numrep2), where
#	@numrep1/2 is the numerical array representation
#	of the string 1/2 under the same bit unit to divide a string into an array.
#
# OUTPUT: \@numrep_and,
#	which is the numerical array representation
#	of the binary string: string 1 & string 2.
#
#	The subroutine returns undef if the process failed.
#
sub bitwise_and_numarrays (\@\@) {

	my ($numrep1, $numrep2) = @_;

	my $size1 = @{$numrep1};
	my $size2 = @{$numrep2};
	unless ($size1 == $size2) {
	    print STDERR "[ERROR 'bitwise_and_numarrays'] CLASHING sizes of the numerical representations: $size1 vs $size2 : $! !\n";
	    return undef;
	}

	my @numrep_and = ();
	for (my $i = 0; $i < $size1; $i++) { $numrep_and[$i] = $numrep1->[$i] & $numrep2->[$i] ; }

	return \@numrep_and;

}	# End of the subroutine 'bitwise_and_numarrays' #


# bitwise_xor_numarrays:
#	This subroutine performs the bitwise XOR operation
#	on two input numerical arrays,
#	regarding them as representing two binary strings
#	under the same scheme (i.e. the same bit unit).
#
# 	THEORY: Because the bitwise operation (except the shift operations)
#	only concerns respective bits, a bitwise operation on the numerical
#	array representation is easily performed by simultaneous operations
#	on the components.
#	Because 0 ^ 0 equals 0, we don't have to care about the
#	actual bit size of the last component.
#
#	NOTE that this subrouine requires that the numerical representations
#	have the same size.
#
# INPUT: (\@numrep1, \@numrep2), where
#	@numrep1/2 is the numerical array representation
#	of the binary string 1/2 under the same bit unit
#	to divide a string into an array.
#
# OUTPUT: \@numrep_and,
#	which is the numerical array representation
#	of the binary string: string 1 ^ string 2.
#
#	The subroutine returns undef if the process failed.
#
sub bitwise_xor_numarrays (\@\@) {

	my ($numrep1, $numrep2) = @_;

	my $size1 = @{$numrep1};
	my $size2 = @{$numrep2};
	unless ($size1 == $size2) {
	    print STDERR "[ERROR 'bitwise_xor_numarrays'] CLASHING sizes of the numerical representations: $size1 vs $size2 : $! !\n";
	    return undef;
	}

	my @numrep_and = ();
	for (my $i = 0; $i < $size1; $i++) { $numrep_and[$i] = $numrep1->[$i] ^ $numrep2->[$i] ; }

	return \@numrep_and;

}	# End of the subroutine 'bitwise_xor_numarrays' #

# select_by_bin_value:
#	This subroutine selects entries associated with
#	a binary value (either 1 or 0).
#
# INPUT: ($kept, \@entries, \@num_array, $denom), where
#	$kept (= 1 or 0) specifies which binary value should be kept;
#	@entries is the set of entries from which we will choose those
#	corresponding to the binary $kept;
#	@num_array is the numerical array representation of the binary string
#	used as a reference;
#	$denom is the unit size (in bit) of pieces into which the binary string was devided (usually 32 bits).
#
# OUTPUT: \@selected, where
#	@selected is the set of entries selected.
#
#	The subroutine returns undef if it failed on the way.
#
sub select_by_bin_value ($\@\@$) {

	my ($kept, $entries, $num_array, $denom) = @_;
	my $size = @{$entries};

	unless (($kept == 1) or ($kept == 0)) {
	    print STDERR "[ERROR 'select_by_bin_value'] An ILLEGAL INPUT: \$kept = $kept: $! !\n\n";
	    return undef;
	}

	my $mod = ($size - 1) % $denom;
	my $qtnt = int (($size - 1) / $denom);
	##my $size_last_bits = $mod + 1;

	my @selected = ();
	for (my $i = $qtnt; $i >= 0; $i--) { # Outer for-loop #
	    my $max = ($i == $qtnt) ? $mod : $denom -1 ;
	    my $num = $num_array->[$i];
	    for (my $j = $max; $j >= 0; $j--) { # Inner for-loop #
		my $bit = $num & 1; # Take out the last bit #
		($bit == $kept)
			and (unshift @selected, $entries->[$denom * $i + $j]) ;

		$num >>= 1; # Shift to the right #
	    }			# End of the inner for-loop #
	}			# End of the outer for-loop #

	return \@selected ;

}		# End of the subroutine 'select_by_bin_value' #


# count_binary_states:
#	This subroutine counts the binary states (or the bits taking '1' and '0')
#	in a binary string in a numerical array representation.
#
# INPUT: (\@num_array, $bin_size, $denom), where
#	@num_array is the numerical array representation of the binary string;
#	$bin_size is the total length of the binary string;
#	$denom is the unit size (in bit) of pieces into which the binary string was devided (usually 32 bits).
#
# OUTPUT: ($ct_0, $ct_1), where
#
#	The subroutine returns (undef, undef) if it failed on the way.
#
sub count_binary_states (\@$$) {

	my ($num_array, $bin_size, $denom) = @_;

	my $mod = ($bin_size - 1) % $denom;
	my $qtnt = int (($bin_size - 1) / $denom);
	##my $size_last_bits = $mod + 1;

	my $ct_0 = my $ct_1 = 0;
	for (my $i = $qtnt; $i >= 0; $i--) { # Outer for-loop #
	    my $max = ($i == $qtnt) ? $mod : $denom -1 ;
	    my $num = $num_array->[$i];
	    for (my $j = $max; $j >= 0; $j--) { # Inner for-loop #

		my $bit = $num & 1; # Take out the last bit #
		$bit ?  $ct_1++ : $ct_0++ ;

		$num >>= 1; # Shift to the right #
	    }			# End of the inner for-loop #
	}			# End of the outer for-loop #

	return ($ct_0, $ct_1);

}		# End of the subroutine 'count_binary_states' #


# format_bin_pattern:
#	This subroutine creates a binary string
#	from that representated by an integer array.
#
# INPUT: (\@num_array, $size, $denom), where
#	@num_array is the numerucal array representation of the input binary string;
#	$size specifies the size of the binary string;
#	$denom is the unit size (in bit) of pieces into which the binary string was divided
#	(usually 32 bits).
#
# OUTPUT: \$bin_string.
#
sub format_bin_pattern (\@$$) {

	my ($num_array, $size, $denom) = @_;

	my $mod = ($size - 1) % $denom;
	my $qtnt = int (($size - 1)/ $denom);
	my $size_last_bits = $mod + 1;

	my $bin_string = '';
	for (my $i = 0; $i < $qtnt; $i++) { # For-loop #
	    my $num = $num_array->[$i];
	    $bin_string .= sprintf "%0${denom}b", $num;
	}			# End of the for-loop #

	my $last_num = $num_array->[$qtnt];
	$bin_string .= sprintf "%0${size_last_bits}b", $last_num;

	return \$bin_string;

}	# End of the subroutine 'format_bin_pattern' #


# sbset2numrep_binstring:
#	This subroutine generates a numerical array representation
#	of a bit string, which in turn represents the presence(1)/absence(0)
#	status of the components in a set.
#
# INPUT: (\@sbset, $size, $denom), where
#	@sbset list indices, starting from 0 and ending with $size - 1,
#	of the entries in the subset of a whole set;
#	$size is the number of the  entries in the whole set;
#	$denom is the unit size (in bit) of pieces into which the binary string will be divided (unsually 32 bits).
#
# OUTPUT: \@numrep,
#	which is the numerical array representation of the
#	bit string representing the presence (1)/absence (0) of the entries
#	in the subset.
#
sub sbset2numrep_binstring (\@$$) {

	my ($sbset, $size, $denom) = @_;

	($size>0) or (return []);

	my $sbsize = @{$sbset};

	my @numrep = ();
	my $last_entry = my $last_residue = my $last_num = 0;

	for (my $k=0; $k < $sbsize; $k++) { # Outer for-loop #
	    my $entry = $sbset->[$k];
	    my $residue = $last_residue + $entry  - $last_entry;
	    my $num = $last_num;

	    while ($residue >= $denom) { # Inner while-loop #
		($num == 0) or ($last_residue == $denom - 1)
			or ($num <<= $denom - 1 - $last_residue);

		push @numrep, $num;
		$residue -= $denom;
		$num = $last_residue = 0;
	    }			# End of the inner while-loop #

	    my $diff_residue = $residue - $last_residue;
	    ($num>0) and ($diff_residue >0) and ($num <<= $diff_residue);
	    $num++; # Set the last bit to '1'

	    $last_entry = $entry;
	    $last_residue = $residue;
	    $last_num = $num;
	}			# End of the outer for-loop #

		## Stuff the remaining bit pieces ##

	($last_num == 0) or ($last_residue == $denom - 1)
	    or ($last_num <<= $denom - 1 - $last_residue);
	push @numrep, $last_num;

	my $mod = ($size - 1) % $denom;
	my $qtnt = int (($size - 1)/ $denom);
	my $size_last_bits = $mod + 1;

	my $ct_pads = $qtnt + 1 - scalar (@numrep);
	if ($ct_pads>0) {
	   for (1 .. $ct_pads) { push @numrep, 0; }
	}

		## Final Adjustment ##
	if ($size_last_bits < $denom) {
	    my $last_bits = pop @numrep;
	    ($last_bits>0) and ($last_bits >>= $denom - $size_last_bits);
	    push @numrep, $last_bits;
	}

	return \@numrep;

}	# End of the subroutine 'sbset2numrep_binstring' #


# make_monobit_pattern:
#	This subroutine generates a numerical array that represents
#	a bit string consisting solely of '1' or '0'.
#
# INPUT: ($bit, $size, $denom), where
#	$bit should be either '1' or '0' (a binary numerical value);
#	$size specifies the size of the binary string;
#	$denom is the unit size (in bit) of pieces into which the binary string was divided
#
# OUTPUT: \@mono_pattern,
#	which is the numerical array representing the bit string we want.
#
sub make_monobit_pattern ($$$) {

	my ($bit, $size, $denom)  = @_;

	my $mod = ($size - 1) % $denom;
	my $qtnt = int (($size - 1)/ $denom);
	my $size_last_bits = $mod + 1;

	my @mono_pattern = ();
	for (my $i = 0; $i <= $qtnt; $i++) { # Outer for-loop #

	    my $size_bits = ($i == $qtnt) ? $size_last_bits : $denom;
	    my $num = $bit;

	    for (my $j = 1; $j < $size_bits; $j++) { # Inner for-loop #
		$num <<= 1; # Shift to the left
		$num += $bit;
	    }				# End of the inner for-loop #

	    push @mono_pattern, $num;
	}			# End of the outer for-loop #

	return \@mono_pattern;

}	# End of the subroutine 'make_monobit_pattern' #


# match_bin_arrays:
#	This array compares two arrays of binary series.
#
# INPUT: (\@binarray1, \@binarray2), where
#	@binarrray1/2 are arrays storing binary series.
#
# OUTPUT: 1 if @binarray1 and @binarray2 match each other,
#	  0 if they don't match each other,
#	  undef if there is some error or anomaly.
#
sub match_bin_arrays (\@\@) {

	my ($binarray1, $binarray2) = @_;
	my $size1 = scalar @{$binarray1};
	my $size2 = scalar @{$binarray2};

	unless ($size1 == $size2) {
	    print STDERR "[ERROR 'match_bin_arrays'] The two binary arrays have CLASHING SIZES: $size1 vs $size2: $! !\n\n";
	    return undef;
	}

	for (my $i=0; $i<$size1; $i++) { # For-loop #
	     if ($binarray1->[$i] ^ $binarray2->[$i]) { # The binary arrays differ in the $i th component
		return 0;
	     }
	}		# End of the for-loop #

		# The binary arrays match each other #

	return 1;

}	# End of the subroutine 'match_bin_arrays' #


1;	# a 'spell' which makes the module 'use'-able and 'require'-able #


