package MyMalnManipu_UH;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.3: Copyright (C) 2013 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyMalnManipu_UH.pm:
#	This package consists of subroutines that are intended for
#	Giddy Landan's MSA improvement project.
#
# {MyMalnManipu.pm:
#	This module manipulates multiple or pairwise alignments
#	in a more sophisticated manner than the module "MyMalnIO.pm".
# }
#
#

use strict;
use warnings;
use Exporter;

our @ISA = qw( Exporter);

our @EXPORT_OK =  qw(
	div_msa_by_gappatterns cls_blcs_by_patterns

	examine_patterns_di_tri_chrm
	sort_patterns_di_tri_chrm

	readin_score_matrix reduce_score_mtrx
	cluster_chars_upgma_like group_chars_by_sim 
	repaint_alignment
	chop_aln_via_len make_segid2segaln

	sop_lin_score_msa sum_gap_open_penalties

	msa2len_grps
	map_parsim_length_states

	compact_msa_gapinfo
	min_pairwise_indels make_indel_matrix
	map_incmpl_parsim_gap_states

	msa2column_set
	count_respairs_in_aln
	matrix_dist_extJC_via_effct
	extract_aln_mtrx
	extract_aln_mtrx2

	compare_2msa_respairs
	raw_cmp_2msa_respairs
	);

use MyFile_manipu qw( open_input_file2 open_output_file );
use MyArrayManipu qw( group_equal_measures multiple_sort compare_arrays );
use MyBin_Oper qw(
	bitwise_or_numarrays bitwise_xor_numarrays
	count_binary_states
	sbset2numrep_binstring match_bin_arrays
	);

use MyMalnManipu_pep qw( smpl_compare_seq_pair );
use MyNucSeq qw( print_fasta );

use MyTaxonomy qw( fetch_ext_offsprings_nr );
use MyTreeManipu_II qw();
use MyTreeManipu_III qw( make_node2abs_depth );
use MyTreeMap_II qw(
	map_parsim_states_sankoff_clm locate_parsim_changes
	);

use MyPhyl_Partitions qw( upper_bound_eff_count );

=comment out for the moment
use MyMalnManipu qw(
	pad_5p_end pad_3p_end merge_pad_align rm_gapcolumn rm_gapcolumn_codon
	assign_number map_resnum2alncoord separate_nogap_sites separate_resnum
	compare_malign diff_malign construct_blocks block_starts_ends
	monosite_picker id_indicator id_indicator_aa print_block_maln equiv_columns
        trim_diff_maln maln_identities block_maln_identities
	block_maln_id_distribution id_sliding_window bl_id_sliding_window
	average_id_proportions pick_poor_windows pick_bl_poor_windows
	mask_frames compl_mask_frames extract_bl_selected_frames select_in_phase
        select_unamb_frames sel_complete_codon_frames sel_codon_frames_in_ref
	extract_codon_alignment extract_codon_aln_with_ref pr_in_oden_dec
	chop_alignment);
=cut

our $DENOM_BIN = 32; ### The unit length of pieces into which the binary string will be divided.

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_STAT = $MyTreeMap_II::INDX_STAT;


sub cluster_by_patterns (\@);

# cluster_by_patterns:
#	This subroutine clusters entries with the same pattern.
#
# INPUT: (\@patterns), where
#	@{$patterns[$b]} is an array of 32-bit binary numbers representing a gap pattern
#	of the $b th block
#	('1' for an ungapped subsequence, and '0' for a gap);
#
# OUTPUT: (\@cls_patterns, \@cls_indices), where
#	@{$cls_patterns[$c]} is the gap pattern of the $c th concatenated block;
#	@{$cls_indices[$c]} lists the indices of the blocks that contribute to the $c th concatenated block.
#
#	The subroutine returns (undef, undef) if it failed on the way.
#
sub cluster_by_patterns (\@) {

	my ($patterns) = @_;
	my $ct_blcs = @{$patterns};
	if ($ct_blcs == 0) { return ([], []); }

		## Add the patterns for the blocks their own indices,
		## then sort the patterns.

	my $size_pattern = scalar (@{$patterns->[0]});
	my @indices_srt = (0 .. $size_pattern);
	my @methods_srt = ();
	for (@indices_srt) { push @methods_srt, '+'; }

	my @patterns_indx = ();
	for (my $b = 0; $b < $ct_blcs; $b++) { # The 1st for-loop #
	    push @{$patterns->[$b]}, $b;
	    push @patterns_indx, $patterns->[$b];
	}				# End of the 1st for-loop #

	my $std_patterns_indx = multiple_sort (@patterns_indx, @indices_srt, @methods_srt);

		## After clustering the blocks with the same patterns,
		## extract the indices that were added before,
		## and record the pattern for each cluster.
	pop @indices_srt;
	pop @methods_srt;
	my $cls_patterns_indx = group_equal_measures (@{$std_patterns_indx}, @indices_srt, @methods_srt);

	my @cls_patterns = my @cls_indices = ();
	foreach my $class (@{$cls_patterns_indx}) { # The 2nd for-loop #
	    my @indices = ();
	    foreach my $pattern_indx (@{$class}) { # Inner foreach-loop #
		my $indx = pop @{$pattern_indx};
		push @indices, $indx;
	    }				# End of the inner foreach-loop #
	    push @cls_patterns, $class->[0];
	    push @cls_indices, \@indices;
	}				# End of the 2nd for-loop #

	return (\@cls_patterns, \@cls_indices);

}	# End of the subroutine 'cluster_by_patterns' #


# div_msa_by_gappatterns:
#	This subroutine divides an MSA into blocks,
#	each consisting of contigous sites with the same gap pattern.
#
#	NOTE that the subroutine chews up the input alignment.
#
# INPUT: (\@in_msa, $gap), where
#	${$in_msa[$k]} is the $k th sequence embedded in the input MSA;
#	$gap is a character denoting a gap.
#
# OUTPUT: (\@blocks, \@gp_patterns), where
#	${$blocks->[$b][$k]} is the $k th subsequence (or string of gaps) in the $b th block;
#	@{$gp_patterns->[$b]} is an array of 32-bit binary numbers,
#	representing a gap pattern of the $b th block.
#	The status of the $k th sequence is represented by the ($k % 32) th bit
#	of the int ($k /32) th element
#	('1' for an ungapped sequence, '0' for a gap).
#
#	It returns (undef, undef) if the process failed.
#
sub div_msa_by_gappatterns (\@$) {

	my ($in_msa, $gap) = @_;

	my $ct_seqs = @{$in_msa};
	unless ($ct_seqs>0) {
	    print STDERR "[ERROR 'div_msa_by_gappatterns'] NO INPUT SEQUENCE: $! !\n\n";
	    return (undef, undef);
	}

	my $len_seq = length ${$in_msa->[0]};

	my @blocks = my @gp_patterns = ();

	for (my $l=0; $l<$len_seq; $l++) { # The outer for-loop #

	    my @chars = my @pattern = ();
	    for (my $k=0; $k<$ct_seqs; $k++) { # The 1st inner for-loop #
		my $char = substr (${$in_msa->[$k]}, 0, 1, '');
		push @chars, $char;

		my $bit = ($char eq $gap) ? 0 : 1;

		my $qtnt = int($k/$DENOM_BIN);
		my $mod = $k % $DENOM_BIN;

		if ($mod > 0) { # Append a bit #
		   ##$pattern[$qtnt] = $pattern[$qtnt] << 1; # Rewritten as below
		   $pattern[$qtnt] <<= 1;
		   $pattern[$qtnt] += $bit;
		} else {
		    push @pattern, $bit;
		}
	    }				# End of the 1st inner for-loop #

	    if (@blocks == 0) { # The 1st column #
		my @column = ();
		foreach my $char (@chars) { push @column, \$char; }
		push @blocks, \@column;
		push @gp_patterns, \@pattern;
		next;
	    }

	    my $last_pattern = $gp_patterns[$#gp_patterns];

	    my $if_match = match_bin_arrays (@pattern, @{$last_pattern});
	    unless (defined $if_match) {
		print STDERR "[ERROR 'div_msa_by_gappatterns'] 'match_bin_arrays' FAILED at the $l th  site: $! !\n\n";
		return (undef, undef);
	    }

	    unless ($if_match) { # Renew the block #
		my @column = ();
		foreach my $char (@chars) { push @column, \$char; }
		push @blocks, \@column;
		push @gp_patterns, \@pattern;
		next;
	    }

		# Incorporate the column into the last block #

	    my $last_block = $blocks[$#blocks];
	    for (my $k=0; $k<$ct_seqs; $k++) { ${$last_block->[$k]} .= $chars[$k]; }

	}		# End of the outer for-loop #

	return (\@blocks, \@gp_patterns);

}	# End of the subroutine 'div_msa_by_gappatterns' #


# cls_blcs_by_patterns:
#	This subroutine clusters blocks of the same pattern.
#	Then concatenate the blocks in each cluster.
#
# INPUT: (\@blocks, \@patterns), where
#	${$blocks[$b]->[$k]} is the $k th subsequence (or string of gaps) in the $b th block;
#	@{$patterns[$b]} is an array of 32-bit binary numbers representing a gap pattern
#	of the $b th block
#	('1' for an ungapped subsequence, and '0' for a gap);
#
# OUTPUT: (\@cls_blocks, \@cls_patterns, \@cls_indices), where
#	${$cls_blocks[$c]->[$k]} is the $k th subsequence in the $c th concatenated block of the same pattern;
#	@{$cls_patterns[$c]} is the gap pattern of the $c th concatenated block;
#	@{$cls_indices[$c]} lists the indices of the blocks that contribute to the $c th concatenated block.
#
#	The subroutine returns (undef, undef, undef) if it failed on the way.
#
sub cls_blcs_by_patterns (\@\@) {

	my ($blocks, $patterns) = @_;

	my $ct_blcs = @{$blocks};
	unless ($ct_blcs == @{$patterns}) {
	    print STDERR "[ERROR 'cls_blcs_by_patterns'] The number of patterns (", scalar (@{$patterns}), ") CLASHES with the number of blocks ($ct_blcs): $! !\n\n";
	    return (undef, undef, undef);
	}
	unless ($ct_blcs > 0) {
	    return ([], [], []);
	}

	my ($cls_patterns, $cls_indices) = cluster_by_patterns (@{$patterns});
	unless (defined $cls_patterns) {
	    print STDERR "[ERROR 'cls_blcs_by_patterns'] The subroutine 'cluster_by_patterns' FAILED: $! !\n\n";
	    return (undef, undef, undef);
	}

	my $ct_cls = @{$cls_patterns};
	unless ($ct_cls>0) {
	    return ([], [], []);
	}
	my $ct_seqs = @{$blocks->[0]};

		## Concatenate the blocks in each cluster ##

	my @cls_blocks = ();
	foreach my $indices (@{$cls_indices}) { # Outer foreach-loop #
	    my @cnct = ();
	    foreach (1..$ct_seqs) { my $tmp = ''; push @cnct, \$tmp; }
	    foreach my $indx (@{$indices}) { # Middle foreach-loop #
		my $block = $blocks->[$indx];
		for (my $k=0; $k<$ct_seqs; $k++) { ${$cnct[$k]} .= ${$block->[$k]}; }
	    }			# End of the middle foreach-loop #
	    push @cls_blocks, \@cnct;
	}		# End of the outer foreach-loop #

	return (\@cls_blocks, $cls_patterns, $cls_indices);

}	# End of the subroutine 'classify_blcs_by_patterns' #


# examine_patterns_di_tri_chrm:
#	This satellite subroutine of 'sort_patterns_di_tri_chrm'
#	does the actual job of examining the patterns in the alignment.
#
#	NOTE that the subroutine eats up the input alignment.
#
# INPUT: (\@align), where
#	${$align[$k]} is the $k th sequence in the alignment.
#
# OUTPUT: (\@mono_sites, \@raw_di_sites, \@raw_tri_sites, \%others), where
#	@mono_sites is a list of the positions of 1-character sites;
#	@{$raw_di_sites[$j]} = ($position, @bin) is the information on the $j th
#	site with a two-character pattern.
#	@bin is an array of 32-bit numbers representing the binary pattern,
#	which is a (#{sequences} - 1)-bit string where '0' means being identical to the last character and '1' means being different from the last character,
#	and @positions is the list of positions;
#	@{$raw_tri_sites[$k]} = ($position, @bin_1, @bin_2) is the information on the $k th three-character pattern.
#	@bin_1, @bin_2 are the 'elementary' binary patterns compatible with the three-character pattern (in numerically ascending order). The 'elementary' here means that the 3rd binary pattern is given by the binary operation (@bin_1 | @bin_2) ;
#	%others = (#{characters} => \@positions, ...).
#
#	NOTE: The site position in the alignment starts with 0.
#
sub examine_patterns_di_tri_chrm (\@) {

	my ($align) = @_;

	my $ct_seqs = @{$align};
	unless ($ct_seqs>0) {
	    print STDERR "[ERROR 'examine_patterns_di_tri_chrm'] An ILLEGAL INPUT: NO SEQUENCES in the alignment: $! !\n\n";
	    return (undef, undef, undef, undef);
	}

	my $len_seq = length ${$align->[0]};
	my $ct_digits = $ct_seqs - 1;  

	my $size_numrep = int (($ct_digits - 1)/$DENOM_BIN) + 1;
	my @indices_srt = (0 .. $size_numrep - 1);
	my @methods_srt = ();
	for (1 .. $size_numrep) { push @methods_srt, '+'; }

	my @mono_sites = my @raw_di_sites = my @raw_tri_sites = ();
	my %others;

	for (my $l=0; $l<$len_seq; $l++) { # Outer for-loop #

	    my $ref_char = substr (${$align->[$ct_digits]}, 0, 1, '');
	    my %char2seq; # ($character => \@seq_nums, ...) for the characters different from that in the last sequence.
			  #  NOTE that the sequence number starts with 0, and excludes the last sequence.

	    for (my $k=0; $k < $ct_digits; $k++) { # Inner for-loop # 
		my $char = substr (${$align->[$k]}, 0, 1, '');
		($char eq $ref_char) and next; # Skip if the sequence has the same character as the last sequence 
		(defined $char2seq{$char}) or ($char2seq{$char} = []);
		push @{$char2seq{$char}}, $k;
	    }				# End of the inner for-loop #

	    my @chars = keys %char2seq;
	    my $ct_chars = 1 + scalar (@chars) ;
	    if ($ct_chars > 3) {

		(defined $others{$ct_chars}) or ($others{$ct_chars} = []); 
		push @{$others{$ct_chars}}, $l;

	    } elsif ($ct_chars == 1) { # Mono-color sites

		push @mono_sites, $l;

	    } elsif ($ct_chars == 2) { # Di-color sites

		my $char = $chars[0];
		my $bin = sbset2numrep_binstring (@{$char2seq{$char}}, $ct_digits, $DENOM_BIN);
		unless (defined $bin) {
		    print STDERR "[ERROR 'examine_patterns_di_tri_chrm'] The subroutine 'sbset2numrep_binstring' FAILED for the character '$char': $! !\n";
		    return (undef, undef, undef, undef);
		}
		push @raw_di_sites, [$l, @{$bin}];

	    } else { # Tri-color sites #

		my ($char1, $char2) = @chars;

		my $bin1 = sbset2numrep_binstring (@{$char2seq{$char1}}, $ct_digits, $DENOM_BIN);
		unless (defined $bin1) {
		    print STDERR "[ERROR 'examine_patterns_di_tri_chrm'] The subroutine 'sbset2numrep_binstring' FAILED for the character '$char1': $! !\n";
		    return (undef, undef, undef, undef);
		}
		my $bin2 = sbset2numrep_binstring (@{$char2seq{$char2}}, $ct_digits, $DENOM_BIN);
		unless (defined $bin2) {
		    print STDERR "[ERROR 'examine_patterns_di_tri_chrm'] The subroutine 'sbset2numrep_binstring' FAILED for the character '$char2': $! !\n";
		    return (undef, undef, undef, undef);
		}

		my $sgn = compare_arrays (@{$bin1}, @{$bin2}, @indices_srt, @methods_srt);
		if ($sgn > 0) { # @{$bin1} < @{$bin2} should always hold #
		    ($bin1, $bin2) = ($bin2, $bin1);
		}

		push @raw_tri_sites, [$l, @{$bin1}, @{$bin2}];
	    }

	}			# End of the outer for-loop #

	return (\@mono_sites, \@raw_di_sites, \@raw_tri_sites, \%others);

}	# End of the subroutine 'examine_patterns_di_tri_chrm' #


# sort_patterns_di_tri_chrm:
#	This subroutine examines the patterns of the columns
#	in an alignment, and enumerates the two-color and three-color
#	patterns and the sites that take such patterns.
#
#	NOTE that the subroutine eats up the input alignment.
#
# INPUT: (\@align), where
#	${$align[$k]} is the $k th sequence in the alignment.
#
# OUTPUT: (\@mono_sites, \@di_sites, \@tri_sites, \%others), where
#	@mono_sites is a list of the positions of 1-character sites;
#	@{$di_sites[$j]} = (\@bin, \@positions) is the information on the $j th
#	two-character pattern.
#	@bin is an array of 32-bit numbers representing the binary pattern,
#	which is a (#{sequences} - 1)-bit string where '0' means being identical to the last character and '1' means being different from the last character,
#	and @positions is the list of positions;
#	@{$tri_sites[$k]} = (\@bin_1, \@bin_2, \@bin_3, \@positions) is the information on the $k th three-character pattern.
#	@bin_1, @bin_2, and @bin_3 are the binary patterns compatible with the three-character pattern (in numerically ascending order);
#	%others = (#{characters} => \@positions, ...).
#
#	NOTE: The site position in the alignment starts with 0.
#
sub sort_patterns_di_tri_chrm (\@) {

	my ($align) = @_;

	my $ct_seqs = @{$align};
	unless ($ct_seqs>0) {
	    print STDERR "[ERROR 'sort_patterns_di_tri_chrm'] An ILLEGAL INPUT: NO SEQUENCES in the alignment: $! !\n\n";
	    return (undef, undef, undef, undef);
	}

	my $len_seq = length ${$align->[0]};
	my $ct_digits = $ct_seqs - 1;  

	my $size_numrep = int (($ct_digits - 1) /$DENOM_BIN) + 1;
	
=for DEBUGGING the subroutine
	print STDERR "\n< Subroutine 'sort_patterns_di_tri_chrm' >\n\n";
	print STDERR "[ Inputs ]\n\n";
	print STDERR "Sequence length =  $len_seq .\n";
	print STDERR "#{sequences} = $ct_seqs .\n\n";
	print STDERR "size of numerical representation = $size_numrep .\n\n";
=cut

	# @{$raw_di_sites->[$j]} = ($position, @bin) is the information on the $j th
	# site with a two-character pattern.
	# @bin is an array of 32-bit numbers representing the binary pattern,
	# which is a (#{sequences} - 1)-bit string where '0' means being identical to the last character and '1' means being different from the last character,
	# and @positions is the list of positions;
	# @{$raw_tri_sites->[$k]} = ($position, @bin_1, @bin_2) is the information on the $k th three-character pattern.
	# @bin_1, @bin_2 are the 'elementary' binary patterns compatible with the three-character pattern (in numerically ascending order). The 'elementary' here means that the 3rd binary pattern is given by the binary operation (@bin_1 | @bin_2) ;
	#
	my ($mono_sites, $raw_di_sites, $raw_tri_sites, $others)
		= examine_patterns_di_tri_chrm (@{$align}); 
	unless (defined $mono_sites) {
	    print STDERR "[ERROR 'sort_patterns_di_tri_chrm'] The subroutine 'examine_patterns_di_tri_chrm' FAILED: $! !\n\n";
	    return (undef, undef, undef, undef);
	}

=for DEBUGGING the subroutine
	print STDERR "[ Output of 'examine_patterns_di_tri_chrm' ]\n\n";
	print STDERR '@{$raw_di_sites} = (', "\n";
	foreach my $info (@{$raw_di_sites}) {
	    print STDERR "\t\[", join (', ', @{$info}), "\],\n";
	}
	print STDERR "\t\) .\n\n";

	print STDERR '@{$raw_tri_sites} = (', "\n";
	foreach my $info (@{$raw_tri_sites}) {
	    print STDERR "\t\[", join (', ' , @{$info}), "\],\n";
	}
	print STDERR "\t\) .\n\n";
=cut

		# Sort and compile the set of di-color sites #

=for DEBUGGING the subroutine
	print STDERR "\n --- Now I will sort and subdivide the set of di-color sites ...\n\n";
=cut

		## A BUG FIXED on March 31 (Wed), 2010 ##
	my @indices_srt_di = (1 .. $size_numrep);
	##my @indices_srt_di = (1 .. $ct_digits);

	my @methods_srt_di = ();
	for (@indices_srt_di) { push @methods_srt_di, '+'; }

	my $std_raw_di_sites = multiple_sort (@{$raw_di_sites}, @indices_srt_di, @methods_srt_di);
	unless (defined $std_raw_di_sites) {
	    print STDERR "[ERROR 'sort_patterns_di_tri_chrm'] The subroutine 'multiple_sort' FAILED on the raw set of di-color sites: $! !\n";
	    return (undef, undef, undef, undef);
	}
	my $grpd_raw_di_sites = group_equal_measures (@{$std_raw_di_sites}, @indices_srt_di, @methods_srt_di);
	unless (defined $grpd_raw_di_sites) {
	    print STDERR "[ERROR 'sort_patterns_di_tri_chrm'] The subroutine 'group_equal_measures' FAILED on the sorted raw set of di-color sites: $! !\n";
	    return (undef, undef, undef, undef);
	}

=for DEBUGGING the subroutine
	print STDERR "\n --- Now I will compile the set of di-color sites ...\n\n";
=cut
	my @di_sites = ();
	foreach my $group (@{$grpd_raw_di_sites}) { # 1st outer foreach-loop #

	    my @bin = @{$group->[0]}[@indices_srt_di];

	    my @positions = ();
	    foreach my $site_info (@{$group}) { push @positions, $site_info->[0]; }
	    my @std_positions = sort {$a <=> $b} @positions;

	    push @di_sites, [\@bin, \@std_positions];

	}			# End of the 1st outer foreach-loop #


		# Sort and compile the set of tri-color sites #

=for DEBUGGING the subroutine
	print STDERR "\n --- Now I will sort and subdivide the set of tri-color sites ...\n\n";
=cut

		## A BUG FIXED on March 31 (Wed), 2010 ##
	my @indices_srt_tri = (1 .. 2 * $size_numrep);
	##my @indices_srt_tri = (1 .. 2 * $ct_digits);

	my @methods_srt_tri = ();
	for (@indices_srt_tri) { push @methods_srt_tri, '+'; }

	my $std_raw_tri_sites = multiple_sort (@{$raw_tri_sites}, @indices_srt_tri, @methods_srt_tri);
	unless (defined $std_raw_tri_sites) {
	    print STDERR "[ERROR 'sort_patterns_di_tri_chrm'] The subroutine 'multiple_sort' FAILED on the raw set of tri-color sites: $! !\n";
	    return (undef, undef, undef, undef);
	}
	my $grpd_raw_tri_sites = group_equal_measures (@{$std_raw_tri_sites}, @indices_srt_tri, @methods_srt_tri);
	unless (defined $grpd_raw_tri_sites) {
	    print STDERR "[ERROR 'sort_patterns_di_tri_chrm'] The subroutine 'group_equal_measures' FAILED on the sorted raw set of tri-color sites: $! !\n";
	    return (undef, undef, undef, undef);
	}

=for DEBUGGING the subroutine
	print STDERR "\n --- Now I will compile the the set of tri-color sites ...\n\n";
=cut

		## A BUG FIXED on March 31 (Wed), 2010 ##
	my @indices_di2 = (1+$size_numrep .. 2 * $size_numrep);
	##my @indices_di2 = (1+$ct_digits .. 2 * $ct_digits);

	my @tri_sites = ();
	foreach my $group (@{$grpd_raw_tri_sites}) { # 2nd outer foreach-loop #

	    my @bin1 = @{$group->[0]}[@indices_srt_di];
	    my @bin2 = @{$group->[0]}[@indices_di2];
	    my $bin3 = bitwise_or_numarrays (@bin1, @bin2);
	    unless (defined $bin3) {
		print STDERR "[ERROR 'sort_patterns_di_tri_chrm'] The subroutine 'bitwise_or_numarrays' FAILED while compiling the set of tri-color sites: $! !\n";
		return (undef, undef, undef, undef);
	    }

	    my @positions = ();
	    foreach my $site_info (@{$group}) { push @positions, $site_info->[0]; }
	    my @std_positions = sort {$a <=> $b} @positions;

	    push @tri_sites, [\@bin1, \@bin2, $bin3, \@std_positions];

	}			# End of the 2nd outer foreach-loop #

=for DEBUGGING the subroutine
	print STDERR "\n --- 'sort_patterns_di_tri_chrm' ENDED ---\n\n";
=cut

	return ($mono_sites, \@di_sites, \@tri_sites, $others);

}	# End of the subroutine 'sort_patterns_di_tri_chrm' #


# readin_score_matrix:
#	This subroutine reads in a score matrix in the BLAST package.
#
# INPUT: (\$infile, $comment), where
#	$comment is the character to mark the beginning of a comment line.
#
# OUTPUT: (\@characters, \@score_mtrx), where
#	@characters lists all possible characters;
#	$score_mtrx[$i]->[$j] is the score for the pair $characters[$i] and $characters[$j].
#
sub readin_score_matrix (\$$) {

	my ($infile, $comment) = @_;
	local $/ = "\n";

	my $fh_tmp;
	unless ( open ($fh_tmp, "< ${$infile}" ) ) {
	    print STDERR "[ERROR 'readin_score_matrix'] CanNOT OPEN the input file '${$infile}: $! !\n";
	    return (undef, undef);
	}

	my @chars = ();
	while (my $in = <$fh_tmp>) { # 1st round #
	    ($in =~ m/\S/) or next;
	    ($in =~  m/^\s*$comment/o) and next;
	    @chars = split ' ', $in;
	    last;
	}		#  end of the 1st round #

	my @score_mtrx = ();
	while (my $in = <$fh_tmp>) { # 2nd round #
	    ($in =~ m/\S/) or next;
	    ($in =~  m/^\s*$comment/o) and next;
	    my @row = split /\s+/, $in;
	    ($row[0] =~ m/^[a-zA-Z\*]$/) and (shift @row);
	    push @score_mtrx, \@row;
	}		# end of the 2nd round #
	close ($fh_tmp);

	return (\@chars, \@score_mtrx);

}	# End of the subroutine 'readin_score_matrix' #


# reduce_score_mtrx:
#	This subroutine reduces the score matrix by
#	(i) removing those elements concerning unnecessary characters, and by
#	(ii) retaining only the lower-triangular part of the matrix.
#
# INPUT: (\@score_mtrx, \@characters, \@excluded), where
#	$score_mtrx[$i]->[$j] is the score for the pair $characters[$i] and $characters[$j];
#	@characters lists all possible characters;
#	@excluded lists the characters to be excluded.
#
# OUTPUT: (\@rdcd_chars, \@rdcd_score_mtrx), where
#	@rdcd_chars lists the characters excluding those in @excluded;
#	$[$i]->[$j] (0 <= $j < $i < scalar (@rdcd_chars) ) is the reduced score matrix.
#
sub reduce_score_mtrx (\@\@\@) {

	my ($score_mtrx, $chars, $excl) = @_;

	my $ct_chars = @{$chars};
	my %excl;
	foreach my $char (@{$excl}) { $excl{$char} = 1; }

		# Sort out necessary and unnecessary characters #
	my %excl_indices; # = ($indx => 1, ...) for the indices to be excluded
	my @rdcd_chars = ();
	for (my $i = 0; $i < $ct_chars; $i++) { # The 1st round #
	    my $char = $chars->[$i];
	    if (defined $excl{$char}) {
		$excl_indices{$i} = 1;
	    } else {
		push @rdcd_chars, $char;
	    }
	}			# End of the 1st round #

		# Process the score matrix #
	my @rdcd_mtrx = ();
	for (my $i=0; $i<$ct_chars; $i++) { # The 2nd round #

	    (defined $excl_indices{$i}) and next; # Skip the row to be excluded #
	    my $row = $score_mtrx->[$i];

	    my @rdcd_row = ();
	    for (my $j=0; $j<$i; $j++) { # Inner for-loop #
		(defined $excl_indices{$j}) and next; # Skip the column to be excluded #
		push @rdcd_row, $row->[$j];
	    }			# End of the inner for-loop #

	    push @rdcd_mtrx, \@rdcd_row;
	}			# End of the 2nd round #

	return (\@rdcd_chars, \@rdcd_mtrx);

}	# End of the subroutine 'reduce_score_mtrx' #


# cluster_chars_upgma_like:
#	This subroutine clusters characters using a UPGMA-like algorithm.
#	NOTE that it clusters the taxa with the LARGEST score,
#	not those with the smallest score as in the phylogenetic tree construction.
#
#	ALGORITHM:
#	(i) Compare ALL scores, and
#	cluster the pair with the LARGEST score,
#	creating a new node at the mid-point between the two taxa (= characters).
#	(NOTE that the score will be defined between the new clusters,
#	not between the new nodes);
#
#	(ii) Update the score matrix by replacing the clustered pair of taxa
#	with a new taxon defined by the clustered pair.
#	The score between two clusters (A and B) is given by:
#	scr (A,B) = Sum_(i,j) scr (i, j) / #{pairs with defined distances},
#	where the summation is taken across the pairs (i,j) with i in A, j in B.
#	In the actual calculation, the update of the score is performed as:
#	scr (A, B1+B2) = [#{pairs}_(A, B1) * scr (A, B1) + #{pairs}_(A, B2) * scr (A, B2)] / #{pairs}_(A,B) ,
#	where the cluster B is formed by merging the 2 clusters B1 and B2,
#	and #{pairs}_(A, B) is the number of exterior taxon pairs under the taxon pair (A, B).
#	NOTE the equation
#	#{pairs}_(A, B} = #{pairs}_(A, B1) + #{pairs}_(A, B2) ;
#
#	(iii) Iterate (i) and (ii) untill we are left with a taxon
#	consisting of ALL characters.
#
#
# INPUT: (\@chars, \@score_mtrx), 
#	$chars[$i] is the $i th character ($i=0, ..., #{char} -1);
#	$score_mtrx[$i]->[$j] (0<=$j< $i) is the score between
#	$chars[$i] and $chars[$j].
#	The element $score_mtrx[0] is the reference to an empty array;
#
# OUTPUT: (\%name2nodeid, \%nodeid2char, \%node2pa, \%node2chi, \%br_attr, \%node_attr, \%node2ct_pairs), where
#	%name2nodeid = ($name => $node_id, ...) for $TOP_NODE and exterior nodes corresponding to the characters;
#	%nodeid2char = ($node_ID => $char, ...) for exterior nodes;
#	%node2pa = ($node_ID => $parent_ID, ...) ;
#	%node2chi = ($node_ID => \@child_IDs, ...) ;
#	%br_attr = ($ID of the nearest downstream node => [$length, $bootstrap (=undef)], ...);
#	%node_attr = ($node_ID => ['level' from the bottom], ...) for INTERIOR nodes ;
#	%node2ct_pairs = ($node_id => #{pairs of exterior nodes} under the 2 sister clusters, ...) for interior nodes.
#
#	CONVENTION for the NODE IDs:
#	The node ID $i+1 is assigned to the character $chars[$i] ($i = 0, ... #{taxa}-1),
#	then each interior taxon (or node) is assigned with a node ID that begins with $ct_chars,
#	and that is incremented by 1 as a new taxon (=cluster) is created.
#
sub cluster_chars_upgma_like (\@\@) {

	my ($chars, $score_mtrx) = @_;

	my $ct_chars = @{$chars};
	unless ($ct_chars>0) {
	    print STDERR "[ERROR 'cluster_chars_upgma_like'] ILLEGAL INPUT: NO TAXA are supplied: $! !\n";
	    return (undef, undef, undef, undef, undef, undef, undef);
	}

	unless ($ct_chars == @{$score_mtrx}) {
	    print STDERR "\n[ERROR 'cluster_chars_upgma_like'] ILLEGAL INPUT: The number of ROWS (", scalar (@{$score_mtrx}), ") CLASHES with the number of CHARACTERS ($ct_chars) : $! !\n";
	    return (undef, undef, undef, undef, undef, undef);
	}

		# Initialize the setting #
	my @mtrx_ct_pairs = ();
	my %nodeid2char;
	for (my $i=0; $i<$ct_chars; $i++) { # The preliminary outer for-loop #
	    $nodeid2char{$i+1} = $chars->[$i];
	    my @row_ct_pairs = ();
	    for (my $j = 0; $j<$i; $j++) { push @row_ct_pairs, 1; }
	    push @mtrx_ct_pairs, \@row_ct_pairs;
	}			# End of the preliminary outer for-loop #
	my @sbjct_nodes = (1 .. $ct_chars); # list of subject nodes in exactly the same order as in the score matrix
	my $max_nodeid = $ct_chars;

	my @score_mtrx = @{$score_mtrx};

	my %node2pa; my %node2chi; my %br_attr; my %node2ct_pairs;

	my %node2level; # ($node_ID => 'level' from the bottom, ...), for INTERIOR nodes.


	while (1) { # The main outer while-loop #

	    my $ct_sbjcts = @sbjct_nodes;

		# Search for the closest pair #

	    my @hs_pair = (); # = ($indx_node1, $indx_node2, $score), where $indx_node1/2 is the index in the score matrix, of the highest-scoring pair.
	    for (my $i=0; $i<$ct_sbjcts; $i++) { # The 1st middle for-loop #
		my $row_scores = $score_mtrx[$i];
		for (my $j=0; $j<$i; $j++) { # inner for-loop #
		    my $score = $row_scores->[$j];

			# Skip pairs with undefined or unavailable distances #
		    ## (defined $score) or next;

		    unless (@hs_pair>0) { # 'Initialize' the highest-scoring pair
			@hs_pair = ($i, $j, $score);
			next;
		    }
		    ($score > $hs_pair[2]) or next; # No need to update

			# Update the highest-scoring pair #
		     @hs_pair = ($i, $j, $score);
		
		}			# End of the inner for-loop #
	    }				# End of the 1st middle for-loop #

	    (@hs_pair>0) or last; # We cannot cluster the nodes any more. #

		# Cluster the highest-scoring pair #

	    my ($i, $j, $score) = @hs_pair;
	    $max_nodeid++;

	    $node2ct_pairs{$max_nodeid} = $mtrx_ct_pairs[$i]->[$j];

	    ##my @clustered = @sbjct_nodes[$i,$j];
	    my @clustered = @sbjct_nodes[$j,$i]; # Just for a better visualization

	    $node2chi{$max_nodeid} = \@clustered;
	    $node2level{$max_nodeid} = $score/2; ## REVISED on Apr 21 (Wed), 2010 ##

	    foreach my $nodeid (@clustered) { # The 2nd middle foreach-loop #
		$node2pa{$nodeid} = $max_nodeid;

		   ## REVISED on Apr 21 (Wed), 2010 ## 
		my $br_len  = $score/2;
		my $level = $node2level{$nodeid};
		(defined $level) and ($br_len -= $level);
		##my $br_len = my $level = $score/2;
		##$node2level{$nodeid} = $level;
		##my $children = $node2chi{$nodeid};
		##((defined $children) and (@{$children}>0))
		##    and ($br_len -= $node2level{$children->[0]});
		my @attr = ();
		$attr[$INDX_LENGTH] = $br_len;
		$br_attr{$nodeid} = \@attr;
	    }			# End of the 2nd middle foreach-loop #

	    unless ($ct_sbjcts > 2) { # We do not have the subjects to cluster any longer.
		@sbjct_nodes = ($max_nodeid);
		last;
	    }

		# Update the subject nodes &
		# matrices for scores & #{pairs} #

	    my @new_sbjct_nodes = my @new_score_mtrx = my @new_mtrx_ct_pairs = ();

	    my @mgd_row_scores = my @mgd_row_ct_pairs = ();
		# merged information on the $i th and $j th subjects #

	    for (my $k=0; $k< $ct_sbjcts; $k++) { # The 3rd middle for-loop #
		(($k == $i) or ($k == $j)) and next; # skip the clustered subjects #

			# squeeze pairs of remaining taxa into the new arrays #

		push @new_sbjct_nodes, $sbjct_nodes[$k];

		my $row_scores = $score_mtrx[$k];
		my $row_ct_pairs = $mtrx_ct_pairs[$k];
		my @new_row_scores = my @new_row_ct_pairs = ();
		for (my $l=0; $l<$k; $l++) { # The inner for-loop #
		    (($l == $i) or ($l == $j)) and next; # Skip the clustered subjects

		    push @new_row_scores, $row_scores->[$l];
		    push @new_row_ct_pairs, $row_ct_pairs->[$l];
		}			# End of the inner for-loop #

		push @new_score_mtrx, \@new_row_scores;
		push @new_mtrx_ct_pairs, \@new_row_ct_pairs;

			# pairs containing a new (merged) taxon #

		my $score1 = ($k>$i) ? $score_mtrx[$k]->[$i] : $score_mtrx[$i]->[$k];
		my $ct1 = ($k > $i) ? $mtrx_ct_pairs[$k]->[$i] : $mtrx_ct_pairs[$i]->[$k];
		my $score2 = ($k>$j) ? $score_mtrx[$k]->[$j] : $score_mtrx[$j]->[$k];
		my $ct2 = ($k>$j) ? $mtrx_ct_pairs[$k]->[$j] : $mtrx_ct_pairs[$j]->[$k];
		my $sum_ct = $ct1 + $ct2;
		push @mgd_row_ct_pairs, $sum_ct;
		##if ($sum_ct>0) {
		   my $sum_score = ($ct1>0) ? $ct1 * $score1 : 0;
		   ($ct2>0) and ($sum_score += $ct2 * $score2);
		   push @mgd_row_scores, $sum_score/$sum_ct;
		##} else {
		    ##push @mgd_row_scores, undef;
		##}

	    }			# End of the 3rd middle for-loop #

		# Finish updating the arrays #

	    push @new_sbjct_nodes, $max_nodeid;
	    push @new_score_mtrx, \@mgd_row_scores;
	    push @new_mtrx_ct_pairs, \@mgd_row_ct_pairs;

	    @sbjct_nodes = @new_sbjct_nodes;
	    @score_mtrx = @new_score_mtrx;
	    @mtrx_ct_pairs = @new_mtrx_ct_pairs;

	}	# End of the main outer while-loop #

		#  Create a hash %name2nodeid for each of the remaining subject nodes,
		# which will be the top-node of a tree made of nodes connected to it.

	my @set_name2nodeid = ();
	foreach my $topnode_id (@sbjct_nodes) { # The trailing foreach-loop #
	    my %name2nodeid = ($TOP_NODE => $topnode_id); 
	    my @curr_nodes = ($topnode_id);
	    while (@curr_nodes>0) { # The middle while-loop #
	        my @child_nodes = ();
		foreach my $nodeid (@curr_nodes) { # inner foreach-loop #

		    my $name = $nodeid2char{$nodeid};
		    (defined $name) and ($name2nodeid{$name} = $nodeid);

		    my $children = $node2chi{$nodeid};
		    (defined $children) and (push @child_nodes, @{$children});
		}			# End of the inner foreach-loop #

		@curr_nodes = @child_nodes;
	    }			# End of the middle while-loop #

	    push @set_name2nodeid, \%name2nodeid;
	}			# End of the trailing foreach-loop #

	unless (1 == @set_name2nodeid) {
	    print STDERR "[ERROR 'cluster_chars_upgma_like'] ANOMALOOUS NUMBER (", scalar (@sbjct_nodes), ") of finally remained nodes: ", join (', ', @sbjct_nodes), " : $! !\n";
	    return (undef, undef, undef, undef, undef, undef, undef);
	}

	my $name2nodeid = $set_name2nodeid[0];
	my %node_attr;
	foreach my $node (keys %node2level) { # The 2nd trailing foreach-loop #
	    my $level = $node2level{$node};
	    (defined $level)  or next;
	    my @attr = ();
	    $attr[$INDX_LENGTH] = $level;
	    $node_attr{$node} = \@attr;
	}				# End of the 2nd trailing foreach-loop #

	return ($name2nodeid, \%nodeid2char, \%node2pa, \%node2chi, \%br_attr, \%node_attr, \%node2ct_pairs);

}	# End of the subroutine 'cluster_chars_upgma_like' #


# group_chars_by_sim:
#	This subroutine groups the characters into clusters consisting of members
#	whose 'similarity levels' are higher than or equal to the specified threshold.
#	If a character is NOT 'similar' enough to any other characters,
#	it comprizes a single-member cluster.
#
# INPUT: ($topnode_id, \%node2chi, \%nodeid2char, \%node_attr, $indx_level, $thrsh), where
#	$topnode_id is the top node of the clustering (tree);
#	%node2chi = ($node_id => \@children, ...) for the interior nodes;
#	%nodeid2char = ($node_id => $character, ...) for the exterior nodes;
#	%node_attr = ($node_id => $attr, ...) for the interior nodes,
#	where $attr->[$indx_level], or $attr if $indx_level is undefined,
#	gives the similarity level assigned to the node;
#	$thrsh is the threshold of the similarity level.
#
# OUTPUT: \@clusters, where
#	@{$clusters[$k]} = ($sbtopnode_id, \@chars) for the $k th cluster,
#	with $sbtopnode_id is the top node (or the 'last common ancestor') of the cluster,
#	and @chars lists the characters belonging to the cluster.
#
sub group_chars_by_sim ($\%\%\%$$) {

	my ($topnode_id, $node2chi, $nodeid2char, $node_attr, $indx_level, $thrsh) = @_;

	my @clusters = ();

	my @subjects = ($topnode_id);

	while (@subjects > 0) { # Outer while-loop #

	    my @new_subjects = ();
	    foreach my $id_sb (@subjects) { # Middle foreach-loop #
		my $attr = $node_attr->{$id_sb};
		unless (defined $attr) {
		    my $char = $nodeid2char->{$id_sb};
		    unless (defined $char) {
			print STDERR "[ERROR 'group_chars_by_sim'] The interior node '$id_sb' does NOT have the LEVEL defined: $! !\n";
			return undef;
		    }
			# This should be the exterior node #
		    push @clusters, [$id_sb, [$char]];
		    next;
		}
		    # Interior node #
		my $level = (defined $indx_level) ? $attr->[$indx_level] : $attr ;
		if ($level >= $thrsh) {
			# The cluster satisfies the condition #
		    my $ext_nodes = fetch_ext_offsprings_nr ($id_sb, %{$node2chi});
		    my @chars = ();
		    foreach my $node (@{$ext_nodes}) { # Inner foreach-loop #
			my $char = $nodeid2char->{$node};
			unless (defined $char) {
			    print STDERR "[ERROR 'group_chars_by_sim'] NO CHARACTER is assigned to the exterior offspring '$node' of the  node '$id_sb': $! !\n";
			    return undef;
			}
			push @chars, $char;
		    }				# End of the inner foreach-loop #
		    push @clusters, [$id_sb, \@chars];
		    next;
		}

			# Keep the child nodes for the next subjects #
		my $children = $node2chi->{$id_sb};
		unless (defined $children) {
		    print STDERR "[ERROR 'group_chars_by_sim'] The EXTERIOR NODE '$id_sb' has a LEVEL ($level) assigned ... It's impossible : $! !!\n";
		    return undef;
		}
		push @new_subjects, @{$children};

	    }			# End of the middle foreach-loop #

		# Update the set of subjects #
	    @subjects = @new_subjects;

	}		# End of the outer while-loop #

	return \@clusters;

}	# End of the subroutine 'group_chars_by_sim' #


# repaint_alignment:
#	This subroutine 'repaint' the input alignment
#	using a transformation code from the old character set
#	to the new character set.
#	It keeps a character intact if it is NOT in the transformation code.
#
#	NOTE that the subroutine chews up the input alignment.
#
# INPUT: (\@in_aln, \%old2new), where
#	${$in_aln[$k]} is the $k th sequence in the input alignment;
#	%old2new = ($old_character => $new_character, ...)
#	gives the transformation code
#	from the old character set to the new character set;
#
# OUTPUT: \@out_aln,
#	which  stores the alignment after transformation.
#
sub repaint_alignment (\@\%) {

	my ($in_aln, $old2new) = @_;
	my $ct_seqs = @{$in_aln};

	my @out_aln = ();

	for (my $k=0; $k<$ct_seqs; $k++) { # Outer for-loop #

	    my $old_seq = $in_aln->[$k];
	    my $len = length ${$old_seq};
	    my $new_seq = '';
	    for (my $l=0; $l <$len; $l++) { # Inner for-loop #
		my $old_char = substr (${$old_seq}, 0, 1, '');
		my $new_char = $old2new->{$old_char};

		  # Keep the old character intact if it is NOT in the transfer code #
		(defined $new_char) or ($new_char = $old_char);

		$new_seq .= $new_char;
	    }			# End of the inner for-loop #

	    push @out_aln, \$new_seq;
	}		# End of the outer for-loop #

	return \@out_aln;

}	# End of the subroutine 'repaint_alignment' #


# chop_aln_via_len:
#	This subroutine chops an input alignment into segments
#	according to the input information on the segment lengths.
#	It checks if the total segment length equals the input alignment length
#	and dies otherwise.
#	NOTE that the subroutine eats up the input alignment.
#	(CREATED on Oct 1 (Fri), 2010)
#
# INPUT: (\@inaln, \@seglens, $indx_seglen), where
#	${$inaln[$k]} is the $k th sequence embedeed in the input alignment;
#	$seglens[$s]->[$indx_seglen], or $seglens[$s] if $indx_segaln is undefined,
#	gives the length of the $s th segment.
#
# OUTPUT: \@list_segaln, where
#	${$list_segaln[$s]->[$k]} is the part of the $k th sequence
#	embedded in the $s th segment of the input alignment.
#
sub chop_aln_via_len (\@\@$) {

	my ($inaln, $seglens, $indx_seglen) = @_;

	my $ct_sequences = @{$inaln};
	unless ($ct_sequences>0) {
	    print STDERR "[ERROR 'chop_aln_via_len'] NO SEQUENCES in the input alignment: $! !\n";
	    return undef;
	}
	my $ct_segs = @{$seglens};
	unless ($ct_segs>0) {
	    print STDERR "[ERROR 'chop_aln_via_len'] There are NO SEGMENTS that the alignment will be chopped into: $! !\n";
	    return undef;
	}

		# Check if the two alignment lengths are consistent with each other #
	my $len_full_aln = length ${$inaln->[0]};
	my $tot_seglen = 0;
	foreach my $seginfo (@{$seglens}) {$tot_seglen += (defined $indx_seglen) ? $seginfo->[$indx_seglen] : $seginfo; }
	unless ($len_full_aln == $tot_seglen) {
	    print STDERR "[ERROR 'chop_aln_via_len'] The total length of the segments ($tot_seglen) CLASHes with the length of the full alignment ($len_full_aln): $! !\n";
	    return undef;
	}

		## Chop the input alignment into segments of specified lengths ##
	my @list_segaln = ();

	foreach my $seginfo (@{$seglens}) { # Main foreach-loop #
	    my $seglen = (defined $indx_seglen) ? $seginfo->[$indx_seglen] : $seginfo ;
	    my @segaln = ();
	    foreach my $seq (@{$inaln}) { # Inner foreach-loop #
		my $sbseq = substr (${$seq}, 0, $seglen, '');
		push @segaln, \$sbseq;
	    }			# End of the inner foreach-loop #

	    push @list_segaln, \@segaln;
	}			# End of the main foreach-loop #

	return \@list_segaln;

}	# End of the subroutine 'chop_aln_via_len' #


# make_segid2segaln:
#	This subroutine chops an input alignment into segments
#	according to the input information on the segment lengths.
#	Then makes a hash table associating each segment ID to the corresponding segment alignment.
#	NOTE that this subroutine eats up the input alignment.
#	(CREATED on Oct 1 (Fri), 2010)
#
# INPUT: (\@inaln, \@segids, $indx_segid, \@seglens, $indx_seglen), where
#	${$inaln[$k]} is the $k th sequence embedeed in the input alignment;
#	$segids[$s]->[$indx_segid], or $segids[$s] if $indx_segid is undefined,
#	gives the ID of the $s th segment;
#	$seglens[$s]->[$indx_seglen], or $seglens[$s] if $indx_segaln is undefined,
#	gives the length of the $s th segment.
#
# OUTPUT: \%segid2segaln, where
#	%segid2segaln = ($segment_ID => \@segaln, ...) with
#	${$segaln[$k]} is a part of the $k th sequence 
#	embedded in the segment of the alignment.
#
sub make_segid2segaln (\@\@$\@$) {

	my ($inaln, $segids, $indx_segid, $seglens, $indx_seglen) = @_;

	my $ct_sequences = @{$inaln};
	unless ($ct_sequences>0) {
	    print STDERR "[ERROR 'make_segid2segaln'] NO SEQUENCES in the input alignment: $! !\n";
	    return undef;
	}
	my $ct_segs = @{$segids};
	unless ($ct_segs>0) {
	    print STDERR "[ERROR 'make_segid2segaln'] There are NO SEGMENTS that the alignment will be chopped into: $! !\n";
	    return undef;
	}
	unless ($ct_segs == @{$seglens}) {
	    print STDERR "[ERROR 'make_segid2segaln'] The number of segment IDs ($ct_segs) CLASHes with the number of segment lengths (", scalar (@{$seglens}), "): $! !\n";
	    return undef;
	}

	my $list_segaln = chop_aln_via_len (@{$inaln}, @{$seglens}, $indx_seglen);
	unless (defined $list_segaln) {
	    print STDERR "[ERROR 'make_segid2segaln'] 'chop_aln_via_len' FAILED: $! !\n";
	    return undef;
	}
	unless ($ct_segs == @{$list_segaln}) {
	    print STDERR "[ERROR 'make_segid2segaln'] The number of segments (", scalar (@{$list_segaln}), ") output by 'chop_aln_via_len' CLASHes with the number of segment IDs ($ct_segs): $! !\n";
	    return undef;
	}

	my %segid2segaln;
	for (my $s=0; $s<$ct_segs; $s++) { # Main for-loop #
	    my $segid = (defined $indx_segid) ? $segids->[$s][$indx_segid] : $segids->[$s];
	    my $segaln = $list_segaln->[$s];
	    $segid2segaln{$segid} = $segaln;
	}		# End of the main for-loop #

	return \%segid2segaln;

}	# End of the subroutine 'make_segid2segaln' #


# sop_lin_score_msa:
#	This subroutine calculates the sum-of-pair (SOP) score of
#	a multiple sequence alignment (MSA) in a linear gap penalty scheme.
#
#	NOTE that this subroutine EATS UP the input alignment.
#
#	ALGORITHM:
#	In a linear gap penalty scheme, the SOP score can be calculated
#	as the summation of the contributions from the respective columns.
#	For each column, the score is calculated as:
#	Score(column) = Sum_{1<=n<n'<=N} f(c(n)) * f(c(n'))
#	= Sum_{1<=c<=C} f(c) * (f(c) -1)/2 * S(c,c)  + Sum_{1<=c<c'<=C} f(c)* f(c') * S(c,c'),
#	where n, n' (=1,...,N) label the sequences in the alignment,
#	 c, c' (=1,..,C) label the characters (including a gap '-'),
#	and S(c,c') is the score between the characters c and c'.
#	Whe regard S(-, -) = 0, and S(-, N) = - $gext_pen  here.
#
# INPUT: (\@in_align, \@sc_mtrx, \%char2indx, $gap, $gext_pen), where
#	${$in_align->[$n]} is the $n th sequence embedded in the input alignment;
#	$sc_mtrx[$c]->[$d] (0<=$d<=$c < $C) is the score between the $d th and $c th characters;
#	%char2indx = ($character => $index, ...) tells us where the character is in the scoring matrix. We ASSUME that $character is given in an upper-case letter;
#	$gap is the character representing a gap;
#	$gext_pen is the gap-extension penalty.
#
# OUTPUT: $sop_score, which is the SOP score of the MSA
#	under the specified scoring scheme.
#
sub sop_lin_score_msa (\@\@\%$$) {

	my ($in_align, $sc_mtrx, $char2indx, $gap, $gext_pen) = @_;

	my $ct_seqs = @{$in_align};
	unless ($ct_seqs>0) {
	    print STDERR "[ERROR 'sop_lin_score_msa'] NO SEQUENCE in the input alignment: $! !\n";
	    return undef;
	}

	my $aln_len = length ${$in_align->[0]};
	unless ($aln_len>0) {
	    print STDERR "[ERROR 'sop_lin_score_msa'] NO COLUMNS in the input alignment: $! !\n";
	    return undef;
	}

	my $sop_score = 0;

	for (my $l=0; $l<$aln_len; $l++) { # Main outer for-loop #

	    my %indx2ct; # ($index_char => #{characters}, ...)
	    my $ct_gaps = 0;

	    for (my $n=0; $n<$ct_seqs; $n++) { # 1st middle for-loop #
		my $char = substr (${$in_align->[$n]}, 0, 1, '');
		if ($char eq $gap) {
		    $ct_gaps++;
		    next;
		}

		$char = uc $char;
		my $indx = $char2indx->{$char};
		unless (defined $indx) {
		    print STDERR "[ERROR 'sop_lin_score_msa'] The character '$char' in the $l th site of the $n th sequence is NOT in the list : $! !\n";
		    return undef;
		}
		if (defined $indx2ct{$indx}) {
		    $indx2ct{$indx}++;
		} else {
		    $indx2ct{$indx} = 1;
		}
		
	    }			# End of the 1st middle for-loop #

		# First, count the gap-extension penalty #

	     my $sop_score_clm = - $ct_gaps * ($ct_seqs - $ct_gaps);

		# Then sum the scores between characters #

	     my @indices = sort {$a <=> $b} keys %indx2ct;
	     my $ct_indx = @indices;
	     for (my $c1=0; $c1<$ct_indx; $c1++) { # 2nd middle for-loop #

		my $indx1 = $indices[$c1];
		my $ct1 = $indx2ct{$indx1};

		my $row_score = $sc_mtrx->[$indx1];

		$sop_score_clm += $row_score->[$indx1] * $ct1 * ($ct1-1)/2 ;

		for (my $c2=0; $c2<$c1; $c2++) { # Inner for-loop #
		    my $indx2 = $indices[$c2];
		    my $ct2 = $indx2ct{$indx2};
		    $sop_score_clm += $row_score->[$indx2] * $ct1 * $ct2 ;
		}			# End of the inner for-loop #

	     }				# End of the 2nd middle for-loop #

	      $sop_score += $sop_score_clm;

	}			# End of the main outer for-loop #

	return $sop_score;


}	# End of the subroutine 'sop_lin_score_msa' #


# sum_gap_open_penalties:
#	This subroutine calculates the summation of gap-opening penalties
#	in an input MSA using a numerical array representation of gap patterns.
#	It assumes that the columns franking the MSA have no gap.
#
#	ALGORITHM:
#	Consider two consecutive blocks A and B.
#	Assume there are Ngr sequences with gaps in A and residues in B,
#	Nrg sequences with residues in A and gaps in B,
#	Nrr sequences with residues in both A and B,
#	and Ngg sequences with gaps in both A and B.
#	And the total is N = Ngr + Nrg + Nrr + Ngg.
#	Then, the total gap penalty imposed on this junction is:
#	$go_pen/2 * (Nrg * (Ngg + Nrr) + Ngr * (Ngg + Nrr) + 2 * Nrg * Ngr)
#	= $go_pen/2 * (Nrg * (N - Nrg) + Ngr * (N-Ngr)).
#	This might actually be an over-counting, I will think about it later.
#
#	At both edges, we impose the penalty of $go_pen/2 * Nr * Ng.
#
#	Nrg and Ngr are counted as follows:
#	NA := Nrr + Nrg = #{'1' in A},
#	NB := Nrr + Ngr = #{'1' in B},
#	N^ := Nrg + Ngr = #{'1' in (A ^ B)}.
#	So using a trigonometric relationship,
#	Nrg = (NA - NB + N^)/2,
#	Ngr = (NB - NA + N^)/2.
# 
#
# INPUT: (\@gap_patterns, $ct_seqs, $bin_denom, $go_pen), where
#	@{$gap_patterns->[$b]} gives the numerical array representation of
#	the binary string representing a gap pattern in the $b th block
#	of the input MSA. In the binary string, '1' and '0' represent
#	an ungapped residue and a gap, respectively ;
#	$ct_seqs is the number of sequences in the input MSA;
#	$bin_denom is the unit of bits into which the binary string is split;
#	$go_pen is the gap-opening penalty.
#
# OUTPUT: $sum_gop, which is the summation of the gap-opening penalties
#	imposed on the input MSA.
#
sub sum_gap_open_penalties (\@$$$) {

	my ($gap_patterns, $ct_seqs, $bin_denom, $go_pen) = @_;

	unless ($ct_seqs>0) {
	    print STDERR "[ERROR 'sum_gap_open_penalties'] NO SEQUENCES in the input alignment: $! !\n";
	    return undef;
	}
	my $ct_blks = @{$gap_patterns};
	unless ($ct_blks>0) {
	    print STDERR "[ERROR 'sum_gap_open_penalties'] NO BLOCKS in the input alignment: $! !\n";
	    return undef;
	}


	my $pattern_a = $gap_patterns->[0];
	my ($cmpl_a, $na) = count_binary_states (@{$pattern_a}, $ct_seqs, $bin_denom);

	my $sum_gop = - $go_pen/2 * $cmpl_a * $na;

	for (my $b=1; $b<$ct_blks; $b++) { # Main for-loop #

	    my $pattern_b = $gap_patterns->[$b];
	    my $a_xor_b = bitwise_xor_numarrays (@{$pattern_a}, @{$pattern_b});

	    my ($cmpl_b, $nb) = count_binary_states (@{$pattern_b}, $ct_seqs, $bin_denom);
	    my ($cmpl_xor, $n_xor) = count_binary_states (@{$a_xor_b}, $ct_seqs, $bin_denom);

	    my $nrg = ($na - $nb + $n_xor)/2;
	    my $ngr = ($nb - $na + $n_xor)/2;

	    $sum_gop -= $go_pen/2 * ($nrg *($ct_seqs - $nrg) + $ngr * ($ct_seqs - $ngr));

	    $pattern_a = $pattern_b;
	    ($cmpl_a, $na) = ($cmpl_b, $nb);
	}			# End of the main for-loop #

	$sum_gop -= $go_pen/2 * $cmpl_a * $na;

	return $sum_gop;

}	# End of the subroutine 'sum_gap_open_penalties' #


# msa2len_grps:
#	This subroutine receives an MSA,
#	and classifies the sequences into length groups.
#	It assumes that the input alignment contains only residues or gaps.
#
# INPUT: (\@align, $gap), where
#	${$align[$k]} is the $k th sequence embedded in the input alignment;
#	$gap is the character specifying a gap.
#
# OUTPUT: \@len_grps, where
#	@{$len_grps[$g]} = [$seqlen, \@indices] is for the $g th sequence group
#	(in descending order of the sequence length), with
#	@indices containing the indices of the sequences with that length.
#
sub msa2len_grps (\@$) {

	my ($align, $gap) = @_;

	my $ct_seqs = @{$align};
	unless ($ct_seqs>0) {
	    print STDERR "[ERROR 'msas2len_grps'] NO SEQUENCE in the input alignment: $! !\n";
	    return undef;
	}

	my %len2indices; # ($length => \@indices, ...)

	for (my $k=0; $k<$ct_seqs; $k++) { # Main for-loop #
	    my $seq = $align->[$k];
	    my $cp_seq = ${$seq};

	    my $len_seq = length $cp_seq;
	    my $ct_gaps = 0;
	    for (my $l=0; $l<$len_seq; $l++) { # Inner for-loop #
		my $char = substr ($cp_seq, 0, 1, '');
		($char eq $gap) and ($ct_gaps++);
	    }			 # End of the inner for-loop #
	    $len_seq -= $ct_gaps;

	    my $indices = $len2indices{$len_seq};
	    (defined $indices) or ($indices = $len2indices{$len_seq} = []);
	    push @{$indices}, $k;
	}		# End of the main for-loop #

	my @len_grps = ();

	foreach my $len (sort {$b <=> $a} keys %len2indices) { # Trailing foreach-loop #
	    push @len_grps, [$len, $len2indices{$len}];
	}			# End of the trailing foreach-loop #

	return \@len_grps;

}	# End of the subroutine 'msa2len_grps' #


# map_parsim_length_states:
#	This subroutine maps onto the nodes the length states
#	that give the parsimonious changes along the tree.
#	The subroutine was made by using the framework of 'map_parsim_states_sankoff_1clm'
#	in the module 'MyTreeMap_II.pm'.
#	(CREATED on Oct 13 (Wed), 2010)
#
# map_parsim_states_sankoff_1clm:
#	This variant of 'map_parsim_states_sankoff_msa' specializes in
#	handling just a single column in a MSA,
#	namely comparing just a single trait across taxa.
#
# map_parsim_states_sankoff_msa:
#	This subroutine performs Sankoff's parsimony algorithm on a MSA,
#	and serially find a set of parsimonious maps of states onto nodes for
#	each of the columns in the input MSA.
#
#	NOTE that this subroutine eats up the input MSA.
#
# INPUT: ($topnode_id, \%node2children, \%node2parent, \%name2id, \@seqnames, \@len_grps, $fl_full_mode), where
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2parent = ($node_id => $parent_id, ...) for the non-top nodes;
#	%name2id = ($seqname => $node_id, ...) for the exterior nodes and the top node;
#	$seqnames[$k] is the name of the $k th sequence in the input MSA;
#	@{$len_grps[$g]} = ($length, \@seq_indices) is for the $g th length group,
#	with @seq_indices containing the indices of the sequences with that length;
#	$fl_full_mode = 1 if you want the full mode,
#			= 0 if the brief mode is enough.
#
# OUTPUT: \@set_node2parsstat, where
#	%{$set_node2parsstat[$opt]} = ($node_id => \@state_info, ...) for the $opt th
#	option of the parsimonious states of the clumn mapped onto the nodes,
#	with
#	@state_info = ($state, $sum_penal), where
#	$state is the length state on the node,
#	$sum_penal is the minimum summation of the penalties (= #{length changes}, NOT the total changed length) across the subtree under the node.
#	If $fl_full_mode = 1, the information on ALL the nodes is given.
#	If $fl_full_mode = 0, ONLY the information on the TOP node  is given.
#
sub map_parsim_length_states ($\%\%\%\@\@$) {

	my ($topnode_id, $node2ch, $node2pa, $name2id, $seqnames, $len_grps, $fl_full_mode) = @_;

	my $ct_seqs = @{$seqnames};
	unless ($ct_seqs>0) {
	    print STDERR "[ERROR 'map_parsim_length_states'] NO SEQUENCE in the input MSA: $! !\n";
	    return undef;
	}

	my $ct_states = @{$len_grps};
	unless ($ct_states>0) {
	    print STDERR "[ERROR 'map_parsim_length_states'] NO LENGTH STATES in the input: $! !\n";
	    return undef;
	}
	my @state_list = my @pen_mtrx = ();
	for (my $i=0; $i<$ct_states; $i++) {
	    push @state_list, $len_grps->[$i][0]; 
	    my @row = ();
	    for (my $j=0; $j<$ct_states; $j++) { push @row, ($i == $j) ? 0 : 1; }
	    push @pen_mtrx, \@row;
	}

=for DEBUGGING the subroutine
	print STDERR "\n< Subroutine 'map_parsim_length_states' (#{sequences} = $ct_seqs; #{states} = $ct_states) >\n\n";
	print STDERR join ("\t", '#State_ID', 'State', 'seq,indices'), "\n\n";
	for (my $i=0; $i<$ct_states; $i++) {
	    my $state = $state_list[$i];
	    my $cnct_indices = join (',', @{$len_grps->[$i][1]});
	    print STDERR join ("\t", $i, $state, $cnct_indices), "\n";
	}
	print STDERR "\n\n";
=cut

		# Prepare auxiliary arrays/hashes #

	# %{$node2depth} = ($node_id => #{nodes from the top node to itself}, ...);
	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
	unless (defined $node2depth) {
	    print STDERR "[ERROR 'map_parsim_length_states'] The subroutine 'make_node2abs_depth' FAILED : $! !\n";
	    return undef;
	}

		# Because $index should equal to the index for the array @{$len_grps},
		# this array is unnecessary any more.
	##my %state2indx; # ($state => $index in @state_list, ...)
	##for (my $i = 0; $i<$ct_states; $i++) { $state2indx{$state_list[$i]} = $i; }

	my @node_ids; # $node_ids[$k] is the exterior node ID corresponding to the $k th sequence in @{$msa}
	for (my $k=0; $k<$ct_seqs; $k++) { # The 2nd preliminary for-loop #
	    my $name = $seqnames->[$k];
	    my $node_id = $name2id->{$name};
	    unless (defined $node_id) {
		print STDERR "[ERROR 'map_parsim_length_states'] NODE ID is UNDEFINED for the $k th sequence (name: $name): $! !\n";
		return undef;
	    }
	    $node_ids[$k] = $node_id;
	}			# End of the 2nd preliminary for-loop #

		## Main Process ##

	my %exnode2state_indx; # ($node_id => $index of the state, ...) for the exterior nodes
	for (my $g = 0; $g < $ct_states; $g++) {  # The 1st main for-loop #
	    my ($state, $seq_indices) = @{$len_grps->[$g]};

	    ### my $indx = $state2indx{$state}; ## Actually $indx = $g must hold !!

	    foreach my $seq_indx (@{$seq_indices}) { # Inner foreach-loop #
	        my $node_id = $node_ids[$seq_indx];
	        $exnode2state_indx{$node_id} = $g;
	    }				# End of the inner foreach-loop #
	}				# End of the 1st main for-loop #

=for DEBUGGING the subroutine
	print STDERR '%exnode2state_indx = (', "\n";
	print STDERR "#Node_ID => index of the state\n\n";
	foreach my $node (sort {$a <=> $b} keys %exnode2state_indx) {
	    print STDERR $node, '  =>  ', $exnode2state_indx{$node}, ",\n";
	}
	print STDERR ")\n\n";
=cut

	my $set_node2parsstat =  map_parsim_states_sankoff_clm (%{$node2ch}, %{$node2pa}, %{$node2depth}, %exnode2state_indx, @pen_mtrx, $fl_full_mode) ;
	unless (defined $set_node2parsstat) {
	    print STDERR "[ERROR 'map_parsim_length_states'] 'map_marsim_states_sankoff_msa' FAILED: $! !\n";
	    return undef;
	}
	my $ct_sets = @{$set_node2parsstat};

=for DEBUGGING the subroutine
	print STDERR "\n\n --- 'map_parsim_states_sankoff_clm' found $ct_sets sets of parsimonious states ...\n\n";
=cut

		## Translate the character index back to the character ##

	for (my $s = 0; $s<$ct_sets; $s++) { # The 2nd main for-loop #
	    my $node2parsstat = $set_node2parsstat->[$s];

=for DEBUGGING the subroutine
	print STDERR "[ The $s th set (before state-conversion) ]\n\n";
	print STDERR join ("\t", '#Node_ID', 'State', '#{changes}'), "\n\n";
	foreach my $nodeid (sort {$a <=> $b} keys %{$node2parsstat}) {
	    print STDERR join ("\t", $nodeid, @{$node2parsstat->{$nodeid}}), "\n";
	}
	print STDERR "\n\n";
=cut

	    foreach my $nodeid (keys %{$node2parsstat}) { # The inner foreach-loop #
		my $parsstat = $node2parsstat->{$nodeid};
		my $state_indx = $parsstat->[0];
		my $state = $state_list[$state_indx];
		unless (defined $state) {
		    print STDERR "[ERROR 'map_parsim_length_states'] NO LENGH STATE CORRESPONDs to the state index '$state_indx' for the node ID $nodeid: $! !\n";
		    return undef;
		}
		$parsstat->[0] = $state;
	    }			# End of the inner foreach-loop #
	}			# End of the 2nd main for-loop #

=for DEBUGGING the subroutine
	print STDERR "\n=== 'map_parsim_length_states' ENDED !! ===\n\n";
=cut

	return $set_node2parsstat;

}	# End of the subroutine 'map_parsim_length_states' #


# compact_msa_gapinfo:
#	This subroutine constructs a 'compactified' representation
#	of gaps out of an input MSA.
#	The compactified representation is constructed as follows:
#	(1) Divide the input MSA into blocks of contiguous columns with the same gap pattern;
#	(2) In each block, a sequence is represented by a character of $ungap or $gap
#	   if the sequence in that block is ungapped or consists only of gaps, respectively;
#	(3) Concatenate the contributions from all blocks to each sequence;
#	(4) Also record the lengths of the blocks.
#
# INPUT: (\@in_msa, $gap, $ungap), where
#	${$in_msa[$k]} is the $k th sequence embedded in the input MSA;
#	$gap is a character denoting a gap in the MSA;
#	$ungap will represent an ungapped segment in the output.
#
# OUTPUT: (\@gaprep_seqs, \@block_lengths), where
#	$gaprep_seqs[$k] is the compact representation of gaps in the $k th sequence,
#	which is a concatenated string of $gap (or a stretch of gaps) or $ungap (for an ungapped segment) representing each block;
#	$block_lengths[$b] is the length of the $b th block.
#
#	It returns (undef, undef) if the process failed.
#
sub compact_msa_gapinfo (\@$$) {

	my ($in_msa, $gap, $ungap) = @_;

	my $ct_seqs = @{$in_msa};
	unless ($ct_seqs>0) {
	    print STDERR "[ERROR 'compact_msa_gapinfo'] NO INPUT SEQUENCE: $! !\n\n";
	    return (undef, undef);
	}

	my $len_seq = length ${$in_msa->[0]};
	unless ($len_seq>0) {
	    print STDERR "[ERROR 'compact_msa_gapinfo'] The input sequence has a ZERO LENGTH: $! !\n\n";
	    return (undef, undef);
	}

		# Copy the sequences #
	my @cp_msa = ();
	foreach my $seq (@{$in_msa}) { # Preliminary foreach-loop # 
	    my $cp_seq = ${$seq};
	    push @cp_msa, \$cp_seq;
	}			# End of the preliminary foreach-loop #


	my @gaprep_seqs = my @bl_lens = ();

	my $last_gap_pattern; # gap pattern of the last column

	for (my $l=0; $l<$len_seq; $l++) { # The outer for-loop #

	    my @clm_gp_pattern = ();
	    for (my $k=0; $k<$ct_seqs; $k++) { # The 1st inner for-loop #

		my $char = substr (${$cp_msa[$k]}, 0, 1, '');
		push @clm_gp_pattern, (($char eq $gap) ? $gap : $ungap);

	    }				# End of the 1st inner for-loop #

	    unless (defined $last_gap_pattern) { # The 1st column #
		push @bl_lens, 1;
		@gaprep_seqs = @clm_gp_pattern;
		$last_gap_pattern = join ('', @clm_gp_pattern);
		next;
	    }

	    my $gap_pattern = join ('', @clm_gp_pattern);

	    if ($gap_pattern eq $last_gap_pattern) { # Renew the block #
		    # Just increase the length of the last block by 1 #
		$bl_lens[$#bl_lens]++;
		next;
	    }

		# Create a new block #

	    for (my $k=0; $k<$ct_seqs; $k++) { # the 2nd inner for-loop #
		$gaprep_seqs[$k] .= $clm_gp_pattern[$k];
	    }			# End of the 2nd inner for-loop #
	    push @bl_lens, 1;

		# Update the last gap pattern #
	    $last_gap_pattern = $gap_pattern;

	}		# End of the outer for-loop #

	return (\@gaprep_seqs, \@bl_lens);

}	# End of the subroutine 'compact_msa_gapinfo' #


# min_pairwise_indels:
#	This subroutine receives two gap states and
#	outputs the minimum list of insertions and deletions
#	required to transfer from the 1st sequence to the 2nd one.
#	The ALGORITHM is the following:
#	(1) Divide the string pair into blocks of different gap patterns;
#	(2) If the two blocks are only separated by a block consisting only of gaps,
#	they are merged together.
#	(3) The pattern ($gap, $ungap) is regarded as the insertion,
#	    and the pattern ($ungap, $gap) is regarded as the deletion.
#
# INPUT: (\$gapstat1, \$gapstat2, $gap), where
#	$gapstat1 and $gapstat2 are the gap states of the 1st and the 2nd sequences,
#	each made by concatenating $gap (for a stretch of gaps) and
#	$ungap (or an ungapped stretch) that represent a block;
#	$gap is a character representing a stretch of gaps along a block.
#
# OUTPUT: (\@list_ins, \@list_del), where
#
#	@{$list_ins[$i]} = (list of the indices (0-based) specifying the blocks
#	that comprise the $i th insertion from $gapstat1 to $gapstat2);
#
#	@{$list_del[$d]} = (list of indices (0-based) specifying the blocks
#	that comprise the $d th deletion from $gapstat1 to $gapstat2);
#
sub min_pairwise_indels (\$\$$) {

	my ($rf_gapstat1, $rf_gapstat2, $gap) = @_;

		# Make copies of gap states, so that the originals won't be lost. #
	my $gapstat1 = ${$rf_gapstat1};
	my $gapstat2 = ${$rf_gapstat2};
	my $ct_blocks = length $gapstat1;

		## First Step: Divide the gap state pair into
		## clusters of blocks with the same gap patterns

	my @block_clusters = ();
		# element = [$if_gap1, $if_gap2, \@block_indices]
		# with $if_gap1/2 = 1 when the block is occupied by gaps , = 0 otherwise,
		# @block_indices contains the indices of contiguous blocks (possibly separated by gap-only blocks) with the same gap pattern.

	for (my $b = 0; $b<$ct_blocks; $b++) { # 1st for-loop #

	    my $char1 = substr ($gapstat1, 0, 1, '');
	    my $char2 = substr ($gapstat2, 0, 1, '');

	    my $if_gap1 = ($char1 eq $gap) ? 1 : 0;
	    my $if_gap2 = ($char2 eq $gap) ? 1 : 0;

	    $if_gap1 and $if_gap2 and next; # Skip gap-only blocks

		# The 1st block #
	    unless (@block_clusters>0) {
		push @block_clusters, [$if_gap1, $if_gap2, [$b]];
		next;
	    }

	    my $last_cluster = $block_clusters[$#block_clusters];
	    my ($lif_gap1, $lif_gap2, $last_indices) = @{$last_cluster};

	    if (($if_gap1 == $lif_gap1) and ($if_gap2 == $lif_gap2)) {
		# Incorporate the block into the last cluster #
		push @{$last_indices}, $b;
	    } else {
		# create a new cluster #
		push @block_clusters, [$if_gap1, $if_gap2, [$b]];
	    }
	}			# End of the 1st for-loop #

		## Second Step: Sort the clusters into insertions and deletions ##

	my @list_ins = my @list_del = ();
	foreach my $cluster (@block_clusters) { # 2nd foreach-loop #
	    my ($if_gap1, $if_gap2, $indices) = @{$cluster};
	    if ($if_gap1) {
		$if_gap2 or (push @list_ins, $indices); # 1: gap, 2: ungap
	    } elsif ($if_gap2) {
			# 1: ungap, 2: gap
		push @list_del, $indices;
	    }
	}			# End of the 2nd foreach-loop #

	return (\@list_ins, \@list_del);

}	# End of the subroutine 'min_pairwise_indels' #


# make_indel_matrix:
#	This subroutine receives a list of gap states,
#	and makes three matrices to represent minimum indels required for
#	the transition between two gap states in the list.
#
# INPUT: (\@gapstates, $gap), where
#	@gapstates lists gap states, each of which is a concatenated string
#	of $gap (for a stretch of a gap) or $ungap (for ungapped stretch of residues)
#	representing a block.
#
# OUTPUT: (\@minct_indels, \@ins, \@del), where
#	$minct_indels[$i]->[$j] (0 <= $i, $j < $ct_states)
#	gives the minimum number of indels required from the $i th to the $j th gap state;
#	@{$ins[$i]->[$j]} (0 <= $j < $i < $ct_states) lists the
#	clusters of blocks inserted from the $i th gap state to the $j th one;
#	@{$del[$i]->[$j]} (0 <= $j < $i < $ct_states) lists the
#	clusters of blocks deleted from the $i th gap state to the $j th one.
#	(See the description of 'min_pairwise_indels' for more details
#	on the list of clusters of inserted/deleted blocks.)
#
sub make_indel_matrix (\@$) {

	my ($gapstates, $gap) = @_;

	my $ct_states = @{$gapstates};

	my @minct_indels = my @ins = my @del = ();
	for (1 .. $ct_states) { push @minct_indels, []; }

	for (my $i=0; $i<$ct_states; $i++) {  # Outer for-loop #
	    my $gapstat1 = $gapstates->[$i];

	    $minct_indels[$i]->[$i] = 0;

	    my @row_ins = my @row_del = (); 
	    for (my $j=0; $j<$i; $j++) { # Inner for-loop #
		my $gapstat2 = $gapstates->[$j];

		if ($gapstat1 eq $gapstat2) { # If the two states happen to be identical
		    $row_ins[$j] = [];
		    $row_del[$j] =  [];
		    $minct_indels[$i]->[$j] = $minct_indels[$j]->[$i] = 0;
		}
		my ($list_ins, $list_del) = min_pairwise_indels ($gapstat1, $gapstat2, $gap);
		unless (defined $list_ins) {
		    print STDERR "[ERROR 'make_indel_matrix'] The subroutine 'min_pairwise_indels' FAILED on the $i th row \@ $j th column: $! !\n";
		    return (undef, undef, undef);
		}

		my $ct_indels = scalar (@{$list_ins}) + scalar (@{$list_del});
		$row_ins[$j] = $list_ins;
		$row_del[$j] = $list_del;
		$minct_indels[$i]->[$j] = $minct_indels[$j]->[$i] = $ct_indels;

	    }			# End of the inner for-loop #

	    $ins[$i] = \@row_ins;
	    $del[$i] = \@row_del;
	}			# End of the outer for-loop #

	return (\@minct_indels, \@ins, \@del);

}		# End of the subroutine 'make_indel_matrix' #


# map_incmpl_parsim_gap_states:
#	This subroutine receives an output of 'compact_msa_gapinfo'
#	and maps gap states onto the tree under an 'imcomplete' Sankoff parsimony algorithm,
#	which means we seek the parsimonious indel scenarios ONLY WITHIN the set
#	of gap patterns observed in the MSA.
#
#	This subroutine was made by using the framework of 'map_parsim_length_states'.
#	(CREATED on Oct 14 (Thu), 2010) 
#
# INPUT: ($topnode_id, \%node2children, \%node2parent, \%name2id, \@seqnames, \@gaprep_seqs, $gap, $fl_full_mode), where
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2parent = ($node_id => $parent_id, ...) for the non-top nodes;
#	%name2id = ($seqname => $node_id, ...) for the exterior nodes and the top node;
#	$seqnames[$k] is the name of the $k th sequence in the input MSA;
#
#	$gaprep_seqs[$k] is the compact representation of gaps in the $k th sequence,
#	which is a concatenated string of $gap (or a stretch of gaps) or $ungap (for an ungapped segment) representing each block;
#	$gap is a character representing a stretch of gaps (in a block of a particular gap pattern in MSA);	
#
#	$fl_full_mode = 1 if you want the full mode,
#			= 0 if the brief mode is enough.
#
# OUTPUT: (\@set_node2gapstat, \@ins, \@del, \@list_gapstates), where
#	%{$set_node2gapstat[$opt]} = ($node_id => \@state_info, ...) for the $opt th
#	option of the 'incomplete' parsimonious gap states mapped onto the nodes,
#	with
#	@state_info = ($state, $sum_penal), where
#	$state is the gap state on the node,
#	$sum_penal is the minimum summation of the penalties (= #{indel events}) across the subtree under the node.
#	If $fl_full_mode = 1, the information on ALL the nodes is given.
#	If $fl_full_mode = 0, ONLY the information on the TOP node  is given;
#
#	@{$ins/del[$i]->[$j][$c]} (0 <= $j < $i < $ct_states) gives the 
#	indices specifying the blocks comprising $c th insertion/deletion event
#	from $list_gapstates[$i] to $list_gapstates]$j];
#
#	@list_gapstates non-redundantly lists all the gap states observed in the MSA.
#
sub map_incmpl_parsim_gap_states ($\%\%\%\@\@$$) {

	my ($topnode_id, $node2ch, $node2pa, $name2id, $seqnames, $gaprep_seqs, $gap, $fl_full_mode) = @_;

	my $ct_seqs = @{$seqnames};
	unless ($ct_seqs>0) {
	    print STDERR "[ERROR 'map_incmpl_parsim_gap_states'] NO SEQUENCE in the input MSA: $! !\n";
	    return (undef, undef, undef, undef);
	}
	unless (@{$gaprep_seqs} == $ct_seqs) {
	    print STDERR "[ERROR 'map_incmpl_parsim_gap_states'] The number of gap representations of sequences (", scalar (@{$gaprep_seqs}), ") CLASHES with the number of sequence IDs ($ct_seqs): $! !\n";
	    return (undef, undef, undef, undef); 
	}

	my %gapstat2indices; # ($gap_state => \@indices in @{$gaprep_seqs}, ...)
	for (my $k=0; $k<$ct_seqs; $k++) {
	    my $gapstat = $gaprep_seqs->[$k];
	    my $indices = $gapstat2indices{$gapstat};
	    (defined $indices) or ($indices = $gapstat2indices{$gapstat} = []);
	    push @{$indices}, $k;
	}
	my @list_gapstates = sort keys %gapstat2indices;
	my $ct_states = @list_gapstates;

	my ($pen_mtrx, $ins, $del) = make_indel_matrix (@list_gapstates, $gap);
	unless (defined $pen_mtrx) {
	    print STDERR "[ERROR 'map_incmpl_parsim_gap_states'] The subroutine 'make_indel_matrix' FAILED: $! !\n";
	    return (undef, undef, undef, undef);
	}

=for DEBUGGING the subroutine
	print STDERR "\n< Subroutine 'map_incmpl_parsim_gap_states' (#{sequences} = $ct_seqs; #{states} = $ct_states) >\n\n";
	print STDERR join ("\t", '#State_ID', 'gap_state', 'seq,indices'), "\n\n";
	for (my $i=0; $i<$ct_states; $i++) {
	    my $state = $list_gapstates[$i];
	    my $cnct_indices = join (',', $gapstat2indices{$state});
	    print STDERR join ("\t", $i, $state, $cnct_indices), "\n";
	}
	print STDERR "\n\n";
=cut

		# Prepare auxiliary arrays/hashes #

	# %{$node2depth} = ($node_id => #{nodes from the top node to itself}, ...);
	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
	unless (defined $node2depth) {
	    print STDERR "[ERROR 'map_incmpl_parsim_gap_states'] The subroutine 'make_node2abs_depth' FAILED : $! !\n";
	    return (undef, undef, undef);
	}

	my @node_ids; # $node_ids[$k] is the exterior node ID corresponding to the $k th sequence in @{$msa}
	for (my $k=0; $k<$ct_seqs; $k++) { # The 2nd preliminary for-loop #
	    my $name = $seqnames->[$k];
	    my $node_id = $name2id->{$name};
	    unless (defined $node_id) {
		print STDERR "[ERROR 'map_incmpl_parsim_gap_states'] NODE ID is UNDEFINED for the $k th sequence (name: $name): $! !\n";
		return undef;
	    }
	    $node_ids[$k] = $node_id;
	}			# End of the 2nd preliminary for-loop #

		## Main Process ##

	my %exnode2state_indx; # ($node_id => $index of the state, ...) for the exterior nodes
	for (my $g = 0; $g < $ct_states; $g++) {  # The 1st main for-loop #

	    my $state = $list_gapstates[$g];
	    my $seq_indices = $gapstat2indices{$state};
	    ##my ($state, $seq_indices) = @{$len_grps->[$g]};

	    foreach my $seq_indx (@{$seq_indices}) { # Inner foreach-loop #
	        my $node_id = $node_ids[$seq_indx];
	        $exnode2state_indx{$node_id} = $g;
	    }				# End of the inner foreach-loop #
	}				# End of the 1st main for-loop #

=for DEBUGGING the subroutine
	print STDERR '%exnode2state_indx = (', "\n";
	print STDERR "#Node_ID => index of the state\n\n";
	foreach my $node (sort {$a <=> $b} keys %exnode2state_indx) {
	    print STDERR $node, '  =>  ', $exnode2state_indx{$node}, ",\n";
	}
	print STDERR ")\n\n";
=cut

	my $set_node2gapstat =  map_parsim_states_sankoff_clm (%{$node2ch}, %{$node2pa}, %{$node2depth}, %exnode2state_indx, @{$pen_mtrx}, $fl_full_mode) ;
	unless (defined $set_node2gapstat) {
	    print STDERR "[ERROR 'map_incmpl_parsim_gap_states'] 'map_marsim_states_sankoff_msa' FAILED: $! !\n";
	    return undef;
	}
	my $ct_sets = @{$set_node2gapstat};

=for DEBUGGING the subroutine
	print STDERR "\n\n --- 'map_parsim_states_sankoff_clm' found $ct_sets sets of 'incomplete' parsimonious gap states ...\n\n";
=cut

		## Translate the state index back to the gap state ##

	for (my $s = 0; $s<$ct_sets; $s++) { # The 2nd main for-loop #
	    my $node2gapstat = $set_node2gapstat->[$s];

=for DEBUGGING the subroutine
	print STDERR "[ The $s th set (before state-conversion) ]\n\n";
	print STDERR join ("\t", '#Node_ID', 'State', '#{changes}'), "\n\n";
	foreach my $nodeid (sort {$a <=> $b} keys %{$node2gapstat}) {
	    print STDERR join ("\t", $nodeid, @{$node2gapstat->{$nodeid}}), "\n";
	}
	print STDERR "\n\n";
=cut

	    foreach my $nodeid (keys %{$node2gapstat}) { # The inner foreach-loop #
		my $gapstat_info = $node2gapstat->{$nodeid};
		my $state_indx = $gapstat_info->[0];
		my $state = $list_gapstates[$state_indx];
		##my $state = $state_list[$state_indx];
		unless (defined $state) {
		    print STDERR "[ERROR 'map_incmpl_parsim_gap_states'] NO GAP STATE CORRESPONDs to the state index '$state_indx' for the node ID $nodeid: $! !\n";
		    return undef;
		}
		$gapstat_info->[0] = $state;
	    }			# End of the inner foreach-loop #

	}			# End of the 2nd main for-loop #

=for DEBUGGING the subroutine
	print STDERR "\n=== 'map_incmpl_parsim_gap_states' ENDED !! ===\n\n";
=cut

	return ($set_node2gapstat, $ins, $del, \@list_gapstates);
	##return $set_node2parsstat;

}	# End of the subroutine 'map_incmpl_parsim_gap_states' #


# msa2column_set:
#	This subroutine converts an input alignment, which is a set of embedded sequences,
#	into a set of columns in the alignment, in which each column is represented as
#	a concatenated string of characters.
#	NOTE that this subroutine eats up the input alignment.
#
# INPUT: \@align, where
#	${$align[$k]} is the $k th sequence embedded in the input alignment.
#
# OUTPUT: \@set_columns, where
#	$set_columns[$l] is a string made by concatenating ALL the characters
#	in the $l th column.
#
#	The subroutine returns undef if the process failed.
#
sub msa2column_set (\@) {

	my ($align) = @_;

	my $ct_seqs = @{$align};
	unless ($ct_seqs>0) {
	    print STDERR "[ERROR 'msa2column_set'] NO SEQUENCES in the input MSA: $! !\n";
	    return undef;
	}

	my $aln_len = length ${$align->[0]};
	($aln_len >0) or (return []);

	my @set_columns = ();

	for (my $l=0; $l<$aln_len; $l++) { # The main outer for-loop #

	    my $column = '';
	    for (my $k=0; $k<$ct_seqs; $k++) {
		my $char = substr (${$align->[$k]}, 0, 1, '');
		$column .= $char;
	    }
	    push @set_columns, $column;
	}			# End of the main outer for-loop #

	return \@set_columns;

}	# End of the subroutine 'msa2column_set' #


# count_respairs_in_aln:
#	This subroutine counts the number of residue pairs
#	in a multiple alignment.
#	The total number of residue pairs is just a summation over columns
#	of the number of residue pairs in each column.
#	The number of residue pairs in a column is calculated as:
#	$ct * ($ct - 1) /2, where $ct is the number of residues (excluding gaps) in the column.
#
#	NOTE that this subroutine eats up the input alignment.
#	(CREATED on Feb 9 (Wed), 2011)
#
# INPUT: (\@in_aln, $gap), where
#	${$in_aln[$k]} is the $k th sequence embedded in the input alignment;
#	$gap is the character representing a gap.
#
# OUTPUT: ($totct_respairs, \@ct_respairs_clm), where
#	$totct_respairs is the total number of residue pairs in the alignment;
#	$ct_respairs_clm[$l] is the number of residue pairs in the $l th column.
#
sub count_respairs_in_aln (\@$) {

	my ($in_aln, $gap) = @_;

	my $ct_seqs = @{$in_aln};
	($ct_seqs>0) or (return (0, []));
	my $len_aln = length ${$in_aln->[0]};
	($len_aln>0) or (return (0, []));

	unless ($ct_seqs>1) {
	    my @tmp = ();
	    for (1 .. $len_aln) { push @tmp, 0; }
	    return (0, \@tmp);
	}

	my $totct_rps = 0;
	my @ct_rps_clm = (); 

	for (my $l=0; $l<$len_aln; $l++) { # Outer for-loop #

	    my $ct_res = 0;
	    for (my $k=0; $k<$ct_seqs; $k++) { # Inner for-loop #

		my $char = substr (${$in_aln->[$k]}, 0, 1, '');
		($char eq $gap) or ($ct_res++);
	    }			# End of the inner for-loop #
	    my $ct_rps = $ct_res * ($ct_res - 1) /2;
	    $totct_rps += $ct_rps;
	    push @ct_rps_clm, $ct_rps;

	}	# End of the outer for-loop #

	return ($totct_rps, \@ct_rps_clm);

}		# End of the subroutine 'count_respairs_in_aln' #


# matrix_dist_extJC_via_effct:
#	This revised version of the subroutine 'matrix_dist_extJC'
#	(in the module 'MyMalnManipu_pep.pm') uses effective counts
#	of observed substitutions, instead of raw counts as the predecessor used.
#	(CREATED on June 29 (Tue), 2010)
#
#	ALGORITHM:
#	When we observed r substitutions out of N sites,
#	the effective count r_eff is calculated as N * p_eff,
#	where $p_eff satisfies the following equation:
#
#	I_$p_eff (r+1, N-r+1) = P0 * I_B (r+1, N-r+1) + (1 - P0) * I_A (r+1, N-r+1) ,
#
#	where I_p (x,y) = int_0^p dt t**(x-1) * (1-t)**(y-1) is the incomplete beta function
#	(see 'upper_bound_eff_count' in the module 'MyPhyl_Partitions.pm' for derivation).
#
#	Here the theoretical probability p of substitutions are required to be within the range
#	A <= p <= B (in the extended JC model, A = 0 and B = 1 - 1/$c),
#	and P0 is the probability that the substiution probability p falls within A <= p < $p_eff
#	under the uniform prior across A <= p <= B.
#	If we take P0 = 1/2, $p_eff gives the median of the possible values of p.
#
# matrix_dist_extJC:
#	This subroutine constructs a matrix of pairwise distances
#	between peptide sequences in a MSA using an extended JC model.
#	It ignores sites containing gaps.
#	The extended JC dstance $d between sequences with the p-distance $p is:
#	$d = - ($c -1)/$c  * log (1 - $c/($c-1) * $p),
#	where $c is the number of possible characters.
#	When $c = infinity, it should approach the Poisson distance (- log (1 - $p) ).
#	When $c = 4, it is the ordinary Jukes-Cantor model.
#
#	[We set an 'upper-bound' $d_up of the distance. 
#	If ($p > ($c - 1)/$c) or $d > $d_up,
#	the distance is automatically set to be $d_up.]
#	--- became Obsolete on June 29 (Tue), 2010, because the 'effective count' method
#	naturally restricts $p within the open interval (0, 1 - 1/$c).
#
#	The subroutine was made from the framework of 'matrix_gamma_dist'.
#	(CREATED on June 24 (Thu), 2010)
#
# INPUT: (\@align, $gap, $c, \@parameters), where
#	${$align[$k]} is the $k th sequence embedded in the alignment;
#	$c is the number of possible characters, which should be >= 2;
#	@parameters = ($tgt_prb, $resol, $maxite), with
#	$tgt_prb corresponding to P0 in the above description,
#	$resol being the resolution you desire for the numerical solution for the inverse incomplete beta,
#	and $maxite being the maximum iteration count for the inverse incomplete beta.
#
# OUTPUT: (\@mtrx_dist, \@mtrx_effct_diffs, \@mtrx_ct_diffs, \@mtrx_ct_sites), where
#	$mtrx_dist[$i]->[$j] (0 <= $j < $i <= $ct_seqs - 1) is the distance between the $i th and $j th sequences;
#	$mtrx_effct_diffs[$i]->[$j] is the effective number of sites different between the $i th and the $j th sequences;
#	$mtrx_ct_diffs[$i]->[$j] is the observed number of sites different between the $i th and the $j th sequences;
#	$mtrx_ct_sites[$i]->[$j] is the number of sites containing NO gaps in the alignment between the $i th and $j th sequences. It is undef if the alignment has NO effective sites.
#
sub matrix_dist_extJC_via_effct (\@$$\@) {

	my ($align, $gap, $c, $parameters) = @_;

	unless ($c>=2) {
	    print STDERR "[ERROR 'matrix_dist_extJC_via_effct'] ILLEGAL INPUT: The number of possible characters ($c) is LESS THAN 2: $! !\n";
	    return (undef, undef, undef, undef);
	}

	my $ct_seqs = @{$align};
	($ct_seqs>0) or (return ([], [], [], []));

	my $len = length ${$align->[0]};
	unless ($len>0) {
	    print STDERR "[ERROR 'matrix_dist_extJC_via_effct'] NO SITES in the alignment !\n";
	    return (undef, undef, undef, undef);
	}

	my ($tgt_prb, $resol,  $maxite) = @{$parameters};

	my ($p_lw, $p_up) = (0, 1 - 1/$c);

	my @mtrx_dist = my @mtrx_effct_diffs = my @mtrx_ct_diffs = my @mtrx_ct_sites  = ();
	for (my $i=0; $i<$ct_seqs; $i++) { # Outer for-loop #

	    my @row_dist = my @row_effct_diffs = my @row_ct_diffs = my @row_ct_sites = ();
	    for (my $j=0; $j<$i; $j++) {  # Inner for-loop #
		my ($ct_sites, $ct_diff) = smpl_compare_seq_pair (${$align->[$i]}, ${$align->[$j]}, $gap);
		unless (defined $ct_sites) {
		    print STDERR "[ERROR 'matrix_dist_extJC_via_effct'] The subroutine 'smpl_compare_seq_pair' FAILED between the $i th sequence and the $j th sequence in the alignment: $! !\n";
		    return (undef, undef, undef, undef);
		}
		   ## ADDED in creation of 'matrix_dist_extJC_via_effct' (on June 29 (Tue), 2010) ##
		unless ($ct_sites>0) {
		    print STDERR "[ERROR 'matrix_dist_extJC_via_effct'] NO non-gapped SITES in the alignment between the $i th sequence and the $j th sequence: $! !\n";
		    return (undef, undef, undef, undef);
		}
		my ($effct_diff, $error_x, $error_y, $itect) = upper_bound_eff_count ($ct_sites, $ct_diff, $tgt_prb, $p_lw, $p_up, $resol, $maxite);
		unless (defined $effct_diff) {
		    print STDERR "[ERROR 'matrix_dist_extJC_via_effct'] The subroutine 'upper_bound_eff_count' FAILED on the sequence pair \($i, $j\): $! !\n";
		    return (undef, undef, undef, undef);
		}
		
		push @row_effct_diffs, $effct_diff;

		push @row_ct_sites, $ct_sites;
		push @row_ct_diffs, $ct_diff;


		   ## REVISED in creation of 'matrix_dist_extJC_via_effct' (on June 29 (Tue), 2010) ##
		my $p = $effct_diff/$ct_sites;
		my $dist = - ($c -1)/$c * log (1 - $c/($c -1) * $p) ;

=for the predecessor 'matrix_dist_extJC'
		if ($ct_sites == 0) {
		    push @row_dist, undef;
		    next;
		}
		if ($ct_diff == 0) {
		    push @row_dist, 0;
		    next;
		}
		my $p = $ct_diff/$ct_sites;
		my $dist = ($c * $ct_diff < ($c -1) * $ct_sites) ? 
			- ($c -1)/$c * log (1 - $c/($c -1) * $p) : undef;
		((defined $dist) and ($dist < $d_up)) or ($dist = $d_up); # Force the distance  equal to the upper bound in an unacceptable situation
=cut

		push @row_dist, $dist;
	    }			# End of the inner for-loop #

	    push @mtrx_ct_sites, \@row_ct_sites;
	    push @mtrx_ct_diffs, \@row_ct_diffs;
	    push @mtrx_effct_diffs, \@row_effct_diffs;
	    push @mtrx_dist, \@row_dist;
	}			# End of the outer for-loop #

	return (\@mtrx_dist, \@mtrx_effct_diffs, \@mtrx_ct_diffs, \@mtrx_ct_sites);

}		# End of the subroutine 'matrix_dist_extJC_via_effct' #


# extract_aln_mtrx:
#	This subroutine extracts an input alignment
#	written in a text matrix format.
#
# INPUT: \$tmpinfile,
#
# OUTPUT: \@seqaln, where
#	${$seqaln[$k]} is the $k th sequence embedded in the alignment.
#
sub extract_aln_mtrx (\$) {

	my ($tmpinfile) = @_;
	local $/ = "\n";

	my $fh_tmpin = open_input_file2 (${$tmpinfile});
	unless (defined $fh_tmpin) {
	    print STDERR "[ERROR 'extract_aln_mtrx'] 'open_input_file2' FAILED on the input file '${$tmpinfile}': $! !\n";
	    return undef;
	}
	my @seqaln = ();

	while (my $tmpin = <$fh_tmpin>) { # Main while-loop #
	    ($tmpin =~ m/\S/) or next;
	    $tmpin =~ s/\n$//;
	    push @seqaln, \$tmpin;
	}			# End of the main while-loop #
	close ($fh_tmpin);

	return \@seqaln;

}	# End of the subroutine 'extract_aln_mtrx' #


# extract_aln_mtrx2:
#	This revised version copes with a new format
#	which records multiple MSAs,
#	each prepended by a header starting with '#'.
#	(CREATED on Jan 20 (Thu), 2011)
#
# extract_aln_mtrx:
#	This subroutine extracts an input alignment
#	written in a text matrix format.
#
# INPUT: \$tmpinfile,
#
# OUTPUT: \@seqaln, where
#	${$seqaln[$k]} is the $k th sequence embedded in the alignment.
#
sub extract_aln_mtrx2 (\$) {

	my ($tmpinfile) = @_;
	local $/ = "\n";

	my $fh_tmpin = open_input_file2 (${$tmpinfile});
	unless (defined $fh_tmpin) {
	    print STDERR "[ERROR 'extract_aln_mtrx2'] 'open_input_file2' FAILED on the input file '${$tmpinfile}': $! !\n";
	    return undef;
	}
	my @seqaln = ();

	my $ct_header = 0;
	while (my $tmpin = <$fh_tmpin>) { # Main while-loop #
	    ($tmpin =~ m/\S/) or next;

	    if ($tmpin =~ m/^\#/) { # ADDED in creation of 'extract_aln_mtrx2'
		$ct_header++;
		($ct_header > 1) ? last : next ;
	    }

	    $tmpin =~ s/\n$//;
	    push @seqaln, \$tmpin;
	}			# End of the main while-loop #
	close ($fh_tmpin);

	return \@seqaln;

}	# End of the subroutine 'extract_aln_mtrx2' #


# compare_2msa_respairs:
#	This subroutine compares two MSAs at the residue-pair level
#	using Dr. Giddy Landan's program 'msa_set_score'.
#
# INPUT: (\@msa1, \@msa2, \$exec_cmp_2msa, \$tmp_work), where
#	${$msa1->[$k]} and ${$msa2->[$k]} are the $k th sequence embedded
#	in the 1st MSA and the 2nd MSA, respectively;
#	$exec_cmp_2msa is the ABSOLUTE path to the executable file comparing two MSAs;
#	$tmp_work is the path to the temporary working file.
#
# OUTPUT: ($ct_respairs, $tot_respair_score, $ct_clms, $tot_clm_score), where
#	$ct_respairs is the number of residue pairs scored in the 1st MSA;
#	$tot_respair_score is the total residue pair scores,
#	which is basically the number of residue pairs correctly reproduced in the 2nd MSA;
#	$ct_clms is the number of columns scored in the 1st MSA;
#	$tot_clm_score is the total column scores,
#	which is basically the number of columns correctly reproduced in the 2nd MSA.
#
sub compare_2msa_respairs (\@\@\$\$) {

	my ($msa1, $msa2, $exec_cmp_2msa, $tmp_work) = @_;

	my $dis_len = 60; # Maximum length of the sequence displayed in a line

	my $ct_seqs = @{$msa1};
	unless (@{$msa2} == $ct_seqs) {
	    print STDERR "[ERROR 'compare_2msa_respairs'] ILLEGAL INPUT: The two MSAs have CONFLICTING NUMBER OF SEQUENCES: $ct_seqs for MSA1, and ", scalar (@{$msa2}), " for MSA2: $! !\n";
	    return (undef, undef, undef, undef);
	}

=for TESTING the subroutine
	unless (( -f ${$exec_cmp_2msa} ) and ( -x ${$exec_cmp_2msa} )) {
	    print STDERR "[ERROR 'compare_2msa_respairs'] CanNOT FIND the executable file '${$exec_cmp_2msa}': $!  !\n";
	    return (undef, undef, undef, undef);
	}
=cut

	unless ( -d ${$tmp_work} ) {
	    unless (0 == system ("mkdir -p ${$tmp_work}" ) ) {
		print STDERR "[ERROR 'compare_2msa_respairs'] FAILED to CREATE the temporary working directory '${$tmp_work}': $? : $! !\n";
		return (undef, undef, undef, undef);
	    }
	}
	unless (( -r ${$tmp_work} ) and ( -w ${$tmp_work} )) {
	    unless (0 == system ("chmod u+rw ${$tmp_work}") ) {
		print STDERR "[ERROR 'compare_2msa_respairs'] FAILED to CHANGE permission MODES of the directory'${$tmp_work}': $? : $! !\n";
		return (undef, undef, undef, undef);
	    }
	}

	(my $orig_dir = ` pwd `) =~ s/\s+$//;

	unless ( chdir ${$tmp_work} ) {
	    print STDERR "[ERROR 'compare_2msa_respairs'] FAILED to change the directory to '${$tmp_work}': $! !\n";
	    return (undef, undef, undef, undef);
	}

	my @tmp_seqids = ();
	for (my $k=0; $k<$ct_seqs; $k++) { push @tmp_seqids, sprintf ("seq%04d", $k); }

	my $infile1 = 'msa1.fasta';
	my $infile2 = 'msa2.fasta';
	my $out_pref = 'out';
	my $outfile1 = join ('_', $out_pref, 'res_pair.scr');
	my $outfile2 = join ('_', $out_pref, 'col_col.scr');
	my $logfile = 'logfile.txt';

		# Prepare input MSA files #

	my $fh_out1 = open_output_file ($infile1, 0);
	unless (defined $fh_out1) {
	    print STDERR "[ERROR 'compare_2msa_respairs'] 'open_output_file' FAILED on the 1st input MSA file '$infile1': $! !\n";
	    return (undef, undef, undef, undef);
	}
	unless ( print_fasta (@tmp_seqids, @{$msa1}, $dis_len, *${$fh_out1}) ) {
	    print STDERR "[ERROR 'compare_2msa_respairs'] 'print_fasta' FAILED to output the 1st MSA: $! !\n";
	    return (undef, undef, undef, undef);
	}
	close ($fh_out1);

	my $fh_out2 = open_output_file ($infile2, 0);
	unless (defined $fh_out2) {
	    print STDERR "[ERROR 'compare_2msa_respairs'] 'open_output_file' FAILED on the 1st input MSA file '$infile2': $! !\n";
	    return (undef, undef, undef, undef);
	}
	unless ( print_fasta (@tmp_seqids, @{$msa2}, $dis_len, *${$fh_out2}) ) {
	    print STDERR "[ERROR 'compare_2msa_respairs'] 'print_fasta' FAILED to output the 2nd MSA: $! !\n";
	    return (undef, undef, undef, undef);
	}
	close ($fh_out2);

=for TESTING the subroutine
		# Run the program to compare the MSAs #

	unless (0 == system ( "(time ${$exec_cmp_2msa} $infile1 $out_pref -m $infile2) &> $logfile" ) ) {
	    print STDERR "[ERROR 'compare_2msa_respairs'] FAINED to RUN '${$exec_cmp_2msa}' (via 'system') to compare the two MSAs: $? : $! !\n";
	    return (undef, undef, undef, undef);
	}

	my @outputs = ();

	foreach my $outfile ($outfile1, $outfile2) { # Outer foreach-loop #
	    unless (( -f $outfile ) and ( -r $outfile )) {
		print STDERR "[ERROR 'compare_2msa_respairs'] Found NO readable output file '$outfile': $! !\n";
		return (undef, undef, undef, undef);
	    }
	    my $fh_in = open_input_file2 ($outfile);
	    unless (defined $fh_in) {
		print STDERR "[ERROR 'compare_2msa_respairs'] 'open_input_file2' FAILED on the output file '$outfile': $! !\n";
		return (undef, undef, undef, undef);
	    } 

	    my $ct_sbs = my $tot_score = 0;
	    while (my $in = <$fh_in>) { # Inner while-loop #
		($in =~ m/^\#/) and next;
		my @fields = split ' ', $in;
		(@fields > 1) or next;

		$ct_sbs++;
		my $score = $fields[$#fields];
		$tot_score += $score;
	    }			# End of the inner while-loop #
	    close ($fh_in);
	    push @outputs, $ct_sbs, $tot_score;
	}			# End of the outer foreach-loop #
=cut

	unless ( chdir $orig_dir ) {
	    print STDERR "[ERROR 'compare_2msa_respairs'] FAILED to get back to the original directory '$orig_dir': $! !\n";
	    return (undef, undef, undef, undef);
	}

	my ($ct_respairs, $tot_respair_score, $ct_clms, $tot_clm_score) = (20, 10, 10, 5); ## For TESTING the script ###
	##my ($ct_respairs, $tot_respair_score, $ct_clms, $tot_clm_score) = @outputs;

		### RESTART FROM HERE ###

	return  ($ct_respairs, $tot_respair_score, $ct_clms, $tot_clm_score) ;

}	# End of the subroutine 'compare_2msa_respairs' #


# raw_cmp_2msa_respairs:
#	This subroutine conducts Dr. Giddy Landan's program
#	that compares two MSAs at the residue-pair level.
#	Then it appends the raw output files to a set of specified
#	output files.
#	(CREATED on May 2 (Mon), 2011)
#
# compare_2msa_respairs:
#	This subroutine compares two MSAs at the residue-pair level
#	using Dr. Giddy Landan's program 'msa_set_score'.
#
# [INPUT: (\@msa1, \@msa2, \$exec_cmp_2msa, \$tmp_work), where] -- for the predecessor 'compare_2msa_respairs'
# INPUT: (\@msa1, \@msa2, \$exec_cmp_2msa, \$tmp_work, \@fh_outs, $tmpout_pref, $header), where
#	${$msa1->[$k]} and ${$msa2->[$k]} are the $k th sequence embedded
#	in the 1st MSA and the 2nd MSA, respectively;
#	$exec_cmp_2msa is the ABSOLUTE path to the executable file comparing two MSAs;
#	$tmp_work is the path to the temporary working directory;
#	*${$fh_outs[$i]} is the file-handle for the $i th output file;
#	$tmpout_pref is the prefix of the names of the temporary output files;
#	$header is the header of each record.
#	(Each of the record ends with "\n#END\n" in the temporary file.
#	In the output file, a further "\n" will be appended);
#
# [OUTPUT: ($ct_respairs, $tot_respair_score, $ct_clms, $tot_clm_score), where
#	$ct_respairs is the number of residue pairs scored in the 1st MSA;
#	$tot_respair_score is the total residue pair scores,
#	which is basically the number of residue pairs correctly reproduced in the 2nd MSA;
#	$ct_clms is the number of columns scored in the 1st MSA;
#	$tot_clm_score is the total column scores,
#	which is basically the number of columns correctly reproduced in the 2nd MSA.] --- for the predecessor 'compare_2msa_respairs'
# OUTPUT: 1 if the process succeeded,
#	  0 if it failed.
#
##sub compare_2msa_respairs (\@\@\$\$) {
##	my ($msa1, $msa2, $exec_cmp_2msa, $tmp_work) = @_;
#
sub raw_cmp_2msa_respairs (\@\@\$\$\@$$) {

	my ($msa1, $msa2, $exec_cmp_2msa, $tmp_work, $fh_outs, $tmpout_pref, $header) = @_;

	my $dis_len = 60; # Maximum length of the sequence displayed in a line

		# ADDED in creation of 'raw_cmp_2msa_respairs' on May 5 (Thu), 2011 #
	my @tmpout_suffs = qw( col_col.scr msa.scr res_pair.scr res_pair_col.scr res_pair_res.scr res_pair_seq.scr res_pair_seq_pair.scr );
	my $ct_outs = @{$fh_outs};
	unless ($ct_outs == @tmpout_suffs) {
	    print STDERR "[ERROR 'raw_cmp_2msa_respairs'] ILLEGAL INPUT: The number ($ct_outs) of output file-handles CLASHES with what it is supposed to be (", scalar (@tmpout_suffs), "): $! !\n";
	    return 0;
	}

	my $ct_seqs = @{$msa1};
	unless (@{$msa2} == $ct_seqs) {
	    print STDERR "[ERROR 'raw_cmp_2msa_respairs'] ILLEGAL INPUT: The two MSAs have CONFLICTING NUMBER OF SEQUENCES: $ct_seqs for MSA1, and ", scalar (@{$msa2}), " for MSA2: $! !\n";
	    return 0;
	}

=for TESTING the subroutine
	unless (( -f ${$exec_cmp_2msa} ) and ( -x ${$exec_cmp_2msa} )) {
	    print STDERR "[ERROR 'raw_cmp_2msa_respairs'] CanNOT FIND the executable file '${$exec_cmp_2msa}': $!  !\n";
	    ##return (undef, undef, undef, undef);
	    return 0;
	}
=cut

	unless ( -d ${$tmp_work} ) {
	    unless (0 == system ("mkdir -p ${$tmp_work}" ) ) {
		print STDERR "[ERROR 'raw_cmp_2msa_respairs'] FAILED to CREATE the temporary working directory '${$tmp_work}': $? : $! !\n";
		return 0;
	    }
	}
	unless (( -r ${$tmp_work} ) and ( -w ${$tmp_work} )) {
	    unless (0 == system ("chmod u+rw ${$tmp_work}") ) {
		print STDERR "[ERROR 'raw_cmp_2msa_respairs'] FAILED to CHANGE permission MODES of the directory'${$tmp_work}': $? : $! !\n";
		return 0;
	    }
	}

	(my $orig_dir = ` pwd `) =~ s/\s+$//;

	unless ( chdir ${$tmp_work} ) {
	    print STDERR "[ERROR 'raw_cmp_2msa_respairs'] FAILED to change the directory to '${$tmp_work}': $! !\n";
	    return 0;
	}

	my @tmp_seqids = ();
	for (my $k=0; $k<$ct_seqs; $k++) { push @tmp_seqids, sprintf ("seq%04d", $k); }

	my $infile1 = 'msa1.fasta';
	my $infile2 = 'msa2.fasta';
	my @tmpoutfiles = ();
	foreach my $suff (@tmpout_suffs) { push @tmpoutfiles, join ('_', $tmpout_pref, $suff); }
=for the predecessor 'raw_cmp_2msa_respairs'
	my $out_pref = 'out';
	my $outfile1 = join ('_', $out_pref, 'res_pair.scr');
	my $outfile2 = join ('_', $out_pref, 'col_col.scr');
=cut
	my $logfile = 'logfile.txt';

		## Prepare input MSA files ##

	my $fh_tmpin1 = open_output_file ($infile1, 0);
	unless (defined $fh_tmpin1) {
	    print STDERR "[ERROR 'raw_cmp_2msa_respairs'] 'open_output_file' FAILED on the 1st input MSA file '$infile1': $! !\n";
	    return 0;
	}
	unless ( print_fasta (@tmp_seqids, @{$msa1}, $dis_len, *${$fh_tmpin1}) ) {
	    print STDERR "[ERROR 'raw_cmp_2msa_respairs'] 'print_fasta' FAILED to output the 1st MSA: $! !\n";
	    return 0;
	}
	close ($fh_tmpin1);

	my $fh_tmpin2 = open_output_file ($infile2, 0);
	unless (defined $fh_tmpin2) {
	    print STDERR "[ERROR 'raw_cmp_2msa_respairs'] 'open_output_file' FAILED on the 1st input MSA file '$infile2': $! !\n";
	    return 0;
	}
	unless ( print_fasta (@tmp_seqids, @{$msa2}, $dis_len, *${$fh_tmpin2}) ) {
	    print STDERR "[ERROR 'raw_cmp_2msa_respairs'] 'print_fasta' FAILED to output the 2nd MSA: $! !\n";
	    return 0;
	}
	close ($fh_tmpin2);

=for TESTING the subroutine
		# Run the program to compare the MSAs #

	unless (0 == system ( "(time ${$exec_cmp_2msa} $infile1 $tmpout_pref -m $infile2) &> $logfile" ) ) {
	    print STDERR "[ERROR 'raw_cmp_2msa_respairs'] FAINED to RUN '${$exec_cmp_2msa}' (via 'system') to compare the two MSAs: $? : $! !\n";
	    return 0;
	}

		# Transfer the temporary outputs to the corresponding output files #

	for (my $i=0; $i<$ct_outs; $i++) { # Outer for-loop #

	    my $tmpoutfile = $tmpoutfiles[$i];

	    unless (( -f $tmpoutfile ) and ( -r $tmpoutfile )) {
		print STDERR "[ERROR 'raw_cmp_2msa_respairs'] Found NO readable temporary output file '$tmpoutfile': $! !\n";
		return 0;
	    }
	    my $fh_tmpin = open_input_file2 ($tmpoutfile);
	    unless (defined $fh_tmpin) {
		print STDERR "[ERROR 'raw_cmp_2msa_respairs'] 'open_input_file2' FAILED on the temporary output file '$tmpoutfile': $! !\n";
		return 0;
	    } 

	    my $fh_out = $fh_outs->[$i];

	    print $fh_out $header, "\n"; # Print out the header #

	    while (my $tmpin = <$fh_tmpin>) {  print $fh_out $tmpin; }

	    close ($fh_tmpin);

	    print $fh_out "\n";
	}			# End of the outer for-loop #
=cut

	unless ( chdir $orig_dir ) {
	    print STDERR "[ERROR 'raw_cmp_2msa_respairs'] FAILED to get back to the original directory '$orig_dir': $! !\n";
	    return 0;
	}

	return 1;

=for the predecessor 'compare_2msa_respairs'
	my ($ct_respairs, $tot_respair_score, $ct_clms, $tot_clm_score) = (20, 10, 10, 5); ## For TESTING the script ###
	##my ($ct_respairs, $tot_respair_score, $ct_clms, $tot_clm_score) = @outputs;
	return  ($ct_respairs, $tot_respair_score, $ct_clms, $tot_clm_score) ;
=cut

}	# End of the subroutine 'raw_cmp_2msa_respairs' #


1;   # a 'spell' which enables the module 'use'-able and 'require'-able #


