package MyMalnManipu_pep;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.3: Copyright (C) 2013 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyMalnManipu_pep.pm:
#	This module is basically a collection of subroutines
#	each of which is forged by revising  a subroutine in the module
#	'MyMalnManipu.pm' so that it can handle peptide sequences.
#	The module only contains subroutines whose revision is necessary to cope with
#	peptide alignments.
#	Thus the module lacks subroutines which can innately handle both nucleotide
#	and peptide alignments, and those which can NOT EVEN IN PRINCIPLE convert to
#	the peptide counterparts.
#	You should find such subroutines in the original module 'MyMalnManipu.pm'.
#
# MyMalnManipu.pm:
#	This module manipulates multiple or pairwise alignments
#	in a more sophisticated manner than the module "MyMalnIO.pm".
#
# NOTE: This package is almost exclusively for nucleotide multiple-alignments.
#
# REVISED on Nov 19 (Fri), 2004:
#	The subroutine 'compare_columns', which is  a satellite of 'compare_malign',
#	no longer depends on the recursion technique.


use strict;
use warnings;
use Exporter;

use MyMalnManipu qw( block_starts_ends id_sliding_window );

our @ISA = qw( Exporter);

our @EXPORT_OK = qw( separate_nogap_sites_pep monosite_picker_pep id_indicator_aa
		print_block_maln_pep maln_identities_pep block_maln_identities_pep
		block_maln_id_distribution_pep bl_id_sliding_window_pep
		mask_frames_pep compl_mask_frames_pep select_unamb_frames_pep
		pr_block_maln_dec_pep
		pep_diff2dist pep_pairwise_diffs 
		smpl_compare_seq_pair matrix_gamma_dist matrix_dist_extJC
		   );
		###
		### Subroutines which you can find in the module 'MyMalnManipu.pm':
		###
		### pad_5p_end pad_3p_end merge_pad_align rm_gapcolumn rm_gapcolumn_codon
                ### assign_number separate_resnum compare_malign diff_malign
		### construct_blocks extract_bl_selected_frames block_starts_ends id_indicator 
		### equiv_columns trim_diff_maln average_id_proportions
 		### pick_poor_windows pick_bl_poor_windows
 		### select_in_phase sel_complete_codon_frames sel_codon_frames_in_ref
		### extract_codon_alignment extract_codon_aln_with_ref pr_in_oden_dec_pep
		### chop_alignment

###sub pick_bl_poor_windows (\@\@\@\@$$);
###sub pick_poor_windows (\@\@\@\@$$);
###sub compare_columns (\@\@$);
###sub compare_malign_fw (\@\@\@\@);
###sub compare_malign_bw (\@\@\@\@);

##sub phase_column (\@);
##sub id_indicator (\@);

sub id_indicator_aa (\@);


# separate_nogap_sites_pep: This subroutine receives the sequence part of a multiple alignment,
#			  separates columns into two groups according to whether it contains
#			  gaps '-' (or X) or not,
#			  and returns the lists of columns thus separated and of their positions
#			  (under the "start = 0" convention). 
#
sub separate_nogap_sites_pep (\@\@) {

	my ($rf_align, $rf_positions) = @_;

	my $ct_seqs = @{$rf_align};
	$ct_seqs  or (return ([],[],[],[]) );

	my @copy = ();
	foreach my $rf_seq (@{$rf_align}) { push @copy, ${$rf_seq}; }

	my $aln_len = length $copy[0];
	my $fl_pos = (@{$rf_positions} == $aln_len);

	my @nogap_sites = my @gapped_sites = my @nogap_pos = my @gapped_pos = ();

	for (my $l=0; $l <$aln_len; $l++) {	# the outer for #

		my @column = ();
		my $ct_gaps = 0;
		for (my $i=0; $i < $ct_seqs; $i++) {	# the inner for #

			my $char = substr ( $copy[$i], 0, 1, '');
			( ($char eq '-') or (uc($char) eq 'X') ) and ($ct_gaps++);
			push @column, $char;
		}				# end of the inner for #

		my $pos = $fl_pos ? $rf_positions->[$l] : $l;
		if ($ct_gaps) {

			push @gapped_sites , \@column;
			push @gapped_pos, $pos;

		} else {

			push @nogap_sites, \@column;
			push @nogap_pos, $pos;
		}
	}					# end of the outer for #

	return (\@nogap_sites, \@nogap_pos, \@gapped_sites, \@gapped_pos);

}			# end of the subroutine separate_nogap_sites_pep #



# monosite_picker_pep: This subroutine makes a list of
#		positions in the alignment where the column
#		consists only of residues of one kind.
#		This subroutine adopts the (START = 0) convention.
#
sub monosite_picker_pep (\@) {

	my ($rf_seqs) = @_;

	my @list_mono = ();
	my $ct_seqs = @{$rf_seqs};
	($ct_seqs) or (return \@list_mono ) ;

	my @copy = ();
	foreach my $seq (@{$rf_seqs}) { push @copy, ${$seq}; }
	my $seq_len = length (${$rf_seqs->[0]});
	my $first_seq = shift @copy;

	OUTER:
	#foreach my $l  (1 .. $seq_len) {
	foreach my $l  (0 .. $seq_len -1) {

		my $first_char = substr($first_seq, 0, 1, '');
		(defined $first_char) or last;
		$first_char = uc $first_char;
		##($first_char =~ m/[ACGTU]/ ) and ($ct_id++);
		##my $ct_id = 0;
		##($first_char =~ m/[A-Z]/ ) and ($first_char !~ m/[BZXJO]/ ) and ($ct_id++);
		my $ct_id = ($first_char =~ tr/ACDEFGHIKLMNPQRSTUVWY// );

		for (my $i=0; $i<@copy; $i++) {
			my $char = substr ($copy[$i], 0, 1, '');
			(defined $char) or (last OUTER);  # I regard the sequence has already ended. #
			(uc($char) eq $first_char) and $ct_id++;
		}

		($ct_id == $ct_seqs) and (push @list_mono, $l);
	}

	return \@list_mono;
}				# end of the subroutine monosite_picker_pep #


# id_indicator_aa: This subroutine makes a line which indicates
#		the positions of columns in the alignment
#		consisting only of one kind of amino acids.
#
sub id_indicator_aa (\@) {

	my ($rf_seqs) = @_;

	my $id_ind = '';
	my $ct_seqs = @{$rf_seqs};
	($ct_seqs) or (return \$id_ind ) ;

	my @copy = ();
	foreach my $seq (@{$rf_seqs}) { push @copy, uc ${$seq}; }
	my $seq_len = length (${$rf_seqs->[0]});
	my $first_seq = shift @copy;

	foreach my $l  (1 .. $seq_len) {

		my $first_char = substr($first_seq, 0, 1, '');
		#my $ct_id = ( ($first_char eq '-') or ($first_char eq 'N') ) ? 0 : 1;
		my $ct_id = ($first_char =~ tr/ACDEFGHIKLMNPQRSTUVWY// );
			# NOTE: U = 'selenocystein'

		for (my $i=0; $i<@copy; $i++) {
			my $char = substr ($copy[$i], 0, 1, '');
			($char eq $first_char) and $ct_id++;
		}

		$id_ind .= ($ct_id == $ct_seqs) ? '*' : ' ';
	}

	return \$id_ind;
}				# end of the subroutine id_indicator_aa #


# print_block_maln_pep: This subroutine prints out a  block-separated multiple- alignment.
#			It returns the count of completely identical columns in the alignment. 
#
sub print_block_maln_pep (\@\@\@\@\*$) {

	my ($rf_bl_columns, $rf_bl_frames, $rf_resnums, $rf_seqIDs, $fh, $offset) = @_;

	my $ct_blocks = @{$rf_bl_columns};
	my $ct_seqs = @{$rf_seqIDs};
	unless ($ct_blocks) { print $fh "\n"; return 0; }
 	unless (@{$rf_bl_frames} == $ct_blocks) {

		print STDERR "Illegal input for \"print_block_maln_pep\".\n";
		return undef;
	}

	my ($rf_bl_starts, $rf_bl_ends) = block_starts_ends (@{$rf_resnums}, @{$rf_bl_frames});

	unless (defined $rf_bl_starts) {

		print STDERR "Subroutine \"block_starts_ends\" failed.\n";
		return undef;
	}

	my @labels = ();
	foreach my $seqID (@{$rf_seqIDs}) {

		my $label = $seqID . ' ' x (15 - length($seqID) );
		push @labels, $label;
	}

	my $id_ct = 0;

	for (my $i=0; $i < $ct_blocks; $i++) { # the outer for #

	    my $frame = $rf_bl_frames->[$i];
	    my $ct_columns = @{$frame};
	    my $frame_start = $frame->[0] + 1;
	    my $frame_end = $frame->[$#{$frame}] + 1;
	    printf $fh "Block_%d (frame:%7d  --%7d, %7d columns)\n", $i+1+ $offset, $frame_start, $frame_end, $ct_columns;

	    my @reservoir = ();
	    for (my $j=0; $j < $ct_seqs; $j++) { # the inner for #
		my $start = $rf_bl_starts->[$i][$j];
		my $end = $rf_bl_ends->[$i][$j];
		unless (defined $start) { 
			print STDERR "The starting position of block $i, sequence $rf_seqIDs->[$j] is NOT defined.\n";
			return undef;
		}
		unless (defined $end) { 
			print STDERR "The ending position of block $i, sequence $rf_seqIDs->[$j] is NOT defined.\n";
			return undef;
		}
		$start = ($start eq '-') ? (' ' x 4 . '-' x 3) : sprintf ("%7d", $start);
		$end = ($end eq '-') ? (' ' x 4 . '-' x 3) : sprintf ("%7d", $end);
		my $seq = '';
		##foreach my $column (@{$rf_bl_columns->[$i]}) { $seq .= $column->[$j]; }
		for (my $k=0; $k<@{$rf_bl_columns->[$i]}; $k++) {

			unless (defined $rf_bl_columns->[$i][$k]) {

				print STDERR "The column $k of block $i is NOT defined.\n";
				return undef;
			}
			unless (defined $rf_bl_columns->[$i][$k][$j]) {

				print STDERR "The sequence $rf_seqIDs->[$j] is NOT defined at the column $k of block $i.\n";
				return undef;
			}
			$seq .= $rf_bl_columns->[$i][$k][$j];
		}

		my $line = $labels[$j] . $start . '  ' . $seq . $end . "\n";
		print $fh $line;
		push @reservoir, \$seq;

	    }				# end of the inner for #

	    my $id_ind = id_indicator_aa (@reservoir);
	    $id_ct +=  (${$id_ind} =~ tr/\*//);

	    print $fh ' ' x 24 . ${$id_ind} . "\n\n";
	}				# end of the outer for #

	return $id_ct;
}				# end of the subroutine print_block_maln_pep #


# maln_identities_pep: This subroutine receives a list of columns in a multiple alignment
#		of PEPTIDE sequences,
#		  counts the number of matched pairs for each column,
#		  and returns a list of numbers thus counted.
#
sub maln_identities_pep (\@) {

	my ($rf_columns) = @_;

	my $ct_columns = @{$rf_columns};
	($ct_columns) or ( return ([], []) );

	my @column_id_cts = ();
	my $tot_id_ct = 0;
	my $mono_ct = 0;	# the count of monomorphic sites (columns) #	
	my $tot_ct_pairs = 0;

	my $ct_seqs = @{$rf_columns->[0]};
	my $max_ct_pairs = $ct_seqs * ($ct_seqs -1) / 2;
	my $ct_nogap_clms = 0;

	foreach my $column (@{$rf_columns}) {	# the outer foreach #

		my %ct_residues;	# key = residue identity, value = its occurrence in a column #
		my $ct_nogaps = 0;
		foreach my $char (@{$column}) {

			$char = uc $char;
			##( ($char eq '-') or ($char eq 'N') ) and next;
			($char =~ tr/ACDEFGHIKLMNPQRSTUVWY// ) or next;
				# NOTE: U = 'selenocystein'

			(defined $ct_residues{$char}) ? ($ct_residues{$char}++) : ($ct_residues{$char} = 1);
			$ct_nogaps++;
		}

		($ct_nogaps == $ct_seqs) and  ($ct_nogap_clms++);
		my $ct_pairs = $ct_nogaps * ($ct_nogaps -1) /2;
		$tot_ct_pairs += $ct_pairs;

		my $id_ct = 0;
		foreach my $char (keys %ct_residues) {

			my $ct = $ct_residues{$char};
			($ct == 1) or ($id_ct += $ct* ($ct -1) /2 );
		}

		$tot_id_ct += $id_ct;
		push @column_id_cts, $id_ct;
		($id_ct  == $max_ct_pairs) and ($mono_ct++);
	}					# end of the outer foreach #

	#my $av_id_ct = $tot_id_ct/($ct_pairs*$ct_columns);
	my $av_id_ct = ($tot_ct_pairs) ? $tot_id_ct/$tot_ct_pairs : 0;
	my $av_mono_ct = ($ct_nogap_clms) ? $mono_ct/ $ct_nogap_clms : 0;
	my @column_id_stat = ($tot_id_ct, $mono_ct, $tot_ct_pairs, $ct_nogap_clms, $av_id_ct, $av_mono_ct);
	return (\@column_id_cts, \@column_id_stat);

}			# end of the subroutine maln_identities_pep #

# block_maln_identities_pep: This subroutine receives a list of blocks of columns
#	in a PEPTIDE alignment,
#	do similar things with the subroutine 'maln_identities_pep',
#	and returns a list of numbers distributed to the blocks.
#
sub block_maln_identities_pep (\@) {

	my ($rf_blocks) = @_;

	(@{$rf_blocks}) or ( return ([],[]) );

	my @block_id_cts = my @bl_id_stats = ();
	foreach my $block (@{$rf_blocks}) { 

		my ($column_id_cts, $column_id_stat)  = maln_identities_pep (@{$block});
		push @block_id_cts, $column_id_cts;
		push @bl_id_stats, $column_id_stat;
	}

	return (\@block_id_cts, \@bl_id_stats) ;

}			# end of the subroutine block_maln_identities_pep #


# maln_id_distribution_pep: This subroutine receives a list of columns in a PEPTIDE multiple alignment ,
#		  counts the number of matched pairs for each column,
#		  and returns a list of numbers thus counted, PLUS a list of numbers of
#		  residue pairs each for a column.
#
sub maln_id_distribution_pep (\@) {

	my ($rf_columns) = @_;

	my $ct_columns = @{$rf_columns};
	($ct_columns) or ( return ([], []) );

	my @column_id_cts = ();
	my @column_pair_cts = ();
	#my @column_mono_distri = ();

	my $ct_seqs = @{$rf_columns->[0]};
	my $max_ct_pairs = $ct_seqs * ($ct_seqs -1) / 2;

	foreach my $column (@{$rf_columns}) {	# the outer foreach #

		my %ct_residues;	# key = residue identity, value = its occurrence in a column #
		my $ct_nogaps = 0;
		foreach my $char (@{$column}) {

			$char  = uc $char ;
			##( ($char eq '-') or ($char eq 'X') ) and next;
			($char =~ tr/ACDEFGHIKLMNPQRSTUVWY// ) or next;
				# NOTE: U = 'selenocystein'

			(defined $ct_residues{$char}) ? ($ct_residues{$char}++) : ($ct_residues{$char} = 1);
			$ct_nogaps++;
		}

		my $id_ct = 0;
		foreach my $char (keys %ct_residues) {

			my $ct = $ct_residues{$char};
			($ct == 1) or ($id_ct += $ct* ($ct -1) /2 );
		}

		push @column_id_cts, $id_ct;
		my $ct_pairs = $ct_nogaps * ($ct_nogaps -1) /2;
		push @column_pair_cts, $ct_pairs;
		#push @column_mono_distri, ( ($id_ct == $max_ct_pairs) ? 1 : 0 );
	}					# end of the outer foreach #

	return (\@column_id_cts, \@column_pair_cts);

}			# end of the subroutine maln_id_distribution_pep #


# block_maln_id_distribution_pep: This subroutine receives a list of blocks of columns
#		in a PEPTIDE MULTIPLE ALIGNMENT ,
#		do similar things to the subroutine 'maln_id_distribution_pep',
#		and returns a list of numbers each for each block.
#
sub block_maln_id_distribution_pep (\@) {

	my ($rf_bl_columns) = @_;

	(@{$rf_bl_columns}) or ( return ([],[]) );

	my @block_id_distri = my @bl_pair_distri = ();
	foreach my $block (@{$rf_bl_columns}) { 

		my ($column_id_distri, $column_pair_distri)  = maln_id_distribution_pep (@{$block});
		push @block_id_distri, $column_id_distri;
		push @bl_pair_distri, $column_pair_distri;
		#push @bl_mono_distri, $column_mono_distri;
	}

	return (\@block_id_distri, \@bl_pair_distri) ;

}			# end of the subroutine block_maln_id_distribution_pep_pep #



# bl_id_sliding_window_pep: This subroutine conducts sliding-window analysis
#			on the identity distribution for each of blocks.
#
sub bl_id_sliding_window_pep (\@\@$$) {

	my ($rf_bl_columns, $rf_bl_frames, $window_size, $step_size) = @_;

	my $bl_ct = @{$rf_bl_columns};
	$bl_ct or (return ([],[]) );

	my $ct_seqs = @{$rf_bl_columns->[0][0]};

	my ($bl_id_distri, $bl_pair_distri) = block_maln_id_distribution_pep (@{$rf_bl_columns});
	my @matched_propts = my @mono_column_propts = my @wd_starts = my @wd_ends  = ();

	for (my $i=0; $i< $bl_ct; $i++) {

		my ($bl_mtchd_propts, $bl_monoclm_propts, $bl_wd_starts, $bl_wd_ends)
		= id_sliding_window (@{$bl_pair_distri->[$i]}, @{$bl_id_distri->[$i]}, @{$rf_bl_frames->[$i]}, $ct_seqs, $window_size, $step_size);

		push @matched_propts, $bl_mtchd_propts;
		push @mono_column_propts, $bl_monoclm_propts;
		push @wd_starts, $bl_wd_starts;
		push @wd_ends, $bl_wd_ends;
	}		

	return (\@matched_propts, \@mono_column_propts, \@wd_starts, \@wd_ends, $bl_id_distri, $bl_pair_distri);

} 			# end of the subroutine bl_id_sliding_window_pep #



# mask_frames_pep : This subroutine receives the sequence part of a PEPTIDE multiple-alignment and a list of frames,
#		and returns sequences  with the specified frames masked.
#
sub mask_frames_pep (\@\@) {

	my ($rf_align, $rf_frames) = @_;

	my $ct_seq = @{$rf_align};
	$ct_seq or (return [] );

		# a part of the fatal-bug fixation (on Feb 18 (Wed), 2004 ) #
	#unless (@{$rf_frames}) {
	#	my @tmp =  @{$rf_align};
	#	return \@tmp;
	#}

	my @out_align = my @copy = ();
	foreach my $align (@{$rf_align}) {
		my $tmp = '';
		push @out_align , \$tmp;
		#push @copy, ${$align};
		my $seqcp = ${$align};
		push @copy, \$seqcp;	# a bug fixation on March 20 (Sat), 2004 #
					# accompanied by the change from $copy[$i]  to ${$copy[$i]} below #
	}
		# the rest of the fatal-bug fixation (on Feb 18 (Wed), 2004 ) #
	(@{$rf_frames}) or (return \@copy);
	
	my $len_align = length ${$copy[0]};

	my $expected_frame = 0;
	foreach my $frame (@{$rf_frames}) {

		my $len_unmasked = $frame - $expected_frame;
		if ($len_unmasked > 0) {
			for (my $i=0; $i<$ct_seq; $i++) {
				${$out_align[$i]} .= substr ( ${$copy[$i]}, 0, $len_unmasked, '');
			}
				
		} elsif ($len_unmasked < 0) {
			print STDERR "negative unmasked length (expected_frame = $expected_frame, real_frame = $frame).\n";
			return undef;
		}

		for (my $i=0; $i<$ct_seq; $i++) {
			my $char = substr (${$copy[$i]}, 0, 1, '');
			${$out_align[$i]} .= ( ($char eq '*') or ($char eq '-') ) ? '-' : 'X';
		}

		$expected_frame = $frame + 1;

	}				# end of the foreach #

	for (my $i=0; $i<$ct_seq; $i++) { ${$out_align[$i]} .= ${$copy[$i]}; }

	return \@out_align;

}				# end of the subroutine mask_frames_pep #


# compl_mask_frames_pep : This subroutine receives the sequence part of
#		a PEPTIDE multiple-alignment and a list of frames,
#		and returns sequences with the frames complementary to the specified ones  masked.
#
sub compl_mask_frames_pep (\@\@) {

	my ($rf_align, $rf_frames) = @_;

	my $ct_seq = @{$rf_align};
	$ct_seq or (return [] );

		# fixing a fatal bug (on Feb 18 (Wed), 2004) #
	#unless (@{$rf_frames}) {		
	#	my @tmp =  @{$rf_align};
	#	return \@tmp;
	#}

	my @out_align = my @copy = ();
	foreach my $align (@{$rf_align}) {
		my $tmp = '';
		push @out_align , \$tmp;
		push @copy, ${$align};
	}

	my $len_align = length $copy[0];

	my $expected_frame = 0;
	foreach my $frame (@{$rf_frames}) {

		my $len_masked = $frame - $expected_frame;
		if ($len_masked > 0) {
			for (my $i=0; $i<$ct_seq; $i++) {
				my $tmp = substr ( $copy[$i], 0, $len_masked, '');
				$tmp =~ tr/a-zA-Z/X/;
				${$out_align[$i]} .= $tmp;
			}
				
		} elsif ($len_masked < 0) {
			print STDERR "negative masked length (expected_frame = $expected_frame, real_frame = $frame).\n";
			return undef;
		}

		for (my $i=0; $i<$ct_seq; $i++) {
			${$out_align[$i]} .= substr ($copy[$i], 0, 1, '');
		}

		$expected_frame = $frame + 1;

	}				# end of the foreach #

	for (my $i=0; $i<$ct_seq; $i++) { 

		$copy[$i] =~ tr/a-zA-Z/X/;
		${$out_align[$i]} .= $copy[$i];
	}

	return \@out_align;

}				# end of the subroutine compl_mask_frames_pep #


# select_unamb_frames_pep:
#	 This subroutine recieves a PEPTIDE alignment and a subset of its frames,
#	chooses frames which consists solely of unambiguous bases,
#	and returns a list of those frames.
#
# 	The subroutine basically follows the frame numbering convention that (start)=0.
#
sub select_unamb_frames_pep (\@\@) {

	my ($rf_align, $rf_selframe) = @_;

	(@{$rf_selframe}) or ( return []); 

	my @copy = my @ng_frames = ();
	foreach my $rf_seq (@{$rf_align}) {
		push @copy, uc ${$rf_seq};
	}

	my $exp_frame = 0;
	foreach my $frame (@{$rf_selframe}) {

		if ( (my $diff = $frame - $exp_frame) > 0)  {

			for (my $i=0; $i<@copy; $i++) { substr($copy[$i], 0, $diff, ''); }

		} elsif ($diff < 0) {

			print STDERR "Error in diff_malign:\n";
			print STDERR "expected frame ($exp_frame) exceeds the real one ($frame).\n";
			return undef;
		}

		my $fl_nogap = 1; 
		for (my $i=0; $i<@copy; $i++) {

			my $res = substr($copy[$i], 0, 1, '');
			( $res  =~ tr/ACDEFGHIKLMNPQRSTUVWY// ) or ($fl_nogap = 0);
				# Note: 'U' = selenocystein #
		}

		$fl_nogap and (push @ng_frames, $frame);
		$exp_frame = $frame + 1;
		
	}

	return \@ng_frames;

}			# end of the subroutine select_unamb_frames_pep #


# pr_block_maln_dec_pep: This subroutine prints out a  block-separated 
#			PEPTIDE multiple- alignment in a decorated form.
#			It returns the count of completely identical columns in the alignment. 
#
sub pr_block_maln_dec_pep (\@\@\@\@\*$) {

	my ($rf_bl_columns, $rf_bl_frames, $rf_resnums, $rf_seqIDs, $fh, $bnum) = @_;

	my $ct_blocks = @{$rf_bl_columns};
	my $ct_seqs = @{$rf_seqIDs};
	unless ($ct_blocks) { print $fh "\n"; return 0; }
 	unless (@{$rf_bl_frames} == $ct_blocks) {
		print STDERR "Illegal input for \"print_block_maln_dec\".\n";
		return undef;
	}

	my ($rf_bl_starts, $rf_bl_ends) = block_starts_ends (@{$rf_resnums}, @{$rf_bl_frames});

	unless (defined $rf_bl_starts) {
		print STDERR "Subroutine \"block_starts_ends\" failed.\n";
		return undef;
	}

	my @labels = ();
	foreach my $seqID (@{$rf_seqIDs}) {

		my $label = $seqID . ' ' x (15 - length($seqID) );
		push @labels, $label;
	}

	my $id_ct = 0;

	for (my $i=0; $i < $ct_blocks; $i++) { # the outer for #

	    my $frame = $rf_bl_frames->[$i];
	    my $ct_columns = @{$frame};
	    my $frame_start = $frame->[0] + 1;
	    my $frame_end = $frame->[$#{$frame}] + 1;
	    printf $fh "> (frame: %d -- %d, %d columns) ", $frame_start, $frame_end, $ct_columns;

	    my @reservoir = my @intervals = ();
	    for (my $j=0; $j < $ct_seqs; $j++) { # the inner for #
		my $start = $rf_bl_starts->[$i][$j];
		my $end = $rf_bl_ends->[$i][$j];
		unless (defined $start) { 
			print STDERR "The starting position of block $i, sequence $rf_seqIDs->[$j] is NOT defined.\n";
			return undef;
		}
		unless (defined $end) { 
			print STDERR "The ending position of block $i, sequence $rf_seqIDs->[$j] is NOT defined.\n";
			return undef;
		}
		#$start = ($start eq '-') ? (' ' x 4 . '-' x 3) : sprintf ("%7d", $start);
		#$end = ($end eq '-') ? (' ' x 4 . '-' x 3) : sprintf ("%7d", $end);
		my $seq = '';
		for (my $k=0; $k<@{$rf_bl_columns->[$i]}; $k++) {

			unless (defined $rf_bl_columns->[$i][$k]) {
				print STDERR "The column $k of block $i is NOT defined.\n";
				return undef;
			}
			unless (defined $rf_bl_columns->[$i][$k][$j]) {
				print STDERR "The sequence $rf_seqIDs->[$j] is NOT defined at the column $k of block $i.\n";
				return undef;
			}
			$seq .= $rf_bl_columns->[$i][$k][$j];
		}

		#my $line = $labels[$j] . $start . '  ' . $seq . $end . "\n";
		#print $fh $line;
		push @reservoir, \$seq;
		push @intervals, "$start -- $end";

	    }				# end of the inner for #

	    my $id_ind = id_indicator_aa (@reservoir);
	    $id_ct +=  (${$id_ind} =~ tr/\*//);

	    print $fh " (intervals: " . join( '; ', @intervals) . ")\n";

	   for (my $j=0; $j<$ct_seqs; $j++) { print $fh "$labels[$j]${$reservoir[$j]}\n"; }
	   print $fh '>' . ' ' x 14 . ${$id_ind} . "\n";
	   print $fh "\n";
	}				# end of the outer for #

	print $fh "\n";
	return $id_ct;
}				# end of the subroutine pr_block_maln_dec #


# pep_pairwise_diffs: This subroutine receives a PEPTIDE pairwise alignment
#			and returns the counts of sites
#			in which the amino acid residues 
#			of the 2 sequences differ to each other.
#			It also returns the total number of
#			alignment sites consisting solely of 
#			unambiguous characters.
#			If the sequence length are different in the
#			input alignment, the subroutine assumes
#			that the alignment is left justified.
#
sub pep_pairwise_diffs (\@) {

	my ($rf_pairaln) = @_;

	unless (@{$rf_pairaln} == 2) {
		print STDERR "Illegal input for \'count_nt_subst_pair\'.\n";
		return (undef, undef);
	}

		# make a copy of the sequence pair #
	my $cp_seq0 = uc ${$rf_pairaln->[0]};
	my $cp_seq1 = uc ${$rf_pairaln->[1]};

	my $aln_len = (length ($cp_seq0 ) < length ($cp_seq1) ) ?
			length ($cp_seq0) : length ($cp_seq1) ;

	my $unamb_ct = my $diff = my $transition = 0;
	for (my $i=0; $i<$aln_len; $i++) {

		my $char0 = substr($cp_seq0, 0, 1, '');
		my $char1 = substr($cp_seq1, 0, 1, '');

			# skip columns with gaps and/or ambiguous amino acids
		($char0 =~ tr/ACDEFGHIKLMNPQRSTUVWY\*// ) or next;
			# U: selenocystein, *: stop codon #

		$unamb_ct++;

		($char0 eq $char1) and next;	# skip invariable sites

		$diff++;

	}		# end of the outer for #

	return ($unamb_ct, $diff);

}		# end of the subroutine pep_pairwise_diffs #


# pep_diff2dist: This subroutine receives the count of unambiguous sites as well as
#	that of amino acid differences for a peptide pairwise alignment,
#	then calculates the Poisson distance and the distance with Kimura's correction.
#
# INPUT: (#{unamb_sites}, #{diffs})
#
# OUTPUT: (p_dist = #{diffs}/#{unamb_sites}, Poisson distance, Kimura's distance)
#
sub pep_diff2dist ($$) {

	my ($ct_unamb, $ct_diffs) = @_;

	($ct_unamb > 0) or (return (undef, undef, undef) );

	my $p_dist = $ct_diffs/$ct_unamb ;

	my $poisson = ($p_dist < 1) ? ( - log ( 1 - $p_dist) ) : undef ; 

	my $exp_kimura = 1 - $p_dist - $p_dist * $p_dist /5;
	my $kimura = ($exp_kimura > 0) ? ( - log ($exp_kimura) ) : undef;

	return ($p_dist, $poisson, $kimura );

}		# end of the subroutine pep_diff2dist #


# smpl_compare_seq_pair:
#	This satellite subroutine of 'matrix_gamma_dist'
#	makes a simple comparison of the two sequences
#	in a pairwise alignment site-by-site.
#
# INPUT: (\$seq1, \$seq2, $gap), where
#	$seq1 and $seq2 are the sequences embedded in the input pairwise alignment;
#
# OUTPUT: ($ct_sites, $ct_diffs), where
#	$ct_sites is the number of sites NOT containing gaps;
#	$ct_diffs is the number of sites where the characters differ between the sequences.
#
sub smpl_compare_seq_pair (\$\$$) {

	my ($seq1, $seq2, $gap) = @_;

	my $cp_seq1 = ${$seq1};
	my $cp_seq2 = ${$seq2};

	my $len = length $cp_seq1;
	unless ($len  == length $cp_seq2) {
	    print STDERR "[ERROR 'smpl_compare_seq_pair'] ILLEGAL INPUT: The LENGTHS DIFFER between the sequences ($len vs ", length (${$cp_seq2}), "): $! !\n";
	    return (undef, undef); 
	}

	my $ct_sites = my $ct_diffs = 0;

	for (my $l=0; $l<$len; $l++) { # Main for-loop #
	    my $char1 = substr ($cp_seq1, 0, 1, '');
	    my $char2 = substr ($cp_seq2, 0, 1, '');
	    (($char1 eq $gap) or ($char2 eq $gap)) and next; # Skip if there is a gap
	    $ct_sites++;
	    ($char1 eq $char2) or ($ct_diffs++);
	}		# End of the main for-loop #

	return ($ct_sites, $ct_diffs);

}	# End of the subroutine 'smpl_compare_seq_pair' #


# matrix_gamma_dist:
#	This subroutine constructs a matrix of pairwise gamma distances
#	between peptide sequences in a MSA.
#	It ignores sites containing gaps.
#	The gamma distance $d between sequences with the p-distance $p is:
#	$d = $alpha * [(1 - $p)** (-1/$alpha) -1 ],
#	where $alpha is the shape parameter of the gamma distribution.
#	When $alpha = infinity, it should approach the Poisson distance (- log (1 - $p) ).
#
# INPUT: (\@align, $gap, $alpha), where
#	${$align[$k]} is the $k th sequence embedded in the alignment;
#	$gap is the character representing a gap;
#	$alpha is the shape parameter of the gamma distribution.
#	If $alpha = 0, the Poisson distance - log (1 - $p) is calculated.
#
# OUTPUT: (\@mtrx_dist, \@mtrx_ct_diffs, \@mtrx_ct_sites), where
#	$mtrx_dist[$i]->[$j] (0 <= $j < $i <= $ct_seqs - 1) is the distance between the $i th and $j th sequences;
#	$mtrx_ct_diffs[$i]->[$j] is the number of sites different between the $i th and the $j th sequences;
#	$mtrx_ct_sites[$i]->[$j] is the number of sites containing NO gaps in the alignment between the $i th and $j th sequences. It is undef if the alignment has NO effective sites.
#
sub matrix_gamma_dist (\@$$) {

	my ($align, $gap, $alpha) = @_;

	my $ct_seqs = @{$align};
	my $len = length ${$align->[0]};

	my @mtrx_dist = my @mtrx_ct_diffs = my @mtrx_ct_sites  = ();
	for (my $i=0; $i<$ct_seqs; $i++) { # Outer for-loop #

	    my @row_dist = my @row_ct_diffs = my @row_ct_sites = ();
	    for (my $j=0; $j<$i; $j++) {  # Inner for-loop #
		my ($ct_sites, $ct_diff) = smpl_compare_seq_pair (${$align->[$i]}, ${$align->[$j]}, $gap);
		unless (defined $ct_sites) {
		    print STDERR "[ERROR 'matrix_gamma_dist'] The subroutine 'smpl_compare_seq_pair' FAILED between the $i th sequence and the $j th sequence in the alignment: $! !\n";
		    return (undef, undef, undef);
		}
		push @row_ct_sites, $ct_sites;
		push @row_ct_diffs, $ct_diff;
		if ($ct_sites == 0) {
		    push @row_dist, undef;
		    next;
		}
		if ($ct_diff == 0) {
		    push @row_dist, 0;
		    next;
		}
		my $poi_dist = - log (1 - $ct_diff /$ct_sites);
		my $gdist = ($alpha == 0) ? $poi_dist :
			$alpha * (exp ($poi_dist / $alpha) -1);
		push @row_dist, $gdist;
	    }			# End of the inner for-loop #

	    push @mtrx_ct_sites, \@row_ct_sites;
	    push @mtrx_ct_diffs, \@row_ct_diffs;
	    push @mtrx_dist, \@row_dist;
	}			# End of the outer for-loop #

	return (\@mtrx_dist, \@mtrx_ct_diffs, \@mtrx_ct_sites);

}		# End of the subroutine 'matrix_gamma_dist' #


# matrix_dist_extJC:
#	This subroutine constructs a matrix of pairwise distances
#	between peptide sequences in a MSA using an extended JC model.
#	It ignores sites containing gaps.
#	The extended JC dstance $d between sequences with the p-distance $p is:
#	$d = - ($c -1)/$c  * log (1 - $c/($c-1) * $p),
#	where $c is the number of possible characters.
#	When $c = infinity, it should approach the Poisson distance (- log (1 - $p) ).
#	When $c = 4, it is the ordinary Jukes-Cantor model.
#
#	We set an 'upper-bound' $d_up of the distance. 
#	If ($p > ($c - 1)/$c) or $d > $d_up,
#	the distance is automatically set to be $d_up.
#
#	The subroutine was made from the framework of 'matrix_gamma_dist'.
#	(CREATED on June 24 (Thu), 2010)
#
# INPUT: (\@align, $gap, $c, $d_up), where
#	${$align[$k]} is the $k th sequence embedded in the alignment;
#	$c is the number of possible characters, which should be >= 2;
#	$d_up is the upper-bound of the distance.
#
# OUTPUT: (\@mtrx_dist, \@mtrx_ct_diffs, \@mtrx_ct_sites), where
#	$mtrx_dist[$i]->[$j] (0 <= $j < $i <= $ct_seqs - 1) is the distance between the $i th and $j th sequences;
#	$mtrx_ct_diffs[$i]->[$j] is the number of sites different between the $i th and the $j th sequences;
#	$mtrx_ct_sites[$i]->[$j] is the number of sites containing NO gaps in the alignment between the $i th and $j th sequences. It is undef if the alignment has NO effective sites.
#
sub matrix_dist_extJC (\@$$$) {

	my ($align, $gap, $c, $d_up) = @_;

	my $ct_seqs = @{$align};
	my $len = length ${$align->[0]};

	my @mtrx_dist = my @mtrx_ct_diffs = my @mtrx_ct_sites  = ();
	for (my $i=0; $i<$ct_seqs; $i++) { # Outer for-loop #

	    my @row_dist = my @row_ct_diffs = my @row_ct_sites = ();
	    for (my $j=0; $j<$i; $j++) {  # Inner for-loop #
		my ($ct_sites, $ct_diff) = smpl_compare_seq_pair (${$align->[$i]}, ${$align->[$j]}, $gap);
		unless (defined $ct_sites) {
		    print STDERR "[ERROR 'matrix_dist_extJC'] The subroutine 'smpl_compare_seq_pair' FAILED between the $i th sequence and the $j th sequence in the alignment: $! !\n";
		    return (undef, undef, undef);
		}
		push @row_ct_sites, $ct_sites;
		push @row_ct_diffs, $ct_diff;
		if ($ct_sites == 0) {
		    push @row_dist, undef;
		    next;
		}
		if ($ct_diff == 0) {
		    push @row_dist, 0;
		    next;
		}
		my $p = $ct_diff/$ct_sites;
		my $dist = ($c * $ct_diff < ($c -1) * $ct_sites) ? 
			- ($c -1)/$c * log (1 - $c/($c -1) * $p) : undef;
		((defined $dist) and ($dist < $d_up)) or ($dist = $d_up); # Force the distance  equal to the upper bound in an unacceptable situation
		push @row_dist, $dist;
	    }			# End of the inner for-loop #

	    push @mtrx_ct_sites, \@row_ct_sites;
	    push @mtrx_ct_diffs, \@row_ct_diffs;
	    push @mtrx_dist, \@row_dist;
	}			# End of the outer for-loop #

	return (\@mtrx_dist, \@mtrx_ct_diffs, \@mtrx_ct_sites);

}		# End of the subroutine 'matrix_dist_extJC' #



1;		# a 'spell' which enables the module 'use'-able and 'require'-able #


