package MyMalnManipu_suppl;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.3: Copyright (C) 2013 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyMalnManipu_suppl.pm:
#	This module contains some subroutines
#	each of which manipulates alignments in a manner
#	that cannot be performed by the subroutines
#	in the module 'MyMalnManipu.pm'.
#
# MyMalnManipu.pm:
#	This module manipulates multiple or pairwise alignments
#	in a more sophisticated manner than the module "MyMalnIO.pm".
#
# NOTE: This package is almost exclusively for nucleotide multiple-alignments.
#
# REVISED on Nov 19 (Fri), 2004:
#	The subroutine 'compare_columns', which is  a satellite of 'compare_malign',
#	no longer depends on the recursion technique.
#

use strict;
use warnings;

use MyArrayManipu qw( compare_arrays compare_interval_sets );
use MyArrayManipu_num qw( union_nmarray );
use MyMisc qw( count_spec_chars string_intvl_set );

use MyNucSeq qw( print_sequence );
use MyMalnIO qw( extract_clw pr_ODEN_input pr_clustal_like_lite );
use MyMalnManipu qw( rm_gapcolumn2 );
use  MyBlastp_parser_05 qw( score_pairalign );

use MyFile_manipu qw( open_output_file );
use MyTreeManipu_II qw(
	extract_sgltree_from_file
	br_output_tree br_output_tree_sep
	);

use Exporter;

our @ISA = qw( Exporter);

our @EXPORT_OK = qw(
	rm_spec_sites rm_multi_spec_sites concatenate_alignments
	extract_gap_blocks extract_indels_eqbd
	pad_concate_blocks gap_indicator
	splice_blocks_in_aln pick_sbset_aln
	chop_aln_by_qubds gaps_in_sglseq
	maxmin_seqlen
	examine_sel_alnd_regions
	bundle_refine_maln_by_clw
	extract_unmasked_regions
	clip_specregions_maln
	extract_specregions_maln mask_site
	mask_spec_regions mask_unspec_regions
	examine_tgt_regions_pwaln cnct_examine_tgt_regions_pwaln
	);


our $TOP_NODE = $MyTreeManipu_II::TOP_NODE ;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH ;


# rm_spec_sites:
#	This subroutine extracts sites that contains NO specified characters.
#	NOTE that this subroutine eats up the input alignment.
#
# INPUT: (\@align, $aln_beg, $char_spec), where
#	${$align[$m]} is the $m th sequence in the alignment;
#	$aln_beg is the start site number in the coordinate
#	frame representing the alignment;
#	$char_spec is the character specified to be removed.
#
# OUTPUT: (\@nspec_align, \@site_coords), where
#	${$ng_align[$m]} is the $m th sequence in the alignment
#	consisting only of sites without any gaps;
#	@site_nos is the alignment coordinates of such sites.
#
sub rm_spec_sites (\@$$) {

	my ($align, $aln_beg, $char_spec) = @_;

	my $ct_seq = @{$align};
	($ct_seq>0) or (return ([], []) );

	my @nspec_aln = my @site_coords = ();
	for (1 .. $ct_seq) { # The 1st outer for-loop #
	    my $tmp = '';
	    push @nspec_aln, \$tmp;
	}		# End of the 1st outer for-loop #

	my $aln_len = length ${$align->[0]};
	for (my $l=0; $l<$aln_len; $l++) { # The 2nd outer for-loop #

	    my @residues = ();
	    my $ct_specs = 0;
	    for (my $m=0; $m<$ct_seq; $m++) { # The 1st inner for-loop #
		my $res = substr (${$align->[$m]}, 0, 1, '');
		push @residues, $res;
		($res eq $char_spec) and ($ct_specs++);
	    }			# End of the 1st inner for-loop #

	    ($ct_specs>0) and next;

	    push @site_coords, $aln_beg + $l;

	    for (my $m=0; $m<$ct_seq; $m++) { # The 2nd inner for-loop #
		${$nspec_aln[$m]} .= $residues[$m];
	    }			# End of the 2nd inner for-loop #

	}			# End of the 2nd outer for-loop #

	return (\@nspec_aln, \@site_coords);

}		# End of the subroutine 'rm_spec_sites' #


# rm_multi_spec_sites:
#	This subroutine extracts sites that contains NO specified characters.
#	We can specify multiple characters.
#	NOTE that this subroutine eats up the input alignment.
#
# INPUT: (\@align, $aln_beg, \@char_spec, $fl_case_ignore), where
#	${$align[$m]} is the $m th sequence in the alignment;
#	$aln_beg is the start site number in the coordinate
#	frame representing the alignment;
#	@char_spec is a list of characters specified to be removed;
#	$fl_case_ignore
#		=1 if you want to ignore the difference between the upper and lower cases,
#		=0 if you want to differentiate between the two cases.
#
# OUTPUT: (\@nspec_align, \@site_coords), where
#	${$nspec_align[$m]} is the $m th sequence in the alignment
#	consisting only of sites without any specified characters;
#	@site_coords is the alignment coordinates of such sites.
#
sub rm_multi_spec_sites (\@$\@$) {

	my ($align, $aln_beg, $char_spec, $fl_ci) = @_;

	my $ct_seq = @{$align};
	($ct_seq>0) or (return ([], []) );

		## Make a hash serving as multi-flags ##
	my %specified ; # ($character => 1, ...) for those to be removed.
	foreach my $char (@{$char_spec}) {
		$specified{$char} = 1;
		if ($fl_ci) {
		   $specified{uc $char} = $specified{ lc $char} = 1;
		}
	}

		## Initialize the output ##
	my @nspec_aln = my @site_coords = ();
	for (1 .. $ct_seq) { # The 1st outer for-loop #
	    my $tmp = '';
	    push @nspec_aln, \$tmp;
	}		# End of the 1st outer for-loop #

		## The main process ##
	my $aln_len = length ${$align->[0]};
	for (my $l=0; $l<$aln_len; $l++) { # The 2nd outer for-loop #

	    my @residues = ();
	    my $ct_specs = 0;
	    for (my $m=0; $m<$ct_seq; $m++) { # The 1st inner for-loop #
		my $res = substr (${$align->[$m]}, 0, 1, '');
		push @residues, $res;
		(defined $specified{$res}) and ($ct_specs++);
	    }			# End of the 1st inner for-loop #

	    ($ct_specs>0) and next;

	    push @site_coords, $aln_beg + $l;

	    for (my $m=0; $m<$ct_seq; $m++) { # The 2nd inner for-loop #
		${$nspec_aln[$m]} .= $residues[$m];
	    }			# End of the 2nd inner for-loop #

	}			# End of the 2nd outer for-loop #

	return (\@nspec_aln, \@site_coords);

}		# End of the subroutine 'rm_multi_spec_sites' #


# concatenate_alignments:
#	This satellite subroutine of 'cnct_examine_tgt_regions_pwaln'
#	just concatenates a number of (multiple or pairwise) alignments.
#
# INPUT: (\@set_alns, \@set_coords), where
#	${$set_alns[$r]->[$m]} is the $m th sequence embedded
#	in the $r th alignment;
#	@{$set_coords[$r]->[$m]} = ($beg_aln, $end_aln) is the
#	beginning/ending coordinates of the $r th alignment
#	in the coordinate system of the $m th sequence.
#
# OUTPUT: (\@cnct_align, \@cnct_coords), where
#	${$cnct_align[$m]} is the $m th sequence embedded in the
#	concatenated alignment;
#	${$cnct_coords[$m]} = "[$beg_1,$end_1][...]" is a string representation
#	of the boundary coordinates of the $m th sequence.
#
sub concatenate_alignments (\@\@) {

	my ($set_alns, $set_coords) = @_;

	my $ct_alns = @{$set_alns};
	($ct_alns>0) or (return []);
	my $ct_seqs = @{$set_alns->[0]};

	my @cnct_align = my @cnct_coords = ();
	for (1 .. $ct_seqs) {
	    my $null_str = my $null_str2 = '';
	    push @cnct_align, \$null_str;
	    push @cnct_coords, \$null_str2;
	}

	for (my $r=0; $r<$ct_alns; $r++) { # The outer for-loop #
	    my $align = $set_alns->[$r];
	    unless (@{$align} == $ct_seqs) {
		my $ct_seqs2 = @{$align};
		print STDERR "[ERROR 'concatenate_alignments'] The $r th alignment has $ct_seqs2 sequences,\n";
		print STDERR "    which CLASHES the number of sequences in the 0 th alignment $ct_seqs !\n";
		return (undef, undef) ;
	    }

	    my $coords = $set_coords->[$r];
	    (defined $coords) or ($coords = []);

	    for (my $m=0; $m<$ct_seqs; $m++) { # The inner for-loop #
		${$cnct_align[$m]} .= ${$align->[$m]};

		my ($beg, $end) = (defined $coords->[$m]) ?
			@{$coords->[$m]} : ();
		(defined $beg) or ($beg = 'NA');
		(defined $end) or ($end = 'NA');

		${$cnct_coords[$m]} .= "\[$beg\,$end\]";
	    }			# End of the inner for-loop #

	}		# End of the outer for-loop #

	return (\@cnct_align, \@cnct_coords);

}			# End of the subroutine 'concatenate_alignments' #


# extract_gap_blocks:
#	This subroutine extracts blocks of sites with the same gap patterns
#	from a multiple alignment.
#	NOTE that this subroutine eats up the input alignment.
#
# INPUT: (\@align, $start_frame, \@seq_starts, $gap_char), where
#	${$align[$m]} is the $m th sequence in the alignment;
#	$start_frame is the coordinate of the alignment start
#	in the alignment coordinate frame;
#	$seq_starts[$m] is the coordinate of the residue of the $m th sequence that
#	first appear in the alignment examined;
#	$gap_char and $ng_char represents, respectively, a gap and a residue that is not a gap.
#
# OUTPUT: \@list_gap_blcs, where
#	@{$list_gap_blcs[$k]} = ($blstart_frame, $blend_frame, \@blstart_seq, \@blend_seq, \@pattern, \@sbalign), where
#	$blstart/end_frame is the starting/ending coordinate of the block in terms of the alignment frame;
#	$blstart/blend_seq[$m] = the coordinate of the residue of the $m th sequence
#			that starts/ends the block. More precisely, the former
#	gives the coordinate of the residue that is nearest on the right side to the left boundary,
#	and the latter is that which is nearest on the left side to the right boundary;
#	@pattern = (the sequence numbers to which the gap is assigned);
#	$sbalign[$m] = the contribution of the $m th sequence to the block.
#
sub extract_gap_blocks (\@$\@$) {

	my ($align, $start_frame, $seq_starts, $gap_char) = @_;

	my @list_gap_blcs = ();
	my $ct_seqs = @{$align};
	($ct_seqs>0) or (return []);

	my $alnlen = length ${$align->[0]};
	($alnlen>0) or (return []);

	unless (@{$seq_starts}>0) {
	    for (my $k=0; $k<$ct_seqs; $k++) { push @{$seq_starts}, 1; }
	}

	my @curr_resnum = ();
	foreach my $coord (@{$seq_starts}) { push @curr_resnum, $coord-1; }

	for (my $l=0; $l<$alnlen; $l++) { # The outer for-loop #

	    my $curr_frame = $start_frame + $l;
	    my @gapped_seqs = ();
	    my @residues = ();
	    ##my @blend_seq = ();
	    my @blstart_seq = ();
	    for (my $m=0; $m<$ct_seqs; $m++) { # the 1st inner for-loop #
		my $res = substr (${$align->[$m]}, 0, 1, '');
		push @residues, $res;
		##push @blend_seq, $curr_resnum[$m] + 1;
		push @blstart_seq, $curr_resnum[$m] + 1;
		($res eq $gap_char) ? (push @gapped_seqs, $m) : ($curr_resnum[$m]++) ;
	    }			# end of the 1st inner for-loop #

	    my $ct_gaps = @gapped_seqs;

		# No gap is contained in the current site #
	    ($ct_gaps>0) or next;

		# See if the upstream block is adjacent to the current site,
		# and if the current gap pattern matches the last one.
		# Otherwise, store the site as a new gap block.
		#
	    ##my @blstart_seq = @curr_resnum;
	    my @blend_seq = @curr_resnum;
	    my @new_block = ($curr_frame, $curr_frame, \@blstart_seq, \@blend_seq, \@gapped_seqs, \@residues);
	    (@list_gap_blcs>0) or ((push @list_gap_blcs, \@new_block) and next);

	    my $last_block = $list_gap_blcs[$#list_gap_blcs];
	    my @last_pattern = @{$last_block->[4]};
	    (($last_block->[1] == $curr_frame-1) and (@gapped_seqs == @last_pattern))
		 or ((push @list_gap_blcs, \@new_block) and next);
	    my $fl_match = 1;
	    for (my $k=0; $k<$ct_gaps; $k++) { # the 2nd inner for-loop #
		($gapped_seqs[$k] == $last_pattern[$k]) and next;
		$fl_match = 0;
		next;
	    }				# end of the 2nd inner for-loop #
	    $fl_match or ((push @list_gap_blcs, \@new_block) and next);

		# If the current gap pattern matches the last one,
		# append the current site to the last gap block.
	    $last_block->[1] = $curr_frame;
	    $last_block->[3] = \@blend_seq;
	    for (my $k=0; $k<$ct_seqs; $k++) { $last_block->[5][$k] .= $residues[$k]; }

	}			# end of the outer for-loop #

	return \@list_gap_blcs;

}		# end of the subroutine 'extract_gap_blocks' #


# union_indel_blocks:
#	This satellite subroutine of the subroutine 'extract_indel_blocks'
#	merges two sets of indel blocks by putting together
#	the blocks of the same starting and ending frame coordinates.
#
# INPUT: (\@blocks1, \@blocks2), where
#	@{$blocks1/2[$k]} = ($start, $end, \@seqids), with
#	$start and $end are starting and ending coordinates of the $k th indel
#	in terms of the alignment frame coordinate, and
#	@seqids is a list of IDs  of the sequences possessing the indel in question.
#	We ASSUME that indel blocks are pre-sorted according to the order
#	$start ('+') --> $end ('+').
#
# OUTPUT: \@union, where
#	@{$union[$k]} = ($start, $end, \@seqids),
#	which was constructed by merging the indels with the same $start and $end,
#	with @seqids being the union of the sequence IDs from both @blocks1 and @blocks2.
#
sub union_indel_blocks (\@\@) {

	my ($blocks1,$blocks2) = @_;
	my @union = ();
	my @indices = (0, 1); my @methods = ('+', '+');

	my @sarray1 = @{$blocks1};
	my @sarray2 = @{$blocks2};

	my $item1 = shift @sarray1;
	my $item2 = shift @sarray2;

	while (1) {	# the 'infinite' while loop #

		unless (defined $item1) {
			if (defined $item2) {
				push @union, $item2;
				@sarray2 and (push @union, @sarray2); # cram the remaining elements #
			}
			last;
		}

		unless (defined $item2) {
			if (defined $item1) {
				push @union, $item1;
				@sarray1 and (push @union, @sarray1); # cram the remaining elements #
			}
			last;
		}

		my $sgn = compare_arrays (@{$item1}, @{$item2}, @indices, @methods);

		##if ($item1 eq $item2) {	# if-elsif-else #
		if ($sgn == 0) {		# if-elsif-else #
			##push @union, $item1;
				# merge the lists of sequences #
			my @merged = @{$item1};
			$merged[2] = union_nmarray (@{$item1->[2]}, @{$item2->[2]});
			push  @union, \@merged;
			$item1 = shift @sarray1;
			$item2 = shift @sarray2;

		##} elsif ($item1 lt $item2) {
		} elsif ($sgn < 0) {
			push @union, $item1;
			$item1 = shift @sarray1;

		} else {
			push @union, $item2;
			$item2 = shift @sarray2;
		}		# end of the if_elsif_else #

	}			# end of the 'infinite' while loop #

	return \@union;

}		# end of the subroutine 'union_indel_blocks' #


# extract_indels_eqbd:
#	This subroutine extracts insertions/deletions
#	from all the sequences in a multiple alignment,
#	and puts together the indels with the same starting and ending coordinates
#	in terms of the alignment coordinate frame.
#	NOTE that this subroutine eats up the input alignment.
#
# INPUT: (\@align, $start_frame, \@seq_starts, $gap_char), where
#	${$align[$m]} is the $m th sequence in the alignment;
#	$start_frame is the coordinate of the alignment start
#	in the alignment coordinate frame;
#	$seq_starts[$m] is the coordinate of the residue of the $m th sequence that
#	first appear in the alignment examined;
#	$gap_char and $ng_char represents, respectively, a gap and a residue that is not a gap.
#
# OUTPUT: \@list_indels, where
#	@{$list_indels[$k]} = ($blstart_frame, $blend_frame, \@blstart_seq, \@blend_seq, \@pattern, \@sbalign), where
#	$blstart/end_frame is the starting/ending coordinate of the block in terms of the alignment frame;
#	$blstart/blend_seq[$m] = the coordinate of the residue of the $m th sequence
#			that starts/ends the block. More precisely, the former
#	gives the coordinate of the residue that is nearest on the right side to the left boundary,
#	and the latter is that which is nearest on the left side to the right boundary;
#	@pattern = (the sequence numbers to which the gap is assigned);
#	$sbalign[$m] = the contribution of the $m th sequence to the block.
#
sub extract_indels_eqbd (\@$\@$) {

	my ($align, $start_frame, $seq_starts, $gap_char) = @_;

	my $ct_seqs = @{$align};
	($ct_seqs>0) or (return []);

	my $alnlen = length ${$align->[0]};
	($alnlen>0) or (return []);

	unless (@{$seq_starts}>0) {
	    for (my $k=0; $k<$ct_seqs; $k++) { push @{$seq_starts}, 1; }
	}

=for debugging
	print STDERR "\n< Subroutine 'extract_indels_eqbd' >\n\n";
=cut

	my $indel_coords = []; # element = ($start, $end, [list_seqids]);

	for (my $m =0; $m<$ct_seqs; $m++) { # the 1st outer for-loop #

	    my $cp_seq = ${$align->[$m]};
	    my @seq_indel_coords = ();

	    for (my $l=0; $l<$alnlen; $l++) { # the inner for-loop #
		my $curr_frame = $start_frame + $l;
		my $res = substr ($cp_seq, 0, 1, '');
		($res eq $gap_char) or next;
		(@seq_indel_coords>0) or ((push @seq_indel_coords, [$curr_frame, $curr_frame, [$m]]) and next);
		my $last_indel = $seq_indel_coords[$#seq_indel_coords];
		($last_indel->[1] == $curr_frame -1) ?
		    ($last_indel->[1] = $curr_frame) :
		    (push @seq_indel_coords, [$curr_frame, $curr_frame, [$m]]) ;
	    }				# end of the inner for-loop # 

	    $indel_coords = union_indel_blocks (@{$indel_coords}, @seq_indel_coords);

	}				# end of the 1st outer for-loop #


		# Flesh out the information on the blocks of indels #

	my @list_indels = ();
	my $last_start = $start_frame;
	##my @last_starts = ();
	##foreach my $coord (@{$seq_starts}) { push @last_starts, $coord-1; }
	my @last_starts = @{$seq_starts};

	while (my $indel_info = shift @{$indel_coords}) { # the 2nd outermost while-loop #
	    my ($blstart, $blend, $pattern) = @{$indel_info};
	    my $len_up_margin = $blstart - $last_start;

	    if ($len_up_margin>0) {

=for debugging
		print STDERR " --- To extract the interval \[$blstart,$blend\], I have to cut of the upstream region of length $len_up_margin. ---\n";
=cut
		for (my $m=0; $m<$ct_seqs; $m++) { # the 1st middle for-loop #
		    my $ct_gaps = 0;
		    for (my $l=0; $l<$len_up_margin; $l++) { # the inner for-loop #
			my $res = substr (${$align->[$m]}, 0, 1, '');
			($res eq $gap_char) and ($ct_gaps++);
		    }				# end of the inner for-loop #

		    $last_starts[$m] += $len_up_margin - $ct_gaps;
		}			# end of the 1st middle for-loop #

=for debugging
		print STDERR "\n{ The resulting alignment }\n\n";
		for (my $m=0; $m<$ct_seqs; $m++) {
		    printf STDERR "Sequence_%-2d\t%s\n", $m, ${$align->[$m]};
		}
		print STDERR "\n";
=cut

	    }

	    my @subalign = ();
	    my @blstart_seq = my @blend_seq = @last_starts;
	    my $len_region = $blend - $blstart + 1;

	    for (my $m=0; $m<$ct_seqs; $m++) { # the 2nd middle for-loop #
		##($gap_char eq substr (${$align->[$m]}, 0, 1) ) or ($blstart_seq[$m]++);
		my $subseq = substr (${$align->[$m]}, 0, $len_region);
		my $copy = $subseq;
		my $ct_gaps = 0;
		for (my $l=0; $l<$len_region; $l++) { # the inner for-loop #
		    ($gap_char eq substr ($copy, 0, 1, '')) and ($ct_gaps++);
		}				# end of the inner for-loop #
		##$blend_seq[$m] += $len_region - $ct_gaps;
		$blend_seq[$m] += $len_region - $ct_gaps -1;

		push @subalign, $subseq;
	    }				# end of the 2nd middle for-loop #

	    push @list_indels, [$blstart, $blend, \@blstart_seq, \@blend_seq, $pattern, \@subalign];

	    $last_start = $blstart;

	}				# end of the 2nd outermost while-loop #
	
	return \@list_indels;

}		# end of the subroutine 'extract_indels_eqbd' #



# pad_concate_blocks:
#	This subroutine receives a string made from concatenated blocks
#	of a sequence, and reconstructs a 'masked' sequence
#	by padding missing regions with series of a specified character.
#
#	NOTE that this subroutine chews up the input string, so you should
#	copy the input beforehaand if you want to use it later on.
#
# INPUT: (\$concate_blocks, \@block_coords, $beg_frame, $end_frame, $pad), where
#	$concate_blocks is a string made by concatenating blocks;
#	@{$block_coords[$k]} = ($beg, $end) gives the boundaries of the $k th block;
#	$beg/end_frame gives the beginning and the end of the coordinate frame
#	spanned by the sequence (or the alignment);
#	$pad is the charecter used to pad the input string.
#
# OUTPUT: \$padded_seq, where
#	$padded_seq is the sequence made by padding the input with a series of
#	a specified character.
#
sub pad_concate_blocks (\$\@$$$) {

	my ($concate_blocks, $block_coords, $beg_frame, $end_frame, $pad) = @_;

	my $totlen_blks = length ${$concate_blocks};
	my $ct_blocks = @{$block_coords};

	if ($beg_frame > $end_frame) {
	    print STDERR "[ERROR 'pad_concate_blocks'] ILLEGAL INPUT: \(\$beg_frame\, \$end_frame\) = ($beg_frame, $end_frame) .\n";
	    return undef;
	} 
	if (($totlen_blks ==0) and ($ct_blocks ==0)) {
	    my $padded_seq = $pad x ($end_frame - $beg_frame + 1);
	    return $padded_seq;
	} elsif (($totlen_blks ==0) or ($ct_blocks ==0)) {
	    print STDERR "[ERROR 'pad_concate_blocks'] ILLEGAL INPUT: \(\$totlen_blks\, \$ct_blocks\) = ($totlen_blks, $ct_blocks) .\n";
	    return undef;
	}

	my $beg_coord = $block_coords->[0][0];
	my $end_coord = $block_coords->[$ct_blocks-1][1];
	if ($end_coord > $end_frame) {
	    print STDERR "[ERROR 'pad_concate_blocks'] ILLEGAL INPUT: The end of the last block ($end_coord) is DOWNSTREAM of the frame ending ($end_frame) \!\n";
	}

	my $padded_seq = '';
	my $pres_coord = $beg_frame;

	my $totlen_blks2 =0;
	for (my $k=0; $k<$ct_blocks; $k++) { # The main for-loop #

	   my ($beg_blk, $end_blk) = @{$block_coords->[$k]};

	   if ($pres_coord < $beg_blk) {
		$padded_seq .= $pad x ($beg_blk - $pres_coord);
	   } elsif ($pres_coord > $beg_blk) {
		print STDERR "[ERROR 'pad_concate_blocks'] ILLEGAL INPUT: The beginning of the ", $k+1, " th block ($beg_blk) is UPSTREAM of the present coordinate ($pres_coord) !\n";
		return undef;
	   }

	    my $len_blk = $end_blk - $beg_blk + 1;
	   unless ($len_blk>0) {
		print STDERR "[ERROR 'pad_concate_blocks'] ILLEGAL INPUT: The beginning of the ", $k +1, " th block ($beg_blk) is DOWNSTREAM of its end ($end_blk) !\n";
		return undef;
	    }

	    $padded_seq .= substr (${$concate_blocks}, 0, $len_blk, '');
	    $totlen_blks2 += $len_blk;

	    $pres_coord = $end_blk + 1;

	}			# End of the main for-loop #

	unless ($totlen_blks == $totlen_blks2) {
	    print STDERR "[ERROR 'pad_concate_blocks'] ILLEGAL INPUT: The actual length of the concatenated blocks ($totlen_blks) does NOT agree with its calculated length ($totlen_blks2) !\n";
	    return undef;
	}

	($pres_coord < $end_frame + 1) and ($padded_seq .= $pad x ($end_frame + 1 - $pres_coord)); 

	return \$padded_seq;

}		# end of the subroutine 'pad_concate_blocks' #


# gap_indicator:
#	This subroutine makes a string that indicates the positions of gaps
#	out of a set of intervals.
#
# INPUT: (\@gap_intvls, $beg_frame, $end_frame, $gap, $non_gap), where
#	@{$gap_intvls[$m]} = ($start, $end) is the $m th intervals occupied by gaps,
#	NOTE that the intervals do NOT overlap each other;
#	$beg/end_frame is the coordinate of the beginning/ending of the alignment frame;
#	$gap and $non_gap represent that a site is occupied by a gap and a residue, respectively.
#
# OUTPUT: \$gap_indicator, where
#	$gap_indicator is a string that indicates the positions of gaps.
#
sub gap_indicator (\@$$$$) {

	my ($gap_intvls, $beg_frame, $end_frame, $gap, $non_gap) = @_;

	my $ct_intvls = @{$gap_intvls};
	
	unless ($ct_intvls>0) { # NO gaps #
	    my $indicator = $non_gap x ($end_frame - $beg_frame + 1);
	    return \$indicator;
	}
	my $end_coord = $gap_intvls->[$#{$gap_intvls}][1];
	if ($end_coord > $end_frame) {
	    print STDERR "[ERROR 'gap_indicator'] ILLEGAL INPUT: the coordinate of the end of the blocks ($end_coord) exceeds the end of the coordinate frame ($end_frame) !\n";
	    return undef;
	}

	my $indicator = '';
	my $pres_coord = $beg_frame;

	foreach my $intvl (@{$gap_intvls}) { # The main foreach-loop #
	    my ($gap_beg, $gap_end) = @{$intvl};

	    if ($pres_coord < $gap_beg) { # Region not containing gaps #
		$indicator .= $non_gap x ($gap_beg - $pres_coord);
	    } elsif ($pres_coord > $gap_beg) {
		print STDERR "[ERROR 'gap_indicator'] The present coordinate ($pres_coord) exceeds the beginning coordinate of the gap interval \[$gap_beg\, $gap_end\] !\n";
		return undef;
	    }

		# Region consisting of gaps #
	    $indicator .= $gap x ($gap_end - $gap_beg + 1);
	    $pres_coord = $gap_end + 1;

	}			# End of the main foreach-loop #

	($pres_coord < $end_frame + 1) and ($indicator .= $non_gap x ($end_frame + 1 - $pres_coord)); 

	return \$indicator;

}		# End of the subroutine 'gap_indicator' #


# splice_blocks_in_aln:
#	This subroutine concatenates specified sub-regions (= 'block's) of an alignment.
#	NOTE that this subroutine eats up the input alignment.
#
# INPUT: (\@seqs, \@blocks, $beg_frame), where
#	${$seqs[$k]} is the $k th sequence;
#	@{$blocks[$m]} = ($start, $end) is the $m th block to be concatenated;
#	$beg_frame is the coordinate of the beginning of the alignment frame.
#
# OUTPUT: (\@spliced, \@rm_regions), where
#	${$spliced[$k]} is the concatenated regions of the $k th sequence;
#	@{$rm_regions[$n]} stores the sub-alignment of the $n th region that has been removed.
#
sub splice_blocks_in_aln (\@\@$) {

	my ($seqs, $blocks, $beg_frame) = @_;

	my $ct_seqs = @{$seqs};
	($ct_seqs>0) or (return ([], []));
	my $ct_blcs = @{$blocks};
	($ct_blcs>0) or (return ([], [$seqs]));

	my $end_frame = $beg_frame + length (${$seqs->[0]}) -1;
	my $end_coord = $blocks->[$#{$blocks}][1];
	if ( $end_coord > $end_frame) {
	    print STDERR "[ERROR 'splice_blocks_in_aln'] ILLEGAL INPUT: the coordinate of the end of the blocks ($end_coord) exceeds the end of the coordinate frame ($end_frame) !\n";
	    return (undef, undef);
	}

	my @spliced = my @rm_regions = ();
	for (1 .. $ct_seqs) { my $init = ''; push @spliced, \$init; }

	my $pres_coord = $beg_frame;

	foreach my $region (@{$blocks}) { # The main foreach-loop #
	    my ($beg, $end) = @{$region};

	    if ($pres_coord < $beg) {
		my @removed = ();
		foreach my $seq (@{$seqs}) { # The 1st inner foreach-loop #
		    my $upstream_margin = substr (${$seq}, 0, $beg - $pres_coord, '');
		    push @removed, \$upstream_margin;
		}		# End of the 1st inner foreach-loop #
		push @rm_regions, \@removed;
	    } elsif ($pres_coord > $beg) {
		print STDERR "[ERROR 'splice_blocks_in_aln'] THe present coordinate ($pres_coord) exceeds the beginning coordinate of the region \[$beg\, $end\] !\n";
		return (undef, undef);
	    }

	    my $sblen = $end - $beg + 1;
	    for (my $k=0; $k<$ct_seqs; $k++) { # The 2nd inner foreach-loop #
		my $seq  = $seqs->[$k];
		my $sbseq = substr (${$seq}, 0, $sblen, '');
		${$spliced[$k]} .= $sbseq;
	    }			# End of the 2nd inner foreach-loop #

	    $pres_coord = $end + 1;
	}			# End of the main foreach-loop #

	($pres_coord < $end_frame +1) and (push @rm_regions, $seqs);

	return (\@spliced, \@rm_regions);

}		# End of the subroutine 'splice_blocks_in_aln' # 


# pick_sbset_aln;
#	This subroutine receives a multiple alignment and 
#	picks out a subset.
#
# INPUT: (\@in_seqs, \@in_seqids, \@rlv_seqinfo, $indx_in_seqid, $indx_out_seqid),
#	where
#	${$in_seqs[$k]} = the $k th sequence;
#	$in_seqids[$k] = the ID of the $k th sequence;
#	($rlv_in_seqid, $rlv_out_seqid)
#	   = @{$rlv_seqinfo[$k]}[$indx_in_seqid, $indx_out_seqid] for the $k th sequence.
#
# OUTPUT: (\@rlv_seqs, \@rlv_seqids), where
#	@rlv_seqs and @relv_seqids store the relevant sequences and their IDs,
#	respectively, in the same order that they appear in @rlv_seqinfo.
#	NOTE that @rlv_seqs stores COPIES of the relevant sequences.
#
sub pick_sbset_aln (\@\@\@$$) {

	my ($in_seqs, $in_seqids, $rlv_seqinfo, $indx_in_id, $indx_out_id) = @_;

	my $ct_seqs = @{$in_seqs};

	my %seqid2indx; 
	for (my $k =0; $k<$ct_seqs; $k++) { $seqid2indx{$in_seqids->[$k]} = $k; }

	my @rlv_seqs = my @rlv_seqids = ();

	foreach my $seqinfo (@{$rlv_seqinfo}) { # The main foreach-loop #

	    my ($in_seqid, $out_seqid) = @{$seqinfo}[$indx_in_id, $indx_out_id];
	    my $indx = $seqid2indx{$in_seqid};
	    unless (defined $indx) {
		print STDERR "[ERROR 'pick_sbset_aln'] The sequence '$in_seqid' is NOT FOUND in the alignment !\n";
		return (undef, undef);
	    }
	    my $cp_seq = ${$in_seqs->[$indx]};
	    push @rlv_seqs, \$cp_seq;
	    push @rlv_seqids, $in_seqids->[$indx];

	}			# End of the main foreach-loop #

	return (\@rlv_seqs, \@rlv_seqids);

}		# End of the subroutine 'pick_sbset_aln' #


# examine_leftmost_site:
#	This satellite subroutine of 'chop_aln_by_qubds'
#	examines the leftmost site by chopping it of the alignment.
#
# INPUT: (\@src_aln, \@dsn_aln, \@curr_coords, \$curr_frame, $ct_seqs, $char_gap, $fl_append), where
#	${$src_aln[$k]} and ${$dsn_aln[$k]} are the $k th sequence
#	in the alignment to be examined and that to store the resulting site, respectively;
#	$curr_coords[$k] is the coordinate of the $k th sequence in the left-neighboring site;
#	$curr_frame is the coordinate of the alignment frame in the left-neighboring site; 
#	$fl_append = 1 if you want to append the extracted site to @dsn_aln,
#		   = 0 if you want to (re)build @dsn_aln from scratch.
#
# OUTPUT: 1 if the process succeeded,
#	  0 if it failed.
#
sub examine_leftmost_site (\@\@\@\$$$$) {

	my ($src_aln, $dsn_aln, $curr_coords, $curr_frame,
	$ct_seqs, $char_gap, $fl_append) = @_;

	for (my $k=0; $k<$ct_seqs; $k++) { # The 1st inner for-loop #
	    my $char = substr (${$src_aln->[$k]}, 0, 1, '');
	    $char or (return 0);
	    ($char eq $char_gap) or ($curr_coords->[$k]++);
	    $fl_append ? (${$dsn_aln->[$k]} .= $char) : ($dsn_aln->[$k] = \$char);
	}				# End of the 1st inner for-loop #
	${$curr_frame}++;

	return 1;

}		# End of the subroutine 'examine_leftmost_site' #


# chop_aln_by_qubds:
#	This subroutine chops alignments using the specified
#	query coordinates as cut-off points.
#
# INPUT: (\@align, $frame_start, \@begs, \@ends, $char_gap, \@boundaries), where
#	${$align[$k]} = the $k th sequence embedded in the input alignment;
#	$begs[$k] and $ends[$k] are the start and end of the alignment in terms of the $k th sequence coordinate;
#	$boundaries[$m] is the $m th boundary to be used as a cutting point,
#	including the both ends (IN THE HALF-INTEGER REPRESENTATION).
#
# OUTPUT:  (\@chop_aln, \@gaps), where
#	@{$chop_aln[$m]} = ($frame_beg, $frame_end, \@reg_begs, \@rg_ends, \@rg_aln)
#	for the region betweem the $m th and $m +1 th boundaries;
#	@{$gaps[$m]} = ($frame_beg, $frame_end, \@rg_begs, \@rg_ends, \@rg_aln)
#	is for the $m th gap that coincides with the $m th boundary of the query.
#
sub chop_aln_by_qubds (\@$\@\@$\@) {

	my ($align, $frame_start, $begs, $ends, $char_gap, $bds) = @_;

	my $ct_seqs = @{$align};
	($ct_seqs>0) or (return ([], []));
	my $ct_bds = @{$bds};
	unless ($ct_bds>1) {
	    print STDERR "[ERROR 'chop_aln_by_qubds'] ILLEGAL INPUT: Only $ct_bds boundaries are provided.\n";
	    return (undef, undef);
	}
	my ($leftmost, $rightmost) = @{$bds}[0,$#{$bds}];
	my $qu_beg = ($begs->[0] == int ($begs->[0]) ) ?
		$begs->[0] - 1/2 : $begs->[0] ;
	my $qu_end = ($ends->[0] == int ($ends->[0]) ) ?
		$ends->[0] + 1/2 : $ends->[0] ;

	if ($qu_beg < $leftmost) {
	    print STDERR "[ERROR 'chop_aln_by_qubds'] ILLEGAL INPUT: The query start ($qu_beg) COMES TO THE LEFT of the leftmost boundary ($leftmost)!\n";
	    return (undef, undef);
	}
	if ($qu_end > $rightmost) {
	    print STDERR "[ERROR 'chop_aln_by_qubds'] ILLEGAL INPUT: The query end ($ends->[0]) COMES TO THE RIGHT of the rightmost boundary ($rightmost)!\n";
	    return (undef, undef);
	}

	my $qu_len = 0;
	my $cp_quseq = ${$align->[0]};
	while (my $char = substr ($cp_quseq, 0, 1, '') ) { ($char eq $char_gap) or ($qu_len++); }
	unless ($qu_end == $qu_beg + $qu_len) {
	    print STDERR "[ERROR 'chop_aln_by_qubds'] ILLEGAL INPUT: The query length $qu_len does NOT EQUAL the length of the interval (qu_beg\,qu_end) = ($qu_beg, $qu_end) !\n"; 
	    return (undef, undef);
	}

	
		# Initialize the output data #
	my @chop_aln = my @gaps = (); 
	my $curr_frame = $frame_start -1;
	my $end_frame = $frame_start -1 + length ${$align->[0]} ;
	my @curr_coords = ();
	foreach my $beg (@{$begs}) { # The 1st foreach-loop #
	    push @curr_coords, ($beg == int ($beg)) ? $beg - 1 : int ($beg);
	}			# End of the 1st foreach-loop #

	for (my $r=1; $r<$ct_bds; $r++) { # The 2nd outer for-loop #
	    my ($lbd, $rbd) = @{$bds}[$r-1, $r];
	    if (($rbd <= $qu_beg) or ($qu_end < $lbd) or ($curr_frame >= $end_frame)) {
		push @chop_aln, [];
		push @gaps, [];
		next;
	    }

		# examine the first characters #
	    my @reg_alns = ();
	    examine_leftmost_site (@{$align}, @reg_alns, @curr_coords, $curr_frame, $ct_seqs, $char_gap, 0); # Initialize @reg_alns

		# Check whether the left-neighboring gap exits or not #
	    if ($curr_coords[0] < $lbd) {
		my @gp_starts = @curr_coords;
		for (my $k=0; $k<$ct_seqs; $k++) { (${$reg_alns[$k]} eq $char_gap) and ($gp_starts[$k]++); }
		my $fstart_gp = $curr_frame;

		while ((0< length ${$align->[0]})
		   and ($char_gap eq substr (${$align->[0]}, 0, 1) )) { # The 1st inner while-loop #
		   examine_leftmost_site (@{$align}, @reg_alns, @curr_coords, $curr_frame, $ct_seqs, $char_gap, 1); # Append the site to @reg_alns
		}			# End of the 1st inner while-loop #
		my $fend_gp = $curr_frame;
		my @gp_ends = @curr_coords;
		my @gp_alns = @reg_alns;
		push @gaps, [$fstart_gp, $fend_gp, \@gp_starts, \@gp_ends, \@gp_alns];
		unless (0< length ${$align->[0]}) { # No sequence left
		  push @chop_aln, [];
		  next;
		}
	        examine_leftmost_site (@{$align}, @reg_alns, @curr_coords, $curr_frame, $ct_seqs, $char_gap, 0); # Re-initialize @reg_alns
	    } else {
		push @gaps, []; 
	    }

	    $rbd = int $rbd;

	    my @rg_starts = @curr_coords;
	    for (my $k=0; $k<$ct_seqs; $k++) { (${$reg_alns[$k]} eq $char_gap) and ($rg_starts[$k]++); }
	    my $fstart_rg = $curr_frame;

			## Bug fixed on Tuesday, December 5, 2006
	    ##while (0< length ${$align->[0]}) { # The 2nd inner while-loop #
	    while ((0< length ${$align->[0]}) and ($curr_coords[0] < $rbd)) { # The 2nd inner while-loop #
		examine_leftmost_site (@{$align}, @reg_alns, @curr_coords, $curr_frame, $ct_seqs, $char_gap, 1); # Append the site to @reg_alns
		##($curr_coords[0] == $rbd) and last;
	    }				# end of the 2nd inner while-loop #
	    my $fend_rg = $curr_frame;
	    my @rg_ends = @curr_coords;
	    push @chop_aln, [$fstart_rg, $fend_rg, \@rg_starts, \@rg_ends, \@reg_alns];
	}		# End of the 2nd outer for-loop #  

	if (0< length ${$align->[0]}) {
	    my @end_gap = ();
	    examine_leftmost_site (@{$align}, @end_gap, @curr_coords, $curr_frame, $ct_seqs, $char_gap, 0); # Initialize @end_gap
	    my @gp_starts = @curr_coords;
	    for (my $k=0; $k<$ct_seqs; $k++) { (${$end_gap[$k]} eq $char_gap) and ($gp_starts[$k]++); }
	    my $fstart_gp = $curr_frame;

	    while ((0 < length ${$align->[0]}) and
	    ($char_gap eq substr (${$align->[0]}, 0, 1) )) { # The 3rd outer while-loop #
		examine_leftmost_site (@{$align}, @end_gap, @curr_coords, $curr_frame, $ct_seqs, $char_gap, 1); # Append the site to @end_gap
	    }
	    my $fend_gp = $curr_frame;
	    my @gp_ends = @curr_coords;
	    push @gaps, [$fstart_gp, $fend_gp, \@gp_starts, \@gp_ends, \@end_gap];
	} else {
	    push @gaps, [];
	}

	if (0 < length ${$align->[0]}) {
	    print STDERR "[ERROR 'chop_aln_by_qubds'] The alignment appears to have A SUPERFLUOUS PORTION in the downstream of (qu_coord = $qu_end):\n\n";
	    foreach my $seq (@{$align}) { print STDERR '"', ${$seq}, "\"\n"; }
	    print STDERR "\n";
	    return (undef, undef);
	}

	return (\@chop_aln, \@gaps);

}		# End of the subroutine 'chop_aln_by_qubds' #


# gaps_in_sglseq:
#	This subroutine examines gaps in a single sequence.
#	NOTE that this subroutine eats up the input sequence.
#
# INPUT: (\$seq, $seq_beg, $seq_end, $frame_beg, $char_gap), where
#	$seq is the sequence to be examined;
#	$seq_beg/end is the sequence coordinate for the sequence start/end.
#
# OUTPUT: (\@list_gaps, \@list_ng_frags), where
#	@{$list_gaps[$r]} = ($frame_gp_beg, $frame_gp_end, $seq_gap_coord, $gap_len)
#		for the $r th leftmost gap;
#	@{$list_ng_frags[$s]} = ($frame_fg_beg, $frame_fg_end, $seq_fg_beg, $seq_fg_end, \$frag)
#		for the $s th leftmost fragment.
#	NOTE that most of the coordinates are in the INTEGER representation,
#	except $seq_gap_coord, which is in the HALF-INTEGER  representation.
#
sub gaps_in_sglseq (\$$$$$) {

	my ($seq, $seq_beg, $seq_end, $frame_beg, $char_gap) = @_;

	(length (${$seq}) > 0) or (return ([], []));

	my $curr_frame = $frame_beg -1;
	my $curr_coord = ($seq_beg == int $seq_beg) ? $seq_beg - 1 : int $seq_beg ; 

	my @list_gaps = my @list_ng_frags = ();

	my $fl_last_gap = -1; # =1 if the last site was the gap, =0 no gap, = -1 unknown

=for TESTING the subroutine
	print STDERR "\n<< Subroutine 'gaps_in_sglseq' >>\n\n";
	print STDERR "< Input Sequence (beg=$seq_beg, end=$seq_end, frame_beg = $frame_beg) >\n\n";
	print_sequence (${$seq}, 50, *STDERR);
	print STDERR "\n";
=cut

	while (my $char = substr (${$seq}, 0, 1, '') ) { # The main while-loop #

	    $curr_frame++;

	    if ($char eq $char_gap) { # The site is occupied by a gap #
		if ($fl_last_gap == 1) { # The last site was also a gap #
		    $curr_coord  = int ($curr_coord) + 1/2;
		    $list_gaps[$#list_gaps]->[3]++; # elongate the gap #
		} else {
				# finish the last fragment, if any #
		    (@list_ng_frags>0) and 
		      ( @{$list_ng_frags[$#list_ng_frags]}[1,3]
			= ($curr_frame-1, $curr_coord) );

		    $curr_coord  = int ($curr_coord) + 1/2;

				# set a new gap #
		    push @list_gaps, [$curr_frame, undef, $curr_coord, 1];
		    $fl_last_gap = 1;
		}
	    } else { # The site is filled with a residue
		if ($fl_last_gap == 0) { # The last site was also a residue #
		    $curr_coord = int ($curr_coord) + 1;
		    ${$list_ng_frags[$#list_ng_frags]->[4]} .= $char; # elongate the fragment
		} else {
		    		# finish the last gap, if any #
		    (@list_gaps>0) and ( $list_gaps[$#list_gaps]->[1] = $curr_frame -1 );

		    $curr_coord = int ($curr_coord) + 1;

				# set a new fragment #
		    push @list_ng_frags, [$curr_frame, undef, $curr_coord, undef, \$char];
		    $fl_last_gap = 0;
		}
	    }
	}				# End of the main while-loop #


		## The Finishing Touches ##

	if ($fl_last_gap == 1) { #  The last site is occupied with a gap #
		# finish the last gap #
	    $list_gaps[$#list_gaps]->[1] = $curr_frame;
	} elsif ($fl_last_gap == 0) { # The last site is occupied with a residue #
		# finish the last fragment #
	    @{$list_ng_frags[$#list_ng_frags]}[1,3] = ($curr_frame, $curr_coord);
	} else {
	    print STDERR "[ERROR 'gaps_in_sglseq'] Found an ILLEGAL FIANL FLAG: '$fl_last_gap' !\n\n";
	    return (undef, undef);
	}

=for debugging the script
	print STDERR "\n< Output >\n\n";
	print STDERR "{ List of Gaps }\n";
	print STDERR join ("\t", 'Frame_beg', 'Frame_end', 'Seq_coord', 'Length'), "\n";
	foreach my $info_gap (@list_gaps) { print STDERR join ("\t", @{$info_gap}), "\n"; }
	print STDERR "\n\n{ List of Fragments }\n";
	print STDERR join ("\t", 'Frame_beg', 'Frame_end', 'Seq_beg', 'Seq_end'), "\n";
	foreach my $info_frag (@list_ng_frags) {
	    print STDERR join ("\t", @{$info_frag}[0..3]), "\n";
	    print_sequence (${$info_frag->[4]}, 50, *STDERR);
	    print STDERR "\n";
	}
	
#	@{$list_gaps[$r]} = ($frame_gp_beg, $frame_gp_end, $seq_gap_coord, $gap_len)
#		for the $r th leftmost gap;
#	@{$list_ng_frags[$s]} = ($frame_fg_beg, $frame_fg_end, $seq_fg_beg, $seq_fg_end, \$frag)
#		for the $s th leftmost fragment.
=cut

	return (\@list_gaps, \@list_ng_frags); 

}		# end of the subroutine 'gaps_in_sglseq' #


# maxmin_seqlen:
#	This subroutine receives a set of sequences,
#	and returns the maximum and minimun lengths in the set.
#
# INPUT: (\@set_seqs, $indx_len), where
#	the $k th sequence is ${$set_seqs[$k]->[$indx_len]} for $indx_len >= 0,
#	and ${$set_seqs[$k]}  for $indx_len < 0.
#
# OUTPUT: ($maxlen, $minlen, $indx_max, $indx_min), where
#	$indx_max/min is the index pointing the sequence with maximum/minimum length.
#	 It returns (-1,-1,-1,-1) if the set is empty.
#
sub maxmin_seqlen (\@$) {

	my ($set_seqs, $indx_len) = @_;

	my $ct_seqs = @{$set_seqs};
	($ct_seqs>0) or (return (-1, -1, -1, -1));

	my $indx_max = my $indx_min = 0;
	my $maxlen = my $minlen = ($indx_len<0) ?
		length (${$set_seqs->[0]}) : length (${$set_seqs->[0][$indx_len]}) ;

	for (my $k=1; $k<$ct_seqs; $k++) { # The main for-loop #
	    my $rf_seq = ($indx_len<0) ? $set_seqs->[$k] : $set_seqs->[$k][$indx_len]; 
	    my $seqlen = length ${$rf_seq};
	    ($maxlen < $seqlen) and (($maxlen, $indx_max) = ($seqlen, $k));
	    ($minlen > $seqlen) and (($minlen, $indx_min) = ($seqlen, $k));
	}			# # end of the main for-loop #

	return ($maxlen, $minlen, $indx_max, $indx_min);

}		# end of the subroutine 'maxmin_seqlen' #


# examine_sel_alnd_regions:
#	This subroutine examines a multiple alignment,
#	and finds regions of consecutive sites not containing
#	either gaps or specified characters.
#	NOTE that this subroutine eats up the alignment while processing it.
#
# 	REVISED on Wednasday, March 21, 2007:
#	The subroutine now returns a list of dismissed regions as well.
#
# INPUT: (\@align, $beg_frame, \@begs, \@ends, $char_gap, \%dismissed), where
#	${$align[$m]}  is the $m th sequence embedded in the alignment;
#	$begs[$m] and $ends[$m] are the beginning and ending residue numbers
#	of the $m th sequence in the alignment;
#	%dismissed = ($char => some defined value, ...) specifies characters
#	to be dismissed.
#
# OUTPUT: (\@sbaligns, \@frames, \@sbaln_coords, \@dismissed), where
#	${$sbaligns[$r]->[$m]} is the $m th sequence in the $r th subalignment
#	consisting only of sites without gaps or specified characters;
#	@{$frames[$r]} = ($beg_frame, $end_frame) are the frame coordinates
#	of the beginning and ending of the $r th subalignment;
#	@{$sbaln_coords[$r]->[$m]} = ($beg, $end) are the beginning and ending
#	coordinates of the $r th subalignment interms of the $m th sequence;
#	${$dismissed[$s]->[$m]} is the $m th sequence embedded in the dismissed region
#	which is flanked by the $s -1 th and $s th subalignments kept in @sbaligns.
#	Note that the -1 th and the scalar (@sbaligns) lsubalignments mean
#	the Upstream and the Downstream flanking regions, respectively.
#
sub examine_sel_alnd_regions (\@$\@\@$\%) {

	my ($align, $beg_frame, $begs, $ends, $char_gap, $dismissed) = @_;

	my $ct_seqs = @{$align};
	($ct_seqs>0) or (return ([], [], []) );

		## Initialize ##

	my @dismissed = ([]);
	my @curr_coords = ();
	for (my $k=0; $k<$ct_seqs; $k++) { # The 1st outer for-loop #
	    my $beg = $begs->[$k];
	    ($beg == int ($beg)) or ($beg = $begs->[$k] = int ($beg) + 1);
	    $ends->[$k] = int ($ends->[$k]);
	    push @curr_coords, $beg - 1;
	    my $null_str = '';
	    push @{$dismissed[0]}, \$null_str;
	}				# End of the 1st outer for-loop #

	my $aln_len = length ( ${$align->[0]} );

	my @sbaligns = my @frames = my @sbaln_coords = ();

		## Process the main bulk part ##

	my $fl_relv = 0;
	for (my $l=0; $l<$aln_len; $l++) { # The 2nd outer for-loop #
	    my $ct_irrelv = 0;
	    my @site = ();
	    my @last_coords = @curr_coords;
	    for (my $k=0; $k<$ct_seqs; $k++) { # The 1st inner for-loop #
		my $char = substr (${$align->[$k]}, 0, 1, '');
		push @site, $char;
		if ($char eq $char_gap) {
		   $ct_irrelv++;
		} else {
		    $curr_coords[$k]++;
		    (defined $dismissed->{$char}) and ($ct_irrelv++);
		}
	    }				# End of the 1st inner for-loop #

	    if ($ct_irrelv>0) { # The site is to be dismissed #
		if ($fl_relv) { # A relevant region ended at the last site #
		    $fl_relv = 0;
		    my @new_dismissed = ();
		    my $last_sbaln_coords = $sbaln_coords[$#sbaln_coords];
		    for (my $k=0; $k<$ct_seqs; $k++) { # The 2nd inner for-loop #
			push @{$last_sbaln_coords->[$k]}, $last_coords[$k];
			push @new_dismissed, \$site[$k];
		    }			# End of the 2nd inner for-loop #
		    push @dismissed, \@new_dismissed;

		    my $last_frame = $frames[$#frames];
		    push @{$last_frame}, $beg_frame + $l -1;
		} else {  # The dismissed region continues #
		    my $last_dismissed = $dismissed[$#dismissed];
		    for (my $k=0; $k<$ct_seqs; $k++) { # The 3rd inner for-loop #
			${$last_dismissed->[$k]} .= $site[$k];
		    }			# End of the  3rd inner for-loop #
		}
	    } else {
		if ($fl_relv) { # The relevant region continues #
		    my $last_sbalign = $sbaligns[$#sbaligns];
		    for (my $k=0; $k<$ct_seqs; $k++) { # The 4th inner for-loop #
			${$last_sbalign->[$k]} .= $site[$k];
		    }				# End of the 4th inner for-loop #
		} else { # A relevant region begins at this site #
		   $fl_relv = 1;
		   my @new_sbaln = my @new_sbaln_coords = ();
		   for (my $k=0; $k<$ct_seqs; $k++) { # The 5th inner for-loop #
			push @new_sbaln, \$site[$k];
			push @new_sbaln_coords, [$curr_coords[$k]];
		   }			# End of the 5th inner for-loop #
		   push @sbaligns, \@new_sbaln;
		   push @sbaln_coords, \@new_sbaln_coords;
		   push @frames, [$beg_frame + $l]; 
		}
	    }

	}				# End of the 2nd outer for-loop #

		## Finish ##

	if ($fl_relv) { # A relevant region continues to the end of the alignment #
	   my @new_dismissed = ();
	   my $last_sbaln_coords = $sbaln_coords[$#sbaln_coords];
	   for (my $k=0; $k<$ct_seqs; $k++) { # The 3rd outer for-loop #
		push @{$last_sbaln_coords->[$k]}, $curr_coords[$k];
		my $null_str = '';
		push @new_dismissed, \$null_str;
	   }				# End of the 3rd outer for-loop #
	   my $last_frame = $frames[$#frames];
	   push @{$last_frame}, $beg_frame + $aln_len - 1;
	   push @dismissed, \@new_dismissed;
	}

	return (\@sbaligns, \@frames, \@sbaln_coords, \@dismissed);

}		# End of the subroutine 'examine_sel_alnd_regions' #


# bundle_refine_maln_by_clw:
#	This subroutine receives a multiple alignment,
#	then conducts a series of processes,
#	from constructing a guidetree to refining the alignment.
#	NOTE that this subroutine eats up the input alignment.
#
# INPUT: (\@in_aln, \$bin_clw, \$option_gdtree, \$option_refaln, \$tmp_dir,
#		$char_gap, $char_mask, $minct_tossgap), where
#	${$in_aln[$k]} is the $k th sequence embedded in the alignment;
#	$bin_clw is the path to a binary file of 'clustalw';
#	$option_gdtree is the option to be used for the guidetree construction,
#	and $option_refaln is the option to be used for the refinement of the alignment.
#	Using the options, we can specify the type of sequences, for example;
#	$tmp_dir is the path to the temporary working directory.
#
# OUTPUT: (\@out_aln, \$log_gdtree, \$log_refaln), where
#	${$out_aln[$k]} is the $k th sequence embedded in the refined alignment,
#	with the same sequence order as the input;
#	$log_gdtree and $log_refaln are the paths to the log files of the
#	guide tree construction and the alignment refinement, respectively,
#	if the process failed. They are undef if the processes succeeded.
#
sub bundle_refine_maln_by_clw (\@\$\$\$\$$$$) {

	my ($in_aln, $bin_clw, $opt_gdtree, $opt_refaln, $tmp_dir,
		$char_gap, $char_mask, $minct_tossgap) = @_;

	$opt_gdtree = ${$opt_gdtree};
	$opt_refaln = ${$opt_refaln};

		# Keep the same order as the input alignment #
	$opt_refaln =~ s/\-outorder\=\S+\s*//i;
	$opt_refaln = ' -outorder=input ' . $opt_refaln;
	my @char_spec = ($char_gap, $char_mask);

	my $ct_seqs = @{$in_aln}; 

	$tmp_dir = ${$tmp_dir};
	my $tmp_inaln = "$tmp_dir/in.aln";
	my $tmp_gdtree = "$tmp_dir/in.ph";
	my $tmp_outaln = "$tmp_dir/out.aln";
	my $log_gdtree = "$tmp_dir/log_gdtree.txt";
	my $log_refaln = "$tmp_dir/log_refaln.txt";

	($ct_seqs>1) or (return ($in_aln, undef, undef) ); # Do nothing #

		## ADDED on Tuesday, May 22, 2007 ##
	foreach my $tmpfile ($tmp_inaln, $tmp_gdtree, $tmp_outaln, $log_gdtree, $log_refaln) {
	    if ( -f $tmpfile ) {
		unless (0 == system ("mv -f $tmpfile $tmpfile.bk") ) {
		    print STDERR "[WARNING 'bundle_refine_maln_by_clw'] FAILED to rename the temporary file '$tmpfile' by appending the suffix '.bk' !\n";
		}
	    }
	}

	my @labels = my @cp_aln = ();
	for (my $k=0; $k<$ct_seqs; $k++) {
	    my $cp_seq = ${$in_aln->[$k]};
	    push @cp_aln, \$cp_seq;
	    push @labels, sprintf ("SEQ%04d", $k+1);
	}
	pr_clustal_like_lite (@labels, @{$in_aln}, 60, $tmp_inaln);

	if ($ct_seqs>3) { # Make a guidetree #
		# First examine the number of effective counts ##
	    my ($eff_aln, $coords) = rm_multi_spec_sites (@cp_aln, 1, @char_spec, 1);
	    my $ct_effsites = length (${$eff_aln->[0]}) ;
	    ($ct_effsites >= $minct_tossgap) and ($opt_gdtree .= ' -tossgaps ');

	    unless (0 == system ("(time ${$bin_clw} -tree -infile=$tmp_inaln $opt_gdtree) > $log_gdtree 2>&1" ) ) {
		print STDERR "[ERROR 'bundle_refine_maln_by_clw'] FAILED to conduct 'system (\"clustalw -tree -infile=$tmp_inaln ... \")'\n";
		print STDERR "  with the option '$opt_gdtree' : $?\n\n";
		return (undef, \$log_gdtree, undef);
	    }
	    $opt_refaln .= " -usetree=$tmp_gdtree "; ### For TESTING the subroutine
	}

	unless (0 == system ("(time ${$bin_clw} -align -infile=$tmp_inaln -outfile=$tmp_outaln $opt_refaln) > $log_refaln 2>&1") ) {

		## REVISED on Tuesday, May 22, 2007 ##
	    my $fl_mod_gdtree = 0;
	    if (( -f $tmp_gdtree ) and ( -r $tmp_gdtree) and ( -w $tmp_gdtree)) {

		# Try a slight modification of the guide tree #

		print STDERR "[CAUTION 'bundle_refine_maln_by_clw'] FAILED the 1st attempt of refining the alignment, so I will try a slight midification of the guide tree...\n";

###=Disable for TESTING the script
		my %node2pa; my %node2chi; my %name2id; my %id2name; my %br_attr; 
	        my $fl_ext_gdtree = extract_sgltree_from_file ($tmp_gdtree, %node2pa, %node2chi, %name2id, %id2name, %br_attr, '#') ;
		if ($fl_ext_gdtree) { # inner if 1
		    my $topnode_id = $name2id{$TOP_NODE};
		    my $sbtop_nodes = $node2chi{$topnode_id};
		    if ((defined $sbtop_nodes) and (@{$sbtop_nodes}>0)) { # inner if 2
			my $sbtop1 = $sbtop_nodes->[0];
			if ((defined $br_attr{$sbtop1}) and (defined $br_attr{$sbtop1}->[$INDX_LENGTH])) { # inner if 3
			    $br_attr{$sbtop1}->[$INDX_LENGTH] += 0.001;
			    my $fh_gdtree = open_output_file ($tmp_gdtree, 0);
			    if (br_output_tree ($topnode_id, %node2chi, %id2name, %br_attr, *${$fh_gdtree}, 0, 0) ) { # inner if 4
			        $fl_mod_gdtree = 1;
			    }  # end of inner if 4
			    close ($fh_gdtree);
			}  # end of inner if 3
		    } # end of inner if 2
		} # end of inner if 1 
###=cut
	    }
    
	    my $fl_retry = 0;
	    if ($fl_mod_gdtree) {
		# If the guide tree was successfully modified,
		# retry to refine the alignment.
		if (0 == system ("(time ${$bin_clw} -align -infile=$tmp_inaln -outfile=$tmp_outaln $opt_refaln) >> $log_refaln 2>&1")) {
		    print STDERR "    ... Succeeded in re-trying to refine the alignment !!\n\n";
		    $fl_retry = 1;
		}
	    }

	    unless ($fl_retry) {
		print STDERR "[ERROR 'bundle_refine_maln_by_clw'] FAILED to conduct 'system (\"clustalw -align -infile=$tmp_inaln ...\")'\n";
		print STDERR " with the program {clustalw} = ${$bin_clw} \n";
		print STDERR " and with the option '$opt_refaln' : $?\n\n";
		##print STDERR " with the option '$opt_refaln' : $?\n\n";
		return (undef, \$log_gdtree, \$log_refaln);
	    }
		## End of the revision on Tuesday, May 22, 2007 ##
	}


	(my $out_aln, my $labels2, undef, undef) = extract_clw ($tmp_outaln, 1);

	unless ((defined $out_aln) and (@{$out_aln} == $ct_seqs)) {
	    print STDERR "[ERROR 'bundle_refine_maln_by_clw'] FAILED to extract the refined alignment '$tmp_outaln': $! !\n\n";
	    return (undef, \$log_gdtree, \$log_refaln);
	}
	for (my $k=0; $k<$ct_seqs; $k++) {
	    unless ($labels[$k] eq $labels2->[$k]) {
		print STDERR "[ERROR 'bundle_refine_maln_by_clw'] The ID of the $k th sequence in the input alignment (%s) CLASHES that in the output algnment (%s) !\n\n",
		$labels[$k], $labels2->[$k];
		return (undef, undef, undef);
	    }
	}

	$out_aln = rm_gapcolumn2 (@{$out_aln}, $char_gap);

	return ($out_aln, undef, undef);

}		# End of the subroutine 'bundle_refine_maln_by_clw' #


# extract_unmasked_regions: 
#	This subroutine extracts regions of an alignment
#	that do NOT contain any masking characters.
#
#	NOTE that this subroutine eats up the input alignment.
#
# INPUT: (\@align, $indx_tgt, $beg_frame, $beg_tgt, $end_tgt, \%mask, $char_gap),
#	where
#	${$align[$m]} is the $m th sequence embedded in the alignment;
#	$indx_tgt gives the 'target sequence' whose coordinates will be monitored;
#	$beg_tgt and $end_tgt are the beginning and ending coordinates of the
#	regions of the 'target sequence' that is embedded in the alignment;
#	%mask = ($char => 1, ...) for the characters that represents masked residues;
#	$char_gap represents a gap.
#
# OUTPUT: (\@frames_unmasked, \@tgtcoords_unmasked, \@alns_unmasked), where
#	@{$frames_unmasked[$k]} = ($beg_rgfr, $end_rgfr) gives the frame of the $k th
#	unmasked region ;
#	@{$tgtcoords_unmasked[$k]} = ($beg_rg, $end_rg) specifies
#		the target coordinates of the boundaries of the $k th region ;
#	${$alns_unmasked[$k]->[$m]}  is the subsequence of the $m th sequence 
#	  embedded in the alignment for the $k th unmasked region .
#
sub extract_unmasked_regions (\@$$$$\%$) {

	my ($align, $indx_tgt, $beg_frame, $beg_tgt, $end_tgt, $mask, $char_gap) = @_;

	my $ct_seqs = @{$align};
	($ct_seqs>0) or (return ([], [], []) );
	my $min_alnlen = length ${$align->[0]} ;
	for (my $m=1; $m<$ct_seqs; $m++) {
	   my $alnlen = length ${$align->[$m]} ;
	   ($min_alnlen > $alnlen) and ($min_alnlen = $alnlen);
	}

	my $curr_frame = $beg_frame - 1;
	my $curr_tgt = $beg_tgt - 1;

=for DEBUGGING the script
	print STDERR "\n< Subroutine 'extract_unmasked_regions' >\n\n"; 
	print STDERR "Input_sequences:\n\n";
	foreach my $seq (@{$align}) {  print STDERR ${$seq}, "\n"; }
	print STDERR "\n";
=cut

	my @frames_um = my @tgtcoords_um = my @alns_um = ();
	for (my $l=0; $l<$min_alnlen; $l++) { # The outer for-loop #

	   my @chars = ();
	   my $ct_mask = 0;
	   for (my $m=0; $m<$ct_seqs; $m++) { # The 1st inner for-loop #
		my $char = substr ( ${$align->[$m]} , 0, 1, '' );
		push @chars, \$char;
		(defined $mask->{$char}) and ($ct_mask++);
	   }				# End of the 1st inner for-loop #

	   my $fl_tgt_gap = (${$chars[$indx_tgt]} eq $char_gap) ? 1 : 0 ;

		# Update the coordinates #
	   $curr_frame++;
	   $fl_tgt_gap or ($curr_tgt++);

	   ($ct_mask>0) and next; # Skip if the site contains masked residues #

	    unless ((@frames_um>0) and ($frames_um[$#frames_um]->[1] >= $curr_frame -1)) {
			# Start a new unmasked region #
		push @frames_um, [$curr_frame, $curr_frame];
		push @tgtcoords_um, [$curr_tgt + $fl_tgt_gap, $curr_tgt];
		push @alns_um, \@chars;
		next;
	    }

			# Extend the last unmasked region by a site #
	    $frames_um[$#frames_um]->[1] = $curr_frame;
	    $tgtcoords_um[$#tgtcoords_um]->[1] = $curr_tgt;
	    my $last_rgaln = $alns_um[$#alns_um];
	    for (my $m = 0; $m<$ct_seqs; $m++) { # The 2nd inner for-loop #
		${$last_rgaln->[$m]} .= ${$chars[$m]};
	    }			# End of the 2nd inner for-loop #

	}			# End of the outer for-loop #

=for DEBUGGING the script
	print STDERR "  --- 'extract_unmasked_regions' ENDED ---\n\n";
	print STDERR "Remmants of the input:\n\n";
	foreach my $seq (@{$align}) { print STDERR ${$seq} , "\n"; }
	print STDERR "\n\n";
=cut

	return (\@frames_um, \@tgtcoords_um, \@alns_um);

}		# End of the subroutine 'extract_unmasked_regions' #


# clip_specregions_maln:
#	This subroutine cuts out regions of multiple alignments
#	specified by the alignment coordinates.
#	The regions can overlap with one another, as long as the boundaries
#	are arranged in ascending order of starting coordinates.
#
#	NOTE1: The subroutine does NEITHER calculate the boundary residue numbers,
#	 NOR check the consistency of the coordinates.
#	NOTE2: The subroutine eats up the input alignment.
#
# INPUT: (\@align, $beg_frame, \@bd_coords), where
#	${$align[$m]} is the $m th sequence embedded in the alignment;
#	$beg_frame is the beginning coordinate of the alignment coordinate frame;
#	@{$bd_coords[$r]} = ($beg_rg, $end_rg) gives the alignment coordinates
#	of the boundaries for the $r th region. The elements are supposed to be
#	PRE-SORTED in ascending order of $beg_rg.
#
# OUTPUT: \@set_sb_malns, where
#	${$set_sb_malns[$r]->[$m]} is the $m the sequence embedded in the $r th
#	sub-alignment.
#
sub clip_specregions_maln (\@$\@) {

	my ($align, $beg_frame, $bd_coords) = @_;

	my $ct_seqs = @{$align};
	my $ct_rgs = @{$bd_coords};

		## A BUG fixed on May 11 (Mon), 2009 ##
	unless (($ct_seqs>0) and ($ct_rgs>0)) {
	###unless (($ct_seqs>0) or ($ct_rgs>0)) {
	    print STDERR "[ERROR 'clip_specregions_maln'] ILLEGAL INPUT: (#{sequences}, #{regions}) = ($ct_seqs, $ct_rgs): $! !\n\n";
	    return undef;
	}

	my $len_aln = length (${$align->[0]});
	my $end_frame = $beg_frame + $len_aln - 1;

	my $left_bd = $bd_coords->[0][0];
	unless (defined $left_bd) {
	    print STDERR "[ERROR 'clip_specregions_maln'] ILLEGAL INPUT: The left boundary of the 0 th region is UNDEFINED:  $! !\n\n";
	    return undef;
	}

	my $right_bd = $left_bd;
	foreach my $bd (@{$bd_coords}) { ($right_bd < $bd->[1]) and ($right_bd = $bd->[1]); }
	unless (($left_bd >= $beg_frame) and ($right_bd <= $end_frame)) {
	    print STDERR "[ERROR 'clip_specregions_maln'] ILLEGAL INPUT: The boundaries of the ALL regions ($left_bd, $right_bd) do NOT fall within the alignment coordinate frame ($beg_frame, $end_frame): $! !\n\n";
	    return undef;
	}

	my @set_sb_malns = ();

	my $last_beg = $beg_frame;
	for (my $r=0; $r<$ct_rgs; $r++) { # The main for-loop #

		## A BUG Fixed on May 11 (Mon), 2009 ##
	    my ($beg_rg, $end_rg) = @{$bd_coords->[$r]};
	    ##my ($beg_rg, $end_rg) = @{$bd_coords};

	    my $len_sbaln = $end_rg - $beg_rg + 1;
	    unless ($len_sbaln > 0) {
		print STDERR "[ERROR 'clip_specregions_maln'] ILLEGAL INPUT: The beginning point ($beg_rg) is in the DOWNSTREAM of the ending poit ($end_rg) for the $r th region: $! !\n\n";
		return undef;
	    }
	    my $left_margin = $beg_rg - $last_beg;
	    unless ($left_margin >= 0) {
		print STDERR "[ERROR 'clip_specregions_maln'] ILLEGAL INPUT: The beginning point ($beg_rg) of the $r th region is in the UPSTREAM of the last beginning point ($last_beg) : $! !\n\n";
		return undef;
	    }
	    if ($left_margin >0) { # Trim off the left margin #
		for (my $m=0; $m<$ct_seqs; $m++) { # The 1st inner for-loop #
		    substr (${$align->[$m]}, 0, $left_margin, '');
		}				# End of the 1st inner for-loop #
	    }

	    my @sb_maln = ();
	    for (my $m=0; $m<$ct_seqs; $m++) { # The 2nd inner for-loop #
		my $sbseq = substr (${$align->[$m]}, 0, $len_sbaln); 
		push @sb_maln, \$sbseq;
	    }				# End of the 2nd inner for-loop #

	    push @set_sb_malns, \@sb_maln;
	    $last_beg = $beg_rg;
	}				# End of the main for-loop #

	return \@set_sb_malns;

}		# End of the subroutine 'clip_specregions_maln' #


# extract_subregions_maln:
#	This satellite subroutine of 'examine_tgt_regions_pwaln'
#	extracts NON-OVERLAPPING subregions of the input multiple alignment
#	that are EXPLICITLY specified by the boundary coordinates
#	given in either frame coordinates or
#	the coordinate of one of the sequences.
#
#	NOTE that this subroutine eats up the input alignment.
#
# INPUT: (\@align, $indx_ref, $beg_frame, \@aln_begs, \@aln_ends, \@rg_bounds, $char_gap),
#	where
#	${$align[$m]} is the $m th sequence embedded in the alignment;
#	$indx_ref specifies the 'reference sequence', whose coordinates
#	provides the 'reference frame' according to which the alignment is cut.
#	If $indx_ref < 0, the alignment frame will be used instead;
#	$beg_frame is the beginning coordinate of the alignment frame;
#	$aln_begs/ends[$m] is the coordinates of the beginnig/ending of
#	the $m th sequence within the alignment;
#	@{$rg_bounds[$r]} = ($rg_beg, $rg_end) gives the boundaries of the $r th
#	subalignment to be extracteds, expressed in the coordinate system
#	of the $indx_ref th sequence (or the alignment frame if $indx_ref< 0) .
#
# OUTPUT: (\@sb_malns, \@sbaln_frames, \@sbaln_coords, \@remnants), where
#	${$sb_malns[$r]->[$m]} is the $m th sequence embedded in the $r th sub-alignment;
#	@{$sbaln_frames[$r]} = ($beg, $end) gives the boundaries of the $r th sub-alignment
#	 in the alignment frame;
#	@{$sbaln_coords[$r]->[$m]} = ($sb_beg, $sb_end) are the beginning/ending residue numbers in the $m th sequence in the $r th sub-alignment.
#	(To put it more precisely, $sb_beg is the number of the right-neighboring residue of the left boundary, and $sb_end is the number of the left-neighboring residue of the right boundary);
#	${$remnants[$r]->[$m]} is the $m th sequence embedded in the left-franking region of the $m th sub-alignment.
#	The $m = scalar (@rg_bounds) th element gives sequences in the right-frankking region.
#
sub extract_specregions_maln (\@$$\@\@\@$) {

	my ($align, $indx_ref, $beg_frame, $aln_begs, $aln_ends,
		 $rg_bounds, $char_gap) = @_;

	my $ct_seqs = @{$align};
	($ct_seqs>0) or ( return ([], [], [], []) );
	my $ct_rgs  = @{$rg_bounds};
	($ct_rgs>0) or (return ([], [], [], $align) );

		# Initialize #
	my @spec = ($char_gap);
	my $aln_len = length ( ${$align->[0]} ); # The minimum length of the alignment #

	my $curr_frame = $beg_frame - 1;
	my $end_frame = $curr_frame + $aln_len;

	my @curr_coords = ();
	for (my $m=0; $m<$ct_seqs; $m++) { # The 1st outer for-loop #
	    my $beg = $aln_begs->[$m];
	    ($beg == int ($beg)) or ($beg = $aln_begs->[$m] = int ($beg) + 1);
	    my $end = $aln_ends->[$m] = int ($aln_ends->[$m]) ;

	    my $len_in_aln = length (${$align->[$m]});
	    ($aln_len > $len_in_aln) and ($aln_len = $len_in_aln);
	    my $ct_gaps = count_spec_chars (${$align->[$m]}, @spec, 0);
	    my $seqlen = $len_in_aln - $ct_gaps;
	    unless ($seqlen == $end - $beg + 1) {
		my $seqlen2 = $end - $beg + 1;
		print STDERR "[CAUTION 'extract_specregions_maln'] The length of the $m th sequence embedded in the alignment ($seqlen) CLASHES with that expected from the boundary coordinates: $seqlen2 =  $end (end) - $beg (beg) + 1 !\n";
	    }

	    push @curr_coords, $beg - 1;
	}			# End of the 1st outer for-loop #

	my $begin_aln = ($indx_ref<0) ? $beg_frame : $aln_begs->[$indx_ref] ;
	my $end_aln = ($indx_ref<0) ? $end_frame : $aln_ends->[$indx_ref] ; 
	if ($rg_bounds->[0][0] < $begin_aln) {
	    my $leftmost = $rg_bounds->[0][0];
	    print STDERR "[ERROR 'extract_specregions_maln'] The leftmost boundary of the regions you desire ($rg_bounds->[0][0]) is UPSTREAM OF THE BEGINNING of the alignment ($begin_aln) regarding the $indx_ref th sequence !\n";
	    return (undef, undef, undef, undef);
	}

	if ($end_aln < $rg_bounds->[$#{$rg_bounds}][1]) {
	    my $rightmost = $rg_bounds->[$#{$rg_bounds}][1];
	    print STDERR "[ERROR 'extract_specregions_maln'] The rightmost boundary of the regions you desire ($rightmost) is DOWNSTREAM OF THE END of the alignment ($end_aln) regarding the $indx_ref th sequence !\n";
	    return (undef, undef, undef, undef);
	}


	my @sbalns = my @sb_frames = my @sb_coords = my @remnants = ();

	my @cp_rg_bounds = @{$rg_bounds};

	my $ref_coord = ($indx_ref<0) ? $curr_frame : $curr_coords[$indx_ref];

	while (my $rg_begend = shift @cp_rg_bounds) { # The 2nd outer while-loop #

	    my ($rg_beg, $rg_end) = @{$rg_begend};

	    ($rg_beg > $rg_end) and next; # Skip Null boundaries #

		# Extract the left flaking region #
	    my @left_flank = ();
	    for (1 .. $ct_seqs) { my $null_str = ''; push @left_flank, \$null_str; }

	    while ($ref_coord < $rg_beg) { # the 1st middle while-loop #
		$curr_frame++;

		my @site = ();
		for (my $m=0; $m<$ct_seqs; $m++) { # The 1st inner for-loop #
		    my $char = substr (${$align->[$m]}, 0, 1, '');
		    push @site, \$char;
		    ($char eq $char_gap) or ($curr_coords[$m]++);
		}			# End of the 1st inner for-loop #

		$ref_coord = ($indx_ref<0) ? $curr_frame : $curr_coords[$indx_ref] ;

		if ($ref_coord < $rg_beg) { # Still in the left-flanking region #
		    for (my $m=0; $m<$ct_seqs; $m++) { ${$left_flank[$m]} .= ${$site[$m]}; }
		} elsif ($ref_coord == $rg_beg) { # We reached a region to be extracted #
		    push @sbalns, \@site;
		    my @bd_coords = ();
		    for (my $m=0; $m<$ct_seqs; $m++) { # The 2nd inner for-loop #
			my $beg_coord =  my $end_coord = $curr_coords[$m];
			(${$site[$m]} eq $char_gap) and ($beg_coord++);
			push @bd_coords, [$beg_coord, $end_coord];
		    }				# End of the 2nd inner for-loop #
		    push @sb_coords, \@bd_coords;
		    push @sb_frames, [$curr_frame, $curr_frame];
		    last;
		} else { # Normally, this cannot happen #
		    print STDERR "[ERROR 'extract_specregions_maln'] The coordinate ($ref_coord) EXCEEDED THE LEFT_BOUNDARY ($rg_beg) of the region to be extracted !!\n";
		    return (undef, undef, undef, undef);
		}

	    }			# End of the 1st middle while-loop #

	    push @remnants, \@left_flank;

		# Extract the region to be extracted #

	    my $curr_sbaln = $sbalns[$#sbalns];
	    my $curr_bdcoords = $sb_coords[$#sb_coords];
	    my $curr_sbframe = $sb_frames[$#sb_frames];

	    while ($ref_coord < $rg_end) { # The 2nd middle while-loop #
		$curr_frame++;
		my @site = ();
		for (my $m=0; $m<$ct_seqs; $m++) { # The 3rd inner for-loop #
		    my $char = substr (${$align->[$m]}, 0, 1, '');
		    ${$curr_sbaln->[$m]} .= $char;
		    ($char eq $char_gap) or ($curr_coords[$m]++);
		}			# End of the 3rd inner for-loop #

		$ref_coord = ($indx_ref<0) ? $curr_frame : $curr_coords[$indx_ref] ;

		($ref_coord == $rg_end) or next;

		$curr_sbframe->[1] = $curr_frame;
		for (my $m=0; $m<$ct_seqs; $m++) { # The 4th inner for-loop # 
		    $curr_bdcoords->[$m][1] = $curr_coords[$m];
		}			# End of the 4th inner for-loop #
		last;
	    }			# End of the 2nd middle while-loop #

	}			# End of the 2nd outer while-loop #


		# Put the rightmost flanking region into the set of remnants #
	push @remnants, $align;

	return (\@sbalns, \@sb_frames, \@sb_coords, \@remnants);

}		# End of the subroutine 'extract_specregions_maln' #


# mask_site:
#	This satellinte subroutine of 'mask_spec_regions'
#	masks a site of an alignment in a manner specified by the input arguments.
#	The subroutine NEVER maskes gaps, because it may confound the
#	coordinate frame.
#
# INPUT: (\@site, $char_gap, $char_mask, $indx_ref, $fl_all, $fl_caseignore), where
#	$site[$m] gives the residue (or gap) of the $m th sequence
#	residing in the site in question;
#	$indx_ref specifies the sequence to be masked when $indx_ref >= 0,
#	$indx_ref < 0 orders that ALL sequences to be masked;
#	$fl_all = True indicates that ALL sequences should be masked;
#	$fl_caseignore = True ignores the differences  between lowercase
#	and uppercase letters,
#		= False respects the differences.
#
# OUTPUT: \@masked_site,
#	which is the masked version of @site.
#
sub mask_site (\@$$$$$) { 

	my ($site, $char_gap, $char_mask,
		$indx_ref, $fl_all, $fl_caseignore) = @_;

	my $uc_char_gap = uc $char_gap;
	my @masked_site = @{$site}; 

	unless ($fl_all or ($indx_ref< 0)) { # Mask ONLY SPECIFIED Sequence #
	    my $char_ref = $masked_site[$indx_ref];
	    unless ( ($char_ref eq $char_gap) or
		($fl_caseignore and (uc ($char_ref) eq $uc_char_gap)) ) {
		$masked_site[$indx_ref] = $char_mask;
	     }
	    return \@masked_site;
	}

		# Mask ALL residues #

	my $ct_seq = @masked_site;

	for (my $m=0; $m<$ct_seq; $m++) { # The main for-loop #
	    my $char = $masked_site[$m];
	    unless ( ($char eq $char_gap) or
		($fl_caseignore and (uc ($char) eq $uc_char_gap)) ) {
		$masked_site[$m] = $char_mask;
	    }
	}			# end of the main for-loop #

	return \@masked_site;

}		# end of the subroutine 'mask_site' #


# mask_spec_regions:
#	This satellite subroutine of 'examine_tgt_regions_pwaln'
#	masks the regions of the input alignment
#	that are EXPLICITLY specified by the input list of boundaries.
#
# 	NOTE that this subroutine eats up the input alignment.
#
# INPUT: (\@align, $beg_frame, $end_frame, \@beg_coords, \@end_coords, $indx_ref,
#		\@masked_rgs, $char_gap, $char_mask, $fl_all), where
#	${$align[$m]} is the $m th sequence embedded in the alignment;
#	$beg/end_frame gives the frame coordinate of the beginning/ending
#	site of the alignment;
#	$beg/end_coords[$m] is the beginning/ending coordinate of the $m th sequence;
#	$indx_ref provides the 'reference coordinate frame'  of the masked regions.
#	$indx_ref >= 0 appoints the $indx_ref th sequence as a reference,
#	$indx_ref < 0 refers to the alignment frame ;
#	$aln_beg/end is the beginning/end of the alignment in the coordinate system
#	specified by $indx_ref;
#	@{$masked_rgs[$r]} = ($rg_beg, $rg_end) is the beginning/end of the $r th region that is to be masked,
#	in the coordinate system specified by $indx_ref;
#	When $fl_all = 1, ALL sequences are masked in the region.
#	        = 0 ONLY the REFERENCE sequence is masked (if $indx_ref >=0).
#	(When $indx_ref < 0, ALL sequences are masked regardless of the value of $fl_all.)
#
# OUTPUT: (\@masked_aln, \@masked_frame, \@masked_coords), where
#	${$masked_aln[$m]} is the $m th sequence embedded in the masked alignment;
#	@{$masked_frame[$r]} = ($fr_beg, $fr_end) is the frame coordinates of the
#	beginning and ending of the $r th region that are masked;
#	@{$masked_coords[$r]->[$m]} = ($rg_sbbeg, $rg_sbend) is the beginning/end
#	coordinate of the $r th region in the $m th sequence;
#
sub mask_spec_regions (\@$$\@\@$\@$$$) {

	my ($align, $beg_frame, $end_frame, $beg_coords, $end_coords, $indx_ref,
		$masked_rgs, $char_gap, $char_mask, $fl_all) = @_;

	my $ct_seqs = @{$align};
	($ct_seqs>0) or (return ([], [], []));
	my $ct_rgs = @{$masked_rgs};
	($ct_rgs>0) or (return ($align, [], []));
		# It does NOT have to mask  ANY region 

 	my ($aln_beg, $aln_end) = ($indx_ref < 0) ? ($beg_frame, $end_frame) :
		($beg_coords->[$indx_ref], $end_coords->[$indx_ref] ) ;
	my $min_seqlen = length ${$align->[0]};
	for (my $m = 1; $m<$ct_seqs; $m++) { # The 1st for-loop #
	    my $seqlen = length ${$align->[$m]};
	    ($min_seqlen > $seqlen) and ($min_seqlen = $seqlen);
	}			# End of the 1st for-loop #

	my $len_frame = $min_seqlen;
	if ($indx_ref >= 0) {
	    $len_frame = length ${$align->[$indx_ref]} ;
	    my @gap = ($char_gap);
	    my $ct_gap = count_spec_chars (${$align->[$indx_ref]}, @gap, 1) ;
	    $len_frame -= $ct_gap;
	}

	unless ($len_frame == $aln_end - $aln_beg + 1) {
	    my $len_frame2 = $aln_end - $aln_beg + 1;
	    print STDERR "[ERROR 'mask_spec_regions'] The length of the reference frame ($len_frame) CLASHES with that expected from the boundaries:  $len_frame2 = $aln_end (end) - $aln_beg (begin) + 1 !\n";
	    return (undef, undef, undef) ;
	}
	my $leftmost = $masked_rgs->[0][0];
	my $rightmost = $masked_rgs->[$#{$masked_rgs}][1];
	if ($leftmost < $aln_beg) {
	    print STDERR "[ERROR 'mask_spec_regions'] The leftmost boundary of the masked region is UPSTREAM OF THE BEGINNING of the alignment !\n";
	    return (undef, undef, undef) ;
	}
	if ($aln_end < $rightmost ) {
	    print STDERR "[ERROR 'mask_spec_regions'] The rightmost boundary of the masked region is DOWNSTREAM OF THE END of the alignment !\n";
	    return (undef, undef, undef) ;
	}

	my $curr_frame = $beg_frame - 1;
	my @curr_coords = ();
	foreach  my $beg (@{$beg_coords}) { push @curr_coords, $beg - 1; }

	my $ref_coord = ($indx_ref < 0) ? $curr_frame : $curr_coords[$indx_ref];
	my @cp_masked_rgs = @{$masked_rgs};

		## Initialize the output ##
	my @masked_aln = my @masked_frame = my @masked_coords = ();
	for (1 .. $ct_seqs) { my $null_str = ''; push @masked_aln, \$null_str; }

	while (my $masked_rg = shift @cp_masked_rgs) { # The 2nd outer while-loop
	    my ($masked_beg, $masked_end) = @{$masked_rg};

	    ($masked_beg > $masked_end) and next; # Skip Null Regions #


		# Process the preceding region #

	    while ($ref_coord < $masked_beg) { # The 1st middle while-loop #
		$curr_frame++;
		my @site = ();
		for (my $m=0; $m<$ct_seqs; $m++) { # The 1st inner for-loop #
		    my $char = substr (${$align->[$m]}, 0, 1, '');
		    push @site, $char;
		    ($char eq $char_gap) or ($curr_coords[$m]++);
		}			# End of the 1st inner for-loop #

		$ref_coord = ($indx_ref<0) ? $curr_frame : $curr_coords[$indx_ref];

		if ($ref_coord < $masked_beg) { # Still in the left-flanking region #
		    for (my $m=0; $m<$ct_seqs; $m++) { ${$masked_aln[$m]} .= $site[$m]; }

		} elsif ($ref_coord == $masked_beg) { # We've reached a region to be masked #
		    my $masked_site = mask_site (@site, $char_gap, $char_mask, $indx_ref, $fl_all, 1);

		    my @bd_coords = ();
		    for (my $m=0; $m<$ct_seqs; $m++) { # The 2nd inner for-loop #
			${$masked_aln[$m]} .= $masked_site->[$m];
			my $beg_coord = my $end_coord = $curr_coords[$m];
			($site[$m] eq $char_gap) and ($beg_coord++);
			push @bd_coords, [$beg_coord, $end_coord];
		    }				# End of the 2nd inner for-loop #
		    push @masked_coords, \@bd_coords;
		    push @masked_frame, [$curr_frame, $curr_frame];
		    last;

		} else { # Normally, this cannot happen #
		    print STDERR "[ERROR 'mask_spec_regions'] The coordinate ($ref_coord) EXCEEDED THE LEFT BOUDNARY ($masked_beg) of the region to be masked !!\n";
		    return (undef, undef, undef);
		}
	    }				# End of the 1st middle while-loop #

		# Mask the major region #

	    my $curr_bdcoords = $masked_coords[$#masked_coords];
	    my $curr_msframe = $masked_frame[$#masked_frame];

	    while ($ref_coord < $masked_end) { # The 2nd middle while-loop # 
		$curr_frame++;
		my @site = ();
		for (my $m=0; $m<$ct_seqs; $m++) { # The 3rd inner for-loop #
		    my $char = substr (${$align->[$m]}, 0, 1, '');
		    push @site, $char;
		    ($char eq $char_gap) or ($curr_coords[$m]++) ;
		}			# End of the 3rd inner for-loop #

		my $masked_site = mask_site (@site, $char_gap, $char_mask, $indx_ref, $fl_all, 1);
		for (my $m=0; $m<$ct_seqs; $m++) { ${$masked_aln[$m]} .= $masked_site->[$m]; }

		$ref_coord = ($indx_ref<0) ? $curr_frame : $curr_coords[$indx_ref] ;
		($ref_coord == $masked_end) or next;

		$curr_msframe->[1] = $curr_frame;
		for (my $m=0; $m<$ct_seqs; $m++) { # The 5th inner for-loop #
		    $curr_bdcoords->[$m][1] = $curr_coords[$m];
		}			# End of the 5th inner for-loop #
		last;
	    }			# End of the 2nd middle while-loop #

	}				# End of the 2nd outer while-loop

		# Graft the rightmost flanking region onto the masked alignment # 
	for (my $m=0; $m<$ct_seqs; $m++) { ${$masked_aln[$m]} .= ${$align->[$m]}; }

	return (\@masked_aln, \@masked_frame, \@masked_coords);

}		# End of the subroutine 'mask_spec_regions' #



# mask_unspec_regions:
#	This subroutine masks regions NOT SPECIFIED by the input arguments.
#	NOTE that this subroutine EATS UP the input alignment.
#
# INPUT: (\@align, $beg_frame, \@begs, \@ends, $indx_ref, \@unmasked,
#		$char_gap, $char_mask, $fl_all), where
#	${$align[$m]} is the $m th sequence embedded in the alignment;
#	$beg_frame is the beginning coordinate of the alignment frame;
#	$begs[$m] and $ends[$m] gives the start and end coordinates of the $m th sequence in the alignment;
#	@{$unmasked[$r]} = ($um_beg, $um_end) is the $r th region that should be
#	kept UNmasked, in terms of the $indx_ref th sequence coordinates
#	(or the alignment frame if $indx_ref < 0).
#	This subroutine DOES NOT CARE whether the specified unmasked regions
#	sticks out of the entire alignment or not;
#	$fl_all = 1 if ALL sequences are to be masked,
#		= 0 if only the reference sequence is to be masked
#			(when $indx_ref>=0).
#	When $indx_ref < 0, ALL sequences are masked unconditionally.
#
# OUTPUT: (\@masked_aln, \@frames_unmasked, \@coords_unmasked), where
#	@masked_aln stores sequences embedded in the masked alignment;
#	@{$frames_unmasked[$r]} = ($fbeg, $fend) gives the frame of the $r th unmasked region;
#	@{$coords_unmasked[$r]->[$m]} = ($sbbeg, $sbend) is
#	the $rh th unmasked region in terms of the $m th sequence coordinate;
#	they are empty if the corresponding region is not included in 
#	the aligned regions.
#
sub mask_unspec_regions (\@$\@\@$\@$$$) {

	my ($align, $beg_frame, $begs, $ends, $indx_ref,
		$unmasked, $char_gap, $char_mask, $fl_all) = @_;

	($indx_ref < 0) and ($fl_all = 1);
	my $ct_seqs = @{$align};
	($ct_seqs>0) or (return ([], [], []));

	my $curr_frame = ($beg_frame == int $beg_frame) ?  $beg_frame -1 : int $beg_frame ;
	my @masked_aln = my @curr_coords = ();
	my $minlen = length ${$align->[0]};
	for (my $m=0; $m<$ct_seqs; $m++) { # The 1st for-loop #
	    my $init = '';
	    push @masked_aln , \$init;
	    my $len = length ${$align->[$m]};
	    ($minlen > $len) and ($minlen = $len);
	    my $beg = $begs->[$m];
	    push @curr_coords, ($beg == int $beg) ? $beg - 1 : int $beg ;
	}			# End of the 1st for-loop #

	my $end_frame = $beg_frame + $minlen - 1;

	my @frames_um = my @coords_um = ();

	my @cp_umrgs = @{$unmasked};

	while (my $um_rg = shift @cp_umrgs) { # The 2nd outer while-loop #

	    my ($um_beg, $um_end) = @{$um_rg};

	    my $curr_coord = ($indx_ref< 0) ? $curr_frame : $curr_coords[$indx_ref] ;
	    my @rg_coords = ();
	    my @rg_frame = ();
	    if ($curr_coord < $um_beg) {
		# The current site is on the left of the 'unmasked region'

		while ($curr_frame < $end_frame) { # The 1st middle while-loop #
		    $curr_frame++;
		    my @tmp_site = ();
		    for (my $m=0; $m<$ct_seqs; $m++) { # the 1st inner for-loop #
			my $site = substr (${$align->[$m]}, 0, 1, '');
			push @tmp_site, $site;
			($site ne $char_gap) and ($curr_coords[$m]++);
		    }				# End of the 1st inner for-loop #
		    $curr_coord = ($indx_ref<0) ?
			$curr_frame : $curr_coords[$indx_ref];
		    if ($curr_coord < $um_beg) {
			my $masked_site = mask_site (@tmp_site, $char_gap, $char_mask, $indx_ref, $fl_all, 1);
			for (my $m=0; $m<$ct_seqs; $m++) { ${$masked_aln[$m]} .= $masked_site->[$m]; }
		    } elsif ($curr_coord == $um_beg) {
			for (my $m=0; $m<$ct_seqs; $m++) {# The 3rd inner for-loop
			    ${$masked_aln[$m]} .= $tmp_site[$m];
			    my $curr_beg = ($tmp_site[$m] eq $char_gap) ? $curr_coords[$m] + 1 : $curr_coords[$m];
			    push @rg_coords, [$curr_beg, $curr_coords[$m]];
			}			# End of the 3rd inner for-loop # 
			@rg_frame = ($curr_frame, $curr_frame);
			last;

		    } else {
			print STDERR "[ERROR 'mask_unspec_regions'] an UNEXPECTED RESULT: current coordinate ($curr_coord) EXCEEDED the beginning of the right-neighboring unmasked region ($um_beg) while scanning from the left !\n\n";
			return undef;
		    }
		}	# End of the 1st middle while-loop #
	    }

	    if ($curr_coord < $um_end) {
		# The current site is in the midst of the 'unmasked region' 
		if ((@rg_frame == 0) and ($curr_frame < $end_frame)) {
		    @rg_frame = ($curr_frame+1, $curr_frame+1);
		    for (my $m=0; $m<$ct_seqs; $m++) { # The 2nd middle for-loop #
			my $next_beg = ($char_gap ne substr (${$align->[$m]}, 0, 1) ) ?
				$curr_coords[$m] + 1 : $curr_coords[$m];
			push @rg_coords, [$curr_coords[$m]+1, $next_beg]; 
		    }				# end of the 2nd middle for-loop #
		}

		while ($curr_frame < $end_frame) { # The 3rd middle while-loop #
		    $curr_frame++;
		    my @tmp_site = ();
		    for (my $m=0; $m<$ct_seqs; $m++) { # the 4th inner for-loop #
			my $site = substr (${$align->[$m]}, 0, 1, '');
			push @tmp_site, $site;
			($site ne $char_gap) and ($curr_coords[$m]++);
		    }				# End of the 4th inner for-loop #
		    $curr_coord = ($indx_ref<0) ?
			$curr_frame : $curr_coords[$indx_ref];
		    if ($curr_coord > $um_end) {
			print STDERR "[ERROR 'mask_unspec_regions'] an UNEXPECTED RESULT: current coordinate ($curr_coord) EXCEEDED the end of the unmasked region ($um_end) while scanning its inside !\n\n";
			return undef;
		    } else {
				## Bug FIXED just after the creation of
				## this subroutine (on Dec 25 (Tue), 2007)
			for (my $m=0; $m<$ct_seqs; $m++) { ${$masked_aln[$m]} .= $tmp_site[$m]; }
			##my $masked_site = mask_site (@tmp_site, $char_gap, $char_mask, $indx_ref, $fl_all, 1); 
			##for (my $m=0; $m<$ct_seqs; $m++) { ${$masked_aln[$m]} .= $masked_site->[$m]; }
			($curr_coord == $um_end) and last;
		    }
		}	# End of the 3rd middle while-loop #

		for (my $m=0; $m<$ct_seqs; $m++) { $rg_coords[$m]->[1] = $curr_coords[$m]; }
		$rg_frame[1] = $curr_frame;
	    }

		# The current site is now on the right of the 'unmasked region' # 
		# ($curr_coord >= $um_end)

	    push @frames_um, \@rg_frame;
	    push @coords_um, \@rg_coords;

	}			# End of the 2nd outer while-loop # 

		# Mask the remaining (downstream) regions #

	while ($curr_frame < $end_frame) { # The 3rd outer while-loop #
	    $curr_frame++;
	    my @tmp_site = ();
	    for (my $m=0; $m<$ct_seqs; $m++) { push @tmp_site, substr (${$align->[$m]}, 0, 1, ''); }
	    my $masked_site = mask_site (@tmp_site, $char_gap, $char_mask, $indx_ref, $fl_all, 1);
	    for (my $m=0; $m<$ct_seqs; $m++) { ${$masked_aln[$m]} .= $masked_site->[$m]; }
	}			# End of the 3rd outer while-loop #

	return (\@masked_aln, \@frames_um, \@coords_um);

}			# End of the subroutine 'mask_unspec_regions' #


# examine_tgt_regions_pwaln:
#	This subroutine examines regions of a pairwise alignment
#	that are specified by a set of intervals in the coordinate of
#	a target sequence.
#	As for the UNspecified regions,
#	the subroutine removes the external regions,
#	and masks the internal residues in between the specified regions,
#	before calculating the score.
#
# INPUT: (\@pwaln, $indx_tgt, $frame_beg, \@aln_begs, \@aln_ends, \@regions, 
#	\%score_matrix, \@cmp_parameters), where
#	@pwaln = (\$qu_seq, \$sb_seq) is the sequences aligned;
#	$indx_tgt = -1, 0, and +1 if we use
#	the alignment frame, query sequence, and subject coordinates, respectively;
#	$frame_beg is the beginning coordinate of the alignment frame;
#	$aln_{begs/ends}[$m] is the beginning/ending coordinate of the sequence
#	${$pwaln[$m]} embeddded in the alignment;
#	@{$regions[$r]} = ($beg, $end) is the boundaries of the $r th region
#	in terms of the coordinate system specified by $indx_tgt;
#	@cmp_parameters = ($mtrx_delim, $gapopen, $gapext, $char_gap,
#			$lambda, $k_r2b, $char_mask);
#
# OUTPUT: (#{total_sites}, #{identical_res}, total_score, #{gaps}, #{res}_qu, #{res}_sb)
#	for the pairwise alignment with the flanking regions removed and
#	with the internal unspecified residues masked.
#
sub examine_tgt_regions_pwaln (\@$$\@\@\@\%\@) {

	my ($pwaln, $indx_tgt, $frame_beg, $aln_begs, $aln_ends, $regions,
		$score_matrix, $cmp_parameters) = @_;

	my ($mtrx_delim, $gopen, $gext, $char_gap, $lambda, $k_r2b, $char_mask)
		= @{$cmp_parameters};

	unless ((defined $pwaln) and (@{$pwaln}>1)) {
	    print STDERR "[ERROR 'examine_tgt_regions_pwaln'] The INPUT pairwise alignment is CRIPPLED !!\n";
	    return (undef undef, undef, undef, undef, undef);
	}
	(@{$regions}>0) or (return (0,0,0,0,0,0));

	my $left_bd = $regions->[0][0];
	my $right_bd = $regions->[$#{$regions}][1];
	my @extbd_rgs = ([$left_bd, $right_bd]);

	my $cp_quseq = ${$pwaln->[0]};
	my $cp_sbseq = ${$pwaln->[1]};
	my @cp_pwaln = (\$cp_quseq, \$cp_sbseq);

=for TESTING the script
	my $frame_end = $frame_beg - 1 + length ($cp_quseq);
	print STDERR "\n< Subroutine 'examine_tgt_regions_pwaln' >\n\n";
	print STDERR "{ Input algnment (frame = \[$frame_beg\,$frame_end\]) }\n\n";
	my @labels = ('query', 'sbjct');
	for (my $k=0; $k<2; $k++) { $labels[$k] .=  sprintf ("\[%s\,%s\]", $aln_begs->[$k], $aln_ends->[$k]); }
	pr_ODEN_input (@labels, @cp_pwaln, 60, *STDERR);
	my $str_regions = string_intvl_set (@{$regions}, 0, 1) ;
	print STDERR "\n{ target regions (index = $indx_tgt): ${$str_regions} }\n\n";
	print STDERR "  --- Now conduct 'extract_specregions_maln' ---\n\n";
=cut

	my ($sb_pwalns, $sbaln_frames, $sbaln_coords, $remnants)
		= extract_specregions_maln (@cp_pwaln, $indx_tgt, $frame_beg,
			@{$aln_begs}, @{$aln_ends}, @extbd_rgs, $char_gap);

	unless ((defined $sb_pwalns) and (@{$sb_pwalns}==1)) {
	    print STDERR "[ERROR 'examine_tgt_regions_pwaln'] The subroutine 'extract_specregions_maln' FAILED !\n";
	    return (undef, undef, undef, undef, undef, undef);
	}

=for TESTING the script
	printf STDERR "\n{ The resulting trimmed alignment (frame = \[%s\,%s\])}\n\n",
		@{$sbaln_frames->[0]};
	my @labels2 = ('query', 'sbjct');
	for (my $k=0; $k<2; $k++) { $labels2[$k] .= sprintf ("\[%s\,%s\]", @{$sbaln_coords->[0][$k]} ); }
	pr_ODEN_input (@labels2, @{$sb_pwalns->[0]}, 60, *STDERR);
	print STDERR "\n";
	
=cut

	(undef, undef, my $masked_regions) = compare_interval_sets (@{$regions}, @extbd_rgs) ;
	my $sbaln_begframe = $sbaln_frames->[0][0];
	my $sbaln_endframe = $sbaln_frames->[$#{$sbaln_frames}][1];
	my @sbaln_begs = my @sbaln_ends = ();
	foreach my $bd_coords (@{$sbaln_coords->[0]}) {
	    push @sbaln_begs, $bd_coords->[0];
	} 
	foreach my $bd_coords (@{$sbaln_coords->[$#{$sbaln_coords}]}) {
	    push @sbaln_ends, $bd_coords->[1];
	}

=for TESTING the script
	my $str_mask_regions = string_intvl_set (@{$masked_regions}, 0, 1);
	print STDERR "\n --- Now conduct 'mask_spec_regions' ---\n\n";
	print STDERR "{ Regions to be masked (frame = $indx_tgt): ${$str_mask_regions} }\n\n";
=cut

	my ($masked_pwaln, $masked_frame, $masked_coords)
	  = mask_spec_regions (@{$sb_pwalns->[0]}, $sbaln_begframe, $sbaln_endframe,
		@sbaln_begs, @sbaln_ends, $indx_tgt,
		@{$masked_regions}, $char_gap, $char_mask, 1);

=for TESTING the script
	my $str_masked_frame = string_intvl_set (@{$masked_frame}, 0, 1);
	print STDERR "{ Final masked alignment (masked frames = ${$str_masked_frame}) }\n\n";
	for (my $k=0; $k<2; $k++) { $labels2[$k] .= '/'; }
	foreach my $masked_bd (@{$masked_coords}) { 
	  for (my $k=0; $k<2; $k++) { $labels2[$k] .= sprintf ("\[%s\,%s\]", @{$masked_bd->[$k]}); }
	}
	pr_ODEN_input (@labels2, @{$masked_pwaln}, 60, *STDERR);
	print STDERR "\n";
=cut

	my $info_pwaln_score = score_pairalign (${$masked_pwaln->[0]}, ${$masked_pwaln->[1]},
		%{$score_matrix}, $mtrx_delim, $gopen, $gext, $char_gap) ;

	unless ((defined $info_pwaln_score) and (@{$info_pwaln_score}>5)) {
	    print STDERR "[ERROR 'examine_tgt_regions_pwaln'] 'score_pairailgn' FAILED !\n";
	    return (undef, undef, undef, undef, undef, undef);
	}

=for TESTING the subroutine
	print STDERR "\n === The subroutine 'examine_tgt_regions_pwaln' now ENDED ===\n\n";
=cut

	my ($totct_sites, $ct_res_qu, $ct_res_sb, $ct_id, $ct_gap, $tot_score)
		= @{$info_pwaln_score};

	return ($totct_sites, $ct_id, $tot_score, $ct_gap, $ct_res_qu, $ct_res_sb);

}		# End of the subroutine 'examine_tgt_regions_pwaln' #


# cnct_examine_tgt_regions_pwaln:
#	This subroutine first extracts regions of a pairwise alignment
#	that are specified by a set of intervals represented
#	in the coordinate of a target sequence.
#	Then it concatenates the extracted sub-alignments
#	and calculates the score of the resulting 'spliced' alignment.
#
# INPUT: (\@pwaln, $indx_tgt, $frame_beg, \@aln_begs, \@aln_ends, \@regions, 
#	\%score_matrix, \@cmp_parameters), where
#	@pwaln = (\$qu_seq, \$sb_seq) is the sequences aligned;
#	$indx_tgt = -1, 0, and +1 if we use
#	the alignment frame, query sequence, and subject coordinates, respectively;
#	$frame_beg is the beginning coordinate of the alignment frame;
#	$aln_{begs/ends}[$m] is the beginning/ending coordinate of the sequence
#	${$pwaln[$m]} embeddded in the alignment;
#	@{$regions[$r]} = ($beg, $end) is the boundaries of the $r th region
#	in terms of the coordinate system specified by $indx_tgt;
#	@cmp_parameters = ($mtrx_delim, $gapopen, $gapext, $char_gap,
#			$lambda, $k_r2b, $char_mask);
#
# OUTPUT: (#{total_sites}, #{identical_res}, total_score, #{gaps}, #{res}_qu, #{res}_sb)
#	for the pairwise alignment with the flanking regions removed and
#	with the internal unspecified residues masked.
#
sub cnct_examine_tgt_regions_pwaln (\@$$\@\@\@\%\@) {

	my ($pwaln, $indx_tgt, $frame_beg, $aln_begs, $aln_ends, $regions,
		$score_matrix, $cmp_parameters) = @_;

	my ($mtrx_delim, $gopen, $gext, $char_gap, $lambda, $k_r2b, $char_mask)
		= @{$cmp_parameters};

	unless ((defined $pwaln) and (@{$pwaln}>1)) {
	    print STDERR "[ERROR 'cnct_examine_tgt_regions_pwaln'] The INPUT pairwise alignment is CRIPPLED !!\n";
	    return (undef undef, undef, undef, undef, undef);
	}
	(@{$regions}>0) or (return (0,0,0,0,0,0));

	my $left_bd = $regions->[0][0];
	my $right_bd = $regions->[$#{$regions}][1];
	my @extbd_rgs = ([$left_bd, $right_bd]);

	my $cp_quseq = ${$pwaln->[0]};
	my $cp_sbseq = ${$pwaln->[1]};
	my @cp_pwaln = (\$cp_quseq, \$cp_sbseq);

=for TESTING the script
	my $frame_end = $frame_beg - 1 + length ($cp_quseq);
	print STDERR "\n< Subroutine 'cnct_examine_tgt_regions_pwaln' >\n\n";
	print STDERR "{ Input algnment (frame = \[$frame_beg\,$frame_end\]) }\n\n";
	my @labels = ('query', 'sbjct');
	for (my $k=0; $k<2; $k++) { $labels[$k] .=  sprintf ("\[%s\,%s\]", $aln_begs->[$k], $aln_ends->[$k]); }
	pr_ODEN_input (@labels, @cp_pwaln, 60, *STDERR);
	my $str_regions = string_intvl_set (@{$regions}, 0, 1) ;
	print STDERR "\n{ target regions (index = $indx_tgt): ${$str_regions} }\n\n";
	print STDERR "  --- Now conduct 'extract_specregions_maln' ---\n\n";
=cut

	my ($sb_pwalns, $sbaln_frames, $sbaln_coords, $remnants)
		= extract_specregions_maln (@cp_pwaln, $indx_tgt, $frame_beg,
			@{$aln_begs}, @{$aln_ends}, @{$regions}, $char_gap);
			##@{$aln_begs}, @{$aln_ends}, @extbd_rgs, $char_gap);
			## --> REVISED on Wednesday, April 11, 2007 ##

	unless ((defined $sb_pwalns) and (@{$sb_pwalns}>0)) {
	    print STDERR "[ERROR 'cnct_examine_tgt_regions_pwaln'] The subroutine 'extract_specregions_maln' FAILED !\n";
	    return (undef, undef, undef, undef, undef, undef);
	}

	my $ct_pwalns = @{$sb_pwalns};

=for TESTING the script
	print STDERR "\n{ The extracted alignments }\n\n";
	for (my $r=0; $r<$ct_pwalns; $r++) {
	    my @labels2 = ('query', 'sbjct');
	    printf STDERR "( The $r th alignment (frame = \[%s\,%s\]) )\n\n",
		@{$sbaln_frames->[$r]};
	    for (my $k=0; $k<2; $k++) {
		$labels2[$k] .= sprintf ("\[%s\,%s\]", @{$sbaln_coords->[$r][$k]} );
	    }
	    pr_ODEN_input (@labels2, @{$sb_pwalns->[$r]}, 60, *STDERR);
	    print STDERR "\n";
	}
=cut

	my ($cnct_sb_pwalns, $cnct_sbaln_coords)
		= concatenate_alignments (@{$sb_pwalns}, @{$sbaln_coords}) ;

=for TESTING the script
	my $str_cnct_frames = string_intvl_set (@{$sbaln_frames}, 0, 1);
	print STDERR "{ Final concatenated alignment (frames = ${$str_cnct_frames}) }\n\n";
	my @labels3 = (
		'query' . ${$cnct_sbaln_coords->[0]},
		'sbjct' . ${$cnct_sbaln_coords->[1]}
		) ;
	pr_ODEN_input (@labels3, @{$cnct_sb_pwalns}, 60, *STDERR);
	print STDERR "\n";
=cut

	my $info_pwaln_score = score_pairalign (
		${$cnct_sb_pwalns->[0]}, ${$cnct_sb_pwalns->[1]},
		%{$score_matrix}, $mtrx_delim, $gopen, $gext, $char_gap) ;

	unless ((defined $info_pwaln_score) and (@{$info_pwaln_score}>5)) {
	    print STDERR "[ERROR 'cnct_examine_tgt_regions_pwaln'] 'score_pairailgn' FAILED !\n";
	    return (undef, undef, undef, undef, undef, undef);
	}

=for TESTING the subroutine
	print STDERR "\n === The subroutine 'cnct_examine_tgt_regions_pwaln' now ENDED ===\n\n";
=cut

	my ($totct_sites, $ct_res_qu, $ct_res_sb, $ct_id, $ct_gap, $tot_score)
		= @{$info_pwaln_score};

	return ($totct_sites, $ct_id, $tot_score, $ct_gap, $ct_res_qu, $ct_res_sb);

}		# End of the subroutine 'cnct_examine_tgt_regions_pwaln' #


1;	# a 'spell' which enables the module 'use'-able and 'require'-able #


