package MyNucSeq;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.3: Copyright (C) 2013 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyNucSeq.pm: This module performs variuos manipulation on nucleotide sequences.
#
# Revised on Aug 9 (Sat), 2003: I enables the subroutine "reverse_traslation" to return
#				the original length of the relevant cDNA.
#
# REVISED on Sep 21 (Tue), 2004 : The subroutine can now properly take account of the
#			existence of selenocysteins (coded with 'U').

use strict;
use warnings;
use Exporter;
#use MyMalnIO qw(pairwise_difference);

our @ISA = qw( Exporter);
our @EXPORT = qw( );
our @EXPORT_OK = qw(
		sub_sequence extract_multisites_sglseq
		retrieve_correspondence sequential_fastacmd
		batch_fastacmd sgl_fastacmd revcom
		dna2peptide translate_frame translate_codon_aln
		reverse_translation back_translate_pepseq reverse_translation_smart
		back_translate_pepseq_lenient reverse_translation_smart_lenient
		pepaln2codonaln pepaln2codonaln_lite pepaln2codonaln_rg_lite
		print_sequence print_subsequence print_fasta
		pr_multifasta_sbseqs mask_seq_regions pr_mask_multifasta
		);

use MyArrayManipu qw( merge_integer_intervals compare_interval_sets );
use MyFile_manipu qw( open_input_file );

sub print_sequence (\$$\*);

our %GENETIC_CODE = (
    
    'TCA' => 'S',    # Serine
    'TCC' => 'S',    # Serine
    'TCG' => 'S',    # Serine
    'TCT' => 'S',    # Serine
    'TCN' => 'S',    # Serine # ambiguous third base 
    'TCY' => 'S',    # Serine # ambiguous third base (pyrimidine)
    'TCR' => 'S',    # Serine # ambiguous third base (purine)
    'TTC' => 'F',    # Phenylalanine
    'TTT' => 'F',    # Phenylalanine
    'TTY' => 'F',    # Phenylalanine # ambiguous third base (pyrimidine)
    'TTA' => 'L',    # Leucine
    'TTG' => 'L',    # Leucine
    'TTR' => 'L',    # Leucine # ambiguous third base (purine )
    'TAC' => 'Y',    # Tyrosine
    'TAT' => 'Y',    # Tyrosine
    'TAY' => 'Y',    # Tyrosine # ambiguous third base (pyrimidine)
    #'TAA' => '_',    # Stop
    'TAA' => '*',    # Stop
    #'TAG' => '_',    # Stop
    'TAG' => '*',    # Stop
    'TAR' => '*',    # Stop # ambiguous third base (purine)
    'TGC' => 'C',    # Cysteine
    'TGT' => 'C',    # Cysteine
    'TGY' => 'C',    # Cysteine # ambiguous third base (pyrimidine)
    #'TGA' => '_',    # Stop
    'TGA' => '*',    # Stop
    'TGG' => 'W',    # Tryptophan
    'CTA' => 'L',    # Leucine
    'CTC' => 'L',    # Leucine
    'CTG' => 'L',    # Leucine
    'CTT' => 'L',    # Leucine
    'CTN' => 'L',    # Leucine  # ambiguous 3rd base
    'CTR' => 'L',    # Leucine  # ambiguous 3rd base (purine)
    'CTY' => 'L',    # Leucine  # ambiguous 3rd base (pyrimidine)
    'CCA' => 'P',    # Proline
    'CCC' => 'P',    # Proline
    'CCG' => 'P',    # Proline
    'CCT' => 'P',    # Proline
    'CCN' => 'P',    # Proline # ambiguous 3rd base
    'CCY' => 'P',    # Proline # ambiguous 3rd base (pyrimidine)
    'CCR' => 'P',    # Proline # ambiguous 3rd base (purine)
    'CAC' => 'H',    # Histidine
    'CAT' => 'H',    # Histidine
    'CAY' => 'H',    # Histidine # ambiguous 3rd base (pyrimidine)
    'CAA' => 'Q',    # Glutamine
    'CAG' => 'Q',    # Glutamine
    'CAR' => 'Q',    # Glutamine # ambiguous 3rd base (purine)
    'CGA' => 'R',    # Arginine
    'CGC' => 'R',    # Arginine
    'CGG' => 'R',    # Arginine
    'CGT' => 'R',    # Arginine
    'CGN' => 'R',    # Arginine # ambiguous 3rd base
    'CGR' => 'R',    # Arginine # ambiguous 3rd base (purine)
    'CGY' => 'R',    # Arginine # ambiguous 3rd base (pyrimidine)
    'ATA' => 'I',    # Isoleucine
    'ATC' => 'I',    # Isoleucine
    'ATT' => 'I',    # Isoleucine
    'ATY' => 'I',    # Isoleucine # ambiguous 3rd base (pyrimidine)
    'ATG' => 'M',    # Methionine
    'ACA' => 'T',    # Threonine
    'ACC' => 'T',    # Threonine
    'ACG' => 'T',    # Threonine
    'ACT' => 'T',    # Threonine
    'ACN' => 'T',    # Threonine # ambiguous 3rd base
    'ACY' => 'T',    # Threonine # ambiguous 3rd base (pyrimidine) 
    'ACR' => 'T',    # Threonine # ambiguous 3rd base (purine)
    'AAC' => 'N',    # Asparagine
    'AAT' => 'N',    # Asparagine
    'AAY' => 'N',    # Asparagine # ambiguous 3rd base (pyrimidine)
    'AAA' => 'K',    # Lysine
    'AAG' => 'K',    # Lysine
    'AAR' => 'K',    # Lysine # ambiguous 3rd base (purine)
    'AGC' => 'S',    # Serine
    'AGT' => 'S',    # Serine
    'AGY' => 'S',    # Serine # ambiguous 3rd base (pyrimidine)
    'AGA' => 'R',    # Arginine
    'AGG' => 'R',    # Arginine
    'AGR' => 'R',    # Arginine # ambiguous 3rd base (purine)
    'GTA' => 'V',    # Valine
    'GTC' => 'V',    # Valine
    'GTG' => 'V',    # Valine
    'GTT' => 'V',    # Valine
    'GTN' => 'V',    # Valine # ambigous 3rd base
    'GTY' => 'V',    # Valine # ambigous 3rd base (pyrimidine) 
    'GTR' => 'V',    # Valine # ambigous 3rd base (purine)
    'GCA' => 'A',    # Alanine
    'GCC' => 'A',    # Alanine
    'GCG' => 'A',    # Alanine
    'GCT' => 'A',    # Alanine
    'GCN' => 'A',    # Alanine # ambigous 3rd base
    'GCY' => 'A',    # Alanine # ambigous 3rd base (pyrimidine)
    'GCR' => 'A',    # Alanine # ambigous 3rd base (purine)
    'GAC' => 'D',    # Aspartic Acid
    'GAT' => 'D',    # Aspartic Acid
    'GAY' => 'D',    # Aspartic Acid # ambiguous 3rd base (pyrimidine)
    'GAA' => 'E',    # Glutamic Acid
    'GAG' => 'E',    # Glutamic Acid
    'GAR' => 'E',    # Glutamic Acid # ambigous 3rd base (purine)
    'GGA' => 'G',    # Glycine
    'GGC' => 'G',    # Glycine
    'GGG' => 'G',    # Glycine
    'GGT' => 'G',    # Glycine
    'GGN' => 'G',    # Glycine # ambiguous 3rd base (pyrimidine)
    'GGY' => 'G',    # Glycine # ambiguous 3rd base (purine)
    'GGR' => 'G',    # Glycine # ambiguous 3rd base
    '---' => '-'     # codon gap
    );


# sub_sequence:
#	This subroutine just extracts a sub-sequence of the input sequence.
#
# INPUT: (\$in_seq, $start_in, $beg_sbseq, $end_sbseq, $fl_stop), where
#	$in_seq is the input sequence that begins with the coordinate = $start_in;
#	$beg_sbseq and $end_sbseq are the coordinates of the start and end positions
#	of the subsequences to be extracted, in the same frame as given by $start_in;
#	$fl_stop = true if you have to consider the possibility of missing last stop character.
#	         = false otherwise
#
# OUTPUT: \$sbseq, which is the subsequence extracted.
#
sub sub_sequence (\$$$$$) {

	my ($in_seq, $start_in, $beg_sbseq, $end_sbseq, $fl_stop) = @_;

	if ($beg_sbseq < $start_in) {
	    print STDERR "[ERROR 'sub_sequence'] The start of the subsequence ($beg_sbseq) is UPSTREAM of the start of the input ($start_in) !\n";
	    return undef;
	}
	if ($beg_sbseq > $end_sbseq) {
	    print STDERR "[ERROR 'sub_sequence'] The start of the subsequence ($beg_sbseq) is DOWNSTREAM of its end ($end_sbseq) !\n";
	    return undef;
	}

	my $cp_inseq = ${$in_seq};
	my $end_inseq = $start_in + length ($cp_inseq) - 1;
	if ($end_sbseq > $end_inseq) {
	    if ($fl_stop and ($end_sbseq == $end_inseq + 1)) {
		$end_inseq++;
		$cp_inseq .='*';
	    } else {
		print STDERR "[ERROR 'sub_sequence'] The end of the subsequence ($end_sbseq) is DOWNSTREAM of the end of the input ($end_inseq) !\n";
		return undef;
	    }
	}
	
	($beg_sbseq>$start_in) and substr ($cp_inseq, 0, $beg_sbseq - $start_in, ''); # Remove unnecessary upstream region #

	my $sbseq = substr ($cp_inseq, 0, $end_sbseq - $beg_sbseq + 1);

	return \$sbseq;

}		#  End of the subroutine 'sub_sequence' #

# extract_multisites_sglseq:
#	This subroutine extracts multiple sites from a single sequence
#	embedded in an alignment.
#	We ASSUME that the coordinates in @framecoords are arranged in numerically
#	increasing order.
#
# INPUT: (\$seq_embd, \@framecoords, $coord_base), where
#	$seq_embd is a sequence possibly embedded in an alignment;
#	@framecoords gives the alignment frame coordinate (instead of the residue number
#	or sequence coordinate (not counting gaps));
#	$coord_base gives the base (or the starting coordinate)
#	of the aligment coordinate system that @framecoords follow.
#	($coord_base = 1 for 1-based, 0 for 0-based).
#
# OUTPUT: \@extracted_sites,  where
#	$extracted_sites[$k] is the $framecoords[$k] th site in the alignment.
#
#	The subroutine returns undef if some process failed.
#
sub extract_multisites_sglseq (\$\@$) {

	my ($seq_embd, $framecoords, $coord_base) = @_;

	my $ct_coords = @{$framecoords};
	($ct_coords>0) or (return []);
	my $aln_len = length ${$seq_embd};

	my ($beg_frame, $end_frame) = @{$framecoords}[0, $ct_coords-1];

		## A BUG FIXED on July 2 (Wed), 2008 ##
	unless ((0 <= $beg_frame) and ($end_frame <= $aln_len + $coord_base -1)) {
	##unless ((0 <= $beg_frame) and ($end_frame < $aln_len)) {
		print STDERR "[ERROR 'extract_multisites_sglseq'] ILLEGAL INPUT: (beg_frame, end_frame) = ($beg_frame, $end_frame) vs alignment length = $aln_len : $!\n";
		return undef;
	}
	my $cp_seq_embd = ${$seq_embd};

	my @extracted_sites = ();

		## REVISED on June 28 (Sat), 2008 ##
		## (It now can cope with more general cases, instead of only 0-based cases.) ##
	my $last_coord = $coord_base -1;
	##my $last_coord = -1;
	for (my $k=0; $k<$ct_coords; $k++) { # The main for-loop #
	    my $pres_coord = $framecoords->[$k];


		## A BUG FIXED on July 2 (Wed), 2008 ##
	    unless ($pres_coord <= $aln_len + $coord_base -1) {
	    ##unless ($pres_coord < $aln_len) {
		print STDERR "[ERROR 'extract_multisites_sglseq'] Found an alignment COORDINATE ($pres_coord) NOT LESS THAN the alignment length ($aln_len) : $! !\n";
		return undef;
	    }
	    my $left_margin = $pres_coord-1 -$last_coord;
	    unless ($left_margin >=0) {
		print STDERR "[ERROR 'extract_multisites_sglseq'] NEGATIVE LEFT MARGIN ($left_margin):  present coordinate = $pres_coord vs last coordinate = $last_coord : $! !\n";
		return undef;
	    }
	    ($left_margin>0) and (substr ($cp_seq_embd, 0, $left_margin, '') );
	    my $site = substr ($cp_seq_embd, 0, 1, '');
	    push @extracted_sites, $site;

	    $last_coord = $pres_coord;
	}		# End of the main for-loop #

	return \@extracted_sites;

}		# End of the subroutine 'extract_multisites_sglseq' #


# retrieve_correspondence: This subroutine reads in correspondence table between peptides and cDNAs,
#			and stores the information into  a hash. 
#
sub retrieve_correspondence (\$\%$) {

	my ($rf_intable, $rf_pep2cdna, $pre_species) = @_;

	my $filename = ` ls ${$rf_intable}* `;
	(defined $filename) or die "File ${$rf_intable}\* does NOT exist: $!\n";

	if ($filename =~ m/\.gz$/ ) {

		open (IN, " gunzip -c ${$rf_intable}.gz | ") or die "\nCannot folk or open ${$rf_intable}.gz: $!\n";
	} else {
		 open (IN, "< ${$rf_intable}") or die "\nCannot open ${$rf_intable}: $!\n";
	}

	while (my $in = <IN>) {

		#my ($pep_id) = ( $in =~ m/^($pre_species\d+)/o );
		my ($pep_id, $prefix) = ( $in =~ m/^(([A-Z]+)\d+)/ );
		(defined $pep_id) or next;
		($prefix eq  $pre_species) or next;
		my @fields = split ' ', $in;
		my @cdna_ids = @fields[5..$#fields];
		$rf_pep2cdna->{$pep_id} = \@cdna_ids;
	}
	close (IN);

}				# end of the subroutine retrieve_correspondence #


# sequential_fastacmd:
#	This subroutine conducts 'fastacmd' in the BLAST package
#	in a sequential manner to retrieve a set of sequences.
#	It carefully checks the status of the output file.
#
#	NOTE:
#	BE SURE TO CHECK the availability of $exec_fastacmd and $db_seqs
#	before using this subroutine.
#
# INPUT: (\$exec_fastacmd, \$outfile, \$db_seqs, \@list_seqs, $fl_update), where
#	@list_seqs stores the IDs of the sequences to be retrieved;
#	$fl_update = 1 if you construct the output file from scratch,
#	    = 0 if you just append the sequences to a pre-exinting file, if any.
#
# OUTPUT: #{headers} if the process succeeded,
#	  0 if it failed.
#
sub sequential_fastacmd (\$\$\$\@$) { 

	my ($exec_fastacmd, $outfile, $db_seqs, $list_seqs, $fl_update) = @_;

	my $ct_seqs = @{$list_seqs};

	unless ($ct_seqs>0) {
	   print STDERR "[ERROR 'sequential_fastacmd'] The list contains NO SEQUENCES to be retrieved !\n";
	   return 0;
	}

	if ($fl_update and ( -f ${$outfile} )) {
	    unless (0 == system ("mv ${$outfile} ${$outfile}.bk") ) {
		print STDERR "[ERROR 'sequential_fastacmd'] FAILED to rename the pre-existing sequence file '${$outfile}' by appending the suffix '.bk': $? !\n";
		return 0;
	    }
	}

	for (my $m=0; $m<$ct_seqs; $m++) { # The main for-loop #
	    my $seqid = $list_seqs->[$m];

	    unless (0 == system ("${$exec_fastacmd} -d ${$db_seqs} -s $seqid >> ${$outfile}") ) {
		print STDERR "[ERROR 'sequential_fastacmd'] FAILED to retrieve the sequence '$seqid' via 'fastacmd': $? !\n\n"; 
		return 0;
	    }
	}			# End of the main for-loop #

	unless (( -f ${$outfile} ) and ( -r ${$outfile} )) {
	    print STDERR "[ERROR 'sequential_fastacmd'] CANNOT FIND a readable sequence file '${$outfile}': $? !\n\n";
	    return 0;
	}


		# Check the number of sequences retrieved #
	local $/ = "\n"; 
	my $fh_tmp = open_input_file (${$outfile});
	my $ct_header = 0;
	while (my $in = <$fh_tmp>) { ($in =~ m/^\>/) and ($ct_header++); }
	close ($fh_tmp);

	if ($fl_update) {
	    unless ($ct_header == $ct_seqs) {
		print STDERR "[ERROR 'sequential_fastacmd'] The number of headers ($ct_header) CLASHES with that of the sequences to be retrieved ($ct_seqs) : $! !\n";
		return 0;
	    }
	}

	return $ct_header;

}		# End of the subroutine 'sequential_fastacmd' #


# batch_fastacmd : This subroutine fetches sequences specified by a list of sequence IDs,
#			using the NCBI software "fastacmd".
#
sub batch_fastacmd (\@\$) {

	my ($rf_seqIDs, $rf_database) = @_;
	my @list_cdna = ();

	foreach my $seqID (@{$rf_seqIDs}) {

		my $sequence = `fastacmd -d ${$rf_database}  -s $seqID `;
		#my ($seq_ID) = ($sequence =~ m/^\>(\S+)/ );
		#(defined $sequence) or next;
		($sequence) or next;
		$sequence =~ s/^[^\n]+\n//;
		$sequence =~ s/\s+//g; 	# remove all while spaces #

		#my $seq_len = length $sequence;
		#print "Subject:\t$seqID(length=$seq_len)\n\n";

		push @list_cdna, [$seqID, \$sequence];
	}

	return \@list_cdna;

}			# end of the subroutine batch_fastacmd #


# sgl_fastacmd: This subroutine's main purpose is to fetch a single sequence via 'fastacmd'
#		while saving the memory space as much as possible.
#
# OUTPUT: (\$sequence, $length, \$headeer_description)
#
sub sgl_fastacmd ($\$\$) {

	my ($seqid, $database, $exec_dir) = @_;
	local $/ = "\n";

		## REVISED on Monday, April 23, 2007 ##
	my $fh;
	unless ( open ($fh, "${$exec_dir}/fastacmd -d ${$database} -s $seqid |") ) {
	    print STDERR "\n[ERROR 'sgl_fastacmd'] Cannot fork '${$exec_dir}/fastacmd -d ${$database} -s $seqid': $!\n";
	    return (undef, undef, undef);
	}
	##open (my $fh, "${$exec_dir}/fastacmd -d ${$database} -s $seqid |")
	##	or die "\nCannot fork ${$exec_dir}/fastacmd -d ${$database} -s $seqid: $!\n";

	my $header = <$fh>;
	until (!(defined $header) or ($header =~ m/^\>/ ) ) { $header = <$fh>; }

	(defined $header) or ($header = 'NO SEQUENCE');

	my ($seqid2) = ($header =~ m/^\>(\S+)/ );
	unless (defined $seqid2) {
		print STDERR "[ERROR 'sgl_fastacmd'] Unusual output of 'fastacmd'?: $header \n";
		close ($fh);
		return (undef, undef, \$header); 
	}
	$seqid2 =~ s/^lcl\|// ;
	unless ($seqid eq $seqid2) {
		print STDERR "[ERROR 'sgl_fastacmd'] Discrepancy in the input ID ($seqid) and the output ID ($seqid2).\n";
		close ($fh);
		return (undef, undef, \$header);
	}

	(undef, my $desc) = split /\s+/, $header, 2;

	my $seq = '';
	my $len = 0;
	while (my $in = <$fh> ) {

		$in =~ tr/\ \t\n//d ;
		$seq .= $in;
		$len += length ($in);
	}
	close ($fh);

	return (\$seq, $len, \$desc);

}		# end of the subroutine sgl_fastacmd #


# revcom: A subroutine to compute the reverse complement of DNA sequence
#
sub revcom (\$) {

    my($rf_dna) = @_;

    # First reverse the sequence
    my $revcom = reverse ${$rf_dna};

    # Next, complement the sequence, dealing with upper and lower case
    # A->T, T->A, C->G, G->C, R(purine) -> Y(pyrimidine), Y->R
    #$revcom =~ tr/ACGTacgt/TGCAtgca/;
    $revcom =~ tr/ACGTRYacgtry/TGCAYRtgcayr/;

    return \$revcom;
}

# dna2peptide: A subroutine to translate DNA sequence into a peptide
#
sub dna2peptide  (\$) {

    my($rf_dna) = @_;

		# copy the original string,
		# and convert lower-case letters into capital ones #
	my $cp_dna = uc ${$rf_dna};
		# Initialize variables
	my $protein = '';
	my $times = int (length($cp_dna)/3);

    # Translate each three-base codon to an amino acid, and append to a protein 
    #for(my $i=0; $i < (length($cp_dna) - 2) ; $i += 3) {
        #$protein .= codon2aa( substr($cp_dna,$i,3) );
    #}

	for (my $i=0; $i<$times; $i++) {

		my $codon = substr( $cp_dna, 0 , 3, '');

		#$codon = uc $codon; # convert lower-case letters to capital ones #
		my $aa = $GENETIC_CODE{$codon};
		$protein .= (defined $aa) ? $aa : 'X';
	}

    return \$protein;

}			# end of the subroutine dna2peptide #


# translate_frame: A subroutine to translate a frame of DNA (or RNA )
#
sub translate_frame (\$$$) {

	my($rf_seq, $start, $end) = @_;

	my $cp_seq = ${$rf_seq}; # copy the original string #

	
	#$cp_seq =~ s/\s+//g;	# remove white spaces #
	$cp_seq =~ tr/Uu/Tt/;	# convert an RNA sequence to an equvalent DNA sequence #

	my $len = length ($cp_seq);

    # To make the subroutine easier to use, you won't need to specify
    #  the end point-it will just go to the end of the sequence
    #  by default.
    #unless($end) {
    #    $end = length($cp_seq);
    #}

	($start > 1) and substr ($cp_seq, 0, $start -1, ''); # trim the extra foreend #

	if ($end and ($end < $len) and ($end >= $start) ) {

		$cp_seq = substr($cp_seq, 0, $end - $start +1);

	}


    # Finally, calculate and return the translation
        #return dna2peptide ( substr ( $seq, $start - 1, $end -$start + 1) );
        return dna2peptide ($cp_seq);

}				# end of the subroutine translate_frame #


# translate_codon_aln: This subroutine translates codon alignments into peptide alignments.
#			Whether all the padded sequences have the same length or not,
#			The subroutine translates each of the sequences as long  as possible
#			from the alignment start to the end.
#
# PREREQUISITE: The codon alignments are all "in phase".
#
sub translate_codon_aln (\@) {

	my ($codon_aln) = @_;

	my $ct_seq = @{$codon_aln};
	$ct_seq or (return []);

	my @translated = ();

	foreach my $nuc_seq (@{$codon_aln}) {
		my $end = 3 * int (length (${$nuc_seq}) / 3) ;
		my $pep_seq = translate_frame (${$nuc_seq}, 1, $end);
		push @translated, $pep_seq;
	}

	return \@translated;

}		# end of the subroutine translate_codon_aln #


# rm_ambig_sites_aa: This subroutine removes sites including ambiguous sites
#		from the list output by "pairwise_difference".
#
sub rm_ambig_sites_aa (\@) {

	my ($rf_difflist) = @_;

	@{$rf_difflist} or (return $rf_difflist);

	my @output = ();

	foreach my $rf_site (@{$rf_difflist}) {

		("$rf_site->[1]$rf_site->[2]" =~ m/BUZX/i ) or (push @output, $rf_site) ;
	}

	return \@output;
}			# end of the subroutine rm_ambig_sites #


# exact_match: This subroutine compares two sequences,
#		and returns 1 if they exactly match (except ambiguities),
#		and returns 0 otherwise .
#
sub exact_match (\$\$) {

	my ($rf_seq1, $rf_seq2) = @_;

	my $seq1 = uc ${$rf_seq1};	# a minor revision on Sep 3 (Wed), 2003 #
	my $seq2 = uc ${$rf_seq2};	# a minor revision on Sep 3 (Wed), 2003 #
	my $len1 = length ($seq1);
	my $len2 = length ($seq2);

	($len1 == $len2) or (return 0);

	for (1..$len1) {

		my $char1 = substr($seq1, 0, 1, '');
		my $char2 = substr($seq2, 0, 1, '');

		if ($char1 eq $char2) {  next; 
		} elsif ( ($char1 eq '*') and ( uc($char2)  eq 'U') ) { next;
		} elsif ( ( uc($char1) eq 'U') and ($char2 eq '*') ) { next;
		} elsif ( "$char1$char2" =~ m/[BZX]/) { next;
		} else {
			return 0;
		} 
		 

	}

	return 1;

}			# end of the subroutine exact_match #


# embed_peptide: Given a peptide sequence and
#		 a (translated) candidate of its origainal cDNA sequence (and a frame),
#		this subroutine seeks to embed the peptide into the translated cDNA,
#		and returns the start and end positions of the coding region
#		in the UNTRANSLATED cNDA.
#
# REVISED on Sep 21 (Tue), 2004 : The subroutine can now properly take account of the
#			existence of selenocysteins (coded with 'U').
#
sub embed_peptide(\$\$$) {

	my ($rf_peptide, $rf_tr_cdna, $frame) = @_;

	$frame or ($frame = 1);
	my $cp_pep = ${$rf_peptide};
	my $len_pep = length ($cp_pep);
	my $cp_tr_cdna = ${$rf_tr_cdna};
	#my $len_tr_cdna = length ($cp_tr_cdna);
	my $len_margin  = length ($cp_tr_cdna)  - $len_pep;
	($len_margin < 0) and (return (undef, undef) );

		# we require the beginning position to be positive #
	my $begin = ($frame < 0) ? -$frame: $frame;

	my $subject = substr($cp_tr_cdna, 0, $len_pep, '');
	my $fl_succeed = 0;
	#my $ct_stop = ($cp_pep =~ tr/Xx*// );
	my $ct_stop = ($cp_pep =~ tr/XxuU*// ); # REVISED on  Sep 21 (Tue), 2004 #

	while(1) {			# an infinite whlie loop #

		my $sb_ct = ($subject =~ tr/*// );
	   if ($sb_ct <= $ct_stop ) {

		if ( exact_match ($cp_pep, $subject) ) { 

			$fl_succeed = 1;
			last;
		}

		#my @dummy = ();
		#my ($rf_difflist , undef) = pairwise_difference ($cp_pep, $subject, @dummy);
		#unless (@{$rf_difflist}) {
		#  	$fl_succeed = 1;
		#	last;
		#}

		#$rf_difflist = rm_ambig_sites_aa(@{$rf_difflist});
		#unless (@{$rf_difflist}) {
		#	 $fl_succeed = 1;
		#	 last;
		#}
	    }

		$len_margin or last;

			# advance the frame by 3 #
		substr($subject, 0, 1, '');
		$subject .= substr($cp_tr_cdna, 0, 1, '');
		$begin += 3;
		$len_margin--;

	}				# end of the infinite while loop #

	if ($fl_succeed) {
		my $end = $begin + 3* $len_pep-1 ;
		return ($begin, $end);
	} else {
		return (undef, undef);
	}

}			# end of the subroutine embed_peptide #

# reverse_translation: Given a peptide sequence and a set of candidates for its cDNA counterpart,
#			this subroutine examines which is the authentic counterpart,
#			and returns its ID and "reverse translated" sequences.
#
# INPUT: \$pep_seq, [ [cdnaID, \$cdna_seq]_1, [cdna_ID, \$cdna_seq]_2, ... ],
# OUTPUT: (authentic_cdna_ID, \$reverse_translated_seq, begin_pos, end_pos).
#
# (begin_pos < 0 and end_pos < 0 when reverse complement was used )
#
sub reverse_translation (\$\@) {

	my ($rf_pep, $rf_cdna_list) = @_;

	@{$rf_cdna_list} or (return (undef, undef, undef, undef) );

	foreach my $i (1..3) {			# outer foreach #
		foreach my $rf_cdna (@{$rf_cdna_list}) {# inner foreach #

			my $rf_tr_cdna = translate_frame (${$rf_cdna->[1]}, $i, undef);
			my ($begin, $end) = embed_peptide(${$rf_pep},${$rf_tr_cdna},$i);
			(defined $begin ) or next;

			my $orig_len = (${$rf_cdna->[1]} =~ tr/a-zA-Z// );
			my $sub_cdna = substr (${$rf_cdna->[1]}, 0, $end);
			($begin >  1) and  substr ($sub_cdna, 0, $begin -1, '');

			return ($rf_cdna->[0], \$sub_cdna, $begin, $end, $orig_len);

		}					# end of the inner foreach #
	}					# end of the outer foreach #

	# if the true cnda has not been found yet, try the reverse translation #

	foreach my $i (1..3) {			# outer foreach #
		foreach my $rf_cdna (@{$rf_cdna_list}) {# inner foreach #

			my $revcom = revcom (${$rf_cdna->[1]});
			my $rf_tr_cdna = translate_frame (${$revcom}, $i, undef);
			my ($begin, $end) = embed_peptide(${$rf_pep}, ${$rf_tr_cdna}, $i);
			(defined $begin) or next;

			my $orig_len = (${$rf_cdna->[1]} =~ tr/a-zA-Z// ) ;
			my $sub_cdna = substr (${$revcom}, 0, $end);
			($begin >  1) and  substr ($sub_cdna, 0, $begin -1, '');

			return ($rf_cdna->[0], \$sub_cdna, -$begin, -$end, $orig_len);

		}					# end of the inner foreach #
	}					# end of the outer foreach #

	return (undef, undef, undef, undef, undef);

}			# end of the subroutine reverse_translation #


# back_translate_pepseq: This subroutine tries to 'back-translate' a peptide sequence (the 0th argument)
#			by using a cDNA sequence received as the 1st argument.
#
# INPUT: (\$pep_sequence, \$cdna_sequence)
# OUTPUT: (\$back_translated_peptide_sequence, start_coding_on_cDNA, end_coding_on_cDNA, list_unusual).
#
# PREREQUISITE: The peptide and cDNA sequences do NOT contain extra characters,
#		like white spaces or anything other than alphabets, stop '*', and gap '-' .
#		
# CAUTION: Although this subroutine leaves the input peptide sequence intact,
#	 it DOES BITE OFF the left end of the input cDNA sequence.
#
sub back_translate_pepseq (\$\$) {

	my ($pep_seq, $cdna_seq) = @_;

	my $pep_len = length (${$pep_seq});
	my $max_start = length (${$cdna_seq}) - 3 * $pep_len + 1; 

	OUTER:
	for (my $start=1; $start <= $max_start; $start++) {

		my $cp_cdna = ${$cdna_seq};

			# A revision accompanying the bug fix
			# around the rear end of the loop,
			# both made on Tuesday, July 4, 2006
		($start>1) and substr ($cp_cdna, 0, $start-1, '');

		my $cp_pep = ${$pep_seq};
		my $bt_pep = '';
		my @unusual = (); # element = [pos_aa, aa, pos_codon_1st_char, codon]

		for (my $l = 0; $l < $pep_len; $l++) { 

			my $aa = uc substr ($cp_pep, 0, 1, '');
			my $codon = uc substr ($cp_cdna, 0, 3, '');
			$codon =~ tr/U/T/;

			my $coded_aa = $GENETIC_CODE{$codon};
			(defined $coded_aa) or ($coded_aa = 'X');

			if ($coded_aa eq $aa) {
				$bt_pep .= $codon;

			} elsif ( ($aa eq 'U') and ($codon eq 'TGA') ) { 
				$bt_pep .= $codon;
				push @unusual, [$l+1, $aa, $start + 3 * $l, $codon];

			} elsif ( ($aa eq 'X') and ($codon =~ m/[YRN]$/ ) ) { # withhold the judgement #
				$bt_pep .= $codon;
				push @unusual, [$l+1, $aa, $start + 3 * $l, $codon];

			} elsif ( ($codon =~ m/[KMSWBDHV]$/ ) and ( $aa =~ m/[LIVSPTARG]/ ) ) { # withhold the judgement #
				$bt_pep .= $codon;
				push @unusual, [$l+1, $aa, $start + 3 * $l, $codon];

			} elsif ( ($codon =~ m/^[^ATGC\-]/ ) and ($aa =~ m/[LRBZ]/ ) ) { # withhold the judgement #
				$bt_pep .= $codon;
				push @unusual, [$l+1, $aa, $start + 3 * $l, $codon];

			} else {  # failure #
					# A bug-fix on Tuesday, July 4, 2006
				##substr (${$cdna_seq}, 0, 1, '');  # bite off the left end
				next OUTER;
			}
		}

		## succeeded in the back-translation ##

		return (\$bt_pep, $start, $start + 3 * $pep_len -1, \@unusual);
		
	}

		## Failed to back-translate the cDNA ##

	return (undef, undef, undef, undef);

}		# end of the subroutine back_translate_pepseq #


# reverse_translation_smart: This subroutine does essentially the same job as 'reverse_translation',
#			in a cleverer manner.
#
# reverse_translation: Given a peptide sequence and a set of candidates for its cDNA counterpart,
#			this subroutine examines which is the authentic counterpart,
#			and returns its ID and "reverse translated" sequences.
#
# INPUT: \$pep_seq, [ [cdnaID, \$cdna_seq]_1, [cdna_ID, \$cdna_seq]_2, ... ],
# OUTPUT: (authentic_cdna_ID, \$reverse_translated_seq, begin_pos, end_pos).
#
# (begin_pos < 0 and end_pos < 0 when reverse complement was used )
#
sub reverse_translation_smart (\$\@) {

	my ($rf_pep, $rf_cdna_list) = @_;

	@{$rf_cdna_list} or (return (undef, undef, undef, undef, undef, undef) );

	foreach my $rf_cdna (@{$rf_cdna_list} ) {

		my $cdna_id = $rf_cdna->[0];
		(my $cp_cdna_seq = ${$rf_cdna->[1]}) =~ tr/\ \t\n//d; # remove white spaces
		my $orig_len = length $cp_cdna_seq ;

		my ($rf_btpep, $begin, $end, $unusual)
			 = back_translate_pepseq (${$rf_pep}, $cp_cdna_seq);

		(defined $rf_btpep) and
			(return  ($cdna_id, $rf_btpep, $begin, $end, $orig_len, $unusual) );
	}

	# if a cnda counterpart has not been found yet, try the reverse translation #

	foreach my $rf_cdna (@{$rf_cdna_list}) {

		my $cdna_id = $rf_cdna->[0];
		my $revcom =  revcom ( ${$rf_cdna->[1]});
		${$revcom} =~ tr/\ \t\n//d;
		my $orig_len = length ${$revcom};

		my ($rf_btpep, $begin, $end, $unusual)
			 = back_translate_pepseq (${$rf_pep}, ${$revcom}); 

		(defined $rf_btpep) and 
			( return ($cdna_id, $rf_btpep, -$begin, -$end, $orig_len, $unusual) );
	}

		# found no cDNA counterpart #

	return (undef, undef, undef, undef, undef, undef) ;

}		# end of the subroutine reverse_translation_smart #


# back_translate_pepseq_lenient:
#	This subroutine tries to 'back-translate' a peptide sequence (the 0th argument)
#	by using a cDNA sequence received as the 1st argument.
#	The subroutine is a revised version of 'back_translate_pepseq'
#	forged by making some of the conditions less strict.
#	(1: Selenocystein vs '*', 2: Tryptophan vs '*')
#
# INPUT: (\$pep_sequence, \$cdna_sequence)
# OUTPUT: (\$back_translated_peptide_sequence, start_coding_on_cDNA, end_coding_on_cDNA, list_unusual).
#
# PREREQUISITE: The peptide and cDNA sequences do NOT contain extra characters,
#		like white spaces or anything other than alphabets, stop '*', and gap '-' .
#		
# CAUTION: Although this subroutine leaves the input peptide sequence intact,
#	 it DOES BITE OFF the left end of the input cDNA sequence.
#
sub back_translate_pepseq_lenient (\$\$) {

	my ($pep_seq, $cdna_seq) = @_;

	my $pep_len = length (${$pep_seq});
	my $max_start = length (${$cdna_seq}) - 3 * $pep_len + 1; 

	OUTER:
	for (my $start=1; $start <= $max_start; $start++) {

		my $cp_cdna = ${$cdna_seq};

			# A revision accompanying the bug fix
			# around the rear end of the loop,
			# both made on Tuesday, July 4, 2006
		($start>1) and substr ($cp_cdna, 0, $start-1, '');

		my $cp_pep = ${$pep_seq};
		my $bt_pep = '';
		my @unusual = (); # element = [pos_aa, aa, pos_codon_1st_char, codon]

		for (my $l = 0; $l < $pep_len; $l++) { 

			my $aa = uc substr ($cp_pep, 0, 1, '');
			my $codon = uc substr ($cp_cdna, 0, 3, '');
			$codon =~ tr/U/T/;

			my $coded_aa = $GENETIC_CODE{$codon};
			(defined $coded_aa) or ($coded_aa = 'X');

			if ($coded_aa eq $aa) {
				$bt_pep .= $codon;

			} elsif ( ($aa eq 'U') and ($codon eq 'TGA') ) { # REVISED on Dec 19 (Sun), 2004 #
			##} elsif ( ($aa eq 'U') and ($coded_aa eq '*') ) {  # -- returned to the original on the same day #
				$bt_pep .= $codon;
				push @unusual, [$l+1, $aa, $start + 3 * $l, $codon];

			##} elsif ( ($aa eq 'W') and ($coded_aa eq '*') ) {  # ADDED on Dec 19 (Sun), 2004 #
			##	$bt_pep .= $codon; 				# --- disabled again on the same day #
			##	push @unusual, [$l+1, $aa, $start + 3 * $l, $codon];

			} elsif ( ($aa eq 'X') and ($codon =~ m/[YRN]$/ ) ) { # withhold the judgement #
				$bt_pep .= $codon;
				push @unusual, [$l+1, $aa, $start + 3 * $l, $codon];

			} elsif ( ($codon =~ m/[KMSWBDHV]$/ ) and ( $aa =~ m/[LIVSPTARG]/ ) ) { # withhold the judgement #
				$bt_pep .= $codon;
				push @unusual, [$l+1, $aa, $start + 3 * $l, $codon];

			} elsif ( ($codon =~ m/^[^ATGC\-]/ ) and ($aa =~ m/[LRBZ]/ ) ) { # withhold the judgement #
				$bt_pep .= $codon;
				push @unusual, [$l+1, $aa, $start + 3 * $l, $codon];

			} else {  # failure #
					# A bug fixed on Tuesday, July 4, 2006 #
				##substr (${$cdna_seq}, 0, 1, '');  # bite off the left end
				next OUTER;
			}
		}

		## succeeded in the back-translation ##

		return (\$bt_pep, $start, $start + 3 * $pep_len -1, \@unusual);
		
	}

		## Failed to back-translate the cDNA ##

	return (undef, undef, undef, undef);

}		# end of the subroutine back_translate_pepseq_lenient #


# reverse_translation_smart_lenient:
#	This subroutine is a revised version of 'reverse_translation_smart',
#	and makes some of the conditions less strict than the predecessor.
#
# reverse_translation_smart:
#	This subroutine does essentially the same job as 'reverse_translation',
#	in a cleverer manner.
#
# reverse_translation: Given a peptide sequence and a set of candidates for its cDNA counterpart,
#			this subroutine examines which is the authentic counterpart,
#			and returns its ID and "reverse translated" sequences.
#
# INPUT: \$pep_seq, [ [cdnaID, \$cdna_seq]_1, [cdna_ID, \$cdna_seq]_2, ... ],
# OUTPUT: (authentic_cdna_ID, \$reverse_translated_seq, begin_pos, end_pos).
#
# (begin_pos < 0 and end_pos < 0 when reverse complement was used )
#
sub reverse_translation_smart_lenient(\$\@) {

	my ($rf_pep, $rf_cdna_list) = @_;

	@{$rf_cdna_list} or (return (undef, undef, undef, undef, undef, undef) );

	foreach my $rf_cdna (@{$rf_cdna_list} ) {

		my $cdna_id = $rf_cdna->[0];
		(my $cp_cdna_seq = ${$rf_cdna->[1]}) =~ tr/\ \t\n//d; # remove white spaces
		my $orig_len = length $cp_cdna_seq ;

		my ($rf_btpep, $begin, $end, $unusual)
			 = back_translate_pepseq_lenient (${$rf_pep}, $cp_cdna_seq);

		(defined $rf_btpep) and
			(return  ($cdna_id, $rf_btpep, $begin, $end, $orig_len, $unusual) );
	}

	# if a cnda counterpart has not been found yet, try the reverse translation #

	foreach my $rf_cdna (@{$rf_cdna_list}) {

		my $cdna_id = $rf_cdna->[0];
		my $revcom =  revcom ( ${$rf_cdna->[1]});
		${$revcom} =~ tr/\ \t\n//d;
		my $orig_len = length ${$revcom};

		my ($rf_btpep, $begin, $end, $unusual)
			 = back_translate_pepseq_lenient (${$rf_pep}, ${$revcom}); 

		(defined $rf_btpep) and 
			( return ($cdna_id, $rf_btpep, -$begin, -$end, $orig_len, $unusual) );
	}

		# found no cDNA counterpart #

	return (undef, undef, undef, undef, undef, undef) ;

}		# end of the subroutine reverse_translation_smart_lenient #


# pepaln2codonaln: This subroutine constructs an alignment of codon sequences
#		using the alignment of peptide counterparts as a template.
#
sub pepaln2codonaln (\@\@\@) {

	my ($rf_pepaln, $rf_pepIDs, $rf_cDNAs) = @_;

	unless (@{$rf_pepaln} == @{$rf_cDNAs}) {

		print STDERR "The number of peptide aligned sequences does NOT agree with that of cDNA sequences\!\n";
		return (undef, undef, 1);
	}
	@{$rf_cDNAs} or (return ([], [], 0) );

	my @codonaln = ();
	my @pepIDs = ();
	my $ct_failure = 0;

	OUTER:
	for (my $i=0; $i<@{$rf_pepaln}; $i++) { # the outer for #

		my $emd_peptide = ${$rf_pepaln->[$i]};
		my $cDNA = ${$rf_cDNAs->[$i][1]};	# BEWARE of the magic number '1' #
		my $emd_codonseq = '';
		my $emd_pep_len = length $emd_peptide;
		my $gap_ct = ($emd_peptide =~ tr/-//);
		my $pep_len = $emd_pep_len - $gap_ct ;
		my $cdna_len = length $cDNA;
		unless ($cdna_len == 3*$pep_len) {
			print STDERR "Discrepancy between peptide length ($pep_len) and cDNA length ($cdna_len) for $rf_pepIDs->[$i] \!\n";
			return (undef, undef, 1);			
		}

		#for (1 .. $emd_pep_len) {	# the inner for #
		for (my $i=0; $i <$emd_pep_len; $i++) { # the inner for 

			my $aa = substr($emd_peptide, 0, 1, '');
			$aa = uc $aa;

			if ($aa eq '-') {# outer if-else #

				$emd_codonseq .= '---';
				next;
			} else {

				my $codon = substr($cDNA, 0, 3, '');
				$codon = uc $codon;
				$codon =~ tr/U/T/;
				my $coded_aa = $GENETIC_CODE{$codon};
				(defined $coded_aa) or ($coded_aa = 'X');

				if ( ($coded_aa eq $aa) or ( ($codon eq 'TGA') and ($aa eq 'U') ) or ($coded_aa eq 'X' ) or ($aa =~ m/[BZX]/ )  ) { # inner if-else #

					$emd_codonseq .= $codon;
				} else {
					printf STDERR "The peptide sequence $rf_pepIDs->[$i] is NOT a proper translation of its cDNA counterpart ($aa vs $coded_aa (codon: $codon) for the site %d in the peptide alignment)  \!\n", $i +1;
					$ct_failure++;
					#$emd_codonseq .= 'NNN';
					$emd_codonseq .= $codon;

				}		# end of the inner if-else #
			}		# end of the outer if-else #

		}				# end of the inner for #

		push @codonaln, \$emd_codonseq;
		push @pepIDs, $rf_pepIDs->[$i];

	}					# end of the outer for #

	return (\@codonaln, \@pepIDs, $ct_failure);

}			# end of the subroutine pepaln2codonaln #


# pepaln2codonaln_lite: This subroutine constructs an alignment of codon sequences
#		using the alignment of peptide counterparts as a template.
#		 Unlike 'pepaln2codonaln', this subroutine is intended to consume
#		as little memory space as  possible.
#		For this purpose, it chews up the input peptide sequences, so BEWARE!!
#
# REMEMBER: sub sgl_fastacmd  ($seqid, $database, $exec_dir) = @_;
#
sub pepaln2codonaln_lite (\@\@\@\$) {

	my ($rf_pepaln, $rf_pepIDs, $rf_databases, $exec_dir) = @_;

	my $ct_seq = @{$rf_pepaln};
	unless ($ct_seq == @{$rf_pepIDs}) {

		print STDERR "The number of peptide aligned sequences does NOT agree with that of peptide IDs\!\n";
		return (undef, undef, 1);
	}

	my @databases = @{$rf_databases};

	if ($ct_seq >  @databases ) {  # assume the remaining sequences can be retrieved by the last database
		my $last_db = $databases[$#databases];
		for (my $i=@databases; $i <$ct_seq; $i++) { push @databases, $last_db; }
	}

	my @codonaln = ();
	my @pepIDs = ();
	my $ct_failure =0;

	OUTER:
	for (my $i=0; $i<$ct_seq; $i++) { # the outer for #

		my $emd_peptide = $rf_pepaln->[$i];
		my $pepid = $rf_pepIDs->[$i];
		my ($cDNA, $cdna_len, undef) = sgl_fastacmd ($pepid, ${$databases[$i]}, ${$exec_dir});
		unless (defined $cDNA) {
			print STDERR "Failed to retrieve the cDNA counterpart of the peptide $pepid .\n";
			return (undef, undef, $ct_failure + 1);
		}

		my $emd_codonseq = '';
		my $emd_pep_len = length ${$emd_peptide};
		my $gap_ct = (${$emd_peptide} =~ tr/-//);
		my $pep_len = $emd_pep_len - $gap_ct ;

		unless ($cdna_len == 3*$pep_len) {
			print STDERR "Discrepancy between peptide length ($pep_len) and cDNA length ($cdna_len) for $pepid \!\n";
			return (undef, undef, $ct_failure +1);			
		}

		for (my $i=0; $i<$emd_pep_len; $i++) {	# the inner for #

			my $aa = substr(${$emd_peptide}, 0, 1, '');
			$aa = uc $aa;

			if ($aa eq '-') {# outer if-else #

				$emd_codonseq .= '---';
				next;
			} else {

				my $codon = substr(${$cDNA}, 0, 3, '');
				$codon = uc $codon;
				$codon =~ tr/U/T/;
				my $coded_aa = $GENETIC_CODE{$codon};
				(defined $coded_aa) or ($coded_aa = 'X');

				if ( ($coded_aa eq $aa) or ( ($codon eq 'TGA') and ($aa eq 'U') ) or ($coded_aa eq 'X' ) or ($aa =~ m/[BZX]/ )  ) { # inner if-else #

					$emd_codonseq .= $codon;
				} else {
					printf STDERR "The peptide sequence $pepid is NOT a proper translation of its cDNA counterpart ($aa vs $coded_aa (codon $codon) for the site %d in the peptide alignment) \!\n", $i +1;
					$ct_failure++;
					$emd_codonseq .= $codon;

				}		# end of the inner if-else #
			}		# end of the outer if-else #

		}				# end of the inner for #

		push @codonaln, \$emd_codonseq;
		push @pepIDs, $pepid;
	}					# end of the outer for #

	return (\@codonaln, \@pepIDs, $ct_failure);

}			# end of the subroutine pepaln2codonaln_lite #


# pepaln2codonaln_rg_lite:
#	This subroutine constructs an alignment of codon SUB sequences
#	using the REGIONAL alignment of peptide counterparts as a template.
#	 Unlike 'pepaln2codonaln', this subroutine is intended to consume
#	as little memory space as  possible.
#	For this purpose, it chews up the input peptide sequences, so BEWARE!!
#
# INPUT: (\@pepaln, \@pepids, \@databases, \%id2region, \$exec_dir), where
#	${$pepaln[i-1}} = the i-th peptide (sub)sequence embedded in the regional alignment,
#	${$pepids[i-1}} = the ID of the i-th peptide (stored in the database)
#	${$databases[i-1]} = the database storing the i-th peptide
#	%id2region = (peptide ID => [begin_coord, end_coord]
#		of the cDNA subsequence which are to be used in the alignment, ...)
#
# OUTPUT: (\@bt_pepaln, \@pepids, \%pepid2cdnalen, $ct_failure), where
#	@bt_pepaln stores the cDNA alignment produced by back-translation,
#	%pepid2cdnalen = (peptide ID => FULL LENGTH of the cDNA counterpart stored in the database, ...)
#
# REMINDER: sub sgl_fastacmd  ($seqid, $database, $exec_dir) = @_;
#
sub pepaln2codonaln_rg_lite (\@\@\@\%\$) {

	my ($rf_pepaln, $rf_pepIDs, $rf_databases, $id2region, $exec_dir) = @_;

	my $ct_seq = @{$rf_pepaln};
	unless ($ct_seq == @{$rf_pepIDs}) {
	   print STDERR "The number of peptide aligned sequences does NOT agree with that of peptide IDs\!\n";
	   return (undef, undef, 1);
	}

	my @databases = @{$rf_databases};

	if ($ct_seq >  @databases ) {  # assume the remaining sequences can be retrieved by the last database
	   my $last_db = $databases[$#databases];
	   for (my $i=@databases; $i <$ct_seq; $i++) { push @databases, $last_db; }
	}

	my @codonaln = ();
	my @pepIDs = ();
	my $ct_failure =0;
	my %pepid2cdnalen;

	OUTER:
	for (my $i=0; $i<$ct_seq; $i++) { # the outer for #

		my $emd_peptide = $rf_pepaln->[$i];
		my $pepid = $rf_pepIDs->[$i];
		my ($cDNA, $cdna_len, undef) = sgl_fastacmd ($pepid, ${$databases[$i]}, ${$exec_dir});
		unless (defined $cDNA) {
			print STDERR "Failed to retrieve the cDNA counterpart of the peptide $pepid .\n";
			return (undef, undef, undef, $ct_failure + 1);
		}

		my $emd_codonseq = '';
		my $emd_pep_len = length ${$emd_peptide};
		my $gap_ct = (${$emd_peptide} =~ tr/-//);
		my $pep_len = $emd_pep_len - $gap_ct ;


			## The part added to process REGIONAL (FRAGMENTAL) alignments ##
		$pepid2cdnalen{$pepid}  = $cdna_len;

		my $cdna_region = $id2region->{$pepid};
		if ((defined $cdna_region) and ($cdna_region->[0] ne 'NA') ) {
		    my ($beg_reg, $end_reg) = @{$cdna_region};
		    ( (defined $end_reg) and ($end_reg ne 'NA') ) or ($end_reg = $cdna_len);

		    if ($end_reg > $cdna_len) {
			print STDERR "ILLEGAL INPUT for 'pepaln2codonaln_rg_lite': the end coordiate of the region ($end_reg) exceeds the cDNA length ($cdna_len) for $pepid \!\n";
			return (undef, undef, undef, $ct_failure+1);
		    }

		    ($beg_reg > 1) and (substr (${$cDNA}, 0, $beg_reg -1, '') );

		    my $diff = $cdna_len - $end_reg;
		    ($diff > 0) and (substr (${$cDNA}, -$diff, $diff, '') );

			# replace the cDNA length with the length of
			# the sub-sequence to be used for the back translation
		    $cdna_len =  $end_reg - $beg_reg + 1;
		}
			## END of the added part ##

		unless ($cdna_len == 3*$pep_len) {
			print STDERR "ILLEGAL INPUT for 'pepaln2codonaln_rg_lite': a discrepancy between peptide length ($pep_len) and cDNA length ($cdna_len) for $pepid \!\n";
			return (undef, undef, undef, $ct_failure +1);			
		}

		for (my $i=0; $i<$emd_pep_len; $i++) {	# the inner for #

			my $aa = substr(${$emd_peptide}, 0, 1, '');
			$aa = uc $aa;

			if ($aa eq '-') {# outer if-else #

				$emd_codonseq .= '---';
				next;
			} else {

				my $codon = substr(${$cDNA}, 0, 3, '');
				$codon = uc $codon;
				$codon =~ tr/U/T/;
				my $coded_aa = $GENETIC_CODE{$codon};
				(defined $coded_aa) or ($coded_aa = 'X');

				if ( ($coded_aa eq $aa) or ( ($codon eq 'TGA') and ($aa eq 'U') ) or ($coded_aa eq 'X' ) or ($aa =~ m/[BZX]/ )  ) { # inner if-else #

					$emd_codonseq .= $codon;
				} else {
					printf STDERR "[CAUTION: 'pepaln2codonaln_rg_lite'] The peptide sequence $pepid is NOT a proper translation of its cDNA counterpart ($aa vs $coded_aa (codon $codon) for the site %d in the peptide alignment) \!\n", $i +1;
					$ct_failure++;
					$emd_codonseq .= $codon;

				}		# end of the inner if-else #
			}		# end of the outer if-else #

		}				# end of the inner for #

		push @codonaln, \$emd_codonseq;
		push @pepIDs, $pepid;
	}					# end of the outer for #

	return (\@codonaln, \@pepIDs, \%pepid2cdnalen, $ct_failure);

}			# end of the subroutine pepaln2codonaln_rg_lite #


# print_sequence: This subroutine prints out a sequence.
#
sub print_sequence (\$$\*) {

	my ($rf_seq, $line_len, $fh) = @_;

	my $cp_seq = ${$rf_seq};	# copy the original sequence #

	my $seq_len = length ($cp_seq);
	my $times = int ($seq_len/$line_len);
	my $residue = $seq_len % $line_len;

	for (my $i=0; $i<$times; $i++) {
	#foreach my $i (0..$times) {

		my $subseq = substr ($cp_seq, 0, $line_len, '');
		print $fh "$subseq\n";
	}

	$residue and (print $fh "$cp_seq\n");

	return 1;
}				# end of the subroutine print_sequence #


# print_subsequence:
#	This subroutine prints out a subregion of a sequence.
#
# INPUT: (\$seq, $beg_coord, $end_coord, $dis_len, \*${$fh}), where
#	$beg_coord and $end_coord are coordinates of the region in the frame
#	where the beginning of $seq is 1.
#	$dis_len is the maximum length of the sequence displayed in a line.
#
# OUTPUT: 1 if the subroutine succeeded,
#	   0 if it failed.
#
sub print_subsequence (\$$$$\*) {

	my ($seq, $beg_coord, $end_coord, $dis_len, $fh) = @_;

	my $len_end_margin = length (${$seq}) - $end_coord;
	my $len_beg_margin = $beg_coord - 1;

	if ($len_end_margin < 0) {
	    printf STDERR "ILLEGAL INPUT for 'print_subsequence': (length, end) = (%d, %d)\n", length (${$seq}), $end_coord;
	    return 0;
	}
	if ($len_beg_margin < 0) {
	    print STDERR "ILLEGAL INPUT for 'print_subsequence': beg = $beg_coord .\n";
	    return 0;
	}

	my $sbseq = ${$seq};
	($len_end_margin > 0) and
		substr ($sbseq, -$len_end_margin, $len_end_margin, '') ;
	($len_beg_margin > 0) and  substr ($sbseq, 0, $len_beg_margin, '') ;

	print_sequence ($sbseq, $dis_len, *${$fh});

}		# end of the subroutine 'print_subsequence' #


# print_fasta: This subroutine prints out multiple sequences in a fasta format.
#
sub print_fasta (\@\@$\*) {

	my ($rf_seqIDs, $rf_seqs, $dis_len, $fh) = @_;
	$dis_len or ($dis_len = 60);

	(@{$rf_seqIDs} == @{$rf_seqs}) or (return 0);

	for (my $i=0; $i<@{$rf_seqIDs}; $i++) {

		print $fh '>' . $rf_seqIDs->[$i] . "\n";
		unless (print_sequence (${$rf_seqs->[$i]}, $dis_len, *${$fh}) ) {
			print STDERR "Output of $rf_seqIDs->[$i] failed.\n";
			return 0;
		}
	}

	return 1;
}			# end of the subroutine print_fasta #


# pr_multifasta_sbseqs:
#	This subroutine first retrieves sequences in the input list,
#	then trims them if necessary,
#	and finally prints them out in a multi-FASTA format.
#
# INPUT: (\@info_seqs, \$outfile, \$exec_dir, $dis_len, $width_margin), where
#	@{$info_seqs[$k]} = ($seq_id, $seq_len, \$input_db, \@intervals)
#	for the $k th sequence, with
#	@{$intervals[$m]} = ($beg, $end) for the $m th interval.
#	$seq_len and \@intervals can be undefined;
#	$exec_dir specifies the directory where the executive program is located;
#	$width_margin is the maximum width of the margin
#	it can take beyond the specified boundaries.
#
# OUTPUT: \@out_intervals, where
#	@{$out_intervals[$k]} = ($beg_out, $end_out) for the $k th sequence.
#
sub pr_multifasta_sbseqs (\@\$\$$$) {

	my ($info_seqs, $outfile, $exec_dir, $dis_len, $width_margin) = @_;

	($width_margin < 0) and ($width_margin = 0);

	my $fh_out;
	unless (open ($fh_out, "> ${$outfile}") ) {
	    print STDERR "[ERROR 'pr_multifasta_sbseqs'] FAILED to open the file: '${$outfile}' !\n";
	    return undef;
	}

	my @out_intervals = ();

	foreach my $seqinfo (@{$info_seqs}) { # The outer foreach-loop #
	    my ($seqid, $seqlen, $input_db, $intervals) = @{$seqinfo};
	    my $ct_intvls = (defined $intervals) ? scalar (@{$intervals}) : 0;

	    my ($seq, $seqlen2, $header) = sgl_fastacmd ($seqid, ${$input_db}, ${$exec_dir});

	    if ((defined $seqlen) and ($seqlen != $seqlen2)) {
		print STDERR "[CAUTION 'pr_multifasta_sbseqs'] The stated sequence length ($seqlen) CLASHES with the measured one ($seqlen2) for the sequence '$seqid', we will trust the latter !\n";
		$seqlen = $seqlen2;
	    }
	    my ($leftbd, $rightbd) = ($ct_intvls>0) ? @{$intervals->[0]} : (1, $seqlen) ;
	    for (my $m=1; $m<$ct_intvls; $m++) { # The inner foreach-loop #
		my ($beg, $end) = @{$intervals->[$m]};
		($beg < $leftbd) and ($leftbd = $beg);
		($rightbd < $end) and ($rightbd = $end);
	    }				# End of the foreach-loop #

		# Take account of the margins #
	    $leftbd = ($leftbd - $width_margin > 1) ? $leftbd - $width_margin : 1 ;
	    $rightbd = ($rightbd + $width_margin < $seqlen) ? $rightbd + $width_margin : $seqlen ;

		# Trim the sequences #
	    ($leftbd > 1) and ( substr (${$seq}, 0, $leftbd - 1, '') );
	    ($rightbd < $seqlen) and ( substr (${$seq}, $rightbd - $seqlen, $seqlen - $rightbd, '') );

	     (($leftbd>1) or ($rightbd < $seqlen)) and ($seqid .= "   \[$leftbd\,$rightbd\]  (org_len = $seqlen)" );
	     print $fh_out "> $seqid", "\n";
	     print_sequence (${$seq}, $dis_len, *${$fh_out});

	     push @out_intervals, [$leftbd, $rightbd];

	}			# end of the outer foreach-loop #

	close ($fh_out);

	return \@out_intervals;

}		# End of the subroutine 'pr_multifasta_sbseqs' #


# mask_seq_regions:
#	This subroutine masks regions that are NOT in the specified list.
#	NOTE that it eats up the input sequence.
#
# INPUT: (\$seq, $fr_beg, $fr_end, \@intact_intvls, $indx_beg, $indx_end, $char_mask), where
#	$fr_beg/end is the beginning/ending coordinate of the entire sequence;
#	@{$intact_intvls[$m]}[$indx_beg, $indx_end] = ($beg, $end)
#	is for the $m th interval that will NOT be masked.
#
# OUTPUT: \$masked_seq.
#
sub mask_seq_regions (\$$$\@$$$) {

	my ($seq, $fr_beg, $fr_end, $intact_intvls,
		$indx_beg, $indx_end, $char_mask) = @_;

	my $seq_len = length ${$seq};

	unless ($seq_len == $fr_end - $fr_beg + 1) {
	    print STDERR "[ERROR 'mask_seq_regions'] The input sequence (length $seq_len) does NOT FIT IN the framework \[$fr_beg\,$fr_end\] !\n";
	    return undef;
	}

	my $masked_seq = '';
	my $exp_beg = $fr_beg ; # Expected beginning of the next interval
				# if there is no gap.

	foreach my $intvl (@{$intact_intvls}) { # The main foreach-loop #
	    my ($beg, $end) = @{$intvl}[$indx_beg, $indx_end];
	    if (($beg>$fr_end) or ($end > $fr_end)) {
		print STDERR "[ERROR 'mask_seq_regions'] ILLEGAL INPUT: THe interval \[$beg\,$end\] WEND BEYOND the right boundary ($fr_end) !\n";
		return undef;
	    }

	    my $left_margin = $beg - $exp_beg;

	    if ($left_margin<0) {
		printf STDERR "[ERROR 'mask_seq_regions'] Found an OVERLAPPING pair of INTERVALS:  last_end = %d,  curr_beg = $beg !\n";
		return undef;
	    } elsif ($left_margin >0) {
		$masked_seq .= $char_mask x $left_margin;
		substr (${$seq}, 0, $left_margin, '');
	    }

	    my $len_intvl = $end - $beg + 1;
	    if ($len_intvl <=0) {
		print STDERR "[ERROR 'mask_seq_regions'] Found an NON-POSITIVE INTERVAL: \[$beg\,$end\] !\n\n";
		return undef;
	    }

	    $masked_seq .= substr (${$seq}, 0, $len_intvl, '');

	    $exp_beg = $end + 1;
	}				# End of the main foreach-loop #

	my $right_margin = ($fr_end + 1 - $exp_beg);
	($right_margin>0) and ($masked_seq .= $char_mask x $right_margin);

	return \$masked_seq;

}		# End of the subroutine 'mask_seq_regions' #


# pr_mask_multifasta:
#	This subroutine first retrieves sequences in the input list,
#	then masks irrelevant regions,
#	and finally prints them out in a multi-FASTA format.
#
# INPUT: (\@info_seqs, \$outfile, \$exec_dir, $dis_len, $width_margin, $char_mask), where
#	@{$info_seqs[$k]} = ($seq_id, $seq_len, \$input_db, \@intervals)
#	for the $k th sequence, with
#	@{$intervals[$m]} = ($beg, $end) for the $m th interval.
#	$seq_len and \@intervals can be undefined;
#	$exec_dir specifies the directory where the executive program is located;
#	$width_margin is the width of the margin
#	it can go beyond the specified boundaries.
#
# OUTPUT: \@out_intervals, where
#	@{$out_intervals[$k]->[$m]} = ($beg_out, $end_out)
#	for the $m th interval of the $k th sequence.
#
sub pr_mask_multifasta (\@\$\$$$$) {

	my ($info_seqs, $outfile, $exec_dir, $dis_len, $width_margin, $char_mask) = @_;

	($width_margin < 0) and ($width_margin = 0);

	my $fh_out;
	unless (open ($fh_out, "> ${$outfile}") ) {
	    print STDERR "[ERROR 'pr_multifasta_sbseqs'] FAILED to open the file: '${$outfile}' !\n";
	    return undef;
	}

	my @out_intervals = ();

	foreach my $seqinfo (@{$info_seqs}) { # The outer foreach-loop #
	    my ($seqid, $seqlen, $input_db, $intervals) = @{$seqinfo};
	    my $ct_intvls = (defined $intervals) ? scalar (@{$intervals}) : 0;

	    my ($seq, $seqlen2, $header) = sgl_fastacmd ($seqid, ${$input_db}, ${$exec_dir});

	    if ((defined $seqlen) and ($seqlen != $seqlen2)) {
		print STDERR "[CAUTION 'pr_mask_multifasta'] The stated sequence length ($seqlen) CLASHES with the measured one ($seqlen2) for the sequence '$seqid', we will trust the latter !\n";
		$seqlen = $seqlen2;
	    }

		# Take account of the margins #
	    my @ext_intvls = ();
	    for (my $m=0; $m<$ct_intvls; $m++) { # The inner foreach-loop #
		my ($beg, $end) = @{$intervals->[$m]};
		push @ext_intvls, [$beg - $width_margin, $end + $width_margin];
	    }				# End of the foreach-loop #

		# Remove overlaps #

	    my $novlp_intvls = merge_integer_intervals (@ext_intvls, 0, 1);

		# Trim the regions to fit in the original framework [1, $seqlen] #
	    my @frame = ([1, $seqlen]);
	    (my $trimmed_intvls, undef, undef)
		= compare_interval_sets (@{$novlp_intvls}, @frame);

		# Mask the sequence #

	    my $masked_seq = mask_seq_regions (${$seq}, 1, $seqlen,
			 @{$trimmed_intvls}, 0, 1, $char_mask);

	    unless (defined $masked_seq) {
		print STDERR "[ERROR 'pr_mask_multifasta'] 'mask_seq_regions' FAILED for the sequence '$seqid' !\n";
		return undef;
	    }

	    my $leftbd = $trimmed_intvls->[0][0];
	    my $rightbd = $trimmed_intvls->[$#{$trimmed_intvls}][1];
	    my $ct_trimmed_intvls = @{$trimmed_intvls};
	    if (($ct_trimmed_intvls>1) or ($leftbd>1) or ($rightbd < $seqlen)) {
		my ($begs , $ends) = @{$trimmed_intvls->[0]};
		for (my $k=1; $k<$ct_trimmed_intvls; $k++) {
		    $begs .= ':' . $trimmed_intvls->[$k][0];
		    $ends .= ':' . $trimmed_intvls->[$k][1];
		}
		$seqid .= "   \[$begs\,$ends\]  (org_len = $seqlen)";
	     }

	     print $fh_out "> $seqid", "\n";
	     print_sequence (${$masked_seq}, $dis_len, *${$fh_out});

	     push @out_intervals, [$leftbd, $rightbd];

	}			# end of the outer foreach-loop #

	close ($fh_out);

	return \@out_intervals;

}		# End of the subroutine 'pr_mask_multifasta' #


1;		# a "spell" which makes the module "use"-able and "require"-able #


