package MyPertProb_locMSA_dawg_smpl;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.6: Copyright (C) 2015 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyPertProb_locMSA_dawg_smpl.pm:
#	This Perl module is a collection of subroutines
#	that perform perturbation calculations of the probability
#	of a given SAMPLE local multiple sequence alignment (MSA),
#	under Dawg's indel model (Cartwright 2005) with the space- and time-homogeneous indel rates.
#	
#	It was created by taking advantage of some subroutines in 'MyPertProb_locPWA_dawg.pm'.
#
#	[CREATION STARTED on May 8 (Fri), 2015, ENDED on ...]
#
# MyPertProb_locPWA_dawg.pm:
#	This Perl module is a collection of subroutines
#	that perform perturbation calculations of the probability
#	of a given local pairwise sequence alignment (PWA),
#	under Dawg's indel model (Cartwright 2005) with the space- and time-homogeneous indel rates.
#	(Creation STARTED on April 28 (Tue), 2015, ENDED on ...)
#

use strict;
use warnings;

use MyArrayManipu_num qw( diff_nmarray log_sum_exp );
	## union_nmarray

use MyPertProb_locPWA_dawg qw(
	mfac_null_2events
	mfac_asites_vs_dgaps_1event  mfac_asites_vs_dgaps_2events
	mfac_agaps_vs_dsites_1event  mfac_agaps_vs_dsites_2events
	mfac_asites_nhom_dsites_2events  mfac_asites_nhom_dsites_3events
	);


=set aside for later use
use MyTreeManipu_II qw(
		);
	## input_tree_file read_tree_file br_output_tree

use MyArrayManipu qw( diff_array union_array nr_array);

use MyMalnIO qw( pr_ODEN_input );

use MyTaxonomy qw(
	fetch_ancestry 
	fetch_last_common_ancestor_ov2
	fetch_common_ancestors
		);
	## fetch_ext_offsprings fetch_all_offsprings
	## fetch_ext_offsprings_br

use MyTreeManipu_III qw(
	make_node2abs_depth make_node2ct_exoffs
	);

use MyTreeMap_II qw(
	locate_parsim_changes
	);
use MyTreeMap_D qw(
	map_parsim_states_dollo_msa
	);
	# ct_presext_dollo_bu map_node2parsstat_dollo_td
	# map_parsim_states_dollo_clm
	# map_parsim_states_dollo_1clm
=cut

use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	mfac_lone_dsites_vs_ddgaps_2events_3rd
	mfac_dgaps2_nested_in_dgaps3_2events
	mfac_dgaps2_nested_in_dgaps3_3events
	);


	# This specifies the level of tests on the main subroutines,
	# 'xxx', ...
	# (0, Nothing; 1, brief; 2, a bit detailed; 3, quite detailed; 5, very detailed)
##our $TEST_LEVEL = 1;
##our $TEST_LEVEL = 2;
##our $TEST_LEVEL = 3;
our $TEST_LEVEL = 5;

	# This specifies the level of tests on the primary satellite subroutines,
	# 'cal_lnlk1_indel_scn_cls', ...
	# (0, Nothing; 1, brief; 2, a bit detailed; 3, quite detailed; 5, very detailed)
##our $TEST_LEVEL1 = 1;
##our $TEST_LEVEL1 = 2;
##our $TEST_LEVEL1 = 3;
our $TEST_LEVEL1 = 5;

	### DEFINITIONS of the SUBROUTINES ###


	## (I) Fundamental subroutines that give basic functions necessary for the practical perturbation calculations. ##

# F_e_neg:
#	This basic subroutine calculates the basic function:
#	F_e- (x,t) := \int_{0}^{t} dt1 exp(-x*t)
#		= (1 - exp(-x*t))/x if x,
#		= 0, := t if x == 0.
#	It is defined in Eq.(SR-8B.2) in Supplementary Results 8B.
#
#	[CREATED on April 28 (Tue), 2015.] ... IMPORTED from 'MyPertProb_locPWA_dawg.pm'.
#
# INPUT: ($x, $t), both of which are real numbers.
#
# OUTPUT: $f_en = F_e- (x,t).
#
sub F_e_neg ($$) {

	my ($x, $t) = @_;

	my $f_en = ($x == 0) ? $t : (1 - exp (-$x*$t))/$x ;

	return $f_en;

}	# End of the subroutine 'F_e_neg'. #



	## (II) Main subroutines that perform the practical perturbation calculations on sample local MSAs. ##


# mfac_lone_dsites_vs_ddgaps_2events_3rd:
#	This subroutine calculates the portion of the multiplication factor contributed by the "3rd type" of the 2-event (next-parsimonious) local indel history.
#	The history is supposed to be consistent with a run of $n_D1 (= Delta L^{D1}) sites in the sequence 1
#	aligned with a run of gaps in each of sequences 2 and 3, all in between a pair of PASs, (i.e., the local MSA in case (III)).
#	And the history is considered under a 3-OTU tree whose internal node is regarded as the root.
#
#	The subroutine uses the formula, Eq.(VI-3.13') in Results VI.
#
#	[CREATED on May 8 (Fri), 2015.]
#
# INPUT: ($n_D1, \@set_params1, \@set_params2, \@set_params3), where
#
#	$n_D1 is the number of sites that sequence 1 has in between the PASs.
#
#	@set_params1/2/3 = ($b1/2/3, $lambda_I1/2/3, $lambda_D1/2/3, \@freqs_I1/2/3, \@freqs_D1/2/3, $cutoff_I1/2/3, $cutoff_D1/2/3)
#	are indel model parameters along the branch between the root and sequence 1/2/3,
#	with
#	$b1/2/3 is the branch length,
#	$lambda_I1/2/3 and $lambda_D1/2/3 are the total insertion rate and the total deletion rate, respectively, per unit branch length and per site;
#	$freqs_I1/2/3[$l] ($l=1,...,$cutoff_I1/2/3) is the relative frequency of $l-residue insertions;
#	$freqs_D1/2/3[$l] ($l=1,...,$cutoff_D1/2/3) is the relative frequency of $l-residue deletions.
#
# OUTPUT: $mfac_3rd, which is the portion of the multiplication factor we desire.
#
#	The subroutine returns undef if it encounters an error.
#
sub mfac_lone_dsites_vs_ddgaps_2events_3rd ($\@\@\@) {

	my ($n_D1, $set_params1, $set_params2, $set_params3) = @_;

	my ($b1, $lambda_I1, $lambda_D1, $freqs_I1, $freqs_D1, $cutoff_I1, $cutoff_D1) = @{$set_params1};
	my ($b2, $lambda_I2, $lambda_D2, $freqs_I2, $freqs_D2, $cutoff_I2, $cutoff_D2) = @{$set_params2};
	my ($b3, $lambda_I3, $lambda_D3, $freqs_I3, $freqs_D3, $cutoff_I3, $cutoff_D3) = @{$set_params3};

	unless (($n_D1 <= $cutoff_D2) and ($n_D1 <= $cutoff_D3)) { # In this case, the 3rd type local history canNOT contribute a probability.
	    return 0;
	}

	my $prob_b1 = exp ( -($lambda_I1 + $lambda_D1) * $n_D1 * $b1 );
	my $prob_b2 = $lambda_D2 * $freqs_D2->[$n_D1] * (1 - exp (-($lambda_I2 + $lambda_D2) * $n_D1 * $b2)) /(($lambda_I2 + $lambda_D2)*$n_D1) ;
	my $prob_b3 = $lambda_D3 * $freqs_D3->[$n_D1] * (1 - exp (-($lambda_I3 + $lambda_D3) * $n_D1 * $b3)) /(($lambda_I3 + $lambda_D3)*$n_D1) ;

	my $mfac_3rd = $prob_b1 * $prob_b2 * $prob_b3 ;

	return $mfac_3rd ;

}	# End of the subroutine 'mfac_lone_dsites_vs_ddgaps_2events_3rd'. #



# tld_mu_P_msa_E:
#	This satellite subroutine of 'mfac_dgaps2_nested_in_dgaps3_2events'
#	calculates the \tilde{\mu}_P[(E)], which is the multiplication factor contributed by
#	a history consisting of an insertion along branch 1 and a deletion along branch 3.
#	The root state in this case is identical to that of sequence 2, i.e., [L, 1,...,i, j,...,\Delta L^{D1}, R].
#
#	The portion is calculated according to Eq.(SR-9.2b') in Supplementary Results.
#
#	[CREATED on May 9 (Sat), 2015.]
#
# INPUT: ($n_D1, $i, $j, \@set_params1, \@set_params2, \@set_params3), where
#
#	$n_D1 is the number of sites that sequence 1 has in between the PASs;
#	Sequence 2 has a run of gaps flanked by the sites homologous to the $i th and the $j th sites in sequence 1.
#	(And sequence 3 has a run of gaps aligned with the whole run of $n_D1 sites in sequence 1.)
#
#	@set_params1/2/3 = ($b1/2/3, $lambda_I1/2/3, $lambda_D1/2/3, \@freqs_I1/2/3, \@freqs_D1/2/3, $cutoff_I1/2/3, $cutoff_D1/2/3)
#	are indel model parameters along the branch between the root and sequence 1/2/3,
#	with
#	$b1/2/3 is the branch length,
#	$lambda_I1/2/3 and $lambda_D1/2/3 are the total insertion rate and the total deletion rate, respectively, per unit branch length and per site;
#	$freqs_I1/2/3[$l] ($l=1,...,$cutoff_I1/2/3) is the relative frequency of $l-residue insertions;
#	$freqs_D1/2/3[$l] ($l=1,...,$cutoff_D1/2/3) is the relative frequency of $l-residue deletions.
#
# OUTPUT: $tmu_P_E, which is the multiplication factor we desire.
#
#	The subroutine returns undef if it encounters an error.
#
sub tld_mu_P_msa_E ($$$\@\@\@) {

	my ($n_D1, $i, $j, $set_params1, $set_params2, $set_params3) = @_;

	my ($b1, $lambda_I1, $lambda_D1, $freqs_I1, $freqs_D1, $cutoff_I1, $cutoff_D1) = @{$set_params1};
	my ($b2, $lambda_I2, $lambda_D2, $freqs_I2, $freqs_D2, $cutoff_I2, $cutoff_D2) = @{$set_params2};
	my ($b3, $lambda_I3, $lambda_D3, $freqs_I3, $freqs_D3, $cutoff_I3, $cutoff_D3) = @{$set_params3};

	my $l_I1 = $j - $i - 1;
	my $l_D3 = $n_D1 - $l_I1;

	unless (($l_I1 > 0) and ($l_D3 > 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_msa_E'] (A) NEGATIVE INDEL LENGTH(S) found: (l_I1, l_D3) = ($l_I1, $l_D3): $! !\n\n";
	    return undef;
	} 

	unless (($l_I1 <= $cutoff_I1) and ($l_D3 <= $cutoff_D3)) { # This history has the occurrence probability = ZERO under the given indel model.
	   return 0;
	}

	my $lambda1 = $lambda_I1 + $lambda_D1;
	my $mu_P_I1 = $lambda_I1 * $freqs_I1->[$l_I1] * (1 - exp (-$lambda1 * $l_I1 * $b1)) /($lambda1 * $l_I1) ;

	my $lambda3 = $lambda_I3 + $lambda_D3;
	my $mu_P_D3 = $lambda_D3 * $freqs_D3->[$l_D3] * (exp ($lambda3 * $l_D3 * $b3) - 1) /($lambda3 * $l_D3);

	my $tmu_P_E = $mu_P_I1 * $mu_P_D3 ;

	return $tmu_P_E ;

}	# End of the subroutine 'tld_mu_P_msa_E'. #


# tld_mu_P_msa_F:
#	This satellite subroutine of 'mfac_dgaps2_nested_in_dgaps3_2events'
#	calculates the \tilde{\mu}_P[(F)], which is the multiplication factor contributed by
#	a history consisting of a deletion along branch 2 and a deletion along branch 3.
#	The root state in this case is identical to that of sequence 1, i.e., [L, 1, ..., \Delta L^{D1}, R].
#
#	The portion is calculated according to Eq.(SR-9.2c') in Supplementary Results.
#
#	[CREATED on May 9 (Sat), 2015.]
#
# INPUT: ($n_D1, $i, $j, \@set_params1, \@set_params2, \@set_params3), where
#
#	$n_D1 is the number of sites that sequence 1 has in between the PASs;
#	Sequence 2 has a run of gaps flanked by the sites homologous to the $i th and the $j th sites in sequence 1.
#	(And sequence 3 has a run of gaps aligned with the whole run of $n_D1 sites in sequence 1.)
#
#	@set_params1/2/3 = ($b1/2/3, $lambda_I1/2/3, $lambda_D1/2/3, \@freqs_I1/2/3, \@freqs_D1/2/3, $cutoff_I1/2/3, $cutoff_D1/2/3)
#	are indel model parameters along the branch between the root and sequence 1/2/3,
#	with
#	$b1/2/3 is the branch length,
#	$lambda_I1/2/3 and $lambda_D1/2/3 are the total insertion rate and the total deletion rate, respectively, per unit branch length and per site;
#	$freqs_I1/2/3[$l] ($l=1,...,$cutoff_I1/2/3) is the relative frequency of $l-residue insertions;
#	$freqs_D1/2/3[$l] ($l=1,...,$cutoff_D1/2/3) is the relative frequency of $l-residue deletions.
#
# OUTPUT: $tmu_P_F, which is the multiplication factor we desire.
#
#	The subroutine returns undef if it encounters an error.
#
sub tld_mu_P_msa_F ($$$\@\@\@) {

	my ($n_D1, $i, $j, $set_params1, $set_params2, $set_params3) = @_;

	my ($b1, $lambda_I1, $lambda_D1, $freqs_I1, $freqs_D1, $cutoff_I1, $cutoff_D1) = @{$set_params1};
	my ($b2, $lambda_I2, $lambda_D2, $freqs_I2, $freqs_D2, $cutoff_I2, $cutoff_D2) = @{$set_params2};
	my ($b3, $lambda_I3, $lambda_D3, $freqs_I3, $freqs_D3, $cutoff_I3, $cutoff_D3) = @{$set_params3};

	my $l_D2 = $j - $i - 1;
	my $l_D3 = $n_D1;

	unless (($l_D2 > 0) and ($l_D3 > 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_msa_F'] (A) NEGATIVE INDEL LENGTH(S) found: (l_D2, l_D3) = ($l_D2, $l_D3): $! !\n\n";
	    return undef;
	} 

	unless (($l_D2 <= $cutoff_D2) and ($l_D3 <= $cutoff_D3)) { # This history has the occurrence probability = ZERO under the given indel model.
	   return 0;
	}

	my $lambda1 = $lambda_I1 + $lambda_D1;
	my $lambda2 = $lambda_I2 + $lambda_D2;
	my $lambda3 = $lambda_I3 + $lambda_D3;

	my $mu_P_D2 = $lambda_D2 * $freqs_D2->[$l_D2] * (exp ($lambda2 * $l_D2 * $b2) - 1) /($lambda2 * $l_D2) ;

	my $mu_P_D3 = $lambda_D3 * $freqs_D3->[$l_D3] * (exp ($lambda3 * $l_D3 * $b3) - 1) /($lambda3 * $l_D3);

	my $tmu_P_F = exp (-($lambda1*$b1 + $lambda2*$b2 + $lambda3*$b3) * $l_D2) * $mu_P_D2 * $mu_P_D3 ;

	return $tmu_P_F ;

}	# End of the subroutine 'tld_mu_P_msa_F'. #


# mfac_dgaps2_nested_in_dgaps3_2events:
#	This subroutine calculates the portion of the multiplication factor contributed by the 2-event (parsimonious) local indel histories.
#	The histories are suppposed to be consistent with a run of $n_D1 gaps in sequence 3 and a run of ($j - $i-1) gaps in sequence 2;
#	the latter is nested in the former, in between a pair of PASs. This local MSA belongs to case (IV).
#	And the histories are considered under a 3-OTU tree whose internal node is regarded as the root.
#	
#	The subroutine uses the formula, Eq.(SR-9.2a) supplemented with Eqs.(SR-92b',2c').
#
#	[CREATED on May 8 (Fri), 2015.]
#
#
# INPUT: ($n_D1, $i, $j, \@set_params1, \@set_params2, \@set_params3), where
#
#	$n_D1 is the number of sites that sequence 1 has in between the PASs;
#	Sequence 2 has a run of gaps flanked by the sites homologous to the $i th and the $j th sites in sequence 1.
#	(And sequence 3 has a run of gaps aligned with the whole run of $n_D1 sites in sequence 1.)
#
#	@set_params1/2/3 = ($b1/2/3, $lambda_I1/2/3, $lambda_D1/2/3, \@freqs_I1/2/3, \@freqs_D1/2/3, $cutoff_I1/2/3, $cutoff_D1/2/3)
#	are indel model parameters along the branch between the root and sequence 1/2/3,
#	with
#	$b1/2/3 is the branch length,
#	$lambda_I1/2/3 and $lambda_D1/2/3 are the total insertion rate and the total deletion rate, respectively, per unit branch length and per site;
#	$freqs_I1/2/3[$l] ($l=1,...,$cutoff_I1/2/3) is the relative frequency of $l-residue insertions;
#	$freqs_D1/2/3[$l] ($l=1,...,$cutoff_D1/2/3) is the relative frequency of $l-residue deletions.
#
# OUTPUT: ($mfac_2ev, \@components), where
#	$mfac_2ev is the portion of the multiplication factor we desire;
#	@components stores the components of $mfac_2ev.
#
#	The subroutine returns undef if it encounters an error.
#
sub mfac_dgaps2_nested_in_dgaps3_2events ($$$\@\@\@) {

	my ($n_D1, $i, $j, $set_params1, $set_params2, $set_params3) = @_;

	unless (($n_D1 > 1) and (0 <= $i) and ($i+1 < $j) and ($j <= $n_D1+1)) {
	    print STDOUT "[ERROR 'mfac_dgaps2_nested_in_dgaps3_2events'] ILLEGAL INPUT: ", '($n_D1, $i, $j) = ', "($n_D1, $i, $j) : $! !\n\n";
	    return (undef, undef); 
	}

	unless (($i != 0) or ($j != $n_D1+1)) {
	    print STDOUT "[ERROR 'mfac_dgaps2_nested_in_dgaps3_2events'] ILLEGAL INPUT: ", '($n_D1, $i, $j) = ', "($n_D1, $i, $j) : $! !\n\n";
	    return (undef, undef); 
	}


	my $tmu_P_E = tld_mu_P_msa_E ($n_D1, $i, $j, @{$set_params1}, @{$set_params2}, @{$set_params3}) ;

	unless (defined $tmu_P_E) {
	    print STDOUT "[ERROR 'mfac_dgaps2_nested_in_dgaps3_2events'] The subroutine 'tld_mu_P_msa_E' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my $tmu_P_F = tld_mu_P_msa_F ($n_D1, $i, $j, @{$set_params1}, @{$set_params2}, @{$set_params3}) ;

	unless (defined $tmu_P_F) {
	    print STDOUT "[ERROR 'mfac_dgaps2_nested_in_dgaps3_2events'] The subroutine 'tld_mu_P_msa_F' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my $mfac_2ev = $tmu_P_E + $tmu_P_F ;
	my @components = ($tmu_P_E, $tmu_P_F);

	return ($mfac_2ev, \@components) ;


}	# End of the subroutine 'mfac_dgaps2_nested_in_dgaps3_2events'. #



# tld_mu_P_msa_G12:
#	This satellite subroutine of 'mfac_dgaps2_nested_in_dgaps3_3events'
#	calculates the \tilde{\mu}_P[(G-1)] + \tilde{\mu}_P[(G-2)], which is the portion of the multiplication factor
#	contributed by the 3-event histories each of which consists of a two-event history equivalent to an insertion along branch 1
#	and a deletion along branch 3.
#	The root state in this case is identical to that of sequence 2, i.e., [L, 1,...,i, j,...,\Delta L^{D1}, R].
#
#	The portion is calculated according to Eq.(SR-9.4a,b) in Supplementary Results.
#
#	[CREATED on May 9 (Sat), 2015.]
#
# INPUT: ($n_D1, $i, $j, \@set_params1, \@set_params2, \@set_params3), where
#
#	$n_D1 is the number of sites that sequence 1 has in between the PASs;
#	Sequence 2 has a run of gaps flanked by the sites homologous to the $i th and the $j th sites in sequence 1.
#	(And sequence 3 has a run of gaps aligned with the whole run of $n_D1 sites in sequence 1.)
#
#	@set_params1/2/3 = ($b1/2/3, $lambda_I1/2/3, $lambda_D1/2/3, \@freqs_I1/2/3, \@freqs_D1/2/3, $cutoff_I1/2/3, $cutoff_D1/2/3)
#	are indel model parameters along the branch between the root and sequence 1/2/3,
#	with
#	$b1/2/3 is the branch length,
#	$lambda_I1/2/3 and $lambda_D1/2/3 are the total insertion rate and the total deletion rate, respectively, per unit branch length and per site;
#	$freqs_I1/2/3[$l] ($l=1,...,$cutoff_I1/2/3) is the relative frequency of $l-residue insertions;
#	$freqs_D1/2/3[$l] ($l=1,...,$cutoff_D1/2/3) is the relative frequency of $l-residue deletions.
#
# OUTPUT: $tmu_P_G12, which is the multiplication factor we desire.
#
#	The subroutine returns undef if it encounters an error.
#
sub tld_mu_P_msa_G12 ($$$\@\@\@) {

	my ($n_D1, $i, $j, $set_params1, $set_params2, $set_params3) = @_;

	my ($b1, $lambda_I1, $lambda_D1, $freqs_I1, $freqs_D1, $cutoff_I1, $cutoff_D1) = @{$set_params1};
	my ($b2, $lambda_I2, $lambda_D2, $freqs_I2, $freqs_D2, $cutoff_I2, $cutoff_D2) = @{$set_params2};
	my ($b3, $lambda_I3, $lambda_D3, $freqs_I3, $freqs_D3, $cutoff_I3, $cutoff_D3) = @{$set_params3};

	my $l_I1 = $j - $i - 1;
	my $l_D3 = $n_D1 - $l_I1;

	unless (($l_I1 > 0) and ($l_D3 > 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_msa_G12'] (A) NEGATIVE INDEL LENGTH(S) found: (l_I1, l_D3) = ($l_I1, $l_D3): $! !\n\n";
	    return undef;
	} 

	unless (($l_I1 <= 2* $cutoff_I1) and ($l_D3 <= $cutoff_D3)) { # This history has the occurrence probability = ZERO under the given indel model.
	   return 0;
	}

	my $lambda1 = $lambda_I1 + $lambda_D1;
	my $lambda3 = $lambda_I3 + $lambda_D3;

	my ($tmu_P_I1_2ev, $components_I1_2ev) = mfac_agaps_vs_dsites_2events ($l_I1, $b1, $lambda_I1, $lambda_D1, @{$freqs_I1}, @{$freqs_D1}, $cutoff_I1, $cutoff_D1);
	unless (defined $tmu_P_I1_2ev) {
	    print STDOUT "[ERROR 'tld_mu_P_msa_G12'] Subroutine 'mfac_agaps_vs_dsites_2events' FAILED : $! !\n\n";
	    return undef;
	}
	##my $mu_P_I1 = $lambda_I1 * $freqs_I1->[$l_I1] * (1 - exp (-$lambda1 * $l_I1 * $b1)) /($lambda1 * $l_I1) ;

	my $mu_P_D3 = $lambda_D3 * $freqs_D3->[$l_D3] * (exp ($lambda3 * $l_D3 * $b3) - 1) /($lambda3 * $l_D3);

	my $tmu_P_G12 = $tmu_P_I1_2ev * $mu_P_D3 ;

	return $tmu_P_G12 ;

}	# End of the subroutine 'tld_mu_P_msa_G12'. #


# tld_mu_P_msa_G34:
#	This satellite subroutine of 'mfac_dgaps2_nested_in_dgaps3_3events'
#	calculates the \tilde{\mu}_P[(G-3)] + \tilde{\mu}_P[(G-4)], which is the portion of the multiplication factor
#	contributed by the 3-event histories each of which consists of an insertion along branch 1
#	and a 2-event history equivalent to a deletion along branch 3.
#	The root state in this case is identical to that of sequence 2, i.e., [L, 1,...,i, j,...,\Delta L^{D1}, R].
#
#	The portion is calculated according to Eq.(SR-9.4c,d) in Supplementary Results.
#
#	[CREATED on May 9 (Sat), 2015.]
#
# INPUT: ($n_D1, $i, $j, \@set_params1, \@set_params2, \@set_params3), where
#
#	$n_D1 is the number of sites that sequence 1 has in between the PASs;
#	Sequence 2 has a run of gaps flanked by the sites homologous to the $i th and the $j th sites in sequence 1.
#	(And sequence 3 has a run of gaps aligned with the whole run of $n_D1 sites in sequence 1.)
#
#	@set_params1/2/3 = ($b1/2/3, $lambda_I1/2/3, $lambda_D1/2/3, \@freqs_I1/2/3, \@freqs_D1/2/3, $cutoff_I1/2/3, $cutoff_D1/2/3)
#	are indel model parameters along the branch between the root and sequence 1/2/3,
#	with
#	$b1/2/3 is the branch length,
#	$lambda_I1/2/3 and $lambda_D1/2/3 are the total insertion rate and the total deletion rate, respectively, per unit branch length and per site;
#	$freqs_I1/2/3[$l] ($l=1,...,$cutoff_I1/2/3) is the relative frequency of $l-residue insertions;
#	$freqs_D1/2/3[$l] ($l=1,...,$cutoff_D1/2/3) is the relative frequency of $l-residue deletions.
#
# OUTPUT: $tmu_P_G34, which is the multiplication factor we desire.
#
#	The subroutine returns undef if it encounters an error.
#
sub tld_mu_P_msa_G34 ($$$\@\@\@) {

	my ($n_D1, $i, $j, $set_params1, $set_params2, $set_params3) = @_;

	my ($b1, $lambda_I1, $lambda_D1, $freqs_I1, $freqs_D1, $cutoff_I1, $cutoff_D1) = @{$set_params1};
	my ($b2, $lambda_I2, $lambda_D2, $freqs_I2, $freqs_D2, $cutoff_I2, $cutoff_D2) = @{$set_params2};
	my ($b3, $lambda_I3, $lambda_D3, $freqs_I3, $freqs_D3, $cutoff_I3, $cutoff_D3) = @{$set_params3};

	my $l_I1 = $j - $i - 1;
	my $l_D3 = $n_D1 - $l_I1;

	unless (($l_I1 > 0) and ($l_D3 > 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_msa_G34'] (A) NEGATIVE INDEL LENGTH(S) found: (l_I1, l_D3) = ($l_I1, $l_D3): $! !\n\n";
	    return undef;
	} 

	unless (($l_I1 <= $cutoff_I1) and ($l_D3 <= 2 * $cutoff_D3)) { # This history has the occurrence probability = ZERO under the given indel model.
	   return 0;
	}

	my $lambda1 = $lambda_I1 + $lambda_D1;
	my $lambda3 = $lambda_I3 + $lambda_D3;

	my $mu_P_I1 = $lambda_I1 * $freqs_I1->[$l_I1] * (1 - exp (-$lambda1 * $l_I1 * $b1)) /($lambda1 * $l_I1) ;

	my ($tmu_P_D3_2ev, $components_D3_2ev) = mfac_asites_vs_dgaps_2events ($l_D3, $b3, $lambda_I3, $lambda_D3, @{$freqs_I3}, @{$freqs_D3}, $cutoff_I3, $cutoff_D3);
	unless (defined $tmu_P_D3_2ev) {
	    print STDOUT "[ERROR 'tld_mu_P_msa_G34'] Subroutine 'mfac_asites_vs_dgaps_2events' FAILED: $! !\n\n";
	    return undef;
	}
	##my $mu_P_D3 = $lambda_D3 * $freqs_D3->[$l_D3] * (exp ($lambda3 * $l_D3 * $b3) - 1) /($lambda3 * $l_D3);

	my $tmu_P_G34 = $mu_P_I1 * $tmu_P_D3_2ev ;

	return $tmu_P_G34 ;

}	# End of the subroutine 'tld_mu_P_msa_G34'. #


# tld_mu_P_msa_H12:
#	This satellite subroutine of 'mfac_dgaps2_nested_in_dgaps3_3events'
#	calculates the \tilde{\mu}_P[(H-1)] + \tilde{\mu}_P[(H-2)], which is the portion of the multiplication factor contributed by
#	the 3-event histories each of which consists of a 2-event history equivalent to a deletion along branch 2
#	and a deletion along branch 3.
#	The root state in this case is identical to that of sequence 1, i.e., [L, 1, ..., \Delta L^{D1}, R].
#
#	The portion is calculated according to Eq.(SR-9.4e,f) in Supplementary Results.
#
#	[CREATED on May 9 (Sat), 2015.]
#
# INPUT: ($n_D1, $i, $j, \@set_params1, \@set_params2, \@set_params3), where
#
#	$n_D1 is the number of sites that sequence 1 has in between the PASs;
#	Sequence 2 has a run of gaps flanked by the sites homologous to the $i th and the $j th sites in sequence 1.
#	(And sequence 3 has a run of gaps aligned with the whole run of $n_D1 sites in sequence 1.)
#
#	@set_params1/2/3 = ($b1/2/3, $lambda_I1/2/3, $lambda_D1/2/3, \@freqs_I1/2/3, \@freqs_D1/2/3, $cutoff_I1/2/3, $cutoff_D1/2/3)
#	are indel model parameters along the branch between the root and sequence 1/2/3,
#	with
#	$b1/2/3 is the branch length,
#	$lambda_I1/2/3 and $lambda_D1/2/3 are the total insertion rate and the total deletion rate, respectively, per unit branch length and per site;
#	$freqs_I1/2/3[$l] ($l=1,...,$cutoff_I1/2/3) is the relative frequency of $l-residue insertions;
#	$freqs_D1/2/3[$l] ($l=1,...,$cutoff_D1/2/3) is the relative frequency of $l-residue deletions.
#
# OUTPUT: $tmu_P_H12, which is the multiplication factor we desire.
#
#	The subroutine returns undef if it encounters an error.
#
sub tld_mu_P_msa_H12 ($$$\@\@\@) {

	my ($n_D1, $i, $j, $set_params1, $set_params2, $set_params3) = @_;

	my ($b1, $lambda_I1, $lambda_D1, $freqs_I1, $freqs_D1, $cutoff_I1, $cutoff_D1) = @{$set_params1};
	my ($b2, $lambda_I2, $lambda_D2, $freqs_I2, $freqs_D2, $cutoff_I2, $cutoff_D2) = @{$set_params2};
	my ($b3, $lambda_I3, $lambda_D3, $freqs_I3, $freqs_D3, $cutoff_I3, $cutoff_D3) = @{$set_params3};

	my $l_D2 = $j - $i - 1;
	my $l_D3 = $n_D1;

	unless (($l_D2 > 0) and ($l_D3 > 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_msa_H12'] (A) NEGATIVE INDEL LENGTH(S) found: (l_D2, l_D3) = ($l_D2, $l_D3): $! !\n\n";
	    return undef;
	} 

	unless (($l_D2 <= 2*$cutoff_D2) and ($l_D3 <= $cutoff_D3)) { # This history has the occurrence probability = ZERO under the given indel model.
	   return 0;
	}

	my $lambda1 = $lambda_I1 + $lambda_D1;
	my $lambda2 = $lambda_I2 + $lambda_D2;
	my $lambda3 = $lambda_I3 + $lambda_D3;

	my ($tmu_P_D2_2ev, $components_D2_2ev) = mfac_asites_vs_dgaps_2events ($l_D2, $b2, $lambda_I2, $lambda_D2, @{$freqs_I2}, @{$freqs_D2}, $cutoff_I2, $cutoff_D2);
	unless (defined $tmu_P_D2_2ev) {
	    print STDOUT "[ERROR 'tld_mu_P_msa_H12'] Subroutine 'mfac_asites_vs_dgaps_2events' FAILED: $! !\n\n";
	    return undef; 
	}
	##my $mu_P_D2 = $lambda_D2 * $freqs_D2->[$l_D2] * (exp ($lambda2 * $l_D2 * $b2) - 1) /($lambda2 * $l_D2) ;

	my $mu_P_D3 = $lambda_D3 * $freqs_D3->[$l_D3] * (exp ($lambda3 * $l_D3 * $b3) - 1) /($lambda3 * $l_D3);

	my $tmu_P_H12 = exp (-($lambda1*$b1 + $lambda2*$b2 + $lambda3*$b3) * $l_D2) * $tmu_P_D2_2ev * $mu_P_D3 ;

	return $tmu_P_H12 ;

}	# End of the subroutine 'tld_mu_P_msa_H12'. #


# tld_mu_P_msa_H34:
#	This satellite subroutine of 'mfac_dgaps2_nested_in_dgaps3_3events'
#	calculates the \tilde{\mu}_P[(H-3)] + \tilde{\mu}_P[(H-4)], which is the portion of the multiplication factor contributed by
#	the 3-event histories each of which consists of a deletion along branch 2
#	and a 2-event history equivalent to a deletion along branch 3.
#	The root state in this case is identical to that of sequence 1, i.e., [L, 1, ..., \Delta L^{D1}, R].
#
#	The portion is calculated according to Eq.(SR-9.4g,h) in Supplementary Results.
#
#	[CREATED on May 9 (Sat), 2015.]
#
# INPUT: ($n_D1, $i, $j, \@set_params1, \@set_params2, \@set_params3), where
#
#	$n_D1 is the number of sites that sequence 1 has in between the PASs;
#	Sequence 2 has a run of gaps flanked by the sites homologous to the $i th and the $j th sites in sequence 1.
#	(And sequence 3 has a run of gaps aligned with the whole run of $n_D1 sites in sequence 1.)
#
#	@set_params1/2/3 = ($b1/2/3, $lambda_I1/2/3, $lambda_D1/2/3, \@freqs_I1/2/3, \@freqs_D1/2/3, $cutoff_I1/2/3, $cutoff_D1/2/3)
#	are indel model parameters along the branch between the root and sequence 1/2/3,
#	with
#	$b1/2/3 is the branch length,
#	$lambda_I1/2/3 and $lambda_D1/2/3 are the total insertion rate and the total deletion rate, respectively, per unit branch length and per site;
#	$freqs_I1/2/3[$l] ($l=1,...,$cutoff_I1/2/3) is the relative frequency of $l-residue insertions;
#	$freqs_D1/2/3[$l] ($l=1,...,$cutoff_D1/2/3) is the relative frequency of $l-residue deletions.
#
# OUTPUT: $tmu_P_H34, which is the multiplication factor we desire.
#
#	The subroutine returns undef if it encounters an error.
#
sub tld_mu_P_msa_H34 ($$$\@\@\@) {

	my ($n_D1, $i, $j, $set_params1, $set_params2, $set_params3) = @_;

	my ($b1, $lambda_I1, $lambda_D1, $freqs_I1, $freqs_D1, $cutoff_I1, $cutoff_D1) = @{$set_params1};
	my ($b2, $lambda_I2, $lambda_D2, $freqs_I2, $freqs_D2, $cutoff_I2, $cutoff_D2) = @{$set_params2};
	my ($b3, $lambda_I3, $lambda_D3, $freqs_I3, $freqs_D3, $cutoff_I3, $cutoff_D3) = @{$set_params3};

	my $l_D2 = $j - $i - 1;
	my $l_D3 = $n_D1;

	unless (($l_D2 > 0) and ($l_D3 > 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_msa_H34'] (A) NEGATIVE INDEL LENGTH(S) found: (l_D2, l_D3) = ($l_D2, $l_D3): $! !\n\n";
	    return undef;
	} 

	unless (($l_D2 <= $cutoff_D2) and ($l_D3 <= 2 * $cutoff_D3)) { # This history has the occurrence probability = ZERO under the given indel model.
	   return 0;
	}

	my $lambda1 = $lambda_I1 + $lambda_D1;
	my $lambda2 = $lambda_I2 + $lambda_D2;
	my $lambda3 = $lambda_I3 + $lambda_D3;

	my $mu_P_D2 = $lambda_D2 * $freqs_D2->[$l_D2] * (exp ($lambda2 * $l_D2 * $b2) - 1) /($lambda2 * $l_D2) ;

	my ($tmu_P_D3_2ev, $components_D3_2ev) = mfac_asites_vs_dgaps_2events ($l_D3, $b3, $lambda_I3, $lambda_D3, @{$freqs_I3}, @{$freqs_D3}, $cutoff_I3, $cutoff_D3);
	unless (defined $tmu_P_D3_2ev) {
	    print STDOUT "[ERROR 'tld_mu_P_msa_H34'] Subroutine 'mfac_asites_vs_dgaps_2events'] FAILED: $! !\n\n";
	    return undef;
	}
	##my $mu_P_D3 = $lambda_D3 * $freqs_D3->[$l_D3] * (exp ($lambda3 * $l_D3 * $b3) - 1) /($lambda3 * $l_D3);

	my $tmu_P_H34 = exp (-($lambda1*$b1 + $lambda2*$b2 + $lambda3*$b3) * $l_D2) * $mu_P_D2 * $tmu_P_D3_2ev ;

	return $tmu_P_H34 ;

}	# End of the subroutine 'tld_mu_P_msa_H34'. #


# tld_mu_P_msa_I:
#	This satellite subroutine of 'mfac_dgaps2_nested_in_dgaps3_3events'
#	calculates the \tilde{\mu}_P[(I)],
#	which is the portion of the multiplication factor contributed by
#	the 3-event histories each of which consists of an insertion along branch 1,
#	a deletion along branch 2, and a deletion along branch 3. 
#	The root state in case (I) is of the form: [L, 1,...,i', j',... \Delta L^{D1}, R]
#	(with i <= i' , i'+2 <= j' <= j, but with (i', j') neq (i, j)).
#	
#	It was created by modifying the obsolete subroutine 'tld_mu_P_msa_I_J'.
#
#	[CREATED as a consequence of the correction of Section SR-9 of Supplementary Results (on May 11 (Mon), 2015).]
#
# tld_mu_P_msa_I_J:
#	This satellite subroutine of 'mfac_dgaps2_nested_in_dgaps3_3events'
#	calculates the \tilde{\mu}_P[(I)] = \tilde{\mu}_P[(J)],
#	each of which is the portion of the multiplication factor contributed by
#	the 3-event histories each of which consists of an insertion along branch 1,
#	a deletion along branch 2, and a deletion along branch 3. 
#	The root state in case (I) is of the form: [L, 1,...,i', j,... \Delta L^{D1}, R] (with i+1 <= i' <= j-2).
#	The root state in case (J) is of the form: [L, 1,...,i, j',... \Delta L^{D1}, R] (with i+2 <= j' <= j-1).
#
#	The portion is calculated according to Eq.(SR-9.4i,j) in Supplementary Results.
#
#	[CREATED on May 9 (Sat), 2015.]
#
# INPUT: ($n_D1, $i, $j, \@set_params1, \@set_params2, \@set_params3), where
#
#	$n_D1 is the number of sites that sequence 1 has in between the PASs;
#	Sequence 2 has a run of gaps flanked by the sites homologous to the $i th and the $j th sites in sequence 1.
#	(And sequence 3 has a run of gaps aligned with the whole run of $n_D1 sites in sequence 1.)
#
#	@set_params1/2/3 = ($b1/2/3, $lambda_I1/2/3, $lambda_D1/2/3, \@freqs_I1/2/3, \@freqs_D1/2/3, $cutoff_I1/2/3, $cutoff_D1/2/3)
#	are indel model parameters along the branch between the root and sequence 1/2/3,
#	with
#	$b1/2/3 is the branch length,
#	$lambda_I1/2/3 and $lambda_D1/2/3 are the total insertion rate and the total deletion rate, respectively, per unit branch length and per site;
#	$freqs_I1/2/3[$l] ($l=1,...,$cutoff_I1/2/3) is the relative frequency of $l-residue insertions;
#	$freqs_D1/2/3[$l] ($l=1,...,$cutoff_D1/2/3) is the relative frequency of $l-residue deletions.
#
# OUTPUT: $tmu_P_I, which is the multiplication factor we desire.
#
#	The subroutine returns undef if it encounters an error.
#
sub tld_mu_P_msa_I ($$$\@\@\@) {

	my ($n_D1, $i, $j, $set_params1, $set_params2, $set_params3) = @_;

	my ($b1, $lambda_I1, $lambda_D1, $freqs_I1, $freqs_D1, $cutoff_I1, $cutoff_D1) = @{$set_params1};
	my ($b2, $lambda_I2, $lambda_D2, $freqs_I2, $freqs_D2, $cutoff_I2, $cutoff_D2) = @{$set_params2};
	my ($b3, $lambda_I3, $lambda_D3, $freqs_I3, $freqs_D3, $cutoff_I3, $cutoff_D3) = @{$set_params3};

	my $l_D2 = $j - $i - 1;
	my $l_D3 = $n_D1;
	##my $l_I1 = $j - $i - 1;
	##my $l_D3 = $n_D1 - $l_I1;

	unless (($l_D2 > 0) and ($l_D3 > 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_msa_I'] (A) NEGATIVE INDEL LENGTH(S) found: (l_D2, l_D3) = ($l_D2, $l_D3): $! !\n\n";
	    return undef;
	} 

	unless (($l_D2 <= $cutoff_I1 + $cutoff_D2) and ($l_D3 <= $cutoff_I1 + $cutoff_D3)) { # In this case, no histories of this type can contribute non-zero probabilities.
	##unless (($l_I1 <= $cutoff_I1 + $cutoff_D2) and ($l_D3 <= $cutoff_D3)) { # In this case, no histories of this type can contribute non-zero probabilities.
	    return 0;
	}

	my $lambda1 = $lambda_I1 + $lambda_D1;
	my $lambda2 = $lambda_I2 + $lambda_D2;
	my $lambda3 = $lambda_I3 + $lambda_D3;

	my $lambda_t = $lambda1 * $b1 + $lambda2 * $b2 + $lambda3 * $b3 ;

	my $tmu_P_I = 0;

	for (my $k=1; $k <= $l_D2 - 1; $k++) { # Main for-loop (over $k = $jp - $ip -1).
	##for (my $k=1; $k <= $j-$i-2; $k++) { # Main for-loop (over $k = $ip - $i = $j - $jp).

	    my $l_I1p = $k;
	    my $l_D2p = $l_D2 - $k;
	    my $l_D3p = $l_D3 - $k;
	    ##my $l_I1p = $l_I1 - $k;
	    ##my $l_D2p = $k;
	    ##my $l_D3p = $l_D3 + $k;

	    (($l_I1p <= $cutoff_I1) and ($l_D2p <= $cutoff_D2) and ($l_D3p <= $cutoff_D3)) or next; # History cannot contribute a probability unless it satisfies these conditions.

	    my $mu_I1 = $lambda_I1 * $freqs_I1->[$l_I1p] * (1 - exp (-$lambda1 * $l_I1p * $b1)) /($lambda1*$l_I1p) ;
	    my $mu_D2 = $lambda_D2 * $freqs_D2->[$l_D2p] * (exp ($lambda2 * $l_D2p * $b2) - 1) /($lambda2*$l_D2p) ;
	    my $mu_D3 = $lambda_D3 * $freqs_D3->[$l_D3p] * (exp ($lambda3 * $l_D3p * $b3) - 1) /($lambda3*$l_D3p) ;

	    my $exp = exp (-$lambda_t * $l_D2p) ;
	    ##my $exp = exp (-$lambda_t * $k) ;

		# MODIFIED as a consequence of the correction of Section SR-9 of Supplementary Results (on May 11 (Mon), 2015). #
	    $tmu_P_I += ($j - $i - $k) * $exp * $mu_I1 * $mu_D2 * $mu_D3 ; # Take account of the number of histories with the identical probability. #
	    ##$tmu_P_I += $exp * $mu_I1 * $mu_D2 * $mu_D3 ;


	}		# End of the main for-loop (over $k= $jp - $ip - 1). #


	return $tmu_P_I ;

}	# End of the subroutine 'tld_mu_P_msa_I'. #


=DISABLED as a consequence of the correction of Section SR-9 of Supplementary Results (on May 11 (Mon), 2015). #

# tld_mu_P_msa_I_J:
#	This satellite subroutine of 'mfac_dgaps2_nested_in_dgaps3_3events'
#	calculates the \tilde{\mu}_P[(I)] = \tilde{\mu}_P[(J)],
#	each of which is the portion of the multiplication factor contributed by
#	the 3-event histories each of which consists of an insertion along branch 1,
#	a deletion along branch 2, and a deletion along branch 3. 
#	The root state in case (I) is of the form: [L, 1,...,i', j,... \Delta L^{D1}, R] (with i+1 <= i' <= j-2).
#	The root state in case (J) is of the form: [L, 1,...,i, j',... \Delta L^{D1}, R] (with i+2 <= j' <= j-1).
#
#	The portion is calculated according to Eq.(SR-9.4i,j) in Supplementary Results.
#
#	[CREATED on May 9 (Sat), 2015.]
#
# INPUT: ($n_D1, $i, $j, \@set_params1, \@set_params2, \@set_params3), where
#
#	$n_D1 is the number of sites that sequence 1 has in between the PASs;
#	Sequence 2 has a run of gaps flanked by the sites homologous to the $i th and the $j th sites in sequence 1.
#	(And sequence 3 has a run of gaps aligned with the whole run of $n_D1 sites in sequence 1.)
#
#	@set_params1/2/3 = ($b1/2/3, $lambda_I1/2/3, $lambda_D1/2/3, \@freqs_I1/2/3, \@freqs_D1/2/3, $cutoff_I1/2/3, $cutoff_D1/2/3)
#	are indel model parameters along the branch between the root and sequence 1/2/3,
#	with
#	$b1/2/3 is the branch length,
#	$lambda_I1/2/3 and $lambda_D1/2/3 are the total insertion rate and the total deletion rate, respectively, per unit branch length and per site;
#	$freqs_I1/2/3[$l] ($l=1,...,$cutoff_I1/2/3) is the relative frequency of $l-residue insertions;
#	$freqs_D1/2/3[$l] ($l=1,...,$cutoff_D1/2/3) is the relative frequency of $l-residue deletions.
#
# OUTPUT: $tmu_P_I_J, which is the multiplication factor we desire.
#
#	The subroutine returns undef if it encounters an error.
#
sub tld_mu_P_msa_I_J ($$$\@\@\@) {

	my ($n_D1, $i, $j, $set_params1, $set_params2, $set_params3) = @_;

	my ($b1, $lambda_I1, $lambda_D1, $freqs_I1, $freqs_D1, $cutoff_I1, $cutoff_D1) = @{$set_params1};
	my ($b2, $lambda_I2, $lambda_D2, $freqs_I2, $freqs_D2, $cutoff_I2, $cutoff_D2) = @{$set_params2};
	my ($b3, $lambda_I3, $lambda_D3, $freqs_I3, $freqs_D3, $cutoff_I3, $cutoff_D3) = @{$set_params3};

	my $l_I1 = $j - $i - 1;
	my $l_D3 = $n_D1 - $l_I1;

	unless (($l_I1 > 0) and ($l_D3 > 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_msa_I_J'] (A) NEGATIVE INDEL LENGTH(S) found: (l_I1, l_D3) = ($l_I1, $l_D3): $! !\n\n";
	    return undef;
	} 

	unless (($l_I1 <= $cutoff_I1 + $cutoff_D2) and ($l_D3 <= $cutoff_D3)) { # In this case, no histories of this type can contribute non-zero probabilities.
	    return 0;
	}

	my $lambda1 = $lambda_I1 + $lambda_D1;
	my $lambda2 = $lambda_I2 + $lambda_D2;
	my $lambda3 = $lambda_I3 + $lambda_D3;

	my $lambda_t = $lambda1 * $b1 + $lambda2 * $b2 + $lambda3 * $b3 ;

	my $tmu_P_I_J = 0;

	for (my $k=1; $k <= $j-$i-2; $k++) { # Main for-loop (over $k = $ip - $i = $j - $jp).

	    my $l_I1p = $l_I1 - $k;
	    my $l_D2p = $k;
	    my $l_D3p = $l_D3 + $k;

	    (($l_I1p <= $cutoff_I1) and ($l_D2p <= $cutoff_D2) and ($l_D3p <= $cutoff_D3)) or next; # History cannot contribute a probability unless it satisfies these conditions.

	    my $mu_I1 = $lambda_I1 * $freqs_I1->[$l_I1p] * (1 - exp (-$lambda1 * $l_I1p * $b1)) /($lambda1*$l_I1p) ;
	    my $mu_D2 = $lambda_D2 * $freqs_D2->[$l_D2p] * (exp ($lambda2 * $l_D2p * $b2) - 1) /($lambda2*$l_D2p) ;
	    my $mu_D3 = $lambda_D3 * $freqs_D3->[$l_D3p] * (exp ($lambda3 * $l_D3p * $b3) - 1) /($lambda3*$l_D3p) ;

	    my $exp = exp (-$lambda_t * $k) ;

	    $tmu_P_I_J += $exp * $mu_I1 * $mu_D2 * $mu_D3 ;


	}		# End of the main for-loop (over $k= $ip - $i = $j - $jp). #


	return $tmu_P_I_J ;

}	# End of the subroutine 'tld_mu_P_msa_I_J'. #

=cut # END of "DISABLED as a consequence of the correction of Section SR-9 of Supplementary Results (on May 11 (Mon), 2015)." #


# tld_mu_P_msa_J:
#	The subroutine was re-labeled as a consequence of the correction of Section SR-9 of Supplementary Results
#	(on May 11 (Mon), 2015).
#
# tld_mu_P_msa_K:
#	This satellite subroutine of 'mfac_dgaps2_nested_in_dgaps3_3events'
#	calculates the \tilde{\mu}_P[(K (=> now J))], which is the portion of the multiplication factor contributed by
#	the 3-event histories each of which consists of a deletion along branch 1,
#	a deletion along branch 2, and a deletion along branch 3. 
#	The root state in this case is of the form: [L, 1,...,k, v'1,...,v'l, k+1,... \Delta L^{D1}, R]
#	(with i <= k <= j-1; l=1,...,min{L_{D1}^{CO}}, L_{D2}^{CO} - (j-i-1), L_{D3}^{CO} - \Delta L^{D1}}).
#
#	The portion is calculated according to Eq.(SR-9.4k) in Supplementary Results.
#
#	[CREATED on May 9 (Sat), 2015.]
#
# INPUT: ($n_D1, $i, $j, \@set_params1, \@set_params2, \@set_params3), where
#
#	$n_D1 is the number of sites that sequence 1 has in between the PASs;
#	Sequence 2 has a run of gaps flanked by the sites homologous to the $i th and the $j th sites in sequence 1.
#	(And sequence 3 has a run of gaps aligned with the whole run of $n_D1 sites in sequence 1.)
#
#	@set_params1/2/3 = ($b1/2/3, $lambda_I1/2/3, $lambda_D1/2/3, \@freqs_I1/2/3, \@freqs_D1/2/3, $cutoff_I1/2/3, $cutoff_D1/2/3)
#	are indel model parameters along the branch between the root and sequence 1/2/3,
#	with
#	$b1/2/3 is the branch length,
#	$lambda_I1/2/3 and $lambda_D1/2/3 are the total insertion rate and the total deletion rate, respectively, per unit branch length and per site;
#	$freqs_I1/2/3[$l] ($l=1,...,$cutoff_I1/2/3) is the relative frequency of $l-residue insertions;
#	$freqs_D1/2/3[$l] ($l=1,...,$cutoff_D1/2/3) is the relative frequency of $l-residue deletions.
#
# [ OUTPUT: $tmu_P_K, which is the multiplication factor we desire.] --- OBSOLETE as of May 11 (Mon), 2015. 
# OUTPUT: $tmu_P_J, which is the multiplication factor we desire.
#
#	The subroutine returns undef if it encounters an error.
#
##sub tld_mu_P_msa_K ($$$\@\@\@) {...} # OBSOLETE as of May 11 (Mon), 2015. #
sub tld_mu_P_msa_J ($$$\@\@\@) {

	my ($n_D1, $i, $j, $set_params1, $set_params2, $set_params3) = @_;

	my ($b1, $lambda_I1, $lambda_D1, $freqs_I1, $freqs_D1, $cutoff_I1, $cutoff_D1) = @{$set_params1};
	my ($b2, $lambda_I2, $lambda_D2, $freqs_I2, $freqs_D2, $cutoff_I2, $cutoff_D2) = @{$set_params2};
	my ($b3, $lambda_I3, $lambda_D3, $freqs_I3, $freqs_D3, $cutoff_I3, $cutoff_D3) = @{$set_params3};

	my $l_D2 = $j - $i - 1;
	my $l_D3 = $n_D1;

	unless (($l_D2 > 0) and ($l_D3 > 0)) {
	    print STDOUT "[ERROR 'tld_mu_P_msa_J'] (A) NEGATIVE INDEL LENGTH(S) found: (l_D2, l_D3) = ($l_D2, $l_D3): $! !\n\n";
	    return undef;
	} 

	if ($i > $j-1) { # In this case, there is no position where exessive sutes can be inserted.
	    return 0;
	}

	my $cutoff = ($cutoff_D1 < $cutoff_D2 - $l_D2) ? $cutoff_D1 : $cutoff_D2 - $l_D2 ;
	($cutoff > $cutoff_D3 - $l_D3) and ($cutoff = $cutoff_D3 - $l_D3); 

	unless ($cutoff > 0) { # In this case, no history of this type can contribute a probability.
	    return 0;
	}

	my $lambda1 = $lambda_I1 + $lambda_D1;
	my $lambda2 = $lambda_I2 + $lambda_D2;
	my $lambda3 = $lambda_I3 + $lambda_D3;

	my $lambda_t = $lambda1 * $b1 + $lambda2 * $b2 + $lambda3 * $b3 ;

	my $tmu_P_J = 0;

	for (my $l=1; $l <= $cutoff; $l++) { # Main for-loop (over the excessive length, $l). 

	    my $l_D1p = $l;
	    my $l_D2p = $l_D2 + $l;
	    my $l_D3p = $l_D3 + $l;

	    my $mu_D1 = $lambda_D1 * $freqs_D1->[$l_D1p] * (exp ($lambda1 * $l_D1p * $b1) - 1) /($lambda1*$l_D1p) ;
	    my $mu_D2 = $lambda_D2 * $freqs_D2->[$l_D2p] * (exp ($lambda2 * $l_D2p * $b2) - 1) /($lambda2*$l_D2p) ;
	    my $mu_D3 = $lambda_D3 * $freqs_D3->[$l_D3p] * (exp ($lambda3 * $l_D3p * $b3) - 1) /($lambda3*$l_D3p) ;

	    my $exp = exp (-$lambda_t * $l_D2p) ;

	    $tmu_P_J += $exp * $mu_D1 * $mu_D2 * $mu_D3 ;


	}		# End of the main for-loop (over the excessive length, $l).

	$tmu_P_J *= ($j - $i) ; # Multiply it by the number of possible positions where the run of excessive sites are inserted.


	return $tmu_P_J ;

}	# End of the subroutine 'tld_mu_P_msa_J'. #


		### RESTART FROM HERE ###


# mfac_dgaps2_nested_in_dgaps3_3events:
#	This subroutine calculates the portion of the multiplication factor contributed by the 3-event (i.e., next-parsimonious) local indel histories.
#	The histories are suppposed to be consistent with a run of $n_D1 gaps in sequence 3 and a run of ($j - $i-1) gaps in sequence 2;
#	the latter is nested in the former, in between a pair of PASs. This local MSA belongs to case (IV).
#	And the histories are considered under a 3-OTU tree whose internal node is regarded as the root.
#	
#	The subroutine uses the formula, Eq.(SR-9.3) supplemented with Eqs.(SR-9.4a-k).
#
#	[CREATED on May 9 (Sat), 2015.]
#
#
# INPUT: ($n_D1, $i, $j, \@set_params1, \@set_params2, \@set_params3), where
#
#	$n_D1 is the number of sites that sequence 1 has in between the PASs;
#	Sequence 2 has a run of gaps flanked by the sites homologous to the $i th and the $j th sites in sequence 1.
#	(And sequence 3 has a run of gaps aligned with the whole run of $n_D1 sites in sequence 1.)
#
#	@set_params1/2/3 = ($b1/2/3, $lambda_I1/2/3, $lambda_D1/2/3, \@freqs_I1/2/3, \@freqs_D1/2/3, $cutoff_I1/2/3, $cutoff_D1/2/3)
#	are indel model parameters along the branch between the root and sequence 1/2/3,
#	with
#	$b1/2/3 is the branch length,
#	$lambda_I1/2/3 and $lambda_D1/2/3 are the total insertion rate and the total deletion rate, respectively, per unit branch length and per site;
#	$freqs_I1/2/3[$l] ($l=1,...,$cutoff_I1/2/3) is the relative frequency of $l-residue insertions;
#	$freqs_D1/2/3[$l] ($l=1,...,$cutoff_D1/2/3) is the relative frequency of $l-residue deletions.
#
# OUTPUT: ($mfac_3ev, \@components), where
#	$mfac_3ev is the portion of the multiplication factor we desire;
#	@components stores the components of $mfac_3ev.
#
#	The subroutine returns undef if it encounters an error.
#
sub mfac_dgaps2_nested_in_dgaps3_3events ($$$\@\@\@) {

	my ($n_D1, $i, $j, $set_params1, $set_params2, $set_params3) = @_;

	unless (($n_D1 > 1) and (0 <= $i) and ($i+1 < $j) and ($j <= $n_D1+1)) {
	    print STDOUT "[ERROR 'mfac_dgaps2_nested_in_dgaps3_3events'] ILLEGAL INPUT: ", '($n_D1, $i, $j) = ', "($n_D1, $i, $j) : $! !\n\n";
	    return (undef, undef); 
	}

	unless (($i != 0) or ($j != $n_D1+1)) {
	    print STDOUT "[ERROR 'mfac_dgaps2_nested_in_dgaps3_3events'] ILLEGAL INPUT: ", '($n_D1, $i, $j) = ', "($n_D1, $i, $j) : $! !\n\n";
	    return (undef, undef); 
	}


	my $tmu_P_G12 = tld_mu_P_msa_G12 ($n_D1, $i, $j, @{$set_params1}, @{$set_params2}, @{$set_params3}) ;

	unless (defined $tmu_P_G12) {
	    print STDOUT "[ERROR 'mfac_dgaps2_nested_in_dgaps3_3events'] The subroutine 'tld_mu_P_msa_G12' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my $tmu_P_G34 = tld_mu_P_msa_G34 ($n_D1, $i, $j, @{$set_params1}, @{$set_params2}, @{$set_params3}) ;

	unless (defined $tmu_P_G34) {
	    print STDOUT "[ERROR 'mfac_dgaps2_nested_in_dgaps3_3events'] The subroutine 'tld_mu_P_msa_G34' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my $tmu_P_H12 = tld_mu_P_msa_H12 ($n_D1, $i, $j, @{$set_params1}, @{$set_params2}, @{$set_params3}) ;

	unless (defined $tmu_P_H12) {
	    print STDOUT "[ERROR 'mfac_dgaps2_nested_in_dgaps3_3events'] The subroutine 'tld_mu_P_msa_H12' FAILED: $! !\n\n";
	    return (undef, undef);
	}

	my $tmu_P_H34 = tld_mu_P_msa_H34 ($n_D1, $i, $j, @{$set_params1}, @{$set_params2}, @{$set_params3}) ;

	unless (defined $tmu_P_H34) {
	    print STDOUT "[ERROR 'mfac_dgaps2_nested_in_dgaps3_3events'] The subroutine 'tld_mu_P_msa_H34' FAILED: $! !\n\n";
	    return (undef, undef);
	}

		# MODIFIED accompanying the correction of Section SR-9 of Supplementary Results (on May 11 (Mon), 2015). #
		#
	my $tmu_P_I = tld_mu_P_msa_I ($n_D1, $i, $j, @{$set_params1}, @{$set_params2}, @{$set_params3}) ;
	##my $tmu_P_I = my $tmu_P_J = tld_mu_P_msa_I_J ($n_D1, $i, $j, @{$set_params1}, @{$set_params2}, @{$set_params3}) ;

	unless (defined $tmu_P_I) {
	    print STDOUT "[ERROR 'mfac_dgaps2_nested_in_dgaps3_3events'] The subroutine 'tld_mu_P_msa_I' FAILED: $! !\n\n";
	    return (undef, undef);
	}

		# MODIFIED accompanying the correction of Section SR-9 of Supplementary Results (on May 11 (Mon), 2015). #
		#
	my $tmu_P_J = tld_mu_P_msa_J ($n_D1, $i, $j, @{$set_params1}, @{$set_params2}, @{$set_params3}) ;
	##my $tmu_P_K = tld_mu_P_msa_K ($n_D1, $i, $j, @{$set_params1}, @{$set_params2}, @{$set_params3}) ;

	unless (defined $tmu_P_J) {
	    print STDOUT "[ERROR 'mfac_dgaps2_nested_in_dgaps3_3events'] The subroutine 'tld_mu_P_msa_J' FAILED: $! !\n\n";
	    return (undef, undef);
	}

		# MODIFIED accompanying the correction of Section SR-9 of Supplementary Results (on May 11 (Mon), 2015). #
		#
	my $mfac_3ev = $tmu_P_G12 + $tmu_P_G34 + $tmu_P_H12 + $tmu_P_H34 + $tmu_P_I + $tmu_P_J;
	my @components = ($tmu_P_G12, $tmu_P_G34, $tmu_P_H12, $tmu_P_H34, $tmu_P_I, $tmu_P_J);
	##my $mfac_3ev = $tmu_P_G12 + $tmu_P_G34 + $tmu_P_H12 + $tmu_P_H34 + $tmu_P_I + $tmu_P_J + $tmu_P_K ;
	##my @components = ($tmu_P_G12, $tmu_P_G34, $tmu_P_H12, $tmu_P_H34, $tmu_P_I, $tmu_P_J, $tmu_P_K);

	return ($mfac_3ev, \@components) ;


}	# End of the subroutine 'mfac_dgaps2_nested_in_dgaps3_3events'. #


1;	# a "spell" which makes the module "use"-able and/or "require"-able #


