package MyPhyML_parser;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.3: Copyright (C) 2013 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyPhyML_parser.pm:
#	This module manipulates various inputs/outputs of the 'phyml' program.
#
#	REVISED on May 12 (Thu), 2011:
#	(1) Now, information outputs are put through STDOUT, instead of STDERR as before the revision.
#	(2) The amount of information outputs can be adjusted by the argument $debug_level.
#	  (As $debug_level is larger, more information will be given.)
#

use strict;
use warnings;

use Exporter;

our @ISA = qw(Exporter);

use MyFile_manipu qw(
	open_input_file open_input_file2  open_output_file
	);

our @EXPORT_OK = qw(
	est_resfreqs_frm_msa
	print_custom_sbst_model
	extract_phyml_statistics
	extract_phyml_statistics2
	fetch_clmwise_likelihoods_phyml
	) ;


	# A Hard-coded definition of the JTT model (extracted from the PAML package) #
	# ADDED accompanying the creation of 'alt_ml_score.pl' on May 2 (Mon), 2011 #
our $JTT_SBST_MODEL = <<END_JTT;
                                                                           
 58                                                                        
 54  45                                                                    
 81  16 528                                                                
 56 113  34  10                                                            
 57 310  86  49   9                                                        
105  29  58 767   5 323                                                    
179 137  81 130  59  26 119                                                
 27 328 391 112  69 597  26  23                                            
 36  22  47  11  17   9  12   6  16                                        
 30  38  12   7  23  72   9   6  56 229                                    
 35 646 263  26   7 292 181  27  45  21  14                                
 54  44  30  15  31  43  18  14  33 479 388  65                            
 15   5  10   4  78   4   5   5  40  89 248   4  43                        
194  74  15  15  14 164  18  24 115  10 102  21  16  17                    
378 101 503  59 223  53  30 201  73  40  59  47  29  92 285                
475  64 232  38  42  51  32  33  46 245  25 103 226  12 118 477            
  9 126   8   4 115  18  10  55   8   9  52  10  24  53   6  35  12        
 11  20  70  46 209  24   7   8 573  32  24   8  18 536  10  63  21  71    
298  17  16  31  62  20  45  47  11 961 180  14 323  62  23  38 112  25  16

0.076748 0.051691 0.042645 0.051544 0.019803 0.040752 0.061830
0.073152 0.022944 0.053761 0.091904 0.058676 0.023826 0.040126
0.050901 0.068765 0.058565 0.014261 0.032102 0.066005

END_JTT


# est_resfreqs_frm_msa:
#	This subroutine estimates the residue frequencies
#	from an input MSA.
#	It keeps the input MSA intact.
#
#	(CREATED in creation of 'compare_ml_msa.bch.extJC.6t.PhyML.pl' on Mar 28 (Mon), 2011)
#
# INPUT: (\@msa, $gap, \%res2freq, $fl_caseignore, $fl_relative), where
#	${$msa->[$k]} is the $k th sequence embedded in the input MSA;
#	$gap is a character representing a gap;
#	%res2freq = ($residue => $frequency, ...) is to be filled in;
#	$fl_caseignore = 1 if you identify lower-case letters with the corresponding upper-case letters (the frequencies will be recorded in terms of upper-base letters),
#	= 0 if you distinguish them;
#	$fl_relative = 1 to compute relative frequencies,
#	= 0 to compute absolute frequencies, or numbers, of the residues.
#
# OUTPUT: $totct_res, which is the total count of the residues.
#
#	It returns undef if it fails.
#
sub est_resfreqs_frm_msa (\@$\%$$) {

	my ($msa, $gap, $res2freq, $fl_caseignore, $fl_relative) = @_;

	$fl_caseignore and ($gap = uc $gap);

	my $ct_seqs = @{$msa};
	my $totct_res = 0;

	for (my $k=0; $k<$ct_seqs; $k++) { # Outermost for-loop #
	    my $cp_seq = ${$msa->[$k]};
	    $fl_caseignore and ($cp_seq = uc $cp_seq);

	    my $seq_len = length $cp_seq;
	    for (my $l=0; $l<$seq_len; $l++) { # Inner for-loop #

		my $res = substr ($cp_seq, 0, 1, '');
		($res eq $gap) and next;

		$totct_res++;

		(defined $res2freq->{$res}) or ($res2freq->{$res} = 0);
		$res2freq->{$res}++;

	    }			# End of the inner for-loop #

	}			# End of the outermost for-loop #

	if ($fl_relative) {
	    unless ($totct_res > 0) {
		print STDOUT "[WARNING 'est_resfreqs_frm_msa'] The total number of residues is ZENO, so I cannot calculate the relative frequencies: $! !\n\n";
		return 0;
	    }
	    foreach my $res (keys %{$res2freq}) { $res2freq->{$res} /= $totct_res; }
	}

	return $totct_res;

}	# End of the subroutine 'est_resfreqs_frm_msa' #


# print_custom_sbst_model:
#	This subroutine prints out a 'custom-made' amino-acid substitution model,
#	which is a combination of a pre-substitution matrix given by a source data file,
#	and user-defined amino-acid frequencies.
#	(CREATED in creation of 'compare_ml_msa.bch.extJC.6t.PhyML.pl' on Mar 29 (Tue), 2011)
#
# INPUT: (\$src_dat, \$outfile, \@aa_freqs, $resct_in_line, $endmark), where
#	$src_dat is the path to the source data file.
#	IT IS ASSUMED that the file follows the same format as PAML data files;
#	$outfile is the path to the output file;
#	@aa_freqs stores user-specified amino-acid frequencies;
#	$resct_in_line is the maximum number of amino-acid frequencies
#	to be displayed in a line;
#	$endmark is a string that marks the end of the substitution model in the output file.
#
# OUTPUT: 1 if the process succeeded,
#	  0 if it failed.
#
sub print_custom_sbst_model (\$\$\@$$) {

	my ($src_dat, $outfile, $aa_freqs, $resct_in_line, $endmark) = @_;
	local $/ = "\n";

	unless (20 == @{$aa_freqs}) {
	    print STDOUT "[ERROR 'print_custom_sbst_model'] The input array \@\{\$aa_freqs\} = (", join (', ', @{$aa_freqs}), ") has ", scalar (@{$aa_freqs}), " elements, which CLASHES with the number of amino acids (20): $! !\n";
	    return 0;
	}

	my $fh_tmpin = open_input_file2 (${$src_dat});
	unless (defined $fh_tmpin) {
	    print STDOUT "[ERROR 'print_custom_sbst_model'] 'open_input_file2' FAILED on the source data file '${$src_dat}': $! !\n";
	    return 0;
	}
	my $fh_tmpout = open_output_file (${$outfile}, 0);
	unless (defined $fh_tmpout) {
	    print STDOUT "[ERROR 'print_custom_sbst_model'] 'open_output_file' FAILED on the output file '${$outfile}': $! !\n";
	    return 0;
	}

	while (my $tmp = <$fh_tmpin>) { # The 1st while-loop #
	    print $fh_tmpout $tmp;
	    ($tmp =~ m/\S/) and last;
	}			# End of the 1st while-loop #

	while (my $tmp = <$fh_tmpin>) { # THe 2nd while-loop #
	    print $fh_tmpout $tmp;
	    ($tmp =~ m/\S/) or last;
	}			# End of the 2nd while-loop #
	close ($fh_tmpin);

	my $quotient = int (20/$resct_in_line);

	for (my $q = 0; $q <= $quotient; $q++) { # Outer for-loop #
	    my $lb = $q*$resct_in_line;
	    my $ub = $lb + $resct_in_line;
	    ($ub > 20) and ($ub = 20);
	    my @out = ();
	    for (my $a = $lb; $a<$ub; $a++) { push @out, $aa_freqs->[$a]; }
	    print $fh_tmpout join (' ', @out), "\n";
	}			# End of the outer for-loop #


	print $fh_tmpout "\n";

	($endmark ne '') and (print $fh_tmpout "\n", $endmark, "\n");

	close ($fh_tmpout);

	return 1;

}	# End of the subroutine 'print_custom_sbst_model' #



# extract_phyml_statistics:
#	This subroutine reads in an output statistics file of PhyML,
#	and extracts statistics recorded on it.
#	(CREATED in creation of 'compare_ml_msa.bch.extJC.6t.PhyML.pl' on Mar 29 (Tue), 2011)
#
# INPUT: (\$instat, \%stat_phyml, $debug_level), where
#	$instat is the path to an output statistics file;
#	%stat_phyml = ($key => $scalar_value ($key is the string between '.' and ':', which ' ' replaced with '_'),
#	   'discrete_gamma' => [$ncate, $alpha] (or undef),
#	   'nuc_freqs' => [$fA, $fC, $fG, $fT] (or undef),
#	   'rel_rates' => [$r_AC, $r_AG, $r_AT, $r_CG, $r_CT, $r_GT] (or undef),
#	   'rate_mtrx' => \@mtrx,
#	with $mtrx[$i]->[$j] giving the instantaneous rate from the $i-th to the $j-th nucleotide (in the order: A, C, G, T) (or undef)
#		), which is to be filled in while the subroutine progresses;
#	$debug_level controls the amount of information outputs.
#	A larger value give more information.
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
##sub extract_phyml_statistics (\$\%) {
##	my ($instat, $stat_phyml) = @_;
#
sub extract_phyml_statistics (\$\%$) {

	my ($instat, $stat_phyml, $debug_level) = @_;
	local $/ = ''; # Use blank lines as input-record-separators #

	my $fh_itmp = open_input_file2 (${$instat});
	unless (defined $fh_itmp) {
	    print STDOUT "[ERROR 'extract_phyml_statistics'] 'open_input_file2' FAILED on the input statistics file '${$instat}': $! !\n";
	    return 0;
	}

	while (my $tmpin = <$fh_itmp>) { # Main while-loop #

	    ($tmpin =~ m/\S/) or next; # Skip blank block (if any)
	    ($tmpin =~ m/^\s*o/) and next; # Skip the header and the footer #

	    unless ($tmpin =~ s/^\s*\.\s*//) {
		$tmpin =~ s/\s+$//;
		if ($debug_level>1) {
		    print STDOUT "[WARNING 'extract_phyml_statistics'] An UNEXTECTED BLOCK: $! !\n";
		    print STDOUT " '$tmpin'.\n";
		    print STDOUT " --- I will skip it ...\n\n";
		}
		next;
	    }
	    $tmpin =~ s/\s+$//;
	    my ($key, $value) = split /\s*\:\s*/, $tmpin, 2;

	    if ($key =~ m/^Discrete\s+gamma\s+model/) {
		if($value =~ m/^Yes/) {
		    my @lines = split /\n+/, $value;
		    unless (@lines == 3) {
			print STDOUT "[WARNING 'extract_phyml_statistics'] Found an ANOMALOUS BLOCK describing the discrete gamma model: $! !\n";
			print STDOUT "'$tmpin' .\n";
			print STDOUT "  --- I will skip it ...\n\n";
		        next;
		    }
		    my ($ncate) = ($lines[1] =~ m/(\S+)\s*$/);
		    my ($alpha) = ($lines[2] =~ m/(\S+)\s*$/);

		    $stat_phyml->{'discrete_gamma'} = [$ncate, $alpha];
		}

	    } elsif ($key =~ m/^Nucleotides\s+frequencies/) {

		(my $chunk = <$fh_itmp>) =~ s/\s+$//;
		my @lines = split /\n+/, $chunk;
		unless (@lines == 4) {
		    print STDOUT "[WARNING 'extract_phyml_statistics'] Found an ANOMALOUS BLOCK describing the nucleotide frequencies: $! !\n";
		    print STDOUT "'$chunk' .\n";
		    print STDOUT "  --- I will skip it ...\n\n";
		    next;
		}
		my @nfreqs = ();
		for (my $b=0; $b< 4; $b++) {
		    my ($freq) = ($lines[$b] =~ m/([^\s\=]+)\s*$/);
		    unless (defined $freq) {
			print STDOUT "[ERROR 'extract_phyml_statistics'] FAILED to extract the frequency from the ", $b + 1, " th line for the nucleotide fequencies: $! !\n";
			print STDOUT " '", $lines[$b], "' .\n\n";
			return 0;
		    }
		    push @nfreqs, $freq;
		}
		$stat_phyml->{'nuc_freqs'} = \@nfreqs;

	    } elsif ($key =~ m/^GTR\s+relative\s+rate\s+parameters/) {

		(my $chunk = <$fh_itmp>) =~ s/\s+$//;
		my @lines = split /\n+/, $chunk;
		unless (@lines == 6) {
		    print STDOUT "[WARNING 'extract_phyml_statistics'] Found an ANOMALOUS BLOCK describing the relative rate parameters: $! !\n";
		    print STDOUT "'$chunk' .\n";
		    print STDOUT "  --- I will skip it ...\n\n";
		    next;
		}
		my @rel_rates = ();
		for (my $b=0; $b< 6; $b++) {
		    my ($rrate) = ($lines[$b] =~ m/(\S+)\s*$/);
		    unless (defined $rrate) {
			print STDOUT "[ERROR 'extract_phyml_statistics'] FAILED to extract the relative rate from the ", $b + 1, " th line for the relative rate parameters: $! !\n";
			print STDOUT " '", $lines[$b], "' .\n\n";
			return 0;
		    }
		    push @rel_rates, $rrate;
		}
		$stat_phyml->{'rel_rates'} = \@rel_rates;

	    } elsif ($key =~ m/^Instantaneous\s+rate\s+matrix/) {

		(my $chunk = <$fh_itmp>) =~ s/\s+$//;
		my @lines = split /\n+/, $chunk;
		unless (@lines == 5) {
		    print STDERR "[WARNING 'extract_phyml_statistics'] Found an ANOMALOUS BLOCK describing the instantaneous rate matrix: $! !\n";
		    print STDERR "'$chunk' .\n";
		    print STDERR "  --- I will skip it ...\n\n";
		    next;
		}
		my @rate_mtrx = ();
		for (my $b=1; $b< 5; $b++) {
		    my @rates = split ' ', $lines[$b];
		    unless (@rates == 4) {
			print STDOUT "[ERROR 'extract_phyml_statistics'] ANOMALOUS INPUT: the ", $b + 1, " th line for the rate matrix has ", scalar (@rates), " elements: $! !\n";
			print STDOUT " '", $lines[$b], "' .\n\n";
			return 0;
		    }
		    push @rate_mtrx, \@rates;
		}
		$stat_phyml->{'rate_mtrx'} = \@rate_mtrx;

	    } elsif ((defined $value) and ($value =~ m/\S/)) {

		$key =~ s/\s/_/g;
		$stat_phyml->{$key} = $value;
	    }

	}			# End of the main while-loop #
	close ($fh_itmp);

	return 1;

}	# End of the subroutine 'extract_phyml_statistics' #


# extract_phyml_statistics2:
#	This revised version of 'extract_phyml_statistics' can handle
#	both new and old formats of the output statistics file of PhyML.
#	(CREATED in creation of 'compare_ml_msa.bch.extJC.6t.PhyML.pl' on Apr 8 (Fri), 2011)
#
# extract_phyml_statistics:
#	This subroutine reads in an output statistics file of PhyML,
#	and extracts statistics recorded on it.
#	(CREATED in creation of 'compare_ml_msa.bch.extJC.6t.PhyML.pl' on Mar 29 (Tue), 2011)
#
# INPUT: (\$instat, \%stat_phyml, $debug_level), where
#	$instat is the path to an output statistics file;
#	%stat_phyml = ($key => $scalar_value ($key is the string between '.' and ':', which ' ' replaced with '_'),
#	   'discrete_gamma' => [$ncate, $alpha] (or undef),
#	   'nuc_freqs' => [$fA, $fC, $fG, $fT] (or undef),
#	   'rel_rates' => [$r_AC, $r_AG, $r_AT, $r_CG, $r_CT, $r_GT] (or undef),
#	   'rate_mtrx' => \@mtrx,
#	with $mtrx[$i]->[$j] giving the instantaneous rate from the $i-th to the $j-th nucleotide (in the order: A, C, G, T) (or undef)
#		), which is to be filled in while the subroutine progresses;
#	$debug_level controls the amount of information outputs.
#	Its larger value gives more information.
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
##sub extract_phyml_statistics2 (\$\%) {
##	my ($instat, $stat_phyml) = @_;
#
sub extract_phyml_statistics2 (\$\%$) {

	my ($instat, $stat_phyml, $debug_level) = @_;

	# REVISED in creation of 'extract_phyml_statistics2' (on Apr 8 (Fri), 2011) #
	local $/ = "\n"; # Use a newline character as input-record-separators #
	##local $/ = ''; # Use blank lines as input-record-separators #
	local $| = 1; # Flush the output buffer everytime an output is given.

	my $fh_itmp = open_input_file2 (${$instat});
	unless (defined $fh_itmp) {
	    print STDOUT "[ERROR 'extract_phyml_statistics2'] 'open_input_file2' FAILED on the input statistics file '${$instat}': $! !\n";
	    return 0;
	}

		# Skip the header part #
		# (ADDED in creation of 'extract_phyml_statistics2'
		# (on Apr 8 (Fri), 2011))
	my $ct_header_lines = 0;
	while (my $tmpin = <$fh_itmp>) { # Preliminary while-loop #
	    ($tmpin =~ m/^\s*o/) or next;
	    $ct_header_lines++;
	    ($ct_header_lines == 2) and last;
	}		# End of the preliminary while-loop #

	# REVISED in creation of 'extract_phyml_statistics2' on Apr 8 (Fri), 2011 #
	my $tmpin = <$fh_itmp>;
	while (defined $tmpin) { # Main while-loop #
	##while (my $tmpin = <$fh_itmp>) { # Main while-loop #

	# REVISED in creation of 'extract_phyml_statistics2' (on Apr 8 (Fri), 2011) #
	    unless ($tmpin =~ m/\S/) { # Skip blank lines (if any)
		$tmpin = <$fh_itmp>;
		next;
	    }
	    ##($tmpin =~ m/\S/) or next; # Skip blank lines (if any)

	# REVISED in creation of 'extract_phyml_statistics2' (on Apr 8 (Fri), 2011) #
	    ($tmpin =~ m/^\s*o/) and last; # Stop when reaching the footer #
	    ##($tmpin =~ m/^\s*o/) and next; # Skip the header and the footer #

	    unless ($tmpin =~ s/^\s*\.\s*//) {
		$tmpin =~ s/\s+$//;
		if ($debug_level>1) {
		    print STDOUT "[WARNING 'extract_phyml_statistics2'] An UNEXTECTED BLOCK: $! !\n";
		    print STDOUT " '$tmpin'.\n";
		    print STDOUT " --- I will skip it ...\n\n";
		}

	# ADDED in creation of 'extract_phyml_statistics2' on Apr 8 (Fri), 2011 #
		$tmpin = <$fh_itmp>;

		next;
	    }
	    $tmpin =~ s/\s+$//;
	    my ($key, $value) = split /\s*\:\s*/, $tmpin, 2;

	    if ($key =~ m/^Discrete\s+gamma\s+model/) {
		if($value =~ m/^Yes/) {

	# REVISED in creation of 'extract_phyml_statistics2' (on Apr 8 (Fri), 2011) #
		    my @lines = ($tmpin);
		    while ($tmpin = <$fh_itmp>) { # 1st inner while-loop #
			($tmpin =~ m/^\s*\./) and last;
			($tmpin =~ m/\S/) or next; # Skip blank lines #
			($tmpin =~ s/\s+$//);
			push @lines, $tmpin;
		    }			# End of the 1st inner while-loop #
		    ##my @lines = split /\n+/, $value;

		    unless (@lines == 3) {
		        print STDOUT "[WARNING 'extract_phyml_statistics2'] Found an ANOMALOUS BLOCK describing the discrete gamma model: $! !\n";

	# REVISED in creation of 'extract_phyml_statistics2' (on Apr 8 (Fri), 2011) #
			my $block = join ("\n", @lines);
		        print STDOUT "'$block' .\n";
		        ##print STDOUT "'$tmpin' .\n";

		        print STDOUT "  --- I will skip it ...\n\n";
		        next;
		    }
		    my ($ncate) = ($lines[1] =~ m/(\S+)\s*$/);
		    my ($alpha) = ($lines[2] =~ m/(\S+)\s*$/);

		    $stat_phyml->{'discrete_gamma'} = [$ncate, $alpha];

		} else { # ADDED in creation of 'extract_phyml_statistics2' (on Apr 8 (Fri), 2011) #
		   $tmpin = <$fh_itmp>;
		}

	    } elsif ($key =~ m/^Nucleotides\s+frequencies/) {

	# REVISED in creation of 'extract_phyml_statistics2' (on Apr 8 (Fri), 2011) #
		my @lines = ();
		while ($tmpin = <$fh_itmp>) { # 1st inner while-loop #
		    ($tmpin =~ m/^\s*\./) and last;
		    ($tmpin =~ m/\S/) or next; # Skip blank lines #
		    ($tmpin =~ s/\s+$//);
		    push @lines, $tmpin;
		}		# End of the 1st inner while-loop #
		##(my $chunk = <$fh_itmp>) =~ s/\s+$//;
		##my @lines = split /\n+/, $chunk;

		unless (@lines == 4) {
		    print STDOUT "[WARNING 'extract_phyml_statistics2'] Found an ANOMALOUS BLOCK describing the nucleotide frequencies: $! !\n";

	# ADDED in creation of 'extract_phyml_statistics2' (on Apr 8 (Fri), 2011) #
		    my $chunk = join ("\n", @lines); # 

		    print STDOUT "'$chunk' .\n";
		    print STDOUT "  --- I will skip it ...\n\n";
		    next;
		}
		my @nfreqs = ();
		for (my $b=0; $b< 4; $b++) {
		    my ($freq) = ($lines[$b] =~ m/([^\s\=]+)\s*$/);
		    unless (defined $freq) {
			print STDOUT "[ERROR 'extract_phyml_statistics2'] FAILED to extract the frequency from the ", $b + 1, " th line for the nucleotide fequencies: $! !\n";
			print STDOUT " '", $lines[$b], "' .\n\n";
			return 0;
		    }
		    push @nfreqs, $freq;
		}
		$stat_phyml->{'nuc_freqs'} = \@nfreqs;

	    } elsif ($key =~ m/^GTR\s+relative\s+rate\s+parameters/) {

	# REVISED in creation of 'extract_phyml_statistics2' (on Apr 8 (Fri), 2011) #
		my @lines = ();
		while ($tmpin = <$fh_itmp>) { # 1st inner while-loop #
		    ($tmpin =~ m/^\s*\./) and last;
		    ($tmpin =~ m/\S/) or next; # Skip blank lines #
		    ($tmpin =~ s/\s+$//);
		    push @lines, $tmpin;
		}		# End of the 1st inner while-loop #
		##(my $chunk = <$fh_itmp>) =~ s/\s+$//;
		##my @lines = split /\n+/, $chunk;

		unless (@lines == 6) {
		    print STDOUT "[WARNING 'extract_phyml_statistics2'] Found an ANOMALOUS BLOCK describing the relative rate parameters: $! !\n";

	# ADDED in creation of 'extract_phyml_statistics2' (on Apr 8 (Fri), 2011) #
		    my $chunk = join ("\n", @lines); # 

		    print STDOUT "'$chunk' .\n";
		    print STDOUT "  --- I will skip it ...\n\n";
		    next;
		}
		my @rel_rates = ();
		for (my $b=0; $b< 6; $b++) {
		    my ($rrate) = ($lines[$b] =~ m/(\S+)\s*$/);
		    unless (defined $rrate) {
			print STDOUT "[ERROR 'extract_phyml_statistics2'] FAILED to extract the relative rate from the ", $b + 1, " th line for the relative rate parameters: $! !\n";
			print STDOUT " '", $lines[$b], "' .\n\n";
			return 0;
		    }
		    push @rel_rates, $rrate;
		}
		$stat_phyml->{'rel_rates'} = \@rel_rates;

	    } elsif ($key =~ m/^Instantaneous\s+rate\s+matrix/) {

	# REVISED in creation of 'extract_phyml_statistics2' (on Apr 8 (Fri), 2011) #
		my @lines = ();
		while ($tmpin = <$fh_itmp>) { # 1st inner while-loop #
		    ($tmpin =~ m/^\s*\./) and last;
		    ($tmpin =~ m/\S/) or next; # Skip blank lines #
		    ($tmpin =~ s/\s+$//);
		    push @lines, $tmpin;
		}		# End of the 1st inner while-loop #
		##(my $chunk = <$fh_itmp>) =~ s/\s+$//;
		##my @lines = split /\n+/, $chunk;

		unless (@lines == 5) {
		    print STDOUT "[WARNING 'extract_phyml_statistics2'] Found an ANOMALOUS BLOCK describing the instantaneous rate matrix: $! !\n";

	# ADDED in creation of 'extract_phyml_statistics2' (on Apr 8 (Fri), 2011) #
		    my $chunk = join ("\n", @lines); # 

		    print STDOUT "'$chunk' .\n";
		    print STDOUT "  --- I will skip it ...\n\n";
		    next;
		}
		my @rate_mtrx = ();
		for (my $b=1; $b< 5; $b++) {
		    my @rates = split ' ', $lines[$b];
		    unless (@rates == 4) {
			print STDOUT "[ERROR 'extract_phyml_statistics2'] ANOMALOUS INPUT: the ", $b + 1, " th line for the rate matrix has ", scalar (@rates), " elements: $! !\n";
			print STDOUT " '", $lines[$b], "' .\n\n";
			return 0;
		    }
		    push @rate_mtrx, \@rates;
		}
		$stat_phyml->{'rate_mtrx'} = \@rate_mtrx;

	    } elsif ((defined $value) and ($value =~ m/\S/)) {

		$key =~ s/\s/_/g;
		$stat_phyml->{$key} = $value;

	# ADDED in creation of 'extract_phyml_statistics2' (on Apr 8 (Fri), 2011) #
		$tmpin = <$fh_itmp>;

	# ADDED in creation of 'extract_phyml_statistics2' (on Apr 8 (Fri), 2011) #
	    } else {
		if ($debug_level>1) {
		    print STDOUT "[WARNING 'extract_phyml_statistics2'] ANOMALOUS INPUT: UNDEFINED VALUE for the key '$key' in the descripition line: $! !\n";
		    print STDOUT " '$tmpin' .\n\n";
		}
		$tmpin = <$fh_itmp>;

	    }

	}			# End of the main while-loop #
	close ($fh_itmp);

	return 1;

}	# End of the subroutine 'extract_phyml_statistics2' #


# fetch_clmwise_likelihoods_phyml:
#	This subroutine reads in a supplementary output file 'xxx__phyml_lk.txt',
#	which records the column-wise likelihoods and column-wise, rate category-wise, likilihoods. 
#
# INPUT: (\$infile, $fl_log), where
#	$infile is the path to the supplementary output file;
#	$fl_log = 1 if you want LOG-likelihoods,
#		= 0 if you want RAW-likelihoods.
#
# OUTPUT: (\@clmwise_lks, \@clm_rcate_wise_lks, \@mean_alpha, \@tags), where
#	$clmwise_lks[$l] is the (log-)likelihood for the $l th column;
#	$clm_rcate_wise_lks[$l]->[$c] is the (log-)likelihood for the $c th rate category in the $l th column;
#	$mean_rcate[$l] is the posterior mean of the gamma shape parameter for the $l th column.
#	@tags = ('Site', 'P(D|M)') for uniform rate,
#	      = ('Site', 'P(D|M)', "P(D|M,rr[1]=$rr1)", ... "P(D|M,rr[C]=$rrC)", 'Posterior mean')
#		when there are C rate categories, with $rr(k) (k=1,..,C) is the representative rate
#		of the k-th category.
#
sub fetch_clmwise_likelihoods_phyml (\$$) {

	my ($infile, $fl_log) = @_;
	local $/ = "\n"; # use a newline character as an input-record-separator #

	my $fh_itmp = open_input_file2 (${$infile});
	unless (defined $fh_itmp) {
	    print STDOUT "[ERROR 'fetch_clmwise_likelihoods_phyml'] 'open_input_file2' FAILED on the input file '${$infile}': $! !\n";
	    return (undef, undef, undef, undef);
	}

	my $header;
	while ($header = <$fh_itmp>) { ($header =~ m/^Site/) and last; }

	my @tags = split /\s+/, $header;
	if (@tags>2) {
	   my $last = pop @tags;
	   if ($last eq 'mean') {
	      $tags[$#tags] .= " $last";
	   } else {
		$header =~ s/\s+$//;
		print STDOUT "[ERROR 'fetch_clmwise_likelihoods_phyml'] Found an ANOMALOUS FIELD TAG LINE: $! !\n";
		print STDOUT " '$header' .\n\n";
		return (undef, undef, undef, undef);
	   }
	}
	my $ct_flds = @tags;

	my $site_no = 0;
	my @clmwise_lks = my @clm_rcate_wise_lks = my @mean_alpha = ();

	while (my $in = <$fh_itmp>) { # Main while-loop #

	    ($in =~ m/^\d/) or next;
	    $site_no++;
	    my @fields = split /\s+/, $in;
	    unless (@fields == $ct_flds) {
		$in =~ s/\s+$//;
		print STDOUT "[ERROR 'fetch_clmwise_likelihoods_phyml'] The number of fields (", scalar (@fields), ") in the $site_no th site CLASHES with that ($ct_flds) expected from the tags: $! !\n";
		print STDOUT " '$in' .\n\n";
		return (undef, undef, undef, \@tags);
	    }
	    my $site_no1 = shift @fields;
	    unless ($site_no1 == $site_no) {
		$in =~ s/\s+$//;
		print STDOUT "[ERROR 'fetch_clmwise_likelihoods_phyml'] The site number '$site_no1' CLASHES with the simple order ($site_no) of the site: $! !\n";
		print STDOUT " '$in' .\n\n";
		return (undef, undef, undef, \@tags);
	    }
	    my $lk = shift @fields;
	    if ($fl_log) {
		push @clmwise_lks, (($lk>0) ? log ($lk) : undef);
	    } else {
		push @clmwise_lks, $lk;
	    }

	    ($ct_flds>2) or next;

	    my $malpha = pop @fields;
	    push @mean_alpha, $malpha;

	    if ($fl_log) {
		my @lnlks = ();
		foreach my $lk1 (@fields) { push @lnlks, (($lk1>0) ? log ($lk1) : undef); }
		push @clm_rcate_wise_lks, \@lnlks;
	    } else {
		push @clm_rcate_wise_lks, \@fields;
	    }

	}		# End of the main while-loop #

	close ($fh_itmp);

	return (\@clmwise_lks, \@clm_rcate_wise_lks, \@mean_alpha, \@tags);

}	# End of the subroutine 'fetch_clmwise_likelihoods_phyml' #


1;	# a "spell" which makes the module "use"-able and/or "require"-able #


