package MyTreeManipu_II_dawg;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.5: Copyright (C) 2014 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyTreeManipu_II_dawg.pm:
#	This package inputs/outputs trees in a modified Newick format used by Dawg, the molecular evolution simulator.
#	(CREATION STARTED on Feb 18 (Tue), 2014, ENDED on Feb xx (yyy), 2014)
#
# MyTreeManipu_II.pm: This package can process more general types of trees,
#	(i.e. NOT limited to binary trees with trichotomous top nodes).
#	It is also consistent with the format of the taxonomy tree which are handled by
#	the module 'MyTaxonomy.pm'.
#
# MyTreeManipu.pm: This package performs basic tree manipulations.
#
#  Subroutines whose substitutes are available in this module ('MyTreeManipu_II.pm')
#	or in 'MyTaxonomy.pm':
#   our @EXPORT = qw(retrieve_top_node confine_outgroup print_unrooted);
#   our @EXPORT_OK = qw(switch_topnode  print_subtree topo_unrooted topo_subtree
#		list_sequences fetch_node_series dist_nodes
#		    );
#
# Subroutines whose substitutes are not available yet:
#   our @EXPORT_OK = ( retrieve_bootstrap  prod_bootstrap );
#

use strict;
use warnings;
use MyArrayManipu qw( diff_array union_array );
use MyFile_manipu qw( open_input_file );
use MyTaxonomy qw( fetch_ancestry fetch_last_common_ancestor_ov2 fetch_common_ancestors
		 fetch_ext_offsprings fetch_all_offsprings );

use Exporter;

our @ISA = qw(Exporter);

##sub rec_fillin_boot (\@\$);
#sub rec_prod_boot (\@\@);


our @EXPORT_OK = qw(
	extract_sgltree_from_file_dawg
	input_tree_file_dawg read_tree_file_dawg input_subtree_dawg
	output_tree_dawg draw_subtree_dawg
	br_output_tree_dawg br_draw_subtree_dawg
	br_output_tree_sep_dawg br_draw_subtree_sep_dawg
		) ;
	## input_tree_with_nodeids 
	## topo_subtree

sub input_tree_file_dawg (\$\%\%\%\%\%\%\$);
sub read_tree_file_dawg (\$\$\%\%\%\%\%\%);
sub draw_subtree_dawg ($\%\%\%$$);
sub br_draw_subtree_dawg ($\%\%\%$$);
sub br_draw_subtree_sep_dawg ($\%\%\%$$$);

our $TOP_NODE = 'TOP_NODE';
our $INDX_LENGTH = 0;
our $INDX_BOOT = 1; # For the predecessor 'MyTreeManipu_II.pm'. # NOTE: IT should be used ONLY for exterior nodes. #
our $INDX_INTNODENAME = 1; # ADDED in creation of 'MyTreeManipu_II_dawg.pm'.

our $START_GROUP = '(';	# the symbol marking the start of the grouping of nodes.
our $END_GROUP = ')';  # the symbol marking the end of the grouping of nodes.


### 5 important hash tables which describe a tree ###
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# [ %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.  ] --- for the predecessor 'MyTreeManipu_II.pm'.
#
#	# The following are for the package 'MyTreeManipu_II_dawg.pm'. #
#
# %branch_attributes = (ID of the nearest downstream node => [length, name_dw_node, ...], ...) .
#	--- for the branches.
#
# %name2id_intnode = ($node_name => $node_ID, ...) for the interior node (including the top node, which will be recoreded if $node_name ne $TOP_NODE). #
#


# extract_sgltree_from_file_dawg:
#	This slightly modified version of 'extract_sgltree_from_file' can correctly extract
#	all information described in the tree in the format specified by Dawg.
#	(This s a 'wrapper' of the new subroutine 'input_tree_file_dawg'.)
#	(CREATED on Feb 19 (Wed), 2014)
#
# extract_sgltree_from_file:
#	This subroutine is a sort of 'wrapper' of the subroutine 'input_tree_file',
#	and reads in a SINGLE TREE n the Newick format and stores the information.
#	If the comment-initiating 'word' is specified,
#	it removes lines starting with the word.
#
# [ INPUT: (\$infile, \%node2parent, \%node2children, \%name2id, \%id2name, \%branch_attributes, $comment), where ] --- for the predecessor 'extract_sgltree_from_file'.
# INPUT: (\$infile, \%node2parent, \%node2children, \%name2id, \%id2name, \%branch_attributes, \%name2id_intnode, \$topnode_name, $comment), where
#	the 1st argument is the path to the input tree file;
#	The 2nd to 7th arguments are hashes whose formats are described above.
#
# OUTPUT: 1 if the process succeeded,
#	  0 if it failed.
#
sub extract_sgltree_from_file_dawg (\$\%\%\%\%\%\%\$$) {

	##my ($infile, $node2pa, $node2chi, $name2id, $id2name, $br_attr, $comment) = @_;
	my ($infile, $node2pa, $node2chi, $name2id, $id2name, $br_attr, $name2id_intnode, $topnode_name, $comment) = @_;
	local $/ = "\n";

	my $fh = open_input_file (${$infile});
	unless (defined $fh) {
	    print STDOUT "[ERROR 'extract_sgltree_from_file_dawg'] FAILED TO OPEN the input tree file '${$infile}': $! !\n\n";
	    return 0;
	}
	my $len_comment = length $comment;
	my $intree = '';
	while (my $in = <$fh>) { # The main while-loop #

	    ($in =~ m/\S/) or next; # skip a blank line #
	    $in =~ s/^\s+//; # Remove preceding white spaces #

		# Skip a comment line #
	   ($len_comment <= length ($in)) and
		($comment eq substr ($in, 0, $len_comment)) and next;
	   $intree .= $in;
	}			# End of the main while-loop #
	close ($fh);

=for DEBUGGING the subroutine
	print STDERR "\n< Subroutine 'extract_sgltree_from_file' >\n\n";
	print STDERR "Input tree to be parsed by 'input_tree_file':\n\n";
	print STDERR $intree;
	print STDERR "\n";
=cut

	unless ( input_tree_file_dawg ($intree, %{$node2pa}, %{$node2chi}, %{$name2id}, %{$id2name}, %{$br_attr}, %{$name2id_intnode}, ${$topnode_name} ) ) {
	    print STDOUT "[ERROR 'extract_sgltree_from_file_dawg'] FAILED to perform the subroutine 'input_tree_file_dawg': $! !\n\n";
	    return 0;
	}

	return 1;

}		# End of the subroutine 'extract_sgltree_from_file_dawg' #


# input_tree_file_dawg:
#	This slightly modified version of 'input_tree_file' can handle
#	input tree in the format specified by Dawg, especially, can extract names of interior nodes.
#	(This is actually an I/O interface of the subroutine 'read_tree_file_dawg'.)
#	(CREATED on Feb 19 (Wed), 2014)
#
# input_tree_file:
#	This subroutine reads in a tree file and stores the information on a whole tree.
#	This is actually an I/O interface of the subroutine 'read_tree_file'.
#
# INPUT is supposed to start with the left parenthesis ('(') , and end with the semicolon ';',
# possibly with some comments written in alphabets,
# and children are supposed to be separated by the comma (',').
#
sub input_tree_file_dawg (\$\%\%\%\%\%\%\$) {

	##my ($input, $node2parent, $node2children, $name2id, $id2name, $branch_attributes) = @_; 
	my ($input, $node2parent, $node2children, $name2id, $id2name, $branch_attributes, $name2id_intnode, $topnode_name) = @_; 

		# MODIFIED in creation of 'input_tree_file_dawg'. #
	${$input} =~ s/\s*\;\s*$//; # remove a superfluous string trailing the main information

	my ($tmp_topnode_name) = (${$input} =~ m/\)\s*([^\)]+)$/); # Extract the name of the top node, if recorded at all.
	my $len_topnode_name = (defined $tmp_topnode_name) ? length ($tmp_topnode_name) : 0;
	($len_topnode_name>0) and ( substr (${$input}, - $len_topnode_name, $len_topnode_name, '') ); # Remove the top node name from the tree.
	${$input} =~ s/\s+$//; # Remove the white spaces remaining on the right-end of the tree.

	###${$input} =~ s/\s*\w*\s*\;\s*$//; # remove a superfluous string trailing the main information
	##${$input} =~ s/[^\)]*\;\s*$//; # remove a superfluous string trailing the main information

	my $ct_lpar = (${$input} =~ tr/(// );
	my $ct_rpar = (${$input} =~ tr/)// );

	unless ($ct_lpar == $ct_rpar) {
	    print STDOUT "[ERROR 'input_tree_file_dawg'] ILLEGAL INPUT: The number of left parentheses ($ct_lpar) CLASHES with that of right parentheses ($ct_rpar): $! !\n";
	    return 0;
	}

	unless (${$input} =~ s/^\s*\(\s*// ) { #  remove the left parenthesis at the beginning of the tree
		print STDOUT "[ERROR 'input_tree_file_dawg'] ILLEGAL input : Tree NOT BEGIN with '(':  $! !\n";
		print STDOUT "'${$input}' .\n";
		return 0;
	}

	my $node_id = 0;

	unless ( read_tree_file_dawg (${$input}, $node_id, %{$node2parent}, %{$node2children},
			%{$name2id}, %{$id2name}, %{$branch_attributes}, %{$name2id_intnode}) ) {

		print STDOUT "[ERROR 'input_tree_file_dawg'] 'read_tree_file_dawg' failed at the node ID $node_id: $! !\n";
		return 0;
	}

	$name2id->{$TOP_NODE} = $node_id;

	if (defined $branch_attributes->{$node_id}) {
		print STDOUT "[CAUTION 'input_tree_file_dawg'] The top_node (ID:$node_id) has somehow information on its 'upper branch', which CANNOT exist:\n";
		print STDOUT ' ' x 5 . '{' . join (', ' , @{$branch_attributes->{$node_id}} ) . "\} .\n";
		undef $branch_attributes->{$node_id};
	}

		## ADDED in creation of 'input_tree_file_dawg'. ##

	if ($len_topnode_name > 0) {
	    $name2id_intnode->{$tmp_topnode_name} = $node_id;
	    ${$topnode_name} = $tmp_topnode_name;
	}
	##($len_topnode_name > 0) and ($name2id_intnode->{$topnode_name} = $node_id);

	foreach my $nodeid (keys %{$branch_attributes}) { # Trailing foreach-loop (over nodes except the top node). #

	    my $ch = $node2children->{$nodeid};
	    ((defined $ch) and (@{$ch}>0)) or next; # Skip exterior nodes. #

	    my $attr = $branch_attributes->{$nodeid};
	    my $name = $attr->[$INDX_INTNODENAME];

	    (defined $name) and ($name2id_intnode->{$name} = $nodeid);

	}			# End of the trailing foreach-loop (over nodes except the top node). #

	return 1;

}		# End of the subroutine 'input_tree_file_dawg' #


# input_subtree_dawg:
#	This slightly revised version of the subroutine 'input_substree' was made
#	to retrieve the complete information from an output of the subroutine 'draw_subtree_dawg'.
#	(CREATED on Feb 19 (Wed), 2014)
#
# input_subtree: This subroutine was made to retrieve the complete information from an
#	output of the subroutine 'draw_subtree'.
#
sub input_subtree_dawg (\$) {

	my ($in_subtree) = @_;
	${$in_subtree} =~ s/\s*\;\s*$//;

	my ($dist, $boot) ;

	my ($appended) = (${$in_subtree} =~ m/([^\)]+)$/ );
	if (defined $appended) {
	    my $app_len = length ($appended);
	    substr (${$in_subtree}, - $app_len, $app_len, ';'); # remove the appended information from the input
	    $appended =~ s/^\s+//;
	}

	my %node2parent = my %node2children = my %name2nodeid = my %nodeid2name = my %branch_attributes = ();
	my %name2id_intnode; # ADDED in creation of 'input_subtree_dawg'. #
	my $topnode_name = ''; # ADDED in creation of 'input_subtree_dawg'. #

	unless ( input_tree_file_dawg (${$in_subtree},
		%node2parent, %node2children, %name2nodeid, %nodeid2name, %branch_attributes, %name2id_intnode, $topnode_name) ) {

		print STDOUT "[ERROR 'input_subtree_dawg'] 'input_tree_file' failed.\n";
		return (undef, undef, undef, undef, undef, undef, undef);
	}

	if ( (defined $appended) and (length ($appended) > 0) ) { # Modified in creation of 'input_subtree_dawg'. #
	##if ( defined $appended) {...} # For the predecessor 'input_subtree'.

			# when -bootlabel=node
		my ($bootstrap, $br_len) = split /\:/, $appended;

=for the predecessor 'input_subtree'
			# when -bootlabel=branch
		($bootstrap eq '') and ( ($br_len, $bootstrap) = ($appended =~ m/^\:([^\[]+)\[([^\]]+)\]/ ) );
=cut

			# when the branch has no bootstrap value displayed

		# MODIFIED to DEBUG the subroutine.
		if (!(defined $br_len) and ($appended =~ s/^\://)) {
		    $br_len = $appended;
		}
		##(defined $br_len) or ( ($br_len  = $appended) =~ s/^\:// ) ;

		( (defined $bootstrap) and ($bootstrap eq '') ) and (undef $bootstrap);

		((defined $br_len) and ($br_len =~ m/\S/)) or ($br_len = undef); # ADDED to DEBUG the subroutine (on Feb 21 (Fri), 2014).

		my $nodename = $bootstrap; # ADDED in creation of 'input_subtree_dawg'. #

		my $topnode = $name2nodeid{$TOP_NODE};

			# MODIFIED in creation of 'input_subtree_dawg'. #
		if (defined $br_len) { # The $topnode shoud NOT be the genuine top node.
		    my @attr = ();
		    $attr[$INDX_LENGTH] = $br_len;
		    if (defined $nodename) {
		       $attr[$INDX_INTNODENAME] = $nodename;
		       $name2id_intnode{$nodename} = $topnode;
		    }
		    $branch_attributes{$topnode} = \@attr;

		} elsif (defined $nodename) { # The node SHOULD be the genuine top node, with a particular name.
		    $topnode_name = $nodename;
		    $name2id_intnode{$nodename} = $topnode;
		}
		##$branch_attributes{$topnode} = [$br_len, $bootstrap];

	}

		# MODIFIED in creation of 'input_subtree_dawg'. #

	return (\%node2parent, \%node2children, \%name2nodeid, \%nodeid2name, \%branch_attributes, \%name2id_intnode, $topnode_name);
	##return (\%node2parent, \%node2children, \%name2nodeid, \%nodeid2name, \%branch_attributes);

}		# End of the subroutine 'input_subtree_dawg' #


# read_tree_file_dawg:
#	This revised version of the subroutine 'read_tree_file' can read
#	a (sub-)tree in the format specified by Dawg, and retrieve the names of interior nodes, if any.
#	(CREATED on Feb 19 (Wed), 2014)
#
# read_tree_file:
#	This subroutine reads in (a part of) a tree-file text
#	and stores the information on the 5 hash tables describing trees.
#
#	REVISED on July 13 (Tue), 2010:
#	The subroutine now can correctly read in significance or other information
#	associated with the exterior node, which was unreadable before the revision.
#
# INPUT is supposed to start with a left parenthesis ('(') ,
#	which must have been removed before being passed to this subroutine,
#	and end with a right parenthesis followed by a semicolon ';',
# 	possibly with some comments (in alphabets) in between,
#	and children are supposed to be separated by commas (',').
#
sub read_tree_file_dawg (\$\$\%\%\%\%\%\%) {

	##my ($input, $node_id, $node2parent, $node2children, $name2id, $id2name, $branch_attributes) = @_;
	my ($input, $node_id, $node2parent, $node2children, $name2id, $id2name, $branch_attributes, $name2id_intnode) = @_;

	my @children = ();

	#while (${$input} !~ m/^\)/ ) ) { # the outermost while #
	while ( !(${$input} =~ s/^\)\s*// ) ) { # the outermost while
				# --- continues untill it reaches the end of the group of nodes

	    if (${$input} =~ s/^\s*\(\s*// ) { # reading in a group of nodes #

		unless (read_tree_file_dawg (${$input}, ${$node_id}, %{$node2parent}, %{$node2children},
				%{$name2id}, %{$id2name}, %{$branch_attributes}, %{$name2id_intnode} ) ) {
			print STDOUT "[ERROR 'read_tree_file_dawg'] 'read_tree_file_dawg' ENDED at node ID  = ${$node_id}' : $! !\n";
			return 0; # error occurred #
		}

		push @children, ${$node_id};

	    } else { # reading in a single node (probably external) #

		my ($tmp) = (${$input} =~ m/^([^\,\)]+)/ );
		unless (defined $tmp)  {
			print STDOUT "[ERROR 'read_tree_file_dawg'] Unexpected input : '" . substr (${$input}, 0, 10) . "'...: $! \n";
			return 0;
		}

		substr (${$input}, 0, length ($tmp) , ''); # remove $tmp from the input

		$tmp =~ s/\s+$//;

		#my ($seqname, $br_len) = ($tmp =~ m/^([^\:]+)\:([^\:]+)/ );

			## REVISED on July 13 (Tue), 2010 ##
		my ($seqname, $br_attr) = split /\:/, $tmp, 2;
		##my ($seqname, $br_len) = split /\:/, $tmp;

		(defined $seqname) or ($seqname = $tmp);

		${$node_id}++;
		push @children, ${$node_id};
		$name2id->{$seqname} = ${$node_id};
		$id2name->{${$node_id}} = $seqname;

			## ADDED on July 13 (Tue), 2010 ##
		if ((defined $br_attr) and ($br_attr ne '')) {

			# when -bootlabel=node
		    my ($bootstrap, $br_len) = split /\:/, $br_attr;

			# when -bootlabel=branch
		    ((defined $br_len) and ($br_len =~ m/\d/)) or ( ($br_len, $bootstrap) = ($br_attr =~ m/^([^\[]*)\[([^\]]+)\]/ ) );

		    if (defined $br_len) {
			($br_len eq '')  and (undef $br_len);
		    } else {
				# Only a single attribute is recorded #
			$br_len = $br_attr;
		    }
		    ((defined $bootstrap) and ($bootstrap eq '')) and (undef $bootstrap);

		    my @attr = ();
		    (defined $br_len) and ($attr[$INDX_LENGTH] = $br_len);
		    (defined $bootstrap) and ($attr[$INDX_BOOT] = $bootstrap);
		    (@attr>0) and ($branch_attributes->{${$node_id} } = \@attr);
		}
		##(defined $br_len) and ($branch_attributes->{${$node_id} } = [$br_len, undef] );

	    }

	    ${$input} =~ s/^\,\s*//; # remove the comma delimiting this group from the following one

	}		# end of the outermost while #

	##${$input} =~ s/^\)\s*//;

	${$node_id}++; # The ID of the node that has just been read in.

	$node2children->{${$node_id}} = \@children;
	foreach my $child (@children) { $node2parent->{$child} = ${$node_id}; }
		
	my ($tmp) = (${$input} =~ m/^([^\,\)]+)/ ); # information belonging to the node in question, or its immediate upper branch 
	if ( defined $tmp) {

		substr (${$input}, 0, length ($tmp), '' ); # remove $tmp from the input
		$tmp =~ s/\s+$//;
			# when -bootlabel=node
		my ($bootstrap, $br_len) = split /\:/, $tmp;

=for the predecessor 'read_tree_file'
			# when -bootlabel=branch
		($bootstrap eq '') and ( ($br_len, $bootstrap) = ($tmp =~ m/^\:([^\[]+)\[([^\]]+)\]/ ) );
=cut

			# when the branch has no bootstrap value displayed
		(defined $br_len) or ( ($br_len  = $tmp) =~ s/^\:// ) ;
		( (defined $bootstrap) and ($bootstrap eq '') ) and (undef $bootstrap);

		my $nodename = $bootstrap; # ADDED in creation of 'read_tree_file_dawg'. #

			## REVISED on July 13 (Tue), 2010 ##
		my @attr = ();
		(defined $br_len) and ($attr[$INDX_LENGTH] = $br_len);

				# MODIFIED in creation of 'read_tree_file_dawg'. #
		(defined $nodename) and ($attr[$INDX_INTNODENAME] = $nodename);
		##(defined $bootstrap) and ($attr[$INDX_BOOT] = $bootstrap);

		(@attr>0) and ($branch_attributes->{${$node_id}} = \@attr);
		##$branch_attributes->{${$node_id}} = [$br_len, $bootstrap];

	}

	return 1;

}		# End of the subroutine 'read_tree_file_dawg' #


=for the predecessor 'MyTreeManipu_II.pm'.

# input_tree_with_nodeids:
#	This subroutine can read in a tree, possibly accompanied by
#	node IDs used in the previous analyses.
#	It ASSUMES that each node ID is given on the immediate right of
#	a cluster representing the node.
#
# INPUT: (\$in_tree, $fl_in_nodeid, \$reserve_nodeid), where
#	$in_tree is a chunk describing a tree;
#	$fl_in_nodeid = 1 if you use node IDs accompanying the input tree,
#			= 0 if you newly assign node IDs;
#	$reserve_nodeid is the ID reserved for  nodes whose input IDs are not assigned in the input tree.
#	It will be incremented by 1 before being used.
#
# OUTPUT: (\%node2pa, \%node2chi, \%name2id, \%id2name, \%br_attr), where
#	%node2pa = ($node_ID => $parent_ID, ...);
#	%node2chi = ($node_ID => \@child_IDs, ...);
#	%name2id = ($node_name => $node_ID, ...) for the nodes bearing names
#	(usually exterior nodes and the top node);
#	%id2name = ($node_ID => $node_name, ...) for the nodes bearing names
#	EXCLUDING the top node;
#	%br_attr = ($ID of the node immediatly downstream of the branch
#		=> \@attr, ...) with
#	($br_len, $boot) = @attr[$INDX_LEN, $INDX_BOOT].
#
sub input_tree_with_nodeids (\$$\$) {

	my ($in_tree, $fl_in_nodeid, $reserve_nodeid) = @_;

	my $input_topnode_id;

##=for DEBUGGING the subroutine
	print STDERR "\n< Subroutinen 'input_tree_with_nodeids' >\n\n";
	print STDERR "{ Raw input tree }\n\n";
	print STDERR ${$in_tree} , "\n";
##=cut

	if ($fl_in_nodeid) {
	    ($input_topnode_id) = (${$in_tree} =~ m/\{\s*([^\s\{\}]+)\s*\}\;\s*$/);
	    unless (defined $input_topnode_id) {
		print STDERR "[CAUTION 'input_tree_with_nodeids'] FAILED to extract topnode ID from the input tree: $! !\n";
		print STDERR ${$in_tree}, "\n";
		##return (undef, undef, undef, undef, undef);
	    }
	    ${$in_tree} =~ s/[^\)]+\;\s*$/\;/;
	}

	my %node2pa; my %node2chi; my %name2id; my %id2name; my %br_attr;
	unless ( input_tree_file (${$in_tree}, %node2pa, %node2chi, %name2id, %id2name, %br_attr) ) {
	    print STDERR "[ERROR 'input_tree_with_nodeids'] The subroutine 'input_tree_file' FAILED to read in the input tree: $!\n";
	    print STDERR ${$in_tree} , "\n";
	    return (undef, undef, undef, undef, undef);
	}

##=for DEBUGGING the subroutine
	print STDERR "\n{ 1st extracted information on the input tree }\n\n"; 
	print STDERR "\#Node_ID --> seq_name\n\n";
	foreach my $nodeid (sort {$a <=> $b} keys %id2name) {
	    print STDERR $nodeid, ' --> ', $id2name{$nodeid}, "\n";
	}
	print STDERR "\n\#Seq_name --> node_ID\n\n";
	foreach my $seqname (sort keys %name2id) {
	    print STDERR $seqname, ' --> ', $name2id{$seqname}, "\n";
	}
	print STDERR "\n\#Node_ID <-- parent_ID\n\n";
	foreach my $nodeid (sort {$a <=> $b} keys %node2pa) {
	    print STDERR $nodeid, ' <-- ', $node2pa{$nodeid}, "\n";
	}
	print STDERR "\n\#Node_ID --> \[IDs,of,children\]\n\n";
	foreach my $nodeid (sort {$a <=> $b} keys %node2chi) {
	    print STDERR $nodeid, ' --> [', join (',', @{$node2chi{$nodeid}}), ']', "\n";
	}
	print STDERR "\n#Dw_node_ID  --> \[branch,attributes\]\n\n";
	foreach my $nodeid (sort {$a <=> $b} keys %br_attr) {
	    print STDERR $nodeid, ' --> [', join (',', @{$br_attr{$nodeid}}), ']', "\n";
	}
	print STDERR "\n";
##=cut

	$fl_in_nodeid or (return (\%node2pa, \%node2chi, \%name2id, \%id2name, \%br_attr) );

		# Replace automatically assigned node IDs
		# with input node IDs

	my $old_topnode_id = $name2id{$TOP_NODE};

	my %in_id2name; my %in_br_attr;
	my %name2in_id;
	my %old_id2in_id; # ( $old_node_ID =>  $input_node_ID, ...)

	    ## Handle the case where the input topnode ID was not extracted ##
	    ## --> MOVED from below for DEBUGGING (on Sep 1 (Mon), 2008)
	unless (defined $input_topnode_id) {
	    print STDERR "[CAUTION 'input_tree_with_nodeids'] The TOP NODE (automatically assigned ID: $old_topnode_id) does NOT HAVE the corresponding INPUT ID: $! !\n";
	    ${$reserve_nodeid}++;
	    $input_topnode_id = ${$reserve_nodeid};
	    print STDERR "  --- So I will assign the new ID '$input_topnode_id' ...\n\n";
	}

		## Made from conditional to mandatory for DEBUGGING
		## (on Sep 1 (Mon), 2008)
	###if (defined $input_topnode_id) {
	    $name2in_id{$TOP_NODE} = $input_topnode_id;
	    (defined $old_topnode_id) and ($old_id2in_id{$old_topnode_id} = $input_topnode_id);
	###}

	   # Extract input node IDs from node names #
	foreach my $old_id (keys %id2name) { # The 1st foreach-loop #
	    my $old_name = $id2name{$old_id};
	    my ($in_name, $in_id) = split /\s*\{\s*/, $old_name;
	    unless (defined $in_id) {
		print STDERR "[CAUTION 'input_tree_with_nodeids'] FAILED to extract the INPUT ID out of the node name '$old_name' (automatically assigned ID: $old_id): $! !\n";
		next;
	    }
	    $in_id =~ s/\s*\}\s*$//;
	    $name2in_id{$in_name} = $in_id;
	    $in_id2name{$in_id} = $in_name;
	    my $attr = $br_attr{$old_id};
	    (defined $attr) and ($in_br_attr{$in_id} = $attr);
	    $old_id2in_id{$old_id} = $in_id;
	}			# End of the 1st foreach-loop #

	    # Extract input node IDs from branch attributes #
	foreach my $old_id (keys %br_attr) { # The 2nd foreach-loop #
	    my $old_attr = $br_attr{$old_id};
	    my ($old_brlen, $old_boot) = @{$old_attr}[$INDX_LENGTH, $INDX_BOOT];
	    my @new_attr = ();

		# The branch has aready been handled #
	    (defined $old_id2in_id{$old_id}) and next;

	    if ((defined $old_boot) and ($old_boot =~ s/^\s*\{\s*//)) {
		my ($in_id, $in_boot) = split /\s*\}\s*/, $old_boot;
		unless (defined $in_id) {
		    print STDERR "[ERROR 'input_tree_with_nodeids'] FAILED to extract the INPUT ID out of the bootstrap value '\{$old_boot' of the node with automatically assigned ID '$old_id': $! !\n";
		    return (undef, undef, undef, undef, undef);
		}
		$old_id2in_id{$old_id} = $in_id;
		$new_attr[$INDX_LENGTH] = $old_brlen;
		((defined $in_boot) and ($in_boot =~ m/\S/)) and ($new_attr[$INDX_BOOT] = $in_boot);
		((defined $new_attr[$INDX_LENGTH]) or (defined $new_attr[$INDX_BOOT])) and ($in_br_attr{$in_id} = \@new_attr);
		next;
	    }
	    if ((defined $old_brlen) and ($old_brlen =~ s/^\s*\{\s*//)) {
		my ($in_id, $in_brlen) = split /\s*\}\s*/, $old_brlen;
		unless (defined $in_id) {
		    print STDERR "[ERROR 'input_tree_with_nodeids'] FAILED to extract the INPUT ID out of the branch length '\{$old_brlen' of the node with automatically assigned ID '$old_id': $! !\n";
		    return (undef, undef, undef, undef, undef);
		}
		$old_id2in_id{$old_id} = $in_id;
		((defined $in_brlen) and ($in_brlen =~ m/\S/)) and ($new_attr[$INDX_LENGTH] = $in_brlen);
		((defined $old_boot) and ($old_boot =~ m/\S/)) and ($new_attr[$INDX_BOOT] = $old_boot);
		((defined $new_attr[$INDX_LENGTH]) or (defined $new_attr[$INDX_BOOT])) and ($in_br_attr{$in_id} = \@new_attr);
		next;
	    }

		# Somehow, the branch has  been left over #
	    print STDERR "[ERROR 'input_tree_with_nodeids'] The INPUT ID has NOT BEEN FOUND out of EITHER bootstrap value '$old_boot' OR branch length '$old_brlen' of the node with the automatically asigned ID '$old_id': $! !\n";
	    return (undef, undef, undef, undef, undef);

	}		# End of the 2nd foreach-loop #

	    ## Handle the case where the input topnode ID was not extracted ##
	    ## --> Moved upward, and current block has been changed to:
	    ## Handle the case in which a dummy top node is suspected
	    ## (on Sep 1 (Mon), 2008)

		## Double conditions were withdrawn on Sep 1 (Mon), 2008
	##unless (defined $input_topnode_id) {
	 ##  if (defined $old_topnode_id) {
			## A BUG FIXING MEASURE on July 17 (Thu), 2008 ##
		if ((2> scalar keys %id2name) and (0 == scalar keys %br_attr)) {
		    # In this case, the 'tree' must have originally consisted of just one node #
		     %node2pa=();
		     %node2chi = ();

			# Further fixed on Sep 1 (Mon), 2008 #
		     my ($sgl_node_id) = keys %in_id2name;
			# BUG fixed on July 17 (Thu), 2008 #
		     ##($input_topnode_id) = keys %in_id2name;
		     ##($input_topnode_id) = keys %id2name;

			## ADDED on Sep 1 (Mon) 2008
		     (defined $old_topnode_id) or ($old_topnode_id = 'NA');

			# Revised on Sep 1 (Mon), 2008 #
		     print STDERR "[CAUTION 'input_tree_with_nodeids'] The tree seems to consist of a single node '$sgl_node_id', and the \"Top node\" '$input_topnode_id' appear just dummy. So I will just dismiss the latter.\n\n";
		     ##print STDERR "[CAUTION 'input_tree_with_nodeids'] The tree seems to consist of a single node '$input_topnode_id', and the \"Top node\" '$old_topnode_id' appear just dummy. So I will just dismiss the latter.\n\n";

		    $input_topnode_id = $sgl_node_id;
			## Moved from below on Sep 1 (Mon), 2008
		    $name2in_id{$TOP_NODE} = $input_topnode_id;

##=MOVED UPWARD for DEBUGGING  (on Sep 1 (Mon), 2008)
		} else {
			# In this case, we consider that the top node ID was just forgotten to be displayed.
		    $input_topnode_id = $old_id2in_id{$old_topnode_id};
		    unless (defined $input_topnode_id) {
			print STDERR "[CAUTION 'input_tree_with_nodeids'] The TOP NODE (automatically assigned ID: $old_topnode_id) does NOT HAVE the corresponding INPUT ID: $! !\n";
			${$reserve_nodeid}++;
			$input_topnode_id = ${$reserve_nodeid};
			print STDERR "  --- So I will assign the new ID '$input_topnode_id' ...\n\n";
			$old_id2in_id{$old_topnode_id} = $input_topnode_id;
		    }
##=cut
		}
			## Moved upward on Sep 1 (Mon), 2008
		###$name2in_id{$TOP_NODE} = $input_topnode_id;
	##   }
	##}

	my %in_nodeid2pa;
	foreach my $old_id (keys %node2pa) { # The 3rd foreach-loop #
	    my $old_pa = $node2pa{$old_id};
	    my $in_id = $old_id2in_id{$old_id};
	    unless (defined $in_id) {
		print STDERR "[ERROR 'input_tree_with_nodeids'] CanNOT FIND the INPUT ID for the node with automatically assigned ID '$old_id' (and with parent '$old_pa'): $! !\n";
		return (undef, undef, undef, undef, undef);
	    }
	    my $in_pa = $old_id2in_id{$old_pa};
	    unless (defined $in_pa) {
		print STDERR "[ERROR 'input_tree_with_nodeids'] CanNOT FIND the INPUT ID for the parent (automatically assigned ID '$old_pa') of the node '$in_id'  (automatically assigned ID '$old_id'): $! !\n";
		return (undef, undef, undef, undef, undef);
	    }
	    $in_nodeid2pa{$in_id} = $in_pa;
	}			# End of the 3rd foreach-loop #

	my %in_nodeid2chi;
	foreach my $old_id (keys %node2chi) { # The 4th foreach-loop #
	    my $old_children = $node2chi{$old_id};
	    my $in_id = $old_id2in_id{$old_id};
	    unless (defined $in_id) {
		print STDERR "[ERROR 'input_tree_with_nodeids'] CanNOT FIND the INPUT ID for the node with automatically assigned ID '$old_id' (and with children ", join (' ', @{$old_children}), "): $! !\n";
		return (undef, undef, undef, undef, undef);
	    }
	    my @in_children = ();
	    foreach my $old_chi (@{$old_children}) { # The inner foreach-loop #
		my $in_chi = $old_id2in_id{$old_chi};
		unless (defined $in_chi) {
		     print STDERR "[ERROR 'input_tree_with_nodeids'] CanNOT FIND the INPUT ID for the child (automarically assigned ID '$old_chi') of the node '$in_id' (automatically assigned ID '$old_id'): $! !\n";
		    return (undef, undef, undef, undef, undef);
		}

		push @in_children, $in_chi;
	    }				# End of the inner foreach-loop #

	    $in_nodeid2chi{$in_id} = \@in_children;
	}			# End of the 4th foreach-loop #

##=for DEBUGGING the subroutine
	print STDERR "\n --- End of the subroutine 'input_tree_with_nodeids' ---\n\n";
##=cut

	return (\%in_nodeid2pa, \%in_nodeid2chi, \%name2in_id, \%in_id2name, \%in_br_attr);

}		# End of the subroutin 'input_tree_with_nodeids' #

=cut # END of "for the predecessor 'MyTreeManipu_II.pm'."


=for the predecessor 'MyTreeManipu_II.pm'.

# copy_tree: This subroutine copies 4 out of 6 hashes which describes a tree.
#	It does NOT copy either %id2name or %name2id_intnode, with the assumption that they will not change forever.
#
sub copy_tree (\%\%\%\%) {

	my ($node2parent, $node2children, $name2id, $branch_attributes) = @_;

	my %cp_node2parent = %{$node2parent};
	my %cp_node2children;
	foreach my $node (%{$node2children}) {
	    my $children = $node2children->{$node};
 	    (defined $children) or next;
	    my @cp_children = @{$children};
	    $cp_node2children{$node} = \@cp_children;
	}
	my %cp_name2id = %{$name2id};
	my %cp_br_attr;
	foreach my $node (%{$branch_attributes}) {
	    my $attributes = $branch_attributes->{$node};
	    (defined $attributes) or next;
	    my @cp_attributes = @{$attributes};
	    $cp_br_attr{$node} =  \@cp_attributes;
	}

	return (\%cp_node2parent, \%cp_node2children, \%cp_name2id, \%cp_br_attr) ;

}		# end of the subroutine copy_tree #

=cut


###=for the moment, comment out the following subroutines in order to test the script little by little

# output_tree_dawg:
#	This modified version of 'output_tree' can also output the names of interior nodes, if any.
#	(CREATED on Feb 19 (Wed), 2014)
#
# output_tree:
#	This subroutine outputs a part of a tree under a specified node into a filehandle.
#	It outputs the daughter nodes in the ascinding order of the sizes of subtrees under them.
#
# INPUT: ($node_id, \%node2children, \%id2name, \%branch_attributes, $name_topnode, $filehandle,
#	$option_bootlabel, $option_nodeid),
#	where $node_id is the ID of the node which is the top of the part of the tree being output,
#	$option_bootlabel = 1 if you dsiplay  bootstrap at the branch, =0 (or undef) at the node.
#	$option_nodeid = 1 if you display node ID, =0 (or undef) otherwise.
#
sub output_tree_dawg ($\%\%\%$\*$$) {

	my ($node_id, $node2children, $id2name, $branch_attributes, $name_topnode, $fh,
		$option_bootlabel, $option_nodeid) = @_;

	my $tree = draw_subtree_dawg ($node_id, %{$node2children}, %{$id2name},
		%{$branch_attributes}, $option_bootlabel, $option_nodeid);

=set aside to see if it is necessary.
		# ADDED in creation of 'output_tree_dawg'. #
	my $attr = $branch_attributes->{$node_id};
	my $nodename = $id2name->{$node_id};
	(defined $nodename) or ($nodename = (defined $attr) ? $attr->[$INDX_INTNODENAME] : '');
=cut

		# when the node is an external node #
	(defined $node2children->{$node_id}) or ( ${$tree} =  "( ${$tree} )" ) ;

		# When the node is a top-node. # (ADDED in creation of 'output_tree_dawg'.) #
	unless (defined $branch_attributes->{$node_id}) {
	    (defined $name_topnode) and (${$tree} .= $name_topnode);
	}

	print $fh ${$tree} . "\;\n" ;

	return 1;

}		# End of the subroutine 'output_tree_dawg' #


# draw_subtree_dawg:
#	This modified version of 'draw_subtree' can also output names of interior nodes, if any.
#	(CREATED on Feg 19 (Wed), 2014)
#
# draw_subtree:
#	This subroutine outputs a part of a tree under a specified node
#	in a string of the Newick format.
#	It outputs the daughter nodes in the ascinding order of
#	the sizes of subtrees under them.
#	So the tree gets more and more complex as
#	it moves from left to right.
#
# INPUT: ($node_id, \%node2children, \%id2name, \%branch_attributes,
#	$option_bootlabel, $option_nodeid),
#	where $node_id is the ID of the node which is the top of
#	the part of the tree being output,
#	$option_bootlabel = 1 if you display bootstrap value at the branch,
#			=0 (or undef) at the node. 
#	$option_nodeid = 1 if you display node ID, =0 (or undef) otherwise.
#
sub draw_subtree_dawg ($\%\%\%$$) {

	my ($node_id, $node2children, $id2name, $branch_attributes,
		$option_bootlabel, $option_nodeid) = @_;

	$option_bootlabel = 0; # Forcefully set it at zero. (ADDED in creation of 'draw_subtree_dawg'.)

	my $tree = '';

	my $children = $node2children->{$node_id};
	if (defined $children) {

	    $tree .= "( ";

	    my @children_info = (); # element = [node_ct, \$subtree];
	    foreach my $child (@{$children}) {
		my $nodes = fetch_all_offsprings ($child, %{$node2children});
		my $node_ct = (defined $nodes) ? @{$nodes} : 0;
		my $subtree =
		draw_subtree_dawg ($child, %{$node2children}, %{$id2name},
			%{$branch_attributes}, $option_bootlabel, $option_nodeid);
			push @children_info , [$node_ct, $subtree];
	    }

	    my @std_children = sort { $a->[0] <=> $b->[0] } @children_info;
	    my $ct_children = @std_children;
	    for (my $i=0; $i <$ct_children; $i++) {
		$i and ($tree .= ', ') ; 
		$tree .=  ${$std_children[$i]->[1]};
	    }

	    $tree .= " )" ;

		# a tentative revision on Nov 7 (Sun), 2004 #
	     $option_nodeid and ($tree .= "{$node_id}");
	     

	} else { # the $node_id is probably an external node #

	    my $nodename = $id2name->{$node_id};
	    unless (defined $nodename) {
		print STDERR "NO NAME for the external node ID $node_id.\n";
		$nodename = $node_id;
	    }
	    $tree .= ($option_nodeid ? ("$nodename" . "{$node_id}") : $nodename ) ;
	}

	my $upper_branch = $branch_attributes->{$node_id};

	if (defined $upper_branch) {

		# MODIFIED in creation of 'draw_subtree_dawg'. #
	    my ($br_len, $nodename) = @{$upper_branch}[$INDX_LENGTH, $INDX_INTNODENAME];
	    (defined $nodename) and ($tree .= $nodename);
	    (defined $br_len) and ($tree .=  ":$br_len");
=for the predecessor 'draw_subtree_dawg'.
	    my ($br_len, $boot) = @{$upper_branch}[$INDX_LENGTH, $INDX_BOOT];
	    if (defined $br_len) {
		if ($option_bootlabel) {
		    $tree .= (defined $boot) ? (":$br_len" . "[$boot]") : ":$br_len";
		} else {
		    $tree .= (defined $boot) ? "$boot:$br_len" : ":$br_len";
		}
	    }
=cut
	}

	return \$tree;

}		# End of the subroutine 'draw_subtree_dawg' #


# br_output_tree_dawg:
#	This modified version of 'br_output_tree' can output the names of interior nodes, if any.
#	(CREATED on Feb 19 (Wed), 2014)
#
# br_output_tree:
#	This subroutine outputs a part of a tree under a specified node
#	into a filehandle.
#	It outputs the daughter nodes just in the order they are stored in the
#	array @{$node2children->{$node_id}}.
#
# INPUT: ($node_id, \%node2children, \%id2name, \%branch_attributes, $name_topnode, $filehandle,
#	$option_bootlabel, $option_nodeid),
#	where $node_id is the ID of the node which is the top of the part of the tree being output,
#	$option_bootlabel = 1 if you dsiplay  bootstrap at the branch, =0 (or undef) at the node.
#	$option_nodeid = 1 if you display node ID, =0 (or undef) otherwise.
#
sub br_output_tree_dawg ($\%\%\%$\*$$) {

	my ($node_id, $node2children, $id2name, $branch_attributes, $name_topnode, $fh,
		$option_bootlabel, $option_nodeid) = @_;

	$option_bootlabel = 0; # Forcefully set it at Zero. (ADDED in creation of 'br_output_tree_dawg'.)

	my $tree = br_draw_subtree_dawg ($node_id, %{$node2children}, %{$id2name},
		%{$branch_attributes}, $option_bootlabel, $option_nodeid);

		# when the node is an external node #
	(defined $node2children->{$node_id}) or ( ${$tree} =  "( ${$tree} )" ) ;

		# When the node is the genuine top-node. # (ADDED in creation of 'br_output_tree_dawg'.) #
	unless (defined $branch_attributes->{$node_id}) {
	   (defined $name_topnode) and (${$tree} .= $name_topnode) ; 
	}

	print $fh ${$tree} . "\;\n" ;

	return 1;

}		# end of the subroutine 'br_output_tree_dawg' #



# br_draw_subtree_dawg:
#	This modified version of 'br_draw_subtree' can also output the names of interior nodes, if any.
#	(CREATED on Feb 19 (Wed), 2014)
#
# br_draw_subtree:
#	This subroutine outputs a part of a tree under a specified node
#	in a string of the Newick format.
#	It outputs the daughter nodes in the order
#	stored in the array @{$node2children->{$node_id}}.
#
# INPUT: ($node_id, \%node2children, \%id2name, \%branch_attributes,
#	$option_bootlabel, $option_nodeid),
#	where $node_id is the ID of the node which is the top of
#	the part of the tree being output,
#	$option_bootlabel = 1 if you display bootstrap value at the branch,
#			=0 (or undef) at the node. 
#	$option_nodeid = 1 if you display node ID, =0 (or undef) otherwise.
#
sub br_draw_subtree_dawg ($\%\%\%$$) {

	my ($node_id, $node2children, $id2name, $branch_attributes,
		$option_bootlabel, $option_nodeid) = @_;

	my $tree = '';

	my $children = $node2children->{$node_id};
	if (defined $children) {

	    $tree .= "( ";

	    my $ct_children = @{$children};
	    for (my $i=0; $i <$ct_children; $i++) {

		$i and ($tree .= ', ') ; 

		my $child = $children->[$i];
		my $subtree =
		br_draw_subtree_dawg ($child, %{$node2children}, %{$id2name},
		   %{$branch_attributes}, $option_bootlabel, $option_nodeid);
		$tree .=  ${$subtree};
	    }

	    $tree .= " )" ;

		# a tentative revision on Nov 7 (Sun), 2004 #
	     $option_nodeid and ($tree .= "{$node_id}");
	     

	} else { # the $node_id is probably an external node #

	    my $nodename = $id2name->{$node_id};
	    unless (defined $nodename) {
		print STDERR "[WANING 'br_draw_subtree'] NO NAME for the external node ID $node_id.\n";
		$nodename = $node_id;
	    }
	    $tree .= ($option_nodeid ? ("$nodename" . "{$node_id}") : $nodename ) ;
	}

	my $upper_branch = $branch_attributes->{$node_id};

	if (defined $upper_branch) {

	    my ($br_len, $nodename) = @{$upper_branch}[$INDX_LENGTH, $INDX_INTNODENAME];
	    (defined $nodename) and ($tree .= $nodename);
	    ((defined $br_len) and ($br_len =~ m/\S/)) and ($tree .= ":$br_len");
=for the predecessor 'br_draw_subtree_dawg'
	    my ($br_len, $boot) = @{$upper_branch}[$INDX_LENGTH, $INDX_BOOT];

		## REVISED on Thursday, Sep 21, 2006 ##
	    if ((defined $br_len) or (defined $boot)) {
		(defined $br_len) or ($br_len = '');
		if ($option_bootlabel) {
		    $tree .= (defined $boot) ? (":$br_len" . "[$boot]") : ":$br_len";
		} else {
		    $tree .= (defined $boot) ? "$boot:$br_len" : ":$br_len";
		}
	    }
=cut
	}

	return \$tree;

}		# End of the subroutine 'br_draw_subtree_dawg' #


# br_output_tree_sep_dawg:
#	This modified version of 'br_output_tree_sep' can also output the names of interior nodes, if any.
#	(CREATED on Feb 19 (Wed), 2014)
#
# br_output_tree_sep:
#	A revised version of 'br_output_tree' with the freedom of choice
#	of the 'separator' to delimit the nodes in the tree.
#
# br_output_tree:
#	This subroutine outputs a part of a tree under a specified node
#	into a filehandle.
#	It outputs the daughter nodes just in the ordered in the
#	array @{$node2children->{$node_id}}.
#
# INPUT: ($node_id, \%node2children, \%id2name, \%branch_attributes, $name_topnode, *${$filehandle},
#	$option_bootlabel, $option_nodeid, $separator),
#	where $node_id is the ID of the node which is the top of the part of the tree being output,
#	$option_bootlabel = 1 if you dsiplay  bootstrap at the branch, =0 (or undef) at the node.
#	$option_nodeid = 1 if you display node ID, =0 (or undef) otherwise.
#
sub br_output_tree_sep_dawg ($\%\%\%$\*$$$) {

	my ($node_id, $node2children, $id2name, $branch_attributes, $name_topnode, $fh,
		$option_bootlabel, $option_nodeid, $separator) = @_;

	$option_bootlabel = 0; # Forcefully set it at ZERO. (ADDED in creation of 'br_output_tree_sep_dawg'.)

	my $tree = br_draw_subtree_sep_dawg ($node_id, %{$node2children}, %{$id2name},
		%{$branch_attributes}, $option_bootlabel, $option_nodeid, $separator);

		# when the node is an external node #
	(defined $node2children->{$node_id}) or ( ${$tree} =  "($separator" . ${$tree} . "$separator)" ) ;

		# When the node is the top-node. # (ADDED in creation of 'br_output_tree_sep_dawg'.)
	unless (defined $branch_attributes->{$node_id}) {
	   (defined $name_topnode) and (${$tree} .= $name_topnode) ; 
	}

	print $fh ${$tree} . "\;\n" ;

	return 1;

}		# end of the subroutine 'br_output_tree_sep_dawg' #


# br_draw_subtree_sep_dawg:
#	This modified version of 'br_draw_subtree_sep' can also output
#	the names of interior nodes, if any.
#	(CREATED on Sep 19 (Wed), 2014)
#
# br_draw_subtree_sep:
#	A revised version of 'br_draw_subtree' with a freedom of choice
#	of the 'sepatator' that delimits the nodes in the output tree.
#
# br_draw_subtree:
#	This subroutine outputs a part of a tree under a specified node
#	in a string of the Newick format.
#	It outputs the daughter nodes in the order
#	stored in the array @{$node2children->{$node_id}}.
#
# INPUT: ($node_id, $node2children, $id2name, $branch_attributes,
#	$option_bootlabel, $option_nodeid, $separator),
#	where $node_id is the ID of the node which is the top of
#	the part of the tree being output,
#	$option_bootlabel = 1 if you display bootstrap value at the branch,
#			=0 (or undef) at the node. 
#	$option_nodeid = 1 if you display node ID, =0 (or undef) otherwise.
#	$separator specifies the separator that delimits the nodes in the output tree.
#
sub br_draw_subtree_sep_dawg ($\%\%\%$$$) {

	my ($node_id, $node2children, $id2name, $branch_attributes,
		$option_bootlabel, $option_nodeid, $separator) = @_;

	my $tree = '';

	my $children = $node2children->{$node_id};
	if (defined $children) {

	    $tree .= "($separator";

	    my $ct_children = @{$children};
	    for (my $i=0; $i <$ct_children; $i++) {

		$i and ($tree .= ',' . $separator) ; 

		my $child = $children->[$i];
		my $subtree =
		br_draw_subtree_sep_dawg ($child, %{$node2children}, %{$id2name},
		   %{$branch_attributes}, $option_bootlabel, $option_nodeid, $separator);
		$tree .=  ${$subtree};
	    }

	    $tree .= "$separator)" ;

		# a tentative revision on Nov 7 (Sun), 2004 #
	     $option_nodeid and ($tree .= "{$node_id}");
	     

	} else { # the $node_id is probably an external node #

	    my $nodename = $id2name->{$node_id};
	    unless (defined $nodename) {
		print STDERR "[WANING 'br_draw_subtree_sep_dawg'] NO NAME for the external node ID $node_id.\n";
		$nodename = $node_id;
	    }
	    $tree .= ($option_nodeid ? ("$nodename" . "{$node_id}") : $nodename ) ;
	}

		### RESTART FROM HERE ###

	my $upper_branch = $branch_attributes->{$node_id};

	if (defined $upper_branch) {

		# MODIFIED in creation of 'br_draw_subtree_sep_dawg'. #
	    my ($br_len, $nodename) = @{$upper_branch}[$INDX_LENGTH, $INDX_INTNODENAME];
	    (defined $nodename) and ($tree .= $nodename);
	    ((defined $br_len) and ($br_len =~ m/\S/)) and ($tree .= ":$br_len");
=for the predecessor 'br_draw_subtree_sep'.
	    my ($br_len, $boot) = @{$upper_branch}[$INDX_LENGTH, $INDX_BOOT];

			## REVISED  on Thursday, Sep 21, 2006 ##
	    if ((defined $br_len) or (defined $boot)) {
		(defined $br_len) or ($br_len = '');
		if ($option_bootlabel) {
		    $tree .= (defined $boot) ? (":$br_len" . "[$boot]") : ":$br_len";
		} else {
		    $tree .= (defined $boot) ? "$boot:$br_len" : ":$br_len";
		}
	    }
=cut
	}

	return \$tree;

}		# end of the subroutine 'br_draw_subtree_sep_dawg' #


=for the predecessor 'MyTreeManipu_II.pm'.

sub topo_subtree ($\%\%\%);

# topo_subtree: This subroutine returns the reference to  a string
#		which shows a simple tree in which each sequence name
#		is replaced by a  simple code indicating its source species.
#		It outputs the daughter nodes in the ascinding order of the sizes of subtrees under them.
#		So the tree gets more and more complex as it moves from left to right.
#
# INPUT: ($node_id, \%node2children, \%id2name, \%name2code),
#
sub topo_subtree ($\%\%\%) {

	my ($node_id, $node2children, $id2name, $name2code) = @_;

	my $topo = '';

	my $children = $node2children->{$node_id};
	if (defined $children) {

	    $topo .= "(";

	    my @children_info = (); # element = [node_ct, \$subtopo];
	    foreach my $child (@{$children}) {
		my $nodes = fetch_all_offsprings ($child, %{$node2children});
		my $node_ct = (defined $nodes) ? @{$nodes} : 0;
		my $subtopo =
		topo_subtree ($child, %{$node2children}, %{$id2name}, %{$name2code});
			push @children_info , [$node_ct, $subtopo];
	    }
	    my @std_children = sort { $a->[0] <=> $b->[0] } @children_info;
	    my $ct_children = @std_children;
	    for (my $i=0; $i <$ct_children; $i++) {
		$i and ($topo .= ',') ; 
		$topo .=  ${$std_children[$i]->[1]};
	    }

	    $topo .= ")" ;

	} else { # the $node_id is probably an external node #

	    my $nodename = $id2name->{$node_id};
	    unless (defined $nodename) {
		print STDERR "NO NAME for the external node ID $node_id.\n";
		$nodename = $node_id;
	    }
	    my $code  = $name2code->{$nodename};
	    unless (defined $code) {
		print STDERR "NO CODE for the external node ID $node_id.\n";
		$code = $node_id;
	    }
	    $topo .= $code;
	}

	return \$topo;

}		# end of the subroutine topo_subtree #

=cut  # END of "for the predecessor 'MyTreeManipu_II.pm'."


1;		# a "spell" which makes the module "use"-able and/or "require"-able #

