package MyTreeManipu_II_spe;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.5: Copyright (C) 2014 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyTreeManipu_II_spe.pm:
#	This module processes more special types of tree inputs and/or outputs.
#
# MyTreeManipu_II.pm: This package can process more general types of trees,
#	(i.e. NOT limited to binary trees with trichotomous top nodes).
#	It is also consistent with the format of the taxonomy tree which are handled by
#	the module 'MyTaxonomy.pm'.
#
# MyTreeManipu.pm: This package performs basic tree manipulations.
#
#  Subroutines whose substitutes are available in this module ('MyTreeManipu_II.pm')
#	or in 'MyTaxonomy.pm':
#   our @EXPORT = qw(retrieve_top_node confine_outgroup print_unrooted);
#   our @EXPORT_OK = qw(switch_topnode  print_subtree topo_unrooted topo_subtree
#		list_sequences fetch_node_series dist_nodes
#		    );
#
# Subroutines whose substitutes are not available yet:
#   our @EXPORT_OK = ( retrieve_bootstrap  prod_bootstrap );
#

use strict;
use warnings;
use MyArrayManipu qw( diff_array union_array );
use MyTaxonomy qw( fetch_ancestry fetch_last_common_ancestor_ov2 fetch_common_ancestors
		 fetch_ext_offsprings fetch_all_offsprings );
use MyTreeManipu_II qw( );

use Exporter;

our @ISA = qw(Exporter);

##sub rec_fillin_boot (\@\$);
#sub rec_prod_boot (\@\@);


our @EXPORT_OK = qw(
	replace_nodeids
	input_ext_tree_file read_ext_tree_file
	input_ext_subtree
			) ;
	## copy_tree
	## output_tree draw_subtree
	## br_output_tree br_draw_subtree topo_subtree
	## br_output_tree_sep br_draw_subtree_sep

sub read_ext_tree_file (\$\$\%\%\%\%\%\%);
## sub draw_subtree ($\%\%\%$$);
## sub br_draw_subtree ($\%\%\%$$);
## sub br_draw_subtree_sep ($\%\%\%$$$);

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
## our $START_GROUP = '('; # the symbol marking the start of the grouping of nodes.
## our $END_GROUP = ')'; # the symbol marking the end of the grouping of nodes.


=Five important hash tables which describe a tree ###
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
#
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.
=cut



# replace_nodeids:
#	This subroutine replaces old node IDs with new ones.
#	If there is no new ID correspinding to an old ID,
#	it assigns a new ID that never disturbs the existing correspondence.
#
# INPUT: (\%node2parent, \%node2children, \%name2id, %id2name, %branch_attributes, %oldid2newid), where
#	%node2parent = ($node_id => $parent_id, ...);
#	%node2children = ($node_id => [a list of child IDs], ...);
#	%name2id = ($node_name => $node_id, ...);
#	%id2name = ($node_id => $node_name, ...);
#	%branch_attributes = ($node_id => [attributes of the branch immediately above the node], ...);
#	%oldid2newid = ($old_node_id => $new_node_id,...).
#
# OUTPUT: 1 if the subroutine succeeded,
#	  0 if it failed.
#
sub replace_nodeids (\%\%\%\%\%\%) {

	my ($node2pa, $node2chi, $name2id, $id2name, $br_attr, $oldid2newid) = @_;

		# Identify the maximum value of the new node ID #
	my $maxid = 0;
	foreach my $oldid (keys %{$oldid2newid}) { # The 1st foreach-loop #
	    my $newid = $oldid2newid->{$oldid};
	    ($maxid < $newid) and ($maxid = $newid);
	}			# End of the 1st foreach-loop #
	my $extra_id = $maxid + 1;

		# replace the values in %{$name2id} #
	foreach my $name (keys %{$name2id}) { # The 2nd foreach-loop #
	    my $id = $name2id->{$name};
	    my $newid = $oldid2newid->{$id};
	    unless (defined $newid) {
		$oldid2newid->{$id} = $newid = $extra_id;
		$extra_id++;
	    }
	    $name2id->{$name} = $newid;
	}				# End of the 2nd foreach-loop #

		# replace node IDs in the remaining hashes #
	my %new_node2pa; my %new_node2chi; my %new_id2name; my %new_br_attr;

		# %{$node2pa} #
	foreach my $node (keys %{$node2pa}) { # The 3rd foreach-loop #
	    my $parent = $node2pa->{$node};
	    my $newid = $oldid2newid->{$node};
	    unless (defined $newid) {
		$oldid2newid->{$node} = $newid = $extra_id;
		$extra_id++;
	    }
	    my $new_pa = $oldid2newid->{$parent};
	    unless (defined $new_pa) {
		$oldid2newid->{$parent} = $new_pa = $extra_id;
		$extra_id++;
	    }
	    $new_node2pa{$newid} = $new_pa;
	}				# End of the 3rd foreach-loop #
	%{$node2pa} = %new_node2pa;

		# %{$node2chi} #
	foreach my $node (keys %{$node2chi}) { # The 4th foreach-loop # 
	    my $newid = $oldid2newid->{$node};
	    unless (defined $newid) {
		$oldid2newid->{$node} = $newid = $extra_id;
		$extra_id++;
	    }
	    my $children = $node2chi->{$node};
	    my @new_children = ();
	    foreach my $ch (@{$children}) { # The inner foreach-loop #
		my $new_ch = $oldid2newid->{$ch};
		unless (defined $new_ch) {
		    $oldid2newid->{$ch} = $new_ch = $extra_id;
		    $extra_id++;
		}
		push @new_children, $new_ch;
	    }
	    $new_node2chi{$newid} = \@new_children;
	}				# End of the 4th foreach-loop #
	%{$node2chi} = %new_node2chi;

		# %{$id2name} #
	foreach my $id (keys %{$id2name}) { # The 5th foreach-loop #
	    my $name = $id2name->{$id};
	    (defined $name) or next;
	    my $newid = $oldid2newid->{$id};
	    unless (defined $newid) {
		$oldid2newid->{$id} = $newid = $extra_id;
		$extra_id++;
	    }
	    $new_id2name{$newid} = $name;
	}			# End of the 5th foreach-loop #
	%{$id2name} = %new_id2name;

		# %{$br_attr} #
	foreach my $id (keys %{$br_attr}) { # The 6th foreach-loop # 
	    my $attr = $br_attr->{$id};
	    (defined $attr) or next;
	    my $newid = $oldid2newid->{$id};
	    unless (defined $newid) {
		$oldid2newid->{$id} = $newid = $extra_id;
		$extra_id++;
	    }
	    $new_br_attr{$newid} = $attr;
	}			# End of the 6th foreach-loop #
	%{$br_attr} = %new_br_attr;

	return 1;

}		# end of the subroutine 'replace_nodeids' #


# input_ext_tree_file:
#	This subroutine reads in a tree file and stores the information on a whole tree.
#	This is actually an I/O interface of the subroutine 'read_tree_file'.
#
# INPUT is supposed to start with the left parenthesis ('(') , and end with the semicolon ';',
# possibly with some comments written in alphabets,
# and children are supposed to be separated by the comma (',').
#
# $fl_dis_nodeids = 1 if you want to use the node IDs displayed in the input tree,
#		  = 0 if you want to use the node ID assigned while reading in the tree.
#
# OUTPUT: \%nodeid2disid if the subroutine succeeded,
#	  undef if it failed,
#	where
#	%nodeid2disid = ($node_id assigned when read in => $node_id displayed in the input, ...).
#
sub input_ext_tree_file (\$\%\%\%\%\%$) {

	my ($input, $node2parent, $node2children, $name2id, $id2name,
		$branch_attributes, $fl_dis_nodeids) = @_; 

	${$input} =~ s/\;\s*$//; # remove trailing white spaces plus the ending mark of the tree.
	my ($top_info) = (${$input} =~ m/([^\)]+)$/ );
	if (defined $top_info) {
	    my $len_topinfo = length ($top_info);
	    substr (${$input}, -$len_topinfo, $len_topinfo, '');
	}

	unless (${$input} =~ s/^\s*\(\s*// ) { # remove the left parenthesis at the beginning of the tree
	    print STDERR "ILLEGAL input for 'input_tree_file' .\n";
	    return undef;
	}

	my $node_id = 0;

	my %nodeid2disid; # ($node_id read in => $node_id displayed in the input, ...)

	unless ( read_ext_tree_file (${$input}, $node_id, %{$node2parent}, %{$node2children},
			%{$name2id}, %{$id2name}, %{$branch_attributes}, %nodeid2disid) ) {

	    print STDERR "'read_tree_file' failed at the node ID $node_id .\n";
	    return undef;
	}

	$name2id->{$TOP_NODE} = $node_id;

	if (defined $branch_attributes->{$node_id}) {
		print STDERR "CAUTION: the top_node (ID:$node_id) has somehow information on its 'upper branch', which CANNOT exist:\n";
		print STDERR ' ' x 5 . '{' . join (', ' , @{$branch_attributes->{$node_id}} ) . "\} .\n";
		undef $branch_attributes->{$node_id};
	}

		## A Distinctive Addition to the predesessor 'input_ext_tree_file' ##
	my ($new_topid) = (defined $top_info) ? ($top_info =~ m/\{([^{}]+)\}/) : (undef);
	(defined $new_topid) and ($nodeid2disid{$node_id} = $new_topid);
	if ($fl_dis_nodeids) {
	    replace_nodeids (%{$node2parent}, %{$node2children}, %{$name2id}, %{$id2name}, %{$branch_attributes}, %nodeid2disid);
	}

	return \%nodeid2disid;

}		# end of the subroutine 'input_ext_tree_file' #


# input_ext_subtree: This subroutine was made to retrieve the complete information from an
#	output of the subroutine 'draw_subtree'.
#
sub input_ext_subtree (\$$) {

	my ($in_subtree, $fl_dis_nodeids) = @_;
	${$in_subtree} =~ s/\s*\;\s*$//;

	my ($dist, $boot) ;

		## REVISED from the predecessor 'input_ext_subtree' ##
	##my ($appended) = (${$in_subtree} =~ m/([^\)]+)$/ );
	my ($appended) = (${$in_subtree} =~ m/([^\)\}]+)$/ );
	if (defined $appended) {
	    my $app_len = length ($appended);
	    substr (${$in_subtree}, - $app_len, $app_len, ';'); # remove the appended information from the input
	}

	my %node2parent = my %node2children = my %name2nodeid = my %nodeid2name = my %branch_attributes = ();

	unless ( input_ext_tree_file (${$in_subtree},
		%node2parent, %node2children, %name2nodeid, %nodeid2name, %branch_attributes, $fl_dis_nodeids) ) {

		print STDERR "'input_tree_file' failed.\n";
		return (undef, undef, undef, undef, undef);
	}

	if ( defined $appended) {

			# when -bootlabel=node
		my ($bootstrap, $br_len) = split /\:/, $appended;

			# when -bootlabel=branch
		($bootstrap eq '') and ( ($br_len, $bootstrap) = ($appended =~ m/^\:([^\[]+)\[([^\]]+)\]/ ) );

			# when the branch has no bootstrap value displayed
		(defined $br_len) or ( ($br_len  = $appended) =~ s/^\:// ) ;
		( (defined $bootstrap) and ($bootstrap eq '') ) and (undef $bootstrap);

		my $topnode = $name2nodeid{$TOP_NODE};
		$branch_attributes{$topnode} = [$br_len, $bootstrap];
	}

	return (\%node2parent, \%node2children, \%name2nodeid, \%nodeid2name, \%branch_attributes );

}		# end of the subroutine 'input_ext_subtree' #


# read_ext_tree_file: This subroutine reads in (a part of) a extended tree-file text
#		and stores the information on the 5 hash tables describing trees.
#
# INPUT is supposed to start with a left parenthesis ('(') ,
#	which must have been removed before being passed to this subroutine,
#	and end with a right parenthesis followed by a semicolon ';',
# 	possibly with some comments (in alphabets) in between,
#	and children are supposed to be separated by commas (',').
#
# 	%nodeid2disid = ($node_id read in => $node_id displayed, ...)
#
sub read_ext_tree_file (\$\$\%\%\%\%\%\%) {

	my ($input, $node_id, $node2parent, $node2children, $name2id, $id2name, $branch_attributes, $nodeid2disid) = @_;

	my @children = ();

	#while (${$input} !~ m/^\)/ ) ) { # the outermost while #
	while ( !(${$input} =~ s/^\)\s*// ) ) { # the outermost while
			# --- continues untill it reaches the end of the group of nodes

	    if (${$input} =~ s/^\s*\(\s*// ) { # reading in a group of nodes #

		unless (read_ext_tree_file (${$input}, ${$node_id}, %{$node2parent}, %{$node2children},
				%{$name2id}, %{$id2name}, %{$branch_attributes}, %{$nodeid2disid} ) ) {
			return 0; # error occurred #
		}

		push @children, ${$node_id};

	    } else { # reading in a single node (probably external) #

		my ($tmp) = (${$input} =~ m/^([^\,\)]+)/ );
		unless (defined $tmp)  {
			print STDERR "Unexpected input : " . substr (${$input}, 0, 10) . "...\n";
			return 0;
		}

		substr (${$input}, 0, length ($tmp) , ''); # remove $tmp from the input

		$tmp =~ s/\s+$//;

		#my ($seqname, $br_len) = ($tmp =~ m/^([^\:]+)\:([^\:]+)/ );
		my ($seqname, $br_len) = split /\:/, $tmp;
		(defined $seqname) or ($seqname = $tmp);
		${$node_id}++;

			## An important addition to the predecessor 'read_tree_file' ##
		$seqname =~ s/\s+$//;
		if ($seqname =~ s/\}$//) {
		    my ($disid) = ($seqname =~ m/\{(\d+)$/ );
		    if (defined $disid) {
			my $len = length ($disid) + 1;
			substr ($seqname, - $len, $len, '');
			$nodeid2disid->{${$node_id}} = $disid;
		    } else {
			$seqname .= '}';
		    }
		}

		push @children, ${$node_id};
		$name2id->{$seqname} = ${$node_id};
		$id2name->{${$node_id}} = $seqname;

		

		(defined $br_len) and ($branch_attributes->{${$node_id} } = [$br_len, undef] );

	    }

	    ${$input} =~ s/^\,\s*//; # remove the comma delimiting this group from the following one

	}		# end of the outermost while #

	##${$input} =~ s/^\)\s*//;

	${$node_id}++; # The ID of the node that has just been read in.

	$node2children->{${$node_id}} = \@children;
	foreach my $child (@children) { $node2parent->{$child} = ${$node_id}; }
		
	my ($tmp) = (${$input} =~ m/^([^\,\)]+)/ ); # information belonging to the immediate upper branch 
	if (defined $tmp) {

	    substr (${$input}, 0, length ($tmp), '' ); # remove $tmp from the input
	    $tmp =~ s/\s+$//;

		## An important addition to the predecessor 'read_tree_file' ##
	    if ($tmp =~ s/^\{//) {
		my ($disid) = ($tmp =~ m/^(\d+)\}/ );
		if (defined $disid) {
	  	   substr ($tmp, 0, length ($disid) + 1, '');
		   $nodeid2disid->{${$node_id}} = $disid;
		} else {
		   $tmp = '{' . $tmp;
		}
	    }

	    	# when -bootlabel=node
	    my ($bootstrap, $br_len) = split /\:/, $tmp;

	    	# when -bootlabel=branch
	    ($bootstrap eq '') and ( ($br_len, $bootstrap) = ($tmp =~ m/^\:([^\[]+)\[([^\]]+)\]/ ) );

	    	# when the branch has no bootstrap value displayed
	    (defined $br_len) or ( ($br_len  = $tmp) =~ s/^\:// ) ;
	    ( (defined $bootstrap) and ($bootstrap eq '') ) and (undef $bootstrap);

	    $branch_attributes->{${$node_id}} = [$br_len, $bootstrap];
	}

	return 1;

}		# end of the subroutine 'read_ext_tree_file' #


=set aside for potential later use

# copy_tree: This subroutine copies 4 out of 5 hashes which describes a tree.
#	This does NOT copy %id2name with the assumption that they will not change forever.
#
sub copy_tree (\%\%\%\%) {

	my ($node2parent, $node2children, $name2id, $branch_attributes) = @_;

	my %cp_node2parent = %{$node2parent};
	my %cp_node2children;
	foreach my $node (%{$node2children}) {
	    my $children = $node2children->{$node};
 	    (defined $children) or next;
	    my @cp_children = @{$children};
	    $cp_node2children{$node} = \@cp_children;
	}
	my %cp_name2id = %{$name2id};
	my %cp_br_attr;
	foreach my $node (%{$branch_attributes}) {
	    my $attributes = $branch_attributes->{$node};
	    (defined $attributes) or next;
	    my @cp_attributes = @{$attributes};
	    $cp_br_attr{$node} =  \@cp_attributes;
	}

	return (\%cp_node2parent, \%cp_node2children, \%cp_name2id, \%cp_br_attr) ;

}		# end of the subroutine copy_tree #


###=for the moment, comment out the following subroutines in order to test the script little by little

# output_tree: This subroutine outputs a part of a tree under a specified node into a filehandle.
#		It outputs the daughter nodes in the ascinding order of the sizes of subtrees under them.
#
# INPUT: ($node_id, $node2children, $id2name, $branch_attributes, $filehandle,
#	$option_bootlabel, $option_nodeid),
#	where $node_id is the ID of the node which is the top of the part of the tree being output,
#	$option_bootlabel = 1 if you dsiplay  bootstrap at the branch, =0 (or undef) at the node.
#	$option_nodeid = 1 if you display node ID, =0 (or undef) otherwise.
#
sub output_tree ($\%\%\%\*$$) {

	my ($node_id, $node2children, $id2name, $branch_attributes, $fh,
		$option_bootlabel, $option_nodeid) = @_;

	my $tree = draw_subtree ($node_id, %{$node2children}, %{$id2name},
		%{$branch_attributes}, $option_bootlabel, $option_nodeid);

		# when the node is an external node #
	(defined $node2children->{$node_id}) or ( ${$tree} =  "( ${$tree} )" ) ;


	print $fh ${$tree} . "\;\n" ;

	return 1;

}		# end of the subroutine output_tree #


# draw_subtree:
#	This subroutine outputs a part of a tree under a specified node
#	in a string of the Newick format.
#	It outputs the daughter nodes in the ascinding order of
#	the sizes of subtrees under them.
#	So the tree gets more and more complex as
#	it moves from left to right.
#
# INPUT: ($node_id, $node2children, $id2name, $branch_attributes,
#	$option_bootlabel, $option_nodeid),
#	where $node_id is the ID of the node which is the top of
#	the part of the tree being output,
#	$option_bootlabel = 1 if you display bootstrap value at the branch,
#			=0 (or undef) at the node. 
#	$option_nodeid = 1 if you display node ID, =0 (or undef) otherwise.
#
sub draw_subtree ($\%\%\%$$) {

	my ($node_id, $node2children, $id2name, $branch_attributes,
		$option_bootlabel, $option_nodeid) = @_;

	my $tree = '';

	my $children = $node2children->{$node_id};
	if (defined $children) {

	    $tree .= "( ";

	    my @children_info = (); # element = [node_ct, \$subtree];
	    foreach my $child (@{$children}) {
		my $nodes = fetch_all_offsprings ($child, %{$node2children});
		my $node_ct = (defined $nodes) ? @{$nodes} : 0;
		my $subtree =
		draw_subtree ($child, %{$node2children}, %{$id2name},
			%{$branch_attributes}, $option_bootlabel, $option_nodeid);
			push @children_info , [$node_ct, $subtree];
	    }

	    my @std_children = sort { $a->[0] <=> $b->[0] } @children_info;
	    my $ct_children = @std_children;
	    for (my $i=0; $i <$ct_children; $i++) {
		$i and ($tree .= ', ') ; 
		$tree .=  ${$std_children[$i]->[1]};
	    }

	    $tree .= " )" ;

		# a tentative revision on Nov 7 (Sun), 2004 #
	     $option_nodeid and ($tree .= "{$node_id}");
	     

	} else { # the $node_id is probably an external node #

	    my $nodename = $id2name->{$node_id};
	    unless (defined $nodename) {
		print STDERR "NO NAME for the external node ID $node_id.\n";
		$nodename = $node_id;
	    }
	    $tree .= ($option_nodeid ? ("$nodename" . "{$node_id}") : $nodename ) ;
	}

	my $upper_branch = $branch_attributes->{$node_id};

	if (defined $upper_branch) {
	    my ($br_len, $boot) = @{$upper_branch}[$INDX_LENGTH, $INDX_BOOT];
	    if (defined $br_len) {
		if ($option_bootlabel) {
		    $tree .= (defined $boot) ? (":$br_len" . "[$boot]") : ":$br_len";
		} else {
		    $tree .= (defined $boot) ? "$boot:$br_len" : ":$br_len";
		}
	    }
	}

	return \$tree;

}		# end of the subroutine draw_subtree #


# br_output_tree:
#	This subroutine outputs a part of a tree under a specified node
#	into a filehandle.
#	It outputs the daughter nodes just in the ordered in the
#	array @{$node2children->{$node_id}}.
#
# INPUT: ($node_id, $node2children, $id2name, $branch_attributes, $filehandle,
#	$option_bootlabel, $option_nodeid),
#	where $node_id is the ID of the node which is the top of the part of the tree being output,
#	$option_bootlabel = 1 if you dsiplay  bootstrap at the branch, =0 (or undef) at the node.
#	$option_nodeid = 1 if you display node ID, =0 (or undef) otherwise.
#
sub br_output_tree ($\%\%\%\*$$) {

	my ($node_id, $node2children, $id2name, $branch_attributes, $fh,
		$option_bootlabel, $option_nodeid) = @_;

	my $tree = br_draw_subtree ($node_id, %{$node2children}, %{$id2name},
		%{$branch_attributes}, $option_bootlabel, $option_nodeid);

		# when the node is an external node #
	(defined $node2children->{$node_id}) or ( ${$tree} =  "( ${$tree} )" ) ;

	print $fh ${$tree} . "\;\n" ;

	return 1;

}		# end of the subroutine 'br_output_tree' #


# br_draw_subtree:
#	This subroutine outputs a part of a tree under a specified node
#	in a string of the Newick format.
#	It outputs the daughter nodes in the order
#	stored in the array @{$node2children->{$node_id}}.
#
# INPUT: ($node_id, $node2children, $id2name, $branch_attributes,
#	$option_bootlabel, $option_nodeid),
#	where $node_id is the ID of the node which is the top of
#	the part of the tree being output,
#	$option_bootlabel = 1 if you display bootstrap value at the branch,
#			=0 (or undef) at the node. 
#	$option_nodeid = 1 if you display node ID, =0 (or undef) otherwise.
#
sub br_draw_subtree ($\%\%\%$$) {

	my ($node_id, $node2children, $id2name, $branch_attributes,
		$option_bootlabel, $option_nodeid) = @_;

	my $tree = '';

	my $children = $node2children->{$node_id};
	if (defined $children) {

	    $tree .= "( ";

	    my $ct_children = @{$children};
	    for (my $i=0; $i <$ct_children; $i++) {

		$i and ($tree .= ', ') ; 

		my $child = $children->[$i];
		my $subtree =
		br_draw_subtree ($child, %{$node2children}, %{$id2name},
		   %{$branch_attributes}, $option_bootlabel, $option_nodeid);
		$tree .=  ${$subtree};
	    }

	    $tree .= " )" ;

		# a tentative revision on Nov 7 (Sun), 2004 #
	     $option_nodeid and ($tree .= "{$node_id}");
	     

	} else { # the $node_id is probably an external node #

	    my $nodename = $id2name->{$node_id};
	    unless (defined $nodename) {
		print STDERR "[WANING 'br_draw_subtree'] NO NAME for the external node ID $node_id.\n";
		$nodename = $node_id;
	    }
	    $tree .= ($option_nodeid ? ("$nodename" . "{$node_id}") : $nodename ) ;
	}

	my $upper_branch = $branch_attributes->{$node_id};

	if (defined $upper_branch) {
	    my ($br_len, $boot) = @{$upper_branch}[$INDX_LENGTH, $INDX_BOOT];

		## REVISED on Thursday, Sep 21, 2006 ##
	    if ((defined $br_len) or (defined $boot)) {
		(defined $br_len) or ($br_len = '');
		if ($option_bootlabel) {
		    $tree .= (defined $boot) ? (":$br_len" . "[$boot]") : ":$br_len";
		} else {
		    $tree .= (defined $boot) ? "$boot:$br_len" : ":$br_len";
		}
	    }
	}

	return \$tree;

}		# end of the subroutine br_draw_subtree #


# br_output_tree_sep:
#	A revised version of 'br_output_tree' with the freedom of choice
#	of the 'separator' to delimit the nodes in the tree.
#
# br_output_tree:
#	This subroutine outputs a part of a tree under a specified node
#	into a filehandle.
#	It outputs the daughter nodes just in the ordered in the
#	array @{$node2children->{$node_id}}.
#
# INPUT: ($node_id, $node2children, $id2name, $branch_attributes, $filehandle,
#	$option_bootlabel, $option_nodeid, $separator),
#	where $node_id is the ID of the node which is the top of the part of the tree being output,
#	$option_bootlabel = 1 if you dsiplay  bootstrap at the branch, =0 (or undef) at the node.
#	$option_nodeid = 1 if you display node ID, =0 (or undef) otherwise.
#
sub br_output_tree_sep ($\%\%\%\*$$$) {

	my ($node_id, $node2children, $id2name, $branch_attributes, $fh,
		$option_bootlabel, $option_nodeid, $separator) = @_;

	my $tree = br_draw_subtree_sep ($node_id, %{$node2children}, %{$id2name},
		%{$branch_attributes}, $option_bootlabel, $option_nodeid, $separator);

		# when the node is an external node #
	(defined $node2children->{$node_id}) or ( ${$tree} =  "($separator" . ${$tree} . "$separator)" ) ;

	print $fh ${$tree} . "\;\n" ;

	return 1;

}		# end of the subroutine 'br_output_tree_sep' #


# br_draw_subtree_sep:
#	A revised version of 'br_draw_subtree' with a freedom of choice
#	of the 'sepatator' that delimits the nodes in the output tree.
#
# br_draw_subtree:
#	This subroutine outputs a part of a tree under a specified node
#	in a string of the Newick format.
#	It outputs the daughter nodes in the order
#	stored in the array @{$node2children->{$node_id}}.
#
# INPUT: ($node_id, $node2children, $id2name, $branch_attributes,
#	$option_bootlabel, $option_nodeid, $separator),
#	where $node_id is the ID of the node which is the top of
#	the part of the tree being output,
#	$option_bootlabel = 1 if you display bootstrap value at the branch,
#			=0 (or undef) at the node. 
#	$option_nodeid = 1 if you display node ID, =0 (or undef) otherwise.
#	$separator specifies the separator that delimits the nodes in the output tree.
#
sub br_draw_subtree_sep ($\%\%\%$$$) {

	my ($node_id, $node2children, $id2name, $branch_attributes,
		$option_bootlabel, $option_nodeid, $separator) = @_;

	my $tree = '';

	my $children = $node2children->{$node_id};
	if (defined $children) {

	    $tree .= "($separator";

	    my $ct_children = @{$children};
	    for (my $i=0; $i <$ct_children; $i++) {

		$i and ($tree .= ',' . $separator) ; 

		my $child = $children->[$i];
		my $subtree =
		br_draw_subtree_sep ($child, %{$node2children}, %{$id2name},
		   %{$branch_attributes}, $option_bootlabel, $option_nodeid, $separator);
		$tree .=  ${$subtree};
	    }

	    $tree .= "$separator)" ;

		# a tentative revision on Nov 7 (Sun), 2004 #
	     $option_nodeid and ($tree .= "{$node_id}");
	     

	} else { # the $node_id is probably an external node #

	    my $nodename = $id2name->{$node_id};
	    unless (defined $nodename) {
		print STDERR "[WANING 'br_draw_subtree_sep'] NO NAME for the external node ID $node_id.\n";
		$nodename = $node_id;
	    }
	    $tree .= ($option_nodeid ? ("$nodename" . "{$node_id}") : $nodename ) ;
	}

	my $upper_branch = $branch_attributes->{$node_id};

	if (defined $upper_branch) {
	    my ($br_len, $boot) = @{$upper_branch}[$INDX_LENGTH, $INDX_BOOT];

			## REVISED  on Thursday, Sep 21, 2006 ##
	    if ((defined $br_len) or (defined $boot)) {
		(defined $br_len) or ($br_len = '');
		if ($option_bootlabel) {
		    $tree .= (defined $boot) ? (":$br_len" . "[$boot]") : ":$br_len";
		} else {
		    $tree .= (defined $boot) ? "$boot:$br_len" : ":$br_len";
		}
	    }
	}

	return \$tree;

}		# end of the subroutine 'br_draw_subtree_sep' #


sub topo_subtree ($\%\%\%);

# topo_subtree: This subroutine returns the reference to  a string
#		which shows a simple tree in which each sequence name
#		is replaced by a  simple code indicating its source species.
#		It outputs the daughter nodes in the ascinding order of the sizes of subtrees under them.
#		So the tree gets more and more complex as it moves from left to right.
#
# INPUT: ($node_id, \%node2children, \%id2name, \%name2code),
#
sub topo_subtree ($\%\%\%) {

	my ($node_id, $node2children, $id2name, $name2code) = @_;

	my $topo = '';

	my $children = $node2children->{$node_id};
	if (defined $children) {

	    $topo .= "(";

	    my @children_info = (); # element = [node_ct, \$subtopo];
	    foreach my $child (@{$children}) {
		my $nodes = fetch_all_offsprings ($child, %{$node2children});
		my $node_ct = (defined $nodes) ? @{$nodes} : 0;
		my $subtopo =
		topo_subtree ($child, %{$node2children}, %{$id2name}, %{$name2code});
			push @children_info , [$node_ct, $subtopo];
	    }

	    my @std_children = sort { $a->[0] <=> $b->[0] } @children_info;
	    my $ct_children = @std_children;
	    for (my $i=0; $i <$ct_children; $i++) {
		$i and ($topo .= ',') ; 
		$topo .=  ${$std_children[$i]->[1]};
	    }

	    $topo .= ")" ;

	} else { # the $node_id is probably an external node #

	    my $nodename = $id2name->{$node_id};
	    unless (defined $nodename) {
		print STDERR "NO NAME for the external node ID $node_id.\n";
		$nodename = $node_id;
	    }
	    my $code  = $name2code->{$nodename};
	    unless (defined $code) {
		print STDERR "NO CODE for the external node ID $node_id.\n";
		$code = $node_id;
	    }
	    $topo .= $code;
	}

	return \$topo;

}		# end of the subroutine topo_subtree #

=cut


1;		# a "spell" which makes the module "use"-able and/or "require"-able #

