package MyTreeMap_MLf;
#
###################################################################################
#	FA_LOLIPOG_P: A package to calculate First-Approximate LOg-LIkelihood for the Pattern Of Gaps,
#		as well as the first-approximate perfect score (described below),
#		of a given MSA (almost exclusively via Perl modules and scripts).
#
#	Version 0.3: Copyright (C) 2013 Kiyoshi Ezawa
#
#    This package is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This package is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License, "GNU_GPL.txt",
#    along with this package.  If not, see <http://www.gnu.org/licenses/>.
#
#    The author can be contacted by e-mailing to <kezawa dot ezawa3 at gmail dot com>
#    (replace " dot " and " at " with "." and "@", respectively).
###################################################################################
#
# MyTreeMap_MLf.pm:
#	This revised version of the package 'MyTreeMap_ML.pm'
#	contains subroutines that perform FAST algorithms 
#	to calculate ML values.
#
# MyTreeMap_ML.pm:
#	This package contains subroutines that perform algorithms
#	to map several ML- (or Bayesian-) related values
#	onto each node of a phylogenetic tree.
#	The package was made by using some frameworks of 'MyTreeMap_II.pm',
#	by taking advantage of the similarity between Sankoff's parsimony algorithm
#	and Felsenstein's peeling algorithm.
#	(CREATED on June 25 (Fri), 2010)
#
# MyTreeMap_II.pm:
#	This package performs more refined algorithms than 'MyTreeMap.pm',
#	to map various properties, or their changes,
#	onto the nodes or the branches of a phylogenetic tree.
#	(CREATED on June 16 (Wed), 2010)
#
# MyTreeMap.pm:
#	This package maps various properties, or their changes,
#	 onto the nodes or the branches
#	of a (molecular) phylogenetic tree.
#	It mainly uses the parsimony principle,
#	but sometimes uses other methods, such as the maximum likelihood.

use strict;
use warnings;
use MyArrayManipu qw( diff_array union_array nr_array);
use MyArrayManipu_num qw( diff_nmarray );

use MyTaxonomy qw(
	fetch_ancestry 
	fetch_last_common_ancestor_ov2
	fetch_common_ancestors
		);
	## fetch_ext_offsprings fetch_all_offsprings
	## fetch_ext_offsprings_br

use MyTreeManipu_II qw(
		);
	## input_tree_file read_tree_file br_output_tree
	##make_node2children
	## input_subtree copy_tree
	## output_tree draw_subtree topo_subtree
	## shift_topnode move_topnode segregate_subset_nodes
	## sum_branch_lengths intra_lineage_distance inter_node_distance
	## fetch_boot_series extract_hidden_bootstrap
	## resolve_multipartite

use MyTreeManipu_III qw(
	make_node2abs_depth
	);

use Exporter;

our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
	map_node2sb_charlikeli_extJC_bo_f
	finish_node2charlikeli_extJC_td_f
	map_likelihoods_extJC_clm_f
	map_likelihoods_extJC_msa_f msa_likelihood_extJC_f
	nonred_msa_clm_patterns_extJC
	map_likelihoods_extJC_msa_fnr msa_likelihood_extJC_fnr
	);

sub log_sum_exp (\@);

our $TOP_NODE = $MyTreeManipu_II::TOP_NODE;
our $INDX_LENGTH = $MyTreeManipu_II::INDX_LENGTH;
our $INDX_BOOT = $MyTreeManipu_II::INDX_BOOT;
##our $INDX_LENGTH = 0;
##our $INDX_BOOT = 1;
##our $START_GROUP = '(';	# the symbol marking the start of the grouping of nodes.
##our $END_GROUP = ')';  # the symbol marking the end of the grouping of nodes.

	## These two indices specify where the information is stored
	## in %nodee2pars_stats = ($node_id => \@stat_info, ...) #
our $INDX_STAT = 0; # The state (or the character) mapped onto the node.
our $INDX_SBTOT_PEN = 1; # The subtotal of penalties on the subtree under $node_id. 

	### 5 important hash tables that are used in this Module
	### in order to describe a tree
#
# %node2parent = (node_ID => parent_ID, ...)  # top_node_ID => undef
# %node2children = (node_ID => [a list of child ID's], ...) # external_node_ID => undef
# %name2id = (node_name => node_ID, ...) # including $TOP_NODE => top_node_ID
# %id2name = (node_ID => node_name, ...)  # NOT including top_node_ID => $TOP_NODE,
#					 # because we can always identify the top-node by
#					# checking if $node2parent{$node_id} is defined or not.
# %branch_attributes = (ID of the nearest downstream node => [length, bootstrap, ...], ...)
#	--- swapped when the top node is shifted.


	### DEFINITIONS of the SUBROUTINES ###

# log_sum_exp:
#	This subroutine calculates the logarithm
#	of the summation of the exponentials of the input numbers,
#	in a manner that does not ruin the benefit of the logarithmic representation.
#	The subroutine uses the identity:
#
#	log [Sum_{$i = 0, ..., $n} exp ($numbers[$i]) ]
#	= $n_max + log [Sum_{$i = 0, ..., $n} exp ($numbers[$i] - $n_max) ],
#	where $n is the count of input numbers, and $n_max = Max (@numbers).
#
# INPUT: \@numbers, which lists the input numbers.
#
# OUTPUT: $log_sum_exp, which equals
#	log [Sum_{$i = 0, ..., $n} exp ($numbers[$i]) ].
#
sub log_sum_exp (\@) {

	my ($numbers) = @_;

	(@{$numbers}>0) or (return undef);

	my $n_max ;
	foreach my $x (@{$numbers}) { # 1st foreach-loop # 
	    ((defined $n_max) and  ($n_max >= $x)) or ($n_max = $x);
	}			# End of the 1st foreach-loop #

	my $sum_small_exp = 0;
	foreach my $x (@{$numbers}) { # 2nd foreach-loop #
		$sum_small_exp += exp ($x - $n_max);
	}			# End of the 2nd foreach-loop #

	my $log_sum_exp = $n_max + log ($sum_small_exp) ;

	return $log_sum_exp;

}	# End of the subroutine 'log_sum_exp' #


# map_node2sb_charlikeli_extJC_bo_f:
#	This is a FAST version of 'map_node2sb_charlikeli_extJC_bo'.
#	This performs a FAST variant of Felsenstein's peeling (pruning) algorithm
#	to calculate the likelihoods in a bottom-up manner.
#	(CREATED on Sept 18 (Sat), 2010)
#
#	THEORY:
#	This takes advantage of the following properties on the extended Jukes-Cantor model:
#	(1) The likelihood for the characters ABSENT from the MSA column are INdependent of the particular type of the character.
#	(2) We can re-write the summation over the character {c'} as follows:
#	Sum_{c'} P (ch(m) = c' | ch(n) = c) * P ({ch(l_i) = c_i}_li in desc(m) | ch(m) = c', extJC)
#	= exp (-$t) * P ({ch(l_i} = c_i}_li in desc(m) | ch(m) = c, extJC) +
#	 Pch(t) * [ Sum_{c'} P ({ch(l_i) = c_i}_li in desc(m) | ch(m) = c', extJC) ],
#	with Pch(t) = 1/$c * (1 - exp (-$t)).
#	So, the summation in the square brackets can be calculated just once
#	and reused for any character c in {c}. [--- Actually, We did NOT exploit this property...]
#	(3) Because of the symmetry, The summation in the [] in (2) can be further reduced as:
#	Sum_{c':pres in clm} P({ch(l_i) = c_i}_li in desc(m) | ch(m) = c', extJC)
#	+ #{c':abs fm clm} * P({ch(l_i) = c_i)_li in desc(m) | ch(m) = c^A, extJC),
#	where c^A is one of the characters absent from the column.
#
# map_node2sb_charlikeli_extJC_bo:
#	This satellite subroutine of 'map_likelihoods_extJC_clm'
#	assign, to each node in a bottom-up manner,
#	the log likelihoods that its exterior offspring nodes
#	have the observed characters given particular characters on the subject node,
#	using an extended JC model of substitution.
#
#	This subroutime was made by using the framework of 'min_penalties_sankoff_bo'
#	in the module 'MyTreeMap_II.pm'.
#	(CREATED on June 25 (Fri), 2010)
#
#	MODEL: The extended JC model is defined with
#	the transition probability from the character $c1 to $c2 along a branch of length $dist:
#	
#	Prob ($c1, $c2 | $dist, $ct_char)
#		= exp (-$t) + 1/$c * (1 - exp (-$t))) if $c1 == $c2,
#		= 1/$c * (1 - exp (-$t)) if $c1 != $c2,
#
#	where $c is the number of mutually transferable characters,
#	and $t = $c/($c - 1) * $dist.
#
#	NOTE that the subroutine regards negative branch length as zero, while cautioning.
#
# [INPUT: (\%node2children, \%node2depth, \%exnode2charindx, $ct_chars, \%br2attr, $indx_len), where] --- for the precesessor
# INPUT: (\%node2children, \%node2depth, \%exnode2charindx, $ct_chars, $ct_pres, \%br2attr, $indx_len), where
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2depth = ($node_id => $depth, ...) for ALL nodes, with
#	$depth defined as the number of nodes from the top node (inclusive) to themselves (inclusive);
#	%exnode2charindx = ($node_id => $indx of the character at the node, ...)
#	for the exterior nodes. The $indx specifies a character in the character-list
#	that is held by the caller of this subroutine;
#	$ct_chars is the number of possible mutually transferable characters;
#	%br2attr = ($node_id => $attr, ...) for the nodes immediately downstream of the branches,
#	with $attr->[$indx_len], or $attr if $indx_len is undefined, giving the branch length.
#
# OUTPUT: \%node2sb_charlikeli, where
#	%node2sb_charlikeli = ($node_id => \@sb_charlikeli, ...) for ALL nodes,
#	with $sb_charlikeli[$i] being the likelihood that the exterior offsprings of $node_id
#	have the characters observed under the condition that $node_id has the $i th character.
#	$sb_charlikeli[$i] is undef if the likelihood in question is zero.
#
##sub map_node2sb_charlikeli_extJC_bo (\%\%\%$\%$) {
##	my ($node2ch, $node2depth, $exnode2charindx, $ct_chars, $br2attr, $indx_len) = @_;

sub map_node2sb_charlikeli_extJC_bo_f (\%\%\%$$\%$) {

	my ($node2ch, $node2depth, $exnode2charindx, $ct_chars, $ct_pres, $br2attr, $indx_len) = @_;
		## The above two lines were REVISED in creation of 'map_node2sb_charlikeli_extJC_bo_f'
		## (on September 18 (Sat), 2010)


		## Auxiliary scalars (ADDED in creation of 'map_node2sb_charlikeli_extJC_bo_f' (on September 18 (Sat), 2010)

		# whether you can reduce the summation (=1) or not (=0).
	my $fl_red =  ($ct_pres + 1 < $ct_chars) ? 1 : 0;
		# The upper-bound of the 'reduced' for-loop to calculate character-wise log-likilihoods ##
	my $ub_redloop = ($ct_pres + 1 < $ct_chars) ? $ct_pres + 1 : $ct_chars;
		# #{characters that are ABSENT from the column}
	my $ct_abs = $ct_chars - $ct_pres;
	my $ln_ct_abs = ($ct_abs>0) ? log ($ct_abs) : undef;
	my $ln_ct_abs_m1 = ($ct_abs>1) ? log ($ct_abs -1) : undef;

	my %node2sb_charlikeli;

	my @nodes = sort {$node2depth->{$b} <=> $node2depth->{$a} } keys %{$node2depth};
	foreach my $node (@nodes) { # Main outer foreach-loop #

	    my $charindx = $exnode2charindx->{$node};
	    if (defined $charindx) { # An exterior node #
		my @sb_charlikeli = ();
		$sb_charlikeli[$charindx] = 0;
		$node2sb_charlikeli{$node} = \@sb_charlikeli;
		next;
	    }

	    my $children = $node2ch->{$node};
	    unless (defined $children) {
		print STDERR "[ERROR 'map_node2sb_charlikeli_bo'] The 'interior' node '$node'  has NO CHILDREN: $! !\n";
		return undef;
	    }

	    my @sb_charlikeli = (); # element = character-wise log-likelihood


		## REVISED in creation of 'map_node2sb_charlikeli_extJC_bo_f' (on September 18 (Sat), 2010) ##
	    for (my $i=0; $i<$ub_redloop; $i++) { # Mid-outer for-loop (over the characters PRESENT in the column (+1) on the node) #
	    ##for (my $i=0; $i<$ct_chars; $i++) { # Mid-outer for-loop (over the possible characters on the node) #
		my $sb_charlikeli = 0;

		foreach my $ch (@{$children}) { # Mid-inner for-loop (over the children) #
		    my $sb_chlk_ch = $node2sb_charlikeli{$ch};
		    unless (defined $sb_chlk_ch) {
			print STDERR "[ERROR 'map_node2sb_charlikeli_bo'] The child node '$ch' of the node '$node' does NOT HAVE CHARACTER-WISE LIKELIHOODs of its sub-tree: $! !\n";
			return undef; 
		    }

		    my $attr_ch = $br2attr->{$ch};
		    unless (defined $attr_ch) {
			print STDERR "[ERROR 'map_node2sb_charlikeli_bo'] NO ATTRIBUTE of the branch between the node '$node' and its child '$ch': $! !\n";
			return undef;
		    }
		    my $dist = (defined $indx_len) ? $attr_ch->[$indx_len] : $attr_ch;
		    if ($dist<0) {
			print STDERR "[CAUTION 'map_node2sb_charlikeli_bo'] NEGATIVE BRANCH LENGTH ($dist) between the node '$node' and its child '$ch', ... I will regard it as ZERO ...\n\n";
			$dist = 0;
		    }
		    unless ($dist > 0) {
			my $sbsb_charlikeli = $sb_chlk_ch->[$i];
			if (defined $sbsb_charlikeli) {
			   $sb_charlikeli += $sbsb_charlikeli;
			   next;
			} else {
			   $sb_charlikeli = undef;
			   last;
			}
		    }

##=The entire block REVISED in creation of 'map_node2sb_charlikeli_extJC_bo_f' (on September 18 (Sat), 2010) ##

		    my $t = $ct_chars * $dist /($ct_chars - 1);
		    my $prob_chg = (1 - exp (- $t))/$ct_chars;
		    my $ln_prob_chg = log ($prob_chg);
		    my $prob_nc = exp (- $t) + $prob_chg;
		    my $ln_prob_nc = log ($prob_nc);

		    my @terms_sbsb_chlk = ();

		### RESTART FROM HERE ###

			# Contribution from the subject character itself (ADDED in creation of 'map_node2sb_charlikeli_extJC_bo_f') #
		    my $self_term = $sb_chlk_ch->[$i];
		    (defined $self_term) and (push @terms_sbsb_chlk, $self_term + $ln_prob_nc);
		    ##(defined $self_term) and (push @terms_sbsb_chlk, $self_term - $t);

			## Contributions from the characters ACTUALLY PRESENT in the column (except the subject character itself) ##
		    for (my $j=0; $j<$ct_pres; $j++) { # Inner for-loop (over the characters PRESENT in the column on the child) #
		    ##for (my $j=0; $j<$ct_chars; $j++) { # Inner for-loop (over the possible characters on the child) #
			($j == $i) and next; ## ADDED in creation of 'map_node2sb_charlikeli_extJC_bo_f'
			my $term = $sb_chlk_ch->[$j];
			(defined $term) or next; # undefined means zero contribution #

			$term +=  $ln_prob_chg;
			##$term +=  ($i == $j) ? $ln_prob_nc : $ln_prob_chg;
			push @terms_sbsb_chlk, $term;
		    }			# End of the inner for-loop (over the possible characters on the child) #

			## Contributions from the characters ABSENT from the column (except the subject character itself) (ADDED in creation of 'map_node2sb_charlikeli_extJC_bo_f') ##
		    if (($ct_abs>1) or (($i< $ct_pres) and ($ct_abs>0)) ) {
			my $term_abs = $sb_chlk_ch->[$ct_pres];
			if (defined $term_abs) {
			   $term_abs += ($i < $ct_pres) ? $ln_ct_abs : $ln_ct_abs_m1;
			   push @terms_sbsb_chlk, $term_abs + $ln_prob_chg;
			}
		    }

##=END of "The entire block REVISED in creation of 'map_node2sb_charlikeli_extJC_bo_f' (on September 18 (Sat), 2010)" ##

		   if (@terms_sbsb_chlk == 1) {
			$sb_charlikeli += $terms_sbsb_chlk[0];

		   } elsif (@terms_sbsb_chlk>1) {
			my $sbsb_charlikeli = log_sum_exp (@terms_sbsb_chlk);
			$sb_charlikeli += $sbsb_charlikeli;

		    } else {
			# This means all terms to be summed are zero,
			# therefore the total character-wise likelihood for this parent-child pair
			# (on this character) is also zero.
			# That in turn makes the character-wise likelihood zero for this character on this sub-tree.
			$sb_charlikeli = undef;
			last;
		    }

		}		# End of the mid-inner for-loop (over the children) #

		$sb_charlikeli[$i] = $sb_charlikeli;

	    }	# End of the mid-outer for-loop (over the characters PRESENT in the column (+1) on the node) #

		## ADDED in revision to 'map_node2sb_charlikeli_extJC_bo_f' (on September 18 (Sat), 2010) ##
	    if ($fl_red) {
			## Pad the elements of @sb_charlikeli for the ABSENT characters with the 1st absent one ##
		my $sb_chlk_abs = $sb_charlikeli[$ct_pres];
		for (my $i = $ub_redloop; $i<$ct_chars; $i++) { $sb_charlikeli[$i] = $sb_chlk_abs; }
	    }

	    $node2sb_charlikeli{$node} = \@sb_charlikeli;

	}			# End of the main outer foreach-loop #

	return \%node2sb_charlikeli;

}		# End of the subroutine 'map_node2sb_charlikeli_extJC_bo_f' #


# finish_node2charlikeli_extJC_td_f:
#	This is a FAST version of 'finish_node2charlikeli_extJC_td'.
#	The theoretical grounds are the same as those used for
#	'map_node2sb_charlikeli_extJC_bo_f'.
#	Simply put, it takes advantage of the symmetry of character-wise
#	likelihoods among characters ABSENT from the MSA column.
#	(CREATED on September 18 (Sat), 2010)
#
# finish_node2charlikeli_extJC_td:
#	This satellite subroutine of 'map_likelihoods_extJC_clm'
#	finishes calculating the character-wise likelihoods on all nodes
#	in a top-down manner, using an extended JC model of substitution.
#	Here a 'character-wise likelihood' on a node means the likelihood
#	that we observe the characters of all the exterior nodes
#	given that the node has a particular character.
#
#	This subroutine was made by using the framework of
#	'enum_node2parsstat_sankoff_td' in the module 'MyTreeMap_II.pm'.
#	(CREATED on June 25 (Fri), 2010)
#
#	ALGORITHM:
#	A character-wise likelihood on eac node
#	is completed by adding the parent contribution to
#	the partial character-wise likelihood that was already calculated
#	via 'map_node2sb_charlikeli_exJC_bo'.
#	The parent contribution is made by stripping the contribution by
#	the present node off the complete character-wise likelihood for the parent node.
#
#	MODEL: The extended JC model is defined with
#	the transition probability from the character $c1 to $c2 along a branch of length $dist:
#	
#	Prob ($c1, $c2 | $dist, $ct_char)
#		= exp (-$t) + 1/$c * (1 - exp (-$t))) if $c1 == $c2,
#		= 1/$c * (1 - exp (-$t)) if $c1 != $c2,
#
#	where $c is the number of mutually transferable characters,
#	and $t = $c/($c - 1) * $dist.
#
#	NOTE that the subroutine regards negative branch length as zero, while cautioning.
#
# [INPUT: (\%node2pa, \%node2depth, \%node2sb_charlikeli, $ct_chars, \%br2attr, $indx_len, $fl_full_mode), where] -- for the predecessor
# INPUT: (\%node2pa, \%node2depth, \%node2sb_charlikeli, $ct_chars, $ct_pres, \%br2attr, $indx_len, $fl_full_mode), where
#	%node2parent = ($node_id => $parent_id,...) for the non-top nodes;
#	%node2depth = ($node_id => $depth, ...) for ALL nodes, with
#	$depth defined as the number of nodes from the top node (inclusive) to themselves (inclusive);
#	%node2sb_charlikeli = ($node_id => \@sb_charlikeli, ...), with
#	$sb_charlikeli[$i] being the likelihood that 
#	the exterior offspring nodes of $node_id have the observed characters
#	under the condition that $node_id has the $i th character;
#	$ct_chars is the number of possible mutually transferable characters;
#	%br2attr = ($node_id => $attr, ...) for nodes immediately downstream of the branches,
#	with $attr->[$indx_len], or $attr if $indx_len is undefined, being the branch length;
#	$fl_full_mode = 1 if you want the full mode,
#			= 0 if the brief mode is enough.
#
# OUTPUT: \%node2charlikeli, where
#	%node2charlikeli = ($node_id => \@charlikeli, ...),
#	with $charlikeli[$i] being the likelihood that
#	all the exterior nodes have the observed characters
#	under the condition that $node_id has the $i th character.
#
#	If $fl_full_mode = 1, the information on ALL the nodes is given.
#	If $fl_full_mode = 0, ONLY the information on the TOP node is given.
#
##sub finish_node2charlikeli_extJC_td (\%\%\%$\%$$) {
##	my ($node2pa, $node2depth, $node2sb_charlikeli, $ct_chars, $br2attr, $indx_len, $fl_full_mode) = @_;
sub finish_node2charlikeli_extJC_td_f (\%\%\%$$\%$$) {

	my ($node2pa, $node2depth, $node2sb_charlikeli, $ct_chars, $ct_pres, $br2attr, $indx_len, $fl_full_mode) = @_;
		## the above two lines were REVISED in creation of 'finish_node2charlikeli_extJC_td_f' (on September 18 (Sat), 2010)

		## Auxiliary scalars (ADDED in creation of 'finish_node2charlikeli_extJC_td_f' (on September 18 (Sat), 2010)

		# whether you can reduce the summation (=1) or not (=0).
	my $fl_red =  ($ct_pres + 1 < $ct_chars) ? 1 : 0;
		# The upper-bound of the 'reduced' for-loop to calculate character-wise log-likilihoods ##
	my $ub_redloop = ($ct_pres + 1 < $ct_chars) ? $ct_pres + 1 : $ct_chars;
		# #{characters that are ABSENT from the column}
	my $ct_abs = $ct_chars - $ct_pres;
	my $ln_ct_abs = ($ct_abs>0) ? log ($ct_abs) : undef;
	my $ln_ct_abs_m1 = ($ct_abs > 1) ? log ($ct_abs -1) : undef;


	my @nodes = sort {$node2depth->{$a} <=> $node2depth->{$b}} keys %{$node2depth};

	my %node2charlikeli;

	foreach my $node (@nodes) { # The main outermost foreach-loop (over all nodes) #

	    my $sb_charlikeli = $node2sb_charlikeli->{$node};
	    unless (defined $sb_charlikeli) {
		print STDERR "[ERROR 'finish_node2charlikeli_extJC_td_f'] The node '$node' has NO PARTIAL CHARACTERWISE LOG-LIKELIHOOD: $! !\n";
		return undef;
	    }

	    my $parent = $node2pa->{$node};
	    unless (defined $parent) { # The Big unless #
		# The top node --> Initialize the %node2charlikeli #

		my @charlikeli = @{$sb_charlikeli};
		$node2charlikeli{$node} = \@charlikeli;

		if ($fl_full_mode == 1) {
		    next;
		} else {
		    return \%node2charlikeli; # Brief version #
		}
	    } 			# End of the Big unless #

		# Prepare the transition probabilities between $node and $parent

	    my $attr = $br2attr->{$node};
	    unless (defined $attr) {
		print STDERR "[ERROR 'finish_node2charlikeli_extJC_td_f'] NO ATTRIBUTE for the branch between the node '$node' and its parent '$parent': $! !\n";
		return undef;
	    }
	    my $dist = (defined $indx_len) ? $attr->[$indx_len] : $attr;
	    if ($dist < 0) {
		print STDERR "[CAUTION 'finish_node2charlikeli_extJC_td_f'] NEGATIVE LENGTH ($dist) for the branch between the node '$node' and its parent '$parent' ! ... I will regard it as ZERO ...\n\n";
		$dist = 0;
	    }
	    my $t = $ct_chars * $dist /($ct_chars - 1);
	    my $prob_chg = (1 - exp (- $t))/$ct_chars;
	    my $prob_nc = exp (- $t) + $prob_chg;
	    my $ln_prob_chg = ($prob_chg>0) ? log ($prob_chg) : undef;
	    my $ln_prob_nc = log ($prob_nc);

		## Retrieve the complete character-wise log-likelihoods
		## for the parent node

	    my $charlikeli_pa = $node2charlikeli{$parent};
	    unless (defined $charlikeli_pa) {
		print STDERR "[ERROR 'finish_node2charlikeli_extJC_td_f'] The parent node '$parent' of the node '$node' has NO CHARACTER-WISE log-LIKELIHOODS: $! !\n";
		return undef;
	    }

	    my @stripped_off_pa = (); # parent character-wise log-likelihood with the contribution by the current node stripped off

		## REVISED in creation of 'finish_node2charlikeli_extJC_td_f' (on September 18 (Sat), 2010) ##
	    for (my $j=0; $j<$ub_redloop; $j++) { # The 1st middle for-loop (over parent characters ACTUALLY PRESENT in the column (+1))  #
	    ##for (my $j=0; $j<$ct_chars; $j++) { # The 1st middle for-loop (over parent characters) #

		my $ln_chlk_pa = $charlikeli_pa->[$j];
		unless (defined $ln_chlk_pa) { # ZERO parental likelihood #
		    $stripped_off_pa[$j] = undef;
		    next;
		}

		if ($dist ==0) {
			# zero branch length between $node and $parent #
			# (in this situation, $sb_charlikeli->[$j] should always be defined) #
		    $stripped_off_pa[$j] = $ln_chlk_pa - $sb_charlikeli->[$j];
		    next;
		}

		my @terms_sbtract = (); # Terms to be subtracted from the parental character-wise log likelihood #

		## ADDED in creation of 'finish_node2charlikeli_extJC_td_f' (on September 18 (Sat), 2010) ##

			# Contribution from the subject character (ADDED in creation of 'finish_node2charlikeli_extJC_td_f') #
		my $self = $sb_charlikeli->[$j];
		(defined $self) and (push @terms_sbtract, $self + $ln_prob_nc);
		
		## REVISED in creation of 'finish_node2charlikeli_extJC_td_f' (on September 18 (Sat), 2010) ##
		for (my $i = 0; $i<$ct_pres; $i++) { # The inner for-loop (over subject characters PRESENT in the column) #
		##for (my $i = 0; $i<$ct_chars; $i++) { # The inner for-loop (over subject characters) #
		    ($i == $j) and next; ## ADDED in creation of 'finish_node2charlikeli_extJC_td_f'
		    my $sb_ln_chlk = $sb_charlikeli->[$i];
		    (defined $sb_ln_chlk) or next;
		    push @terms_sbtract, $sb_ln_chlk + $ln_prob_chg;
		    ##my $tr_prob = ($i == $j) ? $ln_prob_nc : $ln_prob_chg;
		    ##push @terms_sbtract, $sb_ln_chlk + $tr_prob;
		}	# End of the inner for-loop (over subject characters PRESENT in the column) #

			# Contributions from the characcters ABSENT from the column (except the subject character itself) (ADDED in creation of 'finish_node2charlikeli_extJC_td_f') #
		if (($ct_abs>1) or (($j < $ct_pres) and ($ct_abs>0)) ) {
		   my $sb_ln_chlk_abs = $sb_charlikeli->[$ct_pres];
		   if (defined $sb_ln_chlk_abs) {
			$sb_ln_chlk_abs += ($j < $ct_pres) ? $ln_ct_abs : $ln_ct_abs_m1;
			push @terms_sbtract, $sb_ln_chlk_abs + $ln_prob_chg;
		   }
		}


		if (@terms_sbtract==1) {
		    $stripped_off_pa[$j] = $ln_chlk_pa - $terms_sbtract[0];
		} elsif (@terms_sbtract>1) {
		    my $sbtract = log_sum_exp (@terms_sbtract);
		    $stripped_off_pa[$j] = $ln_chlk_pa - $sbtract;
		} else {
		    print STDERR "[ERROR 'finish_node2charlikeli_extJC_td_f'] This is IMPOSSIBLE! NO TERMS  to be STRIPPED OFF the NON-ZERO total log likelihood ($ln_chlk_pa) of the $j th character for the parent '$parent' of the node '$node': $! !\n";
		    return undef;
		}


	    }	   # End of the 1st middle for-loop (over parent characters ACTUALLY PRESENT in the column (+1)) #

	    my @charlikeli = (); # Complete character-wise log-likelihoods for the subject node #

		## REVISED in creation of 'finish_node2charlikeli_extJC_td_f' (on September 18 (Sat), 2010) ##
	    for (my $j=0; $j<$ub_redloop; $j++) { # The 2nd middle for-loop (over subject characters ACTUALLY PRESENT in the column (+1)) #
	    ## for (my $j=0; $j<$ct_chars; $j++) { # The 2nd middle for-loop (over subject characters) #
		my $ln_sb_chlk = $sb_charlikeli->[$j];
		(defined $ln_sb_chlk) or next;

		if ($dist==0) {
			# ZERO branch length between $node and $parent #
			# Stripped-off parental likelihood can be added as it is.

		    my $strpd_off_pa = $stripped_off_pa[$j];

		    (defined $strpd_off_pa) or next; # The character-wise likelihood should be zero in this case.

		    $charlikeli[$j] = $ln_sb_chlk + $strpd_off_pa;
		    next;
		}

		my @terms_contri_pa = (); # Terms for the parental contribution #
			# Contribution from the subject character itself (ADDED in creation of 'finish_node2charlikeli_extJC_td_f') #
		my $self_strpdoff_pa = $stripped_off_pa[$j];
		(defined $self_strpdoff_pa) and (push @terms_contri_pa, $self_strpdoff_pa + $ln_prob_nc);

			## REVISED in creation of 'finish_node2charlikeli_extJC_td_f' (on September 18 (Sat), 2010) ##
		for (my $i=0; $i<$ct_pres; $i++) { # The inner for-loop (over parent characters ACTUALLY PRESENT in the column) #
		##for (my $i=0; $i<$ct_chars; $i++) { # The inner for-loop (over parent characters) #
		    ($i == $j) and next; ## ADDED in creation of 'finish_node2charlikeli_extJC_td_f'
		    my $strpd_off_pa = $stripped_off_pa[$i];
		    (defined $strpd_off_pa) or next;
		    push @terms_contri_pa, $strpd_off_pa + $ln_prob_chg;
		    ##my $tr_prob = ($i == $j) ? $ln_prob_nc : $ln_prob_chg;
		    ##push @terms_contri_pa, $strpd_off_pa + $tr_prob;
		}	# End of the inner for-loop (over parent characters ACTUALLY PRESENT in the column) #

			# Contributions from the characcters ABSENT from the column (except the subject character itself) (ADDED in creation of 'finish_node2charlikeli_extJC_td_f') #
		if (($ct_abs>1) or (($j < $ct_pres) and ($ct_abs>0)) ) {
		   my $strpd_off_pa_abs = $stripped_off_pa[$ct_pres];
		   if (defined $strpd_off_pa_abs) {
			$strpd_off_pa_abs += ($j < $ct_pres) ? $ln_ct_abs : $ln_ct_abs_m1;
			push @terms_contri_pa, $strpd_off_pa_abs + $ln_prob_chg;
		   }
		}


		if (@terms_contri_pa==1) {
		    $charlikeli[$j] = $ln_sb_chlk + $terms_contri_pa[0];
		} elsif (@terms_contri_pa>1) {
		    my $contri_pa = log_sum_exp (@terms_contri_pa);
		    $charlikeli[$j] = $ln_sb_chlk + $contri_pa;
		} else {
			## In this case, the total parental contribution is zero.
			## Therefore, the complete likelihood is also zero for this character on this node. #
		    $charlikeli[$j] = undef;
		}

	    }	    # End of the 2nd middle for-loop (over subject characters ACTUALLY PRESENT in the column (+1)) #

		## ADDED in creation of 'finish_node2charlikeli_extJC_td_f' (on September 18 (Sat), 2010) ##
	    if ($fl_red) {
		# The elements of @charlikeli for the remaining ABSENT characters will be filled with the the element for the 1st ABSENT character.
		my $chlk_abs = $charlikeli[$ct_pres];
		for (my $j=$ub_redloop; $j<$ct_chars; $j++) { $charlikeli[$j] = $chlk_abs; }
	    }

	    $node2charlikeli{$node} = \@charlikeli;

	}	# End of the main outermost foreach-loop (over all nodes) #

	return \%node2charlikeli;

}	# End of the subroutine 'finish_node2charlikeli_extJC_td_f' #


# map_likelihoods_extJC_clm_f:
#	This is a FAST version of 'map_likelihoods_extJC_clm'.
#	This version performs a FAST variant of Felsenstein's peeling algorithm,
#	using the FAST subroutines 'map_node2sb_charlikeli_extJC_bo_f'
#	and 'finish_node2charlikeli_extJC_td_f'.
#	(CREATED on Sept 17 (Fri), 2010)
#
# map_likelihoods_extJC_clm:
#	This satellite subroutine of 'map_likelihoods_extJC_msa'
#	performs Felsenstein's peeling algorithm, with an extended Jukes-Cantor model,
#	on a column of MSA,
#	to map the character-wise log-likelihoods onto nodes.
#
#	This subroutine was made by using the framework of
#	'map_parsim_states_sankoff_clm' in the module 'MyTreeMap_II.pm'.
#
# INPUT: (\%node2children, \%node2parent, \%node2depth, \%exnode2charindx, $ct_chars,$ct_pres, \%br2attr, $indx_len, $fl_full_mode), where
# [INPUT: (\%node2children, \%node2parent, \%node2depth, \%exnode2charindx, $ct_chars, \%br2attr, $indx_len, $fl_full_mode), where] -- for the predecessor
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2parent = ($node_id => $parent_id,...) for the non-top nodes;
#	%node2depth = ($node_id => $depth, ...) for ALL nodes, with
#	$depth defined as the number of nodes from the top node (inclusive) to themselves (inclusive);
#	%exnode2charindx = ($node_id => $indx of the character at the node, ...)
#	for the exterior nodes. The $indx conforms to @pen_mtrx;
#	$ct_chars is the number of possible mutually transferable characters;
#	$ct_pres is the number of characters ACTUALLY PRESENT in the column; (ADDED on September 18 (Sat), 2010)
#	%br2attr = ($node_id => $attr, ...) for the nodes immediately downstream of the branches, with
#	$attr->[$indx_len], or $attr if $indx_len is undefined, giving the branch length;
#	$fl_full_mode = 1 if you want the full mode,
#			= 0 if the brief mode is enough.
#
# OUTPUT: \%node2charlikeli, where
#
#	%node2charlikeli = ($node_id => \@charlikeli, ...)
#	gives the mapping of the character-wise likelihoods onto the nodes
#	for the input MSA column, with
#	$charlikeli[$c] is the log-likelihood that all the exterior nodes exhibit
#	the observed character given the $c th character on the node $node_id.
#
#	If $fl_full_mode = 1, the information on ALL the nodes is given.
#	If $fl_full_mode = 0, ONLY the information on the TOP node is given.
#
##sub map_likelihoods_extJC_clm (\%\%\%\%$\%$$) {
##	## my ($node2ch, $node2pa, $node2depth, $exnode2charindx, $ct_chars, $ct_pres, $br2attr, $indx_len, $fl_full_mode) = @_;
sub map_likelihoods_extJC_clm_f (\%\%\%\%$$\%$$) {

	my ($node2ch, $node2pa, $node2depth, $exnode2charindx, $ct_chars, $ct_pres, $br2attr, $indx_len, $fl_full_mode) = @_;
		## the above two lines were REVISED in creation of 'map_likelihoods_extJC_clm_f'
		## (on September 18 (Sat), 2010) ##


		## Assign character-wise likelihoods for the sub-trees
		## in a bottom-up manner

	# %{$node2sb_charlikeli} = ($node_id => \@sb_charlikeli, ...) for all nodes, where
	# $sb_charlikeli[$c] is the log-likelihood that the exterior nodes under $node_id
	# have the observed characters given the $c th character on $node_id.
	#
		## REVISED in creation of 'map_likelihoods_extJC_clm_f' (on September 18 (Sat), 2010) ##
	my $node2sb_charlikeli = map_node2sb_charlikeli_extJC_bo_f (%{$node2ch}, %{$node2depth}, %{$exnode2charindx}, $ct_chars, $ct_pres, %{$br2attr}, $indx_len);
	##my $node2sb_charlikeli = map_node2sb_charlikeli_extJC_bo (%{$node2ch}, %{$node2depth}, %{$exnode2charindx}, $ct_chars, %{$br2attr}, $indx_len);
	unless (defined $node2sb_charlikeli) {
	    print STDERR "[ERROR 'map_likelihoods_extJC_clm_f'] The subroutine 'map_node2sb_charlikeli_bo_f' FAILED: $! !\n";
	    return undef;
	}

=for TESTING the subroutine 'map_node2sb_charlikeli_extJC_bo_f'
	print STDERR "[ Output of 'map_node2sb_charlikeli_extJC_bo_f' ]\n\n";
	print STDERR '#Node_ID', "\t", '(partial, character-wise, likelihoods)', "\n\n";
	foreach my $node (sort {$a <=> $b} keys %{$node2sb_charlikeli}) {
	    my $sb_charlikeli = $node2sb_charlikeli->{$node};
	    my @output = ();
	    for (my $i=0; $i<$ct_chars; $i++) {
		my $ln_chlk = $sb_charlikeli->[$i];
		push @output, (defined $ln_chlk) ? exp ($ln_chlk) : 0;
	    }
	    print STDERR $node, "\t", '(', join (', ', @output), ")\n";
	}
	print STDERR "\n\n";
=cut

		## Finish mapping the character-wise likelihoods
		## regarding the entire tree onto the nodes in a top-down manner

		## REVISED in creation of 'map_likelihoods_extJC_clm_f' (on September 18 (Sat), 2010) ##
	my $node2charlikeli = finish_node2charlikeli_extJC_td_f (%{$node2pa}, %{$node2depth}, %{$node2sb_charlikeli}, $ct_chars, $ct_pres, %{$br2attr}, $indx_len, $fl_full_mode);
	## my $node2charlikeli = finish_node2charlikeli_extJC_td (%{$node2pa}, %{$node2depth}, %{$node2sb_charlikeli}, $ct_chars, %{$br2attr}, $indx_len, $fl_full_mode);
	unless (defined $node2charlikeli) {
	    print STDERR "[ERROR 'map_likelihoods_extJC_clm_f'] The subroutine 'finish_node2charlikeli_extJC_td_f' FAILED: $! !\n";
	    return undef;
	}

	return $node2charlikeli;

}	# End of the subroutine 'map_likelihoods_extJC_clm_f' # 


# map_likelihoods_extJC_msa_f:
#	This is a FAST version of 'map_likelihoods_extJC_msa'.
#	The only major difference is that this version uses 'map_likelihoods_extJC_clm_f',
#	which is a FAST version of 'map_likelihoods_extJC_clm'.
#	Accompanying this, the subroutine also counts the number of characters
#	ACTUALLY PRESENT in the column, besides the the total number of characters
#	in the 'local' model (specific to the site in question).
#	(CREATED on September 18 (Sat), 2010)
#
# map_likelihoods_extJC_msa:
#	This subroutine performs Felsenstein's peeling algorithm
#	to estimate the log-likelihood values of columns in the input  MSA,
#	given a tree topology and branch lengths,
#	by using an extended Jukes-Cantor model of substitutions.
#
#	This subroutine was made by using the framework of
#	'map_parsim_states_sankoff_msa' in 'MyTreeMap_II.pm'.
#
#	In the full mode, it also calculates the character-wise likelihoods
#	on all interior node.
#	In the bried mode, it only calculates
#	the character-wise likelihoods on the top node,
#	besides the log-likelihood value
#
#	NOTE that this subroutine eats up the input MSA.
#	PREREQUISITE: The input MSA must NOT contain gaps.
#
#	Model: An extended Jukes-Cantor model is defined with
#	the transition probability from the character $c1 to $c2:
#	Prob ($c1, $c2 | $dist, $ct_char)
#		= exp (-$t) + 1/$c * (1 - exp (-$t))) if $c1 == $c2,
#		= 1/$c * (1 - exp (-$t)) if $c1 != $c2,
#	where $c is the number of mutually transferable characters,
#	and $t = $c/($c - 1) * $dist.
#
# INPUT: ($topnode_id, \%node2children, \%node2parent, \%name2id, \@seqnames, \@msa, $minct_chars, \%br2attr, $indx_len, $fl_full_mode), where
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2parent = ($node_id => $parent_id, ...) for the non-top nodes;
#	%name2id = ($seqname => $node_id, ...) for the exterior nodes and the top node;
#	$seqnames[$k] is the name of the $k th sequence in the input MSA;
#	${$msa[$k]} is the $k th sequence embedded in the input MSA;
#	$minct_chars is the assumed number of possible characters.
#	But the actual number of characters in a column can exceed this value;
#	%br2attr = ($node_id => $attr, ...) for nodes immediately downstream of the branches, with
#	$attr->[$indx_len], or $attr if $indx_len is undefined, giving the branch length;
#	$fl_full_mode = 1 if you want the full mode,
#			= 0 if the brief mode is enough.
#
# OUTPUT: (\@set_chars, \@set_node2charlikeli), where
#	@{$set_chars[$l]} is the list of characters examined in the $l th column;
#	%{$set_node2charlikeli[$l]} = ($node_id => \@charlikeli, ...)
#	gives mapping of the character-wise log-likelihood values the $l th site mapped onto the nodes,
#	with $charlikeli[$c] is the log-likelihood of the character $set_chars[$l]->[$c];
#	If $fl_full_mode = 1, the mapping onto ALL the nodes is given.
#	If $fl_full_mode = 0, ONLY the likelihoods on the TOP node is given.
#
sub map_likelihoods_extJC_msa_f ($\%\%\%\@\@$\%$$) {

	my ($topnode_id, $node2ch, $node2pa, $name2id, $seqnames, $msa,
	   $minct_chars, $br2attr, $indx_len, $fl_full_mode) = @_;

	my $ct_seqs = @{$seqnames};
	unless ($ct_seqs>0) {
	    print STDERR "[ERROR 'map_likelihoods_extJC_msa_f'] NO SEQUENCE in the input MSA: $! !\n";
	    return (undef, undef);
	}
	unless ($ct_seqs == @{$msa}) {
	    print STDERR "[ERROR 'map_likelihoods_extJC_msa_f'] The number of sequences (", scalar (@{$msa}), ") in the MSA CLASHES with the number of sequence names ($ct_seqs): $! !\n";
	    return (undef, undef);
	}
	my $ct_clms = length ${$msa->[0]};
	($ct_clms>0) or (return ([], []));

=for TESTING the subroutine
	print STDERR "\n< Subroutine 'map_likelihoods_extJC_msa_f' >\n\n";
	print STDERR "Minimum number of characters: $minct_chars\n\n";
	print STDERR "\$fl_full_mode = $fl_full_mode\n\n";
=cut
		# Prepare auxiliary arrays/hashes #

	# %{$node2depth} = ($node_id => #{nodes from the top node to itself}, ...);
	my $node2depth = make_node2abs_depth ($topnode_id, %{$node2ch});
	unless (defined $node2depth) {
	    print STDERR "[ERROR 'map_likelihoods_extJC_msa_f'] The subroutine 'make_node2abs_depth' FAILED : $! !\n";
	    return undef;
	}

	my @node_ids; # $node_ids[$k] is the exterior node ID corresponding to the $k th sequence in @{$msa}
	for (my $k=0; $k<$ct_seqs; $k++) { # The 2nd preliminary for-loop #
	    my $name = $seqnames->[$k];
	    my $node_id = $name2id->{$name};
	    unless (defined $node_id) {
		print STDERR "[ERROR 'map_likelihoods_extJC_msa_f'] NODE ID is UNDEFINED for the $k th sequence (name: $name): $! !\n";
		return undef;
	    }
	    $node_ids[$k] = $node_id;
	}			# End of the 2nd preliminary for-loop #

		## Main Process ##

	my @set_chars = ();
	my @set_node2charlikeli = ();

	for (my $l=0; $l<$ct_clms; $l++) { # The main outer for-loop #

	    my @char_in_clm = ();
	    for (my $k=0; $k<$ct_seqs; $k++) { # The 1st inner for-loop #
		my $char = substr (${$msa->[$k]}, 0, 1, '');
		push @char_in_clm, $char;
	    }			# End of the 1st inner for-loop #

	    my $charlist = nr_array (@char_in_clm);

		## REVISED in creation of 'map_likelihoods_extJC_msa_f' (on September 18 (Sat), 2010) ##
	    my $ct_chars = my $ct_pres = @{$charlist};
	    ##my $ct_chars = @{$charlist};
	    if ($ct_chars < $minct_chars) {
		# Pad the character list with numbers
		# (, which will never match the actual characters)
		for (my $n = 0; $n<$minct_chars - $ct_chars; $n++) { push @{$charlist}, $n; }
		$ct_chars = $minct_chars;
	    }
	    push @set_chars, $charlist;
	    my %char2indx; # ($character => $index in @{$charlist}, ...)
	    for (my $i = 0; $i<$ct_chars; $i++) { $char2indx{$charlist->[$i]} = $i; }

=for TESTING the subroutine
	print STDERR "\n[ The $l th site ]\n\n";
	print STDERR "Column:  ", join ('', @char_in_clm), "\n\n";
	print STDERR "Character-list:  ", join (', ' , @{$charlist}), "\n\n";
=cut

	    my %exnode2charindx; # ($node_id => $index of the character, ...) for the exterior nodes
	    for (my $k=0; $k<$ct_seqs; $k++) { # The 2nd inner for-loop #

		my $node_id = $node_ids[$k];
		my $char = $char_in_clm[$k];

		my $indx = $char2indx{$char};
		unless (defined $indx) {
		    print STDERR "[ERROR 'map_likelihood_extJC_msa_f'] INDEX is UNDEFINED for the character '$char' in the $l th site of the $k th sequence: $! !\n";
		    return (undef, undef);
		}
		$exnode2charindx{$node_id} = $indx;
	    }				# End of the 2nd inner for-loop #

		## REVISED in creation of 'map_likelihoods_extJC_msa_f' (on September 18 (Sat), 2010) ##
	    my $node2charlikeli =  map_likelihoods_extJC_clm_f (%{$node2ch}, %{$node2pa}, %{$node2depth}, %exnode2charindx, $ct_chars, $ct_pres, %{$br2attr}, $indx_len, $fl_full_mode) ;
	    ##my $node2charlikeli =  map_likelihoods_extJC_clm (%{$node2ch}, %{$node2pa}, %{$node2depth}, %exnode2charindx, $ct_chars, %{$br2attr}, $indx_len, $fl_full_mode) ;
	    unless (defined $node2charlikeli) {
		print STDERR "[ERROR 'map_likelihoods_extJC_msa_f'] 'map_likelihoods_extJC_clm_f' FAILED on the $l th site: $! !\n";
		return (undef, undef);
	    }

=for TESTING the subroutine 'map_likelihoods_extJC_clm'
	print STDERR "\n[ Outputs of 'map_likelihoods_extJC_clm_f' ]\n\n";
	print STDERR '#Node_ID', "\t", "Charwise_likelihoods (", join (', ', @{$charlist}), ")\n\n";
	foreach my $node (sort {$a <=> $b} keys %{$node2charlikeli}) {
	    my $charlikeli = $node2charlikeli->{$node};
	    my @output = ();
	    for (my $i=0; $i<$ct_chars; $i++) {
		my $ln_chlk = $charlikeli->[$i];
		push @output, (defined $ln_chlk) ? exp ($ln_chlk) : 0;
	    }
	    print STDERR $node, "\t", '(', join (', ', @output), ")\n";
	}
	print STDERR "\n\n";
=cut

	    push @set_node2charlikeli, $node2charlikeli;

	}			# End of the main outer for-loop #

=for TESTING the subroutine
	print STDERR "\n--- The subroutine 'map_likelihoods_extJC_msa_f' ENDED ! ---\n\n";
=cut

	return (\@set_chars, \@set_node2charlikeli);

}	# End of the subroutine 'map_likelihoods_extJC_msa_f' #


# msa_likelihood_extJC_f:
#	This is a FAST version of 'msa_likelihood_extJC'.
#	The only difference is that this uses 'map_likelihoods_extJC_msa_f',
#	which is a FAST version of 'map_likelihoods_extJC_msa'.
#	(CREATED on September 17 (Fri), 2010)
#
# msa_likelihood_extJC:
#	This subroutine performs Felsenstein's peeling algorithm
#	to estimate the log-likelihood values of columns in the input MSA,
#	given a tree topology and branch lengths,
#	by using an extended Jukes-Cantor model of substitutions
#	via the subroutine 'map_likelihoods_extJC_msa'.
#	Then it calculates the log-likelihoood of the entire MSA
#	by summing the log-likelihoods over columns.
#
#	NOTE that this subroutine eats up the input MSA.
#	PREREQUISITE: The input MSA must NOT contain gaps.
#
#	ALGORITHM: For each column, character-wise likelihoods:
#	Prob ({column} | {tree}, $chars[$i] at $topnode_id) 
#	are calculated by the subroutine 'map_likelihoods_extJC_msa'.
#	The total likelihood for the column is then calculated as:
#
#	Prob ({column} | {tree} ) = Sum_{$i} $pi[$i] * Prob ({column} | {tree}, $chars[$i] at $topnode_id),
#
#	where $pi[$i] is the frequency of the character $char[$i] at $topnode_id,
#	which is usually approximated with the equilibrium frequency, 1/scalar (@chars) in this case.
#
#	Model: An extended Jukes-Cantor model is defined with
#	the transition probability from the character $c1 to $c2:
#	Prob ($c1, $c2 | $dist, $ct_char)
#		= exp (-$t) + 1/$c * (1 - exp (-$t))) if $c1 == $c2,
#		= 1/$c * (1 - exp (-$t)) if $c1 != $c2,
#	where $c is the number of mutually transferable characters,
#	and $t = $c/($c - 1) * $dist.
#
# INPUT: ($topnode_id, \%node2children, \%node2parent, \%name2id, \@seqnames, \@msa, $minct_chars, \%br2attr, $indx_len), where
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2parent = ($node_id => $parent_id, ...) for the non-top nodes;
#	%name2id = ($seqname => $node_id, ...) for the exterior nodes and the top node;
#	$seqnames[$k] is the name of the $k th sequence in the input MSA;
#	${$msa[$k]} is the $k th sequence embedded in the input MSA;
#	$minct_chars is the assumed number of possible characters.
#	But the actual number of characters in a column can exceed this value;
#	%br2attr = ($node_id => $attr, ...) for nodes immediately downstream of the branches, with
#	$attr->[$indx_len], or $attr if $indx_len is undefined, giving the branch length.
#
# OUTPUT: ($msa_ln_likeli, \@clm_ln_likeli,\@set_chars), where
#	$msa_ln_likeli is the log-likelihood of the entire input MSA (undef if the total likelihood is zefo);
#	$clm_ln_likeli[$l] is the log-likelihood for the $l th column;
#	@{$set_chars[$l]} is the list of characters examined in the $l th column.
#
sub msa_likelihood_extJC_f ($\%\%\%\@\@$\%$) {

	my ($topnode_id, $node2ch, $node2pa, $name2id, $seqnames, $msa,
	   $minct_chars, $br2attr, $indx_len) = @_;

	my $ct_seqs = @{$seqnames};
	unless ($ct_seqs>0) {
	    print STDERR "[ERROR 'msa_likelihood_extJC_f'] NO SEQUENCE in the input MSA: $! !\n";
	    return (undef, undef, undef);
	}
	unless ($ct_seqs == @{$msa}) {
	    print STDERR "[ERROR 'msa_likelihood_extJC_f'] The number of sequences (", scalar (@{$msa}), ") in the MSA CLASHES with the number of sequence names ($ct_seqs): $! !\n";
	    return (undef, undef, undef);
	}
	my $ct_clms = length ${$msa->[0]};
	($ct_clms>0) or (return (0, [], []));

=for TESTING the subroutine
	print STDERR "\n< Subroutine 'msa_likelihood_extJC_f' >\n\n";
	print STDERR "Minimum number of characters: $minct_chars\n\n";
=cut

	# @{$set_chars->[$l]} is the list of characters examined in the $l th column;
	# %{$set_node2charlikeli->[$l]} = ($node_id => \@charlikeli, ...)
	# gives mapping of the character-wise log-likelihood values the $l th site mapped onto the nodes,
	# with $charlikeli[$c] is the log-likelihood of the character $set_chars[$l]->[$c];
	#
	my ($set_chars, $set_node2chlk) = map_likelihoods_extJC_msa_f ($topnode_id, %{$node2ch}, %{$node2pa}, %{$name2id}, @{$seqnames}, @{$msa}, $minct_chars, %{$br2attr}, $indx_len, 0);
	unless (defined $set_node2chlk) {
	    print STDERR "[ERROR 'msa_likelihood_extJC_f'] The subroutine 'map_likelihoods_extJC_msa_f' FAILED: $! !\n";
	    return (undef, undef, undef);
	}
	unless ((@{$set_chars} == $ct_clms) and (@{$set_node2chlk} == $ct_clms)) {
	    print STDERR "[ERROR 'msa_likelihood_extJC_f'] Something is WRONG! The output list of characters for the columns (", scalar (@{$set_chars}), ") or that of character-wise likelihoods (", scalar (@{$set_node2chlk}), ") has (a) CLASHING SIZE(s) with the number of the MSA columns ($ct_clms): $! !\n";
	    return (undef, undef, undef);
	}

	my $msa_ln_lk = 0;
	my @clm_ln_lk = ();
	for (my $l=0; $l<$ct_clms; $l++) {  # The main for-loop #

	    my $charlist = $set_chars->[$l];
	    my $ct_chars = @{$charlist};
	    my $node2chlk = $set_node2chlk->[$l];
	    my $clm_chlks = $node2chlk->{$topnode_id};
	    unless (defined $clm_chlks) {
		print STDERR "[ERROR 'msa_likelihood_extJC_f'] The character-wise log-likelihoods are NOT AVAILABLE for the $l th column: $! !\n";
		return (undef, undef, undef);
	    }

	    my $pi = 1/$ct_chars; # Equilibrium character frequency in the extended JC model.

	    my @terms_clm_ln_lk = (); # element = ln ({character_wise_likelihood} * {freqeuncy of the character at $topnode_id})
	    foreach my $ln_chlk (@{$clm_chlks}) { # Inner foreach-loop #
		(defined $ln_chlk) or next;
		push @terms_clm_ln_lk, $ln_chlk + log ($pi);
	    }			# End of the inner foreach-loop #
	    unless (@terms_clm_ln_lk) {
		push @clm_ln_lk, undef;
		$msa_ln_lk = undef;
		next;
	    }

	    my $clm_ln_lk = log_sum_exp (@terms_clm_ln_lk);
	    push @clm_ln_lk, $clm_ln_lk;
	    (defined $msa_ln_lk) and ($msa_ln_lk += $clm_ln_lk);

	}			# End of the main for-loop #

=for TESTING the subroutne
	print STDERR "\n\n --- The subroutine 'msa_likelihood_extJC_f' ENDED !! ---\n\n";
=cut

	return ($msa_ln_lk, \@clm_ln_lk, $set_chars);

}	# End of the subroutine 'msa_likelihood_extJC_f' #


# nonred_msa_clm_patterns_extJC:
#	This subroutine makes a set of non-redundant column-patterns
#	out of an input MSA, and outputs the set in the form of MSA.
#
#	NOTE that this subroutine eats up the input MSA.
#
#	(CREATED on Sep 17 (Fri), 2010)
#
# INPUT: \@msa_org, where
#	${$msa_org[$k]} is the $k th sequence embedded in the original input MSA.
#
# OUTPUT: (\@nr_msa, \@set_nr_indx, \@set_chars, \@set_char2char), where
#	${$nr_msa[$k]} gives the concatenation of the characters of the $k th sequence in the non-redundant column-patterns;
#	$set_nr_indx[$l] specifies the index (or the column-number in @{$nr_msa}) of the non-redundant column corresponding to the $l th column of the INPUT MSA;
#	@{$set_chars[$l]} is the set of characters in the $l th column of the INPUT MSA;
#	%{$set_char2char->[$l]} = ($char_orig_msa => $char_nr_msa, ...)
#	gives the association from the character in the $l th column of the original  MSA to the character in the corresponding column of the non-redundant MSA.
#
sub nonred_msa_clm_patterns_extJC (\@) {

	my ($msa_org) = @_;

	my $ct_seqs = @{$msa_org};
	unless ($ct_seqs>0) {
	    print STDERR "[ERROR 'nonred_msa_clm_patterns_extJC'] NO SEQUENCE in the input MSA: $! !\n";
	    return (undef, undef, undef, undef);
	}
	my $ct_clms = length ${$msa_org->[0]};
	($ct_clms>0) or (return ([], [], [], []));

		# Auxiliary array/hash #
	my @alpha = qw( a b c d e f g h i j k l m n o p q r s t u v w x y z );
	my %nrpatt2indx; # (nonredundant pattern => $indx to specify the column position in @nr_msa, ...) 

		# Initialize the outputs #
	my @nr_msa = ();
	for (my $k=0; $k<$ct_seqs; $k++) { my $nr_seq = ''; push @nr_msa, \$nr_seq; }
	my @set_nr_indx = my @set_chars = my @set_char2char = ();

	my $ct_nrpatterns = 0;
	for (my $l=0; $l<$ct_clms; $l++) { # Main for-loop (over the MSA columns) #

	    my $indx_alpha = 0;
	    my %char2alpha; # = (character in the MSA column => alphabet to substitute, ...)
	    my @char_in_clm = ();
	    my @nrpattern = (); 
	    for (my $k=0; $k<$ct_seqs; $k++) { # 1st inner for-loop (over the sequences) #
		my $char = substr (${$msa_org->[$k]}, 0, 1, '');
		push @char_in_clm, $char;

		my $alpha = $char2alpha{$char};
		unless (defined $alpha) {
		    $alpha = $alpha[$indx_alpha];
		    $char2alpha{$char} = $alpha;
		    $indx_alpha++;
		}
		push @nrpattern, $alpha;
	    }			# End of the 1st inner for-loop (over the sequences) #

	    my $charlist = nr_array (@char_in_clm);
	    push @set_chars, $charlist;
	    push @set_char2char, \%char2alpha;

		# Check if the non-redundant pattern already exists or not #
	    my $nrpatt = join ('', @nrpattern);
	    my $indx_nrpatt = $nrpatt2indx{$nrpatt};
	    unless (defined $indx_nrpatt) {
			# The non-redundant column-pattern is new #
		$indx_nrpatt = $ct_nrpatterns;
		$nrpatt2indx{$nrpatt} = $indx_nrpatt;
		$ct_nrpatterns++;
		for (my $k=0; $k<$ct_seqs; $k++) { # 2nd inner for-loop (over the sequences) #
		    ${$nr_msa[$k]} .= $nrpattern[$k];
		}	# End of the 2nd inner for-loop (over the sequences) #
	    }

	    push @set_nr_indx, $indx_nrpatt;

	}		# End of the main for-loop (over the MSA columns) #

	return (\@nr_msa, \@set_nr_indx, \@set_chars, \@set_char2char);

}	# End of the subroutine 'nonred_msa_clm_patterns_extJC' #


# map_likelihoods_extJC_msa_fnr:
#	This subroutine is a revised version of 'map_likelihoods_extJC_msa_nr'.
#	The only difference is that this uses the subroutine 'map_likelihoods_extJC_msa_f', which is a FAST version of the subroutine 'map_likelihoods_extJC_msa'.
#	(CREATED on September 17 (Fri), 2010)
#	
# map_likelihoods_extJC_msa_nr:
#	This subroutine is a kind wrapper of the subroutine 'map_likelihoods_extJC_msa'.
#	Basically, this wrapper compiles the input MSA
#	into the set of columns with non-redundant patterns,
#	calculates column-wise likelihoods of non-redundant patterns via 'map_likelihoods_extJC_msa',
#	then, based on that,
#	outputs the column-wise likelihoods of the original input MSA.
#
#	NOTE that this subroutine eats up the input MSA.
#	PREREQUISITE: The input MSA must NOT contain gaps.
#
#	(CREATED on September 17 (Fri), 2010)
#
# map_likelihoods_extJC_msa:
#	This subroutine performs Felsenstein's peeling algorithm
#	to estimate the log-likelihood values of columns in the input  MSA,
#	given a tree topology and branch lengths,
#	by using an extended Jukes-Cantor model of substitutions.
#
#	(See the header of 'msa_likelihoods_extJC_msa' for details...)
#
# INPUT: ($topnode_id, \%node2children, \%node2parent, \%name2id, \@seqnames, \@msa, $minct_chars, \%br2attr, $indx_len, $fl_full_mode), where
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2parent = ($node_id => $parent_id, ...) for the non-top nodes;
#	%name2id = ($seqname => $node_id, ...) for the exterior nodes and the top node;
#	$seqnames[$k] is the name of the $k th sequence in the input MSA;
#	${$msa[$k]} is the $k th sequence embedded in the input MSA;
#	$minct_chars is the assumed number of possible characters.
#	But the actual number of characters in a column can exceed this value;
#	%br2attr = ($node_id => $attr, ...) for nodes immediately downstream of the branches, with
#	$attr->[$indx_len], or $attr if $indx_len is undefined, giving the branch length;
#	$fl_full_mode = 1 if you want the full mode,
#			= 0 if the brief mode is enough.
#
# OUTPUT: (\@set_chars, \@set_node2charlikeli), where
#	@{$set_chars[$l]} is the list of characters examined in the $l th column;
#	%{$set_node2charlikeli[$l]} = ($node_id => \@charlikeli, ...)
#	gives mapping of the character-wise log-likelihood values the $l th site mapped onto the nodes,
#	with $charlikeli[$c] is the log-likelihood of the character $set_chars[$l]->[$c];
#	If $fl_full_mode = 1, the mapping onto ALL the nodes is given.
#	If $fl_full_mode = 0, ONLY the likelihoods on the TOP node is given.
#
sub map_likelihoods_extJC_msa_fnr ($\%\%\%\@\@$\%$$) {

	my ($topnode_id, $node2ch, $node2pa, $name2id, $seqnames, $msa,
	   $minct_chars, $br2attr, $indx_len, $fl_full_mode) = @_;

	my $ct_seqs = @{$seqnames};
	unless ($ct_seqs>0) {
	    print STDERR "[ERROR 'map_likelihoods_extJC_msa_fnr'] NO SEQUENCE in the input MSA: $! !\n";
	    return (undef, undef);
	}
	unless ($ct_seqs == @{$msa}) {
	    print STDERR "[ERROR 'map_likelihoods_extJC_msa_fnr'] The number of sequences (", scalar (@{$msa}), ") in the MSA CLASHES with the number of sequence names ($ct_seqs): $! !\n";
	    return (undef, undef);
	}
	my $ct_clms = length ${$msa->[0]};
	($ct_clms>0) or (return ([], []));

=for TESTING the subroutine
	print STDERR "\n< Subroutine 'map_likelihoods_extJC_msa_fnr' >\n\n";
	print STDERR "Minimum number of characters: $minct_chars\n\n";
	print STDERR "\$fl_full_mode = $fl_full_mode\n\n";
=cut

		## Make a set of non-redundant column pattens
		## out of the input MSA.
	# ${$nr_msa->[$k]} gives the concatenation of the characters of the $k th sequence in the non-redundant column patterns;
	# $set_nr_indx->[$l] specifies the index (or the column-number in @{$nr_msa}) of the non-redundant column corresponding to the $l th column of the INPUT MSA;
	# @{$set_chars->[$l]} is the set of characters in the $l th column of the INPUT MSA;
	# %{$set_char2char->[$l]} = ($char_orig_msa => $char_nr_msa, ...)
	# gives the association from the character in the $l th column of the original  MSA to the character in the corresponding column of the non-redundant MSA.
	my ($nr_msa, $set_nr_indx, $set_chars, $set_char2char)
		= nonred_msa_clm_patterns_extJC (@{$msa});

	unless (defined $nr_msa) {
	    print STDERR "[ERROR 'msa_likelihoods_extJC_msa_fnr'] The subroutine 'nonred_msa_clm_patterns' FAILED: $! !\n";
	    return (undef, undef);

	}
	unless ((@{$set_nr_indx} == $ct_clms) and (@{$set_chars} == $ct_clms) and (@{$set_char2char} == $ct_clms)) {
	    print STDERR "[ERROR 'msa_likelihoods_extJC_msa_fnr'] Something is WRONG ! The output list of nr-column indices (", scalar (@{$set_nr_indx}), "), that of characters for the input MSA columns (", scalar (@{$set_chars}), "), or that of character-to-character correspondence (", scalar (@{$set_char2char}), ") has (a) CLASHING SIZE(s) with the  number of the columns in the ORIGINAL MSA ($ct_clms): $! !\n";
	    return (undef, undef);
	}

	my $ct_clms_nr = length ${$nr_msa->[0]};

		## Calculate likelihoods of non-redundant column-patterns ##

	# @{$set_chars_nr->[$l]} is the list of characters examined in the $l th non-redundant column-pattern;
	# %{$set_node2chlk_nr->[$l]} = ($node_id => \@charlikeli, ...)
	# gives mapping of the character-wise log-likelihood values of the $l th non-redundant column-pattern mapped onto the nodes,
	# with $charlikeli[$c] is the log-likelihood of the character $set_chars_nr[$l]->[$c];
	#
	my ($set_chars_nr, $set_node2chlk_nr) = map_likelihoods_extJC_msa_f ($topnode_id, %{$node2ch}, %{$node2pa}, %{$name2id}, @{$seqnames}, @{$nr_msa}, $minct_chars, %{$br2attr}, $indx_len, $fl_full_mode);
	unless (defined $set_node2chlk_nr) {
	    print STDERR "[ERROR 'msa_likelihoods_extJC_msa_fnr'] The subroutine 'map_likelihoods_extJC_msa_f' FAILED: $! !\n";
	    return (undef, undef);
	}
	unless ((@{$set_chars_nr} == $ct_clms_nr) and (@{$set_node2chlk_nr} == $ct_clms_nr)) {
	    print STDERR "[ERROR 'msa_likelihood_extJC_msa_fnr'] Something is WRONG! The output list of characters for the columns (", scalar (@{$set_chars_nr}), ") or that of character-wise likelihoods (", scalar (@{$set_node2chlk_nr}), ") has (a) CLASHING SIZE(s) with the number of the NON-REDUNDANT column-patterns ($ct_clms_nr): $! !\n";
	    return (undef, undef);
	}

		# Preparation for the final process #

	my @set_char_nr2indx = (); # %{$set_char_nr2indx[$l]} = ($character => $indx, ...)  for the $l-th non-redundant column-pattern.
	for (my $l=0; $l<$ct_clms_nr; $l++) { # The auxiliary for-loop #
	   my $chars_nr = $set_chars_nr->[$l];
	   my $ct = @{$chars_nr};
	   my %char_nr2indx;
	   for (my $i=0; $i<$ct; $i++) { # Inner for-loop #
		my $char_nr = $chars_nr->[$i];
		$char_nr2indx{$char_nr} = $i;
	   }			# End of the inner for-loop #

	   push @set_char_nr2indx, \%char_nr2indx;
	}			# End of the auxiliary for-loop #

		## Convert the likelihoods of non-redundant patterns
		## into the column-wise likelihoods of the original input MSA.

	my @set_node2charlikeli = ();
	# %{$set_node2charlikeli[$l]} = ($node_id => \@charlikeli, ...)
	# gives mapping of the character-wise log-likelihood values the $l th site mapped onto the nodes,
	# with $charlikeli[$c] is the log-likelihood of the character $set_chars[$l]->[$c];

	for (my $l=0; $l<$ct_clms; $l++) { # The final main for-loop (over the columns in the original MSA) #

	    my $cnum_nr = $set_nr_indx->[$l]; # ID number (or index) of the non-redundant column-pattern
	    my $chars_org = $set_chars->[$l]; # Characters in the original column
	    my $ct_chars = @{$chars_org};

		 ## ADDED in DEBUGGING the subroutine 'map_likelihoods_extJC_msa_fnr' (on Sep 17 (Fri), 2010) ##
	    if ($ct_chars < $minct_chars) {
		# Pad the character list with numbers
		# (, which will never match the actual characters)
		for (my $n = 0; $n<$minct_chars - $ct_chars; $n++) { push @{$chars_org}, $n; }
		$ct_chars = $minct_chars;
	    }

	    my $chars_nr = $set_chars_nr->[$cnum_nr]; # Characters in the corresponding non-redundant column-pattern
	    unless (@{$chars_nr} == $ct_chars) {
		print STDERR "[ERROR 'map_likelihoods_extJC_msa_fnr'] The number of characters ($ct_chars) in the $l th original column CLASHES with the number of characters (", scalar (@{$chars_nr}), ") in the corresponding $cnum_nr th non-redundant column-pattern: $! !\n";
		print STDERR "Original     : ", join (' ', @{$chars_org}), " ,\n";
		print STDERR "Non-redundant: ", join (' ', @{$chars_nr}),  " .\n\n";
		return (undef, undef);
	    }
	    my $char2char = $set_char2char->[$l]; # = {$char_org_msa => $char_nr_msa, ...}
	    my $char_nr2indx = $set_char_nr2indx[$cnum_nr];
	    my @indices_nr; # = $indices_nr[$index in @{$chars_org}]  = $index in @{$chars_nr}, ..)
	    for (my $indx_org=0; $indx_org<$ct_chars; $indx_org++) { # 1st inner for-loop (over characters in the original column) #
		my $char_org = $chars_org->[$indx_org];
		my $char_nr = $char2char->{$char_org};

		 ## ADDED in DEBUGGING the subroutine 'map_likelihoods_extJC_msa_fnr' (on Sep 17 (Fri), 2010) ##
		unless (defined $char_nr) {
		    if ($char_org =~ m/^\d+$/) {
			$char_nr = $char_org;
		    } else {
			print STDERR "[ERROR 'map_likelihoods_extJC_msa_fnr'] NO non-redundant character corresponding to the $indx_org th character '$char_org', which is NON-NUMERICAL, in the $l th column: $! !\n";
			return (undef, undef);
		    }
		}
		my $indx_nr = $char_nr2indx->{$char_nr};
		unless (defined $indx_nr) {
		    print STDERR "[ERROR 'map_likelihoods_extJC_msa_fnr'] NO INDEX of the character '$char_nr' corresponding to the $indx_org th character '$char_org' in the $l th column: $! !\n";
		    return (undef, undef);
		}
		$indices_nr[$indx_org] = $indx_nr;
	    }				# End of the 1st inner for-loop (over characters in the original column) #

	    my %node2charlikeli;
	    my $node2chlk_nr = $set_node2chlk_nr->[$cnum_nr]; # = {$node_id => \@chlk_nr, ...)
	    foreach my $node (keys %{$node2chlk_nr}) { # 2nd inner for-loop (over nodes) #
		my $set_chlk_nr = $node2chlk_nr->{$node}; # $set_chlk_nr->[$c] is the likelihood for the character $chars_nr->[$c].
		my @set_chlk_org = ();
		for (my $indx_org =0; $indx_org < $ct_chars; $indx_org++) { # Innermost for-loop (over characters in the original column) #
		    my $indx_nr = $indices_nr[$indx_org];
		    my $chlk = $set_chlk_nr->[$indx_nr];
		    $set_chlk_org[$indx_org] = $chlk;
		}			# End of the innermost for-loop (over characters in the original column) #

		$node2charlikeli{$node} = \@set_chlk_org;

	    }				# End of the 2nd inner for-loop (over nodes) #

	    push @set_node2charlikeli, \%node2charlikeli;

	}			# End of the final main for-loop (over the columns in the original MSA) #

		## Final output ##

	return ($set_chars, \@set_node2charlikeli);

}	# End of the subroutine 'map_likelihoods_extJC_msa_fnr' #


# msa_likelihood_extJC_fnr:
#	This is a FAST version of 'msa_likelihood_extJC_nr'.
#	The only change is that this uses 'map_likelihoods_extJC_msa_fnr',
#	which is a FAST version of 'map_likelihoods_extJC_msa_nr'.
#	(CREATED on September 17 (Fri), 2010)
#
# msa_likelihood_extJC_nr:
#	This revised subroutine of 'msa_likelihood_extJC' conducts
#	a non-redundant calculation of column-wise likelihoods
#	by using the subroutine 'map_likelihoods_extJC_msa_nr',
#	a non-redundant version of 'map_likelihoods_extJC_msa'.
#	The input and output are (must be!!) identical to the
#	ancestor 'msa_likelihood_extJC'.
#
#	NOTE that this subroutine eats up the input MSA.
#	PREREQUISITE: The input MSA must NOT contain gaps.
#
#	(CREATED on September 17 (Fri), 2010)
#
# msa_likelihood_extJC:
#	This subroutine performs Felsenstein's peeling algorithm
#	to estimate the log-likelihood values of columns in the input MSA,
#	given a tree topology and branch lengths,
#	by using an extended Jukes-Cantor model of substitutions
#	via the subroutine 'map_likelihoods_extJC_msa'.
#	Then it calculates the log-likelihoood of the entire MSA
#	by summing the log-likelihoods over columns.
#
#	NOTE that this subroutine eats up the input MSA.
#	PREREQUISITE: The input MSA must NOT contain gaps.
#
#	ALGORITHM: For each column, character-wise likelihoods:
#	Prob ({column} | {tree}, $chars[$i] at $topnode_id) 
#	are calculated by the subroutine 'map_likelihoods_extJC_msa'.
#	The total likelihood for the column is then calculated as:
#
#	Prob ({column} | {tree} ) = Sum_{$i} $pi[$i] * Prob ({column} | {tree}, $chars[$i] at $topnode_id),
#
#	where $pi[$i] is the frequency of the character $char[$i] at $topnode_id,
#	which is usually approximated with the equilibrium frequency, 1/scalar (@chars) in this case.
#
#	Model: An extended Jukes-Cantor model is defined with
#	the transition probability from the character $c1 to $c2:
#	Prob ($c1, $c2 | $dist, $ct_char)
#		= exp (-$t) + 1/$c * (1 - exp (-$t))) if $c1 == $c2,
#		= 1/$c * (1 - exp (-$t)) if $c1 != $c2,
#	where $c is the number of mutually transferable characters,
#	and $t = $c/($c - 1) * $dist.
#
# INPUT: ($topnode_id, \%node2children, \%node2parent, \%name2id, \@seqnames, \@msa, $minct_chars, \%br2attr, $indx_len), where
#	%node2children = ($node_id => \@children, ...) for the interior nodes;
#	%node2parent = ($node_id => $parent_id, ...) for the non-top nodes;
#	%name2id = ($seqname => $node_id, ...) for the exterior nodes and the top node;
#	$seqnames[$k] is the name of the $k th sequence in the input MSA;
#	${$msa[$k]} is the $k th sequence embedded in the input MSA;
#	$minct_chars is the assumed number of possible characters.
#	But the actual number of characters in a column can exceed this value;
#	%br2attr = ($node_id => $attr, ...) for nodes immediately downstream of the branches, with
#	$attr->[$indx_len], or $attr if $indx_len is undefined, giving the branch length.
#
# OUTPUT: ($msa_ln_likeli, \@clm_ln_likeli,\@set_chars), where
#	$msa_ln_likeli is the log-likelihood of the entire input MSA (undef if the total likelihood is zefo);
#	$clm_ln_likeli[$l] is the log-likelihood for the $l th column;
#	@{$set_chars[$l]} is the list of characters examined in the $l th column.
#
sub msa_likelihood_extJC_fnr ($\%\%\%\@\@$\%$) {

	my ($topnode_id, $node2ch, $node2pa, $name2id, $seqnames, $msa,
	   $minct_chars, $br2attr, $indx_len) = @_;

	my $ct_seqs = @{$seqnames};
	unless ($ct_seqs>0) {
	    print STDERR "[ERROR 'msa_likelihood_extJC_fnr'] NO SEQUENCE in the input MSA: $! !\n";
	    return (undef, undef, undef);
	}
	unless ($ct_seqs == @{$msa}) {
	    print STDERR "[ERROR 'msa_likelihood_extJC_fnr'] The number of sequences (", scalar (@{$msa}), ") in the MSA CLASHES with the number of sequence names ($ct_seqs): $! !\n";
	    return (undef, undef, undef);
	}
	my $ct_clms = length ${$msa->[0]};
	($ct_clms>0) or (return (0, [], []));

=for TESTING the subroutine
	print STDERR "\n< Subroutine 'msa_likelihood_extJC' >\n\n";
	print STDERR "Minimum number of characters: $minct_chars\n\n";
=cut

	# @{$set_chars->[$l]} is the list of characters examined in the $l th column;
	# %{$set_node2charlikeli->[$l]} = ($node_id => \@charlikeli, ...)
	# gives mapping of the character-wise log-likelihood values the $l th site mapped onto the nodes,
	# with $charlikeli[$c] is the log-likelihood of the character $set_chars[$l]->[$c];
	#
	my ($set_chars, $set_node2chlk) = map_likelihoods_extJC_msa_fnr ($topnode_id, %{$node2ch}, %{$node2pa}, %{$name2id}, @{$seqnames}, @{$msa}, $minct_chars, %{$br2attr}, $indx_len, 0);
	unless (defined $set_node2chlk) {
	    print STDERR "[ERROR 'msa_likelihood_extJC_fnr'] The subroutine 'map_likelihoods_extJC_msa_fnr' FAILED: $! !\n";
	    return (undef, undef, undef);
	}
	unless ((@{$set_chars} == $ct_clms) and (@{$set_node2chlk} == $ct_clms)) {
	    print STDERR "[ERROR 'msa_likelihood_extJC_fnr'] Something is WRONG! The output list of characters for the columns (", scalar (@{$set_chars}), ") or that of character-wise likelihoods (", scalar (@{$set_node2chlk}), ") has (a) CLASHING SIZE(s) with the number of the MSA columns ($ct_clms): $! !\n";
	    return (undef, undef, undef);
	}

	my $msa_ln_lk = 0;
	my @clm_ln_lk = ();
	for (my $l=0; $l<$ct_clms; $l++) {  # The main for-loop #

	    my $charlist = $set_chars->[$l];
	    my $ct_chars = @{$charlist};
	    my $node2chlk = $set_node2chlk->[$l];
	    my $clm_chlks = $node2chlk->{$topnode_id};
	    unless (defined $clm_chlks) {
		print STDERR "[ERROR 'msa_likelihood_extJC_fnr'] The character-wise log-likelihoods are NOT AVAILABLE for the $l th column: $! !\n";
		return (undef, undef, undef);
	    }

	    my $pi = 1/$ct_chars; # Equilibrium character frequency in the extended JC model.

	    my @terms_clm_ln_lk = (); # element = ln ({character_wise_likelihood} * {freqeuncy of the character at $topnode_id})
	    foreach my $ln_chlk (@{$clm_chlks}) { # Inner foreach-loop #
		(defined $ln_chlk) or next;
		push @terms_clm_ln_lk, $ln_chlk + log ($pi);
	    }			# End of the inner foreach-loop #
	    unless (@terms_clm_ln_lk) {
		push @clm_ln_lk, undef;
		$msa_ln_lk = undef;
		next;
	    }

	    my $clm_ln_lk = log_sum_exp (@terms_clm_ln_lk);
	    push @clm_ln_lk, $clm_ln_lk;
	    (defined $msa_ln_lk) and ($msa_ln_lk += $clm_ln_lk);

	}			# End of the main for-loop #

=for TESTING the subroutne
	print STDERR "\n\n --- The subroutine 'msa_likelihood_extJC_fnr' ENDED !! ---\n\n";
=cut

	return ($msa_ln_lk, \@clm_ln_lk, $set_chars);

}	# End of the subroutine 'msa_likelihood_extJC_fnr' #


1;	# a "spell" which makes the module "use"-able and/or "require"-able #


